# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================



import itertools 
import templates
import os

def main(GENS=2, prop_simple=False):
    """Returns (slim_fname, props_fname, err_fname, fi_fname)"""

    assert GENS > 1
    GENTYPE = "Generator.random"

    # gens and bits
    gens = []
    bit_names = []
    bits = []
    gens_bits = []
    gens_tick = []
    bits_tick = []
    sources = [] # list of sources to be processed by adders in pairs
    for i in range(GENS):
        gens.append("gen%03d: system %s accesses abus;" % (i, GENTYPE))
        bit = "bit%03d" % i
        bit_names.append(bit)
        bits.append("%s: system Bit accesses abus;" % bit)
        gens_bits.append("data port gen%03d.output -> bit%03d.input;" % (i, i))
        gens_tick.append("event port trig.gen_tick -> gen%03d.tick;" % i)
        bits_tick.append("event port trig.bit_tick -> bit%03d.tick;" % i)
        sources.append("bit%03d" % i)
        pass
    
    # adders connected to sources
    level = 0
    adder_names = []
    adders_level_0 = []
    adders = []
    sources_adders = []
    adds_tick = []
    while len(sources) > 1:
        for i in range(len(sources) / 2):            
            s1 = sources.pop(0)
            s2 = sources.pop(0)
            adder = "adder_%d_%03d" % (level, i)
            adder_names.append(adder)
            if level == 0: adders_level_0.append(adder)

            adders.append("%s: system FTAdder accesses abus;" % adder)
            sources_adders.append("data port %s.output -> %s.input1;" % (s1, adder))
            sources_adders.append("data port %s.output -> %s.input2;" % (s2, adder))
            adds_tick.append("event port trig.add_tick -> %s.tick;" % adder)

            sources.append(adder)
            pass
        level += 1        
        pass

    assert len(sources) == 1
    final_output = "%s.output" % sources[0]

    # properties
    properties = []
    all_bits_equal = " and ".join(("gen%03d.output = gen%03d.output" % (i, i+1) for i in range(GENS-1)))
    some_bits_diff = " or ".join(("gen%03d.output != gen%03d.output" % (i, i+1) for i in range(GENS-1)))

    properties.append(("(trig.mode = mode:on_gen and (%s))" % all_bits_equal, "(%s = 0)" % final_output))
    properties.append(("(trig.mode = mode:on_gen and (%s))" % all_bits_equal, "(%s = 1)" % final_output))
    properties.append(("(trig.mode = mode:on_gen and (%s))" % some_bits_diff, "(%s = 0)" % final_output))
    properties.append(("(trig.mode = mode:on_gen and (%s))" % some_bits_diff, "(%s = 1)" % final_output))

    if prop_simple:
        properties = properties[:1]

    # emit the slim file
    sfname = "Adder-%02d.slim" % GENS
    slimf = open(sfname, "w")
    seps = os.linesep + " " * 4
    slimf.write(templates.slim.substitute(gens = seps.join(gens), 
                                          bits = seps.join(bits), 
                                          adders = seps.join(adders), 
                                          gens_bits = seps.join(gens_bits),
                                          sources_adders = seps.join(sources_adders), 
                                          gens_tick = seps.join(gens_tick), 
                                          bits_tick = seps.join(bits_tick), 
                                          adds_tick = seps.join(adds_tick), 
                                          properties = os.linesep.join(("-- AG ((%s) -> AF (%s))" % (p,r) 
                                                                        for p,r in properties)))
                )
    slimf.close()
    
    # emit the properties file
    pfname = "Adder-%02d.propxml" % GENS
    propsf = open(pfname, "w")
    
    ctl_props = (templates.ctl.substitute(name="Prop CTL %02d" % i,
                                          precond=prop[0], 
                                          response=prop[1])                 
                 for i, prop in itertools.izip(itertools.count(), properties))

    ltl_props = (templates.ltl.substitute(name="Prop LTL %02d" % i,
                                          precond=prop[0], 
                                          response=prop[1])                 
                 for i, prop in itertools.izip(itertools.count(), properties))

    # propositional
    propos = (templates.propos.substitute(name="Propositional 00",
                                          formula="(trig.mode = mode:on_gen and (%s) and (%s))" % \
                                              (all_bits_equal, " or ".join(("%s.output = 1" % adder 
                                                                            for adder in adders_level_0)))), 
              )

    propsf.write(templates.props.substitute(properties=os.linesep.join(itertools.chain(propos,ctl_props,ltl_props,))))
    propsf.close()        

    # emits the error file
    efname = "Adder-%02d-Err.slim" % GENS
    errf = open(efname, "w")
    errf.write(templates.error.substitute({}))
    errf.close()

    # emits the fault injection file
    fifname = "Adder-%02d.fixml" % GENS
    fif = open(fifname, "w")
    
    bit_assocs = (templates.slim_fi_bit_assoc.substitute(name=name)
                  for name in bit_names)    
    adder_assocs = (templates.slim_fi_adder_assoc.substitute(name=name)
                    for name in adder_names)

    fif.write(templates.slim_fi.substitute(assocs=os.linesep.join(
                (assoc for assoc in itertools.chain(bit_assocs, adder_assocs)))
                                           ))
    fif.close()
    
    return (sfname, pfname, efname, fifname)
    

# launcher and options handler
if __name__ == "__main__": 
    import sys
    try: bits = int(sys.argv[1])
    except: 
        sys.stderr.write("""Usage: adder-gen.py <num> 
Where <num> is the number of generators.
""")
        sys.exit(1)
        pass

    if bits < 2:
        sys.stderr.write("""The number of bits has to be greater then 1.
""")
        sys.exit(1)
        pass

    main(bits)

