# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================

import booter

from invokers import slim, dump, nusmv
from utils import debug_info as debug, log
from utils import procmon
import slim_compiler.src.model_extension.core as model_ext_core
import slim_compiler.src.smv.api as smv_api
import property_management.src.api as properties_api
import property_management.src.constants as prop_const
from utils.nusmv_process import NuSMVProcess
from utils.fsap_process import FsapProcess

from datetime import datetime
import os
import os.path
import subprocess
import sys
import itertools

import options

def generate_slim(size):
    """Returns the name of the generated files (see gen.main)"""
    import gen
    return gen.main(size)


def compile_slim(slim_fnames, fi_fname, timings):
    """Returns the nominal root and the extended root
    implementations. If fi_name is None, no extension will be
    done"""
    debug_info = debug.create_debug_and_compile_information() 
    compile_info = debug_info.compile_info

    # parsing
    _start_timer("parse", timings)
    parsing_result = slim.parse(compile_info, slim_fnames)
    if not parsing_result: slim.emit_errors(compile_info); quit(-1)
    _stop_timer("parse", timings)
    # root
    root_impl = slim.get_root_impl(compile_info, options.ROOT_NAME)

    # checking
    _start_timer("typecheck", timings)
    checking_result = slim.check(compile_info)
    if not checking_result: slim.emit_errors(compile_info); quit(-1)
    _stop_timer("typecheck", timings)

    # extension
    ext_root_impl = None
    _start_timer("extension", timings)
    if fi_fname is not None:
        try:
            ext_compile_info = slim.do_fault_injection(compile_info, fi_fname)
        except model_ext_core.InvalidFaultInjectionDocumentError, e:
            debug.emit_error_and_quit(exceptions.UserException(str(e)),
                                      debug_info)
            pass
        assert ext_compile_info != None
        ext_root_impl = slim.get_root_impl(ext_compile_info)
        pass    
    _stop_timer("extension", timings)

    return (root_impl, ext_root_impl)


def generate_smv(basename, root_impl, timings):
    """Returns (smv_fname, ord_fname, needs_msat)"""    
    return _generate_smv(basename, root_impl, timings)

def generate_ext_smv(basename, root_impl, timings):
    """Returns (smv_fname, ord_fname, needs_msat)"""    
    return _generate_smv(basename+"_ext", root_impl, timings, "_ext")


def _generate_smv(basename, root_impl, timings, timer_prefix=""):
    """Returns (smv_fname, ord_fname, needs_msat)"""
      
    # generate smv
    _start_timer("smv"+timer_prefix, timings)
    smv_fname, need_msat, timed = slim.convert_to_smv(root_impl, basename+".smv")
    _stop_timer("smv"+timer_prefix, timings)

    # generates BDD ordering    
    _start_timer("ord"+timer_prefix, timings)
    if not need_msat: 
        ord_fname = nusmv.launch_smv_to_bdd_ordering(smv_fname)
    else: ord_fname = None
    _stop_timer("ord"+timer_prefix, timings)

    return (smv_fname, ord_fname, need_msat)


def generate_observers_alarms(basename, root_impl, timings):
    _start_timer("obs_alm", timings)
    fail_obs_alm_fname = basename+ ".fail_obs_alm"    
    fo_file = open(fail_obs_alm_fname, "w")
    fails, num_obs, num_alm = smv_api.dump_failures_observations_alarms(root_impl, fo_file)
    fo_file.close()
    _stop_timer("obs_alm", timings)
    
    return fail_obs_alm_fname    


def generate_props(root_impl, props_fname, timings):
    _start_timer("props", timings)
    formulae_doc = slim.get_formulae_doc(root_impl, props_fname)
    _stop_timer("props", timings)

    smv_formulae_doc = slim.get_formulae_doc(root_impl, props_fname)
    _start_timer("props_smv", timings)
    properties_api.convert_atomic_propositions_to_smv(smv_formulae_doc, root_impl)
    _stop_timer("props_smv", timings)

    if formulae_doc is not None:
        res = [(formula, smv_formula) for formula, smv_formula in zip(formulae_doc, smv_formulae_doc)]
    else: res = []
    return res


def generate_smv_props(basename, props):
    """Generates a file containing all the SMV properties. Returns
    the file name"""
    smv_props_fn = "%s_smv.props" % basename
    smv_props_f = open(smv_props_fn, "w")

    for prop, smv_prop in props:
        smv_props_f.write(properties_api.get_formula_as_smv_string(smv_prop) + os.linesep * 2)
        pass
    
    smv_props_f.close()
    return smv_props_fn


def run_model_checking(smv_fname, ord_fname, need_msat, props, timings, use_bmc=False):
    """Returns a list of filenames of traces."""
    res = []
    _start_timer("mc", timings)
    nusmv_proc = NuSMVProcess(smv_fname, 
                              need_msat=need_msat,
                              stdout=subprocess.PIPE, 
                              stderr=subprocess.STDOUT)
    try:
        for formula, smv_formula in props:
            trace = None
            if use_bmc:
                if formula.get_logic() == prop_const.Logic.LTL:
                    hold,trace= nusmv.model_checking_bmc(nusmv_proc, smv_formula)                
                    pass
                pass
            else: # BDD
                if formula.get_logic() == prop_const.Logic.CTL:
                    hold,trace= nusmv.model_checking(nusmv_proc, smv_formula)
                    pass
                pass
            if trace: res.append(trace)
            pass

        _stop_timer("mc", timings)

    except KeyboardInterrupt: 
        nusmv_proc.kill()
        raise KeyboardInterrupt()

    nusmv_proc.kill()
    return res


def run_gen_ft(smv_fname, obs_alm_fname, need_msat, props, timings, use_bmc=False):
    """Returns a list of (tmp_ev_fname, tmp_gt_fname, tmp_xml_fname)"""
    res = []
    basename = os.path.splitext(smv_fname)[0]
    _start_timer("ftg", timings)
    fsap_proc = FsapProcess(smv_fname, obs_alm_fname, 
                            need_msat=need_msat,
                            stdout=subprocess.PIPE, 
                            stderr=subprocess.STDOUT)

    try:    
        fsap_proc.set_tmp_filename_prefix(basename + "__temp_fsap__")
    
        for formula, smv_formula in  \
                filter(lambda p: properties_api.check_is_propositional_formula(p[0]), props):
            smv_ = properties_api.get_formula_as_smv_string(smv_formula)
            tmp_ev_fname, tmp_gt_fname, tmp_xml_fname = fsap_proc.compute_fault_tree(smv_)
            res.append((tmp_ev_fname, tmp_gt_fname, tmp_xml_fname))
            pass

        _stop_timer("ftg", timings)

    except KeyboardInterrupt: 
        fsap_proc.kill()
        raise KeyboardInterrupt()    
        
    fsap_proc.kill()
    return res


def _start_timer(name, timings): timings[name].start()

def _stop_timer(name, timings): timings[name].stop()
    

# ----------------------------------------------------------------------
class Timings:
    _start = datetime.now()
    _stop = _start
    desc = ""
    _verbose=False
    _is_running = False

    def __init__(self, desc, verbose=False):
        self.desc = desc
        self._verbose = verbose
        return

    def start(self):
        if self._verbose:
            sys.stdout.write(" " + self.desc + "... ")
            sys.stdout.flush()
            pass

        self._start = datetime.now()
        assert not self._is_running
        self._is_running = True
        return

    def stop(self):
        self._stop = datetime.now()
        assert self._is_running
        self._is_running = False

        if self._verbose:
            sys.stdout.write(str(self.get_time()) + '\n')
            pass

        return

    def get_time(self):
        """If still running, a negative value is provided"""
        if self._is_running: pt = self._start - datetime.now()
        else: pt = self._stop - self._start 
        return pt.days * 3600*24 + pt.seconds + pt.microseconds / 1000000.0

    pass # end of class
# ----------------------------------------------------------------------

def_outfname="adder.timings"

# ----------------------------------------------------------------------
# Main function, which runs all tests and takes timings
def run(from_bits, size, interval, do_mc=True, do_ft=True, 
        outfname=def_outfname, 
        use_bmc=False, cleanup=True, verbose=False):

    timings = {'parse' : Timings("SLIM Parsing", verbose),
               'typecheck' : Timings("SLIM Type Checking", verbose),
               'extension' : Timings("Fault Extension", verbose),
               'smv' : Timings("SMV Translation", verbose),
               'ord' : Timings("BDD Ordering", verbose),
               'props' : Timings("SLIM Props Parsing", verbose),
               'props_smv' : Timings("SMV Props Translation", verbose),
               'mc' : Timings("Model Checking", verbose),
               'obs_alm'     : Timings("Observers and Alarms lists", verbose),
               'smv_ext' : Timings("SMV Translation (extended)", verbose),
               'ord_ext' : Timings("BDD Ordering (extended)", verbose),
               'ftg' : Timings("Fault Tree generation", verbose),
               }

    test_kind = "SLIM Compiler" if not do_mc else "BMC" if use_bmc else "BDD"
    testname="Test (%s) from %d to %d (interval:%d), output to '%s'" % (test_kind, from_bits, size, interval, outfname)
    print "Running", testname     

    timers = ('parse', 'typecheck', 'extension', 'smv', 'ord', 'props', 'props_smv', 'mc', 
              'obs_alm', 'smv_ext', 'ord_ext', 'ftg')
    with open(outfname, "w") as outfile:
        out_condition = "Successfully Ended"

        # prints the header
        outfile.write("""# ----------------------------------------------------------------------
# %s
# Started: %s
# ----------------------------------------------------------------------
#
""" % (testname, str(datetime.utcnow()) + " UTC"))
        
        outfile.write("# (1)size " + 
                      " ".join(
                ("(%d)%s" % (i,t) 
                 for i,t in itertools.izip(itertools.count(2), timers)
                 )) + os.linesep)

        for step in range(from_bits, size+1, interval):
            traces = []
            fts = []
            if verbose:
                print "----------------------------------------------------------------------"
                print "Size: %d/%d" % (step, size)
                pass
            try:
                slim, props_f, err, fi = generate_slim(step)
                root_impl, ext_root_impl = compile_slim((slim,err), fi if do_ft else None, 
                                                        timings)

                basename = os.path.splitext(os.path.basename(slim))[0]

                smv, ord, msat = generate_smv(basename, root_impl, timings)                
                props = generate_props(root_impl, props_f, timings)
            
                smv_props_f = generate_smv_props(basename, props)

                if do_mc: traces = run_model_checking(smv, ord, msat, props, timings, use_bmc)
                
                ext_smv = None
                if ext_root_impl is not None:
                    obs_alm = generate_observers_alarms(basename, ext_root_impl, timings)
                    ext_smv, ord, msat = generate_ext_smv(basename, ext_root_impl, timings)
                    fts = run_gen_ft(ext_smv, obs_alm, msat, props, timings, use_bmc)
                    pass                    

            except KeyboardInterrupt: 
                if verbose: print "Aborted!"
                out_condition = "Stopped (by the user)"
                break # exits the loop

            except Exception:
                if verbose: print "Unexpected Error!"
                out_condition = "Unexpectedly Terminated"
                break # exits the loop
                
            finally:
                # saves timings (those which are available...)
                outfile.write("%d " % step)
                outfile.write(" ".join((str(timings[name].get_time()) for name in timers)) + os.linesep)
                outfile.flush()
                pass

            if cleanup: # cleanup all generated files
                for fn in [slim, props_f, smv_props_f, smv, ord, ext_smv] + traces: 
                    if fn is not None and os.path.isfile(fn): os.remove(fn)
                    pass

                for triple in fts:
                    for fn in triple: 
                        if os.path.isfile(fn): os.remove(fn)
                        pass
                    pass
                pass
            
            pass

        # write footer:        
        outfile.write("""
# ----------------------------------------------------------------------
# %s: %s
# ----------------------------------------------------------------------
""" % (out_condition, str(datetime.utcnow()) + " UTC"))        

        pass # with output file

    print "DONE. You can find timings in file '%s'" % outfname
    return


# ----------------------------------------------------------------------
# launcher and options handler
if __name__ == "__main__": 
    import getopt

    def usage():
        sys.stderr.write("""Usage: run.py [-m] [-t] [-s] [-c] [-v] \
              [-o <fname>] [-f <num>] [-i <num>] <max>
Where:
 <max>     is the maximum number of bits

 -m        optionally disables model checking
 -t        optionally disables fault tree generation
 -s        uses SAT technology instead of BDD
 -c        does not cleanup generated files (for debugging)
 -v        verbosely prints out information while running (default is silent)
 -o <fn>   outputs timings to file <fn> (default: %s)
 -f <num>  starts from given number of bits (default: 2)
 -i <num>  defines the intervals between steps (default:1)
""" % def_outfname)
        sys.exit(1)
        return        

    optlist, args = getopt.getopt(sys.argv[1:], 'hmto:f:scvi:')
    
    if len(args) != 1: usage()
    try: max_bits = int(args[0])
    except: usage()
    
    log.init()
    
    from_bits = 2
    interval = 1
    do_mc = True
    do_ft = True
    cleanup = True
    verbose=False
    use_bmc = False
    outfname = def_outfname
    for op, val in optlist:
        if op == '-h': usage()
        if op == '-m': do_mc = False
        elif op == '-t': do_ft = False
        elif op == '-s': use_bmc = True
        elif op == '-c': cleanup = False
        elif op == '-v': verbose = True
        elif op == '-o': outfname = val
        elif op == '-i': 
            try: interval = int(val)
            except: usage()
            pass
        elif op == '-f': 
            try: from_bits = int(val)
            except: usage()
            pass

        else: usage() # invalid option
        pass

    if max_bits < 2:
        sys.stderr.write("The maximum number of bits has to be greater then 1.\n")
        sys.exit(1)
        pass

    if from_bits < 2:
        sys.stderr.write("The minimum number of bits has to be greater then the 1.\n")
        sys.exit(1)
        pass        

    if from_bits > max_bits:
        sys.stderr.write("The minimum number of bits has to be greater then the maximum.\n")
        sys.exit(1)
        pass        

    run(from_bits, max_bits, interval, do_mc, do_ft, outfname, 
        use_bmc, cleanup, verbose)
