# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================



import thread
import gobject
from gtkmvc import Controller
from compass import log_internal_errors

# base for controllers of activities
class ActivityCtrl (Controller):
    """This is the base class for all activities. This class is
    used to control the common part of all activities, i.e. the set
    of properties. ActivityController has a sub-controller (field
    subctrl) controlling the subview"""

    def __init__(self, subctrl, model, view):
        Controller.__init__(self, model, view)
        self.subctrl = subctrl
        return

    pass #end of class
# ----------------------------------------------------------------------


class StartStopActivityCtrl (Controller):
    """This controller is intended to be used as base class of
    controllers which need to launch and stop processes.  The idea
    is to have the view holding two buttons, one for starting the
    process, and one for stopping it. When the 'start' button is
    pressed, a custom method is invoked. After some time, the
    controller calls a method in the view (must be a
    StartStopActivityView instance) to show the stop button. The
    delay is introduced to avoid switching in short activity which
    may confuse the user. If the stop button is clicked a custom
    method is called. When the process is over, the view is
    switched back to the 'start' state. 

    The view is conditioned to show either the start or the stop
    state depedning on the value of the observable property 'busy'
    in the model."""

    def __init__(self, m, v, start_method, stop_method):
        """Class constructor. 
        @param start_method is the custom method to be invoked when
        the start button is clicked. THe method will be invoked in
        a separate thread.

        @param stop_method is the custom method to be invoked
        when the stop button is clicked.
        """
        self.start_method = start_method
        self.stop_method = stop_method
        self.delay = 1500
        Controller.__init__(self, m, v)
        return

    def register_view(self, view):
        for getter, handler in ((view.get_start_button, self.on_button_start_clicked), 
                                (view.get_stop_button, self.on_button_stop_clicked)):
            button = getter().connect('clicked', handler)
            pass

        # initial setting for the view
        if self.model.busy > 0: self.view.show_stop()
        else: self.view.show_start()
        return

    # Runs the activity in a new thread
    def start_activity(self):
        
        # wraps the start_method in order to manage exceptions
        def managed_start_method():
            import sys
            try:
                self.start_method()
            except Exception as e:
                log_internal_errors(*sys.exc_info())
                return
        
        thread.start_new_thread(managed_start_method, ())
        
        #thread.start_new_thread(self.start_method, ())
        return

    # stops the running activity
    def stop_activity(self):
        self.stop_method()
        return

    def show_stop_delay(self):
        """This method is invoked (after a delay) when the busy
        level of the model is greater than zero, to show the 'stop'
        button"""
        if self.model.busy > 0: self.view.show_stop()
        return
    
    def on_button_start_clicked(self, button): 
        self.start_activity()
        return

    def on_button_stop_clicked(self, button): self.stop_activity()

    def property_busy_value_change(self, model, old, new):
        if model is not self.model: return 
        
        if old == 0 and new > 0: gobject.timeout_add(self.delay, self.show_stop_delay)
        elif new == 0: self.view.show_start()

        # grays out the start button to avoid multiple calls
        self.view.get_start_button().set_sensitive(new == 0)
        return

    pass # end of class
# ----------------------------------------------------------------------
