# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by GMV Aerospace and Defence  
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================



from gtkmvc import Controller
from gtkmvc import adapters
import gtk
import gobject

from utils.misc import relpath

# sub-controller
from src.ctrls.slim import SlimCtrl
from src.ctrls.properties import PropertiesCtrl
from src.ctrls.correctness import CorrectnessCtrl
from src.ctrls.performability import PerformabilityCtrl
from src.ctrls.safety import SafetyCtrl
from src.ctrls.fdir import FdirCtrl
from src.ctrls.tool_handlers import ToolHandlersCtrl
from src.ctrls.validation import ValidationCtrl
from src.ctrls.mission import MissionCtrl
from src.ctrls.tfpg import TfpgCtrl

from src.views.application import AskRootView, AskQuitView, AboutView
from src.views.application import WarningDegradedModeView, WarningDegradedModeTempView

from src.models.application import AboutModel


class ApplCtrl (Controller):

    tfpg_activate = False

    # Observable properties
    __observables__ = ("tfpg_activate", )

    
    def __init__(self, model, view):
        Controller.__init__(self, model, view)
        
        # creates all sub-controllers
        self.menu = AppMenuCtrl(model, view, self)

        self.slim = SlimCtrl(model.slim, view.slim)
        self.properties = PropertiesCtrl(model.properties, view.properties)
        self.mission = MissionCtrl(model.mission, view.mission)
        self.correctness = CorrectnessCtrl(model.correctness, view.correctness)
        self.performability = PerformabilityCtrl(model.performability, view.performability)
        self.safety = SafetyCtrl(model.safety, view.safety)
        self.fdir = FdirCtrl(model.fdir, view.fdir)
        self.tfpg = TfpgCtrl(model.tfpg, view.tfpg)
        self.statusbar = StatusBarCtrl(model, view)
        self.validation = ValidationCtrl(model.validation, view.validation)

        self.tools = ToolHandlersCtrl(model.tools, view.tools)

        # previous button (used to control the toggle activity rollback)
        self._prev_activity = None

        # we are interested in seeing if the slim model has failures
        model.slim.register_observer(self)
        model.mission.register_observer(self)
        model.tfpg.register_observer(self)

        return

    def register_view(self, view):
        # setup of widgets
        self.view['window_top'].connect('delete-event', self.on_window_appl_delete_event)

        self.view.tools['window_processes'].connect('delete-event', 
                                                    self.on_window_processes_delete_event)

        # sets geometries saved from previous session
        width, height = self.view.get_top_widget().get_size()
        if self.model.config.has_option("Geometry", 'appl_width'):
            width = self.model.config.getint("Geometry", 'appl_width')
            pass
        if self.model.config.has_option("Geometry", 'appl_height'):
            height = self.model.config.getint("Geometry", 'appl_height')
            pass
        self.view.get_top_widget().resize(width, height)

        # checks degraded mode and show the warning if needed
        opt = "do_not_show_degraded_mode"
        if self.model.is_degraded_mode():
            if self.model.config.getboolean("Dialogs", opt):
                # shows a temporary window with a warning
                gobject.timeout_add(500, self.show_degraded_popup)
                pass
                
            else:
                # show the warning dialog
                v = WarningDegradedModeView(self.view)
                res, do_not_show, uri = v.run()
                if do_not_show: self.model.config.set("Dialogs", opt, "yes")
                if 1 == res: # download
                    import webbrowser
                    webbrowser.open_new(uri)
                    pass
                pass
            self.view.set_degraded_mode()
            pass
        
        self.on_menu_tfpg_item_enabled()
        return

    def register_adapters(self):
        # setup of adapters
        return

    def shutdown_now(self):
        """Quits the application."""
        # saves the geometries

        width, height = self.view.get_top_widget().get_size()
        self.model.config.set("Geometry", 'appl_width', str(width))
        self.model.config.set("Geometry", 'appl_height', str(height))

        self.model.quit()

        # If a tfpg or slim_associations are not saved, a warning message 
        # appears with the possibility to save it or not.
        self.tfpg.tfpgs_not_saved()
        self.tfpg.tfpg_slim_assoc_ctrl.slim_associations_not_saved()

        gtk.main_quit() # say goodbye      
        return       
        
    def ask_user_to_quit(self):
        """Opens a dialog to ask the user to quit
        explicitly. Returns True if the user answered 'yes'"""
        dialog = AskQuitView()
        return dialog.run() == 1


    # ---------------------------------------------------------------
    #        Signal handlers
    # ---------------------------------------------------------------
    def on_window_appl_delete_event(self, win, event):
        if (self.model.can_quit_safely() or self.ask_user_to_quit()):
            self.shutdown_now()
            pass

        return True

    def on_tb_activity_toggled(self, button):
        """The user selected a different activity view"""

        if button.get_active():
            activity = button.get_name()[len("tb_"):]
            
            if activity not in ("model", "validation") and not self.model.get_current_root():
                # root was not selected, asks to the user to specify one:
                v = AskRootView()
                c = AskRootCtrl(self.model.slim, v)
                response = v.run()
                self.model.slim.unregister_observer(c)
                if response != 1: 
                    if self._prev_activity: self.view[self._prev_activity].set_active(True)
                    return 
                pass
                    
            # here we get the name of the button, that is assumed
            # to have a corresponding view:
            self.view.select_view(activity)
        else:
            # disabling: keep for rollback
            self._prev_activity = button.get_name()
            pass

        # The abllons with the errore messages in slim associations
        # are hidden in a different view is selected
        if self.view.curr_activity != "tfpg":
            self.tfpg.tfpg_slim_assoc_ctrl.balloon_in_slim_assoc.hide_all()
            self.tfpg.tfpg_synthesis_ctrl.balloon_in_synth.hide_all()

        return

    def on_window_processes_delete_event(self, window, event):
        self.tools.disable()
        self.view['mi_processes'].set_active(False)
        return True

    def show_degraded_popup(self):
        v = WarningDegradedModeTempView(self.view)
        v.show()
        gobject.timeout_add(50, self.quit_degraded_popup, v)
        return False

    def quit_degraded_popup(self, view):
        res = view.tick()
        if not res: view.destroy()
        return res

    # ---------------------------------------------------------------
    #        Notifications from observable properties
    # ---------------------------------------------------------------
    def property_failures_value_change(self, model, old, new):
        """This notification is called when into the slim model
        there are injected failures"""

        # If a tfpg or slim_associations are not saved, a warning message 
        # appears with the possibility to save it or not.
        self.tfpg.tfpgs_not_saved()
        self.tfpg.tfpg_slim_assoc_ctrl.slim_associations_not_saved()
        self.tfpg.tfpg_slim_assoc_ctrl.clear_slim_assoc_list()
        self.model.tfpg.clear_tfpgs_list()
        self.view.tfpg.clear_subviews()
        
        self.on_menu_tfpg_item_enabled()

        self.view.enable_failures(new>0 and 
                                  not self.model.is_degraded_mode())
        self.view.enable_performability(new>0)
        
        # Feature #595: Removed the dependency of TFPG from Mission
        # the TFPG is enable after loading slim files
        # and loading/creating fault injections.
        self.view.enable_tfpg(new > 0 and not self.model.is_degraded_mode())

        return


    def property_slim_files_num_value_change(self, model, old, new):
        """This notification is called when the number of loaded
        slim files (into the observed slim model) is changed. It is
        used to enable/disable some features based on the presence
        of at least one slim model."""

        # If a tfpg or slim_associations are not saved, a warning message 
        # appears with the possibility to save it or not.
        self.tfpg.tfpgs_not_saved()
        self.tfpg.tfpg_slim_assoc_ctrl.slim_associations_not_saved()
        self.tfpg.tfpg_slim_assoc_ctrl.clear_slim_assoc_list()
        self.model.tfpg.clear_tfpgs_list()
        self.view.tfpg.clear_subviews()

        self.mission.model.clear()
        self.on_menu_tfpg_item_enabled()
        
        if (new > 0 and old == 0) or (new==0 and old > 0):
            self.view.enable_properties(new > 0)            
            self.view.enable_correctness(new > 0 and
                                         (not self.model.is_degraded_mode() or
                                          not model.infinite_domain))
            self.view.enable_mission(new > 0
                                     and not self.model.is_degraded_mode())

            pass

        return

    def property_infinite_domain_value_change(self, model, old, new):
        """If the slim domain has infinite domain, performability is not
           available"""
        self.view.enable_performability(self.slim.model.failures > 0)
        
    def property_open_world_value_change(self, model, old, new):
        self.performability.view.subview.display_ports_warning(new)


    def property_mission_changed_signal_emit(self, model, args):

        # If a tfpg or slim_associations are not saved, a warning message 
        # appears with the possibility to save it or not.
        #self.tfpg.tfpgs_not_saved()
        #self.tfpg.tfpg_slim_assoc_ctrl.slim_associations_not_saved()
        #self.tfpg.tfpg_slim_assoc_ctrl.clear_slim_assoc_list()
        #self.model.tfpg.clear_tfpgs_list()
        #self.view.tfpg.clear_subviews()

        self.on_menu_tfpg_item_enabled()

        # Feature #595: Removed the dependency of TFPG from Mission
        # the TFPG is enable after loading slim files
        # and loading/creating fault injections.
        
#        self.view.enable_tfpg(
#            self.mission.model.get_phase_op_mode_definitions_counter() > 0 and 		
#            self.mission.model.get_phase_op_mode_associations_counter() == \
#                self.mission.model.get_phase_op_mode_definitions_counter() and
#            self.slim.model.failures > 0)
        
        return


    def property_tfpgs_list_changed_signal_emit(self, model, args):

        if self.tfpg.tfpg_synthesis_ctrl.tfpg_synthesized_loaded is False:
            associations = self.tfpg.tfpg_synthesis_ctrl.initialize_tfpg_model_tab_in_synth([""], [""], [""], [""])

        if self.model.tfpg.get_tfpgs_counter() == 0:
            self.tfpg_activate = False
            self.tfpg.view.tfpg_slim_assoc_view['button_load_associations'].set_sensitive(False)
            self.tfpg.tfpg_synthesis_ctrl.curr_tfpg_name = None
            self.tfpg.tfpg_synthesis_ctrl.synthesis_counter = 0
        else: 
            self.tfpg_activate = True
        
        self.on_menu_tfpg_item_enabled()
     
        return


    def on_menu_tfpg_item_enabled(self):
    
        for mitem in self.view['menuitem_file_menu']:
        
            if self.tfpg_activate is False:
                self.view['menu_save_tfpg'].set_sensitive(False)
            else:
                self.view['menu_save_tfpg'].set_sensitive(True)
                
        return


    pass # end of class ApplCtrl
# ----------------------------------------------------------------------


class AppMenuCtrl (Controller):
    """This is the controller of the application menu. It shares
    the view and model of the main application controller"""

    def __init__(self, m, v, parent):
        Controller.__init__(self, m, v)
        self.parent = parent
        # this is used internally to avoid loops with signals:
        self._itsme = False
        return

    # ---------------------------------------------------------------
    #        Signal handlers
    # ---------------------------------------------------------------    

    # File menu
    def on_mi_file_quit_activate(self, mitem):
        self.parent.shutdown_now()
        return

    def on_mi_file_open_activate(self, mitem):
        self.parent.slim.load_slim_files()
        return

    def on_menu_file_reset_activate(self, mitem):
        self.model.reset_user_config()
        return

    def on_mi_act_activate(self, mitem):
        if self._itsme or not mitem.get_active(): return
        self._itsme = True
        activity = mitem.get_name()[len("mi_act_"):]

        # enables the corresponding activity
        self.view['tb_%s' % activity].set_active(True)
        self._itsme = False
        return

    def on_mi_help_about_activate(self, mitem):
        m = AboutModel(self.model.is_degraded_mode())
        v = AboutView()
        c = AboutCtrl(m, v)

        response = v.run()
        c.relieve_model(m)
        return

    def on_mi_processes_toggled(self, mi):
        # shows the tools monitor dialog
        if mi.get_active(): self.parent.tools.enable()
        else: self.parent.tools.disable()
        return 
        
    def on_menu_save_fi_activate(self, mitem):
        self.parent.slim.save_fault_injections()
        return        
        
    def on_menu_load_fi_activate(self, mitem):
        self.parent.slim.load_fault_injections()
        self.parent.on_menu_tfpg_item_enabled()
        return
    
    def on_menu_save_props_activate(self, mitem):
        self.parent.properties.save_properties()
        return
        
    def on_menu_load_props_activate(self, mitem):
        self.parent.properties.load_properties()
        self.parent.on_menu_tfpg_item_enabled()
        return

    def on_menu_save_mission_activate(self, mitem):
        self.parent.mission.save_mission()
        return

    def on_menu_load_mission_activate(self, miter):
        if (self.parent.slim.model.slim_files_num > 0) and (not self.parent.model.is_degraded_mode()):
            if (not self.model.get_current_root()):
                v = AskRootView()
                c = AskRootCtrl(self.model.slim, v)
                response = v.run()
                self.model.slim.unregister_observer(c)
            else:
                self.parent.mission.load_mission()
                self.parent.on_menu_tfpg_item_enabled()
        else:
            self.parent.slim.model.console.log_compiler_fail("ERROR\n")
            self.parent.slim.model.console.log_compiler_fail(" > It's not possible to load Mission informations if the tool is in degraded mode\n" +
                                                             " > or there is no SLIM files loaded.\n")
        return


    def on_menu_save_tfpg_activate(self, miter):	
        if self.parent.tfpg_activate:					
            self.parent.tfpg.save_tfpg()
        return
        

     # Feature #595: Removed the dependency of TFPG from Mission
     # the TFPG is enable after loading slim files
     # and loading/creating fault injections.
        
#    def on_menu_load_tfpg_activate(self, miter):							
#        if (self.parent.mission.model.get_phase_op_mode_definitions_counter() > 0) and \
#                (self.parent.mission.model.get_phase_op_mode_associations_counter() == \
#                     self.parent.mission.model.get_phase_op_mode_definitions_counter()) and \
#                     (self.parent.slim.model.failures > 0):
#            self.parent.tfpg.load_tfpg()
#            self.parent.tfpg_activate = True
#            self.parent.on_menu_tfpg_item_enabled()
#        else:
#            self.parent.slim.model.console.log_compiler_fail("ERROR\n")
#            self.parent.slim.model.console.log_compiler_fail(" > It's not possible to load TFPG information if :\n" +
#                                                             " >    - the tool is in degraded mode\n" +
#                                                             " >    - there is no SLIM files loaded\n" +
#                                                             " >    - there is no Fault Injection activated\n" +
#                                                             " >    - the tab Mission is not completed.\n")
#        return


    def on_menu_load_tfpg_activate(self, miter):
        
        if (self.parent.slim.model.failures > 0):
            if (not self.model.get_current_root()):
                v = AskRootView()
                c = AskRootCtrl(self.model.slim, v)
                response = v.run()
                self.model.slim.unregister_observer(c)
            else:
                self.parent.tfpg.load_tfpg()
                self.parent.tfpg_activate = True
                self.parent.on_menu_tfpg_item_enabled()
        else:
            self.parent.slim.model.console.log_compiler_fail("ERROR\n")
            self.parent.slim.model.console.log_compiler_fail(" > It's not possible to load TFPG information if :\n" +
                                                             " >    - the tool is in degraded mode\n" +
                                                             " >    - there is no SLIM files loaded\n" +
                                                             " >    - there is no Fault Injection activated\n")
        return
        
        
    pass # end of class
# ----------------------------------------------------------------------


class StatusBarCtrl (Controller):
    """This is the controller specifically made for controlling the
    status bar of the application. The view and the model are
    shared with the main application controller."""
    

    # ---------------------------------------------------------------
    #        Signals and other callbacks
    # ---------------------------------------------------------------
    def pulse_progress_bar_delay(self, progress_bar):
        """this is called once with a little delay, to do not show
        the progress bar for fast activities"""
        if self.model.busy > 0:
            self.view.enable_progress_bar(True)

            # time to show the progress bar
            gobject.timeout_add(250,
                                self.pulse_progress_bar, progress_bar, 
                                priority=gobject.PRIORITY_HIGH)
            pass

        return False

    def pulse_progress_bar(self, progress_bar):
        """Used to pulse the progress bar"""
        progress_bar.pulse()
        return self.model.busy > 0
    

    # ---------------------------------------------------------------
    #        Notifications from observable properties
    # ---------------------------------------------------------------
    def property_busy_value_change(self, model, old, new):
        """Busy is observed to report feedback to the user about
        background actitivies, e.g. into the status bar."""
        if new > 0 and old == 0: 
            gobject.timeout_add(1500, self.pulse_progress_bar_delay, self.view['progress'])

        elif old > 0 and new == 0: self.view.enable_progress_bar(False)        
        
        if new == 1: self.view.set_progress_bar_text("1 activity")
        elif new > 1: self.view.set_progress_bar_text("%d activities" % new)
        return

    pass # end of class StatusBarCtrl
# ----------------------------------------------------------------------


class AskRootCtrl (Controller):
    """This is the controller for the dialog used to prompt the
    user for the root implementation selection when requested. The
    model used by this controller is the SlimModel"""
    
    def register_view(self, view):
        tv = view['tv_roots']
        
        view['tv_roots'].set_model(self.model.roots)

        def __fname(col, rend, model, iter, col_num):
            fname = model.get_value(iter, col_num)
            relfname = relpath(fname)
            if len(fname) < len(relfname): rend.set_property('text', fname)
            else: rend.set_property('text', relfname)
            return
            
        # used to show the name of an implementation within a treecolumn
        def __impl_data(col, rend, model, iter, col_num):
            impl = model.get_value(iter, col_num)
            rend.set_property('text', impl.name)
            return
        
        col = gtk.TreeViewColumn('Implementation')
        # root selected
        rend1 = gtk.CellRendererToggle()
        rend1.set_property("activatable", True)
        rend1.connect("toggled", self.on_root_toggled)
        col.pack_start(rend1)
        col.add_attribute(rend1, "active", 2)
       
        # root name
        rend2 = gtk.CellRendererText()
        col.pack_start(rend2)
        col.set_cell_data_func(rend2, __impl_data, 0)
        tv.append_column(col)

        # this is for the filename containing the root
        rend = gtk.CellRendererText()
        col = gtk.TreeViewColumn('Slim file', rend)
        col.set_cell_data_func(rend, __fname, 1)
        tv.append_column(col)
        
        # this is for root selection 
        sel = tv.get_selection()
        sel.set_mode(gtk.SELECTION_NONE)

        return


    # ---------------------------------------------------------------
    #        Signal handlers
    # ---------------------------------------------------------------    
    def on_root_toggled(self, rend, path):
        self.model.toggle_root(path)
        return

    # ---------------------------------------------------------------
    #        Notifications from observable properties
    # ---------------------------------------------------------------
    def property_root_impl_value_change(self, model, old, new):
        """this observes the selected root impl to enable the "OK"
        button"""      
        if not self.view['button_ok']: return
        self.view['button_ok'].set_sensitive(new != None)
        return


    pass # end of class
# ----------------------------------------------------------------------


class AboutCtrl (Controller):
    """Controller for the 'about' dialog. It shares
    the model of the main application controller"""
    anim_timeout = 20

    def register_view(self, view):
        Controller.register_view(self, view)    
        
        self.view.create_background(*view['da_logos'].window.get_size())
        gobject.timeout_add(self.anim_timeout,
                            self.on_anim_timer,
                            priority=gobject.PRIORITY_HIGH)

        if self.model.is_degraded_mode():
            mode = "<u><b>Degraded Mode</b></u>"
            special = "<b>only</b> the <i>CPL</i> license applies"
        else:
            mode = "<u><b>Fully Operative Mode</b></u>"
            special = "<b>both</b> <i>CPL</i> and <i>Background</i> licenses apply"
            pass

        self.view.set_appl_mode("The application is running in %s \nand %s." \
                                    % (mode, special))

        self.view['tv_cpl'].set_buffer(self.model.tb_cpl)
        self.view['tv_fbkbg'].set_buffer(self.model.tb_bg)
        

        # change to evaluation license display if applicable
        if self.model.is_evaluation():
            self.view.set_appl_mode("The application is running in Evaluation mode. The COMPASS Evaluation License applies.")
            notebook = self.view['notebook1']
            while notebook.get_n_pages() != 1:
                notebook.remove_page(0)
                pass
            pass
            page_child = notebook.get_nth_page(0)
            notebook.set_tab_label_text(page_child, "Evaluation License")
        return



    # signals handing
    def on_linkbutton_clicked(self, button):
        import webbrowser
        webbrowser.open(button.get_uri())
        return

    def on_da_logos_expose_event(self, da, event):
        #if 0 == event.count: 
        #            self.view.create_background(event.area.width, event.area.height)
        
        self.view.draw_logos(event.area)
        return

    def on_da_logos_size_allocate(self, da, rect):
        if da.window is None: return
        # here it is resizing
        self.view.create_background(rect.width, rect.height)        
        return

    def on_anim_timer(self):
        if self.view['da_logos'] is None: return False
        self.view.animate_logos()
        self.view.draw_logos()
        return True

    pass # end of class
# ----------------------------------------------------------------------

