# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


import gtk
import gobject
import os

from gtkmvc import Controller
from gtkmvc.adapters import Adapter

from slim_compiler.src.dataStructure.visitors.printVisitor import toSLIMVisitor

import options

class ConsoleCtrl (Controller):
    """This is the controller for the console, handling compiler
    output and logging. The view is at the moment shared with the
    slim controller, which is also the parent controller."""

    def __init__(self, m, v):
        Controller.__init__(self, m, v)
        return

    def register_view(self, view):
        
        def scroll_to_end(buf, tv):
            # as a final step make sure last lines in the buffer are visible
            # by scrolling the view to show them
            tmp = buf.create_mark("__last_position", buf.get_end_iter(), False)
            tv.scroll_mark_onscreen(tmp)

            # get rid of temporary mark
            buf.delete_mark(tmp)

        # connects the buffers in the model
        self.view['tv_compiler'].set_buffer(self.model.compiler_buf)
        self.model.compiler_buf.connect("changed", scroll_to_end, self.view['tv_compiler'] )
 
        self.view['tv_log'].set_buffer(self.model.log_buf)
        self.model.log_buf.connect("changed", scroll_to_end, self.view['tv_log'] )

        self.view['tv_extended_model'].set_buffer(self.model.extended_model_buf)
        self.model.extended_model_buf.connect("changed", scroll_to_end, self.view['tv_extended_model'] )

        self.view['tv_metrics'].set_buffer(self.model.metrics_buf)
        self.model.metrics_buf.connect("changed", scroll_to_end, self.view['tv_metrics'] )

        return

    def expand_output(self):
        """Expands the console to show an error, or any condition
        that maight be interesting"""
        self.view.show_output()
        vp = self.view['vpaned_top']
        gobject.timeout_add(8, self.expand_timer, vp, 
                            vp.get_position(), 
                            max(vp.get_position()-160, vp.get_property("min-position")+1),
                            priority=gobject.PRIORITY_HIGH)
        return

    # ---------------------------------------------------------------
    #        Signal handlers and other callbacks
    # ---------------------------------------------------------------    
    def on_button_clear_log_clicked(self, button): self.model.clear_log()
    
    def on_button_save_log_clicked(self, button): self.save_log()

    def expand_timer(self, vp, orig_pos, dest_pos):
        cur_pos = vp.get_position()
        if cur_pos <= dest_pos: return False
        vp.set_position(cur_pos - 8)
        return True

    # ---------------------------------------------------------------
    #        Notifications from observable properties
    # ---------------------------------------------------------------
    def property_changed_signal_emit(self, slim_model, args):
        e_str = "-- NOTE: " + str(slim_model.failures) + " fault injections specified.\n"
        if slim_model.failures > 0:
            eroot_impl = slim_model.get_extended_root_impl()
            if eroot_impl:
                e_str += toSLIMVisitor().visit(eroot_impl.subpackage.package.compile_information)
                pass
            pass
        self.model.extended_model_buf.set_text(e_str)
        # Use the idle queue to clear the logs, as messages are added using the same queue
        gobject.idle_add(self.model.clear_metrics)
        slim_model.log_metrics()
        return

    def save_log(self):
        """ Save the current logging date to file"""

        dialog = gtk.FileChooserDialog("Save Log",
                                       None,
                                       gtk.FILE_CHOOSER_ACTION_SAVE,
                                       (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                                       gtk.STOCK_SAVE, gtk.RESPONSE_OK))
        dialog.set_default_response(gtk.RESPONSE_OK)
        filter = gtk.FileFilter()
        filter.set_name("Log file")
        filter.add_pattern("*.log")
        dialog.add_filter(filter)
        dialog.set_current_name(options.BASE_FILE_NAME + ".log")
        dialog.set_current_folder(os.curdir)
        dialog.set_do_overwrite_confirmation(True)
        dialog.show()
        response = dialog.run()
        filename = dialog.get_filename()
        dialog.destroy()
        if response == gtk.RESPONSE_OK:
            name = filename
            fd = open(filename, 'w')
            text = self.model.log_buf.get_text(*self.model.log_buf.get_bounds())
            fd.write(text)
            fd.close()

        return

    pass # end of class ConsoleCtrl

    

