# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by GMV Aerospace and Defence  
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================

import os.path
import gtk
import gobject
from gtkmvc import Controller
from src.ctrls.mission_phases_op_modes import PhasesOpModesCtrl
from src.ctrls.mission_configurations import ConfigurationsCtrl
from src.ctrls.mission_phase_op_mode_combinations import PhaseOpModeCombinationsCtrl
from hasdel_tools.src.generator import generate_unique_mission_basename
import hasdel_tools.src.xml.mission as mission


class MissionCtrl (Controller):
    """Represents the Controller of Mission."""

    def __init__(self, model, view):
        """Constructor Method"""
        Controller.__init__(self, model, view)

        self.phases_op_modes = PhasesOpModesCtrl(model, view.phases_op_modes)
        self.configurations = ConfigurationsCtrl(model, view.configurations)
        self.phase_op_mode_combinations = \
            PhaseOpModeCombinationsCtrl(model, view.phase_op_mode_combinations)
    
        self.model.slim_model.register_observer(self)
        self.model.register_observer(self)
        return


    def save_mission(self):
        """Saves a mission document in .mxml format."""

        dialog = gtk.FileChooserDialog("Save Mission",
                                       None,
                                       gtk.FILE_CHOOSER_ACTION_SAVE,
                                       (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                                       gtk.STOCK_SAVE, gtk.RESPONSE_OK))
        dialog.set_default_response(gtk.RESPONSE_OK)
        filter = gtk.FileFilter()
        filter.set_name("AUTOGEF Mission")
        filter.add_pattern("*.mxml")
        dialog.add_filter(filter)
        dialog.set_current_name(os.path.basename(
            generate_unique_mission_basename() + ".mxml"))
        dialog.set_current_folder(os.path.dirname(
            generate_unique_mission_basename() + ".mxml"))
        dialog.set_do_overwrite_confirmation(True)    
        dialog.show()
        response = dialog.run()        
        if response == gtk.RESPONSE_OK:
            name = dialog.get_filename()
            fd = open(dialog.get_filename(), 'w')
            fd.write(str(mission.create_mission_document(self.model)))
            pass
        dialog.destroy()
        return


    def load_mission(self):
        """Loads a mission document in .mxml format."""

        chooser = gtk.FileChooserDialog(title="Open Mission",
                                        action=gtk.FILE_CHOOSER_ACTION_OPEN,
                                        buttons=(gtk.STOCK_CANCEL,gtk.RESPONSE_CANCEL,
                                                 gtk.STOCK_OPEN,
                                                 gtk.RESPONSE_OK))
        chooser.set_select_multiple(False)
        filter = gtk.FileFilter()
        filter.set_name("AUTOGEF Mission")
        filter.add_pattern("*.mxml")
        chooser.add_filter(filter)
        chooser.set_current_folder(os.path.dirname(
            generate_unique_mission_basename() + ".mxml"))

        response = chooser.run()
        if response == gtk.RESPONSE_OK:
            res = self.model.get_mission_doc(chooser.get_filename())
            if res is not None:
                self.console.expand_output()                
                pass            
            pass

        chooser.destroy()
        return


    # ---------------------------------------------------------------
    #        Signal handlers
    # ---------------------------------------------------------------    
    def property_changed_signal_emit(self, model, args):
        """When a change occured in the mission, a signal enables the tb_phases_op_modes
           @param   model:  The Slim Model.
	   @type    model:  SlimModel
           @param   args:  The arguments of Mission Model.
	   @type    args:  string
        """
        self.view["tb_phases_op_modes"].set_active(True)
        self.view.select_view("phases_op_modes")
        return


    def on_tb_mission_toggled(self, button):
        """The user selected a different Mission tab.
           @param   button:  The button selected.
	   @type    button:  gtk.Button
        """
        if button.get_active():
            activity = button.get_name()[len("tb_"):]
            self.view.select_view(activity)
            pass

        return
    

    # ---------------------------------------------------------------
    #        Notifications from observable properties
    # ---------------------------------------------------------------
    def property_phases_counter_value_change(self, model, old, new):
        """When a change occured in the phases, a signal enables the Configurations or not.
           @param   model:  The Mission Model.
	   @type    model:  MissionModel
           @param   old:  The number of old phases.
	   @type    old:  int
           @param   new:  The number of new phases.
	   @type    new:  int
        """
        if (new > 0 and old == 0) or (new == 0 and old > 0):
            self.view.enable_configurations(new > 0 and
                                            model.get_op_modes_counter() > 0)
            self.view.enable_phase_op_mode_combinations(new > 0 and
                                            model.get_op_modes_counter() > 0)
            pass
        return


    def property_op_modes_counter_value_change(self, model, old, new):
        """When a change occured in the op_modes, a signal enables the Configurations or not.
           @param   model:  The Mission Model.
	   @type    model:  MissionModel
           @param   old:  The number of old op_modes.
	   @type    old:  int
           @param   new:  The number of new op_modes.
	   @type    new:  int
        """
        if (new > 0 and old == 0) or (new == 0 and old > 0):
            self.view.enable_configurations(new > 0 and
                                            model.get_phases_counter() > 0)
            pass
        return


    def property_configurations_counter_value_change(self, model, old, new):
        """When a change occured in the configurations, a signal enables the PhasesOpModes Combinations or not.
           @param   model:  The Mission Model.
	   @type    model:  MissionModel
           @param   old:  The number of old configurations.
	   @type    old:  int
           @param   new:  The number of new configurations.
	   @type    new:  int
        """
        if (new > 0 and old == 0) or (new == 0 and old > 0):
            self.view.enable_phase_op_mode_combinations(new > 0)
        return


    def mission_busy_value_change(self, model, old, new):
	"""Busy level, in this case, is 0 if AUTOGEF is not running and
           1 otherwise. If AUTOGEF is active, all the informations contained
           in the "Fdir Synthesis" tab cannot be edited. This could otherwise
           lead to misinterpretations after AUTOGEF lead to a result.
           @param   model:  The Mission Model.
	   @type    model:  MissionModel
           @param   old:  The number of old value.
	   @type    old:  int
           @param   new:  The number of new value.
	   @type    new:  int
        """
        self.phases_op_modes.phases_op_modes_busy_value_change(
            self, model, old, new)
        self.configurations.configurations_busy_value_change(
            self, model, old, new)
        self.phase_op_mode_combinations.phase_op_mode_combinations_busy_value_change(
            self, model, old, new)

        return

pass # end of MissionCtrl class
# ----------------------------------------------------------------------
