# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by GMV Aerospace and Defence  
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================

import os.path
import gtk
import gobject
from gtkmvc import Controller
from src.models.mission import ConfigurationDialogModel
from src.views.mission_configurations import ConfigurationDialogView
from delayedcall import DelayedCall
import hasdel_tools.src.functions as functions


class ConfigurationsCtrl (Controller):
    """Represents the Controller of Configurations."""

    def __init__(self, model, view):
        """Constructor Method"""
        Controller.__init__(self, model, view)
        return


    def register_view(self, view):
        """Sets the windows, dialog, buttons and other parts of View."""
        Controller.register_view(self, view)

        view['button_add_configuration'].set_sensitive(True)
        view['button_edit_configuration'].set_sensitive(False)
        view['button_remove_configurations'].set_sensitive(False)

        tv = view['tv_configurations']
        tv.set_model(self.model.get_configurations_list())
        for col in self.model.get_configurations_cols():
            tv.append_column(col)

        # to control selection of configurations
        sel = self.view['tv_configurations'].get_selection()
        sel.set_mode(gtk.SELECTION_MULTIPLE)
        sel.connect('changed', self.on_configurations_selection_changed)

        return


    def register_adapters(self):
        return


    def open_confirm_configuration_window(self, tree_model=None, paths=None):
        """Sets the Confirm Window of a new configuration edited.
           @param   tree_model:  The tree_model of configuration.
	   @type    tree_model:  TreeModel
           @param   paths:  The path of configuration.
	   @type    paths:  tuple
        """
        
        if paths is None:
            old_configuration = None
        else:
            old_configuration = \
                tree_model.get_value(tree_model.get_iter(paths[0]), 0)

        m = ConfigurationDialogModel(self.model, old_configuration)
        v = ConfigurationDialogView()
        c = ConfigurationDialogCtrl(m, v)

        response = v.run()

        if gtk.RESPONSE_APPLY == response:
            if old_configuration is None:
                self.model.append_configuration(m)
            else:
                self.model.edit_configuration(m, tree_model.get_iter(paths[0]))

        return


    # ---------------------------------------------------------------
    #        Signal handlers
    # ---------------------------------------------------------------    
    def _clear_selections(self):
        """Clears the old configurations selected."""
        sel = self.view['tv_configurations'].get_selection()
        sel.unselect_all()
        return


    def on_button_add_configuration_clicked(self, button):
        """Adds a configuration to the Configurations list.
           @param   button:  The Add Button (self.view['button_add_configuration']).
	   @type    button:  gtk.Button
        """
        self.open_confirm_configuration_window()
        self._clear_selections()
        return


    def on_button_remove_configurations_clicked(self, button):
        """Removes the configuration selected in the list.
           @param   button:  The Remove Button (self.view['button_remove_configurations']).
	   @type    button:  gtk.Button 
        """
        sel = self.view['tv_configurations'].get_selection()
        tree_model, paths = sel.get_selected_rows()
        self.model.remove_configurations(tree_model, paths)
        self._clear_selections()
        return


    def on_configurations_selection_changed(self, selection):
        """Changes the configuration selected.
           @param   selection:  The selection of configuration.
	   @type    selection:  gtk.TreeSelection
        """
        self.view['button_remove_configurations'].set_sensitive(
            selection.count_selected_rows() > 0)
        
        """ In addition, the "Edit" button has to become enabled when
        one configuration is selected."""
        _, paths = selection.get_selected_rows()
        self.view['button_edit_configuration'].set_sensitive(
            selection.count_selected_rows() == 1)
        return


    def on_button_edit_configuration_clicked(self, button):
        """Edits the configuration selected in the list.
           @param   button:  The Edit Button (self.view['button_edit_configuration']).
	   @type    button:  gtk.Button
        """
        sel = self.view['tv_configurations'].get_selection()
        tree_model, paths = sel.get_selected_rows()
        
        self.open_confirm_configuration_window(tree_model, paths)
        self._clear_selections()
        return


    def on_configuration_selected(self, tv, event):
        """Selects the configuration from the list.
           @param   tv:  The Tree View of the configuration (self.view['tv_configurations']).
	   @type    tv:  gtk.TreeView
           @param   event:  The event associated
	   @type    event:  gtk.gdk.Event
        """
        if event.type == gtk.gdk._2BUTTON_PRESS: 
            sel = self.view['tv_configurations'].get_selection()
            tree_model, paths = sel.get_selected_rows()
            if paths: 
                self.open_confirm_configuration_window(tree_model, paths)
                self._clear_selections()
            else: return
        return 


    def configurations_busy_value_change(self, model, old, new):
	"""Busy level, in this case, is 0 if AUTOGEF is not running and
           1 otherwise. If AUTOGEF is active, all the informations contained
           in the "Mission" tab cannot be edited. This could otherwise
           lead to misinterpretations after AUTOGEF lead to a result.
           @param   model:  The Mission Model.
	   @type    model:  MissionModel
           @param   old:  The number of old value.
	   @type    old:  int
           @param   new:  The number of new value.
	   @type    new:  int
        """
	if model is self: return # not interested in introspection

	self.view['button_edit_configuration'].set_sensitive(new == 0)
	self.view['button_remove_configurations'].set_sensitive(new == 0)
	self.view['button_add_configuration'].set_sensitive(new == 0)
	return


pass # end of ConfigurationsCtrl class
# ----------------------------------------------------------------------


class ConfigurationDialogCtrl(Controller):
    """Represents the Dialog Class of Configuration"""

    def __init__(self, model, view, loading_configurations=False):
        """Constructor Method"""
        Controller.__init__(self, model, view)
        self.loading_configurations = loading_configurations
        self.effect_valid1 = False
        self.effect_valid2 = False
        model.register_observer(self)
        return


    def register_view(self, view):
        """Sets the windows, dialog, buttons and other parts of View."""
        self.delay = DelayedCall(100, view.change_color_and_balloon)
        self.spacecraft_correct = False
        old_configuration = self.model.get_old_configuration()

        if old_configuration:
            old_configuration_id = \
                old_configuration.get_configuration_id()
            old_spacecraft = \
                old_configuration.get_spacecraft()

            self.view['entry_cc_configuration_id'].set_text(
                old_configuration_id)
            self.view['entry_cc_spacecraft'].set_text(old_spacecraft)
            self.spacecraft_correct = True

            self._clear_selections()

            pass

        tv = view['tv_cc_associated_opm']
        tv.set_model(self.model.get_associated_op_modes())
        tv.append_column(self.model.get_associated_op_modes_col())

        tv = view['tv_cc_not_associated_opm']
        tv.set_model(self.model.get_not_associated_op_modes())
        tv.append_column(self.model.get_not_associated_op_modes_col())

        view['entry_cc_configuration_id'].connect(
            'changed', self.on_entry_cc_configuration_id_changed)
	view['entry_cc_configuration_id'].connect(
	    "activate", self.on_activate_entry_effect) 

        view['entry_cc_spacecraft'].connect(
            'changed', self.on_entry_cc_spacecraft_changed)
	view['entry_cc_spacecraft'].connect(
	    "activate", self.on_activate_entry_effect) 


        # to control selection of associated op-modes
        sel = self.view['tv_cc_associated_opm'].get_selection()
        sel.set_mode(gtk.SELECTION_MULTIPLE)
        sel.connect('changed',
                    self.on_associated_opm_selection_changed)

        # to control selection of not associated op-modes
        sel = self.view['tv_cc_not_associated_opm'].get_selection()
        sel.set_mode(gtk.SELECTION_MULTIPLE)
        sel.connect('changed', 
                    self.on_not_associated_opm_selection_changed)

        cond = not self.model.get_old_configuration() is None            
        
        view['entry_cc_configuration_id'].set_sensitive(True)
        view['entry_cc_spacecraft'].set_sensitive(cond)
        view['tv_cc_associated_opm'].set_sensitive(cond)
        view['tv_cc_not_associated_opm'].set_sensitive(cond)
        view['button_cc_as_notas'].set_sensitive(False)
        view['button_cc_notas_as'].set_sensitive(False)
        view['button_cc_confirm'].set_sensitive(cond)

        return


    def register_adapters(self):
        return


    # -------------------------------------------------------------
    #         Signal handlers
    # -------------------------------------------------------------
    def _clear_selections(self):
        """Clears the old configurations selected."""
        sel = self.view['tv_cc_associated_opm'].get_selection()
        sel.unselect_all()
        sel = self.view['tv_cc_not_associated_opm'].get_selection()
        sel.unselect_all()
        return


    def on_button_cc_confirm_clicked(self, button):
        """Confirms the new name of the configuration selected in the list.
           @param   button:  The Confirm Button (self.view['button_cc_confirm']).
	   @type    button:  gtk.Button
        """
        configuration_id = \
            self.view['entry_cc_configuration_id'].get_text().decode('utf_8')
        spacecraft = \
            self.view['entry_cc_spacecraft'].get_text().decode('utf_8')

        """The associated opm list and not associated opm list are
        already saved respectively in self.model.associated_opm and
        in self.model.not_associated_opm."""
        self.model.instantiate(configuration_id, spacecraft)
        self.view['confirm_configuration'].response(gtk.RESPONSE_APPLY)
        self.view['confirm_configuration'].destroy()
        return


    def on_activate_entry_effect(self, editable):
        """Confirms the new name of the configuration selected in the list with the key ENTER.
           @param   editable:  The entry
	   @type    editable:  string
        """
	if (self.effect_valid1 or self.effect_valid2):
            # Simulate clicking of configuration
            self.on_button_cc_confirm_clicked(None)
        return


    def on_button_cc_discard_clicked(self, button):
        """Discards the new name of the configuration selected in the list.
           @param   button:  The Discard Button (self.view['button_cc_discard']).
	   @type    button:  gtk.Button
        """
        self.view['confirm_configuration'].response(gtk.RESPONSE_CANCEL)
        self.view['confirm_configuration'].destroy()
        return


    def on_button_cc_notas_as_clicked(self, button):
        """Transfers the configuration selected to the list of not associated op_modes.
           @param   button:  The Transfer Button (self.view['button_cc_notas_as']).
	   @type    button:  gtk.Button
        """
        sel = self.view['tv_cc_not_associated_opm'].get_selection()
        tree, paths = sel.get_selected_rows()
        self.model.notass_to_ass_op_modes_transfer(tree, paths)
        self._clear_selections()
        return


    def on_button_cc_as_notas_clicked(self, button):
        """Transfers the configuration selected to the list of associated op_modes.
           @param   button:  The Transfer Button (self.view['button_cc_as_notas']).
	   @type    button:  gtk.Button
        """
        sel = self.view['tv_cc_associated_opm'].get_selection()
        tree, paths = sel.get_selected_rows()
        self.model.ass_to_notass_op_modes_transfer(tree, paths)
        self._clear_selections()
        return


    def on_entry_cc_configuration_id_changed(self, entry):
        """Changes the new name of the configuration selected in the list.
           @param   entry:  The new entry name
	   @type    entry:  string
        """
        self._clear_selections()
        if len(entry.get_text()) == 0:
            self.view['entry_cc_spacecraft'].set_sensitive(False)
            self.view['tv_cc_associated_opm'].set_sensitive(False)
            self.view['tv_cc_not_associated_opm'].set_sensitive(False)
            self.view['button_cc_notas_as'].set_sensitive(False)
            self.view['button_cc_as_notas'].set_sensitive(False)
            self.view['button_cc_confirm'].set_sensitive(False)
            self.delay.change_color_and_balloon(entry, ["The configuration ID length must be at least 1"])
        else:
            try:
                entry.get_text().decode('utf_8')
            except UnicodeDecodeError:
                # It was not a ascii-encoded unicode string
                self.delay.change_color_and_balloon(
                    entry, ["Please use UTF-8 characters only"])
                #make sure the "Confirm" button is disabled
                self.view['entry_cc_spacecraft'].set_sensitive(False)
                self.view['tv_cc_associated_opm'].set_sensitive(False)
                self.view['tv_cc_not_associated_opm'].set_sensitive(False)
                self.view['button_cc_notas_as'].set_sensitive(False)
                self.view['button_cc_as_notas'].set_sensitive(False)
                self.view['button_cc_confirm'].set_sensitive(False)

            else:
                if self.model.mission_model.name_exist(
                    entry.get_text(),
                    self.model.get_old_configuration().get_configuration_id() \
                        if self.model.get_old_configuration() is not None else None
                    ):
                    
                    if entry.get_text().lower() == "none":
                        self.delay.change_color_and_balloon(entry, ["This name is invalid"])
                    else:
                        self.delay.change_color_and_balloon(entry, ["This name already exist"])
                    self.view['entry_cc_spacecraft'].set_sensitive(False)
                    self.view['tv_cc_associated_opm'].set_sensitive(False)
                    self.view['tv_cc_not_associated_opm'].set_sensitive(False)
                    self.view['button_cc_notas_as'].set_sensitive(False)
                    self.view['button_cc_as_notas'].set_sensitive(False)
                    self.view['button_cc_confirm'].set_sensitive(False)

                else:
                    self.view.change_color_and_balloon(entry, None)
                    self.view['entry_cc_spacecraft'].set_sensitive(True)
                    self.view['tv_cc_associated_opm'].set_sensitive(
                        self.spacecraft_correct)
                    self.view['tv_cc_not_associated_opm'].set_sensitive(
                        self.spacecraft_correct)
                    self.view['button_cc_notas_as'].set_sensitive(False)
                    self.view['button_cc_as_notas'].set_sensitive(False)
                    self.view['button_cc_confirm'].set_sensitive(
                        self.spacecraft_correct and 
                        self.model.get_associated_op_modes_counter() > 0)
		    if (self.spacecraft_correct and self.model.get_associated_op_modes_counter() > 0):
			self.effect_valid1 = True

        return


    def on_entry_cc_spacecraft_changed(self, entry):
        """Changes the new spacecraft of the configuration selected in the list.
           @param   entry:  The new entry spacecraft
	   @type    entry:  string
        """
        self._clear_selections()
        if len(entry.get_text()) == 0:
            self.spacecraft_correct = False
            self.view.change_color_and_balloon(entry, None)
            self.view['tv_cc_associated_opm'].set_sensitive(False)
            self.view['tv_cc_not_associated_opm'].set_sensitive(False)
            self.view['button_cc_notas_as'].set_sensitive(False)
            self.view['button_cc_as_notas'].set_sensitive(False)
            self.view['button_cc_confirm'].set_sensitive(False)
            self.view.change_color_and_balloon(
                entry, ["The spacecraft must have at least a condition"])
            return

        if len(entry.get_text()) > entry.get_width_chars():
            entry.set_width_chars(len(entry.get_text()))

        error_lst, impl_name = self._check_entry_syntax(entry)

        if not error_lst:
            self.spacecraft_correct = True
            self.view.change_color_and_balloon(entry, None)
            self.view['tv_cc_associated_opm'].set_sensitive(True)
            self.view['tv_cc_not_associated_opm'].set_sensitive(True)
            self.view['button_cc_notas_as'].set_sensitive(False)
            self.view['button_cc_as_notas'].set_sensitive(False)
            self.view['button_cc_confirm'].set_sensitive(
                self.model.get_associated_op_modes_counter() > 0)
	    if (self.spacecraft_correct and self.model.get_associated_op_modes_counter() > 0):
		self.effect_valid2 = True
        else:
            self.delay.change_color_and_balloon(entry, error_lst)
            self.spacecraft_correct = False
            self.view['tv_cc_associated_opm'].set_sensitive(False)
            self.view['tv_cc_not_associated_opm'].set_sensitive(False)
            self.view['button_cc_notas_as'].set_sensitive(False)
            self.view['button_cc_as_notas'].set_sensitive(False)
            self.view['button_cc_confirm'].set_sensitive(False)

        return
        

    def _check_entry_syntax(self, entry):
        """Checks the errors of the new entries of Configurations.
           @param   entry:  The new entry name
	   @type    entry:  string
        """
        contents = entry.get_text()
        
        error_lst, impl_name = functions.get_errors_in_atomic_proposition_with_templates(
            contents, self.model.mission_model.slim_model)
        return [l.__str__() for l in error_lst], impl_name


    def on_associated_opm_selection_changed(self, selection):
        """Changes the associated op_mode of the configuration selected.
           @param   selection:  The selection of associated op_mode.
	   @type    selection:  gtk.TreeSelection
        """
        self.view['button_cc_as_notas'].set_sensitive(
            selection.count_selected_rows() > 0)

        return


    def on_not_associated_opm_selection_changed(self, selection):
        """Changes the not associated op_mode of the configuration selected.
           @param   selection:  The selection of not associated op_mode.
	   @type    selection:  gtk.TreeSelection
        """
        self.view['button_cc_notas_as'].set_sensitive(
            selection.count_selected_rows() > 0)

        return
    

    # ---------------------------------------------------------------
    #        Notifications from observable properties
    # ---------------------------------------------------------------
    def property_associated_op_modes_counter_value_change(self, model, old, new):
        """When a change occured in the Configurations, a signal is emitted.
           @param   model:  The Configuration Model.
	   @type    model:  ConfigurationModel
           @param   old:  The number of old associated op_modes.
	   @type    old:  int
           @param   new:  The number of new associated op_modes.
	   @type    new:  int
        """
        if (new > 0 and old == 0) or (new == 0 and old > 0):
            self.view['button_cc_confirm'].set_sensitive(new > 0)
            pass
        return
        

pass # end of ConfigurationDialogCtrl class
# ----------------------------------------------------------------------
