# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


import gtk
from gtkmvc import Controller
import gobject

from src.ctrls.activity import ActivityCtrl, StartStopActivityCtrl
from src.models.filtered_properties import FilteredPropertiesModel
from src.models.performability import ToolSettingsModel
import options

class PerformabilityCtrl (ActivityCtrl):
    """Controller for the performability activity."""
    def __init__(self, model, view):
        ActivityCtrl.__init__(self, 
                              PerformabilitySubCtrl(self, model, view.subview),
                              model, view)

    def register_view(self, view):
        ActivityCtrl.register_view(self, view)

        tv = view['tv_properties']
        tv.set_model(self.model.filtered_props.get_properties_list())

        # selection
        def __rend_select(col, rend, model, iter):
            col_num = FilteredPropertiesModel.SELECT_SINGLE
            rend.set_property("active", model[iter][col_num])
            return
        rend1 = gtk.CellRendererToggle()
        rend1.set_property("activatable", True)
        rend1.connect("toggled", self.on_prop_toggled)
        col = gtk.TreeViewColumn('Name', rend1)
        col.set_cell_data_func(rend1, __rend_select)

        # and name
        rend2 = gtk.CellRendererText()
        col.pack_start(rend2)
        col.set_cell_data_func(rend2, lambda c,r,m,i: r.set_property('text', m[i][FilteredPropertiesModel.PROPERTY].name))
        tv.append_column(col)

        rend = gtk.CellRendererText()
        col = gtk.TreeViewColumn('Formula', rend)
        col.set_cell_data_func(rend, lambda c,r,m,i: r.set_property('text', m[i][FilteredPropertiesModel.PROPERTY].description))
        tv.append_column(col)

        # no selection for this tv:
        tv.get_selection().set_mode(gtk.SELECTION_NONE)
        return

    def on_prop_toggled(self, rend, path):
        child_path = self.model.filtered_props.get_properties_list().convert_path_to_child_path(path)
        self.model.props_model.toggle_property(child_path, multi=False)
        return

class PerformabilitySubCtrl(Controller):
    """Sub-controller for the correctness activity sub-view"""
    def __init__(self, parent, model, view):
        Controller.__init__(self, model, view)

        # chief controller
        self.parent = parent

        # sub-controllers (view is shared)
        self.model_imc = PerformabilityIMCCtrl(self, model.model_imc, view.imc)
        self.model_sim = PerformabilitySimulationCtrl(self, model.model_sim, view.sim)

        self.observe_model(self.model.slim_model)

    def register_view(self, view):
        return

    # ---------------------------------------------------------------
    #        Signal handlers
    # ---------------------------------------------------------------    
    def on_nb_performability_switch_page(self, nb, page, num): 
        pass

    # ---------------------------------------------------------------
    #        Notifications from observable properties
    # ---------------------------------------------------------------
    def property_infinite_domain_value_change(self, model, old, new):
        self.view.display_imc(new > 0 and self.model.slim_model.failures > 0)
    def property_failures_change(self, model, old, new):
        self.view.display_imc(new > 0 and self.model.slim_model.infinite_domain > 0)

class PerformabilitySimulationCtrl(StartStopActivityCtrl):
    def __init__(self, parent, model, view):
        StartStopActivityCtrl.__init__(self, model, view,
                                        self.start_simulation,
                                        self.stop_simulation)
        self.parent = parent
        # we also observe the number of selected properties in the
        # properties model, to enable performability button
        # accordingly
        self.observe_model(self.model.filtered_props)

        self.property_ok = False
        self.input_ok = False

    def register_view(self, view):
        StartStopActivityCtrl.register_view(self, view)

        # Put strategies into combo box
        combo_box = view['combo_strategy']
        combo_box.set_model(gtk.ListStore(gobject.TYPE_STRING))
        cell = gtk.CellRendererText()
        combo_box.pack_start(cell, True)
        combo_box.add_attribute(cell, 'text', 0)
        for strategy in ['asap']: #, 'local', 'progressive']:
            combo_box.append_text(strategy)
        combo_box.set_active(0)
        
        self.view['button_run'].set_sensitive(self.property_ok and self.input_ok)

        self.view['txtview_error_bound'].set_text(str(options.MC_ERRORBOUND))
        view['txtview_confidence'].set_text(str(options.MC_CONFIDENCE))

        self.model.reset()
        return

    def on_error_bound_changed(self, textview):
        self.check_input()

    def on_confidence_changed(self, textview):
        self.check_input()

    def on_strategy_changed(self, combobox):
        self.check_input()

    def check_input(self):
        self.input_ok = True

        textview = self.view['txtview_error_bound']
        error_bound = textview.get_text()
        try:
            fbound = float(error_bound)
            if fbound <= 0.0 or fbound >= 1.0:
                textview.modify_base(gtk.STATE_NORMAL, gtk.gdk.color_parse('red'))
                self.input_ok = False
            elif fbound <= 0.01 or fbound >= 0.20:
                textview.modify_base(gtk.STATE_NORMAL, gtk.gdk.color_parse('yellow'))
            else:
                textview.modify_base(gtk.STATE_NORMAL, gtk.gdk.color_parse('white'))
        except:
            textview.modify_base(gtk.STATE_NORMAL, gtk.gdk.color_parse('red'))
            ok = False

        textview = self.view['txtview_confidence']
        confidence = textview.get_text()
        try:
            fconfidence = float(confidence)
            if fconfidence <= 0.0 or fconfidence >= 1.0:
                textview.modify_base(gtk.STATE_NORMAL, gtk.gdk.color_parse('red'))
                self.input_ok = False
            elif fconfidence > 0.99 or fconfidence <= 0.50:
                textview.modify_base(gtk.STATE_NORMAL, gtk.gdk.color_parse('yellow'))
            else:
                textview.modify_base(gtk.STATE_NORMAL, gtk.gdk.color_parse('white'))
        except:
            textview.modify_base(gtk.STATE_NORMAL, gtk.gdk.color_parse('red'))
            self.input_ok = False

        if self.input_ok:
            self.model.confidence = fconfidence
            self.model.error_bound = fbound
        self.view['button_run'].set_sensitive(self.property_ok and self.input_ok)

    def start_simulation(self):
        gobject.idle_add(self.redraw);
        error_bound = float(self.view['txtview_error_bound'].get_text())
        confidence = float(self.view['txtview_confidence'].get_text())
        strategy = self.view['combo_strategy'].get_active_text()
        self.model.run_simulation(error_bound, confidence, strategy)
        gobject.idle_add(self.redraw, self.model.row);

    def stop_simulation(self):
        self.model.stop_simulation()

    def redraw(self, prop_row=None):
        if prop_row == None or prop_row[FilteredPropertiesModel.ANALYSIS_RESULT] is None or\
            prop_row[FilteredPropertiesModel.ANALYSIS_RESULT].tool != 'slimsim':
            self.view.reset_prob()
        else:
            res = prop_row[FilteredPropertiesModel.ANALYSIS_RESULT]
            self.view.display_prob(res.probability)

    # ---------------------------------------------------------------
    #        Notifications from observable properties
    # ---------------------------------------------------------------            
    def property_single_selected_value_change(self, model, old, new):
        self.model.reset()
        self.view.reset_prob()
        self.property_ok = False
        if new:
            self.property_ok = self.model.set_property()
            if self.property_ok:
                self.redraw(self.model.row)
        else:
            self.redraw()

        self.view['button_run'].set_sensitive(self.property_ok and self.input_ok)

class PerformabilityIMCCtrl(StartStopActivityCtrl):
    def __init__(self, parent, model, view):
        StartStopActivityCtrl.__init__(self, model, view,
                                        self.start_performability,
                                        self.stop_performability)
        self.parent = parent
        # we also observe the number of selected properties in the
        # properties model, to enable performability button
        # accordingly
        self.observe_model(self.model.filtered_props)

        self.tool_settings_ctrl = ToolSettingsCtrl(ToolSettingsModel(), view.tool_settings_sub)
        return

    def register_view(self, view):
        StartStopActivityCtrl.register_view(self, view)
        self.model.reset()
        return
    

    def redraw(self, prop_row=None):
        """Redraws the plotting area. If points is given, it is used instead
        of the internal state.
        """

        # If no result
        if prop_row == None or prop_row[FilteredPropertiesModel.ANALYSIS_RESULT] is None or \
            prop_row[FilteredPropertiesModel.ANALYSIS_RESULT].tool == 'slimsim':
            # set the x axis
            ltime_bound = self.model.get_lower_time_bound()
            utime_bound = self.model.get_upper_time_bound()
            # set the plotted line
            #xpoints = [ p[0] for p in self.model.points ]
            #ypoints = [ p[1] for p in self.model.points ]
            xpoints = []
            ypoints = []
        else:
            res = prop_row[FilteredPropertiesModel.ANALYSIS_RESULT]
            if res.has_stepwise_results():
                # Draw the resulting lines on the canvas.
                xpoints, ypoints = (res.probability_steps[0],[res.probability_steps[1]])
                                
                # set the x axis
                ltime_bound = self.model.get_lower_time_bound(prop_row[FilteredPropertiesModel.PROPERTY].property)
                utime_bound = self.model.get_upper_time_bound(prop_row[FilteredPropertiesModel.PROPERTY].property)
                
                for ypoints_ in ypoints:
                    self.view.draw(xpoints, ypoints_, ltime_bound, utime_bound)
        
                if ypoints != [] and ypoints[0] != []:
                    self.view.display_percentage(ypoints[0][len(ypoints[0]) - 1])
            elif res.is_et_analysis():
                value = res.expected_time[0] if res.expected_time[0] == res.expected_time[1] else res.expected_time
                self.view.display_et(value)
            elif res.is_lra_analysis():
                value = res.lra[0] if res.lra[0] == res.lra[1] else res.lra
                self.view.display_lra(value)
            else:
                raise "This should be unreachable"
        return

    def start_performability(self):
#        self.model.busy += 1
        gobject.idle_add(self.redraw);
        self.model.run_performability(self.tool_settings_ctrl.model)
        gobject.idle_add(self.redraw, self.model.row);
#        self.model.busy -= 1
        return

    def stop_performability(self):
        self.model.stop_performability()
        return

    # ---------------------------------------------------------------
    #        Notifications from observable properties
    # ---------------------------------------------------------------            
    def property_single_selected_value_change(self, model, old, new):
        self.view['button_run'].set_sensitive(new and not
                                              self.model.infinite_model)
        self.model.reset()
        self.view.reset_labels()
        if new:
            self.model.set_property()
            self.redraw(self.model.row)
        else:
            self.redraw()
        return

class ToolSettingsCtrl(Controller):
    def __init__(self, model, view):
        Controller.__init__(self, model, view)

    def register_view(self, view):
        view['txtview_error_bound'].connect('changed', self.on_error_bound_changed)
        view['combo_tool'].connect('changed', self.on_tool_changed)

        # Put model.valid_tools into combo box
        combo_box = view['combo_tool']
        combo_box.set_model(gtk.ListStore(gobject.TYPE_STRING))
        cell = gtk.CellRendererText()
        combo_box.pack_start(cell, True)
        combo_box.add_attribute(cell, 'text', 0)
        for tool in self.model.valid_tools:
            combo_box.append_text(tool)
        combo_box.set_active(0)

        # Init error bound
        view['txtview_error_bound'].set_text(str(self.model.selected_error_bound))

    def register_adapters(self):
        pass

    def on_error_bound_changed(self, textview):
        error_bound = textview.get_text()
        if self.model.is_error_bound_suitable(error_bound):
            self.model.set_error_bound(error_bound)
            if self.model.get_error_bound() >= 1:
                self.view.tint_error_base("yellow")
            else:
                self.view.tint_error_base("white")
        else:
            self.view.tint_error_base("red")

    def on_tool_changed(self, combobox):
        selection = combobox.get_active_text()
        self.model.set_selected_tool(selection);
        if selection == "MRMC":
            self.view.hide_error_bound(True)
        else:
            self.view.hide_error_bound(False)
