# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


import os
import os.path
import gtk
import pango
import gobject
import copy
from gtkmvc import Controller
from src.ctrls.activity import ActivityCtrl, StartStopActivityCtrl
from hasdel_tools.src.generator import generate_unique_tfpg_xml_basename, \
generate_unique_tfpg_human_readable_basename
import hasdel_tools.src.xml.tfpg as tfpg
from delayedcall import DelayedCall
from tfpg_viewer.src.views.tfpg_view import TFPGView
from hasdel_tools.src.xml.tfpg import tfpg_doc_to_tfpg_internal_struct
from hasdel_tools.src.tfpg.tfpg_errors import TimeNotCorrectError, NodeLoopError, \
NodeNotInEdge, FMNodeIsNotDestNode, NotExistingModes, FMNotCorrectError, \
NotExistingNode
from hasdel_tools.src.functions import get_errors_in_atomic_proposition_with_templates
from matplotlib.backends.backend_gtkagg import NavigationToolbar2GTKAgg
import hasdel_tools.src.functions as functions

from src.ctrls.trace import TraceCtrl
from src.views.trace import TraceView
from src.models.trace import TraceViewModel
 
import options


class TfpgCtrl(ActivityCtrl):
    """Represents the Controller of Tfpg."""

    def __init__(self, model, view):
        """Constructor Method"""
        ActivityCtrl.__init__(self, 
                              TfpgSubCtrl(self, model, view.subview), 
                              model, view)

        from src.ctrls.application import StatusBarCtrl
        self.statusbar = StatusBarCtrl(model, view)
        
        # Sub-controllers:
        # (ParentModel -> TfpgModel
        # SubModels -> TfpgViewerModel, TfpgSlimAssocModel, TfpgSynthesisModel, TfpgBehavValModel, TfpgEffectValModel, TfpgErrorsModel
        # ParentView -> TfpgView
        # SubViews -> TfpgViewerView, TfpgSlimAssocView, TfpgSynthesisView, TfpgBehavValView, TfpgEffectValView, TfpgErrorsView)
        
        self.tfpg_viewer_ctrl = TfpgViewerCtrl(model, model.tfpg_viewer_model, view, view.tfpg_viewer_view)
        self.tfpg_slim_assoc_ctrl = TfpgSlimAssocCtrl(model, model.tfpg_slim_assoc_model, view, view.tfpg_slim_assoc_view)
        self.tfpg_synthesis_ctrl = TfpgSynthesisCtrl(self, model, model.tfpg_synthesis_model, view, view.tfpg_synthesis_view)
        self.tfpg_behav_val_ctrl = TfpgBehavValCtrl(model, model.tfpg_behav_val_model, view, view.tfpg_behav_val_view)
        self.tfpg_effect_val_ctrl = TfpgEffectValCtrl(model, model.tfpg_effect_val_model, view, view.tfpg_effect_val_view)
        self.tfpg_errors_ctrl = TfpgErrorsCtrl(model, model.tfpg_errors_model, view, view.tfpg_errors_view)
        
        # The self.tfpg_view_from_viewer is from tools.tfpg_viewer.src.views.tfpg_view
        self.tfpg_view_from_viewer = None
        
        # The self.curr_tfpg_name is the current name of the loaded tfpg
        self.curr_tfpg_name = None
        
        # The self.hr_tfpg is the human_readable filename related to the xml file with the same name in the same dir
        self.hr_tfpg_fn = ""
        
        # The self.tmp_tfpg_name is the temporarly human_readable tfpg name related to the file edited
        self.tmp_tfpg_name = ""
        
        # The self.tfpg_not_saved is the list of tfpg edited and not saved
        self.tfpg_not_saved = []

        self.observe_model(model.slim_model)
        return

   
    def register_view(self, view):
        """Sets the windows, dialog, buttons and other parts of View."""
        ActivityCtrl.register_view(self, view)
        
        # The column of the list of tfpgs
        tv = self.view['tv_tfpg_list']
        tv.set_model(self.model.get_tfpgs_list())
        tv.append_column(self.model.get_tfpgs_col())

        # To control selection of tfpgs
        sel = self.view['tv_tfpg_list'].get_selection()
        sel.set_mode(gtk.SELECTION_BROWSE)
        sel.connect('changed', self.on_tfpgs_selection_changed)

        # At first all buttons are disabled.
        self.view.disable_buttons()
        return


    def save_tfpg(self):
        """Saves a tfpg document in .txml format."""    
        
        dialog = gtk.FileChooserDialog("Save TFPG",
                                       None,
                                       gtk.FILE_CHOOSER_ACTION_SAVE,
                                       (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                                       gtk.STOCK_SAVE, gtk.RESPONSE_OK))
        dialog.set_default_response(gtk.RESPONSE_OK)
        filter = gtk.FileFilter()
        filter.set_name("FAME TFPG")
        filter.add_pattern("*.txml")
        dialog.add_filter(filter)
        

        #LB## If more tfpgs are edited and not saved
        if self.get_tfpg_not_saved():
        
            tfpg_name = self.get_tfpg_not_saved()[0]
            tfpg_filename = self.get_tfpg_not_saved()[1]
            tfpg_structural = self.get_tfpg_not_saved()[3]
                
                
            if tfpg_filename != "":
                file_name, file_extension = os.path.splitext(tfpg_filename)   
                dialog.set_current_name(os.path.basename(file_name + ".txml"))
            else:
                dialog.set_current_name(os.path.basename(generate_unique_tfpg_xml_basename() + ".txml"))
                      
            dialog.set_do_overwrite_confirmation(True)    
            dialog.show()

            response = dialog.run()     
            name = dialog.get_filename()
                           
            if response == gtk.RESPONSE_OK:
                            
               file_name, file_extension = os.path.splitext(name)   
               name_txml = file_name + ".txml"
               tfpg_structural.dump_to_xml(name_txml)
               pass
                   
            dialog.destroy()
         
        #LB## If a selected tfpg not edited is saved
        else:
            
            dialog.set_current_name(os.path.basename(generate_unique_tfpg_xml_basename() + ".txml"))       
            dialog.set_do_overwrite_confirmation(True)    
            dialog.show()
            
            sel = self.view['tv_tfpg_list'].get_selection()
            tree_model, paths = sel.get_selected_rows()
         
            if paths:
                tfpg_structural = tree_model.get_value(tree_model.get_iter(paths[0]), 0)

            response = dialog.run()     
            name = dialog.get_filename()
               
            if response == gtk.RESPONSE_OK:
                
                file_name, file_extension = os.path.splitext(name)
                name_txml = file_name + ".txml"
                tfpg_structural.dump_to_xml(name_txml)

                #LB## When a tfpg is saved, the 3rd field of 'tfpgs_name_fn_saved_list' become True
                ##### and both human readable and xml files are saved.
                for tfpg in self.model.get_tfpgs_name_fn_saved_list():
                    if tfpg[0] == tfpg_structural.get_name() and tfpg[2] is False:
                        tfpg[2] = True
                pass
            
            dialog.destroy()
        return


    def load_tfpg(self):  
        """Loads a normal or a synthesized tfpg in .txml format."""
        chooser = gtk.FileChooserDialog(title="Open TFPG",
                                        action=gtk.FILE_CHOOSER_ACTION_OPEN,
                                        buttons=(gtk.STOCK_CANCEL,gtk.RESPONSE_CANCEL,
                                                 gtk.STOCK_OPEN,
                                                 gtk.RESPONSE_OK))
        chooser.set_select_multiple(False)
        filter = gtk.FileFilter()
        filter.set_name("FAME TFPG")
        filter.add_pattern("*.txml")
        chooser.add_filter(filter)
        chooser.set_current_folder(os.path.dirname(
            generate_unique_tfpg_xml_basename() + ".txml"))

        response = chooser.run()
        if response == gtk.RESPONSE_OK:
            res = self.model.get_tfpg_doc(chooser.get_filename())

            if res is not None:
                        
                tfpg_structural = tfpg_doc_to_tfpg_internal_struct(res)
                self.curr_tfpg_name = tfpg_structural.get_name()

                self.tfpg_viewer_ctrl.curr_tfpg_name = self.curr_tfpg_name
                self.tfpg_slim_assoc_ctrl.curr_tfpg_name = self.curr_tfpg_name

                self.tfpg_view_from_viewer = TFPGView()
                self.tfpg_viewer_ctrl.tfpg_view_from_viewer = self.tfpg_view_from_viewer
                fig = self.tfpg_view_from_viewer.create_tfpg_canvas_figure(tfpg_graph=tfpg_structural)
                fig.mpl_connect('motion_notify_event', self.tfpg_viewer_ctrl.onmotion)
                fig.mpl_connect('button_press_event', self.tfpg_viewer_ctrl.onclick)

                toolbar = NavigationToolbar(fig,None)
                
                self.view.tfpg_viewer_view.clear()
                self.view.tfpg_viewer_view['vbox_toolbar_viewer'].add(toolbar)
                self.view.tfpg_viewer_view['vbox_view'].add(fig)
                self.view.tfpg_viewer_view['vbox_view'].show_all()

                self.view.enable_buttons()
                self.view.tfpg_slim_assoc_view['button_save_associations'].set_sensitive(False)

                row, column = self.view['tv_tfpg_list'].get_cursor()
                if row != None:
                    row_nbr = row[0] + 1
                else:
                    row_nbr = 0
                self.view['tv_tfpg_list'].set_cursor(row_nbr)

                # Finally, it is required that the tfpg model tab is initialized
                associations = self.tfpg_slim_assoc_ctrl.initialize_tfpg_model_tab(
                            tfpg_structural.get_fm_nodes(),
                            tfpg_structural.get_monitored_discrepancies(),
                            tfpg_structural.get_unmonitored_discrepancies(),
                            tfpg_structural.get_modes_list())

                self.model.tfpg_slim_assoc_model.add_associations_list(self.curr_tfpg_name, associations)

                # The focus is on the Viewer Window (0).
                # When there is an error the page switches to the Errors Window (5).
                if self.view.get_active_page() in [0,1,2,3,4,5]: 
                
                    if self.model.get_errors_counter() >= 1:
                        self.view.disable_buttons()
                        self.view.clear_subviews()
                        self.tfpg_errors_ctrl.show_error_message("Error: \n\n" + self.model.error_name + "\n")
                        self.view.show_page(5)
                    else:
                        self.view.show_page(0)
                    
                chooser.destroy()
                self.view.tfpg_slim_assoc_view['button_load_associations'].set_sensitive(True)

            else:

                # The focus is on the Viewer Window (0).
                # When there is an error the page switches to the Errors Window (5).
                if self.view.get_active_page() in [0,1,2,3,4,5]: 
                
                    if self.model.get_errors_counter() >= 1:
                        self.view.disable_buttons()
                        self.view.clear_subviews()
                        self.tfpg_errors_ctrl.show_error_message("Error: \n\n" + self.model.error_name + "\n")
                        self.view.show_page(5)
                    else:
                        self.view.show_page(0)
                        
                chooser.destroy()
                self.view.tfpg_slim_assoc_view['button_load_associations'].set_sensitive(True)
            
        else:
            # response = gtk.RESPONSE_CANCEL
            chooser.destroy()

        return


    def on_button_new_tfpg_clicked(self, action):
        """Creates a tfpg in human readable (.tfpg) format.
                @param   action:  The action to enable the button self.view['button_new_tfpg']
                @type    action:  gtk.Button
        """
        dialog = gtk.FileChooserDialog("New TFPG",
                                       None,
                                       gtk.FILE_CHOOSER_ACTION_SAVE,
                                       (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                                       gtk.STOCK_OPEN, gtk.RESPONSE_OK))
        dialog.set_default_response(gtk.RESPONSE_OK)
        dialog.set_do_overwrite_confirmation(True)

        filter = gtk.FileFilter()
        filter.set_name("FAME TFPG")
        filter.add_pattern("*.tfpg")
        dialog.add_filter(filter)
        dialog.set_current_name(os.path.basename(
            generate_unique_tfpg_human_readable_basename() + ".tfpg"))
        response = dialog.run()        
        
        name = dialog.get_filename()
        if response == gtk.RESPONSE_OK:
            file_name, file_extension = os.path.splitext(name)
            name_tfpg = file_name + ".tfpg"
            name_txml = file_name + ".txml"

        dialog.destroy()
        
        if name_tfpg:
            self.on_button_edit_tfpg_clicked(action, modified_filename=name_tfpg, from_scratch=True)
            self.model.tfpgs_counter += 1
            self.model.tfpgs_list_changed.emit()
            
            tfpg_structural = self.model.parse_concrete_syntax_file(name_tfpg)
            #LB## When a tfpg is edited, the 3rd field of 'tfpgs_name_fn_saved_list' become False
            ##### until the tfpg is saved.
            self.model.get_tfpgs_name_fn_saved_list().append([self.curr_tfpg_name, name_tfpg, False, tfpg_structural, name_txml])
            
        return


    def on_button_edit_tfpg_clicked(self, action, modified_filename=None, from_scratch=False):
        """Function called when the user wants to edit, using a textual editor,
            the visualized structural.
            The second parameter (modified_filename) is used in case the user
            has saved some modifications, but there was a parsing error. In this
            case, an error dialog is shown and the text editor is re-opened
            loading data from modified_filename.
            The third parameter, from_scratch, must be set to True in case we want to create
            a new tfpg from scratch.
        
            @param   action:  The action to enable the button self.view['button_edit_tfpg']
            @type    action:  gtk.Button
            @param   modified_filename: the temporal filename of the unsuccessfully modified file
            @type    modified_filename: string
            @param   from_scratch: in case we want to create a new tfpg
            @type    from_scratch: boolean
        """

        if modified_filename == None or from_scratch is True:
            dialog = gtk.MessageDialog(self.view['window_top'], 0, gtk.MESSAGE_INFO,
                                       gtk.BUTTONS_OK,
                                       "Save and close the external editor to see your changes")
            response = dialog.run()
            dialog.destroy()

        sel = self.view['tv_tfpg_list'].get_selection()
        tree_model, paths = sel.get_selected_rows()
        
        if paths:
            tfpg_structural = tree_model.get_value(tree_model.get_iter(paths[0]), 0)

        
        if modified_filename == None:

            #LB## Checks if there is a tfpg name 
            for tfpg in self.model.get_tfpgs_name_fn_saved_list():
                if tfpg[0] == tfpg_structural.get_name():
                    self.tmp_tfpg_name = tfpg[0]
                    self.hr_tfpg_fn = tfpg[1]
            
            if self.hr_tfpg_fn != "":
                tmp_filename = self.hr_tfpg_fn
                tfpg_structural_loaded = self.model.parse_concrete_syntax_file(tmp_filename)

                #LB## Checks if the tfpg_structural of the human readable file is equal to the tfpg of xml file.
                if not tfpg_structural_loaded.__eq__(tfpg_structural):
                
                    parent_name = 'window_top'

                    error = "The human readable file: '' " + self.curr_tfpg_name + " '' does not match with the loaded xml file." + \
                            " If you click on OK, the human readable file will be overwritten and the xml will be edited!"
                    response = self.view.ui_show_message_dialog("error", str(error), "", parent_name)

                    if response == gtk.RESPONSE_CANCEL: 
                        return
                    else:      
                        tfpg_structural.dump_to_concrete_syntax_file(tmp_filename)    
                        return self.on_button_edit_tfpg_clicked(action, tmp_filename)

            else:

                #LB## If there is not a human readable file related to the loaded xml file,
                ##### a human readable filewith the same name of the xml file is created
                ##### and saved/edited/updated automatically in the same dir.
                for tfpg in self.model.get_tfpgs_name_fn_saved_list():
                    if tfpg[1] == "":
                        txml_filename = tfpg[4]
                        file_name, file_extension = os.path.splitext(txml_filename)
                        tmp_filename = file_name + ".tfpg"
                        tfpg[1] == tmp_filename
                        tfpg_structural.dump_to_concrete_syntax_file(tmp_filename)
                        break
        
        else:
            tmp_filename = modified_filename
               

        ret_code = self.view.ui_run_text_editor(tmp_filename, from_scratch=from_scratch)

        if ret_code == 0:

            try:
                # Parsing the file without building a TfpgDocument
                tfpg_structural = self.model.parse_concrete_syntax_file(tmp_filename)
                
                self.tfpg_view_from_viewer = TFPGView()
                self.tfpg_viewer_ctrl.tfpg_view_from_viewer = self.tfpg_view_from_viewer
                fig = self.tfpg_view_from_viewer.create_tfpg_canvas_figure(tfpg_graph=tfpg_structural)
                fig.mpl_connect('motion_notify_event', self.tfpg_viewer_ctrl.onmotion)
                fig.mpl_connect('button_press_event', self.tfpg_viewer_ctrl.onclick)

                # If the user modifies a tfpg, the old one must be remove from the
                # list on the left and a new one must be added
                # This operation must not be done in case a new tfpg is created
                if from_scratch is False:
                   self.model.remove_tfpgs(tree_model, paths, "Edit")

                self.model.get_tfpgs_list().append([tfpg_structural])

                self.view.tfpg_viewer_view.clear_tfpg_graph()
                self.view.tfpg_viewer_view['vbox_view'].add(fig)
                self.view.tfpg_viewer_view['vbox_view'].show_all()

                # As the new tfpg is added at the end of a list, the selected
                # tfpg must be the last one
                self.view['tv_tfpg_list'].set_cursor(len(self.model.get_tfpgs_list())-1)
                
                self.view.enable_buttons()
                self.view.tfpg_slim_assoc_view['button_save_associations'].set_sensitive(False)

                for tfpg in self.model.get_tfpgs_name_fn_saved_list():
                    if tfpg[0] == self.tmp_tfpg_name:
                        tfpg[0] = tfpg_structural.get_name()
                        tfpg[2] = False
                        tfpg[3] = tfpg_structural
                
            except Exception as ex:
                parent_name = 'window_top'
                response = self.view.ui_show_message_dialog("error", str(ex), "", parent_name)

                if response == gtk.RESPONSE_CANCEL: 
                    
                    message = "Are you sure do you want to exit? All changes will be discarded."
                    response = self.view.ui_show_message_dialog("warning", 
                                message, "", parent_name, buttons=gtk.BUTTONS_YES_NO)

                    if response == gtk.RESPONSE_NO: 
                        return self.on_button_edit_tfpg_clicked(action, tmp_filename)

                else:                
                    return self.on_button_edit_tfpg_clicked(action, tmp_filename)

        else:
            raise Exception("Error while using the external editor")

        return


    def on_button_remove_tfpg_clicked(self, button):
        """Removes the name of the tfpg selected in the list.
            @param   button:  The Remove Button: self.view['button_edit_tfpg'].
            @type    button:  gtk.Button
        """
        sel = self.view['tv_tfpg_list'].get_selection()
        tree_model, paths = sel.get_selected_rows()
        
        if paths:
            tfpg_structural = tree_model.get_value(tree_model.get_iter(paths[0]), 0)
            
            for tfpg_not_saved in self.model.get_tfpgs_name_fn_saved_list():

                if tfpg_not_saved[0] == tfpg_structural.get_name() and tfpg_not_saved[2] is False:
                
                    parent_name = 'window_top'
                    save_error = "Warning: the TFPG '' " + tfpg_not_saved[0] + " '' was not saved. Please click on OK to save the tfpg, " + \
                            "otherwise all changes will be discarded!"
                    save_response = self.view.ui_show_message_dialog("error", save_error, "", parent_name)

                    self.model.get_tfpgs_name_fn_saved_list().remove(tfpg_not_saved)
                    
                    self.tfpg_not_saved = tfpg_not_saved
                    
                    if save_response == gtk.RESPONSE_CANCEL: 
                        message = "Are you sure do you want to exit? All changes will be discarded."
                        response = self.view.ui_show_message_dialog("warning", 
                                            message, "", parent_name, buttons=gtk.BUTTONS_YES_NO)

                        if response == gtk.RESPONSE_NO: 
                            self.save_tfpg()

                    else:                
                        self.save_tfpg()
                        
                        
            # Checks if the slim_assoc file is saved or not
            for slim_assoc_not_saved in self.tfpg_slim_assoc_ctrl.get_slim_assoc_not_saved():
            
                if slim_assoc_not_saved[0] == tfpg_structural.get_name() and slim_assoc_not_saved[2] is False:
                    parent_name = 'window_top'
                    save_error = "Warning: the slim associations related to the TFPG '' " + slim_assoc_not_saved[0] + " '' were not saved. " + \
                                "Please click on OK to save the slim associations, otherwise all changes will be discarded!"
                    save_response = self.view.ui_show_message_dialog("error", save_error, "", parent_name)
                    
                    self.tfpg_slim_assoc_ctrl.get_slim_assoc_not_saved().remove(slim_assoc_not_saved)
                    
                    self.tfpg_slim_assoc_ctrl.curr_slim_assoc_not_saved = slim_assoc_not_saved
                    
                    if save_response == gtk.RESPONSE_CANCEL: 
                        message = "Are you sure do you want to exit? All changes will be discarded."
                        response = self.view.ui_show_message_dialog("warning", 
                                            message, "", parent_name, buttons=gtk.BUTTONS_YES_NO)

                        if response == gtk.RESPONSE_NO: 
                            self.tfpg_slim_assoc_ctrl.save_slim_assoc()

                    else:                
                        self.tfpg_slim_assoc_ctrl.save_slim_assoc()
        
            
        self.model.remove_tfpgs(tree_model, paths, "Remove")

        self.tfpg_view_from_viewer = None
        self.tfpg_viewer_ctrl.tfpg_view_from_viewer = self.tfpg_view_from_viewer
        self.view.clear_subviews()
        self.view.disable_buttons()
        self.view.tfpg_slim_assoc_view['button_load_associations'].set_sensitive(True)
        self._clear_selections()

        return


    def on_button_export_tfpg_clicked(self, button):
        """Function calle dwhen the users clicks the 'Export' button.
            The mean of the 'Export' button is to let the user to
            export the tfpg in .dot format, so that it allows a 
            better "HD representation"
        """
        dialog = gtk.FileChooserDialog("Export TFPG",
                                       None,
                                       gtk.FILE_CHOOSER_ACTION_SAVE,
                                       (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                                       gtk.STOCK_SAVE, gtk.RESPONSE_OK))
        dialog.set_default_response(gtk.RESPONSE_OK)

        png_filter = gtk.FileFilter()
        png_filter.set_name("PNG")
        png_filter.add_pattern("*.png")
        dialog.add_filter(png_filter)

        pdf_filter = gtk.FileFilter()
        pdf_filter.set_name("PDF")
        pdf_filter.add_pattern("*.pdf")
        dialog.add_filter(pdf_filter)

        dot_filter = gtk.FileFilter()
        dot_filter.set_name("DOT")
        dot_filter.add_pattern("*.dot")
        dialog.add_filter(dot_filter)

        dialog.set_do_overwrite_confirmation(True)    
        dialog.show()

        response = dialog.run()

        file_format = None

        if response == gtk.RESPONSE_OK:
            filename = dialog.get_filename()
            selected_filter = dialog.get_filter()
            file_format = selected_filter.get_name()
            
        else: 
            dialog.destroy()
            return
            
        dialog.destroy()
        

        if file_format == "PNG":
            self.tfpg_view_from_viewer.export_to_file(filename, 'png')
        elif file_format == "PDF":
            self.tfpg_view_from_viewer.export_to_file(filename, 'pdf')
        else:
            # default: .dot format
            self.tfpg_view_from_viewer.export_to_file(filename, 'dot')
        return


    def load_tfpg_synthesized(self, tfpg_filename, associations):  
        """Loads a synthesized tfpg in .txml format.

           @param tfpg_filename: the name of the synthesized tfpg
           @type  tfpg_filename: string

           @param associations: the list of associations which led to the synthesis
           @type associations: list
        """
        
        if self.tfpg_synthesis_ctrl.tfpg_synthesized_loaded is True:
        
            res = self.model.get_tfpg_doc(tfpg_filename)

            if res is not None:

                tfpg_structural = tfpg_doc_to_tfpg_internal_struct(res)
                
                self.curr_tfpg_name = tfpg_structural.get_name()
                
                self.tfpg_viewer_ctrl.curr_tfpg_name = self.curr_tfpg_name
                self.tfpg_slim_assoc_ctrl.curr_tfpg_name = self.curr_tfpg_name
                self.tfpg_synthesis_ctrl.curr_tfpg_name = self.curr_tfpg_name

                self.tfpg_view_from_viewer = TFPGView()
                self.tfpg_viewer_ctrl.tfpg_view_from_viewer = self.tfpg_view_from_viewer

                fig = self.tfpg_view_from_viewer.create_tfpg_canvas_figure(tfpg_graph=tfpg_structural)
                fig.mpl_connect('motion_notify_event', self.tfpg_viewer_ctrl.onmotion)
                fig.mpl_connect('button_press_event', self.tfpg_viewer_ctrl.onclick)

                toolbar = NavigationToolbar(fig,None)
                
                self.view.tfpg_viewer_view.clear()
                self.view.tfpg_viewer_view['vbox_toolbar_viewer'].add(toolbar)
                self.view.tfpg_viewer_view['vbox_view'].add(fig)
                self.view.tfpg_viewer_view['vbox_view'].show_all()

                self.view.enable_buttons()
                self.view.tfpg_slim_assoc_view['button_save_associations'].set_sensitive(False)


                # Automatically Loading of SLIM Associations in TFPG SLIM Associations Tab
                # Get the SLIM associations from the TFPG Synthesis Tab of the GUI
               
                # Before loading the associations, the slim associations tab must be initialised
                self.tfpg_slim_assoc_ctrl.initialize_tfpg_model_tab(
                    tfpg_structural.get_fm_nodes(),
                    tfpg_structural.get_monitored_discrepancies(),
                    tfpg_structural.get_unmonitored_discrepancies(),
                    tfpg_structural.get_modes_list())                
                self.tfpg_slim_assoc_ctrl.load_associations_list(associations)
                self.model.tfpg_slim_assoc_model.add_associations_list(self.curr_tfpg_name, associations)
                
               
        return


    def on_tfpgs_selection_changed(self, selection):
        """The selection of tfpgs has been changed. If one
            or more tfpgs are selected, button "Remove" must be 
            sensitive.
                @param   selection:  The Selection of the TFPGs List
                @type    selection:  gtk.TreeSelection
            """

        tree_model, paths = selection.get_selected_rows()

        if paths:
            tfpg_structural = tree_model.get_value(tree_model.get_iter(paths[0]), 0)
            self.curr_tfpg_name = tfpg_structural.get_name()
            self.tfpg_viewer_ctrl.curr_tfpg_name = self.curr_tfpg_name
            self.tfpg_slim_assoc_ctrl.curr_tfpg_name = self.curr_tfpg_name


            self.tfpg_view_from_viewer = TFPGView()
            self.tfpg_viewer_ctrl.tfpg_view_from_viewer = self.tfpg_view_from_viewer
            fig = self.tfpg_view_from_viewer.create_tfpg_canvas_figure(tfpg_structural)
            fig.mpl_connect('motion_notify_event', self.tfpg_viewer_ctrl.onmotion)
            fig.mpl_connect('button_press_event', self.tfpg_viewer_ctrl.onclick)

            toolbar = NavigationToolbar(fig,None)
            self.view.tfpg_viewer_view.clear()
            self.view.tfpg_viewer_view['vbox_toolbar_viewer'].add(toolbar)
            self.view.tfpg_viewer_view['vbox_view'].add(fig)
            self.view.tfpg_viewer_view['vbox_view'].show_all()
                
            self.view.show_page(0)

            self.view['button_remove_tfpg'].set_sensitive(
                selection.count_selected_rows() > 0)

            _, paths = selection.get_selected_rows()
            self.view['button_edit_tfpg'].set_sensitive(
                selection.count_selected_rows() == 1)

            self.view.enable_buttons()
            self.view.tfpg_slim_assoc_view['button_save_associations'].set_sensitive(False)

            ## Also the tfpg model tab must be updated
            self.tfpg_slim_assoc_ctrl.initialize_tfpg_model_tab(
                tfpg_structural.get_fm_nodes(),
                tfpg_structural.get_monitored_discrepancies(),
                tfpg_structural.get_unmonitored_discrepancies(),
                            tfpg_structural.get_modes_list())
            associations = self.model.tfpg_slim_assoc_model.get_slim_associations(tfpg_structural.get_name())
            if associations != None: self.tfpg_slim_assoc_ctrl.load_associations_list(associations)

            self.model.clear()

            self.view.tfpg_slim_assoc_view['button_load_associations'].set_sensitive(True)
            
        return


    def get_curr_tfpg_name(self):
        """Gets the current name of the loaded Tfpg.
            @return   self.curr_tfpg_name:  The current name of the loaded Tfpg.
            @rtype    self.curr_tfpg_name:  string
        """     
        return self.curr_tfpg_name
        

    def _clear_selections(self):
        """Clears the Tfpgs List."""
        sel = self.view['tv_tfpg_list'].get_selection()
        sel.unselect_all()
        return
        
        
    def tfpgs_not_saved(self):
        """After editing a tfpg, if the user quit the GUI,
            removes/modifies slim models, fault injections or the mission
            elements, a warning message appears asking the user what to do:
            to save the tfpgs modified and not saved or not.
            """
            
        for tfpg_not_saved in self.model.get_tfpgs_name_fn_saved_list():
            
            if tfpg_not_saved[2] is False:
                parent_name = 'window_top'
                save_error = "Warning: the TFPG '' " + tfpg_not_saved[0] + " '' was not saved. Please click on OK to save the tfpg, " + \
                        "otherwise all changes will be discarded!"
                save_response = self.view.ui_show_message_dialog("error", save_error, "", parent_name)
                self.tfpg_not_saved = tfpg_not_saved
                    
                if save_response == gtk.RESPONSE_CANCEL: 
                    message = "Are you sure do you want to exit? All changes will be discarded."
                    response = self.view.ui_show_message_dialog("warning", 
                                        message, "", parent_name, buttons=gtk.BUTTONS_YES_NO)

                    if response == gtk.RESPONSE_NO: 
                        self.save_tfpg()

                else:                
                    self.save_tfpg()
                    
        return
        
    def get_tfpg_not_saved(self):
        """Gets the list of edited/updated Tfpgs.
            @return   self.tfpg_not_saved:  The list of edited/updated Tfpgs.
            @rtype    self.tfpg_not_saved:  list of [tfpg_name, human_readable_filename, False, tfpg_structural, txml_filename]
        """     
        return self.tfpg_not_saved
        
             
        
    # ---------------------------------------------------------------
    #        Signal handlers
    # ---------------------------------------------------------------  
    def property_changed_signal_emit(self, model, args):
        """When a change occured in the tfpg, a signal clears the tabs and
            disables the buttons.
            @param   model:  The Slim Model.
            @type    model:  SlimModel
            @param   args:  The arguments of SlimModel.
            @type    args:  string
        """    
            
        self.model.clear_tfpgs_list()
        self.tfpg_slim_assoc_ctrl.clear_slim_assoc_list()
        self.view.tfpg_viewer_view.clear()
        self.view.disable_buttons()
        self.view.tfpg_slim_assoc_view['button_load_associations'].set_sensitive(True)     
        return
        
        
        
    def property_tfpgs_counter_value_change(self, model, old, new):
        """When a change occured in the tfpgs, a signal enables the Effectiveness Validation or not.
            @param   model:  The Tfpg Model.
            @type    model:  TfpgModel
            @param   old:  The number of old tfpgs.
            @type    old:  int
            @param   new:  The number of new tfpgs.
            @type    new:  int
        """
        if (new > 0 and old == 0) or (new == 0 and old > 0):
            self.view.tfpg_effect_val_view.enable_effect_val(new > 0 and
                                            model.get_tfpgs_counter() > 0)
            self.view.tfpg_behav_val_view.enable_behav_val(new > 0 and
                                            model.get_tfpgs_counter() > 0)
            pass
        return


    def on_nb_tfpg_switch_page(self, nb, page, page_num):
        """The user selected a different TFPG tab.
            @param   nb:  The notebook self.view[nb_tfpg].
            @type    nb:  gtk.Notebook
            @param   page:  The pointer to the current view.
            @type    page:  gpointer
            @param   page_num:  The current number of tfpg tab.
            @type    page_num:  int
        """
        
        # The ballons with the error meassages in slim associations 
        # are hidden if a different view is selected
        
        if page_num != 1: 
            self.tfpg_slim_assoc_ctrl.balloon_in_slim_assoc.hide_all()
        if page_num != 2: 
            self.tfpg_synthesis_ctrl.balloon_in_synth.hide_all()

        return

 
pass # end of TfpgCtrl class
# ----------------------------------------------------------------------



class TfpgSubCtrl (Controller):
    """Represents the SubController for the StartStopActivity SubViews"""
    
    def __init__(self, parent, model, view):
        Controller.__init__(self, model, view)

        # Chief controller
        self.parent = parent

        # Sub-controllers (all share the same view)  
        self.tfpg_synthesis_ctrl = TfpgSynthesisCtrl(self.parent, model, model.tfpg_synthesis_model, view, view.tfpg_synthesis_view)
        self.tfpg_behav_val_ctrl = TfpgBehavValCtrl(model, model.tfpg_behav_val_model, view, view.tfpg_behav_val_view)
        self.tfpg_effect_val_ctrl = TfpgEffectValCtrl(model, model.tfpg_effect_val_model, view, view.tfpg_effect_val_view)
        
        return

  
pass # end of class
# ----------------------------------------------------------------------



class TfpgViewerCtrl (Controller):
    """This is a sub-controller for handling TfpgViewer activities"""
        
    def __init__(self, parent_model, model, parent_view, view):
        Controller.__init__(self, model, view)

        self.parent_model = parent_model
        self.parent_view = parent_view
        
        # The self.tfpg_view_from_viewer is from tools.tfpg_viewer.src.views.tfpg_view
        self.tfpg_view_from_viewer = None
        
        # The self.curr_tfpg_name is the current name of the loaded tfpg
        self.curr_tfpg_name = None
        
        self.observe_model(model.slim_model)
        return

    

    def register_view(self, view): 
        Controller.register_view(self, view)
        return
        
        
    def on_button_zoom_in_clicked(self, action):
        """Function called when the user wants to zoom in.
            @param   action:  The action to enable the button self.view['button_zoom_in']
            @type    action:  gtk.Button
            """
        fig = self.tfpg_view_from_viewer.ui_zoom_in()
        fig.mpl_connect('motion_notify_event', self.onmotion)
        fig.mpl_connect('button_press_event', self.onclick)

        toolbar = NavigationToolbar(fig,None)
        self.view.clear()
        self.view['vbox_toolbar_viewer'].add(toolbar)
        self.view['vbox_view'].add(fig)
        self.view['vbox_view'].show_all()

        return


    def on_button_zoom_out_clicked(self, action):
        """Function called when the user wants to zoom out.
            @param   action:  The action to enable the button self.view['button_zoom_out']
            @type    action:  gtk.Button
            """
        fig = self.tfpg_view_from_viewer.ui_zoom_out()
        fig.mpl_connect('motion_notify_event', self.onmotion)
        fig.mpl_connect('button_press_event', self.onclick)

        toolbar = NavigationToolbar(fig,None)
        self.view.clear()
        self.view['vbox_toolbar_viewer'].add(toolbar)
        self.view['vbox_view'].add(fig)
        self.view['vbox_view'].show_all()

        return


    def on_button_caption_clicked(self, action):
        """Function called when the user clicks the informatio button
           to see the caption.
            @param   action:  The action to enable the button self.view['button_caption']
            @type    action:  gtk.Button
            """
        popup = gtk.Window()
        popup.set_position(gtk.WIN_POS_MOUSE)
        popup.set_title( "Caption" )

        hbox = gtk.HBox()
        for ch in hbox.get_children():
            hbox.remove(ch)

        caption = self.tfpg_view_from_viewer.create_caption()
        hbox.add(caption)
        popup.add(hbox)
        popup.set_default_size(400, 600)
        popup.show_all()

        return
        
    
    
    # ------------------------------------------------------------------------
    #       Mouse Listener Function
    # ------------------------------------------------------------------------
    
    def onclick(self, event):
        """Function called when the user presses the mouse button onto the graphical 
            representation.
            Basically, if the user presses over a node or an edge, its information is shown by 
            setting the variable 'component_clicked' to be the selected component.
             @param   event:  The event when the user click on the nodes/edges of TFPG.
             @type    event:  MPL MouseEvent
        """
        component_clicked = self.get_mouse_position(event)
        self.view.clear_tfpg_info()

        if component_clicked != None:
            info = self.tfpg_view_from_viewer.print_component_info(component_clicked, pretty_print=False)
            self.display_related_information_in_tabular_format(component_clicked, info)

        return



    def display_related_information_in_tabular_format(self, component_clicked, info, errors=None):
 
        if "_edg_lbl" in component_clicked:

            # It is an edge

            table = gtk.Table(5, 5, True)
                
            lbl_name = gtk.Label("Name: ")
            table.attach(lbl_name, 0, 1, 0, 1)
            txt_name = gtk.Entry(50)
            txt_name.set_name('name')
            txt_name.set_text(info['name'])
            if errors != None and 'name' in errors.keys():
                txt_name.modify_text(gtk.STATE_NORMAL, gtk.gdk.color_parse("red"))
                txt_name.set_tooltip_text(errors['name'])
            txt_name.connect("changed", self.tfpg_element_info_changed, txt_name, info)
            table.attach(txt_name, 1, 2, 0, 1)
            
            lbl_src = gtk.Label("Source: ")
            table.attach(lbl_src, 0, 1, 1, 2)
            txt_src = gtk.Entry(50)
            txt_src.set_name('src')
            txt_src.set_text(info['src'])
            if errors != None and 'src' in errors.keys():
                txt_src.modify_text(gtk.STATE_NORMAL, gtk.gdk.color_parse("red"))
                txt_src.set_tooltip_text(errors['src'])
            txt_src.connect("changed", self.tfpg_element_info_changed, txt_src, info)
            table.attach(txt_src, 1, 2, 1, 2)
            
            lbl_dest = gtk.Label("Destination: ")
            table.attach(lbl_dest, 0, 1, 2, 3)
            txt_dest = gtk.Entry(50)
            txt_dest.set_name('dst')
            txt_dest.set_text(info['dst'])
            if errors != None and 'dst' in errors.keys():
                txt_dest.modify_text(gtk.STATE_NORMAL, gtk.gdk.color_parse("red"))
                txt_dest.set_tooltip_text(errors['dst'])
            txt_dest.connect("changed", self.tfpg_element_info_changed, txt_dest, info)
            table.attach(txt_dest, 1, 2, 2, 3)

            lbl_tMin = gtk.Label("TMin: ")
            table.attach(lbl_tMin, 2, 3, 0, 1)
            txt_tMin = gtk.Entry(50)
            txt_tMin.set_name('tMin')
            txt_tMin.set_text(str(info['tMin']))
            if errors != None and 'tMin' in errors.keys():
                txt_tMin.modify_text(gtk.STATE_NORMAL, gtk.gdk.color_parse("red"))
                txt_tMin.set_tooltip_text(errors['tMin'])
            txt_tMin.connect("changed", self.tfpg_element_info_changed, txt_tMin, info)
            table.attach(txt_tMin, 3, 4, 0, 1)

            lbl_tMax = gtk.Label("TMax: ")
            table.attach(lbl_tMax, 2, 3, 1, 2)
            txt_tMax = gtk.Entry(50)
            txt_tMax.set_name('tMax')
            txt_tMax.set_text(str(info['tMax']))
            if errors != None and 'tMax' in errors.keys():
                txt_tMax.modify_text(gtk.STATE_NORMAL, gtk.gdk.color_parse("red"))
                txt_tMax.set_tooltip_text(errors['tMax'])
            txt_tMax.connect("changed", self.tfpg_element_info_changed, txt_tMax, info)
            table.attach(txt_tMax, 3, 4, 1, 2)

            lbl_modes = gtk.Label("Modes: ")
            table.attach(lbl_modes, 2, 3, 2, 3)
            txt_modes = gtk.Entry(50)
            txt_modes.set_name('modes')
            txt_modes.set_text(info['modes'])
            if errors != None and 'modes' in errors.keys():
                txt_modes.modify_text(gtk.STATE_NORMAL, gtk.gdk.color_parse("red"))
                txt_modes.set_tooltip_text(errors['modes'])
            txt_modes.connect("changed", self.tfpg_element_info_changed, txt_modes, info)
            table.attach(txt_modes, 3, 4, 2, 3)
            
            button_save_changes = gtk.Button("Save Changes")
            table.attach(button_save_changes, 0, 1, 4, 5)

            button_save_changes.connect("clicked",
                                     self.update_graph,
                                     component_clicked,
                                     info)

            # self.view['vbox_info'].pack_start(table, False, False, 0)
            self.view['vbox_info'].add_with_viewport(table)

        else:

            # It is a node

            table = gtk.Table(2, 5, True)
            
            lbl_name = gtk.Label("Name: ")
            table.attach(lbl_name, 0, 1, 0, 1)
            txt_name = gtk.Entry(50)
            txt_name.set_name('name')
            txt_name.set_text(info['name'])
            if errors != None and 'name' in errors.keys():
                txt_name.modify_text(gtk.STATE_NORMAL, gtk.gdk.color_parse("red"))
                txt_name.set_tooltip_text(errors['name'])
            txt_name.connect("changed", self.tfpg_element_info_changed, txt_name, info)
            table.attach(txt_name, 1, 2, 0, 1)
            
            lbl_type = gtk.Label("Type: ")
            table.attach(lbl_type, 0, 1, 1, 2)
            txt_type = gtk.Entry(50)
            txt_type.set_name('type')
            txt_type.set_text(info['type'])
            if errors != None and 'type' in errors.keys():
                txt_type.modify_text(gtk.STATE_NORMAL, gtk.gdk.color_parse("red"))
                txt_type.set_tooltip_text(errors['type'])
            txt_type.connect("changed", self.tfpg_element_info_changed, txt_type, info)
            table.attach(txt_type, 1, 2, 1, 2)
            
            lbl_monitored = gtk.Label("Monitored: ")
            table.attach(lbl_monitored, 0, 1, 2, 3)
            button_is_monitored = gtk.CheckButton("")
            button_is_monitored.set_active(info['isMonitored'])
            button_is_monitored.connect("toggled", self.tfpg_node_monitored_toggled, info)
            table.attach(button_is_monitored, 1, 2, 2, 3)
            
            button_save_changes = gtk.Button("Save Changes")
            table.attach(button_save_changes, 0, 1, 4, 5)
            
            button_save_changes.connect("clicked",
                                     self.update_graph,
                                     component_clicked,
                                     info)
          
        # self.view['vbox_info'].pack_start(table, False, False, 0)
        self.view['vbox_info'].add_with_viewport(table)   
        self.view['vbox_info'].show_all()

        return


    def tfpg_node_monitored_toggled(self, widget, values):
        """Function called when the user toogles a node monitor.
            @param  widget: The toggle button of Node Monitor.
            @type   widget: gtk.ToogleButton
            @param  values: The dict containing pairs <attribute, value> for the selected element
            @type   values: dict
        """
        values['isMonitored'] = widget.get_active()
        return



    def tfpg_element_info_changed(self, widget, entry, values):
        """Function called when the user changes (without pressing the 'Enter' button or
            giving some confirmation) a value in the vbox_info which appears when the user
            selects a component of the TFPG.
            Essentially, every time the user changes something, the 'values' dict containing
            the information of the element is updated.
            @param  widget: The widget of new entry.
            @type   widget: gtk.Widget
            @param  entry: The new entry.
            @type   entry: gtk.Entry
            @param  values: The dict containing pairs <attribute, value> for the selected element
            @type   values: dict
        """
        values[entry.get_name()] = entry.get_text()
        return



    def update_graph(self, widget, component, values):
        """Updates the graphical representation of the TFPG with some user-defined values.
            This function is called when the user clicks the 'Save changes' buttons, which
            appears in the vbox_info when the user selects a component of the TFPG.
            Essentially, it updated the graph internal structure with the (possibly) new 
            values inserted, and redraws the graph.
            @param  widget: The enter button.
            @type   widget: gtk.Button
            @param  component: The name of the element updated.
            @type   component: string
            @param  values: The dict containing pairs <attribute, value> for the selected element
            @type   values: dict
        """
        tmp_filename = ""
        txml_filename = ""
        error_list = self.tfpg_view_from_viewer.update_graph(component, values)
        if error_list == []:

            sel = self.parent_view['tv_tfpg_list'].get_selection()
            tree_model, paths = sel.get_selected_rows()
            
            if paths:
                tfpg_structural = tree_model.get_value(tree_model.get_iter(paths[0]), 0)
                for tfpg in self.parent_model.get_tfpgs_name_fn_saved_list():
                    if tfpg[0] == self.curr_tfpg_name:
                        tfpg[0] = tfpg_structural.get_name()
                        tmp_filename = tfpg[1]
                        txml_filename = tfpg[4]
                        tfpg[2] = False
            

            fig = self.tfpg_view_from_viewer.create_tfpg_canvas_figure(tfpg_graph=tfpg_structural)
            fig.mpl_connect('motion_notify_event', self.onmotion)
            fig.mpl_connect('button_press_event', self.onclick)

            for ch in self.view['vbox_view'].get_children():
                self.view['vbox_view'].remove(ch)
                   
                self.view['vbox_view'].add(fig)
                self.view['vbox_view'].show_all()

                # GZ: for now we remove all the shown information when the user clicks the 
                #     button save changes (without considering whether all went good or not.
                #     Notice that probably we need a global variable to always which is the 
                #     component clicked such that it is possible modify the same element 
                #     twice sequentially
                self.view.clear_tfpg_info()


            if tmp_filename != "":
            
                tfpg_structural.dump_to_concrete_syntax_file(tmp_filename)  
                
            else:

                #LB## If there is not a human readable file related to the loaded xml file,
                ##### a human readable filewith the same name of the xml file is created
                ##### and saved/edited/updated automatically in the same dir.

                file_name, file_extension = os.path.splitext(txml_filename)
                tmp_filename = file_name + ".tfpg"
                tfpg_structural.dump_to_concrete_syntax_file(tmp_filename)
            
            
        else:
            # self.tfpg_view.ui_show_message_dialog('error', str(error_list[0]),
            #                                       parent_name="window_top")
            ## GZ: maybe this part should not be in the controller, but there should be 
            ##     a function of the model whose return value is used by the view
            self.view.clear_tfpg_info()

            errors = dict()
            if type(error_list[0]) == TimeNotCorrectError:
                errors['tMin'] = str(error_list[0])
                errors['tMax'] = str(error_list[0])

            elif type(error_list[0]) == NodeLoopError:
                errors['src'] = str(error_list[0])
                errors['dst'] = str(error_list[0])

            elif type(error_list[0]) == NodeNotInEdge:
                ## GZ: how to understand whether is the source or the destination?
                errors['src'] = str(error_list[0])
                errors['dst'] = str(error_list[0])

            elif type(error_list[0]) == FMNodeIsNotDestNode:
                errors['dst'] = str(error_list[0])

            elif type(error_list[0]) == NotExistingModes:
                errors['modes'] = str(error_list[0])

            elif type(error_list[0]) == FMNotCorrectError:
                errors['type'] = str(error_list[0])

            elif type(error_list[0]) == NotExistingNode:
                if error_list[0].get_node_role() != None:
                    errors[error_list[0].get_node_role()] = str(error_list[0])

            self.display_related_information_in_tabular_format(component,
                                                               values,
                                                               errors=errors)

            pass

        return
    
        
    def onmotion(self, event):
        """Function called when the user moves the mouse.
            In case the user moves the mouse over a node or an edge, the mouse icon is
            changed and the coordinates are returned.
            @param   event:  The event when the user moves the mouse over a node or an edge of TFPG.
                @type    event:  MPL MouseEvent
        """
        position = self.get_mouse_position(event)
        main_window = self.view[self.view.top].get_parent().get_parent().get_parent().get_parent().get_parent()
        if position != None:
            self.view[self.view.top].get_parent().get_window().set_cursor(gtk.gdk.Cursor(gtk.gdk.HAND2))
        else:
            self.view[self.view.top].get_window().set_cursor(gtk.gdk.Cursor(gtk.gdk.ARROW))

        return
        
        
    def get_mouse_position(self, event):
        """Checks whether the mouse is over a node or an edge. 
            In case the mouse is over a node or an edge, it returns its name.
            Otherwise, it returns None.
            @param   event:  The event when the user moves the mouse over a node or an edge of TFPG.
            @type    event:  MPL MouseEvent
           
            @return   nd:  The name of the node/edge or None.
            @rtype    nd:  string
        """
        for nd in self.tfpg_view_from_viewer.pos.keys():

            (nd_x, nd_y) = self.tfpg_view_from_viewer.pos[nd]

            if "edg" in nd:
                if event.xdata >= nd_x - 30 and event.xdata <= nd_x + 22:
                    if event.ydata >= nd_y - 30 and event.ydata <= nd_y + 30:
                        return nd

            elif "nd_lbl" not in nd:
                if event.xdata >= nd_x - 5 and event.xdata <= nd_x + 5:
                    if event.ydata >= nd_y - 15 and event.ydata <= nd_y + 15:
                        return nd

        return None



pass # end of TfpgViewerCtrl class
# ----------------------------------------------------------------------


class TfpgSlimAssocCtrl (Controller):
    """This is a sub-controller for handling TfpgModel activities"""
        
    def __init__(self, parent_model, model, parent_view, view):
        Controller.__init__(self, model, view)

        self.parent_model = parent_model
        self.parent_view = parent_view
        
        self.balloon_in_slim_assoc = gtk.Window()
        self.balloon_in_slim_assoc.set_decorated(False)
        self.balloon_in_slim_assoc.set_accept_focus(False)
        self.balloon_in_slim_assoc.set_property("skip-taskbar-hint", True)
        self.txtview_in_slim_assoc = gtk.TextView()
        self.txtview_in_slim_assoc.set_property("wrap-mode",gtk.WRAP_WORD)
        self.txtbuf_in_slim_assoc = gtk.TextBuffer()
        self.txtview_in_slim_assoc.set_buffer(self.txtbuf_in_slim_assoc)
        self.balloon_in_slim_assoc.add(self.txtview_in_slim_assoc)

        # The self.curr_tfpg_name is the current name of the loaded tfpg
        self.curr_tfpg_name = None
        self.slim_expr_error = ""
        self.observe_model(model.slim_model)
        
        # The list of slim_associations, related to the loaded tfpg, saved or not
        # The list is composed of: [tfpg_name, slim_assoc_filename, True -> saved, else False]
        self.slim_assoc_not_saved = []
        self.curr_slim_assoc_not_saved = []
        
        return


    def register_view(self, view): 
        Controller.register_view(self, view)
        return
        
            
    def initialize_tfpg_model_tab(self, fm_ls, mon_discr_ls, unmon_discr_ls, modes_ls):
        """It Prepares the tree view of the tfpg model tab.
            Given three input lists representing failure mode variables, monitored discrepancies
            and not monitored discrepancies, essentially this function prepares the 'Tfpg Mode' 
            tab by inserting all the variable names and, in case of failure modes, also the relative
            drop-down list.
            @param    fm_ls: The list of failure modes in string format.
            @type     fm_ls: list
            @param    mon_discr_ls: The list of monitored discrepancies in string format.
            @type     mon_discr_ls: list
            @param    unmon_discr_ls: The list of not-monitored discrepancies in string format.
            @type     unmon_discr_ls: list

                @return    slim_associations: The dict of all slim association (failure, monitored, 
                                         not monitored and modes)
            @rtype     slim_associations: dict
        """

        self.view.clear()

        # Failure modes

        possible_fm_expr = self.get_fm_possible_slim_associations()
        slim_associations = dict()

        fm_liststore = gtk.ListStore(str, str, str, str)
        for fm in fm_ls:
            fm_liststore.append([fm.get_name(), ":=", None, None])
            slim_associations[fm.get_name()] = ("ERROR", None, "FailureMode")

        self.view['tv_failure_modes'].set_model(fm_liststore)

        renderer_name = gtk.CellRendererText()
        column_name = gtk.TreeViewColumn("Name", renderer_name, text=0)
        self.view['tv_failure_modes'].append_column(column_name)

        renderer_def = gtk.CellRendererText()
        column_def = gtk.TreeViewColumn("", renderer_def, text=1)
        self.view['tv_failure_modes'].append_column(column_def)

        renderer_expr = gtk.CellRendererCombo()
        renderer_expr.set_property("editable", True)
        renderer_expr.set_property("has-entry", False)
        renderer_expr.set_property("model", possible_fm_expr)
        renderer_expr.set_property("text-column", 0)
        renderer_expr.connect("edited", self.combo_changed, fm_liststore)

        column_expr = gtk.TreeViewColumn("Expression", renderer_expr, text=2)
        column_expr.set_expand(True)
        self.view['tv_failure_modes'].append_column(column_expr)

        renderer_expr_result = gtk.CellRendererText()
        column_result = gtk.TreeViewColumn("Result", renderer_expr_result, text=3)
        column_result.set_sizing(gtk.TREE_VIEW_COLUMN_FIXED)
        column_result.set_fixed_width(70)
        self.view['tv_failure_modes'].append_column(column_result)

        # Monitored Discrepancies

        mon_liststore = gtk.ListStore(str, str, str, str)
        for mon in mon_discr_ls:
            mon_liststore.append([mon.get_name(), ":=", None, None])
            slim_associations[mon.get_name()] = ("ERROR", None, "MonitoredDiscr")

        self.view['tv_monitored_discr'].set_model(mon_liststore)

        renderer_name_1 = gtk.CellRendererText()
        column_name_1 = gtk.TreeViewColumn("Name", renderer_name_1, text=0)
        self.view['tv_monitored_discr'].append_column(column_name_1)

        renderer_def_1 = gtk.CellRendererText()
        column_def_1 = gtk.TreeViewColumn("", renderer_def_1, text=1)
        self.view['tv_monitored_discr'].append_column(column_def_1)

        renderer_editable_expr_1 = gtk.CellRendererText()
        renderer_editable_expr_1.set_property("editable", True)

        column_editable_expr_1 = gtk.TreeViewColumn("Expression",
            renderer_editable_expr_1, text=2)
        column_editable_expr_1.set_expand(True)
        self.view['tv_monitored_discr'].append_column(column_editable_expr_1)

        renderer_editable_expr_1.connect("edited",
                                         self.association_edited,
                                         mon_liststore,
                                         'tv_monitored_discr')

        renderer_expr_result_1 = gtk.CellRendererText()
        column_result_1 = gtk.TreeViewColumn("Result", renderer_expr_result_1, text=3)
        column_result_1.set_sizing(gtk.TREE_VIEW_COLUMN_FIXED)
        column_result_1.set_fixed_width(70)
        self.view['tv_monitored_discr'].append_column(column_result_1)
        
        sel = self.view['tv_monitored_discr'].get_selection()
        sel.set_mode(gtk.SELECTION_BROWSE)
        sel.connect("changed", self.result_clicked, 'tv_monitored_discr')

        # Unmonitored Discrepancies

        unmon_liststore = gtk.ListStore(str, str, str, str)
        for unmon in unmon_discr_ls:
            unmon_liststore.append([unmon.get_name(), ":=", None, None])
            slim_associations[unmon.get_name()] = ("ERROR", None, "UnmonitoredDiscr")

        self.view['tv_not_monitored_discr'].set_model(unmon_liststore)

        renderer_name_2 = gtk.CellRendererText()
        column_name_2 = gtk.TreeViewColumn("Name", renderer_name_2, text=0)
        self.view['tv_not_monitored_discr'].append_column(column_name_2)

        renderer_def_2 = gtk.CellRendererText()
        column_def_2 = gtk.TreeViewColumn("", renderer_def_2, text=1)
        self.view['tv_not_monitored_discr'].append_column(column_def_2)

        renderer_editable_expr_2 = gtk.CellRendererText()
        renderer_editable_expr_2.set_property("editable", True)

        column_editable_expr_2 = gtk.TreeViewColumn("Expression",
            renderer_editable_expr_2, text=2)
        column_editable_expr_2.set_expand(True)
        self.view['tv_not_monitored_discr'].append_column(column_editable_expr_2)

        renderer_editable_expr_2.connect("edited",
                                         self.association_edited,
                                         unmon_liststore,
                                         'tv_not_monitored_discr')

        renderer_expr_result_2 = gtk.CellRendererText()
        column_result_2 = gtk.TreeViewColumn("Result", renderer_expr_result_2, text=3)
        column_result_2.set_sizing(gtk.TREE_VIEW_COLUMN_FIXED)
        column_result_2.set_fixed_width(70)
        self.view['tv_not_monitored_discr'].append_column(column_result_2)

        sel = self.view['tv_not_monitored_discr'].get_selection()
        sel.set_mode(gtk.SELECTION_BROWSE)
        sel.connect("changed", self.result_clicked, 'tv_not_monitored_discr')

        # Tfpg modes

        modes_liststore = gtk.ListStore(str, str, str, str)
        for mode in modes_ls:
            modes_liststore.append([str(mode), ":=", None, None])
            slim_associations[str(mode)] = ("ERROR", None, "TfpgMode")

        self.view['tv_tfpg_modes'].set_model(modes_liststore)

        renderer_name_3 = gtk.CellRendererText()
        column_name_3 = gtk.TreeViewColumn("Name", renderer_name_3, text=0)
        self.view['tv_tfpg_modes'].append_column(column_name_3)

        renderer_def_3 = gtk.CellRendererText()
        column_def_3 = gtk.TreeViewColumn("", renderer_def_3, text=1)
        self.view['tv_tfpg_modes'].append_column(column_def_3)

        renderer_editable_expr_3 = gtk.CellRendererText()
        renderer_editable_expr_3.set_property("editable", True)

        column_editable_expr_3 = gtk.TreeViewColumn("Expression",
                                                    renderer_editable_expr_3,
                                                    text=2)
        column_editable_expr_3.set_expand(True)
        self.view['tv_tfpg_modes'].append_column(column_editable_expr_3)
        
        renderer_editable_expr_3.connect("edited",
                                         self.association_edited,
                                         modes_liststore,
                                         'tv_tfpg_modes')

        renderer_expr_result_3 = gtk.CellRendererText()
        column_result_3 = gtk.TreeViewColumn("Result", renderer_expr_result_3, text=3)
        column_result_3.set_sizing(gtk.TREE_VIEW_COLUMN_FIXED)
        column_result_3.set_fixed_width(70)
        self.view['tv_tfpg_modes'].append_column(column_result_3)

        sel = self.view['tv_tfpg_modes'].get_selection()
        sel.set_mode(gtk.SELECTION_BROWSE)
        sel.connect("changed", self.result_clicked, 'tv_tfpg_modes')

        self.view['vbox_associations'].show_all()

        # Initially, all the text entries are empty; hence, we cannot save
        self.view['button_save_associations'].set_sensitive(False)

        return slim_associations



    def association_edited(self, widget, path, text, liststore, tv_name):
        """Function called when the user inserts or updates a slim expression
            associated to a tfpg mode.
            Notice that this function is called in the case the user modifies
            a discrepancy. Obviously, it is not called in case of failure modes,
            as they are selectable through a drop-down list.
            @param    widget: The widget of slim_expression.
            @type     widget: gtk.CellRenderedText
            @param    path: The path of the slim_expression.
            @type     path: int
            @param    text: The text of the slim_expression.
            @type     text: string
            @param    liststore: The liststore of the slim_expression.
            @type     liststore: gtk.ListStore
            @param    tv_name: The name of the window (monitored or not monitored)
            @type     tv_name: string
        """
        error_list, impl_name = get_errors_in_atomic_proposition_with_templates(text, self.model.slim_model)
        slim_associations = self.model.get_slim_associations(self.curr_tfpg_name)

        # getting the type of association depending on the tree view we have selected
        _type = None
        if tv_name == "tv_monitored_discr":
            _type = "MonitoredDiscr"
        elif tv_name == "tv_not_monitored_discr":
            _type = "UnmonitoredDiscr"
        elif tv_name == "tv_tfpg_modes":
            _type = "TfpgMode"
        else:
            # !!Warning!! This code should never be reached
            raise Exception("Unknown Tree View")

        if error_list == []:

            check_errors = self.check_entry(text, self.model.slim_model, tv_name)

            if check_errors == []:
                liststore[path][2] = text
                slim_associations[liststore[path][0]] = ("OK", text, _type)
                liststore[path][3] = "OK"
                self.view['button_save_associations'].set_sensitive(self.all_associations_ok())
                self.balloon_in_slim_assoc.hide_all()
 
            else:
                error = str(check_errors[0])
                self.show_association_error(error, tv_name, path)
                liststore[path][2] = text
                slim_associations[liststore[path][0]] = ("ERROR", error, _type)
                liststore[path][3] = "ERROR"
                self.view['button_save_associations'].set_sensitive(False)
                
        else:
            error = str(error_list[0])
            self.show_association_error(error, tv_name, path)
            liststore[path][2] = text
            slim_associations[liststore[path][0]] = ("ERROR", error, _type)
            liststore[path][3] = "ERROR"
            self.view['button_save_associations'].set_sensitive(False)

        self.model.add_associations_list(self.curr_tfpg_name, slim_associations)


        flag = False
        for slim_assoc in self.get_slim_assoc_not_saved():
            if slim_assoc[0] == self.curr_tfpg_name:
                flag = True 
                
        # If new slim expression are inserted, the list is updated    
        if self.all_associations_ok() and flag is False:
            self.slim_assoc_not_saved.append([self.curr_tfpg_name, "", False])
            
            
        # Set the SLIM_ASSOC as saved
        for slim_assoc in self.get_slim_assoc_not_saved():
            if slim_assoc[0] == self.curr_tfpg_name:
                if self.all_associations_ok():
                    slim_assoc[2] = False
                else:
                    slim_assoc[2] = True
                        
        return


    def combo_changed(self, widget, path, text, liststore):
        """Function called when the user chooses which slim expression of the drop-down
            list must be associated to a failure mode.
            Notice that this function is always called within the treeview tv_failure_modes.
            @param    widget: The widget of slim_expression.
            @type     widget: gtk.CellRenderedText
            @param    path: The path of the slim_expression.
            @type     path: int
            @param    text: The text of the slim_expression.
            @type     text: string
            @param    liststore: The liststore of the slim_expression.
            @type     liststore: gtk.ListStore
        """
        
        slim_associations = self.model.get_slim_associations(self.curr_tfpg_name)

        liststore[path][2] = text
        slim_associations[liststore[path][0]] = ("OK", text, "FailureMode")
        liststore[path][3] = "OK"
        self.view['button_save_associations'].set_sensitive(self.all_associations_ok())
        self.balloon_in_slim_assoc.hide_all()

        self.model.add_associations_list(self.curr_tfpg_name, slim_associations)

        # Set the SLIM_ASSOC as saved
        for slim_assoc in self.get_slim_assoc_not_saved():
            if slim_assoc[0] == self.curr_tfpg_name:
                if self.all_associations_ok():
                    slim_assoc[2] = False
                else:
                    slim_assoc[2] = True
        
        return


    def show_association_error(self, text, tv_name, path):
        """Function which shows the error derived from the insertion of a wrong slim expression.
            @param    text: The text of the wrong slim_expression.
            @type     text: string
            @param    tv_name: The name of the window (monitored or not monitored)
            @type     tv_name: string
            @param    path: The path of the wrong slim_expression.
            @type     path: int
            """

        window_x, window_y = self.view[tv_name].get_parent_window().get_origin()
        bottom_left_x = self.view[tv_name].allocation.x + window_x
        bottom_left_y = self.view[tv_name].allocation.y + window_y + 50 + (int(path)*22)

        self.balloon_in_slim_assoc.set_position(gtk.WIN_POS_MOUSE)
        self.txtview_in_slim_assoc.modify_base(gtk.STATE_NORMAL, gtk.gdk.color_parse("indian red"))
        self.txtbuf_in_slim_assoc.set_text(text)
        self.balloon_in_slim_assoc.resize(400, 30)
        self.balloon_in_slim_assoc.show_all()
        self.balloon_in_slim_assoc.move(bottom_left_x, bottom_left_y)

        return


    def all_associations_ok(self):
        """Function which checks that all the associated slim expressions are correct.
            @return    True or False: True if all associated slim expressions are correct, else False.
            @rtype     True or False: boolean
            """
        slim_associations = self.model.get_slim_associations(self.curr_tfpg_name)

        for (k, v) in slim_associations.items():
            (val, error, _type) = v
            if val == "ERROR" : return False

        return True


    def result_clicked(self, selection, tv_name):
        """Function called when the user clicks over one row of the list of slim expressions.
            In case the selected row contains a wrong slim expression, the relative error is
            shown.
            @param    selection: The slim_expression selected.
            @type     selection: gtk.TreeSelection
            @param    tv_name: The name of the window (monitored or not monitored)
            @type     tv_name: string
            """
        self.balloon_in_slim_assoc.hide_all()
        tree_model, tree_iter = selection.get_selected()
        slim_associations = self.model.get_slim_associations(self.curr_tfpg_name)

        if tree_iter != None:
            checked = tree_model[tree_iter][3]
            if checked == "ERROR":
                mode = tree_model[tree_iter][0]
                (val, error, _type) = slim_associations[mode]
                self.show_association_error(error, tv_name, 
                                            tree_model.get_path(tree_iter)[0])

        return


    def check_entry(self, text, slim_model, tv_name):
        """Function which checks a syntactically correct slim expression entered by the user.
            @param    text: The slim_expression.
            @type     text: string
            @param    slim_model: The slim_model.
            @type     slim_model: SlimModel
            @param    tv_name: The name of the window (monitored or not monitored)
            @type     tv_name: string

            @return    error_list: The list with the errors of wrong slim_expressions.
            @rtype     error_list: list
            """
        import slim_compiler.src.smv.api as smv_api

        error_list = []

        ## In case of monitored discrepancies, the following check must be performed
        ## - Expressions on Monitored discrepancies must contain only observable variables
        if tv_name == "tv_monitored_discr" or tv_name == "tv_tfpg_modes":

            error_list, _ = functions.get_errors_in_atomic_proposition_with_templates_observable(text, self.model.slim_model)

        ## In case of not-monitored discrepancies, the following check must be performed.
        ## - Discrepancies expressions must contain only data or mode variables from the system
        ##   This means that event ports and everything related to error models bust be excluded

        elif tv_name == "tv_not_monitored_discr":

            root_impl = slim_model.get_nominal_root_impl()
            _type = root_impl.parent.get_component_type(root_impl.name.split(".")[0])
            event_ports = dict(_type.in_event_ports.items() + _type.out_event_ports.items())

            # GZ: maybe the following should be contained in a function
            # discrepancies expressions are generic boolean expressions
            # to check that they refer only on observable variable, we perform the following steps:
            # - replacement of all boolean operators with whitespaces
            #   notice that a boolean operator is here defined as a string ('and', 'or', ...)
            #   between spaces, such that it is not confused with part of a variabel name, or
            #   as 'not ', if at the beginning of the expression
            # - thus, we split the expression removing whitespaces; what remains is the identifier
            #   of a variable or a boolean value ('true' or 'false')
            if text[:4] == "not ":
                text = text.replace(text[:4], " ")
            identifiers = text.replace(" and ", " ")
            identifiers = identifiers.replace(" iff ", " ")
            identifiers = identifiers.replace(" imp ", " ")
            identifiers = identifiers.replace(" not ", " ")
            identifiers = identifiers.replace(" or ", " ")
            identifiers = identifiers.replace(" xnor ", " ")
            identifiers = identifiers.replace(" xor ", " ")
            identifiers = identifiers.replace(" = ", " ")
            identifiers = " ".join(identifiers.split())
            ids = identifiers.split(" ");
            for ident in ids:
                ident = ident.strip()
                if ident != "false" and ident != "true":
                    if ident in event_ports:
                        error = "%s is an event port and cannot be contained in a discrepancy expression.\n" %(ident)
                        error_list.append(error)

        return error_list


    def load_associations_list(self, associations):
        """It loads in the Tfpg Model tab a given associations dictionary.
            This function is called in case the user switches from a tfpg to another one by
            clicking on the list on the left. In this case, the tfpg-related associations
            list must be loaded.
            @param    associations: The dict of associations.
            @type     associations: dict
        """
        for (k, v) in associations.items():
            var_name = k
            error = self.load_single_association(var_name, associations[k])
            if error is True:
                self.parent_view.tfpg_errors_view.errors_buf.insert_with_tags_by_name(
                    self.parent_view.tfpg_errors_view.errors_buf.get_end_iter(),
                    "ERROR:\n\n It is not possible to load the Slim Associations before " +
                    "to load the TFPG.\n", "errors")
                self.parent_view.show_page(5)
                return

        self.view['vbox_associations'].show_all()

        return


    def load_single_association(self, var_name, assoc):
        """It loads a single association to the Tfpg Model tab.
            This function takes two parameters, var_name, the name of the variable to update
            (first column of the treeview) and assoc, a tuple consisting of a value ("OK" or 
            "ERROR") saying whether it has a valid slim expression associated, and of the 
            associated slim expression.
            @param    var_name: The name of the variable to update.
            @type     var_name: string
            @param    assoc: the tuple containing the slim expresison and a value saying whether it is valid or not.
            @type     assoc: tuple(str, str)
        """

        error = False

        (ok, var_value, _type) = assoc

        if _type == "FailureMode":
            model = self.view['tv_failure_modes'].get_model()
            n = 0
            if model is not None:
                while n < len(model):
                    if model[n][0] == var_name:
                        if ok == "OK":
                            model[n][2] = var_value
                            model[n][3] = ok
                        else:
                            model[n][2] = None
                            self.view['tv_failure_modes'].set_model(model)
                            return
                    n = n+1
            else:
                error = True

        elif _type == "MonitoredDiscr":
            n = 0
            model = self.view['tv_monitored_discr'].get_model()

            if model is not None:
                while n < len(model):
                    if model[n][0] == var_name:
                        if ok == "OK":
                            model[n][2] = var_value
                            model[n][3] = ok
                        else:
                            model[n][2] = None
                            self.view['tv_monitored_discr'].set_model(model)
                            return
                    n = n+1
            else:
                error = True
 
        elif _type == "UnmonitoredDiscr":
            n = 0
            model = self.view['tv_not_monitored_discr'].get_model()

            if model is not None:
                while n < len(model):
                    if model[n][0] == var_name:
                        if ok == "OK":
                            model[n][2] = var_value
                            model[n][3] = ok
                        else:
                            model[n][2] = None
                            self.view['tv_not_monitored_discr'].set_model(model)
                            return
                    n = n+1
            else:
                error = True

        elif _type == "TfpgMode":
            n = 0
            model = self.view['tv_tfpg_modes'].get_model()
            if model is not None:
                while n < len(model):
                    if model[n][0] == var_name:
                        if ok == "OK":
                            model[n][2] = var_value
                            model[n][3] = ok
                        else:
                            model[n][2] = None
                            self.view['tv_tfpg_modes'].set_model(model)
                            return
                    n = n + 1
            else:
                error = True

        else:
            ## !!Warning!! This code should never be reached
            error = True
            raise Exception("Unknown association type: " + str(_type))
         
        return error


    def save_slim_assoc(self):
        """Saves a tfpg-slim_expr associations document in .axml format, without clicking the save button."""
        from hasdel_tools.src.xml.tfpg_slim_associations import create_associations_document

        if self.get_curr_slim_assoc_not_saved():
        
            dialog = gtk.FileChooserDialog("Save Associations",
                                           None,
                                           gtk.FILE_CHOOSER_ACTION_SAVE,
                                           (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                                           gtk.STOCK_SAVE, gtk.RESPONSE_OK))
            dialog.set_default_response(gtk.RESPONSE_OK)
            filter = gtk.FileFilter()
            filter.set_name("FAME TFPG ASSOC")
            filter.add_pattern("*.axml")
            dialog.add_filter(filter)
            dialog.set_current_name(os.path.basename(
                generate_unique_tfpg_xml_basename() + ".axml"))
            dialog.set_do_overwrite_confirmation(True)    
            dialog.show()

            response = dialog.run()

            if response == gtk.RESPONSE_OK:
                filename = dialog.get_filename()
                assoc_doc = create_associations_document(self.model, self.get_curr_slim_assoc_not_saved()[0])
                with open(filename, "w") as f:
                    f.write(str(assoc_doc))
                pass
                
                
                # Set the SLIM_ASSOC as saved
                for slim_assoc in self.get_slim_assoc_not_saved():
                    if slim_assoc[0] == self.get_curr_slim_assoc_not_saved()[0] and slim_assoc[2] is False:
                        slim_assoc[2] = True
                        
            dialog.destroy()
        return
        

    
    def on_button_save_associations_clicked(self, action):
        """Saves a tfpg-slim_expr associations document in .axml format."""
        from hasdel_tools.src.xml.tfpg_slim_associations import create_associations_document

        dialog = gtk.FileChooserDialog("Save Associations",
                                       None,
                                       gtk.FILE_CHOOSER_ACTION_SAVE,
                                       (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                                       gtk.STOCK_SAVE, gtk.RESPONSE_OK))
        dialog.set_default_response(gtk.RESPONSE_OK)
        filter = gtk.FileFilter()
        filter.set_name("FAME TFPG ASSOC")
        filter.add_pattern("*.axml")
        dialog.add_filter(filter)
        dialog.set_current_name(os.path.basename(
            generate_unique_tfpg_xml_basename() + ".axml"))
        dialog.set_do_overwrite_confirmation(True)    
        dialog.show()

        response = dialog.run()
        
        if response == gtk.RESPONSE_OK:
            filename = dialog.get_filename()
            assoc_doc = create_associations_document(self.model, self.curr_tfpg_name)
            with open(filename, "w") as f:
                f.write(str(assoc_doc))
            pass
            
            # Set the SLIM_ASSOC as saved
            for slim_assoc in self.get_slim_assoc_not_saved():
                if slim_assoc[0] == self.curr_tfpg_name and slim_assoc[2] is False:
                    slim_assoc[2] = True
                    
        dialog.destroy()
        return


    def on_button_load_associations_clicked(self, action):
        """Load a tfpg-slim_expr associations document in .axml format.
            @param   action:  The action to enable the button self.view['button_load_associations']
            @type    action:  gtk.Button
            """
            
        chooser = gtk.FileChooserDialog(title="Open TFPG SLIM Associations",
                                        action=gtk.FILE_CHOOSER_ACTION_OPEN,
                                        buttons=(gtk.STOCK_CANCEL,gtk.RESPONSE_CANCEL,
                                                 gtk.STOCK_OPEN,
                                                 gtk.RESPONSE_OK))
        chooser.set_select_multiple(False)
        filter = gtk.FileFilter()
        filter.set_name("FAME TFPG SLIM ASSOC")
        filter.add_pattern("*.axml")
        chooser.add_filter(filter)
        chooser.set_current_folder(os.path.dirname(
            generate_unique_tfpg_xml_basename() + ".axml"))

        response = chooser.run()
        if response == gtk.RESPONSE_OK:
            ## First, we check whether there are validation/syntax errors in the 
            ## associations file
            input_assoc, errors_ls = self.model.get_slim_associations_from_file(
                                                   self.curr_tfpg_name,
                                                   chooser.get_filename())
            if errors_ls == []:
                ## If the associations file is syntactically correct, we check that there
                ## are no errors in the slim expressions
                assoc_ls, error = self.check_associations_list(input_assoc)
            
                if error is False:
                    self.load_associations_list(assoc_ls)
                    self.model.add_associations_list(self.curr_tfpg_name, assoc_ls)
 
                    if self.all_associations_ok() is False:
                        self.model.slim_model.console.log_compiler_fail(
                            " > ERROR\n\t > Some associations contain errors.\n")
                    else:
                        # Append the SLIM_ASSOC as saved
                        self.slim_assoc_not_saved.append([self.curr_tfpg_name, chooser.get_filename(), True])
 
                else:
                    self.parent_view.tfpg_errors_view.errors_buf.insert_with_tags_by_name(
                        self.parent_view.tfpg_errors_view.errors_buf.get_end_iter(),
                        "Slim Association Syntax ERROR:\n\n " + self.slim_expr_error + "\n",
                        "errors")
                    self.parent_view.show_page(5)

            else:
                ## errors_ls != [] -> we must show the syntax/validation errors found in 
                ## the associations file
                for err in errors_ls:
                    self.parent_view.tfpg_errors_view.errors_buf.insert_with_tags_by_name(
                        self.parent_view.tfpg_errors_view.errors_buf.get_end_iter(),
                        str(err),
                        "errors")
                    self.parent_view.show_page(5)
 
        chooser.destroy()
        return


    def check_associations_list(self, input_assoc):
        """Function which checks a syntactically correct list of slim expression.
            @param    input_assoc: The slim association to check.
            @type     input_assoc: string

            @return    valid_assoc: The dict of valid slim associations.
            @rtype     valid_assoc: dict
        """
        error = False
        valid_assoc = dict()

        for (name, assoc) in input_assoc.items():
            (ok, expr, _type) = assoc
            error_list, impl_name = get_errors_in_atomic_proposition_with_templates(
                expr, self.model.slim_model)

            if error_list == []:
                    
                if _type == "FailureMode":
                    model = self.view['tv_failure_modes'].get_model()
                    n = 0
                    if model is not None:
                        while n < len(model):
                            if model[n][0] == name:
                                fm_expr_found = False
                                for row in self.get_fm_possible_slim_associations():
                                    if expr == row[0]: fm_expr_found = True

                                if fm_expr_found is True: valid_assoc[name] = ("OK", expr, _type)
                                else: valid_assoc[name] = ("ERROR", expr, _type)
                            n = n+1

                elif _type == "MonitoredDiscr":

                    model = self.view['tv_monitored_discr'].get_model()
                    n = 0
                    if model is not None:
                        while n < len(model):
                            if model[n][0] == name:
                                if self.check_entry(expr,self.model.slim_model,
                                                    'tv_monitored_discr') == []:
                                    valid_assoc[name] = ("OK", expr ,_type)
                                else:
                                    valid_assoc[name] = ("ERROR", expr, _type)
                            n = n+1
 
                elif _type == "UnmonitoredDiscr":
                    model = self.view['tv_not_monitored_discr'].get_model()
                    n = 0
                    if model is not None:
                        while n < len(model):
                            if model[n][0] == name:
                                if self.check_entry(expr,self.model.slim_model,
                                                    'tv_not_monitored_discr') == []:
                                    valid_assoc[name] = ("OK", expr, _type)
                                else:
                                    valid_assoc[name] = ("ERROR", expr, _type)
                            n = n+1
                                   
                elif _type == "TfpgMode":
                    model = self.view['tv_tfpg_modes'].get_model()
                    n = 0
                    if model is not None:
                        while n < len(model):
                            if model[n][0] == name:
                                if self.check_entry(expr,self.model.slim_model,
                                                    'tv_tfpg_modes') == []:
                                    valid_assoc[name] = ("OK", expr, _type)
                                else:
                                    valid_assoc[name] = ("ERROR", expr, _type)
                            n = n + 1
                else:         
                    ## !!Warning!! This code should never be reached
                    raise Exception("Invalid association type: " + str(_type))
 
            else:
                self.slim_expr_error = "The SLIM Expression ' " + expr + " ' is wrong:\n\t" + str(error_list[0])
                error = True
 
        return valid_assoc, error


    def get_fm_possible_slim_associations(self):
        """Returns the possible slim expressions that can be associated to a failure mode.
            @return    possible_fm_expr: The listore of possible slim expressions that can be 
                                                                associated to a failure mode in string format.
            @rtype     possible_fm_expr: gtk.ListStore
            """
        possible_fm_expr = gtk.ListStore(str)

        for injection in self.model.slim_model.fault_injections:
           err_state = injection[3]
           component_path = ".".join(injection[4])
           expr = "%s.error = error:%s" %(component_path, err_state)
           possible_fm_expr.append([expr])

        return possible_fm_expr


    def get_assoc_doc_from_shell(self, tfpg_doc, filename):
        """Returns the association document to load from the shell.
            @param    tfpg_doc: The tfpg document related to the associations file.
            @type     tfpg_doc: TfpgDocument
            @param    filename: The name of the associations file loaded.
            @type     filename: string
        """

        associations, errors_ls = self.model.get_slim_associations_from_file(tfpg_doc.name, 
                                                                             filename);
        if errors_ls == []:
            tfpg_structural = self.parent_model.get_tfpg_structural(tfpg_doc.name)                                 
            self.initialize_tfpg_model_tab(tfpg_structural.get_fm_nodes(),
                                           tfpg_structural.get_monitored_discrepancies(),
                                           tfpg_structural.get_unmonitored_discrepancies(),
                                           tfpg_structural.get_modes_list())
                                           
            assoc_ls, error= self.check_associations_list(associations)
            
            if error is False:
            
                self.model.add_associations_list(tfpg_doc.name, assoc_ls)  
                                     
                if assoc_ls != None: 
                    self.load_associations_list(assoc_ls)
                        
                    # Append the SLIM_ASSOC as saved
                    self.slim_assoc_not_saved.append([tfpg_doc.name, filename, True])
                    
            else:
                self.parent_view.tfpg_errors_view.errors_buf.insert_with_tags_by_name(
                        self.parent_view.tfpg_errors_view.errors_buf.get_end_iter(),
                        "Slim Association Syntax ERROR:\n\n " + self.slim_expr_error + "\n",
                        "errors")
                self.parent_view.show_page(5)
                return

        else:
            for err in errors_ls:
                self.parent_view.tfpg_errors_view.errors_buf.insert_with_tags_by_name(
                    self.parent_view.tfpg_errors_view.errors_buf.get_end_iter(),
                    str(err),
                    "errors")
                self.parent_view.show_page(5)
        
            return
           
        
    def get_slim_assoc_not_saved(self):
        """Gets the list of edited/updated slim_associations Tfpgs.
            @return   self.slim_assoc_not_saved:  The list of edited/updated slim_associations Tfpgs.
            @rtype    self.slim_assoc_not_saved:  list of [tfpg_name, slim_assoc filename, False]
        """     
        return self.slim_assoc_not_saved
        
        
        
    def slim_associations_not_saved(self):
        """After editing slim_associations, if the user quit the GUI,
            removes/modifies slim models, fault injections, the mission
            elements or the TFPG, a warning message appears asking the user what to do:
            to save the slim associations modified, and not saved, or not.
        """
            
        for slim_assoc_not_saved in self.get_slim_assoc_not_saved():
            
            if slim_assoc_not_saved[2] is False:
            
                parent_name = 'window_top'
                save_error = "Warning: the slim associations related to the TFPG '' " + slim_assoc_not_saved[0] + " '' were not saved. " + \
                            "Please click on OK to save the slim associations, otherwise all changes will be discarded!"
                save_response = self.parent_view.ui_show_message_dialog("error", save_error, "", parent_name)
                
                self.curr_slim_assoc_not_saved = slim_assoc_not_saved
                    
                if save_response == gtk.RESPONSE_CANCEL: 
                    message = "Are you sure do you want to exit? All changes will be discarded."
                    response = self.parent_view.ui_show_message_dialog("warning", 
                                message, "", parent_name, buttons=gtk.BUTTONS_YES_NO)

                    if response == gtk.RESPONSE_NO: 
                        self.save_slim_assoc()

                else:                
                    self.save_slim_assoc()
                    
        return

        
    def get_curr_slim_assoc_not_saved(self):
        """Gets the current list of edited/updated slim_associations Tfpgs.
            @return   self.curr_slim_assoc_not_saved:  The current list of edited/updated slim_associations Tfpgs.
            @rtype    self.curr_slim_assoc_not_saved:  list of [tfpg_name, slim_assoc filename, False]
        """     
        return self.curr_slim_assoc_not_saved
        
        
    def clear_slim_assoc_list(self):
        """Clears the list of Slim Associations saved and not saved."""
        self.slim_assoc_not_saved = []
            
            

pass # end of TfpgSlimAssocCtrl class
# ----------------------------------------------------------------------


class TfpgSynthesisCtrl (StartStopActivityCtrl):
    """This is a sub-controller for handling synthesis activities"""

    synthesis_counter = 0

    # Observable properties
    __observables__ = ("synthesis_counter", )
        
    def __init__(self, parent_ctrl, parent_model, model, parent_view, view):
        """Constructor Method"""
        StartStopActivityCtrl.__init__(self, model, view, 
                                       self.run_tfpg_synth, model.stop_tfpg_synth)

        self.parent_ctrl = parent_ctrl
        self.parent_model = parent_model
        self.parent_view = parent_view

        #Set the TFPG Synthesis TextView
        view['tv_synthesis'].set_buffer(view.synthesis_buf)

        self.balloon_in_synth = gtk.Window()
        self.balloon_in_synth.set_decorated(False)
        self.balloon_in_synth.set_accept_focus(False)
        self.balloon_in_synth.set_property("skip-taskbar-hint", True)
        self.txtview_in_synth = gtk.TextView()
        self.txtview_in_synth.set_property("wrap-mode", gtk.WRAP_WORD)
        self.txtbuf_in_synth = gtk.TextBuffer()
        self.txtview_in_synth.set_buffer(self.txtbuf_in_synth)
        self.balloon_in_synth.add(self.txtview_in_synth)

        self.slim_expr_error = ""
  
        self.tfpg_synthesized_loaded = False
        self.tfpg_synthesized_filename = ""
        self.curr_tfpg_name = None
        
        self.observe_model(model.slim_model)
        return



    def register_view(self, view): 
        """Sets the windows, dialog, buttons and other parts of View."""
        StartStopActivityCtrl.register_view(self, view)     
        return



    def register_adapters(self):
        self.adapt("synthesis_sat_bound")
        return


    
    def run_tfpg_synth(self):
        """Runs TFPG Synthesis"""

        # Clear the TFPG Synthesis Textview
        self.view.synthesis_buf.delete(*self.view.synthesis_buf.get_bounds())

        # Set the name of the synthesized tfpg
        if self.curr_tfpg_name is None:
            self.curr_tfpg_name = "DependencyGraph"

        # Check if there are other loaded tfpgs with the same name
        while self.parent_model.name_exist(self.curr_tfpg_name):
            self.synthesis_counter += 1
            self.curr_tfpg_name = "DependencyGraph_%d" %self.synthesis_counter

        # Get the SLIM associations from the GUI
        input_assoc = self.model.get_slim_associations_in_synth_from_gui(self.curr_tfpg_name,
                                        self.view['tv_failure_modes_in_synth'].get_model(),
                                        self.view['tv_monitored_discr_in_synth'].get_model(),
                                        self.view['tv_not_monitored_discr_in_synth'].get_model(),
                                        self.view['tv_tfpg_modes_in_synth'].get_model())

        # Synthesis performed
        (res, tfpg_res_name, slim_assoc_res_name) = self.model.synthesize_tfpg(self.curr_tfpg_name)

        # Synthesis correct
        if tfpg_res_name is not None and slim_assoc_res_name is not None:
            self.view.synthesis_buf.insert_with_tags_by_name(self.view.synthesis_buf.get_end_iter(),
                    "Synthesis Completed:\n\n The following TFPG file has been generated: \n%s\n" %tfpg_res_name,
                    "synthesis")

            # The internal structure of slim associations is created
            self.model.add_associations_list_in_synth(input_assoc, self.curr_tfpg_name)
            
            # Create the related SLIM associations file
            from hasdel_tools.src.xml.tfpg_slim_associations import create_associations_document_from_list

            assoc_doc = create_associations_document_from_list(input_assoc)
            with open(slim_assoc_res_name, "w") as f:
                f.write(str(assoc_doc))

            # The synthesized tfpg name is automatically loaded in the tfpgs list
            # The synthesized tfpg graph is automatically created in the tfpg viewer tab
            # the realted slim associations are automatically loaded in the tfpg slim assoc tab
            self.tfpg_synthesized_loaded = True
            self.tfpg_synthesized_filename = tfpg_res_name
            self.parent_ctrl.load_tfpg_synthesized(self.tfpg_synthesized_filename, input_assoc)

        else:
            self.view.synthesis_buf.insert_with_tags_by_name(self.view.synthesis_buf.get_end_iter(),
                    "Synthesis ERROR:\n\n An error occurred during the TFPG Synthesis: \n%s\n" %res,
                    "synthesis")

        ## At the end of the TFPG Synthesis, the self.curr_tfpg_name returns to be None,
        ## because the relation between TFPG NAME - SLIM ASSOCIATIONS is created ate the
        ## moment of TFPG Synthesis performing.
        
        self.curr_tfpg_name = None
        
        return



    def initialize_tfpg_model_tab_in_synth(self, fm_ls, mon_discr_ls, unmon_discr_ls, modes_ls):
        """It prepares the tree views of the tfpg slim associations tab.
           Given three input lists representing failure mode variables, monitored discrepancies and
           not monitored discrepancies, essentially this function prepares the 'Slim Associations'
           tab by inserting all the variable names and, in case of failure modes, also the relative
           drop-down list.

           @param fm_ls: the list of failure modes in string format.
           @type  fm_ls: list
           @param    mon_discr_ls: The list of monitored discrepancies in string format.
           @type     mon_discr_ls: list
           @param    unmon_discr_ls: The list of not-monitored discrepancies in string format.
           @type     unmon_discr_ls: list
           
           @return    slim_associations: The dict of all slim association (failure, monitored, 
                                         not monitored and modes)
           @rtype     slim_associations: dict
        """

        self.view.clear_columns()

        # Failure modes
        possible_fm_expr = self.get_fm_possible_slim_associations_in_synth()
        slim_associations = dict()

        fm_liststore = gtk.ListStore(str, str, str, str)

        for fm in fm_ls:
            fm_liststore.append([fm, ":=", None, None])
            slim_associations[fm] = ("ERROR", None, "FailureMode")
            
        self.view['tv_failure_modes_in_synth'].set_model(fm_liststore)
        
        renderer_editable_name = gtk.CellRendererText()
        renderer_editable_name.set_property("editable", True)

        column_editable_name = gtk.TreeViewColumn("Name",
            renderer_editable_name, text=0)
        column_editable_name.set_expand(True)
        column_editable_name.set_resizable(True)
        self.view['tv_failure_modes_in_synth'].append_column(column_editable_name)
        
        renderer_editable_name.connect("edited",
                                         self.association_name_edited_in_synth,
                                         fm_liststore, 'tv_failure_modes_in_synth')
                                         
                                         

        renderer_def = gtk.CellRendererText()
        column_def = gtk.TreeViewColumn("", renderer_def, text=1)
        self.view['tv_failure_modes_in_synth'].append_column(column_def)

            
        renderer_expr = gtk.CellRendererCombo()
        renderer_expr.set_property("editable", True)
        renderer_expr.set_property("has-entry", False)
        renderer_expr.set_property("model", possible_fm_expr)
        renderer_expr.set_property("text-column", 0)
        renderer_expr.connect("edited", self.combo_changed_in_synth, fm_liststore)

        column_expr = gtk.TreeViewColumn("Expression", renderer_expr, text=2)
        column_expr.set_expand(True)
        column_expr.set_resizable(True)
        self.view['tv_failure_modes_in_synth'].append_column(column_expr)

        renderer_expr_result = gtk.CellRendererText()
        column_result = gtk.TreeViewColumn("Result", renderer_expr_result, text=3)
        column_result.set_sizing(gtk.TREE_VIEW_COLUMN_FIXED)
        column_result.set_resizable(True)
        column_result.set_fixed_width(130)
        self.view['tv_failure_modes_in_synth'].append_column(column_result)

        # Monitored Discrepancies

        mon_liststore = gtk.ListStore(str, str, str, str)
        for mon in mon_discr_ls: 
            mon_liststore.append([mon, ":=", None, None])
            slim_associations[mon] = ("ERROR", None, "MonitoredDiscr")

        
        self.view['tv_monitored_discr_in_synth'].set_model(mon_liststore)

      
        renderer_editable_name_1 = gtk.CellRendererText()
        renderer_editable_name_1.set_property("editable", True)

        column_editable_name_1 = gtk.TreeViewColumn("Name",
            renderer_editable_name_1, text=0)
        column_editable_name_1.set_expand(True)
        column_editable_name_1.set_resizable(True)
        self.view['tv_monitored_discr_in_synth'].append_column(column_editable_name_1)

        renderer_editable_name_1.connect("edited",
                                         self.association_name_edited_in_synth,
                                         mon_liststore, 'tv_monitored_discr_in_synth')
                                 

        renderer_def_1 = gtk.CellRendererText()
        column_def_1 = gtk.TreeViewColumn("", renderer_def_1, text=1)
        self.view['tv_monitored_discr_in_synth'].append_column(column_def_1)

                 
        renderer_editable_expr_1 = gtk.CellRendererText()
        renderer_editable_expr_1.set_property("editable", True)

        column_editable_expr_1 = gtk.TreeViewColumn("Expression",
            renderer_editable_expr_1, text=2)
        column_editable_expr_1.set_expand(True)
        column_editable_expr_1.set_resizable(True)
        self.view['tv_monitored_discr_in_synth'].append_column(column_editable_expr_1)

        renderer_editable_expr_1.connect("edited",   
                                         self.association_edited_in_synth,
                                         mon_liststore,
                                         'tv_monitored_discr_in_synth')

        renderer_expr_result_1 = gtk.CellRendererText()
        column_result_1 = gtk.TreeViewColumn("Result", renderer_expr_result_1, text=3)
        column_result_1.set_sizing(gtk.TREE_VIEW_COLUMN_FIXED)
        column_result_1.set_resizable(True)
        column_result_1.set_fixed_width(130)
        self.view['tv_monitored_discr_in_synth'].append_column(column_result_1)
        
        sel = self.parent_view.tfpg_synthesis_view['tv_monitored_discr_in_synth'].get_selection()
        sel.set_mode(gtk.SELECTION_BROWSE)
        sel.connect("changed", self.result_clicked_in_synth, 'tv_monitored_discr_in_synth')

        # Unmonitored Discrepancies

        unmon_liststore = gtk.ListStore(str, str, str, str)
        for unmon in unmon_discr_ls: 
            unmon_liststore.append([unmon, ":=", None, None])
            slim_associations[unmon] = ("ERROR", None, "UnmonitoredDiscr")
        
        self.view['tv_not_monitored_discr_in_synth'].set_model(unmon_liststore)
        
        renderer_editable_name_2 = gtk.CellRendererText()
        renderer_editable_name_2.set_property("editable", True)

        column_editable_name_2 = gtk.TreeViewColumn("Name",
            renderer_editable_name_2, text=0)
        column_editable_name_2.set_expand(True)
        column_editable_name_2.set_resizable(True)
        self.view['tv_not_monitored_discr_in_synth'].append_column(column_editable_name_2)

        renderer_editable_name_2.connect("edited",
                                         self.association_name_edited_in_synth,
                                         unmon_liststore, 'tv_not_monitored_discr_in_synth')
                                         
                                         

        renderer_def_2 = gtk.CellRendererText()
        column_def_2 = gtk.TreeViewColumn("", renderer_def_2, text=1)
        self.view['tv_not_monitored_discr_in_synth'].append_column(column_def_2)

        renderer_editable_expr_2 = gtk.CellRendererText()
        renderer_editable_expr_2.set_property("editable", True)

        column_editable_expr_2 = gtk.TreeViewColumn("Expression",
            renderer_editable_expr_2, text=2)
        column_editable_expr_2.set_expand(True)
        column_editable_expr_2.set_resizable(True)
        self.view['tv_not_monitored_discr_in_synth'].append_column(column_editable_expr_2)

        renderer_editable_expr_2.connect("edited",  
                                         self.association_edited_in_synth,
                                         unmon_liststore,
                                         'tv_not_monitored_discr_in_synth')

        renderer_expr_result_2 = gtk.CellRendererText()
        column_result_2 = gtk.TreeViewColumn("Result", renderer_expr_result_2, text=3)
        column_result_2.set_sizing(gtk.TREE_VIEW_COLUMN_FIXED)
        column_result_2.set_resizable(True)
        column_result_2.set_fixed_width(130)
        self.view['tv_not_monitored_discr_in_synth'].append_column(column_result_2)

        sel = self.view['tv_not_monitored_discr_in_synth'].get_selection()
        sel.set_mode(gtk.SELECTION_BROWSE)
        sel.connect("changed", self.result_clicked_in_synth, 'tv_not_monitored_discr_in_synth')

        # Tfpg modes

        modes_liststore = gtk.ListStore(str, str, str, str)
        for mode in modes_ls:  
            modes_liststore.append([mode, ":=", None, None])
            slim_associations[mode] = ("ERROR", None, "TfpgMode")

        self.view['tv_tfpg_modes_in_synth'].set_model(modes_liststore)
        
        
        renderer_editable_name_3 = gtk.CellRendererText()
        renderer_editable_name_3.set_property("editable", True)

        column_editable_name_3 = gtk.TreeViewColumn("Name",
            renderer_editable_name_3, text=0)
        column_editable_name_3.set_expand(True)
        column_editable_name_3.set_resizable(True)
        self.view['tv_tfpg_modes_in_synth'].append_column(column_editable_name_3)

        renderer_editable_name_3.connect("edited",
                                         self.association_name_edited_in_synth,
                                         modes_liststore, 'tv_tfpg_modes_in_synth')
                                         

        renderer_def_3 = gtk.CellRendererText()
        column_def_3 = gtk.TreeViewColumn("", renderer_def_3, text=1)
        self.view['tv_tfpg_modes_in_synth'].append_column(column_def_3)

        renderer_editable_expr_3 = gtk.CellRendererText()
        renderer_editable_expr_3.set_property("editable", True)

        column_editable_expr_3 = gtk.TreeViewColumn("Expression",
                                                    renderer_editable_expr_3,
                                                    text=2)
        column_editable_expr_3.set_expand(True)
        column_editable_expr_3.set_resizable(True)
        self.view['tv_tfpg_modes_in_synth'].append_column(column_editable_expr_3)
        
        renderer_editable_expr_3.connect("edited", 
                                         self.association_edited_in_synth,
                                         modes_liststore,
                                         'tv_tfpg_modes_in_synth')

        renderer_expr_result_3 = gtk.CellRendererText()
        column_result_3 = gtk.TreeViewColumn("Result", renderer_expr_result_3, text=3)
        column_result_3.set_sizing(gtk.TREE_VIEW_COLUMN_FIXED)
        column_result_3.set_resizable(True)
        column_result_3.set_fixed_width(130)
        self.view['tv_tfpg_modes_in_synth'].append_column(column_result_3)

        sel = self.view['tv_tfpg_modes_in_synth'].get_selection()
        sel.set_mode(gtk.SELECTION_BROWSE)
        sel.connect("changed", self.result_clicked_in_synth, 'tv_tfpg_modes_in_synth')

        self.view['vbox_associations_in_synth'].show_all()

        # Initially, all the text entries are empty; hence, we cannot save
        self.view['button_save_associations_in_synth'].set_sensitive(False)

        return slim_associations



    def association_edited_in_synth(self, widget, path, text, liststore, tv_name):
        """Function called when the user inserts or updates a slim expression
            associated to a tfpg mode.
            Notice that this function is called in the case the user modifies
            a discrepancy. Obviously, it is not called in case of failure modes,
            as they are selectable through a drop-down list.
	    @param    widget: The widget of slim_expression.
            @type     widget: gtk.CellRenderedText
            @param    path: The path of the slim_expression.
            @type     path: int
            @param    text: The text of the slim_expression.
            @type     text: string
	    @param    liststore: The liststore of the slim_expression.
            @type     liststore: gtk.ListStore
            @param    tv_name: The name of the tree view (monitored, not monitored or tfpg mode)
            @type     tv_name: string
        """
        
        error_list, impl_name = get_errors_in_atomic_proposition_with_templates(text, self.model.slim_model)

        if self.curr_tfpg_name is None:
            slim_associations = dict()
        else:

            ## In this case we get the slim associations from the GUI, 
            ## without building the internal structure, because the user
            ## can modify the slim associations in TFPG Synthesis tab.
            ## The internal structure is built at the moment of TFPG 
            ## Synthesis performing and the self.curr_tfpg_name is
            ## the name of the TFPG clicked in the TFPGs List.

            sel = self.parent_view['tv_tfpg_list'].get_selection()
            tree_model, paths = sel.get_selected_rows()
        
            if paths:
                tfpg_structural = tree_model.get_value(tree_model.get_iter(paths[0]), 0)
                slim_associations = self.model.get_slim_associations_in_synth_from_gui(tfpg_structural.get_name(),
                                                          self.view['tv_failure_modes_in_synth'].get_model(),
                                                          self.view['tv_monitored_discr_in_synth'].get_model(),
                                                          self.view['tv_not_monitored_discr_in_synth'].get_model(),
                                                          self.view['tv_tfpg_modes_in_synth'].get_model())
        

        ## Getting the type of association depending on the tree view we have selected
        
        _type = None
        if tv_name == "tv_monitored_discr_in_synth":
            _type = "MonitoredDiscr"
        elif tv_name == "tv_not_monitored_discr_in_synth":
            _type = "UnmonitoredDiscr"
        elif tv_name == "tv_tfpg_modes_in_synth":
            _type = "TfpgMode"
        else:
            raise Exception("Unknown Tree View")

            
        if error_list == []:

            check_errors = self.check_entry_in_synth(text, self.model.slim_model, tv_name)

            if check_errors == []:
                liststore[path][2] = text
                slim_associations[liststore[path][0]] = ("OK", text, _type)
                    
                # The message NOT COMPLETED in the Results Column means that
                # the slim_expressions are not filled in the related fields. 
                # So the button Run Synthesis is disabled until all the fields
                # are completed without errors.
                
                if (liststore[path][0] is None or liststore[path][0] == ""):
                    liststore[path][3] = "NOT COMPLETED"
                else:
                    liststore[path][3] = "OK"
                    self.view.enable_synthesis(self.all_associations_ok_in_synth())
                    self.view['button_save_associations_in_synth'].set_sensitive(self.all_associations_ok_in_synth())
                    liststore.append(["", ":=", None, None])
                               
                self.balloon_in_synth.hide_all()
     
            else:
                error = str(check_errors[0])
                self.show_association_error_in_synth(error, tv_name, path)
                liststore[path][2] = text
                slim_associations[liststore[path][0]] = ("ERROR", error, _type)
                liststore[path][3] = "ERROR"
                self.view.enable_synthesis(False)
                self.view['button_save_associations_in_synth'].set_sensitive(False)
                    
        else:
            error = str(error_list[0])
            self.show_association_error_in_synth(error, tv_name, path)
            liststore[path][2] = text
            slim_associations[liststore[path][0]] = ("ERROR", error, _type)
            liststore[path][3] = "ERROR"
            self.view.enable_synthesis(False)
            self.view['button_save_associations_in_synth'].set_sensitive(False)
            
        return


    def association_name_edited_in_synth(self, widget, path, text, liststore, tv_name):
        """Function called when the user inserts or updates a SLIM Associations name.
            @param    widget: The widget of slim_expression.
            @type     widget: gtk.CellRenderedText
            @param    path: The path of the slim_expression.
            @type     path: int
            @param    text: The text of the name of slim_expression.
            @type     text: string
            @param    liststore: The liststore of the slim_expression.
            @type     liststore: gtk.ListStore
            @param    tv_name: The name of the window
            @type     tv_name: string
        """
        
        if self.curr_tfpg_name is None:
            slim_associations = dict()
        else:
        
            ## In this case we get the slim associations from the GUI, 
            ## without building the internal structure, because the user
            ## can modify the slim associations in TFPG Synthesis tab.
            ## The internal structure is built at the moment of TFPG 
            ## Synthesis performing and the self.curr_tfpg_name is
            ## the name of the TFPG clicked in the TFPGs List.
            
            sel = self.parent_view['tv_tfpg_list'].get_selection()
            tree_model, paths = sel.get_selected_rows()
        
            if paths:
                tfpg_structural = tree_model.get_value(tree_model.get_iter(paths[0]), 0)
                slim_associations = self.model.get_slim_associations_in_synth_from_gui(tfpg_structural.get_name(),
                                        self.view['tv_failure_modes_in_synth'].get_model(),
                                        self.view['tv_monitored_discr_in_synth'].get_model(),
                                        self.view['tv_not_monitored_discr_in_synth'].get_model(),
                                        self.view['tv_tfpg_modes_in_synth'].get_model())
       
        ## Getting the type of association depending on the tree view we have selected
        
        _type = None
        if tv_name == "tv_monitored_discr_in_synth":
            _type = "MonitoredDiscr"
        elif tv_name == "tv_not_monitored_discr_in_synth":
            _type = "UnmonitoredDiscr"
        elif tv_name == "tv_tfpg_modes_in_synth":
            _type = "TfpgMode"
        else:
            _type = "FailureMode"
        
        if text != "":
        
            liststore[path][0] = text

            # The message NOT COMPLETED in the Results Column means that the names 
            # of the SLIM Associaitons are not filled in the related fields. 
            # So the button Run Synthesis is disabled until all the fields
            # are completed without errors.
                
            if (liststore[path][2] is None or liststore[path][2] == ""):
                liststore[path][3] = "NOT COMPLETED"
                self.view.enable_synthesis(False)
                self.view['button_save_associations_in_synth'].set_sensitive(False)
            else:
                slim_associations[liststore[path][0]] = ("OK", liststore[path][2], _type)
                liststore[path][3] = "OK"
                self.view.enable_synthesis(self.all_associations_ok_in_synth())
                self.view['button_save_associations_in_synth'].set_sensitive(self.all_associations_ok_in_synth())
                liststore.append(["", ":=", None, None])
                
            self.balloon_in_synth.hide_all()
        
        else:
            self.view.enable_synthesis(False)
            self.view['button_save_associations_in_synth'].set_sensitive(False)
        
        return



    def combo_changed_in_synth(self, widget, path, text, liststore):
        """Function called when the user chooses which slim expression of the drop-down
            list must be associated to a failure mode.
            Notice that this function is always called within the treeview tv_failure_modes.
            @param    widget: The widget of slim_expression.
            @type     widget: gtk.CellRenderedText
            @param    path: The path of the slim_expression.
            @type     path: int
            @param    text: The text of the slim_expression.
            @type     text: string
            @param    liststore: The liststore of the slim_expression.
            @type     liststore: gtk.ListStore
        """
        
        if self.curr_tfpg_name is None:
            slim_associations = dict()
        else:
            ## In this case we get the slim associations from the GUI, 
            ## without building the internal structure, because the user
            ## can modify the slim associations in TFPG Synthesis tab.
            ## The internal structure is built at the moment of TFPG 
            ## Synthesis performing and the self.curr_tfpg_name is
            ## the name of the TFPG clicked in the TFPGs List.
            
            sel = self.parent_view['tv_tfpg_list'].get_selection()
            tree_model, paths = sel.get_selected_rows()
        
            if paths:
                tfpg_structural = tree_model.get_value(tree_model.get_iter(paths[0]), 0)
                slim_associations = self.model.get_slim_associations_in_synth_from_gui(tfpg_structural.get_name(),
                                        self.view['tv_failure_modes_in_synth'].get_model(),
                                        self.view['tv_monitored_discr_in_synth'].get_model(),
                                        self.view['tv_not_monitored_discr_in_synth'].get_model(),
                                        self.view['tv_tfpg_modes_in_synth'].get_model())

        liststore[path][2] = text
        slim_associations[liststore[path][0]] = ("OK", text, "FailureMode")
        
        # The message NOT COMPLETED in the Results Column means that the  
        # SLIM expression of the combobox are not filled in the related fields. 
        # So the button Run Synthesis is disabled until all the fields
        # are completed without errors.
               
        if (liststore[path][0] is None or liststore[path][0] == ""):
            liststore[path][3] = "NOT COMPLETED"
            self.view.enable_synthesis(False)
            self.view['button_save_associations_in_synth'].set_sensitive(False)
        else:
            liststore[path][3] = "OK"
            self.view.enable_synthesis(self.all_associations_ok_in_synth())
            self.view['button_save_associations_in_synth'].set_sensitive(self.all_associations_ok_in_synth())
            liststore.append(["", ":=", None, None])
            
        self.balloon_in_synth.hide_all()

        return



    def show_association_error_in_synth(self, text, tv_name, path):
        """Function which shows the error derived from the insertion of a wrong slim expression.
           The error is shown in a balloon which appears unders the wrong slim expression
            @param    text: The text of the wrong slim_expression.
            @type     text: string
            @param    tv_name: The name of the window (monitored or not monitored)
            @type     tv_name: string
            @param    path: The path of the wrong slim_expression.
            @type     path: int
        """

        window_x, window_y = self.view[tv_name].get_parent_window().get_origin()
        bottom_left_x = self.view[tv_name].allocation.x + window_x
        bottom_left_y = self.view[tv_name].allocation.y + window_y + 50 + (int(path)*22)

        self.balloon_in_synth.set_position(gtk.WIN_POS_MOUSE)
        self.txtview_in_synth.modify_base(gtk.STATE_NORMAL, gtk.gdk.color_parse("indian red"))
        self.txtbuf_in_synth.set_text(text)
        self.balloon_in_synth.resize(400, 30)
        self.balloon_in_synth.show_all()
        self.balloon_in_synth.move(bottom_left_x, bottom_left_y)

        return



    def all_associations_ok_in_synth(self):
        """Function which checks that all the inserted slim expressions are correct.
            @return    True or False: True if all associated slim expressions are correct, else False.
            @rtype     True or False: boolean
        """
	            
        if self.curr_tfpg_name is None:
            slim_associations = dict()
       
        else:
            
            ## In this case we get the slim associations from the GUI, 
            ## without building the internal structure, because the user
            ## can modify the slim associations in TFPG Synthesis tab.
            ## The internal structure is built at the moment of TFPG 
            ## Synthesis performing and the self.curr_tfpg_name is
            ## the name of the TFPG clicked in the TFPGs List.
            
            sel = self.parent_view['tv_tfpg_list'].get_selection()
            tree_model, paths = sel.get_selected_rows()
        
            if paths:
                tfpg_structural = tree_model.get_value(tree_model.get_iter(paths[0]), 0)
                slim_associations = self.model.get_slim_associations_in_synth_from_gui(tfpg_structural.get_name(),
                                        self.view['tv_failure_modes_in_synth'].get_model(),
                                        self.view['tv_monitored_discr_in_synth'].get_model(),
                                        self.view['tv_not_monitored_discr_in_synth'].get_model(),
                                        self.view['tv_tfpg_modes_in_synth'].get_model())
                                        

        for (k, v) in slim_associations.items():
            (val, error, _type) = v                       
            ## If a SLIM Associations is NOT COMPLETED or ERROR, the return value is False.
            ## So the button Run Synthesis is disabled.
            if val != "OK" : return False
        
        return True



    def result_clicked_in_synth(self, selection, tv_name):
        """Function called when the user clicks over one row of the list of slim expressions.
            In case the selected row contains a wrong slim expression, the relative error is
            shown.
            @param    selection: The slim_expression selected.
            @type     selection: gtk.TreeSelection
            @param    tv_name: The name of the window (monitored or not monitored)
            @type     tv_name: string
	    """
        self.balloon_in_synth.hide_all()
        tree_model, tree_iter = selection.get_selected()

        if tree_iter != None:
            checked = tree_model[tree_iter][3]
            ## If a SLIM Association is wrong, the related error message appears.
            ## If the SLIM Association is not completed, no dialog messages appear.
            if checked == "ERROR":
                mode = tree_model[tree_iter][0]
                (val, error, _type) = slim_associations[mode]
                self.show_association_error_in_synth(error, tv_name, 
                                            tree_model.get_path(tree_iter)[0])
            
        return



    def check_entry_in_synth(self, text, slim_model, tv_name):
        """Function which checks a syntactically correct slim expression entered by the user.
           Notice that this function requires also the slim_model, as it is used to check the
           expressions.
            @param    text: The slim_expression.
            @type     text: string
            @param    slim_model: The slim_model.
            @type     slim_model: SlimModel
            @param    tv_name: The name of the window (monitored, not monitored or tfpg modes)
            @type     tv_name: string

            @return    error_list: The list with the errors of wrong slim_expressions.
            @rtype     error_list: list
        """
        import slim_compiler.src.smv.api as smv_api

        error_list = []

        ## In case of monitored discrepancies, the following check must be performed
        ## - Expressions on Monitored discrepancies must contain only observable variables
        if tv_name == "tv_monitored_discr_in_synth" or tv_name == "tv_tfpg_modes_in_synth":

            error_list, _ = functions.get_errors_in_atomic_proposition_with_templates_observable(text, self.model.slim_model)

        ## In case of not-monitored discrepancies, the following check must be performed.
        ## - Discrepancies expressions must contain only data or mode variables from the system
        ##   This means that event ports and everything related to error models bust be excluded

        elif tv_name == "tv_not_monitored_discr_in_synth":

            root_impl = slim_model.get_nominal_root_impl()
            _type = root_impl.parent.get_component_type(root_impl.name.split(".")[0])
            event_ports = dict(_type.in_event_ports.items() + _type.out_event_ports.items())

            # GZ: maybe the following should be contained in a function
            # discrepancies expressions are generic boolean expressions
            # to check that they refer only on observable variable, we perform the following steps:
            # - replacement of all boolean operators with whitespaces
            #   notice that a boolean operator is here defined as a string ('and', 'or', ...)
            #   between spaces, such that it is not confused with part of a variabel name, or
            #   as 'not ', if at the beginning of the expression
            # - thus, we split the expression removing whitespaces; what remains is the identifier
            #   of a variable or a boolean value ('true' or 'false')
            if text[:4] == "not ":
                text = text.replace(text[:4], " ")
            identifiers = text.replace(" and ", " ")
            identifiers = identifiers.replace(" iff ", " ")
            identifiers = identifiers.replace(" imp ", " ")
            identifiers = identifiers.replace(" not ", " ")
            identifiers = identifiers.replace(" or ", " ")
            identifiers = identifiers.replace(" xnor ", " ")
            identifiers = identifiers.replace(" xor ", " ")
            identifiers = identifiers.replace(" = ", " ")
            identifiers = " ".join(identifiers.split())
            ids = identifiers.split(" ");
            for ident in ids:
                ident = ident.strip()
                if ident != "false" and ident != "true":
                    if ident in event_ports:
                        error = "%s is an event port and cannot be contained in a discrepancy expression.\n" %(ident)
                        error_list.append(error)

        return error_list



    def load_associations_list_in_synth(self, associations):
        """It loads in the associations tab a given dictionary of associations.
            This function is called in case the user switches from a tfpg to another one by
            clicking on the list on the left. In this case, the tfpg-related associations
            list must be loaded.
            @param    associations: The dict of associations.
            @type     associations: dict
        """
        for (k, v) in associations.items():
            var_name = k
            error = self.load_single_association_in_synth(var_name, associations[k])
            if error is True:
                self.parent_view.tfpg_errors_view.errors_buf.insert_with_tags_by_name(
                    self.parent_view.tfpg_errors_view.errors_buf.get_end_iter(),
                    "ERROR:\n\n It is not possible to load the Slim Associations before " +
                    "to load the TFPG.\n", "errors")
                self.parent_view.show_page(5)
                return

        self.view['vbox_associations_in_synth'].show_all()
        
        return



    def load_single_association_in_synth(self, var_name, assoc):
        """It loads a single association to the associations tab.
            This function takes two parameters, var_name, the name of the variable to update
            (first column of the treeview) and assoc, a tuple consisting of a value ("OK" or 
            "ERROR") saying whether it has a valid slim expression associated, and of the 
            associated slim expression.
            @param    var_name: The name of the variable to update.
            @type     var_name: string
            @param    assoc: the tuple containing the slim expresison and a value saying whether it is valid or not.
            @type     assoc: tuple(str, str)
            @return   error: value saying whether the insertion succeeded or not
            @rtype    error: boolean
        """

        error = False

        (ok, var_value, _type) = assoc

        if _type == "FailureMode":
            model = self.view['tv_failure_modes_in_synth'].get_model()
            n = 0
            if model is not None:
                while n < len(model):
                    if model[n][0] == var_name:
                        if ok == "OK":
                            model[n][2] = var_value
                            model[n][3] = ok
                        else:
                            model[n][2] = None
                            self.view['tv_failure_modes_in_synth'].set_model(model)
                            return
                    n = n+1
            else:
                error = True

        elif _type == "MonitoredDiscr":
            n = 0
            model = self.view['tv_monitored_discr_in_synth'].get_model()

            if model is not None:
                while n < len(model):
                    if model[n][0] == var_name:
                        if ok == "OK":
                            model[n][2] = var_value
                            model[n][3] = ok
                        else:
                            model[n][2] = None
                            self.view['tv_monitored_discr_in_synth'].set_model(model)
                            return
                    n = n+1
            else:
                error = True
 
        elif _type == "UnmonitoredDiscr":
            n = 0
            model = self.view['tv_not_monitored_discr_in_synth'].get_model()

            if model is not None:
                while n < len(model):
                    if model[n][0] == var_name:
                        if ok == "OK":
                            model[n][2] = var_value
                            model[n][3] = ok
                        else:
                            model[n][2] = None
                            self.view['tv_not_monitored_discr_in_synth'].set_model(model)
                            return
                    n = n+1
            else:
                error = True

        elif _type == "TfpgMode":
            n = 0
            model = self.view['tv_tfpg_modes_in_synth'].get_model()
            if model is not None:
                while n < len(model):
                    if model[n][0] == var_name:
                        if ok == "OK":
                            model[n][2] = var_value
                            model[n][3] = ok
                        else:
                            model[n][2] = None
                            self.view['tv_tfpg_modes_in_synth'].set_model(model)
                            return
                    n = n + 1
            else:
                error = True

        else:
            error = True
            raise Exception("Unknown association type: " + str(_type))
         
        return error



    def on_button_save_associations_clicked_in_synth(self, action):
        """Saves a tfpg-slim_expr associations document in .axml format."""

        from hasdel_tools.src.xml.tfpg_slim_associations import create_associations_document_from_list

        dialog = gtk.FileChooserDialog("Save Associations",
                                        None,
                                        gtk.FILE_CHOOSER_ACTION_SAVE,
                                        (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                                        gtk.STOCK_SAVE, gtk.RESPONSE_OK))
        dialog.set_default_response(gtk.RESPONSE_OK)
        filter = gtk.FileFilter()
        filter.set_name("FAME TFPG ASSOC")
        filter.add_pattern("*.axml")
        dialog.add_filter(filter)
        dialog.set_current_name(os.path.basename(
                generate_unique_tfpg_xml_basename() + ".axml"))
        dialog.set_do_overwrite_confirmation(True)    
        dialog.show()

        response = dialog.run()
        
        if response == gtk.RESPONSE_OK:
            filename = dialog.get_filename()

            # The internal structure of slim associations is created and then removed
            # Get the SLIM associations from the TFPG Synthesis Tab of the GUI
            input_assoc = self.model.get_slim_associations_in_synth_from_gui("tmp_tfpg",
                                        self.view['tv_failure_modes_in_synth'].get_model(),
                                        self.view['tv_monitored_discr_in_synth'].get_model(),
                                        self.view['tv_not_monitored_discr_in_synth'].get_model(),
                                        self.view['tv_tfpg_modes_in_synth'].get_model())

            assoc_doc = create_associations_document_from_list(input_assoc)
            
            with open(filename, "w") as f:
                f.write(str(assoc_doc))
                pass
                    
            dialog.destroy()
        return



    def on_button_load_associations_clicked_in_synth(self, action):
        """Load a tfpg-slim_expr associations document in .axml format.
           @param   action:  The action to enable the button self.view['button_load_associations']
	   @type    action:  gtk.Button
        """
	    
        chooser = gtk.FileChooserDialog(title="Open TFPG SLIM Associations",
                                        action=gtk.FILE_CHOOSER_ACTION_OPEN,
                                        buttons=(gtk.STOCK_CANCEL,gtk.RESPONSE_CANCEL,
                                                 gtk.STOCK_OPEN,
                                                 gtk.RESPONSE_OK))
        chooser.set_select_multiple(False)
        filter = gtk.FileFilter()
        filter.set_name("FAME TFPG SLIM ASSOC")
        filter.add_pattern("*.axml")
        chooser.add_filter(filter)
        chooser.set_current_folder(os.path.dirname(
            generate_unique_tfpg_xml_basename() + ".axml"))

        response = chooser.run()
        if response == gtk.RESPONSE_OK:
            ## First, we check whether there are validation/syntax errors in the 
            ## associations file
            
            input_assoc, errors_ls, fm_nodes_list, mon_discr_list, unmon_discr_list, tfpg_modes_list = \
                        self.model.get_slim_associations_in_synth_from_file(chooser.get_filename(), self.curr_tfpg_name)
            
            associations = self.initialize_tfpg_model_tab_in_synth(fm_nodes_list,
                                                                   mon_discr_list,
                                                                   unmon_discr_list,
                                                                   tfpg_modes_list)
            
            if errors_ls == []:
                ## If the associations file is syntactically correct, we check that there
                ## are no errors in the slim expressions
                assoc_ls, error = self.check_associations_list_in_synth(input_assoc)
            
                if error is False:
                    self.load_associations_list_in_synth(assoc_ls)
                    self.view.enable_synthesis(True)
 
                else:
                    self.parent_view.tfpg_errors_view.errors_buf.insert_with_tags_by_name(
                        self.parent_view.tfpg_errors_view.errors_buf.get_end_iter(),
                        "In Synthesis Tab -> Slim Association Syntax ERROR:\n\n " + self.slim_expr_error + "\n",
                        "errors")
                    self.parent_view.show_page(5)

            else:
                ## errors_ls != [] -> we must show the syntax/validation errors found in 
                ## the associations file
                for err in errors_ls:
                    self.parent_view.tfpg_errors_view.errors_buf.insert_with_tags_by_name(
                        self.parent_view.tfpg_errors_view.errors_buf.get_end_iter(),
                        str(err),
                        "errors")
                    self.parent_view.show_page(5)
 
        chooser.destroy()
        return



    def check_associations_list_in_synth(self, input_assoc):
        """Function which checks a syntactically correct list of slim expression.
	    @param    input_assoc: The slim association to check.
            @type     input_assoc: string

	    @return    valid_assoc: The dict of valid slim associations.
            @rtype     valid_assoc: dict
        """
        error = False
        valid_assoc = dict()

        for (name, assoc) in input_assoc.items():
            (ok, expr, _type) = assoc
            error_list, impl_name = get_errors_in_atomic_proposition_with_templates(
                expr, self.model.slim_model)

            if error_list == []:
                    
                if _type == "FailureMode":
                    model = self.view['tv_failure_modes_in_synth'].get_model()
                    n = 0
                    if model is not None:
                        while n < len(model):
                            if model[n][0] == name:
                                fm_expr_found = False
                                for row in self.get_fm_possible_slim_associations_in_synth():
                                    if expr == row[0]: fm_expr_found = True

                                if fm_expr_found is True: valid_assoc[name] = ("OK", expr, _type)
                                else: valid_assoc[name] = ("ERROR", expr, _type)
                            n = n+1

                elif _type == "MonitoredDiscr":

                    model = self.view['tv_monitored_discr_in_synth'].get_model()
                    n = 0
                    if model is not None:
                        while n < len(model):
                            if model[n][0] == name:
                                if self.check_entry_in_synth(expr,self.model.slim_model,
                                                    'tv_monitored_discr_in_synth') == []:
                                    valid_assoc[name] = ("OK", expr ,_type)
                                else:
                                    valid_assoc[name] = ("ERROR", expr, _type)
                            n = n+1
 
                elif _type == "UnmonitoredDiscr":
                    model = self.view['tv_not_monitored_discr_in_synth'].get_model()
                    n = 0
                    if model is not None:
                        while n < len(model):
                            if model[n][0] == name:
                                if self.check_entry_in_synth(expr,self.model.slim_model,
                                                    'tv_not_monitored_discr_in_synth') == []:
                                    valid_assoc[name] = ("OK", expr, _type)
                                else:
                                    valid_assoc[name] = ("ERROR", expr, _type)
                            n = n+1
                                   
                elif _type == "TfpgMode":
                    model = self.view['tv_tfpg_modes_in_synth'].get_model()
                    n = 0
                    if model is not None:
                        while n < len(model):
                            if model[n][0] == name:
                                if self.check_entry_in_synth(expr,self.model.slim_model,
                                                    'tv_tfpg_modes_in_synth') == []:
                                    valid_assoc[name] = ("OK", expr, _type)
                                else:
                                    valid_assoc[name] = ("ERROR", expr, _type)
                            n = n + 1
                else:         
                    ## !!Warning!! This code should never be reached
                    raise Exception("Invalid association type: " + str(_type))
 
            else:
                self.slim_expr_error = "The SLIM Expression ' " + expr + " ' is wrong:\n\t" + str(error_list[0])
                error = True
 
        return valid_assoc, error



    def get_fm_possible_slim_associations_in_synth(self):
        """Returns the possible slim expressions that can be associated to a failure mode.
	   @return    possible_fm_expr: The listore of possible slim expressions that can be 
					associated to a failure mode in string format.
           @rtype     possible_fm_expr: gtk.ListStore
	"""
        possible_fm_expr = gtk.ListStore(str)

        for injection in self.model.slim_model.fault_injections:
           err_state = injection[3]
           component_path = ".".join(injection[4])
           expr = "%s.error = error:%s" %(component_path, err_state)
           possible_fm_expr.append([expr])

        return possible_fm_expr



    def on_button_stop_synthesis_clicked(self, button):
        """Stops the synthesis when the button_stop_synthesis is clicked.
           @param   button:  The Stop Button.
           @type    button:  self.view['button_stop_synthesis']
        """
        self.view.synthesis_buf.insert_with_tags_by_name(self.view.synthesis_buf.get_end_iter(),
                    "Warning! TFPG Synthesis Stopped!\n", "synthesis")          
        return



    
pass # end of TfpgSynthesisCtrl class
# ----------------------------------------------------------------------


class TfpgBehavValCtrl (StartStopActivityCtrl):
    """This is a sub-controller for handling behavioural validation activities"""
        
        
    behav_val_counter = 0
   
    # Observable properties
    __observables__ = ("behav_val_counter", )
    
    
    def __init__(self, parent_model, model, parent_view, view):
        """Constructor Method"""
        StartStopActivityCtrl.__init__(self, model, view, 
                                       self.run_behav_val, model.stop_behav_val)

        self.parent_model = parent_model
        self.parent_view = parent_view
        
        self.results_ctrl_system = None
        self.results_ctrl_tfpg = None

        self.observe_model(model.slim_model)
        return


    def register_view(self, view): 
        """Sets the windows, dialog, buttons and other parts of View."""
        StartStopActivityCtrl.register_view(self, view)
        
        view['win_beh_val_detached'].connect("delete-event", 
                                        self.on_window_beh_val_detached_delete_event)
        
        self.trace_is_detached = False
        return
        

    def on_window_beh_val_detached_delete_event(self, win, event):
        """When the delete button is clicked, the hbox results is re-attached."""
        self.view.attach()
        self.trace_is_detached = False
        return True


    def on_button_beh_val_detach_clicked(self, button):
        """When the detach button is clicked, the hbox results is detached."""
        self.view.detach()
        self.trace_is_detached = True
        return
  
        
    def register_adapters(self):
        self.adapt("behav_val_sat_bound")
        self.adapt("behav_val_use_sat")
        return

    
    def run_behav_val(self):
        """Runs TFPG Behavioural Validation"""
        
        if self.behav_val_counter >= 1:
            self.view.clear()

        sel = self.parent_view['tv_tfpg_list'].get_selection()
        tree_model, paths = sel.get_selected_rows()
        
        if paths:
            tfpg_structural = tree_model.get_value(tree_model.get_iter(paths[0]), 0)
            continuous_assoc = self.model.continuous_var_in_associations(tfpg_structural.get_name())
            if continuous_assoc != []:
                error_message = "The following expressions refer to <b>continuous variables</b>"
                error_messages = ""
                for assoc in continuous_assoc:
                    error_messages = "%s\n * %s" %(error_messages, assoc)
                error_messages = '%s\nThe results of the analysis may be not trustable ' %error_messages
                res = self.parent_view.ui_show_message_dialog('warning', str(error_message), 
                                                                         parent_name='window_top',
                                                                         secondary_text=str(error_messages),
                                                                         buttons=gtk.BUTTONS_OK)

            self.model.validate_behavior(tfpg_structural)
            self.behav_val_counter += 1
            
        return 
        
           
           
    # ---------------------------------------------------------------
    #        Notifications from observable properties
    # ---------------------------------------------------------------    


 
    def property_behav_val_result_value_change(self, model, old, new):          
        """When a new diagnosability analysis result is created, this method
            will be advised to display it.
        """
        
        # relieve old results

        if self.results_ctrl_system != None:
            assert self.results_ctrl_tfpg is not None, "Toolchain is broken (DIAG)."
            self.results_ctrl_system.relieve_model(self.results_ctrl_system.model)
            self.results_ctrl_system = None
            self.results_ctrl_tfpg.relieve_model(self.results_ctrl_tfpg.model)
            self.results_ctrl_tfpg = None
            pass

        if new is None: return

        assert old != new, "Toolchain is broken (DIAG)."
        
        if new.success in ("error", "stop", "complete", "bounded_completeness", "no associations"):
            self.view.display_behav_val_results(new.success, None, None, None)
            self.behav_val_counter += 1
            return
            
        if (self.model.behav_val_result.result_system or self.model.behav_val_result.result_tfpg) is None: 
            self.behav_val_counter += 1
            return

        # instantiate the various handlers

        tvm_system = TraceViewModel()
        tvm_tfpg = TraceViewModel()

        trace_system = self.model.behav_val_result.result_system.trace
        trace_tfpg = self.model.behav_val_result.result_tfpg.trace
        counterexample_description = self.model.behav_val_result.counterexample_description

        tvm_system.add_traces(trace_system)
        tvm_tfpg.add_traces(trace_tfpg)

        tvm_system.set_active_trace(trace_system)
        tvm_tfpg.set_active_trace(trace_tfpg)

        trace_view_system = TraceView(windowed=False, filters=True)
        trace_view_tfpg = TraceView(windowed=False, filters=True)

        self.results_ctrl_system = TraceCtrl(tvm_system, trace_view_system)
        self.results_ctrl_tfpg = TraceCtrl(tvm_tfpg, trace_view_tfpg)
        
        # synchronize scrollbars of trace windows

        swtop_C = trace_view_system['sw_top']
        swtop_D = trace_view_tfpg['sw_top']

        adj_v = gtk.Adjustment(
            value=0.00,
            lower=0.00,
            upper=100.00,
            step_incr=1.00,
            page_incr=10.00,
            page_size=10.00)

        adj_h = gtk.Adjustment(
          value=0.00,
          lower=0.00,
          upper=100.00,
          step_incr=1.00,
          page_incr=10.00,
          page_size=10.00)

        swtop_C.set_hadjustment( adj_h )
        swtop_C.set_hadjustment( adj_h )

        swtop_D.set_vadjustment( adj_v )
        swtop_D.set_vadjustment( adj_v )

        # tell the view to display the traces        
        self.view.display_behav_val_results(new.success, counterexample_description, trace_view_system, trace_view_tfpg)
        return

    
pass # end of TfpgBehavValCtrl class
# ----------------------------------------------------------------------


class TfpgEffectValCtrl (StartStopActivityCtrl):
    """This is a sub-controller for handling effectiveness validation activities"""
        
        
    effect_val_counter = 0
   
    # Observable properties
    __observables__ = ("effect_val_counter", )
    
    
    def __init__(self, parent_model, model, parent_view, view):
        """Constructor Method"""
        StartStopActivityCtrl.__init__(self, model, view, 
                                       self.run_effect_val, model.stop_effect_val)

        self.parent_model = parent_model
        self.parent_view = parent_view
        
        self.results_ctrl_A = None
        self.results_ctrl_B = None

        # The treeview of results is created and the column will be filled 
        # with the results in the view.get_messages_list().
        tv = view['treeview_results']
        tv.set_model(model=view.get_messages_list())

        renderer_text = gtk.CellRendererText()
        column_text = gtk.TreeViewColumn("Message/Result", renderer_text, text=0)
        tv.append_column(column_text)
        
        # To control selection of results
        sel = view['treeview_results'].get_selection()
        sel.set_mode(gtk.SELECTION_MULTIPLE)
        sel.connect('changed', self.on_results_selection_changed)
        
        self.observe_model(model.slim_model)
        return


    def register_view(self, view): 
        """Sets the windows, dialog, buttons and other parts of View."""
        StartStopActivityCtrl.register_view(self, view)   
        view['win_eff_val_detached'].connect("delete-event", 
                                        self.on_window_eff_val_detached_delete_event)
        
        self.trace_is_detached = False
        return
        

    def on_window_eff_val_detached_delete_event(self, win, event):
        """When the delete button is clicked, the hbox results is re-attached."""
        self.view.attach()
        self.trace_is_detached = False
        return True


    def on_button_eff_val_detach_clicked(self, button):
        """When the detach button is clicked, the hbox results is detached."""
        self.view.detach()
        self.trace_is_detached = True
        return

    
    def register_adapters(self):
        self.adapt("effect_val_sat_bound")
        self.adapt("effect_val_use_sat")
        return
        
        
    def on_results_selection_changed(self, selection):
        """Changes the results selected.
            @param   selection:  The selection of results.
            @type    selection:  gtk.TreeSelection
        """
        
        # Gets the tree_model and the path of the results selected
        tree_model, paths = selection.get_selected_rows()
        
        
        for effect_val_failure_system_mode in self.model.effect_val_results_list:
                
            # Selects a row and gets the value of this row.
            value = ""
            for path in paths:
                tree_iter = tree_model.get_iter(path)
                value = tree_model.get_value(tree_iter,0)
                break
                
            # Checks if the couple of failure mode - system mode are in row.
            if str(effect_val_failure_system_mode[1]) in value and str(effect_val_failure_system_mode[2]) in value:

                # If the result is None, no traces are shown and the button detach is hidden.
                if (effect_val_failure_system_mode[0].result_A or effect_val_failure_system_mode[0].result_B) is None: 
                    self.view.display_effect_val_results(None, None, "no trace", None, None)
                    return
            
                # Else the traces are created.
                
                # Instantiate the various handlers

                tvm_A = TraceViewModel()
                tvm_B = TraceViewModel()

                trace_A = effect_val_failure_system_mode[0].result_A.trace
                trace_B = effect_val_failure_system_mode[0].result_B.trace

                tvm_A.add_traces(trace_A)
                tvm_B.add_traces(trace_B)

                tvm_A.set_active_trace(trace_A)
                tvm_B.set_active_trace(trace_B)

                trace_view_A = TraceView(windowed=False, filters=True)
                trace_view_B = TraceView(windowed=False, filters=True)

                self.results_ctrl_A = TraceCtrl(tvm_A, trace_view_A)
                self.results_ctrl_B = TraceCtrl(tvm_B, trace_view_B)
                
                # synchronize scrollbars of trace windows

                swtop_A = trace_view_A['sw_top']
                swtop_B = trace_view_B['sw_top']

                adj_v = gtk.Adjustment(
                    value=0.00,
                    lower=0.00,
                    upper=100.00,
                    step_incr=1.00,
                    page_incr=10.00,
                    page_size=10.00)

                adj_h = gtk.Adjustment(
                  value=0.00,
                  lower=0.00,
                  upper=100.00,
                  step_incr=1.00,
                  page_incr=10.00,
                  page_size=10.00)

                swtop_A.set_hadjustment( adj_h )
                swtop_B.set_hadjustment( adj_h )

                swtop_A.set_vadjustment( adj_v )
                swtop_B.set_vadjustment( adj_v )

                # The trace and the button detach are shown.
                self.view.display_effect_val_results(None, None, "change trace", trace_view_A, trace_view_B)

        return



    def run_effect_val(self):
        """Runs TFPG Effectiveness Validation"""
        
        if self.effect_val_counter >= 1:
            self.view.clear()
           
        sel = self.parent_view['tv_tfpg_list'].get_selection()
        tree_model, paths = sel.get_selected_rows()
        
        if paths:

            # retrieve the selected tfpg
            tfpg_structural = tree_model.get_value(tree_model.get_iter(paths[0]), 0)

            self.model.validate_effectiveness(tfpg_structural)
            self.effect_val_counter += 1
            
        return 
           
           
    # ---------------------------------------------------------------
    #        Notifications from observable properties
    # ---------------------------------------------------------------    


    def property_use_bdd_value_change(self, model, old, new):
        """When a the use bdd value is changed, this method
            will be advised to display it.
        """
        self.view['table_sat'].set_sensitive(not new)
        return


    def property_can_use_bdd_value_change(self, model, old, new):
        """When a the can use bdd value is changed, this method
            will be advised to display it.
        """
        if not new:
            if self.view['rb_behav_val_use_bdd'].get_active():
                self.view['rb_behav_val_use_bdd'].set_active(False)
                self.view['rb_effect_val_use_sat'].set_active(True)
                pass
            pass

        self.view['rb_behav_val_use_bdd'].set_sensitive(new)
        return 

 
    def property_effect_val_result_value_change(self, model, old, new):         
        """When a new diagnosability analysis result is created, this method
            will be advised to display it.
        """
        
        # relieve old results

        if self.results_ctrl_A != None:
            assert self.results_ctrl_B is not None, "Toolchain is broken (DIAG)."
            self.results_ctrl_A.relieve_model(self.results_ctrl_A.model)
            self.results_ctrl_A = None
            self.results_ctrl_B.relieve_model(self.results_ctrl_B.model)
            self.results_ctrl_B = None
            pass

        if new is None: return

        assert old != new, "Toolchain is broken (DIAG)."

        if new.success in ("error", "diagnosability_unknown", "diagnosable", "alarms not enabled", "stop"):
            self.view.display_effect_val_results(self.model.last_system_mode_name, 
                                                 self.model.last_failure_mode_name, 
                                                 new.success, None, None)
            self.effect_val_counter += 1
            return

        
        if (self.model.effect_val_result.result_A or self.model.effect_val_result.result_B) is None: 
            self.effect_val_counter += 1
            return
            
        # instantiate the various handlers

        tvm_A = TraceViewModel()
        tvm_B = TraceViewModel()

        trace_A = self.model.effect_val_result.result_A.trace
        trace_B = self.model.effect_val_result.result_B.trace

        tvm_A.add_traces(trace_A)
        tvm_B.add_traces(trace_B)

        tvm_A.set_active_trace(trace_A)
        tvm_B.set_active_trace(trace_B)

        trace_view_A = TraceView(windowed=False, filters=True)
        trace_view_B = TraceView(windowed=False, filters=True)

        self.results_ctrl_A = TraceCtrl(tvm_A, trace_view_A)
        self.results_ctrl_B = TraceCtrl(tvm_B, trace_view_B)
        
        # synchronize scrollbars of trace windows

        swtop_A = trace_view_A['sw_top']
        swtop_B = trace_view_B['sw_top']

        adj_v = gtk.Adjustment(
            value=0.00,
            lower=0.00,
            upper=100.00,
            step_incr=1.00,
            page_incr=10.00,
            page_size=10.00)

        adj_h = gtk.Adjustment(
          value=0.00,
          lower=0.00,
          upper=100.00,
          step_incr=1.00,
          page_incr=10.00,
          page_size=10.00)

        swtop_A.set_hadjustment( adj_h )
        swtop_B.set_hadjustment( adj_h )

        swtop_A.set_vadjustment( adj_v )
        swtop_B.set_vadjustment( adj_v )

        # tell the view to display the traces        
        self.view.display_effect_val_results(self.model.last_system_mode_name,
                                                self.model.last_failure_mode_name, 
                                                new.success, trace_view_A, trace_view_B)
        return
        
        
        
pass # end of TfpgEffectValCtrl class
# ----------------------------------------------------------------------


class TfpgErrorsCtrl (Controller):
    """This is a sub-controller for handling errors activities"""
        
    def __init__(self, parent_model, model, parent_view, view):
        """Constructor Method"""
        Controller.__init__(self, model, view)

        self.parent_model = parent_model
        self.parent_view = parent_view
        
        #Set the TFPG Errors TextView
        view['tv_errors'].set_buffer(view.errors_buf)
       
        self.observe_model(model.slim_model)
        return


    def register_view(self, view): 
        """Sets the windows, dialog, buttons and other parts of View."""
        Controller.register_view(self, view)
        return
        
        
    def show_error_message(self, message):
        """Shows the error message in the errors subview.
            @param    message: The message to be shown.
            @type     message: string

                @return   error_message: The method that returns the error_message
            @rtype    error_message: ErrorsBuf
            
        """
        error_message = self.view.errors_buf.insert_with_tags_by_name(self.view.errors_buf.get_end_iter(), message, "errors")

        return error_message
    
   
pass # end of TfpgErrorsCtrl class
# ----------------------------------------------------------------------


class NavigationToolbar(NavigationToolbar2GTKAgg):
    """Represents the NavigationToolbar of Tfpg View Buttons."""

    toolitems = [t for t in NavigationToolbar2GTKAgg.toolitems if
                 t[0] in ('Home', 'Pan')]


pass # end of NavigationToolbar class
# ----------------------------------------------------------------------
