# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


from gtk import TextBuffer
import pango
import logging
import re

import utils.log
from gtkmvc import Model

class ConsoleModel (Model):
    """This is the model of the output console. At the moment it is
    located within the slim model."""

    def __init__(self):
        Model.__init__(self)
        # extended model
        self.extended_model_buf = TextBuffer()
        
        # compiler
        self.compiler_buf = TextBuffer()

        # metrics
        self.metrics_buf = TextBuffer()
        gh_metrics = utils.log.GuiHandler(self.metrics_buf)
        utils.log.metrics.logger.addHandler(gh_metrics)

        # experimental for logging
        self.log_buf = TextBuffer()
        gh = utils.log.GuiHandler(self.log_buf)
        gh.setLevel(logging.DEBUG)

        utils.log.user.logger.addHandler(gh)
        utils.log.system.logger.addHandler(gh)
        utils.log.testing.logger.addHandler(gh)

        self._create_style_tags()
        return

    def log_compiler_msg(self, text):
        """Write some text into the compiler log. This is capable
        of syntax HL"""
        text_hl = text # FIXME: implement SHL
        self.compiler_buf.insert_with_tags_by_name(self.compiler_buf.get_end_iter(), text_hl, 
                                                   "operation")
        return

    def log_compiler_error(self, fname, linenum, id, msg, desc):
        """Write some error text into the compiler log. This is capable
        of syntax HL"""
        
        if fname: 
            self.compiler_buf.insert_with_tags_by_name(self.compiler_buf.get_end_iter(), 
                                                       fname, "error-fname")
            self.compiler_buf.insert_with_tags_by_name(self.compiler_buf.get_end_iter(), 
                                                       ":%d:\n" % linenum, 
                                                       "error-line")
            pass

        self.compiler_buf.insert_with_tags_by_name(self.compiler_buf.get_end_iter(), 
                                                   "(%s) " % id, "error-line")

        self.compiler_buf.insert_with_tags_by_name(self.compiler_buf.get_end_iter(), 
                                                   msg +"\n", "error-msg")

        if desc: self.compiler_buf.insert_with_tags_by_name(self.compiler_buf.get_end_iter(), 
                                                            desc+"\n", "error-text")
        return


    def log_compiler_ok(self, text):
        """Write some text about successfull operation into the compiler log"""
        self.compiler_buf.insert_with_tags_by_name(self.compiler_buf.get_end_iter(), text,
                                                   "result_ok")
        return


    def log_compiler_fail(self, text):
        """Write some text about failed operation into the compiler log"""
        self.compiler_buf.insert_with_tags_by_name(self.compiler_buf.get_end_iter(), text,
                                                   "result_ko")
        return

    def clear_metrics(self):
        self.metrics_buf.delete(*self.metrics_buf.get_bounds())
        return

    def clear_log(self):
        self.log_buf.delete(*self.log_buf.get_bounds())
        return

    def clear_compiler_msg(self):
        self.compiler_buf.delete(*self.compiler_buf.get_bounds())
        return


    def _create_style_tags(self):
        """Creates some tags for styling output"""

        buf = self.compiler_buf

        buf.create_tag("spacer", pixels_above_lines=12)        

        buf.create_tag("result_ok", family="Monospace",
                       left_margin=12, foreground="dark green",
                       weight=pango.WEIGHT_BOLD,
                       #underline=pango.UNDERLINE_SINGLE,
                       pixels_below_lines=2)

        buf.create_tag("result_ko", family="Monospace",
                       left_margin=12, foreground="red",
                       weight=pango.WEIGHT_BOLD,
                       underline=pango.UNDERLINE_SINGLE,
                       pixels_below_lines=2)

        buf.create_tag("error-fname",
                       pixels_above_lines=6,
                       left_margin=24,
                       background="gray90")

        buf.create_tag("error-line", 
                       left_margin=24,
                       background="gray90",
                       foreground="darkred",
                       weight=pango.WEIGHT_BOLD)

        buf.create_tag("error-msg", 
                       left_margin=24,
                       style=pango.STYLE_OBLIQUE,
                       weight=pango.WEIGHT_BOLD)

        buf.create_tag("error-text", 
                       left_margin=24,
                       foreground="darkblue", family="Monospace")

        buf.create_tag("operation", family="Monospace",
                       foreground="gray40")

        buf.create_tag("normal", family="Sans")
        buf.create_tag("message", pixels_below_lines=0, foreground="gray30")
        buf.create_tag("timestamp", pixels_below_lines=0, foreground="gray40")

        self.log_buf.create_tag("timestamp", pixels_below_lines=0, foreground="gray50")
        self.log_buf.create_tag("level", pixels_below_lines=0, family="Monospace",
                                foreground="darkblue")

        self.metrics_buf.create_tag("timestamp", pixels_below_lines=0, foreground="gray50")
        self.metrics_buf.create_tag("level", pixels_below_lines=0, family="Monospace",
                                    foreground="darkblue")
        self.metrics_buf.create_tag("metrics", family="Monospace",
                                    left_margin=12, foreground="black")

                                    
        return
    
    pass # end of class ConsoleModel

    
