# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================



import os.path
import types
from gtk import ListStore, TreeStore, SORT_ASCENDING
import logging

from gtkmvc import Model
from gtkmvc import observable

from slim_compiler.src.dataStructure.nominalbehaviour import Continuous
from slim_compiler.src.dataStructure.nominalbehaviour import Clock
import slim_compiler.src.smv.support as smvsupport
from slim_compiler.src.model_extension.util import general
        
class FaultInjectionModel(Model):
    """Model of a fault injection window"""

    fault_injections = None
    compile_info = None
    root_impl = None

    # chosen error implemention
    error_implementation = None
    # chosen error state
    error_state = ""    
    # chosen component instance
    nominal_instance = None
    # chosen data port/sc
    data = ""
    # chosen effect
    effect = ""
    
    # ---- the options viewed in the fault injection pop up ------------------
    # tuple of ComponentInstance, string representation
    nominal_instances = ListStore(types.StringType, object)
    # tuple of ErrorImplementation, string representation
    error_implementations = ListStore(types.StringType, object)
    # tuple of string of error state
    error_states = ListStore(types.StringType)
    # tuple of data port/subcomponent strings
    datas = ListStore(types.StringType)
    
    
    def __init__(self, fault_injections, compile_info, root_impl):
        Model.__init__(self)
        self.fault_injections = fault_injections
        self.compile_info = compile_info
        self.root_impl = root_impl
        return


    def listStore_as_dict(self, liststore):
        return dict([(r[0], r[1]) for r in liststore])


    def fill_error_implementations(self):
        """All error implementations have to be shown to the user"""
        self.error_implementations.clear()
        types_and_impls = smvsupport.get_error_component_types_with_impl(self.compile_info)
        for _type, _impls in types_and_impls.iteritems():            
            for impl in _impls: 
                self.error_implementations.append([impl.subpackage.package_name + "::" + impl.name, impl])
        return
        

    # TODO, can make this implementation smarter by only deleting 
    #  unexisting ones and adding new ones 
    def fill_error_states(self, error_implementation):
        """All error states in the currently chosen error implementation is shown to the user"""
        self.error_states.clear()
        error_type = error_implementation.get_referenced_error_type()
        available_states = set(error_type.states.keys()) | set(error_implementation.states.keys())
        starting_states = set([state.name for state in (error_type.get_all_starting_states() | error_implementation.get_all_starting_states())])
        legal_states = available_states - starting_states
        for state in legal_states:
            self.error_states.append([state])
        return


    # TODO, can make this implementation smarter by only deleting 
    #  unexisting ones and adding new ones 
    def fill_nominal_instances(self, error_implementation):
        """All nominal instances are added that either have an 
        existing association with error_implemention, or have no
        associating at all. If they have associating with 
        error implementations other than error_implementation, they
        should not be in the liststore. Also, we filter all 
        component instances away that have no data subcomponents."""
        
        self.nominal_instances.clear()

        def __has_data(component):
            return (len(component.implementation.data_subcomponents) > 0) \
                    or (len(component.type.out_data_ports) > 0)
        
        # get all component instances, filtered by whether they have
        # a data port             
        component_list = set(general.instance_generator(self.root_impl, __has_data))

        # exclude components in this list that have an existing association
        # with a different error model
        fis = filter(lambda x: x[2][1] != error_implementation.name, self.fault_injections)
        to_exclude_paths = set(map(lambda x: x[4], fis))
        component_list = filter(lambda x: x.path not in to_exclude_paths, component_list)   
    
        # Add the list to the liststore
        for component in sorted(component_list, key=lambda x: ".".join(x.path)):
            self.nominal_instances.append([".".join(component.path), component])
        # update identifier of the root component
        # only do this if there is a component
        if len(self.nominal_instances) > 0 and len(self.nominal_instances[0][1].path) == 0: self.nominal_instances[0][0] = "."
        return


    # TODO, can make this implementation smarter by only deleting 
    #  unexisting ones and adding new ones 
    def fill_data(self, component, error_state):
        """All data ports/subcomponents should be added that have no 
        existing injection to currently selected error state"""
        self.datas.clear()
        # get all data ports/subcomponents first 
        data_list = set()
        data_list.update(component.implementation.data_subcomponents.keys())
        data_list.update(component.type.out_data_ports.keys())

        # filter away data subcomponents that are continuous datatypes
        to_exclude = map(lambda y: y.name,
                      filter(lambda x: isinstance(x.data_type, Continuous)
                                       or isinstance(x.data_type, Clock),
                      component.implementation.data_subcomponents.values()))
        data_list.difference_update(to_exclude)
        
        # filter away ports that have already an injection to 
        # the currently selected error state
        to_exclude = set(self.__get_injected_data_ports())
        data_list.difference_update(to_exclude)

        # Add the list of data ports to the liststore
        for d in sorted(data_list): self.datas.append([d])
        return    
    
    
    def __get_injected_data_ports(self):
        for r in self.fault_injections:
            if (r[2][1] == self.error_implementation.name 
                    and r[3] == self.error_state 
                    and r[4] == self.nominal_instance.path):
                yield r[5]
            pass
        pass
