# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


from gtkmvc import Model

import src.models.properties as models_properties
from src.models.properties import PropertiesModel
import property_management.src.api as properties_api
import property_management.src.constants as prop_const


class FilteredPropertiesModel (Model):
    """Model of filtered properties"""

    # Columns indices
    USE = PropertiesModel.USE
    PROPERTY = PropertiesModel.PROPERTY
    SELECT_SINGLE = PropertiesModel.SELECT_SINGLE
    SELECT_MULTI  = PropertiesModel.SELECT_MULTI
    MC_RESULT = PropertiesModel.MC_RESULT
    FT = PropertiesModel.FT
    ANALYSIS_RESULT = PropertiesModel.ANALYSIS_RESULT
    FDA_LIST = PropertiesModel.FDA_LIST

    # this property is True if one property is selected in the
    # single selection
    single_selected = False
    
    # the number of selected properties in the multi selected
    multi_selected_num = 0

    __observables__ = ("single_selected", "multi_selected_num")

    def __init__(self, props_model):
        Model.__init__(self)

        # the properties model
        self.props_model = props_model

        # this is the set of properties filtered
        self.properties = self.props_model.get_properties_list().filter_new()

        # to filter properties on their logic
        self.logic_set = set((prop_const.Logic.CTL, prop_const.Logic.LTL, prop_const.Logic.MITL,))
        self.propositional_only = False
        self.nominal_only = False
        def _prop_logic_filter(model, it):
            prop = model[it][FilteredPropertiesModel.PROPERTY]
            if prop is None: return False
            if not model[it][FilteredPropertiesModel.USE]: return False
            
            # do translation if necessary
            if prop.property.get_logic() not in self.logic_set:
                logic = iter(self.logic_set).next()
                new_property = models_properties.translate_property_to_logic(prop, logic)
                if new_property:
                    model[it][FilteredPropertiesModel.PROPERTY] = new_property
                    prop = model[it][FilteredPropertiesModel.PROPERTY]
            
            if self.nominal_only:
                if not _property_is_nominal(prop.property):
                    return False
                
            if self.propositional_only:
                return properties_api.check_is_propositional_property(prop.property)
            return prop.property.get_logic() in self.logic_set
        
        self.properties.set_visible_func(_prop_logic_filter)
        
        self.properties.connect("row-deleted", lambda tm, it: self._update_counters())
        self.properties.connect("row-changed", lambda tm, pa, it: self._update_counters())
        self.properties.connect("row-inserted", lambda tm, pa, it: self._update_counters())

        self._update_counters()

        return

    def get_properties_list(self):
        """Returns a ListStore containing all the filtered properties"""
        return self.properties

    def get_filtering(self):
        """Returns the currenlty set filtering logic"""
        return None if self.propositional_only else self.logic_set

    def set_filtering(self, logic_set, nominal_only=False):
        """Called to filter properties wrt given logic set. if
        logic_set is None, only propositional formulae are shown"""
        self.nominal_only = nominal_only
        
        if logic_set is None:
            self.propositional_only = True
            pass
        else:
            logic_set = set(logic_set)
            self.propositional_only = False
            self.logic_set = logic_set
            pass

        self.properties.refilter()
        return

    def restore_logic_filtering(self):
        """Restores previous logic filtering if current filtering
        is propositional"""
        if not self.propositional_only: return
        self.propositional_only = False
        self.properties.refilter()
        self._update_counters()
        return
    
    def _update_counters(self):
        """Updates properties single_selected and multi_selected_num"""
        self.single_selected = any((row[FilteredPropertiesModel.SELECT_SINGLE] 
                                    for row in self))
        self.multi_selected_num = len([row for row in self 
                                       if row[FilteredPropertiesModel.SELECT_MULTI]])
        return

    # this implement the iteration protocol
    #def __iter__(self): return self.get_properties_list().__iter__() 
    def __iter__(self):
        self._outer_iterator = None
        self._inner_iterator = None
        return self
    
    def next(self):
        # if we have just started create an iterator for the (outer) rows
        if not self._outer_iterator:
            self._outer_iterator = (0,)
            
        # first check whether the row is a dummy row, in that case immediately
        # start iterating over the children
        try: # we have to convert our tuple path to a GtkTreeIter to check for children
            gtk_tree_iter = self.properties.get_iter(self._outer_iterator)
        except ValueError:
            raise StopIteration
        
        if self.properties.iter_has_child(gtk_tree_iter):
            # if we just started iteration over children create an iterator
            # for the children and return the first row
            if not self._inner_iterator:
                row_index, = self._outer_iterator
                self._inner_iterator = (row_index,0)
                return self.properties[self._inner_iterator]
            # we have an iterator for children, try advancing
            else:
                row_index, child_index = self._inner_iterator
                if child_index + 1 < self.properties.iter_n_children(gtk_tree_iter):
                    self._inner_iterator = (row_index, child_index + 1)
                    return self.properties[self._inner_iterator]
                # otherwise just continue on row level
                else:
                    self._inner_iterator = None
                    # advance outer iterator one step
                    row_index, = self._outer_iterator
                    if row_index + 1 < len(self.properties):
                        self._outer_iterator = (row_index + 1,)
                        return self.next()
                    else:
                        self._inner_iterator = None
                        self._outer_iterator = None
                        raise StopIteration
        else:
            row_index, = self._outer_iterator
            if row_index < len(self.properties):
                self._outer_iterator = (row_index + 1,)
                return self.properties[(row_index,)] #Yes, return the old one
            else:
                self._inner_iterator = None
                self._outer_iterator = None
                raise StopIteration
    
    pass # end of class
# ----------------------------------------------------------------------
# object independent function
def _property_is_nominal(prop):
    """Checks whether the given property does not refer to error states and
    thus does not depend on the extended SLIM model
    
    @return False if property refers to error states
            True else
    @rtype  bool
    """
    indicator_list = ["error:", ".error ", ".error=", ".error!"]
    
    for placeholder in prop.pattern_instance:
        for substr in  indicator_list:
            # this if statement checks whether the placeholder
            # is directly accessible via the .content or via .content.text. 
            # the latter holds if the placeholder is not an integer
            value = ""
            if str(placeholder.content) == None: # if it's not a string
                value = placeholder.content.text
            else:
                value = str(placeholder.content)

            if value.find(substr) > -1:
                return False # a string indicating that we need the error model
                             # was found
    return True
