# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by GMV Aerospace and Defence  
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================

import os.path
import gtk
import gobject
import options
from utils.misc import relpath
from gtkmvc.model import Model
from gtkmvc import observable
from threading import Lock
from hasdel_tools.src.mission_specification.phases_op_modes import Phase, \
    OpMode, PhaseOpModeListStore, load_phases_list, \
    load_op_modes_list
from hasdel_tools.src.mission_specification.configurations import Configuration, \
    ConfigurationListStore, load_configurations_list
from hasdel_tools.src.mission_specification.phase_op_mode_combinations import \
    PhaseOpModeCombination, load_phase_op_mode_combinations_list
from hasdel_tools.src.xml.mission import MissionDocument
import hasdel_tools.src.functions as functions

from src.views.mission import MissionView



class MissionModel(Model):
    """Represents the Model of Mission."""

    phases_counter = 0
    op_modes_counter = 0
    configurations_counter = 0
    phase_op_mode_definitions_counter = 0
    phase_op_mode_associations_counter = 0

    mission_changed = observable.Signal()
    phases_list_changed = observable.Signal()
    op_modes_list_changed = observable.Signal()
    configurations_list_changed = observable.Signal()
    phase_op_mode_combinations_tree_changed = observable.Signal()

    # observable properties
    __observables__ = (
        "mission_changed",
        "phases_list_changed", "op_modes_list_changed",
        "configurations_list_changed",
        "phase_op_mode_combinations_tree_changed",
        "phases_counter", "op_modes_counter", 
        "configurations_counter",
        "phase_op_mode_definitions_counter", "phase_op_mode_associations_counter",)


    def __init__(self, appl_model, slim_model, mission_model=None):
        """Constructor Method"""
        Model.__init__(self)
        
        self.slim_model = slim_model
        self.mission_model = mission_model
        self.lock = Lock()

        self._init_phases_list()
        self._init_op_modes_list()
        self._init_configurations_list()
        self._init_phase_op_mode_combinations_tree()

        self.mission_view = MissionView()
        slim_model.register_observer(self)
        self.register_observer(self)
        return


    def _init_phases_list(self):
        """Sets the phases in the related column."""
        cr_text = gtk.CellRendererText()
        self.phases_col = gtk.TreeViewColumn("Name", cr_text)


        def phase_set_display_options(column, cell_renderer, model, iter):
            """Sets the text, foreground and background of the phases cell_renderers. 
	       @param   column:  The column Phases.
	       @type    column:  gtk.TreeViewColumn
	       @param   cell_renderer:  The cell_renderer of the column Phases.
	       @type    cell_renderer:  gtk.CellRendererText
	       @param   model:  The PhaseOpMode model. 
	       @type    model:  GtkListStore
	       @param   iter:  The iter of the related Phase.
	       @type    iter:  GtkTreeIter
	    """	    
            phase = model.get_value(iter, 0)
            self.mission_view.cell_renderer_set_text(cell_renderer, str(phase))
            return

        self.phases_col.set_cell_data_func(cr_text, phase_set_display_options)
        self.phases_list = PhaseOpModeListStore()
        return


    def _init_op_modes_list(self):
        """Sets the op_modes in the related column."""
        cr_text = gtk.CellRendererText()
        self.op_modes_col = gtk.TreeViewColumn("Name", cr_text)


        def op_mode_set_display_options(column, cell_renderer, model, iter):
            """Sets the text, foreground and background of the op-modes cell_renderers. 
	       @param   column:  The column Op-modes.
	       @type    column:  gtk.TreeViewColumn
	       @param   cell_renderer:  The cell_renderer of the column Op-modes.
	       @type    cell_renderer:  gtk.CellRendererText
	       @param   model:  The PhaseOpMode model. 
	       @type    model:  GtkListStore
	       @param   iter:  The iter of the related Op-mode.
	       @type    iter:  GtkTreeIter
	    """
            op_mode = model.get_value(iter, 0)
            self.mission_view.cell_renderer_set_text(cell_renderer, str(op_mode))
            return

        self.op_modes_col.set_cell_data_func(cr_text, op_mode_set_display_options)
        self.op_modes_list = PhaseOpModeListStore()
        return


    def _init_configurations_list(self):
        """Sets the configurations in the related column."""
        cr_text = gtk.CellRendererText()
        col_1 = gtk.TreeViewColumn("Configuration ID", cr_text)
	col_1.set_resizable(True)
        col_2 = gtk.TreeViewColumn("Allowed spacecraft configuration", cr_text)
	col_2.set_resizable(True)
        col_3 = gtk.TreeViewColumn("Associated Op-modes", cr_text)
	col_3.set_resizable(True)


        def configuration_id_set_display_options(column, cell_renderer, model, iter):
            """Sets the text, foreground and background of the configurations_id cell_renderers. 
	       @param   column:  The column Configurations_id.
	       @type    column:  gtk.TreeViewColumn
	       @param   cell_renderer:  The cell_renderer of the column Configurations_id.
	       @type    cell_renderer:  gtk.CellRendererText
	       @param   model:  The Configurations model. 
	       @type    model:  GtkListStore
	       @param   iter:  The iter of the related Configurations_id.
	       @type    iter:  GtkTreeIter
	    """
            configuration = model.get_value(iter, 0)
            self.mission_view.cell_renderer_set_text(cell_renderer, configuration.get_configuration_id())
            return


        def spacecraft_set_display_options(column, cell_renderer, model, iter):
            """Sets the text, foreground and background of the spacecraft configurations cell_renderers. 
	       @param   column:  The column spacecraft Configurations.
	       @type    column:  gtk.TreeViewColumn
	       @param   cell_renderer:  The cell_renderer of the column spacecraft Configurations.
	       @type    cell_renderer:  gtk.CellRendererText
	       @param   model:  The Configurations model. 
	       @type    model:  GtkListStore
	       @param   iter:  The iter of the related spacecraft Configurations.
	       @type    iter:  GtkTreeIter
	    """
            configuration = model.get_value(iter, 0)
            self.mission_view.cell_renderer_set_text(cell_renderer, configuration.get_spacecraft())
            return


        def associated_op_modes_set_display_options(column, cell_renderer, model, iter):
            """Sets the text, foreground and background of the associated_op_mode configurations cell_renderers. 
	       @param   column:  The column associated_op_mode Configurations.
	       @type    column:  gtk.TreeViewColumn
	       @param   cell_renderer:  The cell_renderer of the column associated_op_mode Configurations.
	       @type    cell_renderer:  gtk.CellRendererText
	       @param   model:  The Configurations model. 
	       @type    model:  GtkListStore
	       @param   iter:  The iter of the related associated_op_mode Configurations.
	       @type    iter:  GtkTreeIter
	    """
            configuration = model.get_value(iter, 0)
            self.mission_view.cell_renderer_set_text(cell_renderer, configuration.get_associated_op_modes_str())
            return

        col_1.set_cell_data_func(cr_text, configuration_id_set_display_options)
        col_2.set_cell_data_func(cr_text, spacecraft_set_display_options)
        col_3.set_cell_data_func(cr_text, associated_op_modes_set_display_options)
        
        self.configurations_cols = [col_1, col_2, col_3]
        self.configurations_list = ConfigurationListStore()
        return


    def _init_phase_op_mode_combinations_tree(self):
        """Sets the phase_op_mode combinations in the related column."""
        cr_text = gtk.CellRendererText()
        col_1 = gtk.TreeViewColumn("Phases", cr_text)
	col_1.set_resizable(True)
        col_2 = gtk.TreeViewColumn("Associated Op-modes", cr_text)
	col_2.set_resizable(True)
        col_3 = gtk.TreeViewColumn("Definition via Observable", cr_text)
	col_3.set_resizable(True)


        def phase_set_display_options(column, cell_renderer, model, iter):
            """Sets the text, foreground and background of the phase combinations cell_renderers. 
	       @param   column:  The column phase Combinations.
	       @type    column:  gtk.TreeViewColumn
	       @param   cell_renderer:  The cell_renderer of the column phase Combinations.
	       @type    cell_renderer:  gtk.CellRendererText
	       @param   model:  The Combinations model. 
	       @type    model:  GtkTreeStore
	       @param   iter:  The iter of the related phase Combinations.
	       @type    iter:  GtkTreeIter
	    """
            phase_op_mode_combination = model.get_value(iter, 0)
            phase = phase_op_mode_combination.get_phase()
            op_mode = phase_op_mode_combination.get_op_mode()

            if op_mode is None:
                self.mission_view.cell_renderer_set_text(cell_renderer, str(phase))
            else:
                cell_renderer.set_property('text', "")   
            return


        def op_mode_set_display_options(column, cell_renderer, model, iter):
            """Sets the text, foreground and background of the op_mode combinations cell_renderers. 
	       @param   column:  The column op_mode Combinations.
	       @type    column:  gtk.TreeViewColumn
	       @param   cell_renderer:  The cell_renderer of the column op_mode Combinations.
	       @type    cell_renderer:  gtk.CellRendererText
	       @param   model:  The Combinations model. 
	       @type    model:  GtkTreeStore
	       @param   iter:  The iter of the related op_mode Combinations.
	       @type    iter:  GtkTreeIter
	    """
            phase_op_mode_combination = model.get_value(iter, 0)
            op_mode = phase_op_mode_combination.get_op_mode()
            
            if op_mode is not None:
                self.mission_view.cell_renderer_set_text(cell_renderer, str(op_mode))
            else:
                cell_renderer.set_property('text', "")   
            return


        def definition_set_display_options(column, cell_renderer, model, iter):
            """Sets the text, foreground and background of the definition combinations cell_renderers. 
	       @param   column:  The column definition Combinations.
	       @type    column:  gtk.TreeViewColumn
	       @param   cell_renderer:  The cell_renderer of the column definition Combinations.
	       @type    cell_renderer:  gtk.CellRendererText
	       @param   model:  The Combinations model. 
	       @type    model:  GtkTreeStore
	       @param   iter:  The iter of the related definition Combinations.
	       @type    iter:  GtkTreeIter
	    """
            phase_op_mode_combination = model.get_value(iter, 0)
            self.mission_view.cell_renderer_set_text(cell_renderer, phase_op_mode_combination.get_definition())
            return

        col_1.set_cell_data_func(cr_text, phase_set_display_options)
        col_2.set_cell_data_func(cr_text, op_mode_set_display_options)
        col_3.set_cell_data_func(cr_text, definition_set_display_options)
        
        self.phase_op_mode_combinations_cols = [col_1, col_2, col_3]
        self.phase_op_mode_combinations_tree = gtk.TreeStore(gobject.TYPE_PYOBJECT)
        return


    def get_phases_list(self):
        """Gets the list of Phases.
           @return   self.phases_list:  The list of Phases.
	   @rtype    self.phases_list:  list
        """
        return self.phases_list


    def get_phases_col(self):
        """Gets the column number of the Phases.
           @return   self.phases_col:  The column number of the Phases.
	   @rtype    self.phases_col:  int
        """
        return self.phases_col


    def get_phases_counter(self):
        """Gets the counter of the number of the Phases.
           @return   self.phases_counter:  The counter of the number of the Phases.
	   @rtype    self.phases_counter:  int
        """
        return self.phases_counter


    def get_op_modes_list(self):
        """Gets the list of OpModes.
           @return   self.op_modes_list:  The list of OpModes.
	   @rtype    self.op_modes_list:  list
        """
        return self.op_modes_list


    def get_op_modes_col(self):
        """Gets the column number of the OpModes.
           @return   self.op_modes_col:  The column number of the OpModes.
	   @rtype    self.op_modes_col:  int
        """
        return self.op_modes_col


    def get_op_modes_counter(self):
        """Gets the counter of the number of the OpModes.
           @return   self.op_modes_counter:  The counter of the number of the OpModes.
	   @rtype    self.op_modes_counter:  int
        """
        return self.op_modes_counter


    def get_configurations_list(self):
        """Gets the list of Configurations.
           @return   self.configurations_list:  The list of Configurations.
	   @rtype    self.configurations_list:  list
        """
        return self.configurations_list


    def get_configurations_cols(self):
        """Gets the column number of the Configurations.
           @return   self.configurations_cols:  The column number of the Configurations.
	   @rtype    self.configurations_cols:  int
        """
        return self.configurations_cols


    def get_configurations_counter(self):
        """Gets the counter of the number of the Configurations.
           @return   self.configurations_counter:  The counter of the number of the Configurations.
	   @rtype    self.configurations_counter:  int
        """
        return self.configurations_counter


    def get_phase_op_mode_combinations_tree(self):
        """Gets the tree of PhaseOpModeCombinations.
           @return   self.phase_op_mode_combinations_tree:  The tree of PhaseOpModeCombinations.
	   @rtype    self.phase_op_mode_combinations_tree:  TreeStore
        """
        return self.phase_op_mode_combinations_tree


    def get_phase_op_mode_combinations_cols(self):
        """Gets the column number of the PhaseOpModeCombinations.
           @return   self.phase_op_mode_combinations_cols:  The column number of the PhaseOpModeCombinations.
	   @rtype    self.phase_op_mode_combinations_cols:  int
        """
        return self.phase_op_mode_combinations_cols


    def get_phase_op_mode_definitions_counter(self):
        """Gets the counter of the number of the definitions of PhaseOpModeCombinations.
           @return   self.phase_op_mode_definitions_counter:  The counter of the number of the 
                                                                definitions of PhaseOpModeCombinations.
	   @rtype    self.phase_op_mode_definitions_counter:  int
        """
        return self.phase_op_mode_definitions_counter


    def get_phase_op_mode_associations_counter(self):
        """Gets the counter of the number of the associations of PhaseOpModeCombinations.
           @return   self.phase_op_mode_associations_counter:  The counter of the number of the 
                                                                associations of PhaseOpModeCombinations.
	   @rtype    self.phase_op_mode_associations_counter:  int
        """
        return self.phase_op_mode_associations_counter


    def append_phase(self, model):
        """Appends the Phase to the Phases List.
           @param   model:  The add dialog model of the Phase.
	   @type    model:  PhaseDialogModel
        """
        self.lock.acquire()
        self.phases_list.append([Phase(model.get_name())])
        self.phases_counter += 1
        self.phases_list_changed.emit()
        self.lock.release();
        return


    def append_op_mode(self, model):
        """Appends the OpMode to the OpModes List.
           @param   model:  The add dialog model of the OpMode.
	   @type    model:  OpModeDialogModel
        """
        self.lock.acquire()
        self.op_modes_list.append([OpMode(model.get_name())])
        self.op_modes_counter += 1
        self.op_modes_list_changed.emit()
        self.lock.release();
        return


    def append_configuration(self, model):
        """Appends the Configuration to the Configurations List.
           @param   model:  The add dialog model of the Configuration.
	   @type    model:  ConfigurationDialogModel
        """
        self.lock.acquire()
        self.configurations_list.append([
            Configuration(model.get_configuration_id(),
                          model.get_spacecraft(),
                          model.get_associated_op_modes())])
        self.configurations_counter += 1
        self.configurations_list_changed.emit()
        self.lock.release()
        return


    def edit_phase(self, model):
        """Edits the Phase of the Phases List.
           @param   model:  The edit dialog model of the Phase.
	   @type    model:  PhaseDialogModel
        """
        self.lock.acquire()
        model.get_old_phase().set_name(model.get_name())
        self.phases_list_changed.emit()
        self.lock.release()
        return


    def edit_op_mode(self, model):
        """Edits the OpMode of the OpModes List.
           @param   model:  The edit dialog model of the OpMode.
	   @type    model:  OpModeDialogModel
        """
        self.lock.acquire()
        model.get_old_op_mode().set_name(model.get_name())
        self.op_modes_list_changed.emit()
        self.lock.release()
        return


    def edit_configuration(self, model, iter):
        """Edits the Configuration of the Configurations List.
           @param   model:  The edit dialog model of the Configuration.
	   @type    model:  ConfigurationDialogModel
           @param   iter:  The iter of the Configuration.
	   @type    iter:  gtk.TreeIter
        """
        self.lock.acquire()
        model.get_old_configuration().set_configuration_id(
            model.get_configuration_id())
        model.get_old_configuration().set_spacecraft(
            model.get_spacecraft())
        self.configurations_list.remove(iter)
        self.configurations_list.append([model.get_old_configuration()])
        self.configurations_list_changed.emit()
        self.lock.release()
        return


    def edit_phase_op_mode_associations(self, model, iter):
        """Edits the association of the PhaseOpModeCombinations Tree.
           @param   model:  The edit dialog model of the association of the PhaseOpModeCombinations Tree.
	   @type    model:  PhaseOpModeCombinationEditAssociationDialogModel
           @param   iter:  The iter of the PhaseOpModeCombinations.
	   @type    iter:  gtk.TreeIter
        """
        self.lock.acquire()
        phase = self.phase_op_mode_combinations_tree.get_value(iter, 0).get_phase()
        associated_op_modes = model.get_associated_op_modes()
        iter_child = self.phase_op_mode_combinations_tree.iter_children(iter)

        while iter_child is not None:
            iter_next_child = self.phase_op_mode_combinations_tree.iter_next(iter_child)
            comb = self.phase_op_mode_combinations_tree.get_value(iter_child, 0)
            iter_ass = associated_op_modes.get_iter_first()

            while iter_ass is not None:
                if associated_op_modes.get_value(iter_ass, 0) == comb.get_op_mode():
                    associated_op_modes.remove(iter_ass)
                    break
                else:
                    iter_ass = associated_op_modes.iter_next(iter_ass)
                pass # end of internal while
            else:
                definition = self.phase_op_mode_combinations_tree.get_value(iter_child, 0).get_definition()

                self.phase_op_mode_associations_counter += -1
                if not definition == "":
                    self.phase_op_mode_definitions_counter += -1

                self.phase_op_mode_combinations_tree.remove(iter_child)

            iter_child = iter_next_child

            pass # end of external while

        for op_mode in associated_op_modes:
            iter_child = self.phase_op_mode_combinations_tree.iter_children(iter)
            while iter_child is not None and self.phase_op_mode_combinations_tree.get_value(iter_child, 0).get_op_mode().get_name() < op_mode[0].get_name():
                iter_child = self.phase_op_mode_combinations_tree.iter_next(iter_child)
            insert_iter = self.phase_op_mode_combinations_tree.insert_before(iter, iter_child)
            self.phase_op_mode_combinations_tree.set_value(
                insert_iter, 0, PhaseOpModeCombination(phase, op_mode[0]))
            self.phase_op_mode_associations_counter += 1

        self.phase_op_mode_combinations_tree_changed.emit()
        self.lock.release()
        return


    def edit_phase_op_mode_definition(self, model):
        """Edits the definition of the PhaseOpModeCombinations Tree.
           @param   model:  The edit dialog model of the definition of the PhaseOpModeCombinations Tree.
	   @type    model:  PhaseOpModeCombinationEditDefinitionDialogModel
           @param   iter:  The iter of the PhaseOpModeCombinations.
	   @type    iter:  gtk.TreeIter
        """
        self.lock.acquire()
        if model.get_old_phase_op_mode_combination().get_definition() == "":
            self.phase_op_mode_definitions_counter += 1

        model.get_old_phase_op_mode_combination().set_definition(model.get_definition())
        self.phase_op_mode_combinations_tree_changed.emit()
        self.lock.release()
        return


    def remove_phases(self, tree_model, paths):
        """Removes the Phase from the Phases List.
           @param   tree_model:  The treemodel of the Phase.
	   @type    tree_model:  PhaseOpModeListStore
           @param   paths:  The paths of the Phase.
	   @type    paths:  tuple
        """
        self.lock.acquire()
        for iter in [tree_model.get_iter(path) for path in paths]:
            self.phases_list.remove(iter)
            self.phases_counter += -1

        self.phases_list_changed.emit()
        self.lock.release()
        return


    def remove_op_modes(self, tree_model, paths):
        """Removes the OpMode from the OpModes List.
           @param   tree_model:  The treemodel of the OpMode.
	   @type    tree_model:  PhaseOpModeListStore
           @param   paths:  The paths of the OpMode.
	   @type    paths:  tuple
        """
        self.lock.acquire()
        for iter in [tree_model.get_iter(path) for path in paths]:
            self.op_modes_list.remove(iter)
            self.op_modes_counter += -1

        self.op_modes_list_changed.emit()
        self.lock.release()
        return


    def remove_configurations(self, tree_model, paths):
        """Removes the Configuration from the Configurations List.
           @param   tree_model:  The treemodel of the Configuration.
	   @type    tree_model:  ConfigurationListStore
           @param   paths:  The paths of the Configuration.
	   @type    paths:  tuple
        """
        self.lock.acquire()
        for iter in [tree_model.get_iter(path) for path in paths]:
            self.configurations_list.remove(iter)
            self.configurations_counter += -1

        self.configurations_list_changed.emit()
        self.lock.release()
        return


    def name_exist(self, name, old_name=None):
        """Checks if there are the same name for different elements.
           @param   name:  The name to check.
	   @type    name:  string
           @param   old_name:  The old_name to check.
	   @type    old_name:  string

           @return   True or False:  True if there are the same names, else False.
	   @rtype    True or False:  boolean
        """
        if (name is None) or (name == "") or (name.lower == "none"):
            return False
        elif name == old_name:
            return False
        else:
            for n in self.phases_list:
                if n[0].get_name() == name:
                    return True
            for n in self.op_modes_list:
                if n[0].get_name() == name:
                    return True
            for n in self.configurations_list:
                if n[0].get_configuration_id() == name:
                    return True
            return False


    def get_existing_names(self):
        """Gets a list with the all phase, op_mode and configuration names in their lists.
           @return   result:  The list of phase, op_mode and configuration names
	   @rtype    result:  list
        """
        result = []
        for phase in self.phases_list:
            result.append(phase[0].get_name())
        for op_mode in self.op_modes_list:
            result.append(op_mode[0].get_name())
        for conf in self.configurations_list:
            result.append(conf[0].get_configuration_id())
        return result


    def clear_phases_list(self):
        """Clears the phases from the list."""
        self.phases_list.clear()
        self.phases_counter = 0
        self.phases_list_changed.emit()
        return


    def clear_op_modes_list(self):
        """Clears the op_modes from the list."""
        self.op_modes_list.clear()
        self.op_modes_counter = 0
        self.op_modes_list_changed.emit()
        return


    def clear_configurations_list(self):
        """Clears the configurations from the list."""
        self.configurations_list.clear()
        self.configurations_counter = 0
        self.configurations_list_changed.emit()
        return


    def clear_phase_op_mode_combinations_tree(self):
        """Clears the phases and op_modes from the tree."""
        self.phase_op_mode_combinations_tree.clear()
        for phase in self.phases_list:
            parent_comb = PhaseOpModeCombination(phase[0])
            self.phase_op_mode_combinations_tree.append(None, [parent_comb])

        self.phase_op_mode_definitions_counter = 0
        self.phase_op_mode_associations_counter = 0
        self.phase_op_mode_combinations_tree_changed.emit()
        return


    def clear(self):
        """Clears all the Mission.""" 
        self.clear_phases_list()
        self.clear_op_modes_list()
        self.clear_configurations_list()
        self.clear_phase_op_mode_combinations_tree()
        self.mission_changed.emit()
        return


    def get_mission_doc(self, filename): 
        """Gets the Mission Document.
           @param   filename:  The path of the loaded Mission file.
	   @type    filename:  string
        """

        self.clear()
        self.lock.acquire()
        self.slim_model.console.log_compiler_msg("Loading Mission '%s'... " % relpath(filename))

        if not os.path.exists(filename):
            self.slim_model.console.log_compiler_fail("ERROR\n")
            self.slim_model.console.log_compiler_fail("\t File does not exists.\n")
            self.lock.release()
            return

        mission_doc, validation_errors = MissionDocument().parse(filename)
        
        
        if len(validation_errors) > 0:
            for err in validation_errors:
                self.slim_model.console.log_compiler_fail("ERROR\n")
                self.slim_model.console.log_compiler_fail("\t File contains validation errors: " + str(err) + "\n")
                self.lock.release()
            return 

        syntax_errors = MissionDocument().get_syntax_errors_in_mission_document(
            mission_doc)


        if len(syntax_errors) > 0:
            for err in validation_errors:
                self.slim_model.console.log_compiler_fail("ERROR\n")
                self.slim_model.console.log_compiler_fail("\t File contains syntax errors: " + str(err) + "\n")
                self.lock.release()
            return 

        phases_list = mission_doc.phases_list
        op_modes_list = mission_doc.op_modes_list
        configurations_list = mission_doc.configurations_list
        phase_op_mode_combinations_list = mission_doc.phase_op_mode_combinations_list

        self.slim_model.console.log_compiler_ok("OK\n")


        """Loading Phases"""
        error_list, num_added, num_contains = load_phases_list(phases_list, self.phases_list, self.get_existing_names())

        if num_added > 0:
            self.phases_counter += num_added
            self.phases_list_changed.emit()

        for error in error_list:
            self.slim_model.console.log_compiler_fail(" > ERROR\n")
            self.slim_model.console.log_compiler_fail("\t > " + str(error) + "\n")

        self.slim_model.console.log_compiler_msg(" > Loaded " + str(num_added) +
                                                 " of " + str(num_contains) + " Phases.\n")


        """Loading Op-Modes"""
        error_list, num_added, num_contains = load_op_modes_list(op_modes_list, self.op_modes_list, self.get_existing_names())

        if num_added > 0:
            self.op_modes_counter += num_added
            self.op_modes_list_changed.emit()

        for error in error_list:
            self.slim_model.console.log_compiler_fail(" > ERROR\n")
            self.slim_model.console.log_compiler_fail("\t > " + str(error) + "\n")

        self.slim_model.console.log_compiler_msg(" > Loaded " + str(num_added) +
                                                 " of " + str(num_contains) + " Op-Modes.\n")

        if (self.phases_counter == 0) and (self.op_modes_counter == 0):
            self.slim_model.console.log_compiler_fail(" > WARNING\n")
            self.slim_model.console.log_compiler_msg("\t > No Phase or no Op-Mode are already defined, " +
                                                     "so the Configurations and the Phase/Op-Mode " + 
                                                     "Combinations won't loaded.\n")
            self.lock.release()
            return


        """Loading Configurations"""
        error_list, num_added, num_contains = load_configurations_list(self.slim_model,
                                                                       configurations_list, 
                                                                       self.configurations_list, 
                                                                       self.get_existing_names(),
                                                                       self.op_modes_list)

        if num_added > 0:
            self.configurations_counter += num_added
            self.configurations_list_changed.emit()

        for error in error_list:
            self.slim_model.console.log_compiler_fail(" > ERROR\n")
            self.slim_model.console.log_compiler_fail("\t > " + str(error) + "\n")

        self.slim_model.console.log_compiler_msg(" > Loaded " + str(num_added) +
                                                 " of " + str(num_contains) + " Configurations.\n")

        if self.configurations_counter == 0:
            self.slim_model.console.log_compiler_fail(" > WARNING\n")
            self.slim_model.console.log_compiler_msg("\t > No Configuration are already defined, " +
                                                     "so the Phase/Op-Mode Combinations won't loaded.\n")
            self.lock.release()
            return


        """Loading Phase/Op-Mode Combinations"""
        error_list, num_added, num_changed, num_ass, num_def, num_contains = load_phase_op_mode_combinations_list(self.slim_model,
                                                                                                                  self.op_modes_list,
                                                                                                                  phase_op_mode_combinations_list, 
                                                                                                                  self.phase_op_mode_combinations_tree)
        
        if (num_added > 0) or (num_changed > 0):
            self.phase_op_mode_associations_counter += num_ass
            self.phase_op_mode_definitions_counter += num_def
            self.phase_op_mode_combinations_tree_changed.emit()

        for error in error_list:
            self.slim_model.console.log_compiler_fail(" > ERROR\n")
            self.slim_model.console.log_compiler_fail("\t > " + str(error) + "\n")
                
        self.slim_model.console.log_compiler_msg(" > Added " + str(num_added) +
                                                 " of " + str(num_contains) + " Phase/Op-Mode Combinations.\n")
        self.slim_model.console.log_compiler_msg(" > Modified " + str(num_changed) +
                                                 " of " + str(num_contains) + " Phase/Op-Mode Combinations.\n")

        self.lock.release()
        return


    # ---------------------------------------------------------------
    #            Notifications from observable properties
    # ---------------------------------------------------------------
    def property_changed_signal_emit(self, model, args):
        """When a change occured in the GUI, a signal clears all.
           @param   model:  The Slim Model.
	   @type    model:  SlimModel
           @param   args:  The arguments of Mission Model.
	   @type    args:  string
        """
        self.clear()
        return


    def property_phases_list_changed_signal_emit(self, model, args):
        """When a change occured in the phases list, a signal clears the phase_op_mode_combinations_tree.
           @param   model:  The Mission Model.
	   @type    model:  MissionModel
           @param   args:  The arguments of Mission Model.
	   @type    args:  string
        """
        self.clear_phase_op_mode_combinations_tree()
        self.mission_changed.emit()
        return


    def property_op_modes_list_changed_signal_emit(self, model, args):
        """When a change occured in the op_modes list, a signal clears the configurations 
           and the phase_op_mode_combinations_tree.
           @param   model:  The Mission Model.
	   @type    model:  MissionModel
           @param   args:  The arguments of Mission Model.
	   @type    args:  string
        """
        self.clear_configurations_list()
        self.clear_phase_op_mode_combinations_tree()
        self.mission_changed.emit()
        return


    def property_configurations_list_changed_signal_emit(self, model, args):
        """When a change occured in the configurations list, a signal says that there was a change.
           @param   model:  The Mission Model.
	   @type    model:  MissionModel
           @param   args:  The arguments of Mission Model.
	   @type    args:  string
        """
        self.mission_changed.emit()
        return


    def property_phase_op_mode_combinations_tree_changed_signal_emit(self, model, args):
        """When a change occured in the phase_op_mode_combinations tree, a signal says that there was a change.
           @param   model:  The Mission Model.
	   @type    model:  MissionModel
           @param   args:  The arguments of Mission Model.
	   @type    args:  string
        """
        self.mission_changed.emit()
        return


pass # end of MissionModel class
# ------------------------------------------------------------------------------


class PhaseDialogModel(Model):
    """Represents the Dialog Class of Phase"""

    def __init__(self, mission_model, old_phase=None):
        """Constructor Method"""
        Model.__init__(self)
        self.mission_model = mission_model
        self.old_phase = old_phase
        return


    def instantiate(self, name):
        """Instantiates the name of the Phase.
           @param   name:  The name of the Phase.
	   @type    name:  string
        """
        self.name = name
        return


    def get_old_phase(self):
        """Gets the old Phase.
           @return   self.old_phase:  The old Phase.
	   @rtype    self.old_phase:  Phase
        """
        return self.old_phase


    def get_name(self):
        """Gets the name of the Phase.
           @return   self.name:  The name of the Phase.
	   @rtype    self.name:  string
        """
        return self.name


pass # end of PhaseDialogModel class
# ------------------------------------------------------------------------------


class OpModeDialogModel(Model):
    """Represents the Dialog Class of OpMode"""

    def __init__(self, mission_model, old_op_mode=None):
        """Constructor Method"""
        Model.__init__(self)
        self.mission_model = mission_model
        self.old_op_mode = old_op_mode
        return


    def instantiate(self, name):
        """Instantiates the name of the OpMode.
           @param   name:  The name of the OpMode.
	   @type    name:  string
        """
        self.name = name
        return


    def get_old_op_mode(self):
        """Gets the old OpMode.
           @return   self.old_op_mode:  The old OpMode.
	   @rtype    self.old_op_mode:  OpMode
        """
        return self.old_op_mode


    def get_name(self):
        """Gets the name of the OpMode.
           @return   self.name:  The name of the OpMode.
	   @rtype    self.name:  string
        """
        return self.name

pass # end of OpModeDialogModel class
# ------------------------------------------------------------------------------


class ConfigurationDialogModel(Model):
    """Represents the Dialog Class of Configuration"""

    associated_op_modes_counter = 0

    # observable properties
    __observables__ = ("associated_op_modes_counter", )


    def __init__(self, mission_model, old_configuration=None):
        """Constructor Method"""
        Model.__init__(self)
        self.mission_model = mission_model
        self.old_configuration = old_configuration
        self.associated_op_modes = PhaseOpModeListStore()
        self.not_associated_op_modes = PhaseOpModeListStore()
        
        if self.old_configuration is None:
            for op_mode in mission_model.op_modes_list:
                self.not_associated_op_modes.append(op_mode)
        else:
            self.associated_op_modes = self.old_configuration.get_associated_op_modes()
            self.associated_op_modes_counter = len(self.associated_op_modes)
            for op_mode in mission_model.op_modes_list:
                for opm in self.associated_op_modes:
                    if op_mode[0] == opm[0]:
                        break
                else:
                    self.not_associated_op_modes.append(op_mode)

        cr_text = gtk.CellRendererText()
        self.associated_op_modes_col = gtk.TreeViewColumn("Name", cr_text)
	self.associated_op_modes_col.set_resizable(True)
        self.not_associated_op_modes_col = gtk.TreeViewColumn("Name", cr_text)
	self.not_associated_op_modes_col.set_resizable(True)


        def op_mode_to_str(column, cell_renderer, model, iter):
            """Converts the op_mode of the Configuration in string format."""
            op_mode = model.get_value(iter, 0)
            cell_renderer.set_property('text', str(op_mode))
            return


        self.associated_op_modes_col.set_cell_data_func(cr_text, op_mode_to_str)
        self.not_associated_op_modes_col.set_cell_data_func(cr_text, op_mode_to_str)
        self.lock = Lock()
        return


    def instantiate(self, configuration_id, spacecraft):
        """Instantiates the id and the spacecraft of the Configuration.
           @param   configuration_id:  The id of the Configuration.
	   @type    configuration_id:  string
           @param   spacecraft:  The spacecraft of the Configuration.
	   @type    spacecraft:  string
        """
        self.configuration_id = configuration_id
        self.spacecraft = spacecraft
        return


    def get_old_configuration(self):
        """Gets the old Configuration.
           @return   self.old_configuration:  The old Configuration.
	   @rtype    self.old_configuration:  Configuration
        """
        return self.old_configuration


    def get_configuration_id(self):
        """Gets the id of the Configuration.
           @return   self.configuration_id:  The id of the Configuration.
	   @rtype    self.configuration_id:  string
        """
        return self.configuration_id


    def get_spacecraft(self):
        """Gets the spacecraft of the Configuration.
           @return   self.spacecraft:  The spacecraft of the Configuration.
	   @rtype    self.spacecraft:  string
        """
        return self.spacecraft


    def get_associated_op_modes(self):
        """Gets the op_modes associated to the Configuration.
           @return   self.associated_op_modes:  The op_modes associated to the Configuration.
	   @rtype    self.associated_op_modes:  list
        """
        return self.associated_op_modes


    def get_associated_op_modes_counter(self):
        """Gets the counter of the number of op_modes associated to the Configuration.
           @return   self.associated_op_modes_counter:  The counter of the number of op_modes associated to the Configuration.
	   @rtype    self.associated_op_modes_counter:  int
        """
        return self.associated_op_modes_counter


    def get_associated_op_modes_col(self):
        """Gets the column number of the op_modes associated to the Configuration.
           @return   self.associated_op_modes_col:  The column number of the op_modes associated to the Configuration.
	   @rtype    self.associated_op_modes_col:  int
        """
        return self.associated_op_modes_col


    def get_not_associated_op_modes(self):
        """Gets the op_modes not associated to the Configuration.
           @return   self.not_associated_op_modes:  The op_modes not associated to the Configuration.
	   @rtype    self.not_associated_op_modes:  list
        """
        return self.not_associated_op_modes


    def get_not_associated_op_modes_col(self):
        """Gets the column number of the op_modes not associated to the Configuration.
           @return   self.not_associated_op_modes_col:  The column number of the op_modes not associated to the Configuration.
	   @rtype    self.not_associated_op_modes_col:  int
        """
        return self.not_associated_op_modes_col


    def ass_to_notass_op_modes_transfer(self, tree_model, paths):
        """Transfers the associated op_modes selected to the related list.
           @param   tree_model:  The treemodel of the Configuration.
	   @type    tree_model:  ConfigurationListStore
           @param   paths:  The paths of the Configuration.
	   @type    paths:  tuple
        """
        self.lock.acquire()
        for iter in [tree_model.get_iter(path) for path in paths]:
            op_mode = self.associated_op_modes.get_value(iter, 0)
            self.associated_op_modes.remove(iter)
            self.not_associated_op_modes.append([op_mode])
            self.associated_op_modes_counter += -1

        self.lock.release()
        return


    def notass_to_ass_op_modes_transfer(self, tree_model, paths):
        """Transfers the not associated op_modes selected to the related list.
           @param   tree_model:  The treemodel of the Configuration.
	   @type    tree_model:  ConfigurationListStore
           @param   paths:  The paths of the Configuration.
	   @type    paths:  tuple
        """
        self.lock.acquire()
        for iter in [tree_model.get_iter(path) for path in paths]:
            op_mode = tree_model.get_value(iter, 0)
            self.not_associated_op_modes.remove(iter)
            self.associated_op_modes.append([op_mode])
            self.associated_op_modes_counter += 1

        self.lock.release()
        return

pass # end of ConfigurationDialogModel class
# ------------------------------------------------------------------------------


class PhaseOpModeCombinationEditAssociationDialogModel(Model):
    """Represents the Dialog Class of Association of the PhaseOpModeCombination"""


    def __init__(self, mission_model, parent, associated_op_modes):
        """Constructor Method"""
        Model.__init__(self)
        self.mission_model = mission_model
        self.parent = parent
        self.associated_op_modes = associated_op_modes
        self.not_associated_op_modes = PhaseOpModeListStore()

        for op_mode in mission_model.op_modes_list:
            for opm in self.associated_op_modes:
                if op_mode[0] == opm[0]:
                    break
            else:
                self.not_associated_op_modes.append(op_mode)
                
        cr_text = gtk.CellRendererText()
        self.associated_op_modes_col = gtk.TreeViewColumn("Name", cr_text)
	self.associated_op_modes_col.set_resizable(True)
        self.not_associated_op_modes_col = gtk.TreeViewColumn("Name", cr_text)
	self.not_associated_op_modes_col.set_resizable(True)


        def op_mode_to_str(column, cell_renderer, model, iter):
            """Converts the op_mode of the PhaseOpModeCombination in string format."""
            op_mode = model.get_value(iter, 0)
            cell_renderer.set_property('text', str(op_mode))
            return


        self.associated_op_modes_col.set_cell_data_func(cr_text, op_mode_to_str)
        self.not_associated_op_modes_col.set_cell_data_func(cr_text, op_mode_to_str)
        self.lock = Lock()
        return


    def get_parent(self):
        """Gets the parent of the Configuration.
           @return   self.parent:  The parent of the PhaseOpModeCombination.
	   @rtype    self.parent:  gtk.TreeIter
	"""
        return self.parent


    def get_old_associated_op_modes(self):
        """Gets the old op_modes associated to the PhaseOpModeCombination.
           @return   self.old_associated_op_modes:  The old op_modes associated to the PhaseOpModeCombination.
	   @rtype    self.old_associated_op_modes:  list
        """
        return self.old_associated_op_modes


    def get_associated_op_modes(self):
        """Gets the op_modes associated to the PhaseOpModeCombination.
           @return   self.associated_op_modes:  The op_modes associated to the PhaseOpModeCombination.
	   @rtype    self.associated_op_modes:  list
        """
        return self.associated_op_modes


    def get_associated_op_modes_col(self):
        """Gets the column number of the op_modes associated to the PhaseOpModeCombination.
           @return   self.associated_op_modes_col:  The column number of the op_modes associated to the PhaseOpModeCombination.
	   @rtype    self.associated_op_modes_col:  int
        """
        return self.associated_op_modes_col


    def get_not_associated_op_modes(self):
        """Gets the op_modes not associated to the PhaseOpModeCombination.
           @return   self.not_associated_op_modes:  The op_modes not associated to the PhaseOpModeCombination.
	   @rtype    self.not_associated_op_modes:  list
        """
        return self.not_associated_op_modes


    def get_not_associated_op_modes_col(self):
        """Gets the column number of the op_modes not associated to the Configuration.
           @return   self.not_associated_op_modes_col:  The column number of the op_modes not associated to the Configuration.
	   @rtype    self.not_associated_op_modes_col:  int
        """
        return self.not_associated_op_modes_col


    def ass_to_notass_op_modes_transfer(self, tree_model, paths):
        """Transfers the associated op_modes selected to the related list.
           @param   tree_model:  The treemodel of the PhaseOpModeCombination.
	   @type    tree_model:  PhaseOpModeCombination Model
           @param   paths:  The paths of the PhaseOpModeCombination.
	   @type    paths:  tuple
        """
        self.lock.acquire()
        for iter in [tree_model.get_iter(path) for path in paths]:
            op_mode = self.associated_op_modes.get_value(iter, 0)
            self.associated_op_modes.remove(iter)
            self.not_associated_op_modes.append([op_mode])

        self.lock.release()
        return


    def notass_to_ass_op_modes_transfer(self, tree_model, paths):
        """Transfers the not associated op_modes selected to the related list.
           @param   tree_model:  The treemodel of the PhaseOpModeCombination.
	   @type    tree_model:  PhaseOpModeListStore
           @param   paths:  The paths of the PhaseOpModeCombination.
	   @type    paths:  tuple
        """
        self.lock.acquire()
        for iter in [tree_model.get_iter(path) for path in paths]:
            op_mode = tree_model.get_value(iter, 0)
            self.not_associated_op_modes.remove(iter)
            self.associated_op_modes.append([op_mode])

        self.lock.release()
        return

pass # end of PhaseOpModeCombinationEditAssociationDialogModel class
# ------------------------------------------------------------------------------


class PhaseOpModeCombinationEditDefinitionDialogModel(Model):
    """Represents the Dialog Class of Definition of the PhaseOpModeCombination"""

    def __init__(self, mission_model, old_phase_op_mode_combination):
        """Constructor Method"""
        Model.__init__(self)
        self.mission_model = mission_model
        self.old_phase_op_mode_combination = old_phase_op_mode_combination
        return


    def instantiate(self, definition):
        """Instantiates the definition of the PhaseOpModeCombination.
           @param   definition:  The definition of the PhaseOpModeCombination.
	   @type    definition:  string
        """
        self.definition = definition
        return


    def get_old_phase_op_mode_combination(self):
        """Gets the old PhaseOpModeCombination.
           @return   self.old_phase_op_mode_combination:  The old PhaseOpModeCombination.
	   @rtype    self.old_phase_op_mode_combination:  gtk.TreeStore
        """
        return self.old_phase_op_mode_combination


    def get_definition(self):
        """Gets the definition of the PhaseOpModeCombination.
           @return   self.definition:  The definition of the PhaseOpModeCombination.
	   @rtype    self.definition:  string
        """
        return self.definition

pass # end of PhaseOpModeCombinationEditDefinitionDialogModel class
# ------------------------------------------------------------------------------
