# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


from gtkmvc import Model
import utils.log
from utils.misc import relpath, get_out_fname
import options
import os
import os.path
import gtk
import pango
import gobject
import copy
from gtkmvc import observable
from src.models.activity import ActivityModel
from slim_compiler.src.dataStructure import nominalbehaviour, expressions
from slim_compiler.src.dataStructure.nominalbehaviour import Bool, Enum
from hasdel_tools.src.tfpg.tfpg import Tfpg, Node, Edge, NodeType
from hasdel_tools.src.tfpg.tfpg_errors import FMNotCorrectError
import hasdel_tools.src.functions as functions
from hasdel_tools.src.xml.tfpg import TfpgDocument, tfpg_doc_to_tfpg_internal_struct, \
tfpg_internal_struct_to_tfpg_doc
from hasdel_tools.src.xml.tfpg_slim_associations import AssociationsDocument
from hasdel_tools.src.tfpg.tfpg_errors import DiscrepancyWithoutIncomingEdge, \
InvalidNameError

from src.views.tfpg import TfpgView
from src.models.tool_handler import FameTfpgHandlerModel
from utils import fame_tfpg_process
from tfpg2smv.tfpg2smv import compute_smv_node_name

from utils.fame_tfpg_process import FameTfpgProcess
import optparse
import tempfile
import logging
from src.models.fdir import FdirDiagResult
from src.models.correctness import ResultModel

from decimal import Decimal
import glob



class TfpgModel (ActivityModel):
    """Model of tfpg panel. It contains submodels for tfpg viewer, synthesis
        validation and errors.
    """

    tfpgs_counter = 0
    errors_counter = 0
    busy = 0
    tfpgs_list_changed = observable.Signal()
    
    # Observable properties
    __observables__ = ( "tfpgs_counter", "errors_counter", "busy", "tfpgs_list_changed",)

    def __init__(self, appl_model, slim_model, mission_model):        
        ActivityModel.__init__(self, appl_model, slim_model, mission_model)
        
        self.appl_model = appl_model
        self.slim_model = slim_model
        self.mission_model = mission_model
        
        # Models managed by TfpgModel

        self.tfpg_viewer_model = TfpgViewerModel(self, self.slim_model, self.mission_model)
        self.tfpg_slim_assoc_model = TfpgSlimAssocModel(self, self.slim_model, self.mission_model)

        self.tfpg_synthesis_model = TfpgSynthesisModel(self, self.slim_model)
        self.tfpg_behav_val_model = TfpgBehavValModel(self, self.slim_model)
        self.tfpg_effect_val_model = TfpgEffectValModel(self, self.slim_model)

        self.tfpg_errors_model = TfpgErrorsModel(self, self.slim_model, self.mission_model)

        # Monitors
        self.appl_model.tools.add_tool("Synthesis", self.tfpg_synthesis_model, "TFPG")
        self.appl_model.tools.add_tool("Behavioural Validation", self.tfpg_behav_val_model, "TFPG")
        self.appl_model.tools.add_tool("Effectiveness Validation", self.tfpg_effect_val_model, "TFPG")
        
        # Observes busy levels
        self.tfpg_synthesis_model.register_observer(self)
        self.tfpg_behav_val_model.register_observer(self)
        self.tfpg_effect_val_model.register_observer(self)
        
        # Sets the column of tfpgs list
        self._init_tfpgs_list()

        # It is the ParentView (TfpgView)
        self.tfpg_view = TfpgView()
        
        # It is the error message to show in the errors_buffer
        self.error_name = ""

        ##LB# It is the list of tfpgs that is composed of: tfpg_name, human_readable filename, saved 
        ##### (True if the tfpg is not modified, else False), tfpg_structural and txml_filename
        self.tfpgs_name_fn_saved_list = []
        
        return


    def _init_tfpgs_list(self):
        """Sets the Tfpgs in the column."""
        cr_text = gtk.CellRendererText()
        self.tfpgs_col = gtk.TreeViewColumn("TFPGs", cr_text)
        self.tfpgs_list = gtk.ListStore(gobject.TYPE_PYOBJECT)


        def tfpg_set_display_options(column, cell_renderer, model, iter):
            """Sets the text, foreground and background of the tfpgs cell_renderers. 
                @param   column:  The column Tfpgs.
                @type    column:  gtk.TreeViewColumn
                @param   cell_renderer:  The cell_renderer of the column Tfpgs.
                @type    cell_renderer:  gtk.CellRendererText
                @param   model:  The Tfpg model. 
                @type    model:  GtkListStore
                @param   iter:  The iter of the related Tfpg.
                @type    iter:  GtkTreeIter
            """
            tfpg = model.get_value(iter, 0)
            if isinstance(tfpg, Tfpg):
                self.tfpg_view.cell_renderer_set_text(cell_renderer, tfpg.get_name())
            return

        self.tfpgs_col.set_cell_data_func(cr_text, tfpg_set_display_options)
        return


    def get_tfpgs_list(self):
        """Gets the list of the tfpgs.
            @return   self.tfpgs_list:  The list of the tfpgs.
            @rtype    self.tfpgs_list:  gtk.ListStore
        """
        return self.tfpgs_list


    def get_tfpgs_col(self):
        """Gets the column of the tfpgs.
            @return   self.tfpgs_col:  The column of the tfpgs.
            @rtype    self.tfpgs_col:  gtk.TreeViewColumn 
        """
        return self.tfpgs_col


    def get_tfpgs_counter(self):
        """Gets the counter of the tfpgs in the list.
            @return   self.tfpgs_counter:  The number of the tfpgs in the list.
            @rtype    self.tfpgs_counter:  int
        """
        return self.tfpgs_counter


    def remove_tfpgs(self, tree_model, paths, state):
        """Removes the tfpgs selected from the list.
            @param   tree_model:  The tree model of the Tfpg.
            @type    tree_model:  TfpgModel.
            @param   paths:  The paths of the Tfpg.
            @type    paths:  list
            @param   state:  If the tfpg is removed or edited.
            @type    state:  string
        """
        if state == "Remove":
            for iter in [tree_model.get_iter(path) for path in paths]:
                self.tfpgs_list.remove(iter)
                   
            self.tfpgs_counter += -1
            self.tfpgs_list_changed.emit()
        else:
        
            for iter in [tree_model.get_iter(path) for path in paths]:
                self.tfpgs_list.remove(iter)
        
        return


    def name_exist(self, name, old_name=None):
        """Checks if there are more tfpg with the same name or a wrong name.
            @param   name:  The name of the Tfpg.
            @type    name:  string
            @param   old_name:  The name of the old Tfpg.
            @type    old_name:  string

            @return   Boolean:  False if there are more tfpg with the same name or 
                                 a wrong name, otherwise True
            @rtype    Boolean:  boolean
        """
        if (name is None) or (name == "") or (name.lower == "none"):
            return False
        elif name == old_name:
            return False
        else:
            for n in self.tfpgs_list:
                if n[0].get_name() == name:
                    return True
            return False


    def get_existing_names(self):
        """Gets a list with the all tfpgs name in the column.
            @return   result:  The list of tfpgs name.
            @rtype    result:  list
        """
        result = []
        for tfpg in self.tfpgs_list:
            result.append(tfpg[0].get_name())
        return result


    def get_errors_counter(self):
        """Gets the counter of the errors in nodes, edges or tfpgs.
            @return   self.errors_counter:  The number of the errors.
            @rtype    self.errors_counter:  int
        """
        return self.errors_counter


    def clear_tfpgs_list(self):
        """Clears the list of Tfpgs."""
        self.tfpgs_name_fn_saved_list = []
        self.tfpgs_list.clear()
        self.tfpgs_counter = 0
        self.tfpgs_list_changed.emit()
        return
        

    def clear(self):
        """Clears all elements of subviews"""
        self.tfpg_view.clear_subviews()
        self.tfpgs_list_changed.emit()
        return
        
   
    def get_tfpg_doc(self, filename):
        """This function loads the a TFPG from the filename, it parses the file and checks
	        validation and syntax errors.
	        Two checks are performed: 
		        - a Discrepancy must have at least one incoming edge
		        - it is not possible to have more tfpgs with the same tfpg name
	        It converts the TfpgDoc into the Tfpg internal struct and it appends the 	
            loaded tfpgs into the Tfpgs List.
            @param   filename:  The path of the loaded tfpg file.
	        @type    filename:  string
	
	        @param   tfpg_doc:  The TfpgDocument of the loaded TFPG.
	        @type    tfpg_doc:  TfpgDocument
        """

        self.busy += 1
        self.clear()

	    #LB# Checks if the TFPG file exists.
        if not os.path.exists(filename):
            self.errors_counter = 1
            self.error_name = "The file: " + relpath(filename) + " does not exist!"
            self.slim_model.console.log_compiler_fail(
                " > ERROR\n" "\t > See the TFPG Errors Window for more details.\n")
            self.busy -= 1
            return
        else: self.errors_counter = 0

	    #LB# Gets the TfpgDocument and any validation errors. Parses the TFPG file.
        tfpg_doc, validation_errors = TfpgDocument().parse(filename)

        if len(validation_errors) > 0:
            for err in validation_errors:
                self.errors_counter = 1
                self.error_name = "The file: " + relpath(filename) + " contains validation errors:\n\t " + str(err)
                self.slim_model.console.log_compiler_fail(
                    " > ERROR\n" "\t > See the TFPG Errors Window for more details.\n")
            self.busy -= 1
            return 
        else: self.errors_counter = 0

        #LB# Gets any syntax errors. 
        syntax_errors = TfpgDocument().get_syntax_errors_in_tfpg_document(tfpg_doc)
        self.slim_model.console.log_compiler_msg("Loading TFPG '%s'... " % filename)

        if len(syntax_errors) > 0:
            for err in syntax_errors:
                self.errors_counter = 1
                self.error_name = "The file: " + relpath(filename) + " contains syntax errors:\n\t " + str(err)
                self.slim_model.console.log_compiler_fail(
                    " > ERROR\n" "\t > See the TFPG Errors Window for more details.\n")
            self.busy -= 1
            return 
        else: self.errors_counter = 0

        self.slim_model.console.log_compiler_ok("OK\n")

        #LB# Converts the TFPG file from the TfpgDocument to the TfpgInternalStruct
        tfpg_internal_struct = tfpg_doc_to_tfpg_internal_struct(tfpg_doc)
       
       
	    #LB# Gets any errors from the Tfpg. 
        error_list = tfpg_internal_struct.get_errors_list()
        for error in error_list:
            self.errors_counter = 1
            self.error_name = "The file: " + relpath(filename) + " contains syntax errors:\n\t" + str(error)
            self.slim_model.console.log_compiler_fail(
                " > ERROR\n" "\t > See the TFPG Errors Window for more details.\n")
            self.busy -= 1
            return
        else: self.errors_counter = 0
        
        #LB# Gets any errors from the Tfpg Edges. 
        for node in tfpg_internal_struct.get_nodes_list():
            for error in node.get_node_errors_list():
                self.errors_counter = 1
                self.error_name = "The file: " + relpath(filename) + " contains syntax errors:\n\t" + str(error)
                self.slim_model.console.log_compiler_fail(
                    " > ERROR\n" "\t > See the TFPG Errors Window for more details.\n")
                self.busy -= 1
                return
        else: self.errors_counter = 0
        
        #LB# Gets any errors from the Tfpg Nodes. 
        for edge in tfpg_internal_struct.get_edges_list():
            for error in edge.get_edge_errors_list():
                self.errors_counter = 1
                self.error_name = "The file: " + relpath(filename) + " contains syntax errors:\n\t" + str(error)
                self.slim_model.console.log_compiler_fail(
                    " > ERROR\n" "\t > See the TFPG Errors Window for more details.\n")
                self.busy -= 1
                return
        else: self.errors_counter = 0
        
        
        #LB# Checks if a Discrepancy must have at least one incoming edge
        error = False
        discrepancy = ""
        nodes_name_type_list = []
        edges_destNode_list = []

        for node in tfpg_internal_struct.get_nodes_list():
            nodes_name_type_list.append([node.get_name(), node.get_type()])

        for edge in tfpg_internal_struct.get_edges_list():
            edges_destNode_list.append(edge.get_destNode().get_name())

        for nd in nodes_name_type_list:
            if (nd[1] != NodeType.FM) and (nd[0] not in edges_destNode_list):
                error = True
                discrepancy = nd[0]

        if error == True:
            self.errors_counter = 1
            self.error_name = "The file: " + relpath(filename) + " contains syntax errors:\n\t" + str(DiscrepancyWithoutIncomingEdge(discrepancy))
            self.slim_model.console.log_compiler_fail(
                    " > ERROR\n" "\t > See the TFPG Errors Window for more details.\n")
            self.busy -= 1
            return
        else: self.errors_counter = 0

        #LB# Checks that it is not possible to have more tfpgs with the same tfpg name
        for tfpg in self.tfpgs_list:
            if tfpg[0].get_name() == tfpg_internal_struct.get_name():
                self.errors_counter = 1
                self.error_name = "The file: " + relpath(filename) + " contains syntax errors:\n\t" + \
                                    "This TFPG " + tfpg[0].get_name() + \
                                    " already exists or there is an other TFPG with the same name."
                self.slim_model.console.log_compiler_fail(
                        " > ERROR\n" "\t > See the TFPG Errors Window for more details.\n")
                self.busy -= 1
                return
        else: self.errors_counter = 0

        #LB# Checks if the tfpg name is valid: it must be not none, not empty and only 
        #LB# composed of alphanumeric + "-" and "_" characters.
        import re
        regex = re.compile('^[a-zA-Z0-9_-]+$')

        if tfpg_internal_struct.get_name() == "none" or tfpg_internal_struct.get_name() == "" or \
            (not regex.match(tfpg_internal_struct.get_name())):
            
            self.errors_counter = 1
            error = InvalidNameError(tfpg_internal_struct.get_name())
            self.error_name = "The file: " + relpath(filename) + " contains syntax errors:\n\t" + str(error)
            self.slim_model.console.log_compiler_fail(
                    " > ERROR\n" "\t > See the TFPG Errors Window for more details.\n")
            self.busy -= 1
            return
        else: self.errors_counter = 0


        """Loading Tfpgs"""

        self.tfpgs_list.append([tfpg_internal_struct])
        self.tfpgs_counter += 1
        self.tfpgs_list_changed.emit()
        self.slim_model.console.log_compiler_msg(" > Loaded the TFPG: " + tfpg_internal_struct.get_name() + ".\n")


        #LB# Checks if there is, in the same current dir, a file.tfpg related to the file.txml loaded
        hr_tfpg_fn = ""                                         #Human Readable Filename
        txml_filename = filename                                #XML Filename
        txml_basename = os.path.basename(txml_filename)         #XML Basename
        txml_fn_no_ext = os.path.splitext(txml_basename)[0]     #XML Basename without extension
                
                
        old_curr_dir = os.getcwd()                              #Get the current dir                      
        curr_dir = os.path.dirname(txml_filename)               #Get the current dir of the XML Filename 
        os.chdir(curr_dir)                                      #Set the current dir of the XML Filename 

        
        for tfpg_basename in glob.glob("*.tfpg"):
        
            tfpg_fn_no_ext = os.path.splitext(tfpg_basename)[0] #Human Readable Basename without extension
                    
            if tfpg_fn_no_ext == txml_fn_no_ext:
                hr_tfpg_fn = os.path.abspath(tfpg_basename)     #Human Readable Filename in the same dir of XML Filename
                      
                                      
        self.tfpgs_name_fn_saved_list.append([tfpg_internal_struct.get_name(), hr_tfpg_fn, True, tfpg_internal_struct, txml_filename])

        os.chdir(old_curr_dir) #Set the old current dir (it is used to return at Code/tools)
        
        self.busy -= 1
        return tfpg_doc


    def get_tfpg_document(self, tfpg_name):
        """Gets the TfpgDocument of the related Tfpg name.
            This function is used for the dump of TFPG in the tool_handler model.
            @param   tfpg_name:  The name of the related TFPG.
            @type    tfpg_name:  string

            @return   tfpg_doc:  The TfpgDocument of the related Tfpg name.
            @rtype    tfpg_doc:  TfpgDocument
        """

        if tfpg_name in self.get_existing_names():
            for el in self.tfpgs_list:
                tfpg_structural = el[0]
                if tfpg_structural.get_name() == tfpg_name:
                    tfpg_doc = tfpg_internal_struct_to_tfpg_doc(tfpg_structural)
		    return tfpg_doc


    def get_tfpg_structural(self, tfpg_name):
        """Gets the Tfpg Structural of the related Tfpg name. 
            @param   tfpg_name:  The name of the related TFPG.
            @type    tfpg_name:  string

            @return   tfpg_structural:  The Tfpg of the related Tfpg name.
            @rtype    tfpg_structural:  Tfpg
        """
        if tfpg_name in self.get_existing_names():
            for el in self.tfpgs_list:
                tfpg_structural = el[0]
                
                if tfpg_structural.get_name() == tfpg_name:
		    return tfpg_structural


    def get_tfpg_filename(self, tfpg_name):
        """Gets the filename of the related Tfpg.
            @return   tfpg[1]:  The filename of the related Tfpg.
            @rtype    tfpg[1]:  string
        """
        for tfpg in self.get_tfpgs_name_fn_saved_list():
            if tfpg[0] == str(tfpg_name):
                return tfpg[4]

    
    def get_tfpgs_name_fn_saved_list(self):
        """Gets the list of name, filename and saved of the related Tfpg.
            @return   self.tfpgs_name_fn_saved_list:  The list of name, filename and saved of the related Tfpg
            @rtype    self.tfpgs_name_fn_saved_list:  list
        """
        return self.tfpgs_name_fn_saved_list
        
        
    def parse_concrete_syntax_file(self, filename):
        """Parses a file written in human readable file and returns the
            corresponding tfpg internal structure.
            @param   filename:  The path of the tfpg file.
            @type    filename:  string

            @return   tfpg_graph:  The Tfpg structural.
            @rtype    tfpg_graph:  Tfpg
        """

        import antlr3

        from tfpg_viewer.grammar.generated.TfpgLexer import TfpgLexer
        from tfpg_viewer.grammar.generated.TfpgParser import TfpgParser
        
        char_stream = antlr3.ANTLRFileStream(filename)
        lexer = TfpgLexer(char_stream)
        tokens = antlr3.CommonTokenStream(lexer)
        parser = TfpgParser(tokens)

        errors = []

        r = parser.tfpg()
        t = r.tree

        tfpg_graph = Tfpg()

        
        
        for child in t.getChildren():

            if str(child) == "NAME":
                tfpg_graph.set_name(str(child.getChildren()[0]))

            elif str(child) == "FM":
                if len(child.getChildren()) == 1:
                    tfpg_graph.add_fm_node(str(child.getChildren()[0]))
                else:
                    errors.append(FMNotCorrectError)

            elif str(child) == "AND":
                if len(child.getChildren()) == 2:
                    tfpg_graph.add_and_node(str(child.getChildren()[0]), True)
                else:
                    tfpg_graph.add_and_node(str(child.getChildren()[0]), False)

            elif str(child) == "OR":
                if len(child.getChildren()) == 2:
                    tfpg_graph.add_or_node(str(child.getChildren()[0]), True)
                else:
                    tfpg_graph.add_or_node(str(child.getChildren()[0]), False)

            elif str(child) == "MODES":
                modes = []
                for ch in child.getChildren():
                    modes.append(str(ch))
                tfpg_graph.set_modes_list(modes)

            elif str(child) == "EDGE":
                edg_name = str(child.getChildren()[0])
                edg_src = str(child.getChildren()[1])
                edg_dst = str(child.getChildren()[2])
                edg_tMin = float(str(child.getChildren()[3]))
                if str(child.getChildren()[4]) == "INFINITY":
                    edg_tMax = -1
                else:
                    edg_tMax = float(str(child.getChildren()[4]))
                 
                 
                edg_modes = [] 
                if str(child.getChildren()[5]) == "ALL": 
                    for m in modes:
                        edg_modes.append(str(m))
                else:      
                    for ch in (child.getChildren()[5]).getChildren():
                        edg_modes.append(str(ch))
                        
                tfpg_graph.add_edge(tfpg_graph.get_node_by_name(edg_src), 
                                    edg_tMin, edg_tMax, edg_modes,
                                    tfpg_graph.get_node_by_name(edg_dst), 
                                    edg_name)
        
        for tfpg_err in tfpg_graph.get_errors_list():
            errors.append(tfpg_err)
        if errors != []:
            for err in errors:
                raise err

        return tfpg_graph
        
        
    # ---------------------------------------------------------------
    #        Notifications from observable properties
    # ---------------------------------------------------------------    
    def property_busy_value_change(self, model, old, new):
        if model is self: return # not interested in introspection
        self.busy += (new - old)
        return

    pass # end of class
    
    
    def property_tfpgs_list_changed_signal_emit(self, model, args):
        """When a change occured in the GUI, a signal clears all.
            @param   model:  The Slim Model.
            @type    model:  SlimModel
            @param   args:  The arguments of Mission Model.
            @type    args:  string
        """
        self.clear()
        return

        
    def gcd_tfpg_propagation_times(self, tfpg_structural):
        """Gets the GCD of tfpg propagation times. It is used to calculate 
            the scaling_factor / sample_rate used in the TFPG Effectiveness Validation
            and FAME Synthesis.
            @param   tfpg_structural: The TFPG Internal Struct.
            @type    tfpg_structural:  TFPG
            
            @param   res:  The GCD of tfpg propagation times.
            @type    res:  float
        """
        # List of timings constants of TFPG without INFINITY (-1)
        timings_list = tfpg_structural.get_timings_constants(True)
        
        # Import the general GCD functions from utils.misc.py
        from utils.misc import compute_greatest_common_divisor
        
        # Compute the GCD of tmin and tmax lists
        gcd_time = compute_greatest_common_divisor(timings_list)
        
        # If all TFPG propagation times are 0, the GCD is set 1.
        if gcd_time == 0: return 1.0
        else: return float(gcd_time)
        
      
pass # end of TfpgModel class
# ----------------------------------------------------------------------


class TfpgViewerModel (Model):
    """Model for Tfpg Viewer."""


    def __init__(self, tfpg_model, slim_model, mission_model):
        Model.__init__(self)
        
        self.tfpg_model = tfpg_model
        self.slim_model = slim_model
        self.mission_model = mission_model
        self.slim_model.register_observer(self)
        return
        

pass # end of TfpgViewerModel class
# ----------------------------------------------------------------------


class TfpgSlimAssocModel (Model):
    """Model for Tfpg Viewer."""


    def __init__(self, tfpg_model, slim_model, mission_model):
        Model.__init__(self)
        
        self.tfpg_model = tfpg_model
        self.slim_model = slim_model
        self.mission_model = mission_model
        self.slim_model.register_observer(self)
        
        # Dictionaries of <tfpg_names, dict(slim_expressions)> pairs
        self.tfpgs_slim_associations = dict()
        
        self.curr_tfpg_name = ""
        
        return
        
        
    def add_associations_list(self, tfpg_name, associations_ls):
        """Function which adds the corrispondent dictionary of <var_name, slim_expr>
            pairs for a given tfpg. Notice that this is usefule in the TFPG Mode tab.
            @param   tfpg_name:  The name of the related TFPG.
            @type    tfpg_name:  string
            @param   associations_ls:  The list of slim_associations
            @type    associations_ls:  list
        """
        for k in self.tfpgs_slim_associations.keys():
            if k == tfpg_name:
                self.tfpgs_slim_associations[tfpg_name] = associations_ls
                return
        
        self.tfpgs_slim_associations[tfpg_name] = associations_ls
        
        return


    def get_slim_associations(self, tfpg_name):
        """Returns the dictionary of <var_name, slim_expr> relative to the specific tfpg.
            @param   tfpg_name:  The name of the related TFPG.
            @type    tfpg_name:  string

            @return   self.tfpgs_slim_associations[k]:  The dictionary relative to the specific tfpg.
            @rtype    self.tfpgs_slim_associations[k]:  dict
        """
        for k in self.tfpgs_slim_associations.keys():

            if k == tfpg_name:
                self.curr_tfpg_name = k
                return self.tfpgs_slim_associations[k]

        return


    def get_slim_associations_from_file(self, tfpg_name, filename):
        """Converts an .axml file into a list of associations.
            @param   tfpg_name:  The name of the related TFPG.
            @type    tfpg_name:  string
            @param   filename:  The path of the tfpg file.
            @type    filename:  string
	
            @return  (assoc_ls, errors_list):  A pair containing the dictionary 
                                               <var_name, slim_expr> related to the 
                                               specific tfpg and the list of errors.
            @rtype   (assoc_ls, errors_list):  tuple
         """
        errors_list = []
        if not os.path.exists(relpath(filename)):
            self.errors_counter = 1
            errors_list.append("ERROR\n\n The file: '" + filename + "' does not exist.\n")
            self.slim_model.console.log_compiler_fail(
                " > ERROR\n" "\t > See the TFPG Errors Window for more details.\n")
            return None, errors_list
        else: self.errors_counter = 0

        assoc_doc, validation_errors = AssociationsDocument().parse(filename)
        
        if len(validation_errors) > 0:
            for err in validation_errors:
                self.errors_counter = 1
                errors_list.append("VALIDATION ERRORS\n\nThe file: '" + filename + 
                                   "' contains validation " + "erros:\n\n \t" + str(err) + "\n")
                self.slim_model.console.log_compiler_fail(
                    " > ERROR\n" "\t > See the TFPG Errors Window for more details.\n")
                return None, errors_list
        else: self.errors_counter = 0

        syntax_errors = AssociationsDocument.get_syntax_errors_in_assoc_document(assoc_doc)

        if len(syntax_errors) > 0:
            for err in syntax_errors:
                self.errors_counter = 1
                errors_list.append("SYNTAX ERRORS\n\n The file: '" + str(filename) + 
                                   "' contains syntax errors:\n\n" +
                                   " \t" + str(err) + "\n")
                self.slim_model.console.log_compiler_fail(
                    " > ERROR\n" "\t > See the TFPG Errors Window for more details.\n")
                return None, errors_list
        else: self.errors_counter = 0
 
        assoc_ls = dict()

        ## An AssociationDocument contains 4 different types of associations.
        ## For each one of them, we must add an associations in assoc_ls; the 3rd parameter
        ## of the tuple we associated to each variable will say which kind of association it is
        for assoc in assoc_doc.failureModes_assoc_ls:
            _id = assoc.get('id')
            expr = assoc.get('expr')
            assoc_ls[_id] = (None, expr, "FailureMode")
 
        for assoc in assoc_doc.monitoredDiscr_assoc_ls:
            _id = assoc.get('id')
            expr = assoc.get('expr')
            assoc_ls[_id] = (None, expr, "MonitoredDiscr")

        for assoc in assoc_doc.unmonitoredDiscr_assoc_ls:
            _id = assoc.get('id')
            expr = assoc.get('expr')
            assoc_ls[_id] = (None, expr, "UnmonitoredDiscr")

        for assoc in assoc_doc.tfpgModes_assoc_ls:
            _id = assoc.get('id')
            expr = assoc.get('expr')
            assoc_ls[_id] = (None, expr, "TfpgMode")

        self.tfpgs_slim_associations[tfpg_name] = assoc_ls
 
        return assoc_ls, errors_list
        
        
    def slim_associations_without_errors(self, associations):
        """Function which checks that all the associated slim expressions are correct.
	        @return    True or False: True if all associated slim expressions are correct, else False.
            @rtype     True or False: boolean
	    """
        for (k, v) in associations.items():
            (val, error, _type) = v
            if val == "ERROR" : return False

        return True
        
          
        
    def get_errors_from_associations(self, tfpg_name):
        """Function which checks that all the associated slim expressions are correct.
            @return    True or False: True if all associated slim expressions are correct, else False.
            @rtype     True or False: boolean
        """

        if tfpg_name is "": self.curr_tfpg_name = tfpg_name
        
        slim_associations = self.get_slim_associations(tfpg_name)

        if slim_associations is None: return ["No SLIM Associations file provided."]

        errors_list = []

        for (k, v) in slim_associations.items():
            (val, error, _type) = v
            
            if val == "ERROR":
                errors_list.append("In the TFPG -> SLIM Association Tab, the " + _type + " slim association '" + k + "' is wrong: \n\t" + error)
            else: continue
        
        if errors_list != []: 
            return errors_list

        return []

pass # end of TfpgSlimAssocModel class
# ----------------------------------------------------------------------


class TfpgSynthesisModel (FameTfpgHandlerModel):
    """Model for handling TFPG Synthesis"""

    busy = 0
    synthesis_sat_bound = options.NUSMV_BMC_LENGTH
        
    # Observable properties
    __observables__ = ("busy", "synthesis_sat_bound", )


    def __init__(self, tfpg_model, slim_model):
        """Constructor Method"""
        FameTfpgHandlerModel.__init__(self, slim_model, tfpg_model)
        
        self.tfpg_model = tfpg_model
        self.slim_model = slim_model

        # Dictionary of <tfpg_names, dict(slim_expressions)> pairs
        self.tfpgs_slim_associations_in_synth = dict()

        self.slim_model.register_observer(self)
        return
        
    def stop_tfpg_synth(self):
        """Stops the TFPG Synthesis"""
        self.stop_fame_tfpg()
        return ("Warning! TFPG Synthesis Stopped!", None)
        

    def synthesize_tfpg(self, tfpg_name):
        """Performs TFPG Synthesis and generates the results."""

        result = None
            
        # Get the SLIM Associations
        associations = self.get_slim_associations_in_synth(tfpg_name)
        
        if associations is not None:
        
            ## This is an other check of the correctness of SLIM Associations.
            ## Can we remove it?
            ok = self.slim_associations_without_errors_in_synth(associations)

            if ok is True:
            
                self.busy += 1
                try:
                    # Synthesis is carried out on the extended SLIM model
                    root_impl = self.slim_model.get_extended_root_impl()
                    
                    # compute smv filename (this could be done more elegantly)
                    self.slim_model.smv_fname = None
                    self.smv_file, self.ord_file, self.need_msat, generated, killed = \
                        self.generate_smv_ord(True)  
                    if killed:
                        self.busy -= 1
                        return # got killed by event!
                    
                    # for effectiveness analysis, mathsat is needed because of dense-time
                    # clock variables
                    self.need_msat = True

                    # the BDD ordering file computed above is useful only for BDD-based analyses
                    self.ord_file = None
                    
                    log_out, logfname = tempfile.mkstemp(suffix='.fametfpgout',
                        prefix=os.path.splitext(self.smv_file)[0], dir=".", text=True)

                    from tfpg2smv.tfpg2smv import dump_extension_for_synthesis_to_smv
                    
                    dump_extension_for_synthesis_to_smv(root_impl, associations, self.smv_file)
                    
                    # It is used to know the dir when the .xml and .txml are generated
                    
                    sa_fm_list_name = os.path.dirname(os.path.splitext(self.smv_file)[0]) + "/" + tfpg_name + "_sa_fm_list.xml"

                    from tfpg2smv.tfpg2smv import dump_discrepancies_for_synthesis_to_xml
                    dump_discrepancies_for_synthesis_to_xml (associations, sa_fm_list_name)

                    self.args = ["-sa_compass", "-sa_compass_task", sa_fm_list_name]

                    tfpg_res_name = os.path.dirname(os.path.splitext(self.smv_file)[0]) + "/" + tfpg_name + ".txml"
                    slim_assoc_res_name = os.path.dirname(os.path.splitext(self.smv_file)[0]) + "/" + tfpg_name + ".axml"

                    system_modes = []
                    failure_modes = []
                    discrepancies = []
                    
                    for (k, v) in associations.items():
                        (val, error, _type) = v

                        if _type == "TfpgMode":
                            system_modes.append(k)
                        elif _type == "FailureMode": 
                            failure_modes.append(k)
                        elif _type == "MonitoredDiscr":
                            discrepancies.append((k, "monitored"))
                        elif _type == "UnmonitoredDiscr": 
                            discrepancies.append((k, "not_monitored"))
                        else:
                            # !!Warning!! This code should never be reached
                            raise Exception("Unknown Association Type")


                    self.start_fame_tfpg()
                    res = self.get_tool().synthesize_tfpg(self.smv_file, tfpg_res_name,
                                                          system_modes, failure_modes, discrepancies,
                                                          tfpg_name, self.synthesis_sat_bound)         
                    self.stop_fame_tfpg()

                    if res == "ok": 
                        result = (res, tfpg_res_name, slim_assoc_res_name)
                        self.slim_model.debug_info.tfpg_filename = tfpg_res_name
                    else: 
                        result = (res, None, None)
              
                finally:
                    self.busy -= 1
              
            else:
                result = ("ERROR in SLIM Associations: some Slim expressions are not correct!", None)
                    
        else:
            result = ("ERROR in SLIM Associations: No SLIM associations are loaded!", None)

        return result



    def add_associations_list_in_synth(self, associations_ls, tfpg_name):
        """Function which adds the corrispondent dictionary of <var_name, slim_expr>
            pairs for a given tfpg.
            @param   tfpg_name:  The name of the related TFPG.
            @type    tfpg_name:  string
            @param   associations_ls:  The list of slim_associations
            @type    associations_ls:  list
        """
        for k in self.tfpgs_slim_associations_in_synth.keys():
            if k == tfpg_name:
                self.tfpgs_slim_associations_in_synth[tfpg_name] = associations_ls
                return
        
        self.tfpgs_slim_associations_in_synth[tfpg_name] = associations_ls
        
        return



    def get_slim_associations_in_synth(self, tfpg_name):
        """Returns the dictionary of <var_name, slim_expr> relative to the specific tfpg.
           @param   tfpg_name:  The name of the related TFPG.
           @type    tfpg_name:  string

           @return   self.tfpgs_slim_associations_in_synth[k]:  The dictionary relative to the specific tfpg.
           @rtype    self.tfpgs_slim_associations_in_synth[k]:  dict
        """
        for k in self.tfpgs_slim_associations_in_synth.keys():

            if k == tfpg_name:
                return self.tfpgs_slim_associations_in_synth[k]

        return



    def get_slim_associations_in_synth_from_file(self, filename, tfpg_name):
        """Converts an .axml file into a list of associations.
           Internally, they are associated to a tfpg called 'tfpg_name'

            @param   filename:  The path of the tfpg file.
            @type    filename:  string
            @param   tfpg_name:  The name of the related TFPG.
            @type    tfpg_name:  string

            @return  (assoc_ls, errors_list):  A pair containing the dictionary 
                                               <var_name, slim_expr> related to the 
                                               specific tfpg and the list of errors.
            @rtype   (assoc_ls, errors_list):  tuple
         """
        errors_list = []
        if not os.path.exists(relpath(filename)):
            self.errors_counter = 1
            errors_list.append("ERROR\n\n The file: '" + filename + "' does not exist.\n")
            self.slim_model.console.log_compiler_fail(
                " > ERROR\n" "\t > See the TFPG Errors Window for more details.\n")
            return None, errors_list
        else: self.errors_counter = 0

        assoc_doc, validation_errors = AssociationsDocument().parse(filename)
        
        if len(validation_errors) > 0:
            for err in validation_errors:
                self.errors_counter = 1
                errors_list.append("VALIDATION ERRORS\n\nThe file: '" + filename + 
                                   "' contains validation " + "erros:\n\n \t" + str(err) + "\n")
                self.slim_model.console.log_compiler_fail(
                    " > ERROR\n" "\t > See the TFPG Errors Window for more details.\n")
                return None, errors_list
        else: self.errors_counter = 0

        syntax_errors = AssociationsDocument.get_syntax_errors_in_assoc_document(assoc_doc)

        if len(syntax_errors) > 0:
            for err in syntax_errors:
                self.errors_counter = 1
                errors_list.append("SYNTAX ERRORS\n\n The file: '" + str(filename) + 
                                   "' contains syntax errors:\n\n" +
                                   " \t" + str(err) + "\n")
                self.slim_model.console.log_compiler_fail(
                    " > ERROR\n" "\t > See the TFPG Errors Window for more details.\n")
                return None, errors_list
        else: self.errors_counter = 0
 
        assoc_ls = dict()
        
        fm_nodes_list = []
        mon_discr_list = []
        unmon_discr_list = []
        tfpg_modes_list = []

        ## An AssociationDocument contains 4 different types of associations.
        ## For each one of them, we must add an associations in assoc_ls; the 3rd parameter
        ## of the tuple we associated to each variable will say which kind of association it is
        for assoc in assoc_doc.failureModes_assoc_ls:
            _id = assoc.get('id')
            expr = assoc.get('expr')
            assoc_ls[_id] = (None, expr, "FailureMode")
            fm_nodes_list.append(_id)
 
        for assoc in assoc_doc.monitoredDiscr_assoc_ls:
            _id = assoc.get('id')
            expr = assoc.get('expr')
            assoc_ls[_id] = (None, expr, "MonitoredDiscr")
            mon_discr_list.append(_id)

        for assoc in assoc_doc.unmonitoredDiscr_assoc_ls:
            _id = assoc.get('id')
            expr = assoc.get('expr')
            assoc_ls[_id] = (None, expr, "UnmonitoredDiscr")
            unmon_discr_list.append(_id)

        for assoc in assoc_doc.tfpgModes_assoc_ls:
            _id = assoc.get('id')
            expr = assoc.get('expr')
            assoc_ls[_id] = (None, expr, "TfpgMode")
            tfpg_modes_list.append(_id)

        self.tfpgs_slim_associations_in_synth[tfpg_name] = assoc_ls
 
        return assoc_ls, errors_list, fm_nodes_list, mon_discr_list, unmon_discr_list, tfpg_modes_list



    def get_slim_associations_in_synth_from_gui(self, tfpg_name,
                                                tv_failure_modes_model,
                                                tv_monitored_discr_model,
                                                tv_not_monitored_discr_model,
                                                tv_tfpg_modes_model):
        """Retrieves and creates an associations dictionary based on the information collected from
           the synthesis tab of the TFPG panel
           Internally, these associations are linked to the tfpg called <tfpg_name>

           @param tfpg_name: the name of the tfpg which the associations (will) belong to
           @type  tfpg_name: string
           @param tv_failure_modes_model: the failure modes <name-expression> associations taken from the gui
           @type  tv_failure_modes_model: gtk.ListStore
           @param tv_monitored_discr_model: the monitored discrepancies <name-expression> associations taken from the gui
           @type  tv_monitored_discr_model: gtk.ListStore
           @param tv_not_monitored_discr_model: the not-monitored discrepancies <name-expression> associations taken from
                                                the gui
           @type  tv_not_monitored_discr_model: gtk.ListStore
           @param tv_tfpg_modes_model: the tfpg modes <name-expression> associations taken from the gui
           @type  tv_tfpg_modes_model: gtk.ListStore
           @return: a dict associating to <tfpg_name> the list of associations (in form of (ok, expr, val)) taken from the gui
           @rtype: dict
        """
 
        assoc_ls = dict()

        for row in tv_failure_modes_model:
        
            if row[0] != "":
            
                _id = row[0]
                expr = row[2]
                ok = row[3]
                assoc_ls[_id] = (ok, expr, "FailureMode")

        for row in tv_monitored_discr_model:
        
            if row[0] != "":
                _id = row[0]
                expr = row[2]
                ok = row[3]
                assoc_ls[_id] = (ok, expr, "MonitoredDiscr")

        for row in tv_not_monitored_discr_model:
        
            if row[0] != "":
            
                _id = row[0]
                expr = row[2]
                ok = row[3]
                assoc_ls[_id] = (ok, expr, "UnmonitoredDiscr")

        for row in tv_tfpg_modes_model:  
          
            if row[0] != "":
                
                _id = row[0]
                expr = row[2]
                ok = row[3]
                assoc_ls[_id] = (ok, expr, "TfpgMode")

        self.tfpgs_slim_associations_in_synth[tfpg_name] = assoc_ls
 
        return assoc_ls



    def slim_associations_without_errors_in_synth(self, associations):
        """Function which checks that all the associated slim expressions are correct.
        @return    True or False: True if all associated slim expressions are correct, else False.
        @rtype     True or False: boolean
        """
        for (k, v) in associations.items():
            (val, error, _type) = v
            if val == "ERROR" : return False
            elif val == "NOT COMPLETED" : return False

        return True

pass # end of TfpgSynthesisModel class
# ----------------------------------------------------------------------


class TfpgBehavValModel (FameTfpgHandlerModel):
    """Model for handling TFPG Behavioural Validation"""

    busy = 0
    behav_val_result = None
    behav_val_sat_bound = options.NUSMV_BMC_LENGTH
    behav_val_use_sat = options.NUSMV_USE_BMC
        
    # Observable properties
    __observables__ = ("busy", "behav_val_result", "behav_val_sat_bound", "behav_val_use_sat", )
    
    
    def __init__(self, tfpg_model, slim_model):
        """Constructor Method"""
        FameTfpgHandlerModel.__init__(self, slim_model, tfpg_model)
        return
        
        
    def generate_smv_file (self, tfpg_structural):

        # compute smv filename (this could be done more elegantly)
        self.slim_model.smv_fname = None
        self.smv_file, self.ord_file, self.need_msat, generated, killed = \
            self.generate_smv_ord(True)
        if killed: return # got killed by event!

        # for effectiveness analysis, mathsat is needed because of dense-time
        # clock variables
        self.need_msat = True

        # the BDD ordering file computed above is useful only for BDD-based analyses
        self.ord_file = None
 
        # behavioral analysis is carried out on the extended SLIM model
        root_impl = self.slim_model.get_extended_root_impl()

        # retrieve TFPG associations
        assoc_model = self.tfpg_model.tfpg_slim_assoc_model
        associations = assoc_model.get_slim_associations(tfpg_structural.get_name())
        ok = assoc_model.slim_associations_without_errors(associations)

        if associations is not None and ok is True:
            ## GZ: old code
            # associations_simple = {}
            # for k in associations:
            #     associations_simple[k] = associations[k][1]

            # call dumping function
            from tfpg2smv.tfpg2smv import dump_extended_sys_as_smv
            dump_extended_sys_as_smv(root_impl, tfpg_structural, associations, self.smv_file)
            
        else:
            self.behav_val_result = TfpgBehvalResult("no associations", None, None, None)

        return
        

    def stop_behav_val(self):
        """Stops the TFPG Behavioural Validation"""
        self.stop_fame_tfpg()
        self.behav_val_result = TfpgDiagResult("stop", None, None)
        return 


    def validate_behavior(self, tfpg_structural):
        """Performs TFPG Behavioural Validation and generates the results."""

        self.busy += 1
        try:
        
            # generate SMV file for analysis
            self.generate_smv_file(tfpg_structural)

            # run completeness check

            self.start_fame_tfpg()
            res = self.get_tool().check_tfpg_completeness(tfpg_structural, self.behav_val_sat_bound)            
            self.stop_fame_tfpg()

            # display result

            result = res['result'] # incomplete/complete/bounded_completeness/error
            counterexample = res['counterexample']
            counterexample_description = res['counterexample_description']

            if counterexample is not None:
                sys_trace = counterexample['sys_trace']
                tfpg_trace = counterexample['tfpg_trace']
                self.behav_val_result = TfpgBehvalResult('incomplete', counterexample_description, sys_trace, tfpg_trace)

            else:

                if result == "error":
                    self.behav_val_result = TfpgBehvalResult(None, None, None, None)
                            
                elif result == "complete":
                    self.behav_val_result = TfpgBehvalResult(result, None, None, None)
                                    
                else:
                    self.behav_val_result = TfpgBehvalResult(result, None, None, None)

        finally:
            self.busy -= 1

        return 



    def continuous_var_in_associations(self, tfpg_name):
        """Check whether some slim expression refer to continuous variables."""

        from slim_compiler.src.smv.walkers import InfiniteDomainFinderWalker
        from slim_compiler.src.dataStructure import nominalbehaviour

        def get_clock_data(comp_impl, names, clock_data):
            """
            Recursive functions which appends to the clock_data list, passed as parameter,
            the complete name of the clock data of the model.
            """

            for name, dsc in comp_impl.data_subcomponents.iteritems():
                if isinstance(dsc.data_type,
                              (nominalbehaviour.Clock,
                               nominalbehaviour.Continuous)):
                    complete_name = "%s.%s" %(".".join(names), name)
                    clock_data.append(complete_name)
            for sc_name in sorted(comp_impl.non_data_subcomponents):
                sc = comp_impl.non_data_subcomponents[sc_name]
                sc_impl = sc.get_referenced_component_implementation()
                names.append(sc_name)
                get_clock_data(sc_impl, names, clock_data)
                del names[-1]
            return

        assoc_model = self.tfpg_model.tfpg_slim_assoc_model
        associations = assoc_model.get_slim_associations(tfpg_name)
        
        root_impl = self.slim_model.get_extended_root_impl()
        compile_info = root_impl.subpackage.package.compile_information

        clock_data = []
        continuous_assoc = []
        walker = InfiniteDomainFinderWalker()
        walker.walk(root_impl)
        
        if walker.has_found_clock_data():
            get_clock_data(root_impl, [], clock_data)
 
        for clock_el in clock_data:
            for _, expr, _ in associations.values():
                if clock_el in expr:
                    continuous_assoc.append("Expression '%s' contains a continuous variable: %s" %(expr, clock_el))

        return continuous_assoc

                                             
pass # end of TfpgBehavValModel class
# ----------------------------------------------------------------------


class TfpgBehvalResult ():
    """ Represents the container for the TFPG Behavioral Validation result,
    having the success flag (a string) and possibly two ResultModel 
    with the traces. """

    def __init__(self, _success, counterexample_description, trace_system_fn, trace_tfpg_fn):
        """Constructor Method"""
        self.success = _success

        # something went wrong
        if _success == "error":
            self.result_system = None
            self.result_tfpg = None
            self.counterexample_description = None
            return

        # there's a counterexample
        if _success == "incomplete":
            assert os.path.isfile(trace_system_fn) and os.path.isfile(trace_tfpg_fn),\
                "For unknown reasons no trace files were generated in TFPG behavioral validation."
                
            self.result_system = ResultModel()
            self.result_tfpg = ResultModel()

            self.result_system.create_trace_from_file(trace_system_fn)
            self.result_tfpg.create_trace_from_file(trace_tfpg_fn)

            self.counterexample_description = counterexample_description
           
        # there's no counterexample
        else:
            self.result_system = None
            self.result_tfpg = None
            self.counterexample_description = None

        return

pass # end of TfpgBehvalResult class
# ----------------------------------------------------------------------


class TfpgEffectValModel (FameTfpgHandlerModel):
    """Model for handling TFPG Effectiveness Validation"""

    busy = 0
    effect_val_result = None
    effect_val_results_list = []
    effect_val_sat_bound = options.NUSMV_BMC_LENGTH
    effect_val_use_sat = options.NUSMV_USE_BMC
    last_failure_mode_name = '[None]'
    last_system_mode_name = '[None]'
    results_list_changed = observable.Signal()
        
    # Observable properties
    __observables__ = ("busy", "effect_val_result", "effect_val_sat_bound", "effect_val_use_sat", 
                        "last_system_mode_name", "last_failure_mode_name", "results_list_changed", 
                        "effect_val_results_list", )
    
    # constants for effectiveness analysis

    SYS_MODE_VAR_NAME = "system_mode"
    TIME_TICK = "time_tick"
    FROZEN_SYS_MODE = True
    DUMP_VALIDATION_PROPS = False
    
    def __init__(self, tfpg_model, slim_model):
        """Constructor Method"""

        # TODO: consider to remove slim_model as a parameter of the constructor.
        #       it is not needed for effectiveness analysis, and not foreseen by
        #       the interface specification documents.

        FameTfpgHandlerModel.__init__(self, slim_model, tfpg_model)
        return

        
    def generate_smv_file (self, tfpg_structural):

        # TODO: this is only temporary; see ticket #73 on trac for a discussion on how
        #       to obtain a scaling factor for the TFPG (not the same as sampling rate!!!).
        
        # The scaling_factor is the GCD of TFPG propagation times.
        scaling_factor = self.tfpg_model.gcd_tfpg_propagation_times(tfpg_structural)
            
        # compute smv filename (this could be done more elegantly)
        self.slim_model.smv_fname = None
        self.smv_file, self.ord_file, self.need_msat, generated, killed = \
            self.generate_smv_ord(True)
        if killed: return # got killed by event!

        # for effectiveness analysis, no mathsat is needed
        self.need_msat = False

        # the BDD ordering file computed above is valid only for the system model,
        # not the TFPG model
        self.ord_file = None
 
        # dump tfpg as smv

        from tfpg2smv.tfpg2smv import dump_tfpg_as_smv
        dump_tfpg_as_smv(tfpg_structural, self.smv_file,
            self.TIME_TICK, scaling_factor,
            self.FROZEN_SYS_MODE, self.DUMP_VALIDATION_PROPS)
            
        return    
       
    def stop_effect_val(self):
        """Stops the TFPG Effectiveness Validation"""
        self.stop_fame_tfpg()
        self.effect_val_result = TfpgDiagResult("stop", None, None)
        return 
        
    def validate_effectiveness(self, tfpg_structural):
        """Performs TFPG Behavioural Effectiveness and generates the results."""
       
        # generate SMV file for analysis
        self.generate_smv_file(tfpg_structural)

        # Observables
        observables = tfpg_structural.get_monitored_discrepancies()
        
        # System Modes
        system_modes_list = tfpg_structural.get_modes_list()
        
        # Failure Modes Targets

        failure_mode_list = []
        
        # the list of failure 
        # modes is composed of TFPG Structural FM nodes.
 
        fm_nodes = tfpg_structural.get_fm_nodes()
        for fm_node in fm_nodes:
            failure_mode_list.append([fm_node.get_name()]) 

        if failure_mode_list == []:
            self.effect_val_result = TfpgDiagResult('alarms_not_enabled', None, None)
            self.stop_fame_tfpg()
            return

        # main effectiveness validation loop
        
        self.busy += 1
        try:
            
            if failure_mode_list and system_modes_list:
                for failure_mode in failure_mode_list:
                    for system_mode in system_modes_list:
                       
                        self.start_fame_tfpg()

                        result = self.get_tool().check_tfpg_diagnosability(observables, failure_mode,
                            self.SYS_MODE_VAR_NAME, self.TIME_TICK, system_mode,
                            self.effect_val_use_sat, self.effect_val_sat_bound)
                              
                        self.last_failure_mode_name = failure_mode
                        self.last_system_mode_name = system_mode
  
                        if result['result'] == 'error': 
                            self.effect_val_result = TfpgDiagResult("error", None, None)
                            return
                                
                        elif result['result'] == 'diagnosable':
                            self.effect_val_result = TfpgDiagResult('diagnosable', None, None)
                                
                        elif result['result'] == 'not_diagnosable':
                            self.effect_val_result = TfpgDiagResult('not_diagnosable', result['critical_pair'][0], result['critical_pair'][1])
                                
                        else:
                            assert result['result'] == 'diagnosability_unknown'
                            self.effect_val_result = TfpgDiagResult('diagnosability_unknown', None, None)

                        # The self.effect_val_results_list is the list of all results with the couple of failure_mode - system_mode
                        # It is used in the function on_results_selection_changed of ctrl.
                        self.effect_val_results_list.append([self.effect_val_result, self.last_failure_mode_name, self.last_system_mode_name])
                        self.stop_fame_tfpg()

            else: self.effect_val_result = TfpgDiagResult("error", None, None)
                            
        finally: self.busy -= 1
     
        return 

pass # end of TfpgEffectValModel class
# ----------------------------------------------------------------------


class TfpgDiagResult ():
    """Represents the container for the Diagnosability Check result,
        having the success flag (a string) and possibly two ResultModel 
        with the traces.
    """

    def __init__(self, _success, trace_A_fn, trace_B_fn):
        """Constructor Method"""
        self.success = _success

        # something went wrong
        if _success == "error":
            self.result_A = None
            self.result_B = None
            return

        # there's a counterexample
        if _success == "not_diagnosable":
            assert os.path.isfile(trace_A_fn) and os.path.isfile(trace_B_fn),\
                "For unknown reasons DIAG has not generated output files."
                
            self.result_A = ResultModel()
            self.result_B = ResultModel()

            self.result_A.create_trace_from_file(trace_A_fn)
            self.result_B.create_trace_from_file(trace_B_fn)
           
        # there's no counterexample
        else:
            self.result_A = None
            self.result_B = None

        return

pass # end of TfpgDiagResult class
# ----------------------------------------------------------------------


class TfpgErrorsModel (Model): 
    """Model for showing TFPG Errors"""

    busy = 0
        
    # Observable properties
    __observables__ = ("busy",)
    
    
    def __init__(self, tfpg_model, slim_model, mission_model):
        """Constructor Method"""
        Model.__init__(self)
        
        self.tfpg_model = tfpg_model
        self.slim_model = slim_model
        self.mission_model = mission_model
        
        self.slim_model.register_observer(self)
        return
        
pass # end of TfpgErrorsModel class
# ----------------------------------------------------------------------

