# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by GMV Aerospace and Defence  
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


import os.path
import subprocess
import tempfile
import threading

from gtkmvc import ModelMT
from utils.smv2sigref_process import Smv2SigrefProcess
from utils.nusmv_process import NuSMVProcess
from utils.fsap_process import FsapProcess
from utils.procmon import ProcMonitor
from utils.main_extension_converter import MainExtensionConverter
import utils.log
import utils.misc
from utils.fame_tfpg_process import FameTfpgProcess
import slim_compiler.src.smv.api as smv_api
import options
import invokers


class ToolHandlerModel (ModelMT):
    """This is a model to be used as base class for tool handlers
    like nusmv, fsap and mrmc. A tool takes some time to produce a
    result, and this class provides an observable properties that
    declares the level of business the tool has each moment. When
    the tool starts an activity, increment the busy level. When the
    tool completes an activity, decrement the busy level.

    The tool handler features also an internal monitor, which can
    measure the total amount of memory and cputime used by the
    underlaying process running the tool."""

    busy = 0
    tool = None
    __observables__ = ("busy", "tool")

    def __init__(self):
        ModelMT.__init__(self)

        #public
        self.tool = None

        #private
        self._mon = None
        self.__memout__ = int(options.MEMOUT)
        self.__timeout__ = int(options.TIMEOUT)
        assert self.__timeout__ is not None

        self.observe_model(self)  # to observe 'tool'
        return

    def set_tool(self, tool):
        """tool is an instance of InteractiveProcess"""
        self.tool = tool
        return

    def get_tool(self):
        """Returns the currently running instance of
        InteractiveProcess that has been previously set"""
        return self.tool

    def set_mem_limit(self, limit):
        """Sets memory limit for the running process, in MB.
        If -1, it is set a no limit"""
        self.__memout__ = int(limit) if int(limit) >= 0 else -1
        if self._mon:
            self._mon.set_memout(self.__memout__ << 10
                                 if self.__memout__ >= 0
                                 else -1)

        return

    def set_time_limit(self, limit):
        """Sets CPU time limit for the running process, in seconds.
        If -1, it is set a no limit"""
        self.__timeout__ = int(limit) if int(limit) >= 0 else -1
        if self._mon:
            self._mon.set_timeout(self.__timeout__)
        assert self.__timeout__ is not None
        return

    def stop_tool(self):
        """Shutsdown the tool, whatever it is doing now"""
        if self.tool is not None:
            self.tool.kill()
            self.tool = None
            pass
        return

    def get_status(self):
        """Returns a tuple containing:
           status : string
           total running cputime in seconds: integer
           partial running cputime in seconds: integer
             (this is considered for timeouts)
           total memory usage in KB: integer (this is considered for memouts)
           peak of memory usage in KB
           """
        if self._mon is None:
            return ("INACTIVE", 0, 0, 0, 0)

        return self._mon.get_info()

    def property_tool_value_change(self, model, old, new):
        """Observes its own tool property to start a monitor
        everytime tool changes"""
        if model is not self:
            return
        if old is not None:
            old.kill()

        assert self.__timeout__ is not None

        if self._mon is not None:
            self._mon.stop()
            pass

        if new is not None:
            self._mon = ProcMonitor(new,
                                    self.__memout__ << 10
                                    if self.__memout__ >= 0
                                    else -1,
                                    self.__timeout__)
            self._mon.start()
            pass

        return
# ----------------------------------------------------------------------


class MRMCHandlerModel(ToolHandlerModel):
    """This is a model that handles MRMC."""
    def ___init___(self):
        ToolHandlerModel.__init__(self)

    def stop_mrmc(self):
        """Shutsdown mrmc, whatever it is doing now"""
        self.stop_tool()
        return

    def start_mrmc(self, tra_file, lab_file, formula_string,
                   stepwise=False, granularity=10, args=[]):
        # checking of whether we should start up MRMC
        if self.tool is not None and self.tool.is_alive():
            return

        self.tool = invokers.mrmc.start_mrmc(tra_file, lab_file,
                                             formula_string, stepwise,
                                             granularity, args)

        self.tool.wait()
        return
# ----------------------------------------------------------------------


class IMCAHandlerModel(ToolHandlerModel):
    """This is a model that handles IMCA."""
    def ___init___(self):
        ToolHandlerModel.__init__(self)

    def stop_imca(self):
        """Shutsdown imca, whatever it is doing now"""
        self.stop_tool()
        return

    def start_imca(self, ma_filename, computation_type_, lower_time_bound,
                   upper_time_bound, error_bound=0.1, stepwise=False,
                   granularity=10, args=[]):
        # checking of whether we should start up IMCA
        if self.tool is not None and self.tool.is_alive():
            return

        self.tool = utils.imca.start_imca(ma_filename, computation_type_,
                                          lower_time_bound, upper_time_bound,
                                          error_bound, stepwise, granularity,
                                          args)

        self.tool.wait()
        return
# ----------------------------------------------------------------------


class Model2ImcHandlerModel(ToolHandlerModel):
    """This is a model that handles Model2Imc."""
    def ___init___(self):
        ToolHandlerModel.__init__(self)

    def stop_model2imc(self):
        """Shutsdown model2imc, whatever it is doing now"""
        self.stop_tool()
        return

    def start_model2imc(self, formula_numbers, output_dir, output_formats):
        """starts a new Model2Imc session"""
        # checking of whether we should start up Model2Imc
        if self.tool is not None and self.tool.is_alive():
            return

        self.tool = utils.model2imc_process.start_model2imc(formula_numbers,
                                                            output_dir,
                                                            output_formats)

        self.tool.wait()
        return
# ----------------------------------------------------------------------
class SlimSimHandlerModel(ToolHandlerModel):
    def ___init___(self):
        ToolHandlerModel.__init__(self)

    def stop_slimsim(self):
        """Shutsdown model2imc, whatever it is doing now"""
        self.stop_tool()
        return


    def start_slimsim(self, xml_filename, prop_filename, step_bound, error_bound, confidence, strategy):
        """starts a new Slimsim session"""
        # checking of whether we should start up Model2Imc
        if self.tool is not None and self.tool.is_alive(): return

        self.tool = invokers.slimsim.start_slimsim(xml_filename, prop_filename, step_bound, error_bound, confidence, strategy)

        self.tool.wait()

        return

# ----------------------------------------------------------------------

class SmvHandlerModel (ToolHandlerModel):
    """This is the base class of all tools which need to generate a
    smv file (along with a variable ordering)"""

    def __init__(self, slim_model,
                 timed_trans_fair_alt_mode=options.SMV_FAIR_TIMED):
        """@param bool timed_trans_fair_alt_mode: if True, the smv file will
        be generated with time transions in fair paths, and
        time-transition alternation with discrete transitions
        (default). When False, no constraints are generated for timed
        transitions, and timed variable is also made available.

        Use set_timed_trans_fair_alt_mode to later change this
        behavour on demand. IN particular, this is used when switching
        between checking of non-timed properties and timed-properties,
        which have different requisites of the smv file at
        engine-level.
        """

        ToolHandlerModel.__init__(self)

        self.slim_model = slim_model
        # this is used to interrupt the ordering file generation process
        self.__kill_ord_event = None
        self.__kill_ord_event_processed = threading.Event()

        self._timed_trans_fair_alt_mode = timed_trans_fair_alt_mode
        return

    def set_timed_trans_fair_alt_mode(self, mode=True):
        """Used to set the timed_trans_fair_alt_mode (see constructor for
        further info). Returns True iff the setting triggers
        regeneration of smv file and restart of tool.
        """

        if mode != self._timed_trans_fair_alt_mode:
            self._timed_trans_fair_alt_mode = mode
            self.stop_nusmv()  # to trigger smv regeneration
            self.start_nusmv()  # restart
            return True

        return False

    def generate_smv_ord(self, use_extended_model):

        """Starts the generation of smv and ordering file. Returns:
        (smv_fname, ord_fname, needs_msat, generated, killed)"""

        self.__kill_ord_event = threading.Event()

        smv_fname, ord_fname, need_msat, generated = \
            self.slim_model.generate_smv_file(
                use_extended_model,
                self.__kill_ord_event,
                generate_time_var=not self._timed_trans_fair_alt_mode,
                generate_time_fairness_alt=self._timed_trans_fair_alt_mode)

        killed = self.__kill_ord_event.isSet()
        self.__kill_ord_event = None
        self.__kill_ord_event_processed.set()

        if killed:
            utils.log.system.debug("SmvHandler got stopped "
                                   "during ordering construction")

        return (smv_fname, ord_fname, need_msat, generated, killed)

    def stop_tool(self):
        if self.__kill_ord_event:  # waiting for ordering to be generated
            self.__kill_ord_event_processed.clear()
            self.__kill_ord_event.set()
            self.__kill_ord_event_processed.wait()  # waits until killed
            pass

        ToolHandlerModel.stop_tool(self)
        return
# ----------------------------------------------------------------------


class NuSMVHandlerModel (SmvHandlerModel):
    """This is a model (to be derived) that owns and controls a
    nusmv process"""

    def __init__(self, slim_model,
                 timed_trans_fair_alt_mode=options.SMV_FAIR_TIMED):
        SmvHandlerModel.__init__(
            self, slim_model,
            timed_trans_fair_alt_mode=timed_trans_fair_alt_mode)

        self.__nusmv_needs_restart = True

        # observes the change signal from the slim model, to know
        # when nusmv must be restarted
        self.slim_model.register_observer(self)
        return

    def stop_nusmv(self):
        """Shutsdown nusmv, whatever it is doing now"""
        self.stop_tool()
        self.__nusmv_needs_restart = True
        return

    def start_nusmv(self, use_extended_model=False):
        """starts a new session of nusmv if not already
        going. Returns True if the tool was actually (re)started,
        False if the tool is currently running already."""

        smv_fname, ord_fname, need_msat, generated, killed = \
            self.generate_smv_ord(use_extended_model)
        if killed:
            return  # got killed by event!

        if self.__nusmv_needs_restart or generated:
            self.stop_nusmv()

        if self.tool is not None and self.tool.is_alive():
            return False

        # constructs a new instance of nusmv
        fd, logfname = tempfile.mkstemp(suffix='.nusmvout',
                                        prefix=os.path.splitext(smv_fname)[0],
                                        dir=".", text=True)
        utils.log.system.debug("Invoking NuSMV, with redirected to " +
                               logfname)

        self.tool = NuSMVProcess(smv_fname, ord_fname, need_msat,
                                 stdout=subprocess.PIPE,
                                 stderr=subprocess.STDOUT,
                                 dumpers=[os.fdopen(fd, "w")])

        self.__nusmv_needs_restart = False
        return True

    # ----------------------------------------------------------------------
    # Observers
    # ----------------------------------------------------------------------
    def property_changed_signal_emit(self, model, args):
        """This is an observer for slim model modification. It is
        used to trigger re-generation and re-compilation of the smv
        file"""
        if model is self.slim_model:
            self.__nusmv_needs_restart = True

        return
# ----------------------------------------------------------------------


class FsapHandlerModel (SmvHandlerModel):
    """This is a model (to be derived) that owns and controls a
    fsap process"""

    # constants used by the class
    DELAY_COUNTER_VAR = "fsap_delay_counter"

    def __init__(self, slim_model,
                 timed_trans_fair_alt_mode=options.SMV_FAIR_TIMED):
        SmvHandlerModel.__init__(
            self, slim_model,
            timed_trans_fair_alt_mode=timed_trans_fair_alt_mode)

        self.__fsap_needs_restart = True

        # these are filled in during the dumping of failures, observations
        # and alarms
        self.failures = -1
        self.observations = -1
        self.alarms = -1

        # this is the SMV file name used by the FSAP instance
        # also used to enable (see start_fsap) and disable (see stop_fsap) the
        # function extend_smv_with_delay_counter
        self.smv_fname = None

        # observes the change signal from the slim model, to know
        # when fsap must be restarted
        self.slim_model.register_observer(self)
        return

    def stop_fsap(self):
        """Shutsdown fsap, whatever it is doing now"""
        self.smv_fname = None
        self.stop_tool()
        self.__fsap_needs_restart = True
        return

    def start_fsap_obs(self):
        self.start_fsap("Observable")
        return

    def start_fsap_alm(self):
        self.start_fsap("Alarm")
        return

    def start_fsap(self, obs_alm=None):
        """starts a new session of fsap if not already going"""

        smv_fname, ord_fname, need_msat, generated, killed = \
            self.generate_smv_ord(True)
        if killed:
            return  # got killed by event!

        if self.__fsap_needs_restart or generated:
            self.stop_fsap()
        if self.tool is not None and self.tool.is_alive():
            return

        self.smv_fname = smv_fname

        if obs_alm is None:
            # dump failure observations alarms
            fail_obs_alm_fname = utils.misc.get_out_fname(
                options.BASE_FILE_NAME + ".fail_obs_alm")

            fo_file = open(fail_obs_alm_fname, "w")
            self.failures, self.observations, self.alarms = \
                smv_api.dump_failures_observations_alarms(
                    self.slim_model.get_extended_root_impl(), fo_file)

        elif obs_alm == "Observable":
            # dump failure observations
            fail_obs_alm_fname = utils.misc.get_out_fname(
                options.BASE_FILE_NAME + ".fail_obs")
            fo_file = open(fail_obs_alm_fname, "w")
            self.failures, self.observations = \
                smv_api.dump_failures_observations(
                    self.slim_model.get_extended_root_impl(), fo_file)
            self.alarms = -1

        elif obs_alm == "Alarm":
            # dump failure alarms
            fail_obs_alm_fname = utils.misc.get_out_fname(
                options.BASE_FILE_NAME + ".fail_alm")
            fo_file = open(fail_obs_alm_fname, "w")
            self.failures, self.alarms = \
                smv_api.dump_failures_alarms(
                    self.slim_model.get_extended_root_impl(), fo_file)
            self.observations = -1

        else:
            return

        fo_file.close()

        # constructs a new instance of fsap
        fd, logfname = tempfile.mkstemp(
            suffix='.fsapout',
            prefix=os.path.splitext(self.smv_fname)[0],
            dir=".", text=True)
        utils.log.system.debug("Invoking Fsap, with redirected to " + logfname)

        self.tool = FsapProcess(self.smv_fname, fail_obs_alm_fname,
                                need_msat=need_msat,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.STDOUT,
                                dumpers=[os.fdopen(fd, "w")])

        # set a prefix for temporary files.
        # this prefix is unusual enough not to interfere with normal files
        self.tool.set_tmp_filename_prefix(
            utils.misc.get_out_fname(options.BASE_FILE_NAME + "__temp_fsap__"))

        self.__fsap_needs_restart = False
        return

    def extend_smv_with_delay_counter(self, smv_prop, delay_counter_bound):
        assert delay_counter_bound > 0
        assert self.smv_fname is not None

        conv = MainExtensionConverter()

        delta_var = None

        if self.slim_model.timed_domain:
            # count time
            delta_var = conv.delta_var()
        else:
            # count steps
            delta_var = "1"

        delay_counter_bound_str = None
        delay_var_type = None

        if self.slim_model.infinite_domain:
            # delay counter bound is float
            delay_counter_bound_str = "%f" % delay_counter_bound
            delay_var_type = "real"
        else:
            # delay bound is int
            assert delay_counter_bound % 1 == 0
            delay_counter_bound_str = "%d" % delay_counter_bound
            delay_var_type = "-1..%d" % delay_counter_bound

        rnd_choice_var = "%s_rnd" % self.DELAY_COUNTER_VAR

        main_extension = \
            "  VAR %s : %s;\n" % (self.DELAY_COUNTER_VAR, delay_var_type) + \
            "  VAR %s : boolean;\n" % rnd_choice_var + \
            "\n" + \
            "  ASSIGN init(%s) :=\n" % (self.DELAY_COUNTER_VAR) + \
            "    case\n" + \
            "      %s & %s : 0;\n" % (smv_prop, rnd_choice_var) + \
            "      TRUE : -1;\n" + \
            "    esac;\n" + \
            "\n"

        main_extension += \
            "  ASSIGN next(%s) :=\n" % (self.DELAY_COUNTER_VAR) + \
            "    case\n" + \
            "      %s = -1 & !next(%s) : -1;\n" % (self.DELAY_COUNTER_VAR, smv_prop) + \
            "      %s = -1 &  next(%s) & !%s : -1;\n" % (self.DELAY_COUNTER_VAR, smv_prop, rnd_choice_var) + \
            "      %s = -1 &  next(%s) &  %s :  0;\n" % (self.DELAY_COUNTER_VAR, smv_prop, rnd_choice_var) + \
            "      %s + %s < %s : %s + %s;\n" % (self.DELAY_COUNTER_VAR, delta_var, delay_counter_bound_str, self.DELAY_COUNTER_VAR, delta_var) + \
            "      TRUE : %s;\n" % (delay_counter_bound_str) + \
            "    esac;\n" + \
            "\n"

        conv.set_main_extension(main_extension)

        out = open(self.smv_fname, 'w')
        conv.dump(self.slim_model.get_extended_root_impl(), out)
        out.close()

        return

    # ----------------------------------------------------------------------
    # Observers
    # ----------------------------------------------------------------------
    def property_changed_signal_emit(self, model, args):
        """This is an observer for slim model modification. It is
        used to trigger re-generation and re-compilation of the smv
        file"""
        self.__fsap_needs_restart = True
        return
# ----------------------------------------------------------------------


class Smv2SigrefHandlerModel (ToolHandlerModel):
    """This is a model that handles SMV2Sigref process"""
    def ___init___(self):
        ToolHandlerModel.__init__(self)
        return

    def stop_smv_to_sigref(self):
        """Shutsdown the process, whatever it is doing now"""
        self.stop_tool()
        return

    def start_smv_to_sigref(self, basename):
        """starts a new Smv2SigRef session"""
        # checking of whether we should start up SigRef
        if self.tool is not None and self.tool.is_alive():
            return

        # constructs a new instance of nusmv
        fd, logfname = tempfile.mkstemp(suffix='.smv2sigrefout',
                                        prefix=basename,
                                        dir=".", text=True)
        utils.log.system.debug("Invoking SMV2Sigref, with redirected to " +
                               logfname)

        # loading up Smv2SigRef
        self.tool = Smv2SigrefProcess(stdout=subprocess.PIPE,
                                      stderr=subprocess.STDOUT,
                                      dumpers=[os.fdopen(fd, "w")])
        return

    def convert_smv_to_sigref(self, smv_fname, prob_fname,
                              ord_fname=None, label_fname=None):
        """Converts the given smv file to sigref. Returns a pair
        (output_xml_sigref, var_ordering_file_name)"""
        sigref, ord = self.tool.build_sigref_inputs(smv_fname,
                                                    prob_fname,
                                                    ord_fname,
                                                    label_fname)
        return (sigref, ord)
# ----------------------------------------------------------------------


class FameTfpgHandlerModel (SmvHandlerModel):
    """This is a model (to be derived) that owns and controls a
    fame_tfpg_process"""

    def __init__(self, slim_model, tfpg_model,
                 timed_trans_fair_alt_mode=options.SMV_FAIR_TIMED):
        SmvHandlerModel.__init__(
            self, slim_model,
            timed_trans_fair_alt_mode=timed_trans_fair_alt_mode)

        self.__fame_tfpg_needs_restart = True
        self.slim_model = slim_model
        self.tfpg_model = tfpg_model

        assert tfpg_model.tfpg_slim_assoc_model is not None

        # TODO: [BB] check whether slim model and tfpg change event
        # handlers are in place

        self.slim_model.register_observer(self)
        self.tfpg_model.register_observer(self)

        self.smv_file = None
        self.ord_file = None
        self.need_msat = False
        self.args = []

        return

    def stop_fame_tfpg(self):
        """Shuts down the NuSMV process, whatever it is doing at the moment."""
        self.stop_tool()
        self.__fame_tfpg_needs_restart = True
        return

    def start_fame_tfpg(self, slim_root=None):
        """Starts a new session of NuSMV if one hasn't been started yet."""

        # at this point, the SMV file must have been created by the subclass
        assert self.smv_file is not None

        if self.__fame_tfpg_needs_restart:
            self.stop_fame_tfpg()
        if self.tool is not None and self.tool.is_alive():
            return

        # constructs a new instance of tfpg_fame
        fd, logfname = tempfile.mkstemp(
            suffix='.fametfpgout',
            prefix=os.path.splitext(self.smv_file)[0],
            dir=".", text=True)

        utils.log.system.debug("Invoking FameTfpg, with redirected to " +
                               logfname)

        self.tool = FameTfpgProcess(self.smv_file, self.ord_file,
                                    self.need_msat, self.args,
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.STDOUT,
                                    dumpers=[os.fdopen(fd, "w")])

        # set a prefix for temporary files.
        # this prefix is unusual enough not to interfere with normal files
        self.tool.set_tmp_filename_prefix(utils.misc.get_out_fname(
            options.BASE_FILE_NAME + "__temp_fametfpg__"))

        self.__fame_tfpg_needs_restart = False
        return

    # ----------------------------------------------------------------------
    # Observers
    # ----------------------------------------------------------------------

    def property_changed_signal_emit(self, model, args):
        """This is an observer for slim model modification. It is
        used to trigger re-generation and re-compilation of the smv
        file."""
        self.__fame_tfpg_needs_restart = True
        return
# ----------------------------------------------------------------------
