# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


from gtkmvc import Model, TreeStoreModel
from gtkmvc.observable import Signal
# validate imports
from slim_compiler.src.testexpr import parseExpression
from slim_compiler.src.smv.converter import SmvConverter
from slim_compiler.src.errorManagement.exceptions import InvalidExpressionException
from slim_compiler.src.parse_property import check_property_tree
# end
from src.models.trace_filter import TraceFilters
import src.utils.nusmv

from slim_compiler.src.smv import api 

import xml.dom
import gtk
import gobject

# ----------------------------------------------------------------------
class TraceStripeModel (Model):
    """A single labelled list of (implicitly timed) values."""

    def __init__(self, name, info=None):
        """info is an optional parameter to attach any needed info
        to the stripe (for example, if it is an input
        var). Interpretation is free and it is up the caller to
        decide about it."""
        Model.__init__(self)

        self.name = name
        self.values = []
        self._info = info
        self._is_error = False #say if this is an error stripe
        self._is_boolean = False
        self._is_delta = False
        return


    def to_xml(self, root_doc, parent_node):
        """@return the stripe node in the case sub stripes must be
        appended to it"""
        
        stripe_node = root_doc.createElement("stripe")

        # name
        name_node = root_doc.createElement("name")
        name_node.appendChild(root_doc.createTextNode(self.name))
        stripe_node.appendChild(name_node)

        # type
        type_node = root_doc.createElement("type")
        if self._is_boolean: type_txt = "boolean"
        else: type_txt = "enumerative"
        type_node.appendChild(root_doc.createTextNode(type_txt))
        stripe_node.appendChild(type_node)

        # steps
        for idx in range(len(self.values)):
            step_node = root_doc.createElement("step")

            sval = self.values[idx]
            sname = "%d" % idx

            # step's name
            sname_node = root_doc.createElement("name")
            sname_node.appendChild(root_doc.createTextNode(sname))
            step_node.appendChild(sname_node)

            # step's value
            sval_node = root_doc.createElement("value")
            sval_node.appendChild(root_doc.createTextNode(sval))
            step_node.appendChild(sval_node)

            stripe_node.appendChild(step_node)
            pass
        
        # finally appends the stripe to the parent node. We return the stripe
        # node in the case this stripe contains sub stripes that have to be
        # appended to it
        parent_node.appendChild(stripe_node)     
        return stripe_node


    def from_xml(self, root_doc, node):
        """
        fills the stripe reading from an xml representation of the trace stripe.
        """
        m = { "name" : None,
              "type" : None}

        steps = {}

        for child in node.childNodes:
            if child.nodeType == xml.dom.Node.ELEMENT_NODE and m.has_key(child.tagName):
                m[child.tagName] = ""
                for text_child in child.childNodes:
                    if text_child.nodeType == xml.dom.Node.TEXT_NODE:
                        m[child.tagName] += text_child.data
                        pass
                    pass
                pass
            
            elif child.nodeType == xml.dom.Node.ELEMENT_NODE and child.tagName=="step":
                # retrieves a single (step_name, value) pair
                m2 = { "name" : None,
                       "value" : None}
                
                for child2 in child.childNodes:
                    if child2.nodeType == xml.dom.Node.ELEMENT_NODE and m2.has_key(child2.tagName):
                        m2[child2.tagName] = ""
                        for text_child in child2.childNodes:
                            if text_child.nodeType == xml.dom.Node.TEXT_NODE:
                                m2[child2.tagName] += text_child.data
                                pass
                            pass
                        pass
                    pass
                assert m2["name"].strip()
                steps[m2["name"].strip()]=m2["value"].strip()
                pass
            pass

        assert m["name"].strip()

        # sets the internals:
        self.name  = m["name"].strip()
        self._is_boolean = (m["type"].strip().lower().rfind("boolean") >= 0)

        # reorders the steps
        step_nums = steps.keys()
        step_nums.sort()
        for step in step_nums: self.add_value(steps[step])    

        return
    
    def set_boolean(self, value=True): self._is_boolean = value

    def is_boolean(self): return self._is_boolean        

    def set_error(self, value): self._is_error = value
    def get_error(self): return self._is_error
    def set_delta(self, value=False): self._is_delta = value
    def is_delta(self): return self._is_delta

    def get_info(self, error=False):
        """
        @param error: if True, error is a valid type, else it isn't
        """
        if error and self.get_error():
            return "error"
        else:
            return self._info

    def get_name(self):
        """@return the stripe's name"""
        return self.name

    def set_name(self, name):
        """Changes the stripe's name"""
        self.name = name
        return

    def get_steps(self):
        """@return the number of currently contained steps"""
        return len(self.values)

    def add_value(self, value):
        """Adds a new step, containing the given value"""
        self.values.append(value)
        return

    def remove_step(self, step):
        """Removes the value at the given step. Reduces the number of steps."""
        del self.values[step]
        return
    
    def set_value_at_step(self, step, value):
        """Sets the value at the given step.
        @param step starts from 0"""
        self.values[step] = value
        return

    def get_value_at_step(self, step):
        """@return the value at given step (initial step is 0)"""
        # input and combinatorial have no last state
        if self.get_steps() == step and self.get_info() != "state":
            return ""
        try:
            return self.values[step]
        except IndexError:
            raise IndexError("step "+str(step)+" out of range in stripe "+str(self.get_name()))

    pass # end of class

#extension of stripe model
class ExtensibleTraceStripeModel(TraceStripeModel):
    """
    This is a TraceStripeModel that can be extended following some rule.
    (like the guided simulation does)
    The extend_stripe method is not implemented so the two ways to use it is to subclass it,
    or modify it.
    """
    
    def __init__(self, name, info=None):
        TraceStripeModel.__init__(self, name, info)
        self.DEFAULT_TRANSITION = "0" #DISCRETE TRANSITION
        self._base_number_of_steps = 0
        self.__first_add = True
        self.new_value = None
        return

    #the extension of the stripe is not used, another stripe (already extended) is loaded
    def extend_stripe(self, steps):
        """
        simulate a new value starting from the last value,
        and add it to the list of values.
        @param steps: the number of steps to simulate and add.
        """
        assert steps > 0, "Invalid steps! "+str(steps)
        self.append_empty_step()
        self.new_value = None
        self._base_number_of_steps = 0
        # something
        raise Exception("To be implemented")
        
        return False

    def append_empty_step(self):
        """
        append an empty step to the stripe
        """
        if self.__first_add:
            self.__first_add = False
            self._base_number_of_steps = self.get_steps()
            pass
        new = "" if not self.is_delta() else self.DEFAULT_TRANSITION #default a non continuous transition
        self.values.append( new )
        self.new_value = new
        return

    def base_steps(self):
        """
        return the base number of steps in the model.
        (minus the values added with force_stripe_value, and extend_stripe)
        """
        return self._base_number_of_steps

    
    def get_forced_value(self):
        """
        @return the value forced into the stripe, or none if no value is forced into.
        """
        return self.new_value

    def force_stripe_value(self, pos, new):
        """
        force a position in the stripe to change value.
        @param pos: the position in the stripe 
        (can be len(pos), if you would to force a 
        new value. when extend_stripe is called, the value forced
        remains unmuted, and the others are simulated)
        @param new: the new value (of the right type)
        """

        maxlen = self.get_steps()
        assert pos <= maxlen and pos >= 0, "Invalid pos! "+str(pos)
        
        if pos == maxlen:
            self.append_empty_step()
            pass

        if self.is_delta():
            try: #try an int conversion
                to_add = int(new)
            except ValueError:
                to_add = self.DEFAULT_TRANSITION #default non continuous transition
        # elif self.is_boolean():
        #     s_new = str(new).lower() #because nusmv takes true or false, not TRUE or FALSE
        #     if s_new == 'false' or s_new == 'true' or not s_new:
        #         to_add = s_new 
        #     else:
        #         to_add = str(bool(s_new)).lower()
                
        else:
            to_add = str(new)
            pass

        self.values[pos] = to_add
        self.new_value =  to_add
        return
    pass
#--------------------------------------------------end class

class TraceModel (TreeStoreModel):
    """The model of a single trace, intended as a set of stripes of
    signals or other.
    @see TraceStripeModel
    """
    ERROR_STRIPE = "_errorSubcomponent"
    ERROR_OLD = "mode"
    ERROR_REPLACE = "error"
    # ----------------------------------------
    #  Observable properties
    # ----------------------------------------
    name = None
    description_changed = Signal() # something changed in the description
    __observables__ = ("name", "description_changed")

    def __init__(self, name=None, dependencies=None, formula=None, 
                 stripe_container=TraceStripeModel):
        """@param dependencies is the list of the property names
        which self is built on.
        @param name if None, a name will be provided by the
        TracesModel this trace will be added to.
        @param stripe_container: is a class that is able to handle a single stripe.
        It must be derived from TraceStripeModel.
        """
        # set of TraceStripeModel's instances, and the corresponding name in SLIM language
        TreeStoreModel.__init__(self, gobject.TYPE_PYOBJECT, gobject.TYPE_STRING) 

        if name: self.name = name
        self.stripe_container = stripe_container
        self.history = [] # keeps track of what happend to this trace
        self.notes = ""
        self.loops = []
        self.max_loop = 0
        if dependencies is None: dependencies = []
        self.dependencies = dependencies
        self.__empty = True #this is only a variable to store some internal results in is_empty function

        self.creation_info = ""
        self.formula = formula        
        self.loops_sep = ", "
        #delta row
        #an iterator that points to the stripe that contains the information about the transitions
        #foreach step i if delta[i] > 0 --> step i is a continuous transition
        #otherwise it is a discrete transition
        self._delta = None 
        return
    
    def rename_error_stripes(self):
        """
        The error stripes will be renamed in this way:
        all the strings dot-separated retrieved from the slim_name, 
        that are equal to the property ERROR_STRIPE (by default equal to "_errorSubcomponent") 
        will be deleted, 
        and if the last string dot-separated is equal to the value in ERROR_OLD (by default equal to "mode")
        it will be replaced by the string in ERROR_REPLACE (by default equal to "error")
        """
        i = self.get_iter_first()
        while i:
            self.__rename(i)
            i = self.iter_next(i)
            pass
        return
        
    def __rename(self, i):
        # i is the iterator to the line to rename
        slim_name = self.get_value(i, 1)
        stripe = self.get_value(i, 0)
        # rename only valid names
        if not slim_name:
            return
        # change some names
        names = slim_name.split(".")
        new_names = [x for x in names if x != self.ERROR_STRIPE]
        if len(new_names) < len(names):
            if new_names[-1] == self.ERROR_OLD: new_names[-1] = self.ERROR_REPLACE
            stripe.set_error(True)
            pass

        new_slim_name = ".".join(new_names)
    
        self.set_value(i, 1, new_slim_name)
        return
        
    def add_event_to_history(self, text):
        self.history.append(text)
        self.description_changed.emit()
        return

    def __sort_asc_desc(self, reverse):
        if reverse:              #1 is the col on which we will sort (the column of the Name)
            self.set_sort_column_id( 1, gtk.SORT_DESCENDING )
        else:
            self.set_sort_column_id( 1, gtk.SORT_ASCENDING )
            pass

        return
    
    #implement a sorting over the stripes
    def sort(self, cmp_function=None, reverse=None):
        """
        if reverse is None, cmp_function will be used,
        otherwise, 
        if reverse is true, the order is descending
        if reverse is false, the order is ascending
        """
        #function used to compare the two stripes
        def comparefunction(treemodel, iter1, iter2, user_data):
            s1 = treemodel.get_value(iter1, 0) #stripe1
            ns1 = treemodel.get_value(iter1, 1) #stripe1 name
            
            s2 = treemodel.get_value(iter2, 0) #stripe2
            ns2 = treemodel.get_value(iter2, 1) #stripe2 name
            
            return cmp_function( (ns1,s1), (ns2,s2) )
              
        if reverse != None:
            self.__sort_asc_desc(reverse)
            return

        if cmp_function:
            #col, func, user_data # 0 is the col on wich we sort if a compare function is passed (the stripe)
            self.set_default_sort_func( comparefunction, None)
            self.set_sort_column_id( -1, 0 ) # I have to set the column id to -1 for the default sorting.. -_-
            return

        return

    def set_dependencies(self, deps):
        if deps != self.dependencies:
            self.dependencies = deps
            self.description_changed.emit()
            pass
        return

    def get_stripe_container(self):
        """
        get the class instance that handles a stripe
        """
        return self.stripe_container

    def get_dependencies(self): return self.dependencies

    def set_formula(self, formula): self.formula = formula
    def get_formula(self): return self.formula

    #an iterator that points to the stripe that contains the information about the transitions
    #foreach step i if delta[i] > 0 --> step i is a continuous transition
    #otherwise it is a discrete transition
    # i is the delta path
    def set_delta(self, i): 
        self._delta = i
        self.get_value(i, 0).set_delta(True) #say to the stripe "you are the delta stripe!"
        return
    
    def get_delta_iter(self): return self._delta
    def get_delta_path(self): return self.get_path(self._delta)

    def get_creation_info(self): return self.creation_info
    def set_creation_info(self, info): self.creation_info = info

    def update_dependency(self, old_name, new_name):
        """Searches old_name in the dependencies, and if found
        substitutes it with new_name. @return True if substitution
        occurs"""
        try: idx = self.dependencies.index(old_name)
        except ValueError: return False

        self.dependencies[idx] = new_name
        self.description_changed.emit()
        return True
    

    def to_xml(self, root_doc, parent_node):
        """
        @return an xml representation of the trace
        """
        top_node = root_doc.createElement("trace")

        # name
        name_node = root_doc.createElement("name")
        name_node.appendChild(root_doc.createTextNode(self.name))
        top_node.appendChild(name_node)

        # dependencies
        deps_node = root_doc.createElement("dependencies")
        
        for dep in self.dependencies:
            dep_node = root_doc.createElement("dependency")
            dep_node.appendChild(root_doc.createTextNode(dep))
            deps_node.appendChild(dep_node)
            pass
        top_node.appendChild(deps_node)

        # loops
        loops_node = root_doc.createElement("loops")
        loops_str = self.loops_sep.join(map(str, self.loops))
        loops_node.appendChild(root_doc.createTextNode(loops_str))
        top_node.appendChild(loops_node)

        # stripes:
        iter = self.get_iter_first() 
        self.to_xml_stripe(iter, root_doc, top_node)
                               
        # notes:
        notes_node = root_doc.createElement("notes")
        notes_node.appendChild(root_doc.createTextNode(self.notes))
        top_node.appendChild(notes_node)

        # creation_info:
        notes_node = root_doc.createElement("creation_info")
        notes_node.appendChild(root_doc.createTextNode(self.creation_info))
        top_node.appendChild(notes_node)       

        # originating property
        if self.formula: txt = self.formula
        else: txt = ""
        notes_node = root_doc.createElement("formula")
        notes_node.appendChild(root_doc.createTextNode(txt))
        top_node.appendChild(notes_node)

        # history
        notes_node = root_doc.createElement("history")
        notes_node.appendChild(root_doc.createTextNode("\\".join(self.history)))
        top_node.appendChild(notes_node)

        parent_node.appendChild(top_node)
        return top_node


    def to_xml_stripe(self, iter, root_doc, parent_node):
        """Recursive service for stripes"""
        while iter:
            stripe = self.get_value(iter, 0)
            stripe_node = stripe.to_xml(root_doc, parent_node)

            # recursive call
            iter_children = self.iter_children(iter)
            while iter_children:
                # TODO this method does not exist! it is useful?
                self.to_xml_stripe_aux(iter_children, root_doc, stripe_node)
                iter_children = self.iter_next(iter_children)
                pass

            iter = self.iter_next(iter)
            pass
        return
    
    def to_xml_stripe_aux(iter_children, root_doc, stripe_node):
        raise Exception("Not implemented")
    
    def from_xml(self, root_doc, node):
        """
        read the trace from an xml representation
        """
        
        m = { "name" : None,
              "loops" : None,
              "notes" : None,
              "creation_info" : None,
              "formula" : None,
              "history" : None }
        
        for child in node.childNodes:
            if child.nodeType == xml.dom.Node.ELEMENT_NODE and m.has_key(child.tagName):
                m[child.tagName] = ""
                for text_child in child.childNodes:
                    if text_child.nodeType == xml.dom.Node.TEXT_NODE:
                        m[child.tagName] += text_child.data
                        pass
                    pass
                pass

            elif child.nodeType == xml.dom.Node.ELEMENT_NODE and child.tagName=="stripe":
                self.from_xml_stripe(root_doc, child, None)
                pass
            pass

        assert m["name"].strip()
        self.name = m["name"].strip()
        self.notes = m["notes"].strip()
        if m["creation_info"]: self.creation_info = m["creation_info"].strip()
        if m["formula"]: self.formula = m["formula"].strip()
        
        if m["history"]: self.history = m["history"].strip().split("\\")

        ls = m["loops"].strip() 
        if ls: self.loops = ls.split(self.loops_sep)

        # collects dependencies:
        deps = node.getElementsByTagName("dependency")
        for dep in deps:
            dep_str = ""
            for text_child in dep.childNodes:
                if text_child.nodeType == xml.dom.Node.TEXT_NODE: dep_str += text_child.data
                pass
            self.dependencies.append(dep_str.strip())
            pass
        
        return

    def from_xml_stripe(self, root_doc, node, iter):
        """Recursive service for nested stripes"""
        stripe = self.stripe_container("")
        stripe.from_xml(root_doc, node)
        slim_name = api.convert_smv_id_to_slim(stripe.get_name())
        
        iter2 = self.append(iter)
        self.set_value(iter2, 0, stripe) # inserts the stripe
        self.set_value(iter2, 1, slim_name) # inserts the corresponding slim name

        for child in node.childNodes:
            if child.nodeType == xml.dom.Node.ELEMENT_NODE and child.tagName=="stripe":
                self.from_xml_stripe(root_doc, child, iter2)
                pass
            pass
        return


    def add_loop_at_step(self, step):
        """Adds a loop from the last state to the state at given step"""
        if step not in self.loops:
            self.loops.append(step)
            self.description_changed.emit()
            pass
        return

    def is_step_in_loop(self, step):
        """@return True if given step is within a loop"""
        for loop in self.loops:
            if step >= loop: return True
            pass
        return False

    def is_step_a_loop(self, step):
        """@return True if the given step is a loop step"""
        return step in self.loops
    
    def foreach_func(self, model, path, iter):
        self.max_loop = max(model.get_value(iter, 0).get_steps(), self.max_loop)
        return
    
    def is_empty(self, col):
        """
        say if this index number (starting from 0) is empty or not, foreach stripe in the model
        at the column indexed by the param col
        """
        #helper TODO there is a bug in this function, __empty is always false..
        def foreach_func(model, path, iter, user_data):
            #search for a non-empty stripe at this step
            col = user_data
            stripe = model.get_value(iter, 0)
            if stripe.is_delta():
                return False #skip delta stripes

            try:
                val = stripe.get_value_at_step(col)
            except IndexError:
                val = False
                pass

            self.__empty = self.__empty and not val    
            return
        #end helper
        self.__empty = True
        self.foreach( foreach_func, col )
            
        if self.__empty:
            self.__empty = True #reset to true 
            return True #it is empty
        else:
            return False #it is not empty
    
    def get_max_steps(self):
        """Returns the maximum steps number in the trace. Traverses
        the entire model"""
        self.max_loop = 0
        self.foreach(self.foreach_func)
        return self.max_loop        

    pass # end of class


#extension of the trace model
NULL = -1        
class ExtensibleTraceModel(TraceModel):
    """
    This class is derived from a TraceModel, 
    and extends its feature, in order to extend the trace (by calling extend_trace)
    and check whether or not a certain value is valid or not for a certain stripe
    (accordingly with the slim_model, if there is one).
    """
    def __init__(self, name=None, dependencies=None, formula=None, 
                 stripe_container=ExtensibleTraceStripeModel,
                 slim_model=None):

        #assert slim_model, "I need the slim model in order to check the values inserted"        
        TraceModel.__init__(self,name, dependencies, formula, stripe_container)
        self.slim_model = slim_model
        self.converter = SmvConverter() #useful for validate new values

        self._has_empty_col = False
        return
        
    def set_slim_model(self, model):
        """
        set the slim model. This enables the validation of the input.
        """
        self.slim_model = model
        return
    
    def get_slim_model(self):
        return self.slim_model

    def set_value(self, iter, n, obj): #override set_value
        #in pos 0, there is the stripe, in pos 1 there is the slim_name of the stripe
        if n == 0 and obj.__class__ != ExtensibleTraceStripeModel:
            raise TypeError("An extensible trace model can't use a normal trace stripe model")

        super(ExtensibleTraceModel,self).set_value(iter, n, obj)
        return
    #helper in order to iter on cols, action is a string with the function to call
    #and params are the params that you would pass to the function
    def __iter_on_stripes(self, action, value=0, params=NULL):
        #@param value -> the value of the row in the treeStore to use 
        i = self.get_iter_first()
        res_ls = []
            
        max_step = self.get_max_steps()
        while i:
            stripe = self.get_value(i, value)
        
            if not callable(action) and not hasattr( stripe, action ):
                raise Exception("Invalid action specified to the stripe! "+str(stripe.get_name()))
    
            if callable(action):
                #takes the trace, the stripe, and the maximum number of steps in the trace
                res_ls.append( action(self, stripe, max_step) )
            else:
                method = getattr( stripe, action )
                if params == NULL:
                    res_ls.append(method())
                else:
                    res_ls.append(method(params))
                    pass
                pass
                
            i = self.iter_next(i)        
                
        return res_ls

    def validate_input(self, new, slim_stripe_name):
        """ say if the value is valid for the stripe 
        @return a tuple with (valid_input,error)
        where valid_input is True, if the input is valid,
        and false if the input is not valid, and the error
        is a string with the error encountered."""
        # convert each costraint in nusmv
        assert type(slim_stripe_name) is str, str(type(slim_stripe_name))

        if not self.slim_model:
            return (True,"") # no possible validations, everything is fine

        if str(new) == "": # base case
            return (True,"")

        expr = slim_stripe_name+'='+str(new)
        try:
            ast = parseExpression(expr)
        except InvalidExpressionException, e:
            return (False,str(e))
        
        try:
            check_property_tree( self.slim_model.compile_info, ast, self.slim_model.get_extended_root_impl() )
        except InvalidExpressionException, e:
            return (False,str(e))
        
        try:
            nusmv_c = self.converter.expression( self.slim_model.get_extended_root_impl(), ast )
        except AssertionError, e:
            return (False,str(e))
        
        return (True,"") if nusmv_c else (False,"the converted expression is empty")


    def get_forced_values(self):
        """@returns a dict with the keys the slim names,
        and the values the foced values into the stripes in slim format"""
        i = self.get_iter_first()
        d = {}
        while i:
            slim_name = self.get_value(i, 1)
            stripe = self.get_value(i, 0)
            if not slim_name:
                i = self.iter_next(i)
                continue
            
            d[slim_name] = stripe.get_forced_value()

            i = self.iter_next(i)
            pass

        return d

    def append_col(self):
        """
        this function append an empty column to the trace model.
        it can append more than one empty step, until the number of steps
        in the stripe is equal to the number of steps in the trace.
        """
        if self._has_empty_col:
            return
        def append_steps(trace, stripe, max_step):
            diff = max_step - stripe.get_steps() + 1 # I have to add 1 in each case
            
            while diff:
                stripe.append_empty_step()
                diff -= 1
                pass
            return
        self.__iter_on_stripes( append_steps ) 
        self._has_empty_col = True
        return

    def extend_trace(self, steps):
        """
        This function compute n new steps, for this trace,
        and add it to the ExtensibleTraceModel.
        @param steps: how many steps do you want to add 
        """
        self.__iter_on_stripes( "extend_stripe", steps ) 
        self._has_empty_col = False
        return
        
    def base_steps(self):
        """ @returns the last step automatically generated, withouth guided costraints. """
        return max( self.__iter_on_stripes( "base_steps" ) )
    pass
#---------------------------------------------------------------------end of the class

class TraceViewModel(Model):
    """
    This class handles the filtering over a TraceModel.
    With this class can be handled sets of filters and TraceModels, but only one 
    is the active filter (the filter that will be applied) and the active model
    (the model on which the filter will be applied).
    @see TraceModel, TraceFilters
    """
    _active_trace = None
    _active_filter = None
    _active_trace_filtered = None
    __observables__ = ("_active_trace","_active_filter","_active_trace_filtered")

    def __init__(self):
        Model.__init__(self)
        self.__filters = TraceFilters() #list of filters to apply
        self.__trace_models = [] #list of trace models 
        self.__max_step = 0        
        return

    # helper function
    def __filter(self, model, iter, data):
        # implements a filtering on rows
        # the filtering on cols is done in ctrls/trace.py using .is_valid_col() function
        row_name = model.get_value(iter, 1)
        stripe = model.get_value(iter, 0)
        
        return data.is_valid_name(str(row_name)) and data.is_valid_type(str(stripe.get_info(error=True)))
        
    def __apply_filter(self):
        if not self._active_trace or not self._active_filter:
            return

        f = self._active_trace.filter_new() #filter for rows (name & type) and cols 
        f.set_visible_func(self.__filter, data=self._active_filter) #filtering 
        self._active_trace_filtered = f#.get_model() #set filtered model
        return

    def __set_max_step(self):
        # set __max_step
        if self._active_filter and self._active_trace:
            
            len_ranges = 0 #count how many cols 
            for start,end in self._active_filter.get_range_iter():
                if start == -1:
                    len_ranges += end
                elif end == -1:
                    len_ranges += self._active_trace.get_max_steps() - start 
                else:
                    len_ranges += end - start + 1
                
            self.__max_step = len_ranges
            
        elif self._active_trace:
            self.__max_step = self._active_trace.get_max_steps()


        return

    #traces functions
    def sort_trace(self, func=None, reverse=None):
        """
        Sort the active trace using the function passed.
        The signature of the function follows:
        def compare( (stripe_name1,stripe1), (stripe_name2,stripe2) )
        Where stripe1 and 2 were the two istances 
        of the TraceStripeModel class.
        The function have to return 
        0 if the stripes are equal, 
        -1 if stripe1 will come before stripe2, 
        1 if stripe2 will come before stripe1.
        If no function is passed, 
        the traces will be ordered by name, 
        in ascending order if reverse == False
        in descending order otherwise.
        If reverse is None, the compare function is used.
        If both 'func' and 'reverse' parameter are set, 
        the reverse parameter is used.
        @see TraceStripeModel
        """
        if self._active_trace: self._active_trace.sort(func, reverse)            
        return

    def add_traces(self, traces):
        """
        add a list of traces to the current list of traces.
        @param list of traces
        """
        if type(traces) is list:
            ts = [t for t in traces if t not in self.__trace_models] #delete existing traces
            self.__trace_models.extend( ts )
        else:
            if traces not in self.__trace_models:
                self.__trace_models.append(traces)

        return

    def add_traces_from_file(self, fname, th=TraceModel):
        """
        @param th: trace handler, a class that can handle a single trace
        """
        reader = src.utils.nusmv.NuSMVTraceReader()
        try:
            m = reader.parse(fname, trace_handler=th)
        except xml.parsers.expat.ExpatError:
            raise Exception("Parse Error")
        
        if not m: raise Exception("Parse Error")
        
        self.add_traces( m )
        
        return m

    def get_traces(self):
        """
        get the current list of traces.
        """
        #i suppose that the user is trusted
        return self.__trace_models

    def del_trace(self, trace):
        """
        delete from the list the selected trace.
        If the trace is the active one, this method unset it.
        You can get the same result using unset_trace()
        @param the trace to remove from the list (the same instance in the list)
        """
        if trace == self._active_trace: self.unset_trace()

        for x in self.__trace_models:
            if x == trace:
                self.__trace_models.remove(trace)
                return
            pass
        return
    
    def get_active_full_trace(self):
        """get the active trace, not filtered"""
        return self._active_trace

    def get_active_trace(self):
        """get the active trace (filtered, if there is an active filter)"""
        return self._active_trace_filtered

    def set_active_trace(self, trace):
        """
        set the trace passed as parameter,
        the active trace (the trace that will be displayed).
        """
        #check if the trace is in the set of valid traces
        assert trace in self.__trace_models, str(trace)+" is not in "+str(self.__trace_models)
        if trace == self._active_trace: return

        self._active_trace = trace

        self._active_trace_filtered = trace
        self.__set_max_step()

        self.__apply_filter()
        assert self._active_trace_filtered

        return

    def unset_trace(self):
        """
        unset the active trace.
        It's better to re-set another trace,
        instead of unset the current one.
        """
        self._active_trace_filtered = self._active_trace = None
        self.__max_step = 0

    #filters functions
    def add_filters(self, filters):
        if type(filters) is not list: self.__filters.add_filters([filters])
        else: self.__filters.add_filters(filters)
        return

    def add_filters_from_xml(self, filters):
        """
        this function load from a file object or a file name a set of filters 
        Remember! only the active filter will be applied.
        @see TraceModel
        @see TraceFilters
        """
        self.__filters.from_xml(filters)
    
    def get_filters(self):
        """
        this function returns a TraceFilters class
        previously loaded and
        saved in the FullTraceModel.
        @see FullTraceModel
        @see TraceFilters
        """
        #i assume that the user is trusted
        return self.__filters

    def del_filter(self, f):
        """
        this function removes a given filter from the set of filters.
        @param the TraceFilter class instance to remove
        """
        return self.__filters.delete_filter(f.get_filter_name())
        return

    def get_active_filter(self):
        return self._active_filter

    def set_active_filter(self, f):
        """
        @param an istance of the TraceFilter class
        @see TraceFilter
        """
        #this method raises keyerror if the name passed is not contained
        #in the filters set
        self.__filters.get_filter( f.get_filter_name() ) 
        
        self._active_filter = f
        self.__set_max_step()
        self.__apply_filter()
        
        return

    def unset_filter(self):
        """disable the active filter"""
        self._active_filter = None
        self._active_trace_filtered = self._active_trace


    #here we define the function used in the ctrl class
    #some functions are called directly from the active_trace,
    #in order to mantain back compatibility with the old model
    def valid_cols_iter(self):
        """
        get an object that provide the list of not filtered cols, asc ordered
        """
        def generator():
            for c in xrange(0,self._active_trace.get_max_steps()):
                if self._active_filter.is_valid_step(c):
                    yield c

        if not self._active_trace:
            return xrange(0,0) #empty iterator

        if not self._active_filter:
            return xrange(0,self._active_trace.get_max_steps())
        
        return generator()
            
        
    def valid_cols(self):
        return [x for x in self.valid_cols_iter()]

    def get_max_step(self):
        return self.__max_step
        
    def is_step_a_loop(self, s):
        if self._active_trace:
            return self._active_trace.is_step_a_loop(s)
        else: return False

    def is_step_in_loop(self, s):
        if self._active_trace:
            return self._active_trace.is_step_in_loop(s)
        else: return False

    def get_value(self, iterator, column): 
        if self._active_trace_filtered and self._active_trace_filtered.iter_is_valid(iterator):
            return self._active_trace_filtered.get_value(iterator, column)
        else: return None
        
    def filter_new(self): #provide further filtering outside this class
        """
        same as the gtk.TreeStore.filter_new() method, on the active trace
        """
        if self._active_trace_filtered:
            return self._active_trace_filtered.filter_new() 
        else:
            return None
    
    def get_delta_stripe(self):
        if self._active_trace:
            di = self._active_trace.get_delta_iter()
            if di:
                return self._active_trace.get_value(di, 0 ) #delta stripe
            else:
                return None
        else:
            return None
        
    def set_delta(self, i): # i must be an iter
        """
        set the delta stripe iterator for the active trace.
        @param i, the iterator for the gtk.TreeStore
        """
        if self._active_trace:
            self._active_trace.set_delta(i)
            return True
        else:
            return False
        
