# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


# The view for slim models and fault injection

import os.path
from gtkmvc import View
import src.utils.globals


class ActivityView (View):
    """This class is used for all activity views, namely
    validation, correctness, performability, safety and fdir.

    This was created to assure a common view showing properties
    list on one side (possibly shown differently case to case), and
    the particular activity subview on the other. Particular views
    can use this class as common base class.

    The subview is available through member self.subview
    """

    glade = os.path.join(src.utils.globals.GLADE_DIR, "activity.glade")
    top = "hpaned_top" 

    def __init__(self, subview):
        """@param subview is the subview to be attached to the activity
        This is passed down by the derived view class."""
        View.__init__(self)

        self.subview = subview        
        # attaches the subview
        self.get_top_widget().pack2(subview.get_top_widget(), 
                                    resize=True, shrink=False)
        return
    pass # end of class
# ----------------------------------------------------------------------


class StartStopActivityView (View):
    """This class is intended to be used as base class of all views
    that have a pair of start/stop buttons which control an
    activity.  The view has two states 'start' and 'stop'. When
    state is 'start', the start button is shown and the activity is
    ready to be run. In 'stop' state the stop button is shown and
    the activity can be cancelled.

    This view must be used in pair with a StartStopActivityCtrl.
    The derived class must overload the methods get_start_button
    and get_stop_button returning the widgets of the buttons
    respectively for starting and stopping the activity.

    The derived class can also implement methods show_start and
    show_stop that shows the respective state (default is simply to
    show/hide corresponding buttons)
    """

    # these two have to be implemented in derived classes
    def get_start_button(self): raise NotImplementedError()
    def get_stop_button(self): raise NotImplementedError()

    def show_start(self): 
        """Default behaviour is to show the start button, and hide
        the stop button"""
        self.get_start_button().show()
        self.get_stop_button().hide()
        return

    def show_stop(self): 
        """Default behaviour is to show the stop button, and hide
        the start button"""
        self.get_start_button().hide()
        self.get_stop_button().show()
        return

    pass # end of class
    
