# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by GMV Aerospace and Defence  
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================



from gtkmvc import View
import cairo, gtk
import random, math

import src.utils.globals

# sub-views:
from src.views.slim import SlimView
from src.views.properties import PropertiesView
from src.views.correctness import CorrectnessView
from src.views.performability import PerformabilityView
from src.views.safety import SafetyView
from src.views.mission import MissionView
from src.views.fdir import FdirView
from src.views.validation import ValidationView
from src.views.tool_handlers import ToolHandlersView
from src.views.tfpg import TfpgView

import options

import gtk
import os.path

class ApplView (View):
    glade = os.path.join(src.utils.globals.GLADE_DIR, "application.glade")
    top = "window_top"

    def __init__(self):
        View.__init__(self)

        # current view
        self.curr_activity = "model"

        # creates all manual widgets
        pb = gtk.ProgressBar()
        pb.set_pulse_step(0.1)
        self['statusbar'].pack_end(pb, expand=False, padding=6)
        self['progress'] = pb
        
        # This map associates names of subviews to the
        # corresponding view instance and page number. It is used
        # by select_view
        self.subviews = {}

        # creates all sub views
        self._create_subviews()
        self.slim = self.subviews["model"][0]
        self.properties = self.subviews["properties"][0]
        self.mission = self.subviews["mission"][0]
        self.correctness = self.subviews["correctness"][0]
        self.performability = self.subviews["performability"][0]
        self.safety = self.subviews["safety"][0]
        self.fdir = self.subviews["fdir"][0]
        self.validation = self.subviews["validation"][0]
        self.tfpg = self.subviews["tfpg"][0]

        self.tools = ToolHandlersView()

        self.enable_failures(False)
        self.enable_properties(False)
        self.enable_mission(False)
        self.enable_correctness(False)
        self.enable_performability(False)

        return

    def select_view(self, name):
        """Selects the corresponding view"""
        self['nb_top'].set_current_page(self.subviews[name][1])
        self['mi_act_%s' % name].set_active(True)
        self.curr_activity = name
        return

    def enable_failures(self, flag):
        """This method is called to enable features that need
        failure model to exist"""

        wids = ('tb_safety', 'mi_act_safety', 
                'tb_fdir', 'mi_act_fdir')
        for wid in wids: self[wid].set_sensitive(flag)
        return
        
    def enable_properties(self, flag):
        """Called to enable/disable access to the properties
        (e.g. when there are no slim files, properties are not
        reachable)"""
        wids = ('tb_properties', 'mi_act_properties')
        for wid in wids: self[wid].set_sensitive(flag)
        return

    def enable_correctness(self, flag):
        """Called to enable/disable access to the correctness view
        (e.g. when there are no slim files, correctness is not
        reachable)"""
        wids = ('tb_correctness', 'mi_act_correctness')
        for wid in wids: self[wid].set_sensitive(flag)
        return

    def enable_performability(self, flag):
        """Called to enable/disable access to the performability view"""
        wids = ('tb_performability', 'mi_act_performability')
        for wid in wids: self[wid].set_sensitive(flag)
        return

    def enable_mission(self, flag):
        """Called to enable/disable access to the mission view
        that need nominal model to exist"""
        wids = ('tb_mission', 'mi_act_mission')
        for wid in wids: self[wid].set_sensitive(flag)
        return

    def enable_tfpg(self, flag):
        """Called to enable/disable access to the tfpg view"""
        wids = ('tb_tfpg', 'mi_act_tfpg')
        for wid in wids: self[wid].set_sensitive(flag)
        return

    def _create_subviews(self):
        # the top-level notebook that contantains all the sub-views

        self._add_sub_view(SlimView(), "model")
        self._add_sub_view(PropertiesView(), "properties")
        self._add_sub_view(MissionView(), "mission")
        self._add_sub_view(ValidationView(), "validation")
        self._add_sub_view(CorrectnessView(), "correctness")
        self._add_sub_view(PerformabilityView(), "performability")
        self._add_sub_view(SafetyView(), "safety")
        self._add_sub_view(FdirView(), "fdir")
        self._add_sub_view(TfpgView(), "tfpg")
        return

    def _add_sub_view(self, subview, name):
        nb = self['nb_top']         
        nb.append_page(subview.get_top_widget())
        self.subviews[name] = (subview, nb.get_n_pages() - 1)
        return
        
    def enable_progress_bar(self, flag):
        """Enable/disables the progress bar in the statusbar. This
        is used during background activities."""
        pb = self['progress']
        if flag: pb.show() 
        else: pb.hide()
        return

    def set_progress_bar_text(self, text):
        """Sets a text in the progress bar"""
        pb = self['progress']
        pb.set_text(text)
        return

    def set_degraded_mode(self):
        """Sets the window title to contain a warning that "degraded
        mode" is on"""
        win = self.get_top_widget()
        win.set_title(win.get_title() + " (Degraded Mode)")
        return

    pass # end of class ApplView
# ----------------------------------------------------------------------

class DialogView (View):
    """Base class for dialogs"""

    def run(self):
        """Runs the dialog, destroys the window and returns the
        response. You can use this method ony once per instance"""
        res = self.get_top_widget().run()
        self.get_top_widget().destroy()
        return res
    pass # end of class
# ----------------------------------------------------------------------


class AskRootView (DialogView):
    """This view is used for the dialog that asks to the user to
    specify the root implementation when not previously
    specified"""

    glade = os.path.join(src.utils.globals.GLADE_DIR, "application.glade")
    top = "dialog_ask_root"

    pass # end of class
# ----------------------------------------------------------------------


class AskQuitView (DialogView):
    """This view is used for the dialog that asks to the user to
    quit expliticly"""

    glade = os.path.join(src.utils.globals.GLADE_DIR, "application.glade")
    top = "dialog_ask_quit"
    pass # end of class
# ----------------------------------------------------------------------


class AboutView (DialogView):
    """This view is for the 'about' dialog"""
    glade = os.path.join(src.utils.globals.GLADE_DIR, "about.glade")
    top = "about_dialog"

    scale_inc_range = (0.0001, 0.0003)
    scale_factor = 0.04
    scale_alpha_factor = 20


    def __init__(self):
        DialogView.__init__(self)            

        class SurfAttr:
            def __init__(self):
                self.alpha = 1.0
                self.scale = 1.0
                self.scale_inc = (random.uniform(*AboutView.scale_inc_range) * 
                                  random.choice((-1,1)))
                return
            pass

        self.surf_logos = [ (cairo.ImageSurface.create_from_png(
                os.path.join(src.utils.globals.IMAGES_DIR,
                             "large/%s-logo.png" % name)), 
                             SurfAttr())
                            for name in ("esa", "fbk", "rwth", "thales")]                

        # widgets setup
        self.setup_view()
        return

    def setup_view(self):
        """Sets up several widgets in the view"""
        # sets the version number in title
        title_lbl = self['about_label_title']
        title_text = title_lbl.get_label()
        version = ".".join(map(str, options.PROG_VER))
        title_lbl.set_markup(title_text % {'version': version })

        # sets images
        self['image_header'].set_from_file(
            os.path.join(src.utils.globals.IMAGES_DIR,
                         "large/about-header.png"))
        
        # sets the da size
        width = max((surf.get_width() for surf, _ in self.surf_logos))

        # this is used also later 
        self.logos_height = sum((surf.get_height() for surf, _ in self.surf_logos))
        
        self['da_logos'].set_size_request(int(width*1.20), 
                                          int(self.logos_height * 1.20))
        return

    def set_appl_mode(self, markup):
        self['label_appl_mode'].set_markup(markup)
        return

    def create_background(self, width, height):
        # this is the surface we draw into:
        self.bg_surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, width, height)

        # this is the context we draw with:
        self.bg_ctx = cairo.Context(self.bg_surface)

        # this is the gradient for the background:
        self.bg_gradient = cairo.LinearGradient(0, 0, 0, height)
        self.bg_gradient.add_color_stop_rgba(0, 0, 0, 0, 1)
        self.bg_gradient.add_color_stop_rgba(1, 1, 1, 1, 1)   

        # this is the context for final drwaing to the drawing area:
        self.ctx = self['da_logos'].window.cairo_create()        
        return
        
    def animate_logos(self):
        for _, info in self.surf_logos:

            if (info.scale >= 1 + AboutView.scale_factor or 
                info.scale <= 1 - AboutView.scale_factor):
                info.scale_inc = -info.scale_inc
                pass
            info.scale += info.scale_inc

            info.alpha = (1.0 - (math.fabs(1.0 - info.scale) * 
                                 AboutView.scale_alpha_factor))
            pass

        return


    def draw_logos(self, rect=None):
        da = self['da_logos']
        width, height = da.window.get_size()

        # prepare the background context
        if rect: self.bg_ctx.rectangle(rect.x, rect.y, rect.width, rect.height)
        else: self.bg_ctx.rectangle(0, 0, width, height)

        # draw the background
        self.bg_ctx.set_source(self.bg_gradient)
        self.bg_ctx.fill()

        # calculates positions for logos
        y_delta = (height - self.logos_height) / (len(self.surf_logos) + 1)
        y = y_delta
        surf_x_y_a = []
        for surf, attr in self.surf_logos:
            x = max(0,(width-surf.get_width())/2)
            surf_x_y_a.append((surf, x, y, attr))
            y += surf.get_height() + y_delta
            pass
        
        # draws logos
        for surf,x,y,attr in surf_x_y_a:
            m = cairo.Matrix()
            m.scale(attr.scale, attr.scale)
            self.bg_ctx.set_matrix(m)
            self.bg_ctx.set_source_surface(surf, x, y)
            self.bg_ctx.paint_with_alpha(attr.alpha)
            pass

        # draw to the final ctx
        if rect: self.ctx.rectangle(rect.x, rect.y, rect.width, rect.height)
        else: self.ctx.rectangle(0, 0, width, height)

        self.ctx.set_source_surface(self.bg_surface, 0, 0)
        self.ctx.paint()
        return
    

    pass # end of class
# ----------------------------------------------------------------------


class WarningDegradedModeView (DialogView):
    """This view is for the 'degraded mode' dialog which is shown at
    startup when the application is not fully installed (i.e. bsome
    background is not available)"""
    glade = os.path.join(src.utils.globals.GLADE_DIR, "application.glade")
    top = "dialog_degraded_mode"
    

    def __init__(self, parent_view):
        DialogView.__init__(self, parent=parent_view)        
        return

    def run(self):
        """Runs the dialog, destroys the window and returns a triple
        with (response_ID, flag, uri). flag tells if the user chose to do not
        show the dialog anymore. uri is the download uri.
        You can use this method ony once per instance"""
        res = self.get_top_widget().run()
        flag = self['cb_degraded_warning'].get_active()
        uri = self['linkbutton_download_part2'].get_uri()
        self.get_top_widget().destroy()
        return (res, flag, uri)

    pass # end of class
# ----------------------------------------------------------------------


class WarningDegradedModeTempView (View):
    glade = os.path.join(src.utils.globals.GLADE_DIR, "application.glade")
    top = "popup_degraded_mode"

    def __init__(self, parent_view):
        View.__init__(self)
        win = self.get_top_widget()
        win.set_transient_for(parent_view.get_top_widget())
        return

    def tick(self):
        """Advances the progress bar of the view. Returns False if the
        bar has reached the end"""
        pb = self['pb_degraded']
        fract = max(0, pb.get_fraction()-0.015)
        pb.set_fraction(fract)
        return fract > 0
        
    def destroy(self):
        self.get_top_widget().destroy()
        return
    pass
