# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


# The view for slim models and fault injection

import gtk
import gobject
import os.path
from gtkmvc import View
import src.utils.globals

from src.views.activity import ActivityView, StartStopActivityView
from src.views.correctness import ResultsView # result for FRA is similar to MC
from balloon import Balloon

class FdirView (ActivityView):    
    
    def __init__(self):
        ActivityView.__init__(self, FdirSubView())

        # The SubViews of FdirView
        self.fda = FdirFdaView()
        self.fia = FdirFiaView()
        self.fra = FdirFraView()
        self.diag = FdirDiagView()
        self.fca = FdirFcaView()

        return

    pass # end of class
# ----------------------------------------------------------------------    


class FdirSubView (View):
    """This is the view of the pure safety, without
    properties"""

    glade = os.path.join(src.utils.globals.GLADE_DIR, "fdir.glade")
    top = "nb_fdir"

    def __init__(self):
        View.__init__(self)  

        # sub views
        self.fda = FdirFdaView()
        self.fia = FdirFiaView()
        self.fra = FdirFraView()
        self.diag = FdirDiagView()
        self.fca = FdirFcaView()

        # connects the sub-views into the notebook:
        nb = self['nb_fdir']
        pages = zip((sv.get_top_widget() 
                     for sv in (self.fda, self.fia, self.fra, self.diag, self.fca)),
                    (gtk.Label(nb.get_tab_label_text(child)) 
                     for child in nb.get_children()))
        for idx in range(nb.get_n_pages()): nb.remove_page(0)
        for child, lbl in pages: nb.append_page(child, lbl)

        return
        
    pass # end of class
# ----------------------------------------------------------------------    


class FdirFdaView (StartStopActivityView):    
    glade = os.path.join(src.utils.globals.GLADE_DIR, "fdir.glade")
    top = "vbox_fda"
    parent = "window_fda" # It is used as parameter for the Balloon Object
    
    def __init__(self):
        StartStopActivityView.__init__(self)
        self.enable_fda(False)
        
        # This window is used to display errors in a tooltip-like fashion
        self.balloon = Balloon(self[FdirFdaView.parent]) 
        return

    def enable_fda(self, flag):
        """Enables all widgets for carrying out fda accordingly to
        the given flag"""
        self['button_run_fda'].set_sensitive(flag)
        if flag: msg = "You can run Fault Detection Analysis"
        else: msg = "Select one property to run Fault Detection Analysis"
        self['label_fda_info'].set_markup(msg)        
        return

    # implemention of virtual methods
    def get_start_button(self): return self['button_run_fda']
    def get_stop_button(self): return self['button_stop_fda']

    def enable_sat(self, flag): 
        self['hbox_fda_sat_bound'].set_sensitive(not flag)
        return
        
        
    def change_color_and_balloon(self, entry, error_lst):
        """Manages the entries colors and displays a balloon with 
            error messages adjacent to the currently edited entry.
            @param   entry:  The reference to an entry that contains the input.
            @type    entry:  gtk.Entry
            @param   error_lst: List of errors, empty when there are none.
            @param   error_lst: list
        """    
        return self.balloon.change_color_and_balloon(entry, error_lst)

    pass # end of class
# ----------------------------------------------------------------------    


class FdirDiagView (StartStopActivityView):    
    glade = os.path.join(src.utils.globals.GLADE_DIR, "fdir.glade")
    top = "vbox_diag"
    parent = "window_diag" # It is used as parameter for the Balloon Object
    
    def __init__(self):
        StartStopActivityView.__init__(self)

        # create and attach result view A
        self.diag_resview_A = ResultsView()
        self.diag_resview_A['SCVDesc'].show_all()
        self.diag_resview_A['label_res_title'].show()
        self.diag_resview_A['button_detach'].hide()
        self['results_container_A'].add(self.diag_resview_A.get_top_widget())

        # create and attach result view B
        self.diag_resview_B = ResultsView()
        self.diag_resview_B['SCVDesc'].show_all()
        self.diag_resview_B['label_res_title'].show()
        self.diag_resview_B['button_detach'].hide()
        self['results_container_B'].add(self.diag_resview_B.get_top_widget())

        # at first, diag is disabled
        self.enable_diag(False)
        self['vbox_satbound'].set_sensitive(False)

        # this will be used for displaying results
        self.result_widget = self['hbox_results']
        self['vbox_diag'].remove(self.result_widget)
        self.trace_window = None
        
        # This window is used to display errors in a tooltip-like fashion
        self.balloon = Balloon(self[FdirDiagView.parent])
        
        return

    def enable_diag(self, flag):
        """Enables all widgets for carrying out diag accordingly to
        the given flag"""

        self['button_run_diag'].set_sensitive(flag)
        if flag: msg = "You can run Diagnosability Analysis."
        else: msg = "Select one property to run Diagnosability Analysis."
        self['label_diag_info'].set_markup(msg)        
        return

    def display_diag_results(self, prop_name, success, trace_view_A, trace_view_B):
        """Redirects the results of a diagnosability analysis run to the
        right GUI widgets."""

        if success == "Error":
            self['label_diag_info'].set_markup('<b>Attention: No result was returned! Error?</b>')
        elif success == "Unknown":
            self['label_diag_info'].set_markup('<b>No counterexample</b> for property <b>%s</b> was found within the given bound.' % prop_name)
        elif success == "True":
            self['label_diag_info'].set_markup('The property <b>%s</b> was found to be <b>diagnosable</b>.' % prop_name)
        else:
            self['label_diag_info'].set_markup('The property <b>%s</b> was found to be <b>not diagnosable</b>.' % prop_name)
            self.diag_resview_A.show_results('<b>Trace A</b>', 'Compliant with diagnosability property.', 
                                             trace_view_A.get_top_widget() if trace_view_A else None)
            self.diag_resview_A.set_icon_stock(gtk.STOCK_YES)
            self.diag_resview_B.show_results('<b>Trace B</b>', 'Not compliant with diagnosability property.', 
                                             trace_view_B.get_top_widget() if trace_view_B else None)
            self.diag_resview_B.set_icon_stock(gtk.STOCK_NO)

            # create window for detached traces
            w = gtk.Window()
            w.set_border_width(12)
            w.set_default_size(640, 480)
            w.connect("delete-event", self.on_trace_window_delete)
            w.add(self.result_widget)
            w.show()
            self.trace_window = w
        return

    def on_trace_window_delete(self, win, event):
        assert self.trace_window, "DIAG: Tried to close non-existing trace window"
        self.trace_window.hide()
        child = self.trace_window.get_child()
        self.trace_window.remove(child)
        self.result_widget = child
        self.trace_window = None
        return

    # implemention of virtual methods
    def get_start_button(self): return self['button_run_diag']
    def get_stop_button(self): return self['button_stop_diag']
    
    
    
    def change_color_and_balloon(self, entry, error_lst):
        """Manages the entries colors and displays a balloon with 
            error messages adjacent to the currently edited entry.
            @param   entry:  The reference to an entry that contains the input.
            @type    entry:  gtk.Entry
            @param   error_lst: List of errors, empty when there are none.
            @param   error_lst: list
        """    
        return self.balloon.change_color_and_balloon(entry, error_lst)
        
        

    pass # end of class
# ----------------------------------------------------------------------       


class FdirFiaView (StartStopActivityView):    
    glade = os.path.join(src.utils.globals.GLADE_DIR, "fdir.glade")
    top = "vbox_fia"
    
    def __init__(self):
        StartStopActivityView.__init__(self)
        return

    def enable_sat(self, flag): 
        self['hbox_fia_sat_bound'].set_sensitive(not flag)
        return

    # implementation of virtual methods
    def get_start_button(self): return self['button_run_fia']
    def get_stop_button(self): return self['button_stop_fia']

    pass # end of class
# ----------------------------------------------------------------------       


class FdirFraView (StartStopActivityView):    
    glade = os.path.join(src.utils.globals.GLADE_DIR, "fdir.glade")
    top = "vbox_fra"
    
    def __init__(self):
        StartStopActivityView.__init__(self)

        # attaches result view
        self.fra_results = ResultsView()
        self['results_fra'].add(self.fra_results.get_top_widget())
        self.enable_fra(False)
        return

    def enable_fra(self, flag):
        """Enables all widgets for carrying out fra accordingly to
        the given flag"""
        self['button_run_fra'].set_sensitive(flag)
        if flag: msg = "You can run Fault Recovery Analysis"
        else: msg = "Select one property to run Fault Recovery Analysis"
        self['label_fra_info'].set_markup(msg)        
        return

    def show_fra_result(self, truth, desc, trace_view, stock_id, fbar_view=None):
        """Show a result coming from Fault Recovery. 
        @param truth is a string holding the truth value of a property.
        @param desc is a markup string with description.
        @param trace_view is TraceView to show, and can be None
        @param stock_id is the stock id of the image in the dialog
        @param fbar_view is a FilterBarView, or None
        """

        title = "<big><b>The property is <tt>%s</tt></b></big>" % truth
        self.fra_results.show_results(title, desc, trace_view.get_top_widget() if trace_view else None)
        
        if trace_view and fbar_view: 
            self.fra_results.set_filter_bar( fbar_view )
        elif not trace_view:
            self.fra_results.unset_filter_bar()

        self.fra_results.set_icon_stock(stock_id)
        return        

    def show_exception(self, title, desc): self.fra_results.show_exception(title, desc)

    # implemention of virtual methods
    def get_start_button(self): return self['button_run_fra']
    def get_stop_button(self): return self['button_stop_fra']

    pass # end of class
# ----------------------------------------------------------------------       


class FdirFcaView (StartStopActivityView):
    
    glade = os.path.join(src.utils.globals.GLADE_DIR, "fdir.glade")
    top = "vbox_fca"
    

    def __init__(self):
        StartStopActivityView.__init__(self)
        return
             


    # implementation of virtual methods
    def get_start_button(self): return self['button_run_fca']
    def get_stop_button(self): return self['button_stop_fca']



    def update(self):
        """Updates the view after a change has been done. All elements are shown
           except for the stop button which is hidden.
        """
        self[self.top].show_all()
        self.get_stop_button().hide()
        return


    def enable_sat(self, flag): 
        """Enable all the components related to the SAT algorithm depending on
           the value of flag.
        """

        self['rb_rec_use_sat'].set_sensitive(not flag)
        self['spinbtn_fca_rec_bound'].set_editable(not flag)
        self['spinbtn_fca_rec_bound'].set_sensitive(not flag)

        self['rb_diag_use_sat'].set_sensitive(not flag)
        self['spinbtn_fca_diag_bound'].set_editable(not flag)
        self['spinbtn_fca_diag_bound'].set_sensitive(not flag)

        self['entry_fca_delay'].set_sensitive(not flag)
        self['radiobutton_fca_diag_finite_delay'].set_sensitive(not flag)
        self['radiobutton_fca_diag_fixed_delay'].set_sensitive(not flag)

        return


    # ---------------------------------------------------------------
    #        Signal handlers
    # ---------------------------------------------------------------
  

    pass # end of class
# ----------------------------------------------------------------------       
