# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by GMV Aerospace and Defence  
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


import gtk
import os.path
from gtkmvc import View
import src.utils.globals

from src.views.mission_phases_op_modes import PhasesOpModesView
from src.views.mission_configurations import ConfigurationsView
from src.views.mission_phase_op_mode_combinations import PhaseOpModeCombinationsView



class MissionView (View):
    """Represents the View of Mission."""

    glade = os.path.join(src.utils.globals.GLADE_DIR, "mission.glade")
    top = "hbox_mission"

    def __init__(self):
        """Constructor Method"""
        View.__init__(self)  

        # current view
        self.curr_activity = "phases_op_modes"

        # This map associates names of subviews to the
        # corresponding view instance and page number. It is used
        # by select_view
        self.subviews = {}

        # created all sub views
        self._create_subviews()
        self.phases_op_modes = self.subviews["phases_op_modes"][0]
        self.configurations = self.subviews["configurations"][0]
        self.phase_op_mode_combinations = self.subviews["phase_op_mode_combinations"][0]

        self.enable_configurations(False)
        self.enable_phase_op_mode_combinations(False)

        return


    def select_view(self, name):
        """Selects the corresponding view.
           @param   name:  The name of the view.
	   @type    name:  string
        """
        self['nb_mission'].set_current_page(self.subviews[name][1])
        self.curr_activity = name
        return


    def enable_configurations(self, flag):
        """Enables/disables the access to the Configurations.
           @param   flag:  The flag that allows the access to the Configurations or not.
	   @type    flag:  boolean
        """
        self['tb_configurations'].set_sensitive(flag)
        return


    def enable_phase_op_mode_combinations(self, flag):
        """Enables/disables the access to the PhaseOpModeCombinations.
           @param   flag:  The flag that allows the access to the PhaseOpModeCombinations or not.
	   @type    flag:  boolean
        """
        self['tb_phase_op_mode_combinations'].set_sensitive(flag)
        return


    def _create_subviews(self):
        """Creates the subviews: phases_op_modes, configurations and phase_op_mode_combinations."""
        self._add_sub_view(PhasesOpModesView(), "phases_op_modes")
        self._add_sub_view(ConfigurationsView(), "configurations")
        self._add_sub_view(PhaseOpModeCombinationsView(), "phase_op_mode_combinations")
        return


    def _add_sub_view(self, subview, name):
        """Adds the corresponding subview to the Mission View.
           @param   subview:  The subview.
	   @type    subview:  gtk.TreeView
           @param   name:  The name of the subview.
	   @type    name:  string
        """
        nb = self['nb_mission']         
        nb.append_page(subview.get_top_widget())
        self.subviews[name] = (subview, nb.get_n_pages() - 1)
        return
        
        
    ################################################################################################
    ######## CELL_RENDERERS of PHASES, OP_MODES, CONFIGURATIONS AND COMBINATIONS FUNCTION ########## 
    ################################################################################################
    
    def cell_renderer_set_text(self, cell_renderer, text):
        """Sets the text of the cell_renderers.
            @param   cell_renderer:  The cell_renderer of the columns.
	    @type    cell_renderer:  gtk.CellRendererText
	    @param   text:  The text to be filled in the cell_renderer of the columns.
	    @type    text:  string
	    
        """        
        if text is None or text == "":
            cell_renderer.set_property('text', "")
        elif text == "None":
            cell_renderer.set_property('text', "none")
        else:
            cell_renderer.set_property('text', text)
            
        return
        
pass # end of MissionView class
# ----------------------------------------------------------------------
