# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================



# The view for slim models and fault injection
import gtk
import os.path
import math
from gtkmvc import View
import src.utils.globals

from property_management.src.patterns import *
from property_management.src.constants import Logic
from property_management.src.constants import PlaceholderType
from property_management.src.properties import *
from property_management.src.formulae import AtomicProposition
import property_management.src.api as properties_api

from src.views.activity import ActivityView, StartStopActivityView
from matplotlib.figure import Figure
#from matplotlib.backends.backend_gtk import FigureCanvasGTK as FigureCanvas
from matplotlib.backends.backend_gtkagg import FigureCanvasGTKAgg as FigureCanvas
from src.views.performability import ToolSettingsSubView
#from matplotlib.backends.backend_gtkcairo import FigureCanvasGTKCairo as FigureCanvas


class SafetyView (ActivityView):    
    
    def __init__(self):
        ActivityView.__init__(self, SafetySubView())
        return

    pass # end of class
# ----------------------------------------------------------------------    


class SafetySubView (View):
    """This is the view of the pure safety, without
    properties"""

    glade = os.path.join(src.utils.globals.GLADE_DIR, "safety.glade")
    top = "nb_safety"

    def __init__(self):
        View.__init__(self)   

        # sub views:
        self.ftg = FtgSubView()
        self.fmea = FmeaSubView()
        self.ftoe = FtoeSubView()
        self.ftv = FtvSubView()
        self.fte = FteSubView()

        # connects the subviews
        nb = self['nb_safety']
        pages = zip((self.ftg.get_top_widget(), 
                     self.fmea.get_top_widget(), 
                     self.ftoe.get_top_widget(),
                     self.ftv.get_top_widget(),
                     self.fte.get_top_widget()), 
                    (gtk.Label(nb.get_tab_label_text(child)) 
                     for child in nb.get_children()))

        for idx in range(nb.get_n_pages()): nb.remove_page(0)
        for child, lbl in pages: nb.append_page(child, lbl)        
        return

    pass # end of class
# ----------------------------------------------------------------------    
        



class FtgSubView (StartStopActivityView):
    glade = os.path.join(src.utils.globals.GLADE_DIR, "safety.glade")
    top = "vbox_ftg"

    def __init__(self):
        StartStopActivityView.__init__(self)
        self.enable_ftg(False)
        return

    def enable_ftg(self, flag):
        """Use to enable/disable controls for performing FTG"""
        self['button_run_ftv'].set_sensitive(flag)
        if not flag: self['label_ftg_info'].set_markup("You have to select <b>one</b> property representing the top level event")
        else: self['label_ftg_info'].set_markup("You can <i>generate</i> a Fault Tree")
        return

    def activate_ft_use_sat(self, flag):
        """Use to activate/deactivate use sat."""
        self['ft_use_sat'].set_active(flag)

    def enable_ft_use_bdd(self, flag):
        """Use to enable/disable bdd"""
        self['ft_use_bdd'].set_sensitive(flag)

    def enable_ft_dynamic(self, flag):
        """Use enable/disable dynamic"""
        self['ft_use_bdd'].set_sensitive(flag)

    def enable_ft_dynamic(self, flag):
        """Use to enable/disable dynamic"""
        self["cb_dyn_ft"].set_sensitive(flag)

    def enable_ft_use_sat(self, flag):
        """Use to enable/disable sat"""
        self['ft_use_sat'].set_sensitive(flag)

    def enable_ft_sat_bound(self, flag):
        """Use to enable/disable sat bound"""
        self['ft_sat_bound_lb'].set_sensitive(flag)
        self['ft_sat_bound'].set_sensitive(flag)

    def enable_ft_use_sbmc(self, flag):
        """Use to enable/disable sat sbmc"""
        self['ft_use_sbmc'].set_sensitive(flag)

    def enable_ft_sbmc_complete(self, flag):
        """Use to enable/disable sbmc complete"""
        self['ft_sbmc_complete'].set_sensitive(flag)
    
    # implemention of virtual methods
    def get_start_button(self): return self['button_run_ftv']
    def get_stop_button(self): return self['button_stop_ftv']    
    pass # end of class
# ----------------------------------------------------------------------    


class FmeaSubView (StartStopActivityView):
    glade = os.path.join(src.utils.globals.GLADE_DIR, "safety.glade")
    top = "vbox_fmea"

    def __init__(self):
        StartStopActivityView.__init__(self)
        self.enable_fmea(False)
        return

    def enable_fmea(self, flag):
        self['button_gen_fmea_table'].set_sensitive(flag)
        if not flag: self.set_info("You have to select <b>one or more</b> properties")
        else: self.set_info("You can <i>generate</i> a FMEA table")
        return

    def enable_sat(self, flag):
        self['hbox_fmea_sat_bound'].set_sensitive(flag)
        return

    def enable_dynamic(self, flag): 
        self['cb_dyn_fmea'].set_sensitive(flag)
        return

    def enable_compact(self, flag): 
        self['cb_compact_fmea'].set_sensitive(flag)
        return
    
    def set_info(self, txt):
        self['label_fmea_info'].set_markup(txt)
        return

    # implemention of virtual methods
    def get_start_button(self): return self['button_gen_fmea_table']
    def get_stop_button(self): return self['button_stop_fmea']    
    pass # end of class
# ----------------------------------------------------------------------    
    

class FtoeSubView (StartStopActivityView):
    glade = os.path.join(src.utils.globals.GLADE_DIR, "safety.glade")
    top = "vbox_ftoe"

    def __init__(self):
        StartStopActivityView.__init__(self)
        self.enable_ftoe(False)
        return

    def enable_ftoe(self, flag):
        """Use to enable/disable controls for performing FTG"""
        self['button_run_ftoe'].set_sensitive(flag)
        if not flag: self['label_ftoe_info'].set_markup("You have to select <b>one or more</b> properties with a fault tree")
        else: self['label_ftoe_info'].set_markup("You can <i>run</i> Fault Tolerance Evaluation")
        return
    
    # implemention of virtual methods
    def get_start_button(self): return self['button_run_ftoe']
    def get_stop_button(self): return self['button_stop_ftoe']
    pass # end of class
# ----------------------------------------------------------------------    

class FtvSubView (StartStopActivityView):
    glade = os.path.join(src.utils.globals.GLADE_DIR, "performability.glade")
    top = "performability_top"
    line = None
    figure = None
    axes = None
    canvas = None

    def __init__(self):
        StartStopActivityView.__init__(self)
        self.figure = Figure()
        self.axes = self.figure.add_subplot(111)
        self.canvas = FigureCanvas(self.figure)
        self.ftvsub = FtvSubSubView(self)
        graph_widget = self['vbox_graph']
        graph_widget.pack_start(self.canvas)
        graph_widget.reorder_child(self.canvas, 0) # reorder the probability entry widget
        hbox_prob = self['hbox_prob']
        graph_widget.reorder_child(hbox_prob, 1)
        ftvsubwidget = self.ftvsub.get_top_widget()
        self['vbox_sub_ftv'].pack_start(ftvsubwidget)
        graph_widget.show_all()
        self.enable_ftv(False)
        return
        
    def get_start_button(self): 
        return self['button_run']
    
    def get_stop_button(self): 
        return self['button_stop']

    def pattern_selection_changed(self, new_pattern, check_callback_method):
        self.ftvsub.pattern_selection_changed(new_pattern, check_callback_method)
    
    def enable_ftv(self, flag):
        """Use to enable/disable controls for performing FTV"""
        self['button_run'].set_sensitive(flag)
        self.ftvsub.enable_ftv(flag)
        return
    
    def display_et(self, number):
        self['label1'].set_label("<b>Result</b>")
        self.canvas.set_visible(False)
        self['label3'].set_label("Estimated Time:")
        self['label_prob'].set_markup('<span color="red"><b>' + str(number) + ' </b></span>')
        
    def display_lra(self, number):
        self['label1'].set_label("<b>Result</b>")
        self.canvas.set_visible(False)
        self['label3'].set_label("Long Run Average:")
        self['label_prob'].set_markup('<span color="red"><b>' + str(number) + ' </b></span>')
        
    def display_percentage(self, number):
        def __round_percentage(number, digits):
            result = 0
            number = number * math.pow(10, digits)
            if number % 5 <= 5:
                result = math.floor(number / 10) / math.pow(10, (digits - 1))
            else:
                result = math.ceil(number / 10) / math.pow(10, (digits - 1))
            return result
        
        probability_percentage = __round_percentage(number * 100, 2)
        
        self['label1'].set_label("<b>Cumulative Distribution Function</b>")
        self.canvas.set_visible(True)
        self['label3'].set_label("Maximum Probability:")
        self['label_prob'].set_markup('<span color="red"><b>' + str(probability_percentage) + ' %</b></span>')
        
    def draw(self, bounds, xpoints, ypoints):
        (lower, upper) = bounds
        self.axes.clear()
        self.axes.set_autoscale_on(False)
        self.axes.set_ylabel("Probability")
        self.axes.set_xlabel("Upper time bound")
        self.axes.set_ylim(0, 1) # set y-axis range
        self.axes.set_xlim(lower, upper) # set x-axis range
        self.axes.plot(xpoints, ypoints, 'b-')
        self.canvas.blit()
        self.canvas.draw_idle()
    
    pass # end of class
# ----------------------------------------------------------------------    

class FtvSubSubView (StartStopActivityView):
    glade = os.path.join(src.utils.globals.GLADE_DIR, "ftverification.glade")
    top = "vbox_ftv"
    
    def __init__(self, parent):
        StartStopActivityView.__init__(self)
        self.parent = parent
        self.enable_ftv(False)
        self.tool_settings_sub = ToolSettingsSubView(self)
        tool_settings_sub_widget = self.tool_settings_sub.get_top_widget()
        base = self[FtvSubSubView.top].get_style().bg
        self['txtview_pattern'].modify_base(gtk.STATE_NORMAL, base[2])
        self['vbox_pattern_settings'].pack_start(tool_settings_sub_widget)
        self['vbox_pattern_settings'].hide_all()
        return
                
    def get_start_button(self):
        return None
    
    def get_stop_button(self):
        return None
        
    def enable_ftv(self, flag):
        if not flag: self['label_ftv_info'].set_markup("You have to select <b>one</b> property with a fault tree and a pattern:")
        else: self['label_ftv_info'].set_markup("You can <i>run</i> Fault Tree Verification")
        return

    pass # end of class
# ----------------------------------------------------------------------    

class FteSubView (StartStopActivityView):
    glade = os.path.join(src.utils.globals.GLADE_DIR, "ftevaluation.glade")
    top = "vbox_fte"
    
    def __init__(self):
        StartStopActivityView.__init__(self)
        self.tool_settings_sub = ToolSettingsSubView(self)
        tool_settings_sub_widget = self.tool_settings_sub.get_top_widget()
        tool_settings_alignment = gtk.Alignment(0.0, 0.0, 1.0, 0)
        tool_settings_alignment.add(tool_settings_sub_widget)
        self['vbox_any'].pack_start(tool_settings_alignment)
        tool_settings_alignment.show()
        self['vbox_any'].reorder_child(tool_settings_alignment,1)

    def get_start_button(self):
        return self['button_start_fte']
    
    def get_stop_button(self):
        return self['button_stop_fte']
    
    def enable_fte(self, flag):
        if not flag: self['label_fte_info'].set_markup("You have to select <b>one</b> property with a fault tree, enter time bounds and a criticality number.")
        else: self['label_fte_info'].set_markup("You can <i>run</i> Fault Tree Evaluation.")
        self['button_start_fte'].set_sensitive(flag)        
        return

    pass # end of class
# ----------------------------------------------------------------------    
