# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================



import gtk
import gobject
import pango
import os
import os.path

import src.utils.globals
from gtkmvc import View 


# ----------------------------------------------------------------------
class TraceView (View):
    """The view of a trace model"""

    glade = os.path.join(src.utils.globals.GLADE_DIR, "trace.glade")

    def __init__(self, windowed=True, filters=True):
        top = ['window_main','window_filter'] if windowed else ['main_vbox']
        super(TraceView, self).__init__(top=top)
        
        if windowed: self['window_main'].set_size_request(650,400)
        else:
            #hide unnecessary widgets
            self['MB'].hide()
            pass

        if filters: self.filterBarView = src.views.trace.FilterBarView(windowed=False)
        self.has_filters = filters

        if self.has_filters:
            #add the filter bar to the view
            self['extras_vbox'].pack_end(self.filterBarView['TBFilter'], expand=False)
            #self.filterBarView['CBStoredFilters'].set_active(0) #the first one is the default active
        else:
                        #hide the unnecessary widget
            self['MIEdit'].hide()
            self['IMISave'].hide()
            self['IIFilter'].hide()
            pass

        return

    def set_horizontal_scroll(self, enable):
        sw = self['sw_top']
        h, v = sw.get_policy()
        if enable: sw.set_policy(gtk.POLICY_AUTOMATIC, v)
        else: sw.set_policy(gtk.POLICY_NEVER, v)
        return

    def set_vertical_scroll(self, enable):
        sw = self['sw_top']
        h, v = sw.get_policy()
        if enable: sw.set_policy(h, gtk.POLICY_AUTOMATIC)
        else: sw.set_policy(h, gtk.POLICY_NEVER)
        return

    def get_scroll_pos(self):
        """Returns the currently scrolling position. This can be
        used as parameter to set_scroll_pos later on a different
        view, to scroll it to the same position."""
        tv = self['treeview']
        return (tv.get_hadjustment().get_value(), tv.get_vadjustment().get_value())

    def set_scroll_pos(self, pos):
        """Scrolls the trace to the given position, as returned by
        a previous call to get_scroll_pos. The given position may
        come from a different trace view instance"""
        def __delayed_call(self, pos):
            tv = self['treeview']
            hadj, vadj = tv.get_hadjustment(), tv.get_vadjustment()
            
            hadj.set_value(min(pos[0], (hadj.upper - hadj.page_size)))
            vadj.set_value(min(pos[1], (vadj.upper - vadj.page_size)))
            return

        gobject.idle_add(__delayed_call, self, pos)
        return

    pass # end of class


class GuidedSimTransitionView (View):    
    glade = os.path.join(src.utils.globals.GLADE_DIR, "guided_simulation.glade")
    top = "top_sim_guided_transitions"

    DISCRETE, TIMED = range(2)

    def __init__(self):
        View.__init__(self)

        self.set_no_constratins()
        self.color_ok = self['sb_time_min'].get_style().base[gtk.STATE_NORMAL]
        self.color_error = gtk.gdk.color_parse("indian red")
        self['label_time_msg'].set_markup("")
        return

    def show_max_constraint(self, flag):
        """Depending on the value of the parameter flag, the gui elements used to specify the max constraint
           of timed transitions are shown"""
        if not flag:
            self['sb_time_max'].hide()
            self['cb_max'].hide()
            self['cb_time_max_unit'].hide()
        else:
            self['cb_max'].show()
            if self['cb_max'].get_active() != 0: 
                self['sb_time_max'].show()
                self['cb_time_max_unit'].show()
            pass
        return           

    def set_no_constratins(self):
        # sets up the view
        self['cb_min'].set_active(0)
        self['cb_max'].set_active(0)
        return


    def set_error(self, flag):
        """Sets if constratins are not valid, or if they are instead"""
        if flag: 
            self['sb_time_min'].modify_base(gtk.STATE_NORMAL, self.color_error)
            self['sb_time_max'].modify_base(gtk.STATE_NORMAL, self.color_error)
            self['label_time_msg'].set_markup("""<b>Constraints are not valid</b>
A timed-transition is possible, but not with given constraints, which are inconsistent.""")
        else: 
            self['sb_time_min'].modify_base(gtk.STATE_NORMAL, self.color_ok)
            self['sb_time_max'].modify_base(gtk.STATE_NORMAL, self.color_ok)
            self['label_time_msg'].set_markup("")
        return
        
    def show_timed_trans_simulation(self, flag):
        """When the slim model is not hybrid, or when timed
        transition is not allowed, the time-transition pane is
        hidden"""
        nb = self['top_sim_guided_transitions']
        nb.set_show_tabs(flag)
        if not flag or True: # It is better to show always the discrete sub-part, otherwise the run button stays disabled
            self.set_guided_simulation_type(GuidedSimTransitionView.DISCRETE)
            pass
        
        return

    def get_min_constr(self):
        """Returns the min time operator, the set value and the time unit in a tuple"""
        if self['cb_min'].get_active() == 0: return (None, None, None)
        if self['cb_time_min_unit'].get_sensitive():
            time_unit = self['cb_time_min_unit'].get_active_text().strip()
        else:
            time_unit = None
        return (self['cb_min'].get_active_text().strip(), self['sb_time_min'].get_text(), time_unit)

    def get_max_constr(self):
        """Returns the max time operator, the set value and the time unit in a tuple"""
        if self['cb_max'].get_active() == 0 or self['cb_min'].get_active() == 1: return (None, None, None)
        if self['cb_time_max_unit'].get_sensitive():
            time_unit = self['cb_time_max_unit'].get_active_text().strip()
        else:
            time_unit = None
        return (self['cb_max'].get_active_text().strip(), self['sb_time_max'].get_text(), time_unit)
        
    def get_guided_simulation_type(self):
        """Returns the currently selected view: returns 
        DISCRETE or TIMED"""
        return (GuidedSimTransitionView.DISCRETE, 
                GuidedSimTransitionView.TIMED,)\
                [self['top_sim_guided_transitions'].get_current_page()]

    def set_guided_simulation_type(self, _type):
        """Sets the currently selected view: DISCRETE or TIMED"""        
        self['top_sim_guided_transitions'].set_current_page(
            (GuidedSimTransitionView.DISCRETE, 
             GuidedSimTransitionView.TIMED,)[_type])

    pass # end of class



#extensible trace view
class ExtensibleTraceView(View):
    """The view of an extensible trace model"""

    glade = os.path.join(src.utils.globals.GLADE_DIR, "guided_simulation.glade")

    def __init__(self, trans_sim_guided_view, windowed=True, filters=True):
        """
        @param trans_sim_guided_ctrl is a GuidedSimTransitionView
        instance used for showing the transition-based guided
        simulation."""
        top = ['window_main'] if windowed else ['hpaned_top']
        super(ExtensibleTraceView, self).__init__(top=top)
        
        #SUB views
        self.tvshow_view = TraceView(windowed=False, filters=filters)
        self.value_view = TraceView(windowed=False, filters=filters)        
        self.trans_view = trans_sim_guided_view

        #sub treeviews
        self.tvshow = self.tvshow_view['treeview']
        self.tvvalue = self.value_view['treeview']

        nb_left = self['nb_left_top']
        nb_left.set_show_tabs(False)
        
        # value-based guided sim
        nb_left.remove_page(1)
        nb_left.insert_page(self.value_view.get_top_widget(), None, 1)

        # trans-based guided sim
        nb_left.remove_page(0)
        nb_left.insert_page(self.trans_view.get_top_widget(), None, 0)

        nb_left.set_current_page(0)

        # trace on right side
        self['hpaned_top'].pack2(self.tvshow_view.get_top_widget())
        self['hpaned_top'].set_position(300)
        #self['hpaned_top'].show_all()

        if windowed: self['window_main'].set_size_request(500,500)

        self.has_filters = filters
        if filters: self.filterBarView = FilterBarView(windowed=False)
            
        # hide toolbars
        self.tvshow_view['extras_vbox'].hide()
        self.value_view['extras_vbox'].hide()
        
        #link together the scrollbar
        sshow = self.tvshow_view['sw_top']
        sadd = self.value_view['sw_top']
        #create the shared adjustment
        adj = gtk.Adjustment(
            value=0.00,
            lower=0.00,
            upper=100.00,
            step_incr=1.00,
            page_incr=10.00,
            page_size=10.00
            )
        #set the shared adjustment
        sshow.set_vadjustment( adj )
        sadd.set_vadjustment( adj )
        
        #connect the row-selected signal in order to keep equal the selection in tvshow and tvvalue
        tvshow_sel = self.tvshow.get_selection()
        tvvalue_sel = self.tvvalue.get_selection()
        tvshow_sel.connect('changed', self._selection_changed, ())
        tvvalue_sel.connect('changed', self._selection_changed, ())
        tvshow_sel.set_mode(gtk.SELECTION_BROWSE)
        tvvalue_sel.set_mode(gtk.SELECTION_BROWSE)

        self.tvshow.connect('unselect-all', self._unselect)
        self.tvvalue.connect('unselect-all', self._unselect)

        return

    def detach_trans_sim_guided_view(self): 
        """Since self is constructed from scratch every time a
        trace has to be constructed, and as instead the trans-based
        guided simulation has to be reused, this getter gives
        access to the internal GuidedSimTransitionView for being
        passed to the new ExtensibleTraceView instance
        constructor."""
        # trans-based guided sim
        nb_left = self['nb_left_top']
        nb_left.remove_page(0)
        return self.trans_view

    def select_trans_based_guided(self):
        self['nb_left_top'].set_current_page(0)
        return

    def select_value_based_guided(self):
        self['nb_left_top'].set_current_page(1)
        return

    def _unselect(self, tv):
        self.tvshow.get_selection().unselect_all()
        self.tvvalue.get_selection().unselect_all()
        return

    def _selection_changed(self, ts, user_data ):
        model,sel = ts.get_selected_rows()
        #keep syncronized the two selections
        add_sel = self.tvvalue.get_selection()
        show_sel = self.tvshow.get_selection()
        for s in sel:
            path = s[0]
            add_sel.select_path(path)
            show_sel.select_path(path)
            pass

        if not len(sel):
            add_sel.unselect_all()
            show_sel.unselect_all()
            pass

        return

    def get_scroll_pos(self):
        """Returns the currently scrolling positions. This can be
        used as parameter to set_scroll_pos later on a different
        view, to scroll it to the same position."""
        return (self.tvshow_view.get_scroll_pos(), 
                self.value_view.get_scroll_pos(), 
                self['hpaned_top'].get_position())

    def set_scroll_pos(self, pos):
        """Scrolls the trace to the given position, as returned by
        a previous call to get_scroll_pos. The given position may
        come from a different trace view instance"""
        self.tvshow_view.set_scroll_pos(pos[0])
        self.value_view.set_scroll_pos(pos[1])
        self['hpaned_top'].set_position(pos[2])
        return


    pass # end of class
#---------------------------------------------------------------------
class FilterBarView (View):
    """The view of a trace model"""

    glade = os.path.join(src.utils.globals.GLADE_DIR, "trace.glade")

    def __init__(self, windowed=True):
        top = ['window_filter_bar'] if windowed else ['TBFilter']
        super(FilterBarView, self).__init__(top=top)
        
        if windowed: self['window_filter_bar'].set_size_request(650,50)
        self.filterView = src.views.trace.FilterView()
        
        # This is used to set tooltips, as it is not always available
        self._set_tooltip = gtk.Widget.set_tooltip_markup \
            if gtk.pygtk_version >= (2,12,0) else None
        
        return

    def valid_tooltip(self):
        if self._set_tooltip:
            self._set_tooltip( self['TENameFilter'], "" )
            pass
        self['TENameFilter'].modify_base( gtk.STATE_NORMAL, gtk.gdk.color_parse("white") )
        
    def errors_tooltip(self, errors):
        ERR = "These regular expressions are <b>not valid</b>:\n"+','.join(errors)+'\nAnd <b>will be skipped</b>'
            
        self['TENameFilter'].modify_base( gtk.STATE_NORMAL, gtk.gdk.color_parse("indian red") )
        if self._set_tooltip:
            self._set_tooltip( self['TENameFilter'], ERR )
            pass
        return
    pass # end of class


class FilterView (View):
    """The view of a filter model"""

    glade = os.path.join(src.utils.globals.GLADE_DIR, "trace.glade")

    def __init__(self):
        top = 'window_filter'
        super(FilterView, self).__init__(top=top)
        
        self['window_filter'].set_size_request(400,300)
        return



# ----------------------------------------------------------------------
class CellRendererTrace (gtk.GenericCellRenderer):

    """Customed ad-hoc cell renderer for displaying traces"""
    
    __gproperties__ = {

        "value" : (gobject.TYPE_STRING, "The current value to show",
                   "The current value to show",
                   None, gobject.PARAM_READWRITE),
        
        "prev-value" : (gobject.TYPE_STRING, "The value shown on the previous (left) cell",
                        "This is the value shown by the previous cell. This can be None.",
                        None, gobject.PARAM_READWRITE),

        "next-value" : (gobject.TYPE_STRING, "The value shown on the next (right) cell",
                        "This is the value shown by the next cell. This can be None.",
                        None, gobject.PARAM_READWRITE),


        "is-boolean" :  (gobject.TYPE_BOOLEAN, "A flag that specify if the displayed value is boolean.",
                         "A flag that specify if the displayed value is boolean.",
                         False, gobject.PARAM_READWRITE ),

        "frame-fg-color" : (gobject.TYPE_STRING, "The colour of the line used to draw the cell frame",
                            "The colour of the line used to draw the cell frame",
                            None, gobject.PARAM_READWRITE),

        "waveform-fg-color" : (gobject.TYPE_STRING, "The colour of the line used to draw the waveform",
                               "The colour of the line used to draw the the waveform",
                               None, gobject.PARAM_READWRITE),
        "waveform-sel-fg-color" : (gobject.TYPE_STRING,
                                   "The colour of the line used to draw the waveform, when the row is selected",
                                   "The colour of the line used to draw the the waveform, when the row is selected",
                                   None, gobject.PARAM_READWRITE),

        "max-width" : (gobject.TYPE_UINT,
                       "The maximum cell width, in pixels",
                       "The maximum cell width, in pixels",
                       0, # min
                       1000, # max
                       100, # default
                       gobject.PARAM_READWRITE),

        "bool-waveform-width" : (gobject.TYPE_UINT,
                       "The waveform line width, in pixels",
                       "The waveform line width, in pixels",
                       0, # min
                       5, # max
                       2, # default
                       gobject.PARAM_READWRITE),
        
        }

    def __init__(self):
        self.__gobject_init__()
        self.is_boolean = False
        self.value = None
        self.prev_value = None
        self.next_value = None
        self.frame_fg_color = None
        self.waveform_fg_color = None
        self.waveform_sel_fg_color = None
        self.bool_waveform_width = 2
        self.max_width = 100 
        return
    
    def do_set_property(self, pspec, value):
        setattr(self, pspec.name.replace("-","_"), value)
        return
    
    def do_get_property(self, pspec):
        return getattr(self, pspec.name.replace("-","_"))

    def on_get_size(self, widget, cell_area):
        if self.value is None: return (0, 0, 0, 0)

        # calculates width and height
        ctx = widget.get_pango_context()
        metrics = ctx.get_metrics(ctx.get_font_description())
        width = pango.PIXELS(metrics.get_approximate_char_width()) * len(self.value) - 1
        
        if cell_area:
            width = max(width, cell_area.width)
            height = cell_area.height
        else: height = 20
        width = min(width, self.max_width)
                
        xpad = self.get_property("xpad")
        ypad = self.get_property("ypad")
       
        calc_width = xpad*2 + width
        calc_height = ypad*2 + height
        if cell_area and width > 0 and height > 0:
            x_offset = self.get_property("xalign") * (cell_area.width - calc_width - xpad)
            y_offset = self.get_property("yalign") * (cell_area.height - calc_height - ypad)
        else:
            x_offset = 0
            y_offset = 0
            pass
        
        return (x_offset, y_offset, calc_width, calc_height)


    def create_text_layout(self, widget, width):
        """This function is called during rendering. The result will
        be a layout built and returned to contain the text of the
        current value, possibly truncated to fit in the maximum
        space. The layout will have a maximum size of the given width"""
     
        ctx = widget.get_pango_context()
        metrics = ctx.get_metrics(ctx.get_font_description())

        text = self.value

        approx_width = pango.PIXELS(metrics.get_approximate_char_width())
        max_chars = width / approx_width
        
        if len(text) > max_chars: text = "..." + text[-max_chars:] # truncates the text
        layout = widget.create_pango_layout(text)
        layout.set_alignment(pango.ALIGN_CENTER)

        return layout
        
    
    def on_render(self, window, widget, background_area, cell_area, expose_area, flags):
        if self.value is None: return        
        draw_rect = expose_area.intersect(cell_area)

        # Draws the area:
        self.draw_cell(window, widget, draw_rect, flags)
        return

    def draw_cell(self, drawable, widget, rect, flags):
        """Does the drawing work"""
        self.draw_frame(drawable, widget, rect, flags)
        self.draw_waveform(drawable, widget, rect, flags)
        return


    def draw_text(self, drawable, widget, rect, flags):
        """Draws the text within (at the centre) the given rect,
        returns the real rect used to draw the text"""
        
        gct = drawable.new_gc()
        gct.set_rgb_fg_color(self.get_best_text_color(widget, flags))

        layout = self.create_text_layout(widget, rect.width)
        text_w, text_h = layout.get_pixel_size()

        x = max(0, (rect.width - text_w)/2)
        res = gtk.gdk.Rectangle(rect.x + x,
                                rect.y + (rect.height - text_h)/2 - 1,
                                text_w, text_h)

        drawable.draw_layout(gct, res.x, res.y, layout)
        return res
        

    def draw_waveform(self, drawable, widget, rect, flags):
        if not self.value:
            return
        wave_rect = rect.copy()        
        if self.is_boolean:
            delta = max(wave_rect.height / 4, 1)
            wave_rect.y += delta
            wave_rect.height -= delta
            self.draw_waveform_boolean(widget, drawable, wave_rect, flags)
            
        else: # an enumerative value:
            delta = max(wave_rect.height / 8, 1)
            wave_rect.y += delta
            wave_rect.height -= delta
            self.draw_waveform_enumerative(widget, drawable, wave_rect, flags)
            pass
        
        return

    def draw_waveform_boolean(self, widget, drawable, rect, flags):
        #self.draw_waveform_enumerative(widget, drawable, rect, flags)
        #return
        col = self.get_best_waveform_color(widget, flags)
        gc = drawable.new_gc(line_width=self.bool_waveform_width)
        gc.set_rgb_fg_color(col)

        hy = rect.y + self.bool_waveform_width / 2
        ly = rect.y + rect.height - self.bool_waveform_width / 2
        
        true_values = ['True']
        if self.value in true_values : 
            drawable.draw_line(gc, rect.x, hy, rect.x+rect.width, hy)
        else: 
            drawable.draw_line(gc, rect.x, ly, rect.x+rect.width, ly)

        if self.prev_value and self.prev_value != self.value:
            drawable.draw_line(gc, rect.x, hy, rect.x, ly+1)
            pass
            
        #if self.next_value not in (self.value, None):
        #    x = rect.x+rect.width
        #    drawable.draw_line(gc, x, hy, x, ly+1)
        #    pass
        
        return


    def draw_waveform_enumerative(self, widget, drawable, rect, flags):
        col = self.get_best_waveform_color(widget, flags)
        gc = drawable.new_gc()
        gc.set_rgb_fg_color(col)
        
        ymid = rect.y + rect.height/2
        ybot = rect.y + rect.height - 1
        xr = rect.x+rect.width

        segms = []
        if self.prev_value is None or self.prev_value != self.value:
            x1 = rect.x+rect.height/2
            segms += [(x1, ybot, rect.x, ymid), (rect.x, ymid, x1, rect.y)]
            if self.next_value is None or self.next_value != self.value:
                x2 = rect.x*2 + rect.width - x1
                segms += [(x1, rect.y, x2, rect.y), (x2, rect.y, xr, ymid),
                          (xr, ymid, x2, ybot), (x2, ybot, x1, ybot)]
            else: segms += [(x1, rect.y, xr, rect.y), (xr, ybot, x1, ybot)]

        elif self.next_value is None or self.next_value != self.value:
            x2 = rect.x + rect.width - rect.height/2
            segms += [(rect.x, rect.y, x2, rect.y), (x2, rect.y, xr, ymid),
                      (xr, ymid, x2, ybot), (x2, ybot, rect.x, ybot)]
        else: segms += [(rect.x, rect.y, xr, rect.y), (xr, ybot, rect.x, ybot)]
            
        drawable.draw_segments(gc, segms)
        self.draw_text(drawable, widget, rect, flags)        
        return
        

    def draw_frame(self, drawable, widget, rect, flags):
        col = self.get_best_frame_color(widget, flags)
        gc = drawable.new_gc()
        gc.set_rgb_fg_color(col)
        drawable.draw_rectangle(gc, False, rect.x, rect.y, rect.width, rect.height)
        return

    def get_best_waveform_color(self, treeview, flags):
        """Returns the best foreground color for the waveform line
        depending on the current context"""

        if self.waveform_fg_color is None:
            st = treeview.get_style()
            if flags & gtk.CELL_RENDERER_SELECTED: col = st.fg[gtk.STATE_SELECTED]
            else: col = st.fg[gtk.STATE_NORMAL]
            return col.copy()

        if self.waveform_sel_fg_color is not None and flags & gtk.CELL_RENDERER_SELECTED:
            return gtk.gdk.color_parse(self.waveform_sel_fg_color)
        return gtk.gdk.color_parse(self.waveform_fg_color)


    def get_best_text_color(self, treeview, flags):
        st = treeview.get_style()
        if flags & gtk.CELL_RENDERER_SELECTED: col = st.fg[gtk.STATE_SELECTED]
        else: col = st.fg[gtk.STATE_NORMAL]
        
        return col.copy()

    def get_best_frame_color(self, treeview, flags):
        """Returns the best foreground color for the frame depending
        on the current context"""

        # if not set by the user through the property, calculates the
        # color of the frame, depending on the cell background and
        # selection status:
        if self.frame_fg_color is None:
            st = treeview.get_style()
            if flags & gtk.CELL_RENDERER_SELECTED: col = st.bg[gtk.STATE_SELECTED]
            elif self.get_property("cell-background-set"):
                col = self.get_property("cell-background-gdk")
            else: col = st.bg[gtk.STATE_NORMAL]
            col = col.copy()
            
        else: col = gtk.gdk.color_parse(self.frame_fg_color)

        # changes slightly the color, depending on the prelit as well:
        if flags & gtk.CELL_RENDERER_PRELIT: col_delta = 12000
        elif self.frame_fg_color is None: col_delta = 7000
        else: col_delta = 0

        light_threshold = 50000
        if col.red < light_threshold: col.red += col_delta
        else: col.red -= col_delta
        if col.green < light_threshold: col.green += col_delta
        else: col.green -= col_delta
        if col.blue < light_threshold: col.blue += col_delta
        else: col.blue -= col_delta
        return col
    
    
    pass # end of class

# ----------------------------------------------------------------------

gobject.type_register(CellRendererTrace)

# To load style for the stripe renderer
#print "Parsing style", os.path.join(src.utils.globals.STYLES_DIR, "styles.rc")
#gtk.rc_parse(os.path.join(src.utils.globals.STYLES_DIR, "styles.rc"))
