# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================

import hasdel_tools.src.parse_definition_via_observable as parse_definition_via_observable

from slim_compiler.src import testexpr
from property_management.src import errors
from slim_compiler.src.dataStructure.nominalbehaviour import Bool, Int, Real, \
    IntRange, Enum

import property_management.src.api as api

import slim_compiler.src.smv.support as smvsupport
import slim_compiler.src.smv.api as smv_api

import os.path
import types
from gtk import ListStore, TreeStore, SORT_ASCENDING
import logging

from gtkmvc import Model
from gtkmvc import observable

import options
from utils import debug_info 
from utils.misc import relpath
import utils.log

from invokers import slim, nusmv
from slim_compiler.src.dataStructure.nominalbehaviour import ComponentImplementation
from slim_compiler.src.dataStructure.compile import CompileInformation


        
def get_errors_in_atomic_proposition_with_templates(atomic_proposition, slim_model):
    """Validates the specified atomic proposition (in SLIM syntax) POSSIBLY
    CONTAINING IMPLEMENTATION NAMES with respect to the SLIM model with the
    specified root component implementation and returns a list containing the
    detected errors.
    This is done in three steps:
    1.) make sure only one component implementation name appears in the AP
    2.) make sure there is at least one instance of this component implementation
        that is reachable from the root
    3.) instantiate the name with a path to a component and check this using
        get_errors_in_atomic_proposition (for every possible implementation
        instance)
        
    @param   atomic_proposition:  The atomic proposition in SLIM format.
    @type    atomic_proposition:  string
    @param   slim_model:  The slim model.
    @type    slim_model:  object

    @return: A list containing the errors in the atomic proposition. This list
             will be empty if the atomic proposition is valid (not necessarily
             true) with respect to the SLIM model.
    @rtype:  A list of errors.InvalidAtomicPropositionError objects. 
    """
    
    # 1st: generate a list of strings of the form "package::component.impl"
    # where "package::" is omitted if package = "__default__"
    impl_names = []
        
    for package in slim_model.compile_info.packages.values():
        for subpackage in [package.private_subpackage, package.public_subpackage]:
            if subpackage.package_name is "__default__":
                impl_names.extend(subpackage.implementations.keys())
            else:
                impl_names.extend([subpackage.package_name + "::" + impl_name
                                   for impl_name
                                   in subpackage.implementations.keys()])
        
    # count how many component implementations the AP refers to
    # and store their names
        
    found_impl_names = []
    for impl_name in impl_names:
        # We have to search for the implementation name followed by a dot
        # in the atomic proposition string. Otherwise substrings might be
        # found.
        if (-1 != atomic_proposition.find(impl_name + ".")):
            found_impl_names.append(impl_name)
        if len(found_impl_names) > 1:
            break
        
    if len(found_impl_names) > 1:
        return ["You may refer to at most one component implementation in " +
                "an atomic proposition"], None
        
    root_implementation = slim_model.get_extended_root_impl()
    if len(found_impl_names) == 1:
        # 2nd: find all instances of the component implementation that is
        # reachable from root. Then check the AP for these instances
        # NOTE: Because of instance based fault injections we have to check
        # ALL reachable instances
        paths = _get_instances_of(found_impl_names[0], slim_model)
        if not paths:
            return ["There is no instance of the specified component " +
                    "implementation which is reachable from the root component."], found_impl_names[0]
            
        assert len(paths) > 0, paths
        for path in paths:
            instantiated_atomic_proposition = \
                atomic_proposition.replace(found_impl_names[0]+".", path)
            errors = api.get_errors_in_atomic_proposition(
                instantiated_atomic_proposition,
                root_implementation)
            if len(errors) > 0:
                return errors, found_impl_names[0]
        return [], found_impl_names[0]
    
    else: # There are no templates, directly check syntax
    
        #LB## This is an ack (related to ticket #100). It gives an error
        ##### if the first char of atomic proposition is "."
        if atomic_proposition[0] == ".":
            atomic_proposition = atomic_proposition[1:]

        return api.get_errors_in_atomic_proposition(atomic_proposition,
                                                    root_implementation), None

pass # end of get_errors_in_atomic_proposition_with_templates
# -----------------------------------------------------------------------

        
def get_errors_in_atomic_proposition_with_templates_observable(atomic_proposition, slim_model):
    """Validates the specified atomic proposition (in SLIM syntax) POSSIBLY
    CONTAINING IMPLEMENTATION NAMES with respect to the SLIM model with the
    specified root component implementation and returns a list containing the
    detected errors.
    This is done in three steps:
    1.) make sure only one component implementation name appears in the AP
    2.) make sure there is at least one instance of this component implementation
        that is reachable from the root
    3.) instantiate the name with a path to a component and check this using
        get_errors_in_atomic_proposition (for every possible implementation
        instance)
        
    @param   atomic_proposition:  The atomic proposition in SLIM format.
    @type    atomic_proposition:  string
    @param   slim_model:  The slim model.
    @type    slim_model:  object

    @return: A list containing the errors in the atomic proposition. This list
             will be empty if the atomic proposition is valid (not necessarily
             true) with respect to the SLIM model.
    @rtype:  A list of errors.InvalidAtomicPropositionError objects. 
    """
    
    # 1st: generate a list of strings of the form "package::component.impl"
    # where "package::" is omitted if package = "__default__"
    impl_names = []
        
    for package in slim_model.compile_info.packages.values():
        for subpackage in [package.private_subpackage, package.public_subpackage]:
            if subpackage.package_name is "__default__":
                impl_names.extend(subpackage.implementations.keys())
            else:
                impl_names.extend([subpackage.package_name + "::" + impl_name
                                   for impl_name
                                   in subpackage.implementations.keys()])
        
    # count how many component implementations the AP refers to
    # and store their names
        
    found_impl_names = []
    for impl_name in impl_names:
        # We have to search for the implementation name followed by a dot
        # in the atomic proposition string. Otherwise substrings might be
        # found.
        if (-1 != atomic_proposition.find(impl_name + ".")):
            found_impl_names.append(impl_name)
        if len(found_impl_names) > 1:
            break
        
    if len(found_impl_names) > 1:
        return ["You may refer to at most one component implementation in " +
                "an atomic proposition"], None
        
    root_implementation = slim_model.get_extended_root_impl()
    if len(found_impl_names) == 1:
        # 2nd: find all instances of the component implementation that is
        # reachable from root. Then check the AP for these instances
        # NOTE: Because of instance based fault injections we have to check
        # ALL reachable instances
        paths = _get_instances_of(found_impl_names[0], slim_model)
        if not paths:
            return ["There is no instance of the specified component " +
                    "implementation which is reachable from the root component."], found_impl_names[0]
            
        assert len(paths) > 0, paths
        for path in paths:
            instantiated_atomic_proposition = \
                atomic_proposition.replace(found_impl_names[0]+".", path)
            errors = _get_errors_in_atomic_proposition(instantiated_atomic_proposition,
                                                       root_implementation)
            if len(errors) > 0:
                return errors, found_impl_names[0]
        return [], found_impl_names[0]
    
    else: # There are no templates, directly check syntax
        errors = _get_errors_in_atomic_proposition(atomic_proposition,
                                                   root_implementation), None
        return errors


pass # end of get_errors_in_atomic_proposition_with_templates_observable
# -----------------------------------------------------------------------

    
def _get_errors_in_atomic_proposition(atomic_proposition, root_implementation):
    """Validates the specified atomic proposition (in SLIM syntax) POSSIBLY
    CONTAINING IMPLEMENTATION NAMES with respect to the SLIM model with the
    specified root component implementation and returns a list containing the
    detected errors.
    
    @param   atomic_proposition:  The atomic proposition in SLIM format.
    @type    atomic_proposition:  string

    @return: error_list: A list containing the errors in the atomic proposition. This list
             will be empty if the atomic proposition is valid (not necessarily
             true) with respect to the SLIM model.
    @rtype:  error_list: A list of errors.InvalidAtomicPropositionError objects. 
    """
    error_list = []
    try:
        atomic_proposition_tree = testexpr.parseExpression(atomic_proposition)
        parse_definition_via_observable.check_definition_via_observable_tree(atomic_proposition_tree,
                                                                             root_implementation)
    except testexpr.InvalidExpressionException, e:
        error_list.append(errors.InvalidAtomicPropositionError(str(e)))
            
    return error_list

pass # end of _get_errors_in_atomic_proposition
# -----------------------------------------------------------------------


def _get_instances_of(impl_name, slim_model):
    """Returns a list of qualified subcomponent names which are instances
    of a given subcomponent implementation.
    @param   impl_name:  The name of the Implementation.
    @type    impl_name:  string
    @param   slim_model:  The slim model.
    @type    slim_model:  object

    @return: A list containing the errors in the atomic proposition. This list
             will be empty if the atomic proposition is valid (not necessarily
             true) with respect to the SLIM model.
    @rtype:  A list of errors.InvalidAtomicPropositionError objects. 
    """
    root = slim_model.root_impl
    paths = []
    if root:
        # The queue is a list of pairs
        # ("path" , ComponentImplementation object)
        queue = [("",root)]
    
        while queue != []:
            current_impl = queue[0][1]
            if current_impl.name == impl_name:
                paths.append(queue[0][0])
            #add the subcomponents of the current subcomponent to the queue
            for id in current_impl.non_data_subcomponents.keys():
                path = queue[0][0] + str(id) + "."
                subcmp = current_impl.non_data_subcomponents[id].get_referenced_component_implementation()
                queue.append((path, subcmp))
            queue.pop(0)
            
    return paths

pass # end of _get_instances_of
# -----------------------------------------------------------------------


def check_value(slim_model, port, observable_name, text, expected_values=[]):
    """Checks if the value of port is correct or not.
    @param   slim_model:  The slim model.
    @type    slim_model:  object
    @param   port: The port.
    @type    port:  The Data Port
    @param   observable_name: The observable_name of the Port.
    @type    observable_name:  string
    @param   text: The text of the Port.
    @type    text:  string
    @param   expected_values: The list of expected_values of the  Port.
    @type    expected_values:  list

    @return  result: True if it is correct, else False.
    @rtype   result: boolean
    """
    if len(text) == 0:
        return (False, ["The value length must be at least 1"])
    else:
        try:
            text.decode('utf_8')
        except UnicodeDecodeError:
            # It was not a ascii-encoded unicode string
            return (False, ["Please use UTF-8 characters only"])

        else:
            error_list = []
            try:
                root_implementation = slim_model.get_extended_root_impl()
                atomic_proposition_tree = testexpr.parseExpression(
                    observable_name + " = " + text.decode('utf_8'))
                parse_definition_via_observable.check_definition_via_observable_tree(
                    atomic_proposition_tree, root_implementation)
            except testexpr.InvalidExpressionException, e:
                error_list.append(errors.InvalidAtomicPropositionError(str(e)))

            if len(error_list) > 0:
                return (False, ["\"" + text.decode('utf_8') + "\" is not valid value for \"" +
                                observable_name + "\""])

            else:
                for n in expected_values:
                    if n[0] == text.decode('utf_8'):
                        return (False, ["\"" + text.decode('utf_8') + "\" is already an expected values."])

                if isinstance(port.data_type, IntRange):
                    if not port.data_type.value_is_valid(int(text.decode('utf_8'))):
                        return (False, ["\"" + text.decode('utf_8') + "\" is out of \"" + 
                                        observable_name + "\" range"])
                    
                return (True, None)

pass # end of check_value
# -----------------------------------------------------------------------
