# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by GMV Aerospace and Defence  
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================

from hasdel_tools.src.errors import PhaseNotExistError, \
    OpModeNotExistError, ExpressionNotCorrectError

import hasdel_tools.src.functions as functions


def load_phase_op_mode_combinations_list(slim_model, op_modes_list, src, dest):
    """Loads the list of phase-op_mode Combinations
           @param   slim_model:  The slim model.
	   @type    slim_model:  object
           @param   op_modes_list:  The list of OpModes
	   @type    op_modes_list:  list
           @param   src:  The src list of Configurations.
	   @type    src:  list
           @param   dest:  The dest list of Configurations.
	   @type    dest:  list
          
           @return   error_list:  The list of errors.
	   @rtype    error_list:  list
           @return   num_added:  The number of phase-op_mode Combinations added to the dest list.
	   @rtype    num_added:  int
           @return   num_contains:  The number of phase-op_mode Combinations in the dest list.
	   @rtype    num_contains:  int 
           @return   num_def:  The number of phase-op_mode Combinations with definition in the dest list.
	   @rtype    num_def:  int 
           @return   num_ass:  The number of phase-op_mode Combinations with association in the dest list.
	   @rtype    num_ass:  int 
           @return   num_changed:  The number of phase-op_mode Combinations changed in the dest list.
	   @rtype    num_changed:  int 
    """

    num_added = 0
    num_contains = 0
    num_def = 0
    num_ass = 0
    num_changed = 0
    error_list = []

    for phase_op_mode_combination in src:
        try:
            num_contains += 1
            phase_name = phase_op_mode_combination.phase.name
            op_mode_name = phase_op_mode_combination.op_mode.name
            definition = phase_op_mode_combination.definition

            iter = dest.get_iter_first()
            while iter is not None:
                phase_op_mode_comb = dest.get_value(iter, 0)
                if phase_op_mode_comb.phase.name == phase_name:
                    iter_child = dest.iter_children(iter)
                    while iter_child is not None:
                        phase_op_mode_comb = dest.get_value(iter_child, 0)
                        if phase_op_mode_comb.op_mode.name == op_mode_name:
                            if definition != phase_op_mode_comb.get_definition():
                                if len(definition) == 0:
                                    if len(phase_op_mode_comb.get_definition()) > 0:
                                        num_def += -1
                                    phase_op_mode_comb.set_definition(definition)
                                    num_changed += 1
                                else:
                                    error_lst = functions.get_errors_in_atomic_proposition_with_templates_observable(definition, slim_model)[0]

                                    if not error_lst:
                                        if len(phase_op_mode_comb.get_definition()) == 0:
                                            num_def += 1
                                        phase_op_mode_comb.set_definition(definition)
                                        num_changed += 1
                                    else:
                                        raise ExpressionNotCorrectError(definition)

                                    pass # end of else of if len(definition)
                                pass # end of if definition != phase_op_mode_comb.get_definition()
                            break # end of if phase_op_mode_comb.op_mode.name == op_mode_name
                        else:
                            iter_child = dest.iter_next(iter_child)
                            pass # end of else if phase_op_mode_comb.op_mode.name == op_mode_name
                        pass # end of while iter_child is not None
                    
                    if iter_child is None:
                        for opm in op_modes_list:
                            if op_mode_name == opm[0].name:
                                can_add = False
                                if len(definition) == 0:
                                    can_add = True
                                else:
                                    error_lst = functions.get_errors_in_atomic_proposition_with_templates_observable(definition, slim_model)[0]
                                    if not error_lst:
                                        can_add = True
                                        num_def += 1
                                    else:
                                        raise ExpressionNotCorrectError(definition)

                                if can_add:
                                    iter_child = dest.iter_children(iter)
                                    while iter_child is not None:
                                        phase_op_mode_comb = dest.get_value(iter_child, 0)
                                        if phase_op_mode_comb.op_mode.name > op_mode_name:
                                            break
                                        iter_child = dest.iter_next(iter_child)
                                    insert_iter = dest.insert_before(iter, iter_child)
                                    dest.set_value(
                                        insert_iter, 0, PhaseOpModeCombination(phase_op_mode_comb.phase, opm[0], definition))
                                    num_ass += 1
                                    num_added += 1
                                break
                        else:
                            raise OpModeNotExistError(op_mode_name)
                    break
                else:
                    iter = dest.iter_next(iter)
                pass # end of external while

            if iter is None:
                raise PhaseNotExistError(phase_name)
            pass # end of try

        except Exception, e:
            error_list.append(e)
        pass # end of for

    return error_list, num_added, num_changed, num_ass, num_def, num_contains


pass # end of load_phase_op_mode_combinations_list
# -----------------------------------------------------------------------


class PhaseOpModeCombination():
    """Represents a PhaseOpModeCombination"""

    def __init__(self, phase, op_mode=None, definition=None):
        """Constructor Method"""
        self.phase = phase
        self.op_mode = op_mode

        if definition is None:
            self.definition = ""
        else:
            self.definition = definition
        return


    def get_phase(self):
        """Gets the phase of the phase-op_mode Combination.
           @return   self.phase:  The phase of the phase-op_mode Combination.
	   @rtype    self.phase:  string
        """
        return self.phase


    def get_op_mode(self):
        """Gets the op_mode of the phase-op_mode Combination.
           @return   self.op_mode:  The op_mode of the phase-op_mode Combination.
	   @rtype    self.op_mode:  string
        """
        return self.op_mode


    def get_definition(self):
        """Gets the definition of the phase-op_mode Combination.
           @return   self.definition:  The definition of the phase-op_mode Combination.
	   @rtype    self.definition:  string
        """
        return self.definition


    def set_definition(self, definition):
        """Sets the definition of the phase-op_mode Combination.
           @param   definition:  The definition of the phase-op_mode Combination.
	   @type    definition:  string
        """
        self.definition = definition
        return


    def __str__(self):
        """Converts all PhaseOpModeCombination attributes to string format.
           @return   self.phase:  The phase of the phase-op_mode Combination in string format.
	   @rtype    self.phase:  string
           @return   self.op_mode:  The op_mode of the phase-op_mode Combination in string format.
	   @rtype    self.op_mode:  string
           @return   self.definition:  The definition of the phase-op_mode Combination in string format.
	   @rtype    self.definition:  string
        """
        return (str(self.phase), str(self.op_mode), self.definition)

pass # end of PhaseOpModeCombination class
# ------------------------------------------------------------------------------
