# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by GMV Aerospace and Defence  
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================

import gtk
import gobject

from hasdel_tools.src.errors import InvalidNameError, \
    NameAlreadyExistsError


def search_phase(phases_list, phase_name):
    """Searches a phase in the phases_list with the key phase_name.
           @param   phases_list:  The list of Phases.
	   @type    phases_list:  list
           @param   phase_name:  The name of the Phase.
	   @type    phase_name:  string
           
           @return   phase[0]:  The found Phase or None.
	   @rtype    phase[0]:  Phase or None
    """
    for phase in phases_list:
        if phase[0].get_name() == phase_name:
            return phase[0]
    return None

pass # end of search_phase
# -----------------------------------------------------------------------


def search_op_mode(op_modes_list, op_mode_name):
    """Searches a op_mode in the op_modes_list with the key op_mode_name.
           @param   op_modes_list:  The list of OpModes.
	   @type    op_modes_list:  list
           @param   op_mode_name:  The name of the OpMode.
	   @type    op_mode_name:  string
           
           @return   op_mode[0]:  The found OpMode or None.
	   @rtype    op_mode[0]:  OpMode or None
    """
    for op_mode in op_modes_list:
        if op_mode[0].get_name() == op_mode_name:
            return op_mode[0]
    return None

pass # end of search_op_mode
# -----------------------------------------------------------------------


def load_phases_list(src, dest, existing_names):
    """Loads the list of Phases.
           @param   src:  The src list of Phases.
	   @type    src:  list
           @param   dest:  The dest list of Phases.
	   @type    dest:  list
           @param   existing_names:  The check of existing names
	   @type    existing_names:  string

           @return   error_list:  The list of errors.
	   @rtype    error_list:  list
           @return   num_added:  The number of phases added to the dest list.
	   @rtype    num_added:  int
           @return   num_contains:  The number of phases in the dest list.
	   @rtype    num_contains:  int 
    """
    num_added = 0
    num_contains = 0
    error_list = []

    for phase in src:
        try:
            phase_name = phase.name
            num_contains += 1
            if (phase_name.lower() == "none") or (phase_name == ""):
                raise InvalidNameError(phase_name)
            elif phase_name in existing_names:
                raise NameAlreadyExistsError(phase_name)
            else:
                dest.append([Phase(phase_name)])
                existing_names.append(phase_name)
                num_added += 1
        except Exception, e:
            error_list.append(e)

        pass # end of for

    return error_list, num_added, num_contains

pass # end of load_phases_list
# -----------------------------------------------------------------------


def load_op_modes_list(src, dest, existing_names):
    """Loads the list of OpModes.
           @param   src:  The src list of OpModes.
	   @type    src:  list
           @param   dest:  The dest list of OpModes.
	   @type    dest:  list
           @param   existing_names:  The check of existing names
	   @type    existing_names:  string

           @return   error_list:  The list of errors.
	   @rtype    error_list:  list
           @return   num_added:  The number of op_modes added to the dest list.
	   @rtype    num_added:  int
           @return   num_contains:  The number of op_modes in the dest list.
	   @rtype    num_contains:  int 
    """
    num_added = 0
    num_contains = 0
    error_list = []

    for op_mode in src:
        try:
            op_mode_name = op_mode.name
            num_contains += 1
            if (op_mode_name.lower() == "none") or (op_mode_name == ""):
                raise InvalidNameError(op_mode_name)
            elif op_mode_name in existing_names:
                raise NameAlreadyExistsError(op_mode_name)
            else:
                dest.append([OpMode(op_mode_name)])
                existing_names.append(op_mode_name)
                num_added += 1
        except Exception, e:
            error_list.append(e)

        pass # end of for

    return error_list, num_added, num_contains

pass # end of load_op_modes_list
# -----------------------------------------------------------------------


class PhaseOpModeListStore(gtk.ListStore):
    """Represents a list of Phases and OpModes"""

    def __init__(self):
        """Constructor Method"""
        gtk.ListStore.__init__(self, gobject.TYPE_PYOBJECT)
        return


    def sort(self):
        """Sorts the phases and op_modes of the list"""
        sup = len(self) - 1
        while sup > 0:
            prev = self.get_value(self[0].iter, 0)
            prev_row = self[0]
            for i, row in enumerate(self):
                curr = self.get_value(row.iter, 0)
                if i > 0:
                    if curr.get_name() < prev.get_name():
                        tmp = prev
                        self.set_value(prev_row.iter, 0, curr)
                        self.set_value(row.iter, 0, tmp)
                    else:
                        prev = curr
                prev_row = row
                if i > sup:
                    break
                pass # end of for
            sup += -1

        return
                
pass # end of PhaseOpModeListStore class
# ------------------------------------------------------------------------------


class Phase():
    """Represents a Phase"""

    def __init__(self, name):
        """Constructor Method"""
        self.name = name
        return


    def get_name(self):
        """Gets the name of the Phase.
           @return   self.name:  The name of the Phase.
	   @rtype    self.name:  string
        """
        return self.name


    def set_name(self, name):
        """Sets the name of the Phase.
           @param   name:  The name of the Phase.
	   @type    name:  string
        """
        self.name = name
        return


    def __str__(self):
        """Converts the Phase to string format.
           @return   self.name:  The name of the Phase in string format.
	   @rtype    self.name:  string
        """
        return self.name

pass # end of Phase class
# ------------------------------------------------------------------------------


class OpMode():
    """Represents a Phase"""

    def __init__(self, name):
        """Constructor Method"""
        self.name = name
        return


    def get_name(self):
        """Gets the name of the OpMode.
           @return   self.name:  The name of the OpMode.
	   @rtype    self.name:  string
        """
        return self.name


    def set_name(self, name):
        """Sets the name of the OpMode.
           @param   name:  The name of the OpMode.
	   @type    name:  string
        """
        self.name = name
        return


    def __str__(self):
        """Converts the OpMode to string format.
           @return   self.name:  The name of the OpMode in string format.
	   @rtype    self.name:  string
        """
        return self.name

pass # end of OpMode class
# ------------------------------------------------------------------------------
