# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================

from tfpg_errors import FMNotCorrectError, TimeNotCorrectError, NodeNotInEdge, \
FMNodeIsNotDestNode, NodeLoopError, NotExistingModes, NotModesInTheEdge, InvalidModeName

from decimal import *
        
class Tfpg():
    """Represents a TFPG"""


    def __init__(self):
        """Constructor Method"""
        self.nodes_list = []
        self.edges_list = []
        self.modes_list = []
        self.name = ""
        self.errors_list = []
        return


    def get_nodes_list(self):
        """Gets the list of the nodes of the TFPG.
            @return   self.nodes_list:  The list of the nodes of the TFPG.
            @rtype    self.nodes_list:  list
        """
        return self.nodes_list


    def get_fm_nodes(self):
        """Gets the list of the failure mode nodes of the TFPG.
            @return   nodes_list:  The list of the failure mode nodes of the TFPG.
            @rtype    nodes_list:  list
        """
        return [node for node in self.nodes_list if node.get_type() == NodeType.FM]


    def get_or_nodes(self):
        """Gets the list of the OR discrepancy nodes of the TFPG.
            @return   nodes_list:  The list of the OR discrepancy nodes of the TFPG.
            @rtype    nodes_list:  list
        """
        return [node for node in self.nodes_list if node.get_type() == NodeType.OR]


    def get_and_nodes(self):
        """Gets the list of the AND discrepancy nodes of the TFPG.
            @return   nodes_list:  The list of the AND discrepancy nodes of the TFPG.
            @rtype    nodes_list:  list
        """
        return [node for node in self.nodes_list if node.get_type() == NodeType.AND]


    def get_monitored_discrepancies(self):
        """Returns the list of monitored discrepancies, i.e. the list of monitored AND or OR nodes.
            @return   nodes_list:  The list of monitored discrepancies of the TFPG.
            @rtype    nodes_list:  list
        """
        return [node for node in self.nodes_list if (node.get_type() == NodeType.AND or 
                node.get_type() == NodeType.OR) and node.is_monitored() is True]


    def get_unmonitored_discrepancies(self):
        """Returns the list of unmonitored discrepancies, i.e. the list of unmonitored AND or OR nodes.
            @return   nodes_list:  The list of unmonitored discrepancies of the TFPG.
            @rtype    nodes_list:  list
        """
        return [node for node in self.nodes_list if (node.get_type() == NodeType.AND or 
                node.get_type() == NodeType.OR) and node.is_monitored() is False]


    def get_edges_list(self):
        """Gets the list of the edges of the TFPG.
            @return   self.edges_list:  The list of the edges of the TFPG.
            @rtype    self.edges_list:  list
        """
        return self.edges_list


    def get_timings_constants(self, exclude_infinity):
        """Gets the list of timings constants of the edges of the TFPG.
            @param   exclude_infinity:  If the INFINITY (-1) time is excluded or not.
            @type    exclude_infinity:  boolean
            @return   timings:  The list of timings constants of the edges of the TFPG
            @rtype    timings:  list of float
        """
        
        
        timings = []
        
        for e in self.edges_list:
        
            getcontext().prec = 20
            tmin = Decimal(str(e.get_tMin()))
            tmax = Decimal(str(e.get_tMax()))
            
            timings.append(tmin)
            
            if tmax >= 0 or not exclude_infinity:
                timings.append(tmax)
        
        return timings


    def get_name(self):
        """Gets the name of the TFPG.
            @return   self.name:  The name of the TFPG.
            @rtype    self.name:  string
        """
        return self.name


    def set_name(self, name):
        """Sets the name of the TFPG.
            @param   name:  The name of the TFPG.
            @type    name:  string
        """
        self.name = name
        return


    def has_node_by_name(self, name):
        """Checks whether there is a node with the same name in the Tfpg.
            @param   name:  The name of the Node.
            @type    name:  string
           
            @return   Boolean:  If there is a node with the same name in the Tfpg.
            @rtype    Boolean:  boolean
        """
        for node in self.nodes_list:
            if node.get_name() == name: return True

        return False


    def get_node_by_name(self, name):
        """If exists, it returns the node having the passed parameter as name
           Notice that this function is useful while converting a TfpgDocument into
           a tfpg internal struct, as in the case of a TfpgDocument, the source node
           of an edge is represented as a string (the name of the node)
            @param  name: the name of the Node.
            @type   name: string

            @return node: The Node or None.
            @rtype  node: Node
        """
        for node in self.nodes_list:
            if node.get_name() == name: return node

        return None


    def has_edge_by_name(self, name):
        """Check whether the is an edge with the same name in the Tfpg.
            @param  name: the name of the Edge.
            @type   name: string

            @return Boolean: True in case there is an edge with the same in the Tfpg
            @rtype  Boolean: boolean
        """
        for edge in self.edges_list:
            if edge.get_name() == name: return True

        return False


    def get_edge_by_name(self, name):
        """If it exists, it is returned the edge having the passed parameter as name.
            @param  name: the name of the Edge.
            @type   name: string

            @return edge: The Edge or None.
            @rtype  edge: Edge
        """
        for edge in self.edges_list:
            if edge.get_name() == name: return edge

        return None


    def add_fm_node(self, name):
        """Adds a Failure Mode Node.
            If errors are found, they are added to the global errors_list
            and None is returned
            @param   name:  The name of the Node.
            @type    name:  string

            @return   node:  The Failure Mode Node.
            @rtype    node:  Node
        """
        ## GZ: it does not make sense to let the user specify also the isMonitored
        ##     parameter, since for Failure Modes it must always be False

        try:
            node = Node(NodeType.FM, name, False)
            self.nodes_list.append(node)
            return node

        except Exception, e:
            self.errors_list.append(e)
            return None


    def add_or_node(self, name, isMonitored):
        """Adds an OR Discrepancy Node.
            @param   name:  The name of the Node.
            @type    name:  string
            @param   isMonitored:  If the Node is monitored.
            @type    isMonitored:  bool

            @return   node:  The OR Discrepancy Node.
            @rtype    node:  Node
        """
        try:
            node = Node(NodeType.OR, name, isMonitored)
            self.nodes_list.append(node)
            return node

        except Exception, e:
            self.errors_list.append(e)
            return None


    def add_and_node(self, name, isMonitored):
        """Adds an AND Discrepancy Node.
            @param   name:  The name of the Node.
            @type    name:  string
            @param   isMonitored:  If the Node is monitored.
            @type    isMonitored:  bool

            @return   node:  The AND Discrepancy Node.
            @rtype    node:  Node
        """
        try:
            node = Node(NodeType.AND, name, isMonitored)
            self.nodes_list.append(node)
            return node

        except Exception, e:
            self.errors_list.append(e)
            return None


    def check_edge(self, edge):
        """Checks whether an edge has been properly built.
            In particular, it performs the following checks:
            - the srcNode must exist
            - the destNode must exist
            - the modes must exist
            - the destNode must not be a FM node
           
            @param   edge: the edge to be checked
            @type    edge: class Edge

            @return  errors_list: the (possibly empty) list of errors
            @rtype   errors_list: list of Exceptions
        """

        errors_list = []

        srcNode = edge.get_srcNode()
        destNode = edge.get_destNode()
        modes = edge.get_modes()
        tMin = edge.get_tMin()
        tMax = edge.get_tMax()
        name = edge.get_name()

        try:
            if not self.has_node_by_name(srcNode.get_name()):
                raise NodeNotInEdge(srcNode.name, name)

            elif not self.has_node_by_name(destNode.get_name()):
                raise NodeNotInEdge(destNode.name, name)
                      
            elif self.has_modes_list(modes):
                raise NotModesInTheEdge(name)

            elif destNode.get_type() == NodeType.FM:
                raise FMNodeIsNotDestNode(destNode.name, name)

            elif self.get_unknown_modes(modes) != []:
                raise NotExistingModes(self.get_unknown_modes(modes))

        except Exception, e:
            errors_list.append(e)

        return errors_list
        


    def add_edge(self, srcNode, tMin, tMax, modes, destNode, name):
        """Adds an Edge.
            @param   srcNode:  The source node of the Edge.
            @type    srcNode:  class Node
            @param   tMin:  The initial time of the Edge.
            @type    tMin:  float
            @param   tMax:  The final time of the Edge.
            @type    tMax:  float
            @param   modes:  The list of modes of the Edge.
            @type    modes:  list
            @param   destNode:  The destination node of the Edge.
            @type    destNode:  class Node
            @param   name:  The name of the Edge.
            @type    name:  string

            @return   node:  The Edge.
            @rtype    node:  Edge
            @return   error_list:  The list of errors.
            @rtype    error_list:  list
        """

        ## GZ: the init method of the Edge class implicitely checks that the Edge
        ##     values are valid. Additionally, other checks must be performed with
        ##     the check_edge() function

        new_edge = Edge(srcNode, tMin, tMax, modes, destNode, name)
        for err in new_edge.get_edge_errors_list():
            self.errors_list.append(err)

        error_list = self.check_edge(new_edge)

        for error in error_list:
            self.errors_list.append(error)

        if self.errors_list == []:
            self.edges_list.append(new_edge)
            return new_edge

        else: return None 


    def get_modes_list(self):
        """Gets the list of the active modes of the TFPG.
            @return   self.modes_list:  The list of the active modes of the TFPG.
            @rtype    self.modes_list:  list of strings
        """
        return self.modes_list


    def set_modes_list(self, modes_list):
        """Sets the list of the modes of the TFPG.

           A check is performed within this function; there cannot be modes 
           having the same name of a node/edge

            @param   modes_list:  The list of the modes of the TFPG.
            @type    modes_list:  list of strings
        """

        try:
            for md in modes_list:
                if self.has_node_by_name(str(md)):
                    raise InvalidModeName(str(md), isNode=True)
                elif self.has_edge_by_name(str(md)):
                    raise InvalidModeName(str(md), isNode=False)

                self.modes_list = modes_list

        except Exception, e:
            self.errors_list.append(e)

        return


    ## GZ: Not sure these function are useful, but maybe in future..
    # def remove_edge(self, del_edg):
    #     """Removes an edge.
    #        @param  del_edg: the Edge to be removed
    #        @type   del_edg: class Edge
    #     """

    #     for edg in self.get_edges_list():
    #         if edg == del_edg:
    #             self.edges_list.remove(edg)

    #     return


    # def remove_node(self, del_nd):
    #     """Removes a node.
    #        @param  del_nd: the Node to be removed
    #        @type   del_nd: class Node
    #     """

    #     for nd in self.get_nodes_list():
    #         if nd == del_nd:
    #             self.nodes_list.remove(nd)

    #     return



    def has_mode(self, mode):
        """Checks whether a mode is contained in the modes list.
            @param   mode:  The mode of the modes_list.
            @type    mode:  string
        """
        for md in self.get_modes_list():
            if str(md) == str(mode): return True
        return False

     
    def has_modes_list(self, modes):
        """Checks whether a modes is not empty.
            @param   modes:  The modes_list.
            @type    modes:  list of string
        """
        for md in modes:
            if md is None: return True
        return False


    def get_unknown_modes(self, modes):
        """Returns the list of modes which are not contained in the structural tfpg.
            For every mode contained in 'modes', check whether it is contained in 
            the list of modes belonging to the structural tfpg; the list of not-contained
            modes is returned. In case all the modes are contained in the strucutural tfpg,
            an empty list is returned.
            @param modes: the list of modes to be checked
            @type  modes: list of string
        """
        unknown_modes = []

        for md in modes:
            if self.has_mode(md) is False:
                unknown_modes.append(md)

        return unknown_modes


    def get_errors_list(self):
        """Gets the list of the errors of the TFPG.
            @return   self.errors_list:  The list of the errors of the TFPG.
            @rtype    self.errors_list:  list
        """
        if self.get_nodes_list() == []:
            self.errors_list.append(Exception("At least one node must be specified"))
        
        return self.errors_list


    def clear(self):
        """Clears the graph, deleting all its nodes and all its edges"""
        del self.nodes_list[0:len(self.nodes_list)]
        del self.edges_list[0:len(self.edges_list)]
        self.name = ""
        return


    def dump_to_concrete_syntax_file(self, out_filename):
        """Saves the current tfpg in concrete syntax format.
            @param out_filename: The name of the file in which the tfpg will be dumped.
            @type  out_filename: string
        """
        with open(out_filename, "w") as f:
            f.write("NAME %s\n" %(self.get_name()))

            for n in self.nodes_list:
                type = n.get_type()
                name = n.get_name()
                if n.is_monitored(): mon = "MONITORED"
                else: mon = ""
                f.write("%s %s %s\n" %(type, name, mon))

            if self.get_modes_list() != []:
                f.write("MODES %s\n" %(",".join(self.get_modes_list())))

            for e in self.edges_list:
                if float(e.get_tMax()) == -1.0:
                    max_val = "INFINITY"
                else:
                    max_val = e.get_tMax()
                f.write("EDGE %s %s %s %s %s (%s)\n" %(
                            e.get_name(),
                            e.get_srcNode().get_name(),
                            e.get_destNode().get_name(),
                            e.get_tMin(),
                            max_val,
                            (",".join(e.get_modes())))
                        )
  
        return


    def dump_to_xml(self, out_filename):
        """Saves the current tfpg in xml format.
            @param out_filename: the name of the file in which the tfpg will be dumped.
            @type out_filename: string
        """
        from hasdel_tools.src.xml.tfpg import tfpg_internal_struct_to_tfpg_doc

        with open(out_filename, "w") as f:
            f.write(str(tfpg_internal_struct_to_tfpg_doc(self)))

        return
        
    
    def __eq__(self, other):
        """Builtin function to compare two tfpgs. Notice that two tfpgs are 
            considered to be equal if all their attributes are equal.
        """
        
        if (self.get_name() == other.get_name()) and (self.get_modes_list() == other.get_modes_list()):
            
            if (len(self.get_nodes_list()) == len(other.get_nodes_list())) and \
                (len(self.get_edges_list()) == len(other.get_edges_list())):

                i = 0
                while (i < len(self.get_nodes_list())):
                    if self.get_nodes_list()[i].__eq__(other.get_nodes_list()[i]): 
                        i += 1
                        pass
                        
                    else: return False
                 
                j = 0
                while (j < len(self.get_edges_list())):
                    if self.get_edges_list()[j].__eq__(other.get_edges_list()[j]): 
                        j += 1
                        pass
                    else: return False
                    
                return True
                
            else: return False
 
        return False

        

pass # end of Tfpg class
# ------------------------------------------------------------------------------


class Edge():
    """Represents a TFPG Edge"""

    def __init__(self, srcNode, tMin, tMax, modes, destNode, name):
        """Constructor Method

           This function creates an edge with the values specified as parameters.
           Notice that these values are set with the internal set_<attribute> 
           functions. Internally, these functions make some comparisons,
           e.g. the set_tMin() function checks that tMin is lower than tMax etc.
           In order to do that, tMax must containe a value, i.e. being different
           from None (otherwise python will complain about that).
           This is the reason why, before setting the values, the attributes are 
           initialized to "standard values".

        """

        ## Initialization of the attributes
        self.name = ""
        self.srcNode = Node(None, "", False)
        self.destNode = Node(None, "", False)
        self.tMin = 0.0
        self.tMax = -1.0
        self.modes = []
        self.edge_errors_list = []  #LB: added this list with all errors of Edges

        ## The constructor parameters are assigned to the internal attributes
        self.set_name(name)
        self.set_srcNode(srcNode)
        self.set_tMin(tMin)
        self.set_tMax(tMax)
        self.set_modes(modes)
        self.set_destNode(destNode)

        return


    def get_srcNode(self):
        """Gets the source node of the Edge.
            @return   self.srcNode:  The source node of the Edge.
            @rtype    self.srcNode:  class Node
        """
        return self.srcNode


    def set_srcNode(self, srcNode):
        """Sets the source node of the Edge.
            @param   srcNode:  The source node of the Edge.
            @type    srcNode: class Node
        """
        
        try:
            if srcNode.name == self.get_destNode().name:
                raise NodeLoopError(srcNode.name)
            else:
                self.srcNode = srcNode
                
        except Exception, e:
            self.edge_errors_list.append(e)

        return


    def get_tMin(self):
        """Gets the initial time of the Edge.
            @return   self.tMin:  The initial time of the Edge.
            @rtype    self.tMin:  float
        """
        return self.tMin


    def set_tMin(self, tMin):
        """Sets the initial time of the Edge.
            @param   tMin:  The initial time of the Edge.
            @type    tMin:  float
        """
        try:
            if float(tMin) < 0 or ((float(self.get_tMax()) < float(tMin)) and \
                (float(self.get_tMax()) != -1.0)):
                raise TimeNotCorrectError(self.get_name(), tMin, self.get_tMax())
            else:
                self.tMin = tMin
                
        except Exception, e:
            self.edge_errors_list.append(e)
        return


    def get_tMax(self):
        """Gets the final time of the Edge.
            @return   self.tMax:  The final time of the Edge.
            @rtype    self.tMax:  float
        """
        return self.tMax


    def set_tMax(self, tMax):
        """Sets the final time of the Edge.
            @param   tMax:  The final time of the Edge.
            @type    tMax:  float
        """
        try:
            if (float(tMax) < float(self.get_tMin())) and float(tMax) != -1.0:
                raise TimeNotCorrectError(self.get_name(), self.get_tMin(), tMax)
            else:
                self.tMax = tMax
                
        except Exception, e:
            self.edge_errors_list.append(e)
        return


    def get_modes(self):
        """Gets the list of modes of the Edge.
            @return   self.modes:  The list of modes of the Edge.
            @rtype    self.modes:  list
        """
        return self.modes


    def set_modes(self, modes):
        """Sets the list of the modes of the Edge.
            @param   modes: the list of the modes of the Edge
            @type    modes: list[string]
        """
        self.modes = modes
        return


    def get_destNode(self):
        """Gets the destination node of the Edge.
            @return   self.destNode:  The destination node of the Edge.
            @rtype    self.destNode:  class Node
        """
        return self.destNode


    def set_destNode(self, destNode):
        """Sets the destination node of the Edge.
            @param   destNode:  The destination node of the Edge.
            @type    destNode:  class Node
        """
        
        try:
            if self.get_srcNode().name == destNode.name:
                raise NodeLoopError(destNode.name)
            else:
                self.destNode = destNode
                
        except Exception, e:
            self.edge_errors_list.append(e)
        return


    def get_name(self):
        """Gets the name of the Edge.
            @return   self.name:  The name of the Edge.
            @rtype    self.name:  string
        """
        return self.name

    def set_name(self, name):
        """Sets the name of the Edge.
            @param   name:  The name of the Edge.
            @type    name:  string
        """
        self.name = name
        return


    def __eq__(self, other):
        """Builtin function to compare two edges. Notice that two edges are 
            considered to be equal if all their attributes are equal.
        """
        if self.get_name() == other.get_name() and \
           self.get_srcNode().get_name() == other.get_srcNode().get_name() and \
           self.get_destNode().get_name() == other.get_destNode().get_name() and \
           str(self.get_tMin()) == str(other.get_tMin()) and \
           str(self.get_tMax()) == str(other.get_tMax()) and \
           self.get_modes() == other.get_modes():
            return True
        else: 
            return False
        
        
    def get_edge_errors_list(self):
        """Gets the errors of the Edge.
            @return   self.edge_errors_list:  The errors of the Edge.
            @rtype    self.edge_errors_list:  list
        """
        return self.edge_errors_list
           

pass # end of Edge class
# ------------------------------------------------------------------------------


class Node():
    """Represents a TFPG Node"""

    def __init__(self, _type, name, isMonitored):
        """Constructor Method.

           This function creates an edge with the values specified as parameters.
           Notice that these values are set with the internal set_<attribute> 
           functions. Internally, these functions make some comparisons,
           e.g. the set_type() function checks that whether the attribute
           isMonitored is False etc.
           In order to do that, isMonitored must containe a value, i.e. 
           being different from None (otherwise python will complain about that).
           This is the reason why, before setting the values, the attributes are 
           initialized to "standard values".
        """
        
        ## Initialization of the attributes
        self._type = None
        self.name = ""
        self.isMonitored = False
        self.node_errors_list = []  #LB: added this list with all errors of Nodes

        ## The constructor parameters are assigned to the internal attributes
        self.set_type(_type)
        self.set_name(name)
        self.set_isMonitored(isMonitored)

        return



    def get_type(self):
        """Gets the type of the Node (FM, AND or OR).
            @return   self.type:  The type of the Node.
            @rtype    self.type:  class NodeType
        """
        return self.type


    def set_type(self, _type):
        """Sets the type of the Node.
            @param   type:  The type of the Node.
            @type    type:  class Node
        """
        try:
            if _type == NodeType.FM and self.is_monitored():
                FMNotCorrectError(self.get_name())
            else:
                self.type = _type
                
        except Exception, e:
            self.node_errors_list.append(e)
        return


    def get_name(self):
        """Gets the name of the Node.
            @return   self.name:  The name of the Node.
            @rtype    self.name:  string
        """
        return self.name


    def set_name(self, name):
        """Sets the name of the Node.
            @param   name:  The name of the Node.
            @type    name:  string
        """
        self.name = name
        return


    def is_monitored(self):
        """Says whether the Node is monitored or not.
            @return   self.isMonitored:  If the Node is monitored.
            @rtype    self.isMonitored:  boolean
        """
        return self.isMonitored


    def set_isMonitored(self, isMonitored):
        """Sets whether the Node is monitored or not.
            @param   isMonitored:  If the Node is monitored.
            @type    isMonitored:  boolean
        """
        try:
            if isMonitored and self.get_type() == NodeType.FM:
                FMNotCorrectError(self.get_name())
            else:
                self.isMonitored = isMonitored
                
        except Exception, e:
            self.node_errors_list.append(e)
        return


    def __eq__(self, other):
        """Builting function to compare two nodes. Notice that two nodes are considered
            to be equal only if all thei attributes are equal
        """
        if self.get_name() == other.get_name() and \
           self.get_type() == other.get_type() and \
           self.is_monitored() == other.is_monitored():
            return True
        else:
            return False
            
            
    def get_node_errors_list(self):
        """Gets the errors of the Node.
            @return   self.node_errors_list:  The errors of the Node.
            @rtype    self.node_errors_list:  list
        """
        return self.node_errors_list


pass # end of Node class
# ------------------------------------------------------------------------------

class NodeType():
    """Represents a TFPG Node type"""
    
    FM = "FM"
    AND = "AND"
    OR = "OR"
    
pass # end of NodeType class
# ------------------------------------------------------------------------------    
