# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by GMV Aerospace and Defence  
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================

import os.path
import copy

from lxml import etree
from slim_compiler.src.model_extension import errors


from hasdel_tools.src.xml.phases import create_phases_document, \
    _PHASES_LIST_TAG, _PHASE_TAG, PhasesDocument, Phase
from hasdel_tools.src.xml.opmodes import create_op_modes_document, \
    _OP_MODES_LIST_TAG, _OP_MODE_TAG, OpModesDocument, OpMode
from hasdel_tools.src.xml.configurations import create_configurations_document, \
    _CONFIGURATIONS_LIST_TAG, _CONFIGURATION_TAG, ConfigurationsDocument, \
    Configuration
from hasdel_tools.src.xml.phase_op_mode_combinations import create_phase_op_mode_combinations_document, \
    _PHASE_OP_MODE_COMBINATIONS_LIST_TAG, _PHASE_OP_MODE_COMBINATION_TAG, \
    PhaseOpModeCombinationsDocument, PhaseOpModeCombination



def create_mission_document(model):
    """Creates the Mission Document.
           @param   model:  The Mission model.
	   @type    model:  object
          
           @return   MissionDocument:  The document of Mission.
	   @rtype    MissionDocument:  object
    """
    return MissionDocument.create(
        create_phases_document(model.get_phases_list()), 
        create_op_modes_document(model.get_op_modes_list()),
        create_configurations_document(model.get_configurations_list()),
        create_phase_op_mode_combinations_document(model.get_phase_op_mode_combinations_tree()))

pass # end of create_mission_document
# ----------------------------------------------------------------------------  
 

def create_mission_document_smv(model, slim_root):
    """Creates the Mission Document with smV identifiers/expressions.
           @param   model:  The Mission model.
	   @type    model:  object
           @param   slim_root:  The slim root.
	   @type    slim_root:  object
          
           @return   MissionDocument:  The document of Mission.
	   @rtype    MissionDocument:  object
    """
    return MissionDocument.create(
        create_phases_document(model.get_phases_list()), 
        create_op_modes_document(model.get_op_modes_list()),
        create_configurations_document(model.get_configurations_list(), format="SMV", slim_root=slim_root),
        create_phase_op_mode_combinations_document(model.get_phase_op_mode_combinations_tree(), format="SMV", slim_root=slim_root))    

pass # end of create_mission_document_smv
# ----------------------------------------------------------------------------  
 

class MissionDocument(etree.ElementBase):
    """Represents the Mission Document"""

    @staticmethod
    def create(phases_list, op_modes_list, configurations_list, phase_op_mode_combinations_list):
        """Creates the Tfpg Document.
           @param   phases_list:  The list of phases.
	   @type    phases_list:  list
           @param   op_modes_list:  The list of op_modes.
	   @type    op_modes_list:  list
           @param   configurations_list:  The list of configurations.
	   @type    configurations_list:  list
           @param   phase_op_mode_combinations_list:  The list of phase_op_mode_combinations.
	   @type    phase_op_mode_combinations_list:  list
          
           @return   element:  The document of Mission.
	   @rtype    element:  object
        """
        element = _parser.makeelement("{" + _XML_NAMESPACE_URI + "}" +
                                      _MISSION_TAG)

        element.append(copy.deepcopy(phases_list))
        element.append(copy.deepcopy(op_modes_list))
        element.append(copy.deepcopy(configurations_list))
        element.append(copy.deepcopy(phase_op_mode_combinations_list))

        return element


    @staticmethod
    def parse(filename):
        """Parse the Mission Document.
           @param   filename:  The name of the file to parse.
	   @type    filename:  string
          
           @return   document_tree.getroot():  The document tree parsed.
	   @rtype    document_tree.getroot():  etree
        """
        error_list = []
        try:
            document_tree = etree.parse(filename, _parser)
            return document_tree.getroot(), error_list

        except etree.XMLSyntaxError, e:
            error_list.append(errors.NonWellFormedXmlError(str(e)))
            return None, error_list


    @staticmethod
    def get_syntax_errors_in_mission_document(mission_document):
        """Gets syntax errors from the Mission Document.
           @param   mission_document:  The Mission Document.
	   @type    mission_document:  object
          
           @return   error_list:  The list of syntax errors.
	   @rtype    error_list:  list
        """
        error_list = []
        try:
            _mission_schema.assertValid(mission_document)
        except etree.DocumentInvalid, e:
            error_list.append(errors.DocumentSyntaxError(str(e)))

        return error_list


    def __str__(self):
        """Converts the elements in string format"""
        return etree.tostring(self, xml_declaration=True, encoding="UTF-8",
                              pretty_print=True)


    phases_list = property(
        lambda self: self[0],
        doc="The Phases list contained in Mission.\n" +
            "@type: PhasesDocument")

    op_modes_list = property(
        lambda self: self[1],
        doc="The OpModes list contained in Mission.\n" +
            "@type: OpModesDocument")

    configurations_list = property(
        lambda self: self[2],
        doc="The Configurations list contained in Mission.\n" +
            "@type: ConfigurationsDocument")

    phase_op_mode_combinations_list = property(
        lambda self: self[3],
        doc="The Phase/Op-Mode Combinations list contained in Mission.\n" +
            "@type: PhaseOpModeCombinationsDocument")

pass # end of MissionDocument class
# ----------------------------------------------------------------------------


# ----------------------------------------------------------------------------
# Note: The following code will be executed once when this module is imported.
# ----------------------------------------------------------------------------

# Define some constants which are used in this module.
# The XML namespace that is used for mission documents:
_XML_NAMESPACE_URI = ""
# The filename of the XML Schema document that defines the structure of valid
# mission documents.
_XSD_FILENAME = os.path.join(os.path.dirname(__file__),
                             "xsd", "mission.xsd")

# Define constants for the XML tag names that are used in mission
# documents."
_MISSION_TAG = "mission"

# Initialize the parser that is used to parse and to create mission
# specification documents.
# XML comments must be discarded. Otherwise the parser would yield data
# structures cluttered with unwanted elements.
# A lookup scheme is used in order to automatically map certain XML elements
# to instances of the adequate classes which are defined in this module.
_parser = etree.XMLParser(remove_comments=True)
_lookup = etree.ElementNamespaceClassLookup()
_parser.set_element_class_lookup(_lookup)
# Associate the various XML element tags with the corresponding classes. 
# Note that all elements are in the same namespace.
_namespace = _lookup.get_namespace(_XML_NAMESPACE_URI)
_namespace[_MISSION_TAG] = MissionDocument
_namespace[_PHASES_LIST_TAG] = PhasesDocument
_namespace[_PHASE_TAG] = Phase
_namespace[_OP_MODES_LIST_TAG] = OpModesDocument
_namespace[_OP_MODE_TAG] = OpMode
_namespace[_CONFIGURATIONS_LIST_TAG] = ConfigurationsDocument
_namespace[_CONFIGURATION_TAG] = Configuration
_namespace[_PHASE_OP_MODE_COMBINATIONS_LIST_TAG] = PhaseOpModeCombinationsDocument
_namespace[_PHASE_OP_MODE_COMBINATION_TAG] = PhaseOpModeCombination

# Parse the XML Schema document that defines the structure of valid mission
# documents. This is needed for validation.
_mission_schema_tree = etree.parse(_XSD_FILENAME)
_mission_schema = etree.XMLSchema(_mission_schema_tree)

