# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by GMV Aerospace and Defence  
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================

import os.path
import copy

from lxml import etree

from hasdel_tools.src.errors import OpModeAlreadyExistsError


def create_op_modes_document(op_modes_list):
    """Creates the OpModes Document.
           @param   op_modes_list:  The list of op_modes.
	   @type    op_modes_list:  list
          
           @return   op_modes_document:  The document of op_modes.
	   @rtype    op_modes_document:  object
    """
    op_modes_document = OpModesDocument.create()

    for op_mode in op_modes_list:
        element = op_modes_document.get_by_name(str(op_mode[0]))

        if element is None:
            new_element = OpMode.create(str(op_mode[0]))
            op_modes_document.add_op_mode(new_element)
            element = op_modes_document.get_by_name(str(op_mode[0]))
            pass

        assert element is not None

    return op_modes_document

pass # end of create_op_modes_document
# ----------------------------------------------------------------------------  
 

class OpModesDocument(etree.ElementBase):
    """Represents the OpModes Document"""

    @staticmethod
    def create():
        """Creates the list of OpModes.     
           @return   list of OpModes:  The document of OpModes list.
	   @rtype    list of OpModes:  object
        """
        return _parser.makeelement("{" + _XML_NAMESPACE_URI + "}" +
                                   _OP_MODES_LIST_TAG)


    def get_by_name(self, name):
        """Gets the op mode from the name.  
           @param   name:  The name of the op mode.
	   @type    name:  string       
          
           @return   op_mode:  The op mode found.
	   @rtype    op_mode:  OpMode
        """
        for op_mode in self:
            if op_mode.name == name:
                return op_mode
        return None


    def add_op_mode(self, op_mode):
        """Adds the op_mode to the op_modes list.
           @param   op_mode:  The OpMode.
	   @type    op_mode:  OpMode
        """
        if self.get_by_name(op_mode.name):
            raise OpModeAlreadyExistsError(op_mode.name)
        else:
            self.append(copy.deepcopy(op_mode))
        return

    def __str__(self):
        """Converts the elements in string format"""
        return etree.tostring(self, xml_declaration=True, encoding="UTF-8",
                              pretty_print=True)

pass # end of OpModesDocument class
# ----------------------------------------------------------------------------


class OpMode(etree.ElementBase):
    """Represents the OpMode"""

    @staticmethod
    def create(name):
        """Creates the OpMode.
           @param   name:  The name of the OpMode.
	   @type    name:  string         
          
           @return   op_mode:  The document of OpMode.
	   @rtype    op_mode:  object
        """

        op_mode = _parser.makeelement(
            "{" + _XML_NAMESPACE_URI + "}" + _OP_MODE_TAG)
        name_element = _parser.makeelement(
            "{" + _XML_NAMESPACE_URI + "}" + _NAME_TAG)
        name_element.text = name
        op_mode.append(name_element)
        return op_mode

    name = property(
        lambda self: self[0].text.strip() if self[0].text is not None else "",
        doc="The name of the Op-Mode.\n" + 
            "@type: string")

pass # end of OpMode class
# ----------------------------------------------------------------------------


# ----------------------------------------------------------------------------
# Note: The following code will be executed once when this module is imported.
# ----------------------------------------------------------------------------

# Define some constants which are used in this module.
# The XML namespace that is used for op-modes documents:
_XML_NAMESPACE_URI = ""
# The filename of the XML Schema document that defines the structure of valid
# op-modes documents.
_XSD_FILENAME = os.path.join(os.path.dirname(__file__),
                             "xsd", "op_modes.xsd")

# Define constants for the XML tag names that are used in op-modes
# documents.
_OP_MODES_LIST_TAG = "opModesList"
_OP_MODE_TAG = "opMode"
_NAME_TAG = "name"

# Initialize the parser that is used to parse and to create op-modes
# specification documents.
# XML comments must be discarded. Otherwise the parser would yield data
# structures cluttered with unwanted elements.
# A lookup scheme is used in order to automatically map certain XML elements
# to instances of the adequate classes which are defined in this module.
_parser = etree.XMLParser(remove_comments=True)
_lookup = etree.ElementNamespaceClassLookup()
_parser.set_element_class_lookup(_lookup)
# Associate the various XML element tags with the corresponding classes. 
# Note that all elements are in the same namespace.
_namespace = _lookup.get_namespace(_XML_NAMESPACE_URI)
_namespace[_OP_MODES_LIST_TAG] = OpModesDocument
_namespace[_OP_MODE_TAG] = OpMode

# Parse the XML Schema document that defines the structure of valid op-modes
# documents. This is needed for validation.
_op_modes_schema_tree = etree.parse(_XSD_FILENAME)
_op_modes_schema = etree.XMLSchema(_op_modes_schema_tree)

