# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by GMV Aerospace and Defence  
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================

import os.path
import copy

from lxml import etree
from hasdel_tools.src.errors import PhaseOpModeCombinationAlreadyExistsError

from hasdel_tools.src.xml.phases import Phase, _PHASE_TAG
from hasdel_tools.src.xml.opmodes import OpMode, _OP_MODE_TAG
import slim_compiler.src.smv.api as smv_api
from slim_compiler.src.testexpr import parseExpression
from property_management.src.api import get_atomic_proposition_in_smv_format as get_smv_format



def create_phase_op_mode_combinations_document(phase_op_mode_combinations_tree, format="SLIM", slim_root=None):
    """Creates the PhaseOpModeCombinations Document.
           @param   phase_op_mode_combinations_tree:  The PhaseOpModeCombinations tree.
	   @type    phase_op_mode_combinations_tree:  TreeStore
           @param   format:  The format of file: SLIM or SMV
	   @type    format:  string
           @param   slim_root:  The slim root.
	   @type    slim_root:  object
          
           @return   phase_op_mode_combinations_document:  The document of PhaseOpModeCombinations.
	   @rtype    phase_op_mode_combinations_document:  object
    """

    phase_op_mode_combinations_document = PhaseOpModeCombinationsDocument.create()

    iter = phase_op_mode_combinations_tree.get_iter_first()

    while iter is not None:
        iter_child = phase_op_mode_combinations_tree.iter_children(iter)

        while iter_child is not None:
            phase_op_mode_combination = phase_op_mode_combinations_tree.get_value(iter_child, 0)
            phase = Phase.create(str(phase_op_mode_combination.get_phase()))
            op_mode = OpMode.create(str(phase_op_mode_combination.get_op_mode()))

            element = phase_op_mode_combinations_document.get_by_phase_op_mode(
                phase.name, op_mode.name)

            if element is None:
                definition = phase_op_mode_combination.get_definition()
                if format=="SMV": # Convert to SMV if needed
                    definition = get_smv_format(definition, slim_root)

                new_element = PhaseOpModeCombination.create(
                    phase, op_mode, 
                    definition) 
                phase_op_mode_combinations_document.add_phase_op_mode_combination(
                    new_element)
                element = phase_op_mode_combinations_document.get_by_phase_op_mode(
                    phase.name, op_mode.name)
                pass

            assert element is not None

            iter_child = phase_op_mode_combinations_tree.iter_next(iter_child)
            pass # end of internal while

        iter = phase_op_mode_combinations_tree.iter_next(iter)
        pass # end of external while

    return phase_op_mode_combinations_document


pass # end of create_phase_op_mode_combinations_document
# ----------------------------------------------------------------------------  
 

class PhaseOpModeCombinationsDocument(etree.ElementBase):
    """Represents the PhaseOpModeCombinations Document"""

    @staticmethod
    def create():
        """Creates the PhaseOpModeCombinations Document.     
           @return   phase_op_mode_combinations:  The document of PhaseOpModeCombinations.
	   @rtype    phase_op_mode_combinations:  object
        """
        return _parser.makeelement("{" + _XML_NAMESPACE_URI + "}" +
                                   _PHASE_OP_MODE_COMBINATIONS_LIST_TAG)


    def get_by_phase_op_mode(self, phase_name, op_mode_name):
        """Gets the phase_op_mode_combination from the phase and op_mode names.  
           @param   phase_name:  The phase_name of the phase_op_mode_combination.
	   @type    phase_name:  string 
           @param   op_mode_name:  The op_mode_name of the phase_op_mode_combination.
	   @type    op_mode_name:  string       
          
           @return   phase_op_mode_combination:  The phase_op_mode_combination found.
	   @rtype    phase_op_mode_combination:  PhaseOpModeCombination
        """
        for phase_op_mode_combination in self:
            if (phase_op_mode_combination.phase.name == phase_name) and \
                    (phase_op_mode_combination.op_mode.name == op_mode_name):
                return phase_op_mode_combination
        return None


    def add_phase_op_mode_combination(self, phase_op_mode_combination):
        """Adds the phase_op_mode_combination to the fd_table document.
           @param   phase_op_mode_combination:  The PhaseOpModeCombination.
	   @type    phase_op_mode_combination:  PhaseOpModeCombination
        """
        if self.get_by_phase_op_mode(phase_op_mode_combination.phase.name,
                                     phase_op_mode_combination.op_mode.name):
            raise PhaseOpModeCombinationAlreadyExistsError(
                phase_op_mode_combination.phase.name, 
                phase_op_mode_combination.op_mode.name)
        else:
            self.append(copy.deepcopy(phase_op_mode_combination))
        return


    def __str__(self):
        """Converts the elements in string format"""
        return etree.tostring(self, xml_declaration=True, encoding="UTF-8",
                              pretty_print=True)

pass # end of PhaseOpModeCombinationsDocument class
# ----------------------------------------------------------------------------


class PhaseOpModeCombination(etree.ElementBase):
    """Represents the PhaseOpModeCombination"""

    @staticmethod
    def create(phase, op_mode, definition):
        """Creates the FDTable Alarm.
           @param   phase:  The phase of the PhaseOpModeCombination.
	   @type    phase:  string    
           @param   op_mode:  The op_mode of the PhaseOpModeCombination.
	   @type    op_mode:  string 
           @param   definition:  The definition of the PhaseOpModeCombination.
	   @type    definition:  string         
          
           @return   phase_op_mode_combination:  The PhaseOpModeCombination.
	   @rtype    phase_op_mode_combination:  object
        """
        phase_op_mode_combination = _parser.makeelement(
            "{" + _XML_NAMESPACE_URI + "}" + _PHASE_OP_MODE_COMBINATION_TAG)

        phase_op_mode_combination.append(copy.deepcopy(phase))

        phase_op_mode_combination.append(copy.deepcopy(op_mode))

        definition_element = _parser.makeelement(
            "{" + _XML_NAMESPACE_URI + "}" + _DEFINITION_TAG)
        definition_element.text = definition
        phase_op_mode_combination.append(definition_element)
        
        return phase_op_mode_combination

    phase = property(
        lambda self: self[0],
        doc="The Phase in this Phase/Op-Mode Combination.\n" + 
            "@type: Phase")

    op_mode = property(
        lambda self: self[1],
        doc="The Op-Mode in this Phase/Op-Mode Combination.\n" + 
            "@type: OpMode")

    definition = property(
        lambda self: self[2].text.strip() if self[2].text is not None else "",
        doc="The definition via Observable of the Phase/Op-Mode Combination.\n" +
            "@type: string")

pass # end of PhaseOpModeCombination class
# ----------------------------------------------------------------------------


# ----------------------------------------------------------------------------
# Note: The following code will be executed once when this module is imported.
# ----------------------------------------------------------------------------

# Define some constants which are used in this module.
# The XML namespace that is used for Phase/Op-Mode combinations documents:
_XML_NAMESPACE_URI = ""
# The filename of the XML Schema document that defines the structure of valid
# Phase/Op-Mode combinations documents.
_XSD_FILENAME = os.path.join(os.path.dirname(__file__),
                             "xsd", "phase_op_mode_combinations.xsd")

# Define constants for the XML tag names that are used in Phase/Op-Mode combinations
# documents.
_PHASE_OP_MODE_COMBINATIONS_LIST_TAG = "phaseOpModeCombinationsList"
_PHASE_OP_MODE_COMBINATION_TAG = "phaseOpModeCombination"
_DEFINITION_TAG = "definition"

# Initialize the parser that is used to parse and to create Phase/Op-Mode combinations
# specification documents.
# XML comments must be discarded. Otherwise the parser would yield data
# structures cluttered with unwanted elements.
# A lookup scheme is used in order to automatically map certain XML elements
# to instances of the adequate classes which are defined in this module.
_parser = etree.XMLParser(remove_comments=True)
_lookup = etree.ElementNamespaceClassLookup()
_parser.set_element_class_lookup(_lookup)
# Associate the various XML element tags with the corresponding classes. 
# Note that all elements are in the same namespace.
_namespace = _lookup.get_namespace(_XML_NAMESPACE_URI)
_namespace[_PHASE_OP_MODE_COMBINATIONS_LIST_TAG] = PhaseOpModeCombinationsDocument
_namespace[_PHASE_OP_MODE_COMBINATION_TAG] = PhaseOpModeCombination
_namespace[_PHASE_TAG] = Phase
_namespace[_OP_MODE_TAG] = OpMode

# Parse the XML Schema document that defines the structure of valid Phase/Op-Mode combinations
# documents. This is needed for validation.
_phase_op_mode_combinations_schema_tree = etree.parse(_XSD_FILENAME)
_phase_op_mode_combinations_schema = etree.XMLSchema(_phase_op_mode_combinations_schema_tree)

