# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


# invokers for mrmc


import os.path
from lxml import etree

import utils.process
import options
import re
import tempfile
from utils import mrmc_process
import subprocess
from slim_compiler.src import deprecated


def launch_sigref_to_mrmc(xml_input_filename):
    """Invokes binary sigref2mrmc using the specified XML file as input.
    
    @param   xml_input_filename:
                The XML document representing the configuration LTS (of a SLIM
                model), including information about the atomic propositions of
                the properties to be checked and a mapping of transition
                labels to properties (or rates). This document uses an
                extended version of the Sigref input syntax.
    @type    xml_input_filename:
                string
    @raises  IOError: If the specified input file does not exist.
    """
    # The specified input file must exist.
    if not os.path.isfile(xml_input_filename):
        raise IOError("File not found: " + xml_input_filename)
    
    # The following code is a hack. It was used to avoid problems with
    # sigref2mrmc in case of XML files containing long sequences of irrelevant
    # blanks. This hack has been disabled because it might lead to errors if
    # the nesting depth of the XML document is very large. This is due to a
    # hard-coded limit in libxml2, which is used by the lxml package.
    # To avoid the problem with blanks, the NuSMV-to-Sigref translator simply
    # does not output irrelevant blanks.
    #
    # -- Start of obsolete hack: ---------------------------------------------
    ## Parse the XML input file, remove all whitespace elements while doing
    ## this and write the result back to the XML input file. This is a
    ## workaround for a problem with libxml2, which is used in sigref2mrmc, to
    ## parse the input file: The input file is parsed using an option that
    ## makes libxml2 ignore blanks which would otherwise result in text nodes
    ## in the XML tree. But it seems that if there are long sequences of blanks
    ## (about 29 in a row), text nodes will be created nevertheless. This leads
    ## to errors because in the input format no text content is allowed.
    #parser = etree.XMLParser(remove_blank_text = True)
    #xml_tree = etree.parse(xml_input_filename, parser)
    ## Important: Don't use pretty printing when writing the XML document
    ## because this would introduce whitespace again.
    #xml_tree.write(xml_input_filename,
    #               xml_declaration = True,
    #               encoding = "UTF-8")
    # -- End of hack. --------------------------------------------------------
    
    # Get the basename of the input file. It will be used as the basename for
    # the output files.
    basename = os.path.splitext(xml_input_filename)[0]
    
    # Invoke sigref2mrmc.
    executable_path = utils.process.get_exec_name(options.SIGREF2MRMC)
    args = [executable_path, xml_input_filename, basename + ".tra",
            basename + ".lab", basename + ".init", "0"]
    
    return utils.process.start_process_and_get_results(
                args, None, options.DISPLAY_CALL_OUTPUT)
            
def launch_mrmc(tra_file, lab_file, formulae):
    """Invokes binary mrmc-1.4 using the specified filenames as input.
    
    @param   tra_file:
                The filename of the .tra file passed to MRMC.
    @param   lab_file:
                The filename of the .lab file passed to MRMC.
    @param   formulae:
                The formulae which should be checked by MRMC.
    @type    tra_file:
                string
    @type    lab_file:
                string
    @type    formulae:
                non-empty list of strings that represent valid probabilistic
                formulae
    @raises  IOError: If one of the specified input files does not exist.
    """
    # The specified input files must exist.
    if not os.path.isfile(tra_file):
        raise IOError("File not found: " + tra_file)         
    if not os.path.isfile(lab_file):
        raise IOError("File not found: " + lab_file)
        
    # Select the right MRMC binary
    executable_path = utils.process.get_exec_name(options.MRMC)
        
    input_script = ""
    for formula in formulae:
        input_script += formula + "\r\n"
    input_script += "quit\r\n"
        
    args = [executable_path, "ctmc", tra_file, lab_file]
    res = utils.process.start_process_and_get_results(args, input_script, False)
    return res

def start_mrmc(tra_filename, lab_filename, formula_string, stepwise=False, granularity=100, args=[]):
    # constructs a new instance of mrmc logging file
    fdlog, logfname = tempfile.mkstemp(suffix='.mrmcout', 
                                    prefix=os.path.splitext(tra_filename)[0], 
                                    dir=None, text=True)
    # construct the commands file
    fdcmd, cmdfname = tempfile.mkstemp(suffix='.mrmccmds', prefix=os.path.splitext(tra_filename)[0], dir=None, text=True)

    if stepwise:
        os.write(fdcmd, "set granularity " + str(granularity) + "\n")
        args.append("-draw")
    os.write(fdcmd, formula_string + "\n")
    os.write(fdcmd, "quit\n")
    os.lseek(fdcmd, 0, 0)
    
    # start up MRMC 
    process = mrmc_process.MRMCProcess(tra_filename, lab_filename, args=args,
                            stdin=fdcmd,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.STDOUT,
                            dumpers=[os.fdopen(fdlog, "w")])

    return process

def run(tra_filename, lab_filename, formula_string, stepwise=False, granularity=100):
    """Start the MRMC process and wait for it to finish.
    Returns exit code and stdout/stderr combined"""
    process = start_mrmc(tra_filename, lab_filename, formula_string, stepwise, granularity)

    process.wait()

    return (process.returncode, (process.get_stdout_text(), process.get_stderr_text()))

def extract_probabilities(output, initial_states_file):
    # retrieve init states
    init = map(int, open(initial_states_file).readlines()[1:])
    assert(len(init) > 0)
       
    concated = ''.join(output)
    r1 = re.compile(r'\$RESULT\:\s+\((.+)\)')
    results = r1.findall(concated)
        
    assert len(results) == 1
    prob_vector = map(float, results[0].split(','))
    return prob_vector[init[0] - 1]

def extract_min_max_probabilities(mrmc_output, initial_states_file):
    def _get_min_probability(result_line, inits):
        prob_vector = map(float, result_line.split(','))
        lowest = 1
        for state in inits:
            lowest = min(prob_vector[state - 1], lowest)
        return lowest            
    
    def _get_max_probability(result_line, inits):
        prob_vector = map(float, result_line.split(','))
        highest = 0
        for state in inits:
            highest = max(prob_vector[state - 1], highest)
        return highest
    
    assert os.path.exists(initial_states_file)

    # retrieve the probability vector
    concated = ''.join(mrmc_output)
    r1 = re.compile(r'\$RESULT\:\s+\((.+)\)')
    results = r1.findall(concated)
    assert len(results) == 1
        
    # retrieve init states
    init = map(int, open(initial_states_file).readlines()[1:])
    assert(len(init) > 0)
    min_ = _get_min_probability(results[0], init)
    max_ = _get_max_probability(results[0], init)
    if min_ == max_:
        return min_ 
    else:
        return (min_, max_)

def _parse_mrmc_cdf_data(stdout, init_filename):
    init_states = map(int, open(init_filename).readlines()[1:])
                    
    xpoints = []
    ypoints = []
    for i in range(0, len(init_states)):
        ypoints.append([])

    lines = stdout.split("\n")
    for line in lines:
        if "# end" in line: break
        if "#" in line:
            line = line.split("#")[1]
            tokens = line.split()
            numbers = [float(i) for i in tokens[0:]]
            xpoints.append(numbers[0])
            for i in range(0, len(init_states)):
                ypoints[i].append(numbers[init_states[i]])
    return (xpoints, ypoints[0])
