# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


# This module contains all high-level functionalities concerning
# SLIM files transformations

import os.path

import antlr3
import antlr3.tree

import options
from utils import log
from utils import misc
from utils import exceptions
from slim_compiler.src.model_extension import fault_injection
import slim_compiler.src.model_extension.core as model_ext_core
from slim_compiler.src.generated.SlimWalkerIdentCheck import \
    SlimWalkerIdentCheck
from slim_compiler.src.dataStructure.derivedlexer import DerivedLexer
from slim_compiler.src.dataStructure.derivedparser import DerivedParser
import slim_compiler.src.smv.api as slim2smv_api
import slim_compiler.src.smv.support as smvsupport
import property_management.src.api as properties_api
import property_management.src.properties as properties_module

# ----------------------------------------------------------------------


from utils.log import log_icd
from slim_compiler.src.dataStructure.visitors.compileVisitor import (
    CheckVisitor)
from slim_compiler.src.dataStructure.visitors.time import (
    TimeDelayClockVisitor, CheckTimeScaleUsageVisitor)
from slim_compiler.src.dataStructure.visitors.expressionVisitor import (
    AnnotateVisitor, TypeVisitor, RangeTransformVisitor)
from slim_compiler.src.dataStructure.visitors.postProcessing import (
    LocatedInfoVisitor, WildcardTransformVisitor, TriggerListTransformVisitor)


@log_icd
def parse(cinfo, files):
    """Parses a set of files compiling them into given compile info
    structure.

    @param cinfo: CompileInformation instance
    @param files: iterable sequence of slim filenames
    @return False if an error occurs during parsing"""
    assert cinfo is not None

    prevErrors = len(cinfo.error_collector.errors)

    for filename in files:
        # sanity checks
        assert os.path.exists(filename), "File %s not found" % filename

        # let the compile_info structure know we are going to process
        # a new file
        cinfo.current_file = os.path.abspath(filename)
        cinfo.files.append(filename)

        char_stream = antlr3.ANTLRFileStream(filename)

        lexer = DerivedLexer(cinfo, char_stream)

        tokens = antlr3.CommonTokenStream(lexer)
        parser = DerivedParser(cinfo, tokens)

        r = parser.specification()

        # start decorating
        if not parser.has_parsing_errors() and not lexer.has_lexing_errors():
            nodes = antlr3.tree.CommonTreeNodeStream(r.tree)
            nodes.setTokenStream(tokens)
            SlimWalkerIdentCheck(nodes).specification(cinfo)

    _post_compile_adaptations(cinfo)

    return len(cinfo.error_collector.errors) == prevErrors


def _post_compile_adaptations(cinfo):
    # Apply some syntactic sugar
    WildcardTransformVisitor().visit(cinfo)
    TriggerListTransformVisitor().visit(cinfo)

    # Fixup locatedinformation information that may be missing
    LocatedInfoVisitor().visit(cinfo)


def check(cinfo):
    """Invokes the type checker over the given cinfo structure. No
    check is carried out if cinfo is None (may be the case when no
    fault injection is performed)"""
    if cinfo is None:
        return True

    # First check if time scales are in use
    CheckTimeScaleUsageVisitor(cinfo).visit(cinfo)

    # Perform checking
    AnnotateVisitor(cinfo).visit(cinfo)
    TypeVisitor(cinfo).visit(cinfo)
    CheckVisitor(cinfo).visit(cinfo)

    # Perform transformations to 'basic' SLIM
    TimeDelayClockVisitor(cinfo).visit(cinfo)
    RangeTransformVisitor().visit(cinfo)
    # Ticket 543: Let SMV converter handle this
    #TimeScaleTransformVisitor(cinfo).visit(cinfo)

    return len(cinfo.error_collector.errors) == 0


def emit_errors(cinfo):
    if len(cinfo.error_collector.errors) != 0:
        log.user.error("Errors: ----------------------------------"
                       "----------------------")
        log.user.error(cinfo.error_collector)
        log.user.error(str(len(cinfo.error_collector.errors)) + " error(s)")
        log.user.info("Files processed: --------------------------"
                      "---------------------")
        map(log.user.info, cinfo.files)
    else:
        log.user.info("No errors found. --------------------------"
                      "---------------------")
        log.user.info("Files processed: --------------------------"
                      "---------------------")
        map(log.user.info, cinfo.files)

    return


def do_fault_injection(cinfo, fi_fname=None):
    """Performs fault injection.

    If fi_fname is not given, a corresponding file can be specified
    in options. Returns extended compile info structure, or None if
    no injection is carried out.

    @raises model_ext_core.InvalidFaultInjectionDocumentError:
                If the fault injection document is invalid.
    """

    if fi_fname is None:
        fi_fname = options.FAULT_INJECTION_FILE_NAME

    if fi_fname:
        injection_doc = fault_injection.parse(fi_fname)
        root_implementation = get_root_impl(cinfo, options.ROOT_NAME)
        # TODO: Raise an exception if the fault injection document is invalid.
        # This can be checked using
        # fault_injection.get_errors_in_injection_document.
        # Note that model_ext_core.create_extended_model will raise an
        # model_ext_core.InvalidFaultInjectionDocumentError if the fault
        # injection document is invalid, but explicit validation would be
        # preferable.
        extended_root_implementation = model_ext_core.create_extended_model(
            root_implementation, injection_doc)

        ext_compile_info = extended_root_implementation.subpackage.package.\
            compile_information

        return extended_root_implementation.subpackage.package.\
            compile_information
    else:
        return None


def convert_to_smv(root_impl, smv_fname=None,
                   generate_time_var=options.SMV_GENERATE_TIME_VAR,
                   generate_time_fairness_alt=options.SMV_FAIR_TIMED):
    """If smv_fname is not given or None, filename is taken from
    the basename and collocated into the output directory. Returns
    a triple (smv_fname, infinite, timed).  smv_fname is the name
    of the generated smv file, infinite and timed are flags telling
    respectively whether infinite and timed domain has been found
    or not.

    Parameters generate_time_var and generate_time_fairness_alt
    control if additional constraints for timed fsm have to be added,
    and will be ignored for discrete (non-timed) FSMs.
    """
    if smv_fname is None:
        smv_fname = misc.get_out_fname(options.BASE_FILE_NAME + ".smv")

    with open(smv_fname, "w") as smv_file:
        infinite, timed = slim2smv_api.convert_to_smv(
            root_impl, smv_file,
            generate_time_var=generate_time_var,
            generate_time_fairness_alt=generate_time_fairness_alt)

    return (smv_fname, infinite, timed)


def get_root_impl(cinfo, root_name=None):
    """Returns a root implementation, taking into account users
    requests. If an error occurs, raises an exception"""
    # checks for root
    if root_name:
        root_impl = smvsupport.get_implementation_from_name(cinfo, root_name)

        if root_impl is None:
            roots = smvsupport.get_root_components(cinfo)
            root_names = ", ".join([x.name for x in roots])
            raise exceptions.UserException(
                "Requested root implementation '%s' not found.\n"
                "Available roots are: %s" % (root_name, root_names))

    else:  # root not specified, search it autonomously
        roots = smvsupport.get_root_components(cinfo)
        if len(roots) == 0:
            raise exceptions.UserException("No roots component found")

        if len(roots) > 1:
            raise exceptions.UserException(
                "There are %d roots candidates: %s\nPlease specify one" %
                (len(roots), ", ".join([x.name for x in roots])))
        root_impl = roots.pop()

    return root_impl


# TODO: This function takes the parameter root_impl but does not use it.
# Maybe the root implementation should be used to check the atomic
# propositions in the formulae document. Otherwise the parameter could be
# removed.
from utils.log import log_icd


@log_icd
def get_formulae_doc(root_impl, prop_fname):
    """Given the name of a file containing a set of formulae,
    returns an instance of
    property_management.src.formulae.FormulaeDocument that
    can be used to iterate over the contained formulas. Each entry
    of the list is a property_management.src.formulae.Formula
    instance"""
    # Return an empty formulae document if no properties document is
    # specified.
    if not prop_fname:
        return properties_api.create_formulae_document_for_properties([])

    prop_doc = properties_module.parse_properties_document(prop_fname)
    errors = properties_module.get_syntax_errors_in_document(prop_doc)

    if len(errors) > 0:
        # Create an error message by concatenating the descriptions of the
        # individual errors, separated by newline characters.
        error_strings = map(str, errors)
        error_message = ("Errors in properties document:\n" +
                         "\n".join(error_strings))
        # XXX: Errors in the properties document will be indicated as user
        # errors (invalid input), but actually this is only true if the
        # properties document was in fact specified by the user and not
        # generated by the toolset. However, it seems that this condition is
        # always satisfied when this function is used.
        raise exceptions.UserException(error_message)

    return properties_api.create_formulae_document_for_properties(prop_doc)
