# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


"""This module contains exception classes related to property management."""

class PropertyManagementError(Exception):
    """The base class for errors that are related to property management."""
    pass


class PlaceholderInstanceAlreadyExistsError(PropertyManagementError):
    """An object indicating that a placeholder instance could not be added to
    a certain pattern instance because there is already an instance for that
    placeholder.
    """
    
    def __init__(self, placeholder_id):
        """Initializes the PlaceholderAlreadyInstantiatedError object.
        
        @param   placeholder_id: The ID of the placeholder.
        @type    placeholder_id: string
        """
        self.placeholder_id = placeholder_id
        
    def __str__(self):
        """Returns a human-readable string representation of the error.
        
        @return: A human-readable error message.
        @rtype: string
        """
        return ('There is already a placeholder instance for the ' +
                'placeholder "%s".') % self.placeholder_id


class FailedToLoadPatternsDocumentError(PropertyManagementError):
    """An object indicating that a certain patterns document could not be
    loaded, for example because it could not be found or it is invalid.
    """
    pass


class NoPatternsForSpecifiedLogicError(FailedToLoadPatternsDocumentError):
    """An object indicating that there is no patterns document for a certain
    logic and thus the document could not be loaded.
    """
    
    def __init__(self, logic):
        """Initializes the NoPatternsForSpecifiedLogicError object.
        
        @param   logic: The logic for which there is no patterns document.
        @type    logic: One of the constants in constants.Logic.
        """
        self.logic = logic
        
    def __str__(self):
        """Returns a human-readable string representation of the error.
        
        @return: A human-readable error message.
        @rtype:  string
        """
        # TODO: Include the name of the logic in the error message.
        return "There are no patterns for the specified logic."


class PatternInstantiationError(PropertyManagementError):
    """The base class for all errors that might occur when instantiating
    formula patterns.
    """
    pass


class PatternNotFoundError(PatternInstantiationError):
    """An error object indicating that a pattern instantiation refers to a
    non-existent pattern.
    """
    
    def __init__(self, pattern_id):
        """Initializes the PatternNotFoundError object.
        
        @param   pattern_id: The ID of the missing pattern.
        @type    pattern_id: string
        """
        self.pattern_id = pattern_id
        
    def __str__(self):
        """Returns a human-readable string representation of the error.
        
        @return: A human-readable error message.
        @rtype:  string
        """
        return ('The property refers to a non-existent pattern. ' +
                'Pattern ID: "%s"') % self.pattern_id


class MissingPlaceholderInstantiationError(PatternInstantiationError):
    """An object indicating that a placeholder is not instantiated in a
    certain property specification.
    """
    
    def __init__(self, placeholder_id):
        """Initializes the PatternNotFoundError object.
        
        @param   placeholder_id: The ID of the placeholder that is not
                                 instantiated.
        @type    placeholder_id: string
        """
        self.placeholder_id = placeholder_id
        
    def __str__(self):
        """Returns a human-readable string representation of the error.
        
        @return: A human-readable error message.
        @rtype:  string
        """
        return ('The placeholder "%s" is not instantiated in the property ' +
                'specification.') % self.placeholder_id
        

class DuplicatePlaceholderInstantiationError(PatternInstantiationError):
    """An error object indicating that a placeholder is instantiated multiple
    times in a certain property specification.
    """
    
    def __init__(self, placeholder_id):
        """Initializes the PatternNotFoundError object.
        
        @param   placeholder_id: The ID of the placeholder that is
                                 instantiated multiple times.
        @type    placeholder_id: string
        """
        self.placeholder_id = placeholder_id
        
    def __str__(self):
        """Returns a human-readable string representation of the error.
        
        @return: A human-readable error message.
        @rtype: string
        """
        return ('The placeholder "%s" is instantiated multiple times in ' +
                'the property specification.') % self.placeholder_id
