# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


"""This module provides functionality to translate a CSL formula in XML format
to MRMC format.
"""

from lxml import etree

from property_management.src import constants
from property_management.src.formula_translation import exceptions
from property_management.src.op_elimination import stochastic


def get_formula_as_string(root):
    """Returns a string representation (in MRMC syntax) of the CSL formula
    that is given in XML format as an ElementTree rooted at the specified
    element.
    
    @param   root: The root element of the ElementTree representing the
                   formula.
    @type    root: etree._Element
    @return: A string representation (in MRMC syntax) of the formula that is
             given in XML format.
    @rtype:  string
    @raises  exceptions.UnexpectedXmlTagError: If the XML representation of
                the formula contains unexpected tags.
    @raises  exceptions.ViolatedRestrictionError: If the formula does not
                satisfy the restrictions that are imposed by the target tool
                (MRMC).
    """
    
    # Associate the tags of the comparison operator elements used in the XML
    # format with their correspondents in MRMC syntax.
    comparison_operators = { "lessThan":     "<",
                             "greaterThan":  ">",
                             "lessEqual":    "<=",
                             "greaterEqual": ">=" }
    
    # Proceed depending on the tag of the specified root element.
    # TODO: Use constants for the XML tag names.
    if root.tag == constants.FORMULAE_NAMESPACE + "cslFormula":
        return get_formula_as_string(root[0])
    
    elif root.tag == constants.CSL_NAMESPACE + "formula":
        return get_formula_as_string(root[0])
        
    elif root.tag == constants.CSL_NAMESPACE + "true":
        return "tt"
        
    elif root.tag == constants.CSL_NAMESPACE + "atomicProposition":
        return root.text
        
    elif root.tag == constants.CSL_NAMESPACE + "and":
        sub_formula1 = get_formula_as_string(root[0])
        sub_formula2 = get_formula_as_string(root[1])
        return "(%s && %s)" % (sub_formula1, sub_formula2)
        
    elif root.tag == constants.CSL_NAMESPACE + "or":
        sub_formula1 = get_formula_as_string(root[0])
        sub_formula2 = get_formula_as_string(root[1])
        return "(%s || %s)" % (sub_formula1, sub_formula2)
        
    elif root.tag == constants.CSL_NAMESPACE + "imp":
        sub_formula1 = get_formula_as_string(root[0])
        sub_formula2 = get_formula_as_string(root[1])
        return "(!%s || %s)" % (sub_formula1, sub_formula2)
        
    elif (root.tag == constants.CSL_NAMESPACE + "eq"):
        sub_formula1 = get_formula_as_string(root[0])
        sub_formula2 = get_formula_as_string(root[1])
        return "((!%s || %s) && ((!%s || %s))" % (  sub_formula1,sub_formula2,
                                                    sub_formula2,sub_formula1)
        
    elif root.tag == constants.CSL_NAMESPACE + "not":
        return "!%s" % get_formula_as_string(root[0])
        
    elif root.tag == constants.CSL_NAMESPACE + "s":
        # The first child of the "s" element contains the comparison operator
        # element. The second child has another child containing the threshold
        # value. The third child represents the sub-formula.
        comparison_operator = root[0][0].xpath("local-name()")
        return "S{%s%s}[%s]" % (comparison_operators[comparison_operator],
                                root[1][0].text,
                                get_formula_as_string(root[2]))
        
    elif root.tag == constants.CSL_NAMESPACE + "p":
        # The structure of the "p" element is equal to the structure of the
        # "s" element (see above).
        comparison_operator = root[0][0].xpath("local-name()")
        return "P{%s%s}[%s]" % (comparison_operators[comparison_operator],
                                root[1][0].text,
                                get_formula_as_string(root[2]))
        
    elif root.tag == constants.CSL_NAMESPACE + "x":
        # The last child element represents the sub-formula.
        sub_formula = get_formula_as_string(root[-1])
        
        # Assume the default values for the lower and upper bound of the
        # operator X. The defaults correspond to an unbounded X.
        lower_bound = "0"
        upper_bound = "infinity"
        # Check whether bounds are specified for the operator X (represented
        # by adequate child elements before the sub-formula child).
        for bound_element in root[0:-1]:
            if bound_element.tag == constants.CSL_NAMESPACE + "lowerBound":
                # Save the lower bound value which is contained in the child
                # element of the "lowerBound" element.
                lower_bound = bound_element[0].text
            elif bound_element.tag == constants.CSL_NAMESPACE + "upperBound":
                # Save the upper bound value which is contained in the child
                # element of the "upperBound" element.
                upper_bound = bound_element[0].text

        # Use the bounded syntax only if at least one bound is set.
        if lower_bound != "0" or upper_bound != "infinity":
            return "X[%s, %s] %s" % (lower_bound, upper_bound, sub_formula)
        else:
            return "X%s" % (sub_formula)
        
    elif root.tag == constants.CSL_NAMESPACE + "u":
        # The last two child elements represent the sub-formulae.
        sub_formula1 = get_formula_as_string(root[-2])
        sub_formula2 = get_formula_as_string(root[-1])
        
        # Assume the default values for the lower and upper bound of the
        # operator U. The defaults correspond to an unbounded U.
        lower_bound = "0"
        upper_bound = "infinity"
        # Check whether bounds are specified for the operator U (represented
        # by adequate child elements before the sub-formula children).
        for bound_element in root[0:-2]:
            if bound_element.tag == constants.CSL_NAMESPACE + "lowerBound":
                # Get the lower bound value which is contained in the child
                # element of the "lowerBound" element.
                lower_bound = bound_element[0].text
            elif bound_element.tag == constants.CSL_NAMESPACE + "upperBound":
                # Get the upper bound value which is contained in the child
                # element of the "upperBound" element.
                upper_bound = bound_element[0].text
                
        # If a lower bound is specified, an upper bound has to be specified as
        # well.
        if lower_bound != "0" and upper_bound == "infinity":
            raise exceptions.ViolatedRestrictionError(
                'If a lower bound is specified for "until" formulae, '
                'an upper bound has to be specified as well.')
        
        # Use the bounded syntax only if at least one bound is set.
        if lower_bound != "0" or upper_bound != "infinity":
            return "%s U[%s, %s] %s" % (sub_formula1,
                                        lower_bound,
                                        upper_bound,
                                        sub_formula2)
        else:
            return "%s U %s" % (sub_formula1, sub_formula2)

    else:
        raise exceptions.UnexpectedXmlTagError(root.tag)
    

def translate_formula(formula_root):
    """Eliminates operators from the specified formulae that are not supported
    by MRMC and returns the resulting formula as a string in MRMC syntax.
    
    @param   formula_root: The root element of the ElementTree representing
                           the formula.
    @type    formula_root: etree._Element
    @return: The translated formula in MRMC syntax.
    @rtype:  string
    @raises  exceptions.UnexpectedXmlTagError: If the XML representation of
                the formula contains unexpected tags.
    @raises  exceptions.ViolatedRestrictionError: If the formula does not
                satisfy the restrictions that are imposed by the target tool
                (MRMC).
    """
    stochastic.eliminate_w(formula_root, constants.CSL_NAMESPACE)
    stochastic.eliminate_g(formula_root, constants.CSL_NAMESPACE)
    stochastic.eliminate_f(formula_root, constants.CSL_NAMESPACE)
    return get_formula_as_string(formula_root)
