# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


"""This module provides functionality to translate a CTL formula in XML format
to NuSMV format.
"""

from lxml import etree

from property_management.src import constants
from property_management.src.formula_translation import exceptions
import property_management.src.op_elimination.ctl as op_elimination_ctl


def get_formula_as_string(root):
    """Return a string representation (in NuSMV syntax) of the CTL formula
    that is given in XML format as an ElementTree rooted at the specified
    element.
    
    @param   root: The root element of the ElementTree representing the
                   formula.
    @type    root: etree._Element
    @return: A string representation (in NuSMV syntax) of the formula that is
             given in XML format.
    @rtype:  string
    @raises  exceptions.UnexpectedXmlTagError: If the XML representation of
                the formula contains unexpected tags.
    @raises  exceptions.ViolatedRestrictionError: If the formula does not
                satisfy the restrictions that are imposed by the target tool
                (NuSMV).
    """

    # Proceed depending on the tag of the specified root element.
    # TODO: Use constants for the XML tag names.
    if root.tag == constants.FORMULAE_NAMESPACE + "ctlFormula":
        return get_formula_as_string(root[0])
    
    elif root.tag == constants.CTL_NAMESPACE + "formula":
        return get_formula_as_string(root[0])
        
    elif root.tag == constants.CTL_NAMESPACE + "true":
        return "true"
        
    elif root.tag == constants.CTL_NAMESPACE + "atomicProposition":
        return root.text
        
    elif root.tag == constants.CTL_NAMESPACE + "and":
        sub_formula1 = get_formula_as_string(root[0])
        sub_formula2 = get_formula_as_string(root[1])
        return "(%s & %s)" % (sub_formula1, sub_formula2)
        
    elif root.tag == constants.CTL_NAMESPACE + "or":
        sub_formula1 = get_formula_as_string(root[0])
        sub_formula2 = get_formula_as_string(root[1])
        return "(%s | %s)" % (sub_formula1, sub_formula2)
        
    elif root.tag == constants.CTL_NAMESPACE + "imp":
        sub_formula1 = get_formula_as_string(root[0])
        sub_formula2 = get_formula_as_string(root[1])
        return "(%s -> %s)" % (sub_formula1, sub_formula2)
        
    elif (root.tag == constants.CTL_NAMESPACE + "eq"):
        sub_formula1 = get_formula_as_string(root[0])
        sub_formula2 = get_formula_as_string(root[1])
        return "(%s <-> %s)" % (sub_formula1, sub_formula2)
        
    elif root.tag == constants.CTL_NAMESPACE + "not":
        return "!%s" % get_formula_as_string(root[0])
        
    elif root.tag == constants.CTL_NAMESPACE + "e":
        return "E%s" % (get_formula_as_string(root[0]))
        
    elif root.tag == constants.CTL_NAMESPACE + "a":
        return "A%s" % (get_formula_as_string(root[0]))
    
    elif root.tag == constants.CTL_NAMESPACE + "x":
        return "X %s" % (get_formula_as_string(root[0]))

    elif root.tag == constants.CTL_NAMESPACE + "u":
        # The last two child elements represent the sub-formulae.
        sub_formula1 = get_formula_as_string(root[-2])
        sub_formula2 = get_formula_as_string(root[-1])
        
        # Assume the default values for the lower and upper bound of the
        # operator U. The defaults correspond to an unbounded U.
        lower_bound = "0"
        upper_bound = "infinity"
        # Check whether bounds are specified for the operator U (represented
        # by adequate child elements before the sub-formula children).
        for bound_element in root[0:-2]:
            if bound_element.tag == constants.CTL_NAMESPACE + "lowerBound":
                # Get the lower bound value which is contained in the child
                # element of the "lowerBound" element.
                lower_bound = bound_element[0].text
            elif bound_element.tag == constants.CTL_NAMESPACE + "upperBound":
                # Get the upper bound value which is contained in the child
                # element of the "upperBound" element.
                upper_bound = bound_element[0].text
                
        # If a lower bound is specified, an upper bound has to be specified as
        # well.
        if lower_bound != "0" and upper_bound == "infinity":
            raise exceptions.ViolatedRestrictionError(
                'If a lower bound is specified for "until" formulae, '
                'an upper bound has to be specified as well.')
        
        # Use the bounded syntax only if the lower or upper bound is set.
        if lower_bound != "0" or upper_bound != "infinity":
            return "[%s BU %s..%s %s]" % (  sub_formula1,
                                            lower_bound,
                                            upper_bound,
                                            sub_formula2)
        else:
            return "[%s U %s]" % (sub_formula1, sub_formula2)
        
    elif root.tag == constants.CTL_NAMESPACE + "f":
        # The last child element represents the sub-formula.
        sub_formula = get_formula_as_string(root[-1])
        
        # Assume the default values for the lower and upper bound of the
        # operator F. The defaults correspond to an unbounded F.
        lower_bound = "0"
        upper_bound = "infinity"
        # Check whether bounds are specified for the operator F (represented
        # by adequate child elements before the sub-formula child).
        for bound_element in root[0:-1]:
            if bound_element.tag == constants.CTL_NAMESPACE + "lowerBound":
                # Save the lower bound value which is contained in the child
                # element of the "lowerBound" element.
                lower_bound = bound_element[0].text
            elif bound_element.tag == constants.CTL_NAMESPACE + "upperBound":
                # Save the upper bound value which is contained in the child
                # element of the "upperBound" element.
                upper_bound = bound_element[0].text

        # If a lower bound is specified, an upper bound has to be specified as
        # well.
        if lower_bound != "0" and upper_bound == "infinity":
            raise exceptions.ViolatedRestrictionError(
                'If a lower bound is specified for "F" formulae, '
                'an upper bound has to be specified as well.')

        # Use the bounded syntax only if at least one bound is set.
        if lower_bound != "0" or upper_bound != "infinity":
            return "BF %s..%s %s" % (lower_bound, upper_bound, sub_formula)
        else:
            return "F %s" % (sub_formula)
        
    elif root.tag == constants.CTL_NAMESPACE + "g":
        # The last child element represents the sub-formula.
        sub_formula = get_formula_as_string(root[-1])
        
        # Assume the default values for the lower and upper bound of the
        # operator G. The defaults correspond to an unbounded G.
        lower_bound = "0"
        upper_bound = "infinity"
        # Check whether bounds are specified for the operator G (represented
        # by adequate child elements before the sub-formula child).
        for bound_element in root[0:-1]:
            if bound_element.tag == constants.CTL_NAMESPACE + "lowerBound":
                # Save the lower bound value which is contained in the child
                # element of the "lowerBound" element.
                lower_bound = bound_element[0].text
            elif bound_element.tag == constants.CTL_NAMESPACE + "upperBound":
                # Save the upper bound value which is contained in the child
                # element of the "upperBound" element.
                upper_bound = bound_element[0].text

        # If a lower bound is specified, an upper bound has to be specified as
        # well.
        if lower_bound != "0" and upper_bound == "infinity":
            raise exceptions.ViolatedRestrictionError(
                'If a lower bound is specified for "G" formulae, '
                'an upper bound has to be specified as well.')

        # Use the bounded syntax only if at least one bound is set.
        if lower_bound != "0" or upper_bound != "infinity":
            return "BG %s..%s %s" % (lower_bound, upper_bound, sub_formula)
        else:
            return "G %s" % (sub_formula)

    else:
        raise exceptions.UnexpectedXmlTagError(root.tag)
    

def translate_formula(formula_root):
    """Eliminates operators from the specified formulae that are not supported
    by NuSMV and returns the resulting formula as a string in NuSMV syntax.
    
    @param   formula_root: The root element of the ElementTree representing
                           the formula.
    @type    formula_root: etree._Element
    @return: The translated formula in NuSMV syntax.
    @rtype:  string
    @raises  exceptions.UnexpectedXmlTagError: If the XML representation of
                the formula contains unexpected tags.
    @raises  exceptions.ViolatedRestrictionError: If the formula does not
                satisfy the restrictions that are imposed by the target tool
                (NuSMV).
    """
    op_elimination_ctl.eliminate_w(formula_root, constants.CTL_NAMESPACE)
    return get_formula_as_string(formula_root)
