# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


"""This module provides functions to eliminate certain operators from CTL
formulae.

The formulae are given in XML format, represented by ElementTree objects.

Currently, the only operator to be eliminated is W. This can be done using the
following function:
eliminate_w(root, namespace)
"""

import copy
from lxml import etree

# Important note: The parser provided by the formulae-module is used to create
# new XML elements. This is necessary in order to have these XML elements
# represented using the custom classes defined in the formulae-module.
import property_management.src.formulae as formulae_module


def eliminate_w(root, namespace):
    """Eliminate the weak until operator W in the formula that is represented
    by the XML ElementTree rooted a the specified element.
    
    This is done by replacing sub-formulae that contain the operator W by
    semantically equivalent sub-formulae that do not contain W but the strong
    until operator U.
    
    The following equivalences hold:
    A[x W y] = !E[!y U !(x or y)]
    E[x W y] = !A[!y U !(x or y)]
    
    @param   root:      The root element of the ElementTree representing the
                        formula.
    @type    root:      etree._Element
    @param   namespace: The XML namespace of the considered logic, surrounded
                        by curly braces. Example:
                        "{http://compass.informatik.rwth-aachen.de/ctl}"
    @type    namespace: string
    """
    # Use a shorter name to refer to the parser for formulae documents, which
    # is used to create new XML elements.
    parser = formulae_module.parser
    
    # Get all "w" elements in the ElementTree representing the formula in
    # depth first order and store them in a list.
    w_elements = list(root.iter(namespace + "w"))
    # Replace these "w" elements by corresponding "u" elements. Do this in
    # reversed order (i.e. bottom-up), to ensure correct results for nested W
    # formulae.
    w_elements.reverse()
    for w_element in w_elements:
        # Create a new element representing the sub-formula "x or y", where x
        # is the first and y the second sub-formula of the current W formula.
        # These sub-formulae are represented by the last child elements of the
        # "w" element. Note that these children are not copied but moved, so
        # they will be removed from their original parent element.
        sub_formula_x = w_element[-2]
        sub_formula_y = w_element[-1]
        x_or_y = parser.makeelement(namespace + "or")
        x_or_y.append(sub_formula_x)
        x_or_y.append(sub_formula_y)
        
        # Create a new element representing the negation of the previous
        # formula, i.e. "!(x or y)".
        # The formula "x or y" has to be wrapped in a "formula" element first.
        x_or_y_wrapper = parser.makeelement(namespace + "formula")
        x_or_y_wrapper.append(x_or_y)
        not_x_or_y = parser.makeelement(namespace + "not")
        not_x_or_y.append(x_or_y_wrapper)
        
        # Create a new element representing the sub-formula "!y".
        # Note that a deep copy is created of the element representing the
        # sub-formula y. This is necessary because y was already appended to
        # another element and would be removed from there without copying.
        not_y = parser.makeelement(namespace + "not")
        not_y.append(copy.deepcopy(sub_formula_y))
        
        # Create a new element representing the sub-formula "!y U !(x or y)".
        # This will be the path formula of the new P formula.
        # The sub-formulae of the U formula have to be wrapped in "formula"
        # elements first.
        not_y_wrapper = parser.makeelement(namespace + "formula")
        not_y_wrapper.append(not_y)
        not_x_or_y_wrapper = parser.makeelement(namespace + "formula")
        not_x_or_y_wrapper.append(not_x_or_y)
        # Create the element representing the U formula.
        u_element = parser.makeelement(namespace + "u")
        # Reuse the bounds of the W operator if there are any.
        # Note that at this point, the sub-formula children of the "w" element
        # have already been removed.
        if len(w_element) > 0:
            # The W operator has at least one bound. Move it to the "u"
            # element. It will be removed from the "w" element.
            u_element.append(w_element[0])
            if len(w_element) > 0:
                # The W operator has an upper and a lower bound. Move the
                # upper bound element as well. It will be removed from the "w"
                # element.
                u_element.append(w_element[0])
        # Add the sub-formulae to the "u" element.
        u_element.append(not_y_wrapper)
        u_element.append(not_x_or_y_wrapper)
        
        # Get the superordinate quantor element ("e" or "a") of the current
        # "w" element.
        quantor_element = w_element.getparent()
        # Create a new quantor element: the complement of the original one.
        if quantor_element.tag == namespace + "a":
            complement_quantor_element = parser.makeelement(namespace + "e")
        else:
            complement_quantor_element = parser.makeelement(namespace + "a")
        # Add the previously constructed path formula to the quantor element.
        complement_quantor_element.append(u_element)
        # The replacement for the W formula is the negation of the formula
        # that has been constructed so far.
        top_level_not_element = parser.makeelement(namespace + "not")
        top_level_not_element.append(complement_quantor_element)
        
        # Replace the W formula with the newly constructed formula.
        quantor_element.getparent().replace(quantor_element,
                                            top_level_not_element)
