# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


"""This module provides the data structures that are used to represent
properties documents. It also provides functions to parse properties documents
and validate them against the adequate XML Schema Definition.

The classes of this module correspond to types of the XML Schema documents
that define the structure of valid properties documents.
"""

import copy

from lxml import etree

from property_management.src import constants
from property_management.src import xml_tags
from property_management.src import errors
from property_management.src import util
from property_management.src import exceptions
import property_management.src.formulae as formulae_module


def parse_properties_document(data_source):
    """Parses the properties document from the specified data source and
    returns a representation of the document using the data structures defined
    in this module (and in the formulae module).
    
    The representation of the document will be in tree structure. For a valid
    document, the root element will be a PropertiesDocument object.
    
    Note: The document will not be validated. This must be done explicitly
    using the function get_syntax_errors_in_document.
    
    @param   data_source: The source of the data to be parsed. This may be a
                          filename or a file-like object.
    @type    data_source: A string or a file-like object.
    @return: A tree representation of the document.
    @rtype:  PropertiesDocument
    """
    # Use the parser that is created when this module is imported (see the end
    # of this file). It creates the adequate objects for the XML elements of
    # the document.
    # The result of the parsing process will be an etree._ElementTree object.
    # Return the root element of this tree (a PropertiesDocument object).
    document_tree = etree.parse(data_source, parser)
    return document_tree.getroot()


def get_syntax_errors_in_document(properties_document):
    """Validates the specified properties document against the adequate XML
    Schema Definition and returns a list containing the syntactic errors.
    
    Note: Syntactic validation does not include a check of the atomic
    propositions with respect to a certain SLIM model.
    
    @return: A list containing the syntactic errors in the document. This list
             will be empty if the document is valid with respect to the
             adequate XML Schema Definition.
    @rtype:  A list of errors.DocumentSyntaxError objects. 
    """
    # Initialize the list of errors to be returned.
    error_list = []
    # Validate the document. If it is invalid, an exception will be raised
    # which can be used to create an error object with an adequate error
    # description.
    # Note: The XML Schema object that is used for validation is created when
    # this module is imported (see the end of this file).
    try:
        _properties_schema.assertValid(properties_document)
    except etree.DocumentInvalid, e:
        error_list.append(errors.DocumentSyntaxError(str(e)))
    
    return error_list


# ----------------------------------------------------------------------------
# Data structures for properties documents.
# ----------------------------------------------------------------------------

class PropertiesDocument(util.SerializableXmlElement):
    """A list of logical properties, represented by Property objects.
    
    Properties can be added to a properties document using the append-method.
    
    Important: PropertiesDocument objects must not be created like this:
    x = PropertiesDocument()
    Instead, the static factory method create must be used.
    This is because this class is a subclass of etree.ElementBase, which must
    not be instantiated directly. Refer to the lxml documentation for details.
    """
    
    @staticmethod
    def create(properties=None):
        """Creates and returns a new PropertiesDocument object.
        
        @param   properties: (Optional) The properties to be added to the new
                             document. Note: Instead of the property objects
                             themselves, deep copies of these objects will be
                             added to the document.
        @type    properties: An iterable over Property objects.
        @return: A new PropertiesDocument object.
        @rtype:  PropertiesDocument
        """
        # Note: The parser that is used to create the object is created when
        # this module is imported (see the end of this file).
        properties_document = parser.makeelement(
                                        constants.PROPERTIES_NAMESPACE +
                                        xml_tags.PROPERTIES_DOCUMENT_TAG)
        if properties is not None:
            for property in properties:
                # Don't add the property object itself but rather a deep copy.
                # This is necessary for the following reason: The property
                # object is an XML element, which might already be used as a
                # child element of some other parent element (e.g. a
                # properties document). Without copying, it would be removed
                # from that parent. See the lxml documentation.
                properties_document.append(copy.deepcopy(property))
        return properties_document
    
    def get_property_by_name(self, name):
        """Returns the property with the specified name.
        
        In a valid properties document the property names are unique.
        
        @return: The property with the specified name or None if no such
                 property exists.
        @rtype:  Property
        """
        for property in self:
            if property.name == name:
                return property
        return None

    def __str__(self):
        """Returns a string representation of this document in XML format.
        
        UTF-8 encoding will be used and an XML header will be included.
        
        @return: A string representation of this document in XML format.
        @rtype:  string
        """
        return etree.tostring(self, xml_declaration=True, encoding="UTF-8",
                              pretty_print=True)
        
        
class Property(util.SerializableXmlElement):
    """A logical property.
    
    A property consists of a pattern instance enriched by a name and a
    description.
    
    Important: Property objects must not be created like this: x = Property()
    Instead, the static factory method create must be used.
    This is because this class is a subclass of etree.ElementBase, which must
    not be instantiated directly. Refer to the lxml documentation for details.
    """
    
    @staticmethod
    def create(pattern_instance, name, description=None):
        """Creates and returns a new Property object.
        
        @param   pattern_instance: The pattern instance to be encapsulated by
                                   property. Note: A deep copy of the pattern
                                   instance will be created.
        @type    pattern_instance: PatternInstance
        @param   name:             The name of the property.
        @type    name:             string
        @param   description:      (Optional) The description of the property.
                                   Default is "".
        @type    description:      string
        @return: A new Property object.
        @rtype:  Property
        """
        # Create the object representing the property. This is an XML element.
        # Note: The parser that is used to create the object is created when
        # this module is imported (see the end of this file).
        property = parser.makeelement(constants.PROPERTIES_NAMESPACE +
                                      xml_tags.PROPERTY_TAG)
        # Add the specified name, description and as child elements of the
        # property XML element. Note that this exact order of the elements is
        # required. However, the description is optional.
        # The pattern instance is already an XML element. Name and description
        # have to be encapsulated by XML elements.
        name_element = parser.makeelement(constants.PROPERTIES_NAMESPACE +
                                          xml_tags.PROPERTY_NAME_TAG)
        name_element.text = name
        property.append(name_element)
        
        if description is not None:
            description_element = parser.makeelement(
                                    constants.PROPERTIES_NAMESPACE +
                                    xml_tags.PROPERTY_DESCRIPTION_TAG)
            description_element.text = description
            property.append(description_element)
        
        # Don't add the specified pattern instance object itself but rather a
        # deep copy. This is necessary for the following reason: The pattern
        # instance XML element might already be used as a child element of
        # some other parent element (e.g. another property). Without copying,
        # it would be removed from that parent. See the lxml documentation.
        property.append(copy.deepcopy(pattern_instance))

        return property
    
    def get_logic(self):
        """Returns a constant identifying the logic that is used by this
        property.
        
        @return: A constant identifying the logic that is used by this
                 property.
        @rtype:  One of the constants in constants.Logic.
        @raises  Exception: If the logic could not be detected.
        """
        # Note: self.pattern_instance.get_logic may raise an Exception.
        return self.pattern_instance.get_logic()

    # The pattern instance is represented by a child element of this XML
    # element. There must be exactly one such child element.
    # Note: There are multiple tags that might be used for that child element,
    # depending on the logic. 
    def _get_pattern_instance(self):
        pattern_instance_tags = [xml_tags.LTL_PATTERN_INSTANCE_TAG,
                                 xml_tags.CTL_PATTERN_INSTANCE_TAG,
                                 xml_tags.MITL_PATTERN_INSTANCE_TAG,
                                 xml_tags.PCTL_PATTERN_INSTANCE_TAG,
                                 xml_tags.CSL_PATTERN_INSTANCE_TAG]
        local_names = ["local-name()='%s'" % tag \
                       for tag in pattern_instance_tags]
        pattern_instances = self.xpath("*[%s]" % " or ".join(local_names))
        assert len(pattern_instances) == 1
        return pattern_instances[0]

    def _set_pattern_instance(self, pattern_instance):
        old_pattern_instance = self._get_pattern_instance()
        # Don't use the specified pattern instance object itself but rather a
        # deep copy. This is necessary for the following reason: The pattern
        # instance object is an XML element, which might already be used as a
        # child element of some other parent element (e.g. another property).
        # Without copying, it would be removed from that parent. See the lxml
        # documentation.
        new_pattern_instance = copy.deepcopy(pattern_instance)
        # Use lxml's replace feature.
        self.replace(old_pattern_instance, new_pattern_instance)
        
    pattern_instance = property(_get_pattern_instance, _set_pattern_instance,
                                doc="The pattern instance that is used to " +
                                    "define this property.\n" +
                                    "Note: Assigning to this property will " +
                                    "create a deep copy of the assigned " +
                                    "object.\n" +
                                    "@type: PatternInstance")

    def get_pattern_id(self):
        return self.pattern_instance.pattern_id

    def stepwise_analysis_possible(self):
        return self.get_pattern_id() not in ["expectedTime", "longRunAverage"]
    
    # The name of this property is stored as the text content of the child
    # element with the adequate tag-name. There must be exactly one such child
    # element.
    def _get_name(self):
        name_elements = self.xpath("*[local-name()='%s']" %
                                   xml_tags.PROPERTY_NAME_TAG)
        assert len(name_elements) == 1
        return name_elements[0].text
    
    def _set_name(self, name):
        name_elements = self.xpath("*[local-name()='%s']" %
                                   xml_tags.PROPERTY_NAME_TAG)
        assert len(name_elements) == 1
        name_elements[0].text = name
        
    name = property(_get_name, _set_name, doc="The name of this property.\n" +
                                              "@type: string")
    # The description of this property is stored as the text content of the
    # child element with the adequate tag-name. There must be at most one such
    # child element (the description is optional).
    def _get_description(self):
        description_elements = self.xpath("*[local-name()='%s']" %
                                          xml_tags.PROPERTY_DESCRIPTION_TAG)
        assert len(description_elements) <= 1
        if len(description_elements) >= 1: # More precisely: len(...) == 1
            # The description element might be empty. In this case, the text
            # attribute will be None. Exclude this case here.
            if description_elements[0].text is not None:
                return description_elements[0].text
        # There is no description element or it is empty, so return the empty
        # string.
        return ""
    
    def _set_description(self, description):
        # If no description exists yet, add an adequate child element.
        # Note that the order of the child elements does not matter, so we can
        # append at the end.
        description_elements = self.xpath("*[local-name()='%s']" %
                                          xml_tags.PROPERTY_DESCRIPTION_TAG)
        assert len(description_elements) <= 1
        if len(description_elements) >= 1: # More precisely: len(...) == 1
            description_elements[0].text = description
        else:
            description_element = parser.makeelement(
                                    constants.PROPERTIES_NAMESPACE +
                                    xml_tags.PROPERTY_DESCRIPTION_TAG)
            description_element.text = description
            self.append(description_element)

    def __str__(self):
        """Returns a string representation of this property in XML format.
        
        UTF-8 encoding will be used and an XML header will be included.
        
        @return: A string representation of this property in XML format.
        @rtype:  string
        """
        return etree.tostring(self, xml_declaration=True, encoding="UTF-8",
                              pretty_print=True)
        
    description = property(_get_description, _set_description,
                           doc="The description of this property.\n" +
                               "@type: string")


class PatternInstance(util.SerializableXmlElement):
    """A pattern instance for a pattern in a certain logic.
    
    A pattern instance is a list of placeholder instances which define the
    content to be inserted instead of the placeholders in the pattern.
    
    Note: The placeholders are identified by their IDs. There must be exactly
    one placeholder instance for each placeholder of the pattern. To ensure
    that there are not duplicate placeholder instances, it is strongly
    recommended to use the method add_placehoder_instance instead of the method
    append (or other list operations that add elements) to add placeholder
    instances to the pattern instance.
    
    Important: PatternInstance objects must not be created like this:
    x = PatternInstance()
    Instead, the static factory method create must be used.
    This is because this class is a subclass of etree.ElementBase, which must
    not be instantiated directly. Refer to the lxml documentation for details.
    """
    
    @staticmethod
    def create(logic, pattern_id):
        """Creates and returns a new pattern instance object referring to the
        pattern in the specified logic with the specified ID.
        
        @param   logic:      A constant identifying the logic of the pattern
                             to be instantiated.
        @type    logic:      One of the constants in constants.Logic.
        @param   pattern_id: The ID of the pattern to be instantiated.
        @type    pattern_id: string
        @return: A new PatternInstance object.
        @rtype:  PatternInstance
        """
        # Create a new object representing a pattern instance for a pattern in
        # the specified logic.
        # Note: The parser that is used to create the object is created when
        # this module is imported (see the end of this file).
        # TODO: Use a dictionary instead of if-elif.
        if logic == constants.Logic.LTL:
            pattern_instance = parser.makeelement(
                    constants.PROPERTIES_NAMESPACE +
                    xml_tags.LTL_PATTERN_INSTANCE_TAG)
        elif logic == constants.Logic.CTL:
            pattern_instance = parser.makeelement(
                    constants.PROPERTIES_NAMESPACE +
                    xml_tags.CTL_PATTERN_INSTANCE_TAG)
        elif logic == constants.Logic.MITL:
            pattern_instance = parser.makeelement(
                    constants.PROPERTIES_NAMESPACE +
                    xml_tags.MITL_PATTERN_INSTANCE_TAG)
        elif logic == constants.Logic.PCTL:
            pattern_instance = parser.makeelement(
                    constants.PROPERTIES_NAMESPACE +
                    xml_tags.PCTL_PATTERN_INSTANCE_TAG)
        elif logic == constants.Logic.CSL:
            pattern_instance = parser.makeelement(
                    constants.PROPERTIES_NAMESPACE +
                    xml_tags.CSL_PATTERN_INSTANCE_TAG)
        # Set the attribute indicating the ID of the pattern to be used.
        pattern_instance.set("patternId", pattern_id)
        return pattern_instance
    
    def get_logic(self):
        """Returns a constant identifying the logic that is used by this
        pattern instance.
        
        @return: A constant identifying the logic that is used by this pattern
                 instance.
        @rtype:  One of the constants in constants.Logic.
        @raises  Exception: If the logic could not be detected.
        """
        # The logic can be deduced from the tag name of the underlying XML
        # element.
        # TODO: Use a dictionary instead of if-elif.
        tag_name = self.xpath("local-name()")
        if tag_name == xml_tags.LTL_PATTERN_INSTANCE_TAG:
            return constants.Logic.LTL
        elif tag_name == xml_tags.CTL_PATTERN_INSTANCE_TAG:
            return constants.Logic.CTL
        elif tag_name == xml_tags.MITL_PATTERN_INSTANCE_TAG:
            return constants.Logic.MITL
        elif tag_name == xml_tags.PCTL_PATTERN_INSTANCE_TAG:
            return constants.Logic.PCTL
        elif tag_name == xml_tags.CSL_PATTERN_INSTANCE_TAG:
            return constants.Logic.CSL
        else:
            # TODO: Use a custom exception class.
            raise Exception(('The logic could not be detected for the tag ' +
                             'name "%s".') % tag_name)
    
    # The pattern ID is stored as an attribute of this XML element.
    def _get_pattern_id(self):
        return self.get("patternId")
    
    def _set_pattern_id(self, pattern_id):
        return self.set("patternId", pattern_id)
    
    pattern_id = property(_get_pattern_id, _set_pattern_id,
                          doc="The ID of the pattern that is " +
                              "instantiated.\n" +
                              "@type: string")
            
    def get_placeholder_instance(self, placeholder_id):
        """Returns the placeholder instance for the placeholder with the
        specified ID.
        
        Note: A valid pattern instance will contain only one placeholder
        instance for a single placeholder.
        
        @param   placeholder_id: The ID of the placeholder to be considered.
        @type    placeholder_id: string
        @return: The placehoder instance for the placeholder with the
                 specified ID or None if no such placeholder instance exists.
        @rtype:  A subclass of PlaceholderInstance.
        """
        for placeholder_instance in self:
            if placeholder_instance.placeholder_id == placeholder_id:
                return placeholder_instance
        return None
    
    def add_placeholder_instance(self, placeholder_instance):
        """Adds the specified placeholder instance to this pattern instance.
        
        @param   placeholder_instance: The placeholder instance to be added.
                                       Note: A deep copy of the placeholder
                                       instance will be created.
        @type    placeholder_instance: PlaceholderInstance
        @raises  exceptions.PlaceholderInstanceAlreadyExistsError:
                    If this pattern instance does already contain an instance
                    for the placeholder referenced by the specified instance.
        """
        placeholder_id = placeholder_instance.placeholder_id
        if self.get_placeholder_instance(placeholder_id):
            raise exceptions.PlaceholderInstanceAlreadyExistsError(
                                                    placeholder_id)
        else:
            # Don't add the specified placeholder instance object itself but
            # rather a deep copy. This is necessary for the following reason:
            # The placeholder instance object is an XML element, which might
            # already be used as a child element of some other parent element
            # (e.g. another pattern instance). Without copying, it would be
            # removed from that parent. See the lxml documentation.
            self.append(copy.deepcopy(placeholder_instance))
            
    # Associate the tags of the XML elements that are considered placeholders
    # with the tags of the corresponding elements that contain the XML
    # fragments to be inserted instead of the placeholders.
    # This is needed for pattern instantiation.
    _placeholder_tag_to_instance_tag_dict = {
            xml_tags.STATE_FORMULA_PLACEHOLDER_TAG:
                    xml_tags.STATE_FORMULA_PLACEHOLDER_INSTANCE_TAG,
            xml_tags.PATH_FORMULA_PLACEHOLDER_TAG:
                    xml_tags.PATH_FORMULA_PLACEHOLDER_INSTANCE_TAG,
            xml_tags.TIME_BOUND_PLACEHOLDER_TAG:
                    xml_tags.TIME_BOUND_PLACEHOLDER_INSTANCE_TAG,
            xml_tags.THRESHOLD_PLACEHOLDER_TAG:
                    xml_tags.THRESHOLD_PLACEHOLDER_INSTANCE_TAG,
            xml_tags.COMPARISON_OPERATOR_PLACEHOLDER_TAG:
                    xml_tags.COMPARISON_OPERATOR_PLACEHOLDER_INSTANCE_TAG}
    
    def create_formula(self, patterns_document):
        """Creates and returns the formula that is represented by this
        instance.
        
        @param   patterns_document: The patterns document that defines the
                                    pattern to be instantiated.
        @type    patterns_document: patterns_module.PatternsDocument
        @return: A new object representing the formula that is represented by
                 this pattern instance.
        @rtype:  formulae_module.Formula
        @raises  exceptions.PatternInstantiationError:
                    If the instantiation of the pattern failed. This might
                    happen if this pattern instance refers to a nonexistent
                    pattern, if not all placeholders are instantiated or if a
                    placeholder is instantiated twice.
        """
        # Get the definition of the pattern to be instantiated.
        pattern = patterns_document.get_pattern_by_id(self.pattern_id) 
        if pattern is None:
            raise exceptions.PatternNotFoundError(self.pattern_id)
    
        # Create a new formula in form of an XML element. This object will be
        # returned at the end. The tag of the element depends on the logic of
        # the formula (and thus the logic of the pattern instance).
        # Note: The parser that is used to create the XML element is created
        # when this module is imported (see the end of this file).
        # TODO: Implement a create-method for the Formula class.
        if self.get_logic() == constants.Logic.LTL:
            formula = parser.makeelement(constants.FORMULAE_NAMESPACE +
                                         xml_tags.LTL_FORMULA_TAG)
        elif self.get_logic() == constants.Logic.CTL:
            formula = parser.makeelement(constants.FORMULAE_NAMESPACE +
                                         xml_tags.CTL_FORMULA_TAG)
        elif self.get_logic() == constants.Logic.PCTL:
            formula = parser.makeelement(constants.FORMULAE_NAMESPACE +
                                         xml_tags.PCTL_FORMULA_TAG)
        elif self.get_logic() == constants.Logic.CSL:
            formula = parser.makeelement(constants.FORMULAE_NAMESPACE +
                                         xml_tags.CSL_FORMULA_TAG)
        elif self.get_logic() == constants.Logic.MITL:
            formula = parser.makeelement(constants.FORMULAE_NAMESPACE +
                                         xml_tags.MITL_FORMULA_TAG)
        else:
            raise exceptions.PatternInstantiationError("Pattern instance " +
                                                       "uses unknown logic.")
        # Initialize the formula element such that its content is a copy of
        # the content of the formula template of the pattern.
        # Note that at this point the formula is invalid because it contains
        # placeholders. These placeholders will be replaced later in this
        # function.
        for child_element in pattern.formula_template:
            formula.append(copy.deepcopy(child_element))
    
        # Construct an XPath expression to find all placeholders in the
        # formula template.
        local_names = \
                ["local-name()='%s'" % tag for tag in \
                 PatternInstance._placeholder_tag_to_instance_tag_dict.keys()]
        placeholders_xpath = ".//*[%s]" % " or ".join(local_names)
        
        # Replace the placeholders elements in the formula template with the
        # XML fragments that are assigned to these placeholders in this
        # pattern instance.
        for placeholder in formula.xpath(placeholders_xpath):
            # Find the placeholder instance that assigns an XML fragment to
            # the current placeholder. It must have the adequate tag.
            placeholder_tag = placeholder.xpath("local-name()")
            placeholder_instance_tag = PatternInstance.\
                    _placeholder_tag_to_instance_tag_dict[placeholder_tag]
            placeholder_instance_xpath = \
                    ("./*[local-name()='%s' and @placeholderId='%s']" %
                     (placeholder_instance_tag, placeholder.get("id")))
            placeholder_instances = self.xpath(placeholder_instance_xpath)
            # There must be exactly one instantiation for the current
            # placeholder.
            if len(placeholder_instances) < 1:
                raise exceptions.MissingPlaceholderInstantiationError(
                                                placeholder.get("id"))
            if len(placeholder_instances) > 1:
                raise exceptions.DuplicatePlaceholderInstantiationError(
                                                placeholder.get("id"))
    
            # Get a copy of the root element of the XML fragment assigned to
            # the placeholder.
            placeholder_content = copy.deepcopy(placeholder_instances[0][0])
            # Replace the placeholder with the associated XML fragment.
            # TODO: placeholder.getparent() should never be None, but it was
            # found to be None sometimes. This is a bug. The following check
            # is useful to investigate this bug but should be replaced by an
            # assertion as soon as the cause of the bug is identified.
            if placeholder.getparent() is not None:
                placeholder.getparent().replace(placeholder, placeholder_content)
            else: 
                import utils.log
                utils.log.system.error("Placeholder '%s' has no parent" % str(placeholder))
                pass
            
        return formula


class PlaceholderInstance(util.SerializableXmlElement):
    """A definition of the content to be inserted instead of a certain
    placeholder in a formula pattern.
    
    This class is a base class for classes that are used to represent a
    concrete kind of placeholder  (like state formula placeholder or time
    bound placeholder).
    """
    
    def get_logic(self):
        """Returns a constant identifying the logic that is used by this
        pattern instance.
        
        @return: A constant identifying the logic that is used by this pattern
                 instance.
        @rtype:  One of the constants in constants.Logic.
        @raises  Exception: If the logic could not be detected.
        """
        # This placeholder instance is represented by an XML element. The
        # logic is determined by the namespace of this element.
        # TODO: This mapping might be useful for other functions as well, so
        # maybe it could be moved to a common module (like "constants").
        namespace_uri_to_logic_dict = {
                constants.LTL_NAMESPACE_URI: constants.Logic.LTL,
                constants.CTL_NAMESPACE_URI: constants.Logic.CTL,
                constants.MITL_NAMESPACE_URI: constants.Logic.MITL,
                constants.PCTL_NAMESPACE_URI: constants.Logic.PCTL,
                constants.CSL_NAMESPACE_URI: constants.Logic.CSL}
        xml_namespace_uri = self.xpath("namespace-uri()")
        try:
            return namespace_uri_to_logic_dict[xml_namespace_uri]
        except KeyError:
            # The XML element has an unexpected namespace.
            # TODO: Use a custom exception class.
            raise Exception("Logic of placeholder instance could not be " +
                            "detected.")
    
    # The placeholder ID is stored as an attribute of this XML element.
    def _get_placeholder_id(self):
        return self.get("placeholderId")
    
    def _set_placeholder_id(self, placeholder_id):
        return self.set("placeholderId", placeholder_id)
    
    placeholder_id = property(_get_placeholder_id, _set_placeholder_id,
                              doc="The ID of the placeholder that is " +
                                  "instantiated.\n" +
                                  "@type: string")
    

class StateFormulaPlaceholderInstance(PlaceholderInstance):
    """A placeholder instance indicating that a certain state formula is to be
    inserted for a certain state formula placeholder in a formula pattern.
    
    This type of placeholder instance is also used for LTL subformulae
    placeholders although actually there are no "state" formulae in LTL.
    """
    
    @staticmethod
    def create(logic, placeholder_id, formula_content):
        """Creates and returns a new state formula placeholder instance
        indicating that the specified formula content is to be inserted for
        the placeholder.
        
        @param   logic:           The logic to be used.
        @type    logic:           One of the constants in constants.Logic.
        @param   placeholder_id:  The ID of the placeholder to be
                                  instantiated.
        @type    placeholder_id:  string
        @param   formula_content: The state formula content to be inserted.
                                  This must be represented by an adequate XML
                                  element. Usually this will be an atomic
                                  proposition.
                                  Note: A deep copy of the state formula
                                  content will be created.
        @type    formula_content: etree._Element, usually
                                  formulae_module.AtomicProposition.
                                  Important: This must never be a
                                  formulae_module.Formula object but rather
                                  its content.
        @return: A new StateFormulaPlaceholderInstance object.
        @rtype:  StateFormulaPlaceholderInstance
        """
        # Get the adequate XML namespace for the specified logic.
        xml_namespace = constants.logic_to_xml_namespace_dict[logic]
        # Create a new XML element representing the placeholder instance in
        # the specified logic.
        # Note: The parser that is used to create the element is created when
        # this module is imported (see the end of this file).
        placeholder_instance = parser.makeelement(
                xml_namespace +
                xml_tags.STATE_FORMULA_PLACEHOLDER_INSTANCE_TAG)
        # The XML element must have exactly one child element, namely the
        # element that represents the formula content to be inserted.
        # Don't add the specified formula content object itself but rather a
        # deep copy. This is necessary for the following reason: The formula
        # content object is an XML element, which might already be used as a
        # child element of some other parent element (e.g. a formula object).
        # Without copying, it would be removed from that parent. See the lxml
        # documentation.
        placeholder_instance.append(copy.deepcopy(formula_content))
        # Set the attribute indicating the ID of the placeholder to be
        # instantiated.
        # TODO: Use a constant for the attribute name.
        placeholder_instance.set("placeholderId", placeholder_id)
        return placeholder_instance
    
    # This placeholder instance is represented by an XML element that has
    # exactly one child element representing the formula content to be
    # inserted.
    def _get_content(self):
        return self[0]
    
    def _set_content(self, formula_content):
        # Don't use the specified formula content object itself but rather a
        # deep copy. This is necessary for the following reason: The formula
        # content object is an XML element, which might already be used as a
        # child element of some other parent element (e.g. a formula object).
        # Without copying, it would be removed from that parent. See the lxml
        # documentation.
        self[0] = copy.deepcopy(formula_content)
        
    content = property(_get_content, _set_content,
                       doc="The state formula content to be inserted for " +
                           "the placeholder.\n" +
                           "Note: Assigning to this property will create a " +
                           "deep copy of the assigned object.\n" +
                           "\n" +
                           "@type: etree._Element, usually " +
                           "formulae_module.AtomicProposition. Important: " +
                           "This must never be a formulae_module.Formula " +
                           "object but rather its content.")
    
    
class TimeBoundPlaceholderInstance(PlaceholderInstance):
    """A placeholder instance indicating that a certain value is to be
    inserted for a certain time bound placeholder in a formula pattern.
    """
    
    @staticmethod
    def create(logic, placeholder_id, time_bound):
        """Creates and returns a new time bound placeholder instance
        indicating that the specified time bound is to be inserted for the
        placeholder.
        
        @param   logic:          The logic to be used.
        @type    logic:          One of the constants in constants.Logic.
        @param   placeholder_id: The ID of the placeholder to be instantiated.
        @type    placeholder_id: string
        @param   time_bound:     The time bound to be inserted. This can be
                                 either an int value or constants.INFINITY,
                                 which is a string representing the special
                                 infinity value.
        @type    time_bound:     int or string
        @return: A new TimeBoundPlaceholderInstance object.
        @rtype:  TimeBoundPlaceholderInstance
        """
        # Get the adequate XML namespace for the specified logic.
        xml_namespace = constants.logic_to_xml_namespace_dict[logic]
        # Create a new XML element representing the placeholder instance in
        # the specified logic.
        # Note: The parser that is used to create the element is created when
        # this module is imported (see the end of this file).
        placeholder_instance = parser.makeelement(
                xml_namespace +
                xml_tags.TIME_BOUND_PLACEHOLDER_INSTANCE_TAG)
        # The XML element must have exactly one child element containing the
        # time bound as text content.
        value_element = parser.makeelement(xml_namespace + xml_tags.VALUE_TAG)
        value_element.text = str(time_bound)
        placeholder_instance.append(value_element)
        # Set the attribute indicating the ID of the placeholder to be
        # instantiated.
        # TODO: Use a constant for the attribute name.
        placeholder_instance.set("placeholderId", placeholder_id)
        return placeholder_instance
    
    # This placeholder instance is represented by an XML element that has
    # exactly one child element containing the time bound as text content.
    def _get_content(self):
        # The time bound might be the special infinity value. In this case it
        # must not be converted to an int.
        if self[0].text == constants.INFINITY:
            return self[0].text
        else:
            # The value may contain a time unit
            # if so, return the string
            try:
                return int(self[0].text)
            except:
                return self[0].text
    
    def _set_content(self, time_bound):
        self[0].text = str(time_bound)
        
    content = property(_get_content, _set_content,
                       doc="The time bound to be inserted for the " +
                           "placeholder.\n" +
                           "This can be either an int value or " +
                           "constants.INFINITY, which is a string " +
                           "representing the special infinity value.\n" +
                           "@type: int or string")
    
    
class ThresholdPlaceholderInstance(PlaceholderInstance):
    """A placeholder instance indicating that a certain float value is to be
    inserted for a certain threshold placeholder in a formula pattern.
    
    Here the term "threshold" refers to a probability threshold in a
    probabilistic formula like "P{>0.5}[some_formula]".
    """
    
    @staticmethod
    def create(logic, placeholder_id, threshold):
        """Creates and returns a new time bound placeholder instance
        indicating that the specified probability threshold is to be inserted
        for the placeholder.
        
        @param   logic:          The logic to be used.
        @type    logic:          One of the constants in constants.Logic.
        @param   placeholder_id: The ID of the placeholder to be instantiated.
        @type    placeholder_id: string
        @param   threshold:      The probability threshold to be inserted.
                                 This must be a value in the range [0,1].
        @type    threshold:      float
        @return: A new ThresholdPlaceholderInstance object.
        @rtype:  ThresholdPlaceholderInstance
        """
        # Get the adequate XML namespace for the specified logic.
        xml_namespace = constants.logic_to_xml_namespace_dict[logic]
        # Create a new XML element representing the placeholder instance in
        # the specified logic.
        # Note: The parser that is used to create the element is created when
        # this module is imported (see the end of this file).
        placeholder_instance = parser.makeelement(
                xml_namespace +
                xml_tags.THRESHOLD_PLACEHOLDER_INSTANCE_TAG)
        # The XML element must have exactly one child element containing the
        # probaility hreshold as text content.
        value_element = parser.makeelement(xml_namespace + xml_tags.VALUE_TAG)
        value_element.text = str(threshold)
        placeholder_instance.append(value_element)
        # Set the attribute indicating the ID of the placeholder to be
        # instantiated.
        # TODO: Use a constant for the attribute name.
        placeholder_instance.set("placeholderId", placeholder_id)
        return placeholder_instance
        
    # This placeholder instance is represented by an XML element that has
    # exactly one child element containing the time bound as text content.
    def _get_content(self):
        return float(self[0].text)
    
    def _set_content(self, threshold):
        self[0].text = str(threshold)
        
    content = property(_get_content, _set_content,
                       doc="The probability threshold to be inserted for" +
                           "the placeholder.\n" +
                           "@type: float")
    

# TODO: Implement a PlaceholderInstance subclasses for comparison operator
# placeholders. (Such placeholders are currently not used in the COMPASS
# toolset.)

# ----------------------------------------------------------------------------
# End of the definition of the data structures.
# ----------------------------------------------------------------------------


def _get_parser_lookup_scheme():
    """Returns a new parser lookup scheme which is used to map certain XML
    elements in properties documents to instances of the adequate classes
    defined in this module and in the formulae-module.
    
    @return: A parser lookup scheme for properties documents.
    @rtype:  etree.ElementNamespaceClassLookup
    """
    # Extend the lookup scheme for formulae documents so that the data
    # structures for properties documents and formulae documents are
    # compatible and can be combined.
    lookup = formulae_module._get_parser_lookup_scheme()
    props_namespace = lookup.get_namespace(constants.PROPERTIES_NAMESPACE_URI)
    props_namespace[xml_tags.PROPERTIES_DOCUMENT_TAG] = PropertiesDocument
    props_namespace[xml_tags.PROPERTY_TAG] = Property
    props_namespace[xml_tags.LTL_PATTERN_INSTANCE_TAG] = PatternInstance
    props_namespace[xml_tags.CTL_PATTERN_INSTANCE_TAG] = PatternInstance
    props_namespace[xml_tags.MITL_PATTERN_INSTANCE_TAG] = PatternInstance
    props_namespace[xml_tags.PCTL_PATTERN_INSTANCE_TAG] = PatternInstance
    props_namespace[xml_tags.CSL_PATTERN_INSTANCE_TAG] = PatternInstance
    # Associate elements representing LTL placeholder instances with the
    # adequate classes.
    ltl_namespace = lookup.get_namespace(constants.LTL_NAMESPACE_URI)
    ltl_namespace[xml_tags.STATE_FORMULA_PLACEHOLDER_INSTANCE_TAG] = \
            StateFormulaPlaceholderInstance
    ltl_namespace[xml_tags.TIME_BOUND_PLACEHOLDER_INSTANCE_TAG] = \
            TimeBoundPlaceholderInstance
    # Associate elements representing CTL placeholder instances with the
    # adequate classes.
    ctl_namespace = lookup.get_namespace(constants.CTL_NAMESPACE_URI)
    ctl_namespace[xml_tags.STATE_FORMULA_PLACEHOLDER_INSTANCE_TAG] = \
            StateFormulaPlaceholderInstance
    ctl_namespace[xml_tags.PATH_FORMULA_PLACEHOLDER_INSTANCE_TAG] = \
            PlaceholderInstance
    ctl_namespace[xml_tags.TIME_BOUND_PLACEHOLDER_INSTANCE_TAG] = \
            TimeBoundPlaceholderInstance
    # Associate elements representing MITL placeholder instances with the
    # adequate classes.
    mitl_namespace = lookup.get_namespace(constants.MITL_NAMESPACE_URI)
    mitl_namespace[xml_tags.STATE_FORMULA_PLACEHOLDER_INSTANCE_TAG] = \
            StateFormulaPlaceholderInstance
    mitl_namespace[xml_tags.PATH_FORMULA_PLACEHOLDER_INSTANCE_TAG] = \
            PlaceholderInstance
    mitl_namespace[xml_tags.TIME_BOUND_PLACEHOLDER_INSTANCE_TAG] = \
            TimeBoundPlaceholderInstance
    # Associate elements representing PCTL placeholder instances with the
    # adequate classes.
    pctl_namespace = lookup.get_namespace(constants.PCTL_NAMESPACE_URI)
    pctl_namespace[xml_tags.STATE_FORMULA_PLACEHOLDER_INSTANCE_TAG] = \
            StateFormulaPlaceholderInstance
    pctl_namespace[xml_tags.PATH_FORMULA_PLACEHOLDER_INSTANCE_TAG] = \
            PlaceholderInstance
    pctl_namespace[xml_tags.TIME_BOUND_PLACEHOLDER_INSTANCE_TAG] = \
            TimeBoundPlaceholderInstance
    pctl_namespace[xml_tags.THRESHOLD_PLACEHOLDER_INSTANCE_TAG] = \
            ThresholdPlaceholderInstance
    pctl_namespace[xml_tags.COMPARISON_OPERATOR_PLACEHOLDER_INSTANCE_TAG] = \
            PlaceholderInstance
    # Associate elements representing CSL placeholder instances with the
    # adequate classes.
    csl_namespace = lookup.get_namespace(constants.CSL_NAMESPACE_URI)
    csl_namespace[xml_tags.STATE_FORMULA_PLACEHOLDER_INSTANCE_TAG] = \
            StateFormulaPlaceholderInstance
    csl_namespace[xml_tags.PATH_FORMULA_PLACEHOLDER_INSTANCE_TAG] = \
            PlaceholderInstance
    csl_namespace[xml_tags.TIME_BOUND_PLACEHOLDER_INSTANCE_TAG] = \
            TimeBoundPlaceholderInstance
    csl_namespace[xml_tags.THRESHOLD_PLACEHOLDER_INSTANCE_TAG] = \
            ThresholdPlaceholderInstance
    csl_namespace[xml_tags.COMPARISON_OPERATOR_PLACEHOLDER_INSTANCE_TAG] = \
            PlaceholderInstance
    return lookup


# ----------------------------------------------------------------------------
# Note: The following code will be executed once when this module is imported.
# ----------------------------------------------------------------------------

# Initialize the parser that is used to parse properties documents.
# XML comments must be discarded. Otherwise the parser would yield data
# structures cluttered with unwanted elements.
parser = etree.XMLParser(remove_comments=True)
# A lookup scheme is used in order to automatically map certain XML elements
# to instances of the adequate classes.
parser.set_element_class_lookup(_get_parser_lookup_scheme())

# Parse the XML Schema document that defines the structure of valid properties
# documents. This is needed for validation.
_properties_schema_tree = etree.parse(constants.PROPERTIES_XSD_PATH)
_properties_schema = etree.XMLSchema(_properties_schema_tree)
