##  ===========================================================================
##  Author: Roberto Cavada <cavada@fbk.eu>
##
##  Copyright (c) 2005 by FBK 
##
##  This is free software; you can redistribute it and/or
##  modify it under the terms of the GNU Lesser General Public
##  License as published by the Free Software Foundation; either
##  version 2 of the License, or (at your option) any later version.
##
##  This software is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  Lesser General Public License for more details.
##
##  You should have received a copy of the GNU Lesser General Public
##  License along with this library; if not, write to the Free Software
##  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA.
##
##  For more information about this software see <http://rat.fbk.eu>
##  or email to the RAT Team <rat@fbk.eu>.
##  Please report bugs to <rat@fbk.eu>.
##
##  ===========================================================================


from gtkmvc.controller import Controller
from models.entries import SignalsModel

import controllers.entry 
import views.entry
import models.entry

import utils.dialogs
import utils.globals

import gtk
import gtk.gdk

import os.path

## ----------------------------------------------------------------------
class EntriesCtrl (Controller):
    """Generic controller for those parts of the GUI that require
    an editable TreeView. See for examples frames for Signals,
    Requirements and Possibilities and Assertion within the Property
    Assurance section."""

    def __init__(self, model, parent_ctrl):
        Controller.__init__(self, model)

        self.pending_dialogs = {}  # the set of signals being created/edited
        self.parent_ctrl = parent_ctrl # to communicate with parent

        return

    def registerView(self, view):
        """Called by the view when it registers itself within self.
        We connect and setup the treeview at this stage"""
        Controller.registerView(self, view)

        self.setup_treeview(view)
        return

    def aux_data_func(self, tv, cell, model, iter, method):
        """Used to show separated parts of objects within the entry model"""
        cell.set_property('text', str(method(model.get_value(iter,0))))
        return

    def setup_treeview(self, view):
        # we create columns and renderers for the treeview

        tv = view['treeview']
        tv.set_model(self.model) # connects the model

        tv.get_selection().set_mode(gtk.SELECTION_MULTIPLE)
        
        rend = gtk.CellRendererText()
        col = gtk.TreeViewColumn("Name", rend)
        col.set_cell_data_func(rend, self.aux_data_func,
                               models.entry.EntryModel.get_name)
        col.set_resizable(True)        
        tv.append_column(col)
                
        rend = gtk.CellRendererText()
        col = gtk.TreeViewColumn("Notes", rend)
        col.set_cell_data_func(rend, self.aux_data_func,
                               models.entry.EntryModel.get_notes)
        col.set_resizable(True)        
        tv.append_column(col)
        
        return

  
    def get_selected_models(self):
        """Returns the list the iterators pointing to selected models"""
        res = []
        tv = self.view['treeview']
        sel = tv.get_selection()

        iter = self.model.get_iter_first()
        while iter:
            if sel.iter_is_selected(iter): res.append(iter)
            iter = self.model.iter_next(iter)
            pass

        return res

    def get_selected_checked_models(self):
        """Returns the list the iterators pointing to selected and checked models"""
        res = []
        tv = self.view['treeview']
        sel = tv.get_selection()

        iter = self.model.get_iter_first()
        while iter:
            if sel.iter_is_selected(iter) and self.model.get_value(iter, 1): res.append(iter)
            iter = self.model.iter_next(iter)
            pass

        return res

    def create_contextual_menu(self):
        """@return a (partially filled) menu for the sake of inherited classes"""

        tv = self.view['treeview']
        sel = tv.get_selection()

        menu = gtk.Menu()

        entries_num = self.model.get_entries_num()
        sels = len(self.get_selected_models())
        is_selected = sels > 0
        are_all_selected = (entries_num == sels)

        # selection submenu
        item_sel = gtk.MenuItem("Selection")
        item_sel.set_sensitive(entries_num > 0)
        item_sel.show()

        submenu = gtk.Menu()
        item = gtk.MenuItem("Select All")
        item.set_sensitive(not are_all_selected)
        item.connect("activate", lambda item, sel: sel.select_all(), sel)
        item.show()
        submenu.append(item)

        item = gtk.MenuItem("Unselect All")
        item.set_sensitive(is_selected)
        item.connect("activate", lambda item, sel: sel.unselect_all(), sel)
        item.show()
        submenu.append(item)

        item = gtk.MenuItem("Invert")
        item.connect("activate", self.select_invert, tv)
        item.show()
        submenu.append(item)

        item_sel.set_submenu(submenu)
        menu.append(item_sel)        
     
        return menu
    

    def type_check_entry(self, entry_view):
        """Called to check if the values entered are fine"""
        return True

    # ------------------------------------------------------------ #
    #                       User's Callbacks                       #
    # ------------------------------------------------------------ #
    def commit_entry(self, m):
        """A entry's asks to commit the corresponding model (passed
        as argument). Before adding the entry, type_check_entry will be
        called."""
        
        assert self.pending_dialogs.has_key(m)
        c,v,e = self.pending_dialogs[m]

        # checks if the model can be added or edited
        existing_model = self.model.get_entry_by_name(v.get_name())

        # not a repeated name:
        if (e and existing_model) or (not e and (existing_model and existing_model != m)):
            utils.dialogs.open_invalid_name_dialog()
            v['entry_name'].grab_focus()
            v['entry_name'].set_position(-1)

        elif self.type_check_entry(v):
            c.update_model() 
            
            if e: self.model.add_new_model(m)
            else: self.model.update_model(m)
            v.destroy()
            del self.pending_dialogs[m]
            pass
        
        return        

    def rollup_entry(self, m):
        assert self.pending_dialogs.has_key(m)
        c,v,e = self.pending_dialogs[m]
        v.destroy()
        del self.pending_dialogs[m]
        return
        

    def edit(self):
        iters = self.get_selected_models()
        for iter in iters: self.edit_entry_at(iter)
        return    

    # ------------------------------------------------------------ #
    #                        GUI's signals                         #
    # ------------------------------------------------------------ #
    def on_treeview_button_press_event(self, treeview, event):
        if event.button == 1 and event.type == gtk.gdk._2BUTTON_PRESS: 
            # double click, edits the selected entry
            pthinfo = treeview.get_path_at_pos(int(event.x), int(event.y))
            
            if pthinfo is None: return False
            
            self.edit_entry_at(self.model.get_iter(pthinfo[0]))
            return True
        
        if event.button == 3:
            # contextual menu:
            menu = self.create_contextual_menu()
            menu.popup(None, None, None, event.button, event.time)
            return True

        return False


    def select_invert(self, item, tv):
        m = tv.get_model()
        sel = tv.get_selection()
        iter = m.get_iter_first()
        while iter:
            if sel.iter_is_selected(iter): sel.unselect_iter(iter)
            else: sel.select_iter(iter)
            iter = m.iter_next(iter)
            pass
            
        return

    pass # end of class


## ----------------------------------------------------------------------
class SignalsCtrl (EntriesCtrl):
    def __init__(self, model, parent_ctrl):
        EntriesCtrl.__init__(self, model, parent_ctrl)
        return

    def setup_treeview(self, view):
        EntriesCtrl.setup_treeview(self, view)

        # Adds the column for type:
        tv = view['treeview']

        rend = gtk.CellRendererText()
        col = gtk.TreeViewColumn("Type", rend)
        col.set_cell_data_func(rend, self.aux_data_func,
                               models.entry.SignalModel.get_type)
        col.set_resizable(True)
        tv.insert_column(col, 1)

        rend = gtk.CellRendererText()
        col = gtk.TreeViewColumn("Kind", rend)
        col.set_cell_data_func(rend, self.aux_data_func,
                               models.entry.SignalModel.get_kind)
        tv.insert_column(col, 2)


        tv.get_selection().connect("changed", self.on_selection_changed)
        return

    def create_new(self):
        m = models.entry.SignalModel(self.model.get_new_instance_name("sgn"))
        c = controllers.entry.SignalCtrl(m, self)
        v = views.entry.SignalView(c, "Create a new signal", m.get_name())
        self.pending_dialogs[m] = (c,v,True) # 3rd arg is "creating mode"
        return

    def edit_entry_at(self, iter):
        m = self.model.get_value(iter, 0)
        if self.pending_dialogs.has_key(m):
            c,v,e = self.pending_dialogs[m]
            v.get_top_widget().present()
        else:
            c = controllers.entry.SignalCtrl(m, self)
            v = views.entry.SignalView(c, "Edit an existing signal", m.get_name())
            self.pending_dialogs[m] = (c,v,False) # 3rd arg is "editing mode"
            pass
        return        

    def delete(self):
        iters = self.get_selected_models()
        num = len(iters)
        if num == 0: return
        if num == 1: txt = "signal"
        else: txt = "signals"
        res = utils.dialogs.open_warning_request_cancel_ok(
                 "You are going to delete %d %s" % (num, txt),  
                 "After deletion is performed, the %s will be not recoverable." % txt)
        if not res: return

        # removes the selected models
        tv = self.view['treeview']
        sel = tv.get_selection()
        sel.unselect_all()

        for iter in iters:
            m = self.model.get_value(iter, 0)
            if self.pending_dialogs.has_key(m):
                c,v,e = self.pending_dialogs[m]
                v.destroy()
                del self.pending_dialogs[m]
                pass
            self.model.remove(iter)
            pass

        self.model.changed = True
        return


    def create_contextual_menu(self):
        menu = EntriesCtrl.create_contextual_menu(self)

        item = gtk.SeparatorMenuItem()
        item.show()
        menu.insert(item, 0)

        item = gtk.MenuItem("Create")
        item.connect("activate", self.parent_ctrl.on_button_new_signal_clicked)
        item.show()
        menu.insert(item, 0)

        is_selected = len(self.get_selected_models()) > 0

        item = gtk.MenuItem("Edit")
        item.connect("activate", self.parent_ctrl.on_button_edit_signal_clicked)
        item.set_sensitive(is_selected)
        item.show()
        menu.insert(item, 1)

        item = gtk.MenuItem("Delete")
        item.connect("activate", self.parent_ctrl.on_button_del_signal_clicked)        
        item.set_sensitive(is_selected)
        item.show()
        menu.insert(item, 2)

        return menu
        
    
    def type_check_entry(self, entry_view):
        """Called to check if the values entered are fine"""
        msg = self.model.type_check_signal(entry_view.get_name(), 
                                           entry_view.get_type())
        entry_view.show_error(msg)
        return msg is None
        

    # ------------------------------------------------------------ #
    #                          Signals                             #
    # ------------------------------------------------------------ #
    def on_selection_changed(self, selection):
        self.parent_ctrl.enable_signals_editing(len(self.get_selected_models()) > 0)
        return

    pass # end of class
    

## ----------------------------------------------------------------------
class PropertiesCtrl (EntriesCtrl):
    def __init__(self, model, parent_ctrl):
        EntriesCtrl.__init__(self, model, parent_ctrl)

        self.checked_all = False
        self.img_checked = gtk.image_new_from_stock('rat-stock-checked', gtk.ICON_SIZE_MENU)
        self.img_checked.show()

        self.img_unchecked = gtk.image_new_from_stock('rat-stock-unchecked', gtk.ICON_SIZE_MENU)
        self.img_unchecked.show()
        return

    def setup_treeview(self, view):
        EntriesCtrl.setup_treeview(self, view)

        tv = view['treeview']

        # Adds the column for toggling:
        rend = gtk.CellRendererToggle()
        #rend.set_property("activatable", True)
        rend.connect("toggled", self.on_check_toggled)
        col = gtk.TreeViewColumn("", rend, active=1, activatable=2)
        col.set_clickable(True)
        col.connect("clicked", self.on_col_toggled_clicked, 1)
        col.set_widget(self.img_checked) # set column pixmap
        tv.insert_column(col, 0)      
        
        # Adds the column for property:
        def prop_aux_data_func(tv, cell, model, iter):
            """Used to show separated parts of objects within the entry model"""
            prop = model.get_value(iter,0).get_property()

            cell.set_property('foreground-set', not model.get_value(iter, 2))
            cell.set_property('background-set', not model.get_value(iter, 2))
            cell.set_property('weight-set', not model.get_value(iter, 2))
            cell.set_property('strikethrough-set', not model.get_value(iter, 2))

            cell.set_property('text', str(prop)) 
            return

        rend = gtk.CellRendererText()
        rend.set_property('foreground', 'red')            
        rend.set_property('background', 'lightgray')            
        rend.set_property('weight', 600)
        rend.set_property('strikethrough', True)

        col = gtk.TreeViewColumn("Property", rend)
        col.set_cell_data_func(rend, prop_aux_data_func)
        col.set_resizable(True)        
        tv.insert_column(col, 1)

        tv.get_selection().connect("changed", self.on_selection_changed)
        return


    def create_contextual_menu(self):
        sel = self.get_selected_models()
        is_selected = len(sel) > 0

        menu = EntriesCtrl.create_contextual_menu(self)

        # Adds check/uncheck all to selection if there is a selection
        sels = self.get_selected_models()
        sels_num = len(sels)
        sel_checked_num = len(self.get_selected_checked_models())
        checked_num = len(self.model.get_checked_properties())
        all_num = self.model.get_entries_num()

        if sels_num > 0:
            # searches the selection menu item
            for item in menu.get_children():
                if item.get_child().get_text() == "Selection":
                    # Adds check/uncheck all item
                    menu_to_append = item.get_submenu()

                    # a separator
                    item = gtk.SeparatorMenuItem()
                    item.show()
                    menu_to_append.append(item)

                    # check all
                    item = gtk.MenuItem("Check All")
                    item.set_sensitive(sels_num > sel_checked_num)
                    item.show()
                    item.connect("activate", self.on_check_sel_activated, sels)
                    menu_to_append.append(item)

                    # uncheck all
                    item = gtk.MenuItem("Uncheck All")
                    item.set_sensitive(sel_checked_num > 0)
                    item.show()
                    item.connect("activate", self.on_uncheck_sel_activated, sels)
                    menu_to_append.append(item)
                    break                
                pass
            pass

        # a separator
        item = gtk.SeparatorMenuItem()
        item.show()
        menu.append(item)

        # check/uncheck all
        item = gtk.MenuItem("Check All")
        item.set_sensitive(all_num > checked_num)
        item.show()
        item.connect("activate", self.on_check_all_activated)
        menu.append(item)

        item = gtk.MenuItem("Uncheck All")
        item.set_sensitive(checked_num>0)
        item.show()
        item.connect("activate", self.on_check_all_activated)        
        menu.append(item)
        
        # a separator
        item = gtk.SeparatorMenuItem()
        item.show()
        menu.append(item)

        item = gtk.MenuItem("Load Into Simulation")
        item.set_sensitive(is_selected)
        item.connect("activate",
                     lambda item, sel: self.load_into_simulation(sel),
                     sel)
        item.show()
        menu.append(item)
        
        return menu


    def type_check_entry(self, entry_view):
        """Called to check if the values entered are fine"""
        msg = self.model.type_check_property(entry_view.get_property())
        entry_view.show_error(msg)
        return msg is None

    # ------------------------------------------------------------------ #
    #                          Callbacks                                 #
    # ------------------------------------------------------------------ #
    def load_into_simulation(self, sel=None):
        if sel is None: sel = self.get_selected_models()
        self.model.load_props_into_simulation(sel)
        return
        

    # ------------------------------------------------------------------ #
    #                          Signals handling                          #
    # ------------------------------------------------------------------ #

    def on_check_toggled(self, renderer, path):
        self.model.toggle_check(path)
        return

    def on_check_all_activated(self, item):
        tvcol = self.view['treeview'].get_column(0)
        tvcol.clicked()
        return

    def on_check_sel_activated(self, item, sels):
        """sels is a list of iterator to the selected models"""
        for iter in sels: 
            if self.model.get_value(iter, 2): self.model.set_value(iter, 1, True)
            pass
        return

    def on_uncheck_sel_activated(self, item, sels):
        """sels is a list of iterator to the selected models"""
        for iter in sels: self.model.set_value(iter, 1, False)
        return

    def on_col_toggled_clicked(self, tvcol, col_idx):
        self.checked_all = not self.checked_all
        iter = self.model.get_iter_first()
        while iter:
            if self.model.get_value(iter, 2): 
                self.model.set_value(iter, col_idx, self.checked_all)
                pass
            iter = self.model.iter_next(iter)
            pass

        if self.checked_all: tvcol.set_widget(self.img_unchecked)
        else: tvcol.set_widget(self.img_checked)
        return
    
    pass # end of class


## ----------------------------------------------------------------------
class RequirementsCtrl (PropertiesCtrl):
    def __init__(self, model, parent_ctrl):
        PropertiesCtrl.__init__(self, model, parent_ctrl)
        return

    def setup_treeview(self, view):
        PropertiesCtrl.setup_treeview(self, view)

        # Adds the column for type:
        tv = view['treeview']

        rend = gtk.CellRendererText()
        col = gtk.TreeViewColumn("Kind", rend)
        col.set_cell_data_func(rend, self.aux_data_func,
                               models.entry.RequirementModel.get_kind)
        tv.insert_column(col, 2)

        return


    def create_new(self):
        m = models.entry.RequirementModel(self.model.get_new_instance_name("req"))
        c = controllers.entry.RequirementCtrl(m, self)
        v = views.entry.RequirementView(c, "Create a new requirement", m.get_name())
        self.pending_dialogs[m] = (c,v,True) # 3rd is "creating mode"
        return

    def edit_entry_at(self, iter):
        m = self.model.get_value(iter, 0)
        if self.pending_dialogs.has_key(m):
            c,v,e = self.pending_dialogs[m]
            v.get_top_widget().present()
        else:
            c = controllers.entry.RequirementCtrl(m, self)
            v = views.entry.RequirementView(c, "Edit an existing requirement", m.get_name())
            self.pending_dialogs[m] = (c,v,False) # 3rd is "editing mode"
            pass
        return

    def delete(self):
        iters = self.get_selected_models()
        num = len(iters)
        if num == 0: return
        if num == 1: txt = "requirement"
        else: txt = "requirements"
        res = utils.dialogs.open_warning_request_cancel_ok(
                 "You are going to delete %d %s" % (num, txt),  
                 "After deletion is performed, the %s will be not recoverable." % txt)
        if not res: return

        # removes the selected models
        tv = self.view['treeview']
        sel = tv.get_selection()
        sel.unselect_all()

        for iter in iters:
            m = self.model.get_value(iter, 0)
            if self.pending_dialogs.has_key(m):
                c,v,e = self.pending_dialogs[m]
                v.destroy()
                del self.pending_dialogs[m]
                pass
            self.model.remove(iter)
            pass
        
        self.model.changed = True
        return

    def create_contextual_menu(self):
        menu = PropertiesCtrl.create_contextual_menu(self)

        item = gtk.SeparatorMenuItem()
        item.show()
        menu.insert(item, 0)

        item = gtk.MenuItem("Create")
        item.connect("activate", self.parent_ctrl.on_button_new_req_clicked)
        item.show()
        menu.insert(item, 0)

        is_selected = len(self.get_selected_models()) > 0

        item = gtk.MenuItem("Edit")
        item.connect("activate", self.parent_ctrl.on_button_edit_req_clicked)
        item.set_sensitive(is_selected)
        item.show()
        menu.insert(item, 1)

        item = gtk.MenuItem("Delete")
        item.connect("activate", self.parent_ctrl.on_button_del_req_clicked)        
        item.set_sensitive(is_selected)
        item.show()
        menu.insert(item, 2)

        return menu

    # ------------------------------------------------------------------ #
    #                          Signals handling                          #
    # ------------------------------------------------------------------ #

    def on_selection_changed(self, selection):
        self.parent_ctrl.enable_reqs_editing(len(self.get_selected_models()) > 0)
        return
    
    pass # end of class


## ----------------------------------------------------------------------
class PA_PropertyCtrl (PropertiesCtrl):

    """This class distinguish a normal PropertyCtrl from a controller
    specific for Property Assurance (simply to save code)
    """

    def __init__(self, model, parent_ctrl):
        super(PA_PropertyCtrl, self).__init__(model, parent_ctrl)

        abort_pixbuf = gtk.gdk.pixbuf_new_from_file(
            os.path.join(utils.globals.IMAGES_DIR, "buttons/gtk-error.png"))

        self.status2stock = {"passed" : ('stock-id', gtk.STOCK_YES),
                             "failed" : ('stock-id', gtk.STOCK_NO),
                             "error"  : ('stock-id', 'rat-stock-aborted'),
                             "stopped": ('stock-id', gtk.STOCK_STOP),
                             }
        return

    def aux_show_status_pix(self, col, cell, model, iter):
        st = self.model.get_value(iter, 0).status
        if self.status2stock.has_key(st): cell.set_property(*self.status2stock[st])
        else: cell.set_property('stock-id', None)
        return

    def aux_show_status_tex(self, col, cell, model, iter):
        st = self.model.get_value(iter, 0).status
        cell.set_property('text', st)
        return

    def setup_treeview(self, view):
        super(PA_PropertyCtrl, self).setup_treeview(view)

        tv = view['treeview']

        # Adds the column for status:
        # Pixbuf
        rendpix = gtk.CellRendererPixbuf()
        rendtex = gtk.CellRendererText()
        col = gtk.TreeViewColumn("Status")
        col.pack_start(rendpix)
        col.pack_start(rendtex)
        col.set_cell_data_func(rendpix, self.aux_show_status_pix)
        col.set_cell_data_func(rendtex, self.aux_show_status_tex)        
        tv.insert_column(col, 1)

        return

    pass # end of class


## ----------------------------------------------------------------------
class PossibilitiesCtrl (PA_PropertyCtrl):
    def __init__(self, model, parent_ctrl):
        PA_PropertyCtrl.__init__(self, model, parent_ctrl)
        return

    def create_new(self):
        m = models.entry.PossibilityModel(self.model.get_new_instance_name("poss"))
        c = controllers.entry.PropertyCtrl(m, self)
        v = views.entry.PossibilityView(c, "Create a new possibility", m.get_name())
        self.pending_dialogs[m] = (c,v,True) # 3rd is "creating mode"
        return

    def edit_entry_at(self, iter):
        m = self.model.get_value(iter, 0)
        if self.pending_dialogs.has_key(m):
            c,v,e = self.pending_dialogs[m]
            v.get_top_widget().present()
        else:
            c = controllers.entry.PropertyCtrl(m, self)
            v = views.entry.PossibilityView(c, "Edit an existing possibility", m.get_name())
            self.pending_dialogs[m] = (c,v,False) # 3rd is "editing mode"
            pass
        return 

    def delete(self):
        iters = self.get_selected_models()
        num = len(iters)
        if num == 0: return
        if num == 1: txt = "possibility"
        else: txt = "possibilities"
        res = utils.dialogs.open_warning_request_cancel_ok(
                 "You are going to delete %d %s" % (num, txt),  
                 "After deletion is performed, the %s will be not recoverable." % txt)
        if not res: return

        # removes the selected models
        tv = self.view['treeview']
        sel = tv.get_selection()
        sel.unselect_all()

        for iter in iters:
            m = self.model.get_value(iter, 0)
            if self.pending_dialogs.has_key(m):
                c,v,e = self.pending_dialogs[m]
                v.destroy()
                del self.pending_dialogs[m]
                pass
            self.model.remove(iter)
            pass

        self.model.changed = True
        return


    def create_contextual_menu(self):
        menu = PropertiesCtrl.create_contextual_menu(self)

        item = gtk.SeparatorMenuItem()
        item.show()
        menu.insert(item, 0)

        item = gtk.MenuItem("Create")
        item.connect("activate", self.parent_ctrl.on_button_new_poss_clicked)
        item.show()
        menu.insert(item, 0)

        is_selected = len(self.get_selected_models()) > 0

        item = gtk.MenuItem("Edit")
        item.connect("activate", self.parent_ctrl.on_button_edit_poss_clicked)
        item.set_sensitive(is_selected)
        item.show()
        menu.insert(item, 1)

        item = gtk.MenuItem("Delete")
        item.connect("activate", self.parent_ctrl.on_button_del_poss_clicked)        
        item.set_sensitive(is_selected)
        item.show()
        menu.insert(item, 2)

        return menu

    
    def aux_show_status_tex(self, col, cell, model, iter):
        st = self.model.get_value(iter, 0).status
        if st == "passed": cell.set_property('markup', '<span color="darkgreen">True</span>')
        elif st == "failed": cell.set_property('markup', '<span color="darkred">False</span>')
        elif st == "error": cell.set_property('markup', '<span background="red">Error</span>')
        elif st is not None: cell.set_property('markup', '<span background="lightgray">%s</span>' % st)
        else: cell.set_property('markup', st)
        return

    # ------------------------------------------------------------------ #
    #                          Signals handling                          #
    # ------------------------------------------------------------------ #

    def on_selection_changed(self, selection):
        self.parent_ctrl.enable_poss_editing(len(self.get_selected_models()) > 0)
        return
    
    pass # end of class


## ----------------------------------------------------------------------
class AssertionsCtrl (PA_PropertyCtrl):
    def __init__(self, model, parent_ctrl):
        PA_PropertyCtrl.__init__(self, model, parent_ctrl)
        return

    def create_new(self):
        m = models.entry.AssertionModel(self.model.get_new_instance_name("asrt"))
        c = controllers.entry.PropertyCtrl(m, self)
        v = views.entry.AssertionView(c, "Create a new assertion", m.get_name())
        self.pending_dialogs[m] = (c,v,True) # 3rd is "creating mode"
        return

    def edit_entry_at(self, iter):
        m = self.model.get_value(iter, 0)
        if self.pending_dialogs.has_key(m):
            c,v,e = self.pending_dialogs[m]
            v.get_top_widget().present()
        else:
            c = controllers.entry.PropertyCtrl(m, self)
            v = views.entry.AssertionView(c, "Edit an existing assertion", m.get_name())
            self.pending_dialogs[m] = (c,v,False) # 3rd is "editing mode"
            pass
        return

    def delete(self):
        iters = self.get_selected_models()
        num = len(iters)
        if num == 0: return
        if num == 1: txt = "assertion"
        else: txt = "assertions"
        res = utils.dialogs.open_warning_request_cancel_ok(
                 "You are going to delete %d %s" % (num, txt),  
                 "After deletion is performed, the %s will be not recoverable." % txt)
        if not res: return

        # removes the selected models
        tv = self.view['treeview']
        sel = tv.get_selection()
        sel.unselect_all()

        for iter in iters:
            m = self.model.get_value(iter, 0)
            if self.pending_dialogs.has_key(m):
                c,v,e = self.pending_dialogs[m]
                v.destroy()
                del self.pending_dialogs[m]
                pass
            self.model.remove(iter)
            pass

        self.model.changed = True
        return


    def create_contextual_menu(self):
        menu = PropertiesCtrl.create_contextual_menu(self)

        item = gtk.SeparatorMenuItem()
        item.show()
        menu.insert(item, 0)

        item = gtk.MenuItem("Create")
        item.connect("activate", self.parent_ctrl.on_button_new_assertion_clicked)
        item.show()
        menu.insert(item, 0)

        is_selected = len(self.get_selected_models()) > 0

        item = gtk.MenuItem("Edit")
        item.connect("activate", self.parent_ctrl.on_button_edit_assertion_clicked)
        item.set_sensitive(is_selected)
        item.show()
        menu.insert(item, 1)

        item = gtk.MenuItem("Delete")
        item.connect("activate", self.parent_ctrl.on_button_del_assertion_clicked)        
        item.set_sensitive(is_selected)
        item.show()
        menu.insert(item, 2)

        return menu


    def aux_show_status_tex(self, col, cell, model, iter):
        st = self.model.get_value(iter, 0).status
        if st == "passed": cell.set_property('markup', '<span color="darkgreen">True</span>')
        elif st == "failed": cell.set_property('markup', '<span color="darkred">False</span>')
        elif st == "error": cell.set_property('markup', '<span background="red">Error</span>')
        elif st is not None: cell.set_property('markup', '<span background="lightgray">%s</span>' % st)
        else: cell.set_property('markup', st)
        return


    # ------------------------------------------------------------------ #
    #                          Signals handling                          #
    # ------------------------------------------------------------------ #

    def on_selection_changed(self, selection):
        self.parent_ctrl.enable_asser_editing(len(self.get_selected_models()) > 0)
        return
    
    pass # end of class
