##  ===========================================================================
##  Author: Roberto Cavada <cavada@fbk.eu>
##
##  Copyright (c) 2007 by FBK 
##
##  This is free software; you can redistribute it and/or
##  modify it under the terms of the GNU Lesser General Public
##  License as published by the Free Software Foundation; either
##  version 2 of the License, or (at your option) any later version.
##
##  This software is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  Lesser General Public License for more details.
##
##  You should have received a copy of the GNU Lesser General Public
##  License along with this library; if not, write to the Free Software
##  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA.
##
##  For more information about this software see <http://rat.fbk.eu>
##  or email to the RAT Team <rat@fbk.eu>.
##  Please report bugs to <rat@fbk.eu>.
##
##  ===========================================================================


from gtkmvc.controller import Controller
from controllers.trace_compact import TraceCompactCtrl
import utils.globals
import gobject

## ----------------------------------------------------------------------
class OutcomeCtrl (Controller):
    """The controller of the component that shows the checking outcomes"""
    
    def __init__(self, model):
        """@param model It is assumed the model contains the
        properties and the assertions sub-models"""

        Controller.__init__(self, model)
        self.auto_expanded = False
        return

    def registerView(self, view):
        Controller.registerView(self, view)

        # connects the stub output text buffer to the corresponding view
        tv = view['textview_output']
        tv.set_buffer(self.model.stub_output_buffer)
        self.set_tv_scrolling(tv)

        return

    def set_tv_scrolling(self, tv):
        """Sets up the scrolling of the given TextView in order to
        show always the last line at the beginning or if the user
        scrolls to the last line. """

        buf = tv.get_buffer()
        end_mark = buf.create_mark("end", buf.get_end_iter(), False)
        buf.connect("changed", self.on_scrollable_tv_changed, tv, end_mark)
        return        


    def auto_expand_after_time(self, time):
        """after given msecs, opens the output window if the stub has
        not finished yet"""
        
        if not self.view['expander_out'].get_expanded():
            self.auto_expanded = False
            gobject.timeout_add(time, self.__on_checking_timeout)
            pass
        return
        

    # ------------------------------------------------------------ #
    #                            Signals                           #
    # ------------------------------------------------------------ #
    def on_expander_activate(self, exp):
        self.auto_expanded = False
        return
    
    def on_button_output_clean_clicked(self, button):
        buf = self.model.stub_output_buffer
        buf.delete(*buf.get_bounds())
        return        

    def on_button_tool_stop_clicked(self, button):
        self.model.stop_stub()
        return

    def on_scrollable_tv_changed(self, buf, tv, end_mark):
        vr = tv.get_visible_rect()

        end = buf.get_end_iter()
        ir = tv.get_iter_location(end)

        if vr.y + vr.height >= ir.y:
            tv.scroll_to_mark(end_mark, 0.05, True, 0.0, 1.0)
            pass
        return

    def on_hb_top_child_detached(self, hb, wid):
        w = self.view.to_window(wid, "Outcomes")
        w.connect("delete-event", self.on_window_destroy)
        return
        
    def on_window_destroy(self, w, event):
        self.view.to_embedded(w)
        return False

    def __on_checking_timeout(self):
        if self.model.busy_level > 0:
            self.view.show_outcomes(True)
            self.auto_expanded = True
            pass
        return False

    # ---------------------------------------- #
    #   Notifications for observed properties  #
    # ---------------------------------------- #

    def property_busy_level_change_notification(self, model, old, new):
        if old != new: self.view['button_tool_stop'].set_sensitive(new > 0)
        self.model.check_panel.busy_level = new
        if new == 0 and self.auto_expanded:
            self.view.show_outcomes(False)
            self.auto_expanded = False
            pass
        
        return

    def property_name_change_notification(self, model, old, new):
        if self.shown_traces.has_key(model):
            # the model is a shown trace, changes the page name:
            self.view.set_page_title(self.shown_traces[model], model.name)
            pass
        return
                
    pass # end of class



# ----------------------------------------------------------------------
class OutcomeTracesCtrl (OutcomeCtrl):
    """The controller of the component that shows the checking outcomes with
    traces"""
    
    def __init__(self, model):
        """@param model It is assumed the model contains the
        properties and the assertions sub-models"""

        OutcomeCtrl.__init__(self, model)

        # the traces currently shown in the outputs window:
        self.shown_traces = {} # associates trace instance and page number

        # self is an observer for the model of traces:
        self.model.traces.registerObserver(self)        
        return

    def registerView(self, view):
        OutcomeCtrl.registerView(self, view)
        return

    def show_trace(self, trace_model):
        if not self.shown_traces.has_key(trace_model):
            # self becomes an observer for this trace
            trace_model.registerObserver(self)

            # creates and attach a trace compact view on the trace
            ctrl = TraceCompactCtrl(trace_model, self.model.traces, self)
            pnum = self.view.add_trace(ctrl, trace_model.name)            
            self.shown_traces[trace_model] = pnum
        else: self.view.switch_to_page(self.shown_traces[trace_model])
        return

    def remove_trace(self, trace_model):
        """Removes the view over a trace from the outputs view. If the
        model is not shown, nothing happens"""
        
        if self.shown_traces.has_key(trace_model):
            pnum = self.shown_traces[trace_model]
            self.view.remove_page(pnum)
            del self.shown_traces[trace_model]
            trace_model.unregisterObserver(self)

            for key in self.shown_traces.keys():
                if self.shown_traces[key] > pnum: self.shown_traces[key] -= 1
                pass
            pass

        return
        

    # ------------------------------------------------------------ #
    #                            Signals                           #
    # ------------------------------------------------------------ #
    

    # ---------------------------------------- #
    #   Notifications for observed properties  #
    # ---------------------------------------- #

    def property_trace_added_change_notification(self, model, old, new):
        if new != old:
            # checks that the category is "new":
            cat_name = self.model.traces.get_value(self.model.traces.iter_parent(new), 0)
            if cat_name == utils.globals.CAT_NEW_NAME:
                # creates a new trace view and adds it to the outputs panel:
                self.show_trace(self.model.traces.get_value(new, 1))
                pass
            pass
        else: self.view.switch_to_page(0) # output
        return

    def property_trace_removed_change_notification(self, model, old, new):
        self.remove_trace(new) # if not shown, nothing will happen
        return            

                
    pass # end of class
    
