##  ===========================================================================
##  Author: Roberto Cavada <cavada@fbk.eu>
##
##  Copyright (c) 2005 by FBK 
##
##  This is free software; you can redistribute it and/or
##  modify it under the terms of the GNU Lesser General Public
##  License as published by the Free Software Foundation; either
##  version 2 of the License, or (at your option) any later version.
##
##  This software is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  Lesser General Public License for more details.
##
##  You should have received a copy of the GNU Lesser General Public
##  License along with this library; if not, write to the Free Software
##  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA.
##
##  For more information about this software see <http://rat.fbk.eu>
##  or email to the RAT Team <rat@fbk.eu>.
##  Please report bugs to <rat@fbk.eu>.
##
##  ===========================================================================

from gtkmvc.controller import Controller

from controllers.trace import TraceCtrl
from controllers.trace_info import TraceInfoCtrl
from controllers.category import CategoryCtrl

from views.trace_info import TraceInfoView
from views.category import CategoryView

import utils.strings
import utils.globals
import utils.dialogs


import gtk
import pango
import math

# ----------------------------------------------------------------------
class TracesCtrl (Controller):
    """This is the controller of the traces window"""

    def __init__(self, traces_model):
        Controller.__init__(self, traces_model)

        self.embedded_categories_names = []
        self.embedded_categories_ctrl = {}

        self.pending_trace_dialogs = {}
        self.pending_cat_dialogs = {}

        self.traces_checked_all = True
        self.cats_checked_all = False
        return

    def registerView(self, view):
        Controller.registerView(self, view)

        # to prevent destruction:
        w = view.get_top_widget()
        w.connect("delete-event", self.on_delete_event)
        
        self.setup_view(view)
        self.model.has_trace_window = True
        return

    def setup_view(self, view):
        tv = view['treeview_cats']
        tv.set_model(self.model.get_categories_model())
        sel = tv.get_selection()
        sel.set_mode(gtk.SELECTION_MULTIPLE)        
        sel.connect("changed", self.on_treeview_cats_sel_changed)
        
        # ------------------------------
        # cat selection, name and description
        cell = gtk.CellRendererToggle()
        cell.set_property('activatable', True)
        
        cell.connect("toggled", self.on_show_cat_toggled, tv.get_model())
        cell.set_property("yalign", 0)
        cell.set_property("ypad", 5)        

        col = gtk.TreeViewColumn('Show', cell)
        col.add_attribute(cell, "active", 2)
        col.set_clickable(True)
        col.connect("clicked", self.on_cat_toggled_clicked, (tv.get_model(), 2))
        
        tv.append_column(col)

        # name and description
        cell = gtk.CellRendererText()
        col = gtk.TreeViewColumn('Category', cell)
        col.set_cell_data_func(cell, self.show_category, tv)
        tv.append_column(col)

        # ------------------------------
        # traces names
        tv = view['treeview_selected_traces']
        tv.set_model(self.model.get_selected_traces_model())

        # selection
        sel = tv.get_selection()
        sel.set_mode(gtk.SELECTION_MULTIPLE)
        sel.connect("changed", self.on_selected_traces_sel_changed)

        cell = gtk.CellRendererToggle()
        cell.set_property('activatable', True)
        cell.connect("toggled", self.on_show_trace_toggled, tv.get_model())
        col = gtk.TreeViewColumn('Show', cell)
        col.add_attribute(cell, "active", 1)
        col.set_clickable(True)
        col.connect("clicked", self.on_trace_toggled_clicked, (tv.get_model(), 1))
        tv.append_column(col)

        cell = gtk.CellRendererText()
        col = gtk.TreeViewColumn('Name', cell)
        col.set_cell_data_func(cell, self.show_trace)
        tv.append_column(col)

        cell = gtk.CellRendererText()
        col = gtk.TreeViewColumn('Category', cell)
        col.add_attribute(cell, "text", 2)
        tv.append_column(col)

        # selects the first cat
        cats = self.model.get_categories_model()
        cats.set_value(cats.get_iter_first(), 2, True)
        self.model.update_selected_traces()
        self.update_selected_traces()
        self.traces_checked_all = True
        return

    def update_selected_traces(self):
        wanted_cats = self.model.get_selected_categories_names()

        # searches the categories that are to be removed and added 
        to_remove = []
        for c in self.embedded_categories_names:
            if not c in wanted_cats: to_remove.append(c)
            pass
        
        to_add = []
        for c in wanted_cats:
            if not self.embedded_categories_ctrl.has_key(c): to_add.append(c)
            pass

        # removes unwanted categories:
        for c in to_remove:
            self.view.remove_embedded_category_view(self.embedded_categories_ctrl[c])
            del self.embedded_categories_ctrl[c]
            pass

        # adds missing categories:
        for c in to_add:
            ctrl = EmbeddedCategoryCtrl(self.model, c)
            self.view.add_embedded_category_view(ctrl)
            self.embedded_categories_ctrl[c] = ctrl
            pass

        # reoders the children:
        self.embedded_categories_names = wanted_cats
        for i in range(len(wanted_cats)):
            c = wanted_cats[i]
            ctrl = self.embedded_categories_ctrl[c]
            self.view.set_embedded_category_pos(ctrl, i)
            pass

        # updates the categories that were already there
        for c in self.embedded_categories_names:
            if c not in to_add:
                ctrl = self.embedded_categories_ctrl[c]
                ctrl.update_selected_traces()
                pass
            pass
        
        return


    def category_renamed(self, old_name, new_name):
        if old_name in self.embedded_categories_names:
            self.embedded_categories_names[self.embedded_categories_names.index(old_name)] = new_name
            self.embedded_categories_ctrl[new_name] = self.embedded_categories_ctrl[old_name]
            del self.embedded_categories_ctrl[old_name]
            
            self.embedded_categories_ctrl[new_name].category_name_changed(new_name)
            pass
        return


    def edit_trace(self, trace):
        """@param trace is the TraceModel instance that must be edited"""
        if self.pending_trace_dialogs.has_key(trace):
            # already open dialog
            c, v = self.pending_trace_dialogs[trace]
            v.get_top_widget().present()
        else:
            c = TraceInfoCtrl(trace, self.model, self)
            v = TraceInfoView(c)
            self.pending_trace_dialogs[trace] = (c,v)
            pass       
        
        return
    

    def edit_category(self, cat_name):
        if self.pending_cat_dialogs.has_key(cat_name):
            # already open dialog
            c, v = self.pending_cat_dialogs[cat_name]
            v.get_top_widget().present()
        else:
            c = CategoryCtrl(self.model, cat_name, self)
            v = CategoryView(c)
            self.pending_cat_dialogs[cat_name] = (c,v)
            pass
        return

    def get_selected_cats(self):
        """Returns the list the iterators pointing to selected categories"""
        res = []
        tv = self.view['treeview_cats']
        sel = tv.get_selection()
        m = tv.get_model()
        
        iter = m.get_iter_first()
        while iter:
            if sel.iter_is_selected(iter): res.append(iter)
            iter = m.iter_next(iter)
            pass
        return res

    def get_selected_traces(self):
        """Returns the list the iterators pointing to selected categories"""
        res = []
        tv = self.view['treeview_selected_traces']
        sel = tv.get_selection()
        m = tv.get_model()
        
        iter = m.get_iter_first()
        while iter:
            if sel.iter_is_selected(iter): res.append(iter)
            iter = m.iter_next(iter)
            pass
        return res

    def create_cats_menu(self):
        """@return a contextual menu to be shown over the categories"""
        menu = gtk.Menu()

        tv = self.view['treeview_cats']
        m, paths = tv.get_selection().get_selected_rows()

        sel_cats = len(paths)
        can_edit = sel_cats > 0
        can_del = False

        # checks if can delete
        for path in paths:
            iter = m.get_iter(path)
            if m.get_value(iter, 3):
                can_del = True
                break
            iter = m.iter_next(iter)
            pass

        item = gtk.MenuItem("Create")
        item.connect("activate", self.on_button_new_cat_clicked)
        item.show()
        menu.append(item)

        item = gtk.MenuItem("Edit")
        item.connect("activate", self.on_button_edit_cat_clicked)
        item.set_sensitive(can_edit)
        item.show()
        menu.append(item)

        item = gtk.MenuItem("Delete")
        item.connect("activate", self.on_button_del_cat_clicked)
        item.set_sensitive(can_del)
        item.show()
        menu.append(item)

        # separator
        item = gtk.SeparatorMenuItem()
        item.show()
        menu.append(item)

        # check/uncheck all
        if self.cats_checked_all: item = gtk.MenuItem("Uncheck all")
        else: item = gtk.MenuItem("Check all")
        item.connect("activate", self.on_check_all_activated, tv)
        item.show()
        menu.append(item)

        self.__create_menu_common(menu, tv, sel_cats)
        return menu


    def create_traces_menu(self):
        """@return a contextual menu to be shown over the traces"""
        menu = gtk.Menu()

        tv = self.view['treeview_selected_traces']
        m, paths = tv.get_selection().get_selected_rows()

        sel_traces = len(paths)
        can_edit = sel_traces > 0
        can_del = sel_traces > 0

        item = gtk.MenuItem("Edit")
        item.connect("activate", self.on_button_edit_trace_clicked)
        item.set_sensitive(can_edit)
        item.show()
        menu.append(item)

        item = gtk.MenuItem("Delete")
        item.connect("activate", self.on_button_del_trace_clicked)
        item.set_sensitive(can_del)
        item.show()
        menu.append(item)

        # separator
        item = gtk.SeparatorMenuItem()
        item.show()
        menu.append(item)

        # check/uncheck all
        if self.traces_checked_all: item = gtk.MenuItem("Uncheck all")
        else: item = gtk.MenuItem("Check all")
        item.connect("activate", self.on_check_all_activated, tv)
        item.show()
        menu.append(item)


        self.__create_menu_common(menu, tv, sel_traces)
        return menu
    

    def __create_menu_common(self, menu, tv, sel_items_num):
        """Auxiliary function for construction of common parts of contextual menu"""
        sel = tv.get_selection()
        
        # separator
        item = gtk.SeparatorMenuItem()
        item.show()
        menu.append(item)

        # sub menu for selection:
        cats_num = self.model.get_cats_num()
        item_sel = gtk.MenuItem("Selection")
        item_sel.set_sensitive(cats_num > 0)
        item_sel.show()

        # selection submenu
        submenu = gtk.Menu()
        item = gtk.MenuItem("Select All")
        item.set_sensitive(sel_items_num < cats_num)
        item.connect("activate", lambda item, sel: sel.select_all(), sel)
        item.show()
        submenu.append(item)

        item = gtk.MenuItem("Unselect All")
        item.set_sensitive(sel_items_num > 0)
        item.connect("activate", lambda item, sel: sel.unselect_all(), sel)
        item.show()
        submenu.append(item)

        item = gtk.MenuItem("Invert")
        item.connect("activate", self.select_invert, tv)
        item.show()
        submenu.append(item)

        item_sel.set_submenu(submenu)
        menu.append(item_sel)
        return


    # ----------------------------------------------------------------------
    #               gtk signals and callbacks
    # ----------------------------------------------------------------------

    def on_check_all_activated(self, item, tv):
        tvcol = tv.get_column(0)
        tvcol.clicked()        
        return

    def select_invert(self, item, tv):
        m = tv.get_model()
        sel = tv.get_selection()
        iter = m.get_iter_first()
        while iter:
            if sel.iter_is_selected(iter): sel.unselect_iter(iter)
            else: sel.select_iter(iter)
            iter = m.iter_next(iter)
            pass
            
        return
    
    def show_category(self, col, cell, model, iter, tv):
        desc_style = "<span size='smaller' %s style='italic'>%s</span>"      
        
        cat_name = model.get_value(iter, 0)
        cat_desc = utils.strings.make_paragraph(model.get_value(iter, 1), 24)
        ntraces = model.get_value(iter, 4)
        if ntraces == 0: traces = "(empty)"
        elif ntraces == 1: traces = "(one trace)"
        else: traces = "(%d traces)" % ntraces
        
        line = "<span weight='bold'>%s</span>  <span foreground='gray40'>%s</span>\n" % (cat_name, traces)
        if tv.get_selection().iter_is_selected(iter): line += desc_style % ("", cat_desc)
        else: line += desc_style % ("foreground='gray40'", cat_desc) 

        cell.set_property('markup', line)
        return

    def show_trace(self, col, cell, model, iter):
        name = model.get_value(iter, 0).name
        cell.set_property('text', name)
        return
    
    def on_show_trace_toggled(self, cell, path, model):
        model[path][1] = not model[path][1]

        # updates only the category that the toggled trace belongs to:
        cat_name = model[path][2]
        ctrl = self.embedded_categories_ctrl[cat_name]
        ctrl.update_selected_traces()
        return

    def on_trace_toggled_clicked(self, tvcol, args):
        """@param args is a pair (model, column_index)"""
        m,idx = args
        self.traces_checked_all = not self.traces_checked_all
        iter = m.get_iter_first()
        while iter:
            m.set_value(iter, idx, self.traces_checked_all)
            iter = m.iter_next(iter)
            pass
        self.update_selected_traces()
        return

    def on_cat_toggled_clicked(self, tvcol, args):
        m,idx = args
        self.cats_checked_all = not self.cats_checked_all
        
        iter = m.get_iter_first()
        while iter:
            m.set_value(iter, idx, self.cats_checked_all)
            iter = m.iter_next(iter)
            pass

        self.model.update_selected_traces()
        self.update_selected_traces()
        self.traces_checked_all = True
        return

    def on_show_cat_toggled(self, cell, path, model):
        model[path][2] = not model[path][2]
        self.model.update_selected_traces()
        self.update_selected_traces()
        self.traces_checked_all = True
        return


    def on_delete_event(self, window, event):
        self.model.has_trace_window = False
        return True


    def on_treeview_selected_traces_button_press_event(self, tv, event):
        if event.button == 1 and event.type == gtk.gdk._2BUTTON_PRESS: 
            # double click, edits the selected entry
            pthinfo = tv.get_path_at_pos(int(event.x), int(event.y))
            
            if pthinfo is None: return False

            m = tv.get_model()
            trace = m.get_value(m.get_iter(pthinfo[0]), 0)
            self.edit_trace(trace)
            return True

        if event.button == 3:
            # contextual menu:
            menu = self.create_traces_menu()
            menu.popup(None, None, None, event.button, event.time)
            return True        
        
        return False
        

    def on_button_edit_trace_clicked(self, button):
        tv = self.view['treeview_selected_traces']
        sel = tv.get_selection()
        m, r = sel.get_selected_rows()
        for path in r:
            iter = m.get_iter(path)
            trace = m.get_value(iter, 0)
            self.edit_trace(trace)
            pass        
        
        return

    def on_button_del_trace_clicked(self, button):
        tv = self.view['treeview_selected_traces']
        sel = tv.get_selection()
        m, r = sel.get_selected_rows()

        to_be_trashed = []
        to_be_deleted = []
        for path in r:
            iter = m.get_iter(path)
            trace = m.get_value(iter, 0)

            cat_name = self.model.get_trace_category(trace)
            if cat_name != utils.globals.CAT_TRA_NAME: to_be_trashed.append((trace, cat_name))
            else: to_be_deleted.append(trace)
            pass

        # move to the trash
        for t, c in to_be_trashed:
            t.add_event_to_history("The user trashed this trace")
            self.model.remove_trace(t, c)
            self.model.add_trace(t, utils.globals.CAT_TRA_NAME)
            pass

        # destroy traces:
        num = len(to_be_deleted)
        if num > 0:
            if num == 1: txt = "trace" 
            else: txt = "traces"
            names = map(lambda t: t.name, to_be_deleted)
            names = utils.strings.make_paragraph(", ".join(names), 48)           
            res = utils.dialogs.open_warning_request_cancel_ok(
                "You are going to delete %d %s" % (num, txt),  
                """Traces are: %s
After deletion is performed, the %s will be not recoverable.""" % (names, txt))
            if res:
                for t in to_be_deleted:
                    # destroys the trace
                    self.model.remove_trace(t, utils.globals.CAT_TRA_NAME)

                    # close any open editing window
                    if self.pending_trace_dialogs.has_key(t):
                        c,v = self.pending_trace_dialogs[t]
                        del self.pending_trace_dialogs[t]
                        v.destroy()
                        pass
                    pass
                pass
            pass
        return


    def on_selected_traces_sel_changed(self, sel):
        selected = len(sel.get_selected_rows()[1]) > 0
        self.view['button_edit_trace'].set_sensitive(selected)
        self.view['button_del_trace'].set_sensitive(selected)
        return

    def on_treeview_cats_sel_changed(self, sel):
        m, paths = sel.get_selected_rows()
        can_del = False
        for path in paths:
            iter = m.get_iter(path)
            if m.get_value(iter, 3):
                can_del = True
                break
            iter = m.iter_next(iter)
            pass
    
        self.view['button_edit_cat'].set_sensitive(len(paths) > 0)
        self.view['button_del_cat'].set_sensitive(can_del)
        return

    def on_button_del_cat_clicked(self, button):
        to_be_deleted = []

        # collects the cateogies to be deleted
        tv = self.view['treeview_cats']
        sel = tv.get_selection()
        m, r = sel.get_selected_rows()
        for path in r:
            iter = m.get_iter(path)
            if m.get_value(iter, 3): to_be_deleted.append(m.get_value(iter, 0))
            pass    

        res, trash = open_confirm_delete_category_dialog(to_be_deleted)
        if not res: return

        for cat_name in to_be_deleted:
            if self.pending_cat_dialogs.has_key(cat_name):
                # close the open dialog
                c, v = self.pending_cat_dialogs[cat_name]
                v.destroy()
                pass

            self.model.remove_category(cat_name, trash)
            pass
         
        return

    def on_button_new_cat_clicked(self, button):
        c = CategoryCtrl(self.model, None, self)
        v = CategoryView(c)        
        return


    def on_treeview_cats_button_press_event(self, tv, event):
        if event.button == 1 and event.type == gtk.gdk._2BUTTON_PRESS: 
            # double click, edits the selected entry
            pthinfo = tv.get_path_at_pos(int(event.x), int(event.y))
            
            if pthinfo is None: return False

            m = tv.get_model()
            cat_name = m.get_value(m.get_iter(pthinfo[0]), 0)
            self.edit_category(cat_name)
            return True

        if event.button == 3:
            # contextual menu:
            menu = self.create_cats_menu()
            menu.popup(None, None, None, event.button, event.time)
            return True

        return False



    def on_button_edit_cat_clicked(self, button):
        tv = self.view['treeview_cats']
        sel = tv.get_selection()
        m, r = sel.get_selected_rows()
        for path in r:
            iter = m.get_iter(path)
            cat_name = m.get_value(iter, 0)
            self.edit_category(cat_name)
            pass
        
        return

        

    # ----------------------------------------------------------------------
    #               user's callback
    # ----------------------------------------------------------------------

    def trace_editing_terminated(self, model):
        """Called by the editing dialog when editing is over"""
        if self.pending_trace_dialogs.has_key(model):
            del self.pending_trace_dialogs[model]
            pass
        return

    def category_editing_terminated(self, cat_name):
        """Called by the editing dialog when editing is over"""
        if cat_name and self.pending_cat_dialogs.has_key(cat_name):
            del self.pending_cat_dialogs[cat_name]
            pass
        return

    # ----------------------------------------------------------------------
    #               Observable properties
    # ----------------------------------------------------------------------
    def property_trace_added_change_notification(self, model, old, new):
        """A trace has been added to the traces model, checks if
        there is something new to show"""
        self.model.update_selected_traces()
        self.update_selected_traces()
        return

    def property_trace_removed_change_notification(self, model, old, new):
        """A trace has been added to the traces model, checks if
        there is something new to show"""
        self.model.update_selected_traces()
        self.update_selected_traces()
        return

    def property_has_trace_window_change_notification(self, model, old, new):
        if new == old: return
        if new: self.view.show()
        else: self.view.hide()
        return


    pass # end of class


## ----------------------------------------------------------------------
class EmbeddedCategoryCtrl (Controller):
    """Controller for an EmbeddedCategoryView within a TracesView"""

    def __init__(self, traces_model, category_name):
        Controller.__init__(self, traces_model)

        self.cat_name = category_name

        self.embedded_traces_names = []
        self.embedded_traces_ctrl = {}
        return
    
    def registerView(self, view):
        Controller.registerView(self, view)

        self.setup_view(view)
        return

    def setup_view(self, view):
        self.set_cat_info()
        self.update_selected_traces()
        return

    def set_cat_info(self):
        cat_text = "<span size='larger' >%s  </span>" \
                   "<span foreground='gray30' size='smaller'>%s</span>"
        self.view['label_cat_text'].set_markup(cat_text % \
                                        (self.cat_name,
                                        self.model.get_category_description(self.cat_name)))
        return

    def category_name_changed(self, new_name):
        self.cat_name = new_name
        self.set_cat_info()
        return
    

    def update_selected_traces(self):
        sel = self.model.get_selected_traces_model()

        trace_names = []
        trace_models = {}
        
        iter = sel.get_iter_first()
        while iter:
            if sel.get_value(iter, 2) == self.cat_name and sel.get_value(iter, 1):
                trace_model = sel.get_value(iter, 0)
                trace_names.append(trace_model.name)
                trace_models[trace_model.name] = trace_model
                pass
            iter = sel.iter_next(iter)
            pass

        # collect the traces to remove and add:
        to_remove = []
        for t in self.embedded_traces_names:
            if not t in trace_names: to_remove.append(t)
            pass
        
        to_add = []
        for t in trace_names:
            if t not in self.embedded_traces_names: to_add.append(t)
            pass

        # removes unwanted traces:
        for t in to_remove:
            self.view.remove_embedded_trace_view(self.embedded_traces_ctrl[t])
            del self.embedded_traces_ctrl[t]
            pass

        # adds missing traces:
        for t in to_add:
            ctrl = EmbeddedTraceCtrl(trace_models[t])
            self.view.add_embedded_trace_view(ctrl)
            self.embedded_traces_ctrl[t] = ctrl
            pass

        # reoders the children:
        self.embedded_traces_names = trace_names
        for i in range(len(trace_names)):
            t = trace_names[i]
            ctrl = self.embedded_traces_ctrl[t]
            self.view.set_embedded_trace_pos(ctrl, i)
            pass

        return        

    pass # end of class
    



## ----------------------------------------------------------------------
class EmbeddedTraceCtrl (Controller):
    """This is a trace controller for begin embedded within an
    EmbeddedCategoryCtrl controller"""
    
    def __init__(self, trace_model):
        Controller.__init__(self, trace_model)

        self.trace = TraceCtrl(trace_model)
        return

    def registerView(self, view):
        Controller.registerView(self, view)

        self.setup_view(view)
        self.set_name_and_description()
        return
    
    def setup_view(self, view):

        # creates tags for showing trace description
        buf = view['textview_trace_info'].get_buffer()
        
        buf.create_tag("notes", background="lightyellow", foreground="darkgreen", editable=False)

        buf.create_tag("description", pixels_below_lines=2, editable=False)
         
        buf.create_tag("dependencies", family="Monospace", weight=pango.WEIGHT_BOLD, editable=False)
        buf.create_tag("history", family="Monospace", foreground="darkblue",
                       left_margin=6, editable=False)
        buf.create_tag("label", foreground="gray20", style=pango.STYLE_ITALIC,
                       variant=pango.VARIANT_SMALL_CAPS, editable=False)
        return


    def set_name_and_description(self):
        self.view['label_trace_name'].set_text(self.model.name)

        # fills up the description
        buf = self.view['textview_trace_info'].get_buffer()
        buf.set_text("")

        # steps nums:
        steps = self.model.get_max_steps()
        if steps == 0: sm = "no steps"
        elif steps == 1: sm = "one single step"
        else: sm = "%d steps" % steps
        buf.insert_with_tags_by_name(buf.get_end_iter(), "This trace contains " + sm, "description")

        # and loops:
        loops = len(self.model.loops)
        if loops == 0: sm = "no loop"
        elif loops == 1: sm = "one loop at step %s" % self.model.loops[0]
        else: sm = "multiple loops at steps " + ", ".join(map(str, self.model.loops))
        buf.insert_with_tags_by_name(buf.get_end_iter(), ", and has %s.\n" % sm, "description")

        # notes
        buf.insert_with_tags_by_name(buf.get_end_iter(), "Notes: ", "label")
        notes_start = buf.get_end_iter()
        if self.model.notes.strip():
            buf.insert_with_tags_by_name(notes_start, self.model.notes+"\n", "notes")
        else: buf.insert_with_tags_by_name(notes_start, "(This trace has no notes)\n", "notes")
        
        # dependencies:
        buf.insert_with_tags_by_name(buf.get_end_iter(), "Dependencies are: ", "label")
        buf.insert_with_tags_by_name(buf.get_end_iter(), ", ".join(self.model.dependencies),
                                     "dependencies")

        buf.insert_with_tags_by_name(buf.get_end_iter(), "\nHistory:\n", "label")

        hlen = len(self.model.history) # length of the history
        # number of digits for alignment:
        hstr = "%0" + "%dd: " % (int(math.log10(30)) + 1)
        for i in range(hlen):
            buf.insert_with_tags_by_name(buf.get_end_iter(),
                                         (hstr % i) + self.model.history[i] + "\n",
                                         "history")
            pass
                    
        return

    # -------------------------------------------------------------------- #
    # Observable properties                                                #
    # -------------------------------------------------------------------- #
    def property_name_change_notification(self, model, old, new):
        if new != old and model == self.model:
            self.view['label_trace_name'].set_text(new)
            pass
        return


    def property_description_changed_change_notification(self, model, old, new):
        """For some reason, it is needed a recalculationof trace's description"""
        self.set_name_and_description()
        return
    

    pass # end of class



# ----------------------------------------------------------------------
def open_confirm_delete_category_dialog(category_names):
    """This dialog is very specific of the traces controller.  @return
    a pair (res, must_trash) where res is True if OK has been
    pressed, and must_trash is True if all contained traces must be
    moved to the trash"""

    num = len(category_names)
    assert num > 0

    if num > 1:
        msg = "You are going to delete %d categories" % num
        details = utils.strings.make_paragraph("Categories names are: " + ", ".join(category_names),
                                               62)
        details += "\nAfter the removal, the categories will be no longer available."
    else: 
        msg = "You are going to delete one category" 
        details = "Category name is " + category_names[0]
        details += "\nAfter the removal, the category will be no longer available."
        pass

    c = utils.dialogs.PopupDialogCtrl()

    import views.dialogs    
    v = views.dialogs.QuestionConfirmCategoryRemoval(c, msg, details)
    res = v.run()
    return (res[0] == gtk.RESPONSE_OK, res[1])
