##  ===========================================================================
##  Authors: Roberto Cavada <cavada@fbk.eu>
##           Ingo Pill <ipill@ist.tugraz.at>
##
##  Copyright (c) 2005 by FBK
##  Copyright (c) 2005 by Graz University of Technology
##
##  This is free software; you can redistribute it and/or
##  modify it under the terms of the GNU Lesser General Public
##  License as published by the Free Software Foundation; either
##  version 2 of the License, or (at your option) any later version.
##
##  This software is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  Lesser General Public License for more details.
##
##  You should have received a copy of the GNU Lesser General Public
##  License along with this library; if not, write to the Free Software
##  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA.
##
##  For more information about this software see <http://rat.fbk.eu>
##  or email to the RAT Team <rat@fbk.eu>.
##  Please report bugs to <rat@fbk.eu>.
##
##  ===========================================================================


from gtkmvc.model import Model
from models.entries import SignalsModel, RequirementsModel
from models.prop_assurance import PropertyAssuranceModel
from models.prop_simulation import PropertySimulationModel
from models.realizability import RealizabilityModel
from models.traces import TracesModel

import models.options

from utils import globals

import utils.dialogs
import utils.utilities
import utils.launcher

import os
import os.path
import xml.dom


# ----------------------------------------------------------------------
class ProjectModel (Model): 
    """A project contains the state of a session. A project can be
    loaded and saved to a file.

    This model contains also a few models, that incapsulate the model
    of respectively the set of Signals, Requirements, Property
    Assurance (PA) and Property Simulation (PS). The model of PA
    contains sub models for the set of possibilities and assertions,
    """

    # -------------------------------------- #
    #      Set of observable properties      #
    # -------------------------------------- #    
    __properties__ = {
        "needs_to_be_saved" : False, # saving status
        "extra_info" : True, # project file holds extra information
        "filename" : None,
        "notes"    : "",    # notes about this project
        "is_new" : True,    # if the project is in its initial state
        "is_project_busy" : False,     # is the project is computing a long task
        "traces_window_shown" : False, # if the traces window is shown
        "active_view" : None, # if PA or PS is shown
        }

    
    # -------------------------------------- #
    #                Methods                 #
    # -------------------------------------- #
    
    def __init__(self, appl_model):
        Model.__init__(self)
               
        # categorized traces
        self.traces = TracesModel(self)
        self.traces.registerObserver(self) # self observes the traces model

        # signals
        self.signals = SignalsModel(self)
        self.signals.registerObserver(self)

        # property simulation
        self.prop_simulation = PropertySimulationModel()
        self.prop_simulation.registerObserver(self)

        # requirements
        self.requirements = RequirementsModel(self, 
                                              self.signals,
                                              self.traces,
                                              self.prop_simulation)
        self.requirements.registerObserver(self)

        # property assurance
        self.prop_assurance = PropertyAssuranceModel(self,
                                                     self.signals,
                                                     self.requirements,
                                                     self.traces,
                                                     self.prop_simulation)
        
        # This model observes the Property Assurance model
        self.prop_assurance.registerObserver(self)

        # realizability
        self.realizability = RealizabilityModel(self.signals, self.requirements)
        self.realizability.registerObserver(self)
        
        # options
        self.options = models.options.ProjectOptionsModel()
        self.application = appl_model
        return
    

    def quit(self):
        """Called when it is time to quit the application"""
        self.signals.quit()
        self.requirements.quit()
        self.prop_assurance.quit()
        return

    def property_changed_or_removed(self, prop_model):
        """A requirement or a property got modified or
        removed. This is used to keep track of all dependencies
        among properties, to reset results of model checking"""
        self.prop_assurance.property_changed_or_removed(prop_model)
        return

    def signal_changed_added_or_removed(self, sgn_model):
        self.requirements.type_check_all()
        self.prop_assurance.type_check_all()
        return
        

    def clear(self):
        """Cleans up the entire model"""
        self.Name = None
        self.signals.clear()
        self.requirements.clear()
        self.prop_assurance.clear()
        self.prop_simulation.clear()
        self.is_new = True
        return        

    def to_xml(self, args):
        thread = utils.launcher.FunctionThread(self.__to_xml,
                                               self.__to_xml_done, args)
        thread.start()
        self.is_project_busy = True
        return


    def __to_xml_done(self, res, args, exit_code):
        """Callback called when to_xml has done"""

        (filename, dummy) = args
        self.is_project_busy = False
        
        if exit_code == 0:
            if self.filename != filename: self.filename = filename
            self.needs_to_be_saved = False
        else:
            utils.dialogs.open_message_dialog("Could not save the project",
                         "An error occurred while trying to save the project");
            pass
        return
    

    def __to_xml(self, args):
        from xml.dom import getDOMImplementation
        impl = getDOMImplementation()
        root = impl.createDocument(None, "project", None)

        (filename, extra_info) = args

        self.signals.to_xml(root, root.documentElement)
        self.signals.changed = False
        
        self.requirements.to_xml(root, root.documentElement)
        self.requirements.changed = False
        
        self.prop_assurance.to_xml(root, root.documentElement)
        self.prop_assurance.changed = False
        
        self.prop_simulation.to_xml(root, root.documentElement)
        self.prop_simulation.changed = False

        # [MP] selectively include extra informations into xml project file
        if extra_info: 
            # traces
            self.traces.to_xml(root, root.documentElement)

            # active view
            node = root.createElement("active_view")
            node.appendChild(root.createTextNode(self.active_view))
            root.documentElement.appendChild(node)

            # notes
            notes_node = root.createElement("notes")
            notes_node.appendChild(root.createTextNode(self.notes))
            root.documentElement.appendChild(notes_node)        

        f = open(filename, "w")
        root.documentElement.writexml(f, addindent="  ", newl=os.linesep)
        f.close()

        root.unlink()
        return

    def from_xml(self, filename):
        from xml.dom.minidom import parse

        root = parse(filename)
        node = root.documentElement
        assert node.tagName == "project"

        signals_set = node.getElementsByTagName("signals")
        for sgns in signals_set:
            self.signals.from_xml(root, sgns)
            sgns.changed = False
            pass

        reqs_set = node.getElementsByTagName("requirements")
        for reqs in reqs_set:
            self.requirements.from_xml(root, reqs)
            reqs.changed = False
            pass

        pas = node.getElementsByTagName("property_assurance")
        for pa in pas:
            self.prop_assurance.from_xml(root, pa)
            pa.changed = False
            pass

        pss = node.getElementsByTagName("property_simulation")
        for ps in pss:
            self.prop_simulation.from_xml(root, ps)
            pa.changed = False
            pass

        catss = node.getElementsByTagName("categories")
        for cats in catss:
            self.traces.from_xml(root, cats)
            pass

        # notes and active_view
        m = { "notes" : "",
              "active_view" : ""
              }
        for child in node.childNodes:
            if child.nodeType == xml.dom.Node.ELEMENT_NODE and child.tagName in m:
                for text_child in child.childNodes:
                    if text_child.nodeType == xml.dom.Node.TEXT_NODE:
                        m[child.tagName] += text_child.data
                        pass
                    pass
                m[child.tagName] = m[child.tagName].strip()
                break
            pass

        self.notes = m["notes"]

        if m["active_view"]: self.active_view = m["active_view"]
        else: m["active_view"] = "pa"

        return


    def save_as(self):
        (fn, extra_info) = utils.dialogs.save_project_dialog(self.application.options.last_visited_path,
                                                             self.extra_info)
        if fn:
            # put extra_info into model
            self.extra_info = extra_info
            
            # adds the project extension if missing:
            fn = utils.utilities.normalize_project_filename(fn)
            self.application.options.last_visited_path = fn

            args = (fn, extra_info)
            self.to_xml(args)
            pass
        return

    def save(self):
        if not self.filename: self.save_as()
        else:
            args = (self.filename, self.extra_info)
            self.to_xml(args)
            pass
        return        

    def open(self, filename):        
        self.clear()
        self.from_xml(filename)
        self.filename = filename
        if self.needs_to_be_saved: self.needs_to_be_saved = False

        # type checks all
        self.requirements.type_check_all()
        self.prop_assurance.type_check_all()
        return


    # ---------------------------------------- #
    #   Notifications for observed properties  #
    # ---------------------------------------- #

    def property_changed_change_notification(self, model, old, new):
        """This notification arrives from signals, requirements,
        property assurance and simulation.  Propagates the notification
        only if needed."""
        if new and not self.needs_to_be_saved: self.needs_to_be_saved = True
        self.is_new = False
        return

    def property_busy_level_change_notification(self, model, old, new):
        if old != new:
            if new == 0 and self.is_project_busy: self.is_project_busy = False
            elif new > 0 and not self.is_project_busy: self.is_project_busy = True
            pass
        return

    def property_has_trace_window_change_notification(self, model, old, new):
        """The traces model has this property to know whether it has an associated
        traces window"""
        if new != old and new != self.traces_window_shown:
            self.traces_window_shown = new
            pass
        
        return

    def property_trace_edited_change_notification(self, model, old, new):
        self.needs_to_be_saved = True
        return

    def property_trace_added_change_notification(self, model, old, new):
        self.needs_to_be_saved = True
        return

    def property_trace_removed_change_notification(self, model, old, new):
        self.needs_to_be_saved = True
        return

    def property_user_category_change_notification(self, model, old, new):
        """The user changed, created or removed a category."""
        self.needs_to_be_saved = True
        return
        
    pass # end of class


