##  ===========================================================================
##  Author: Roberto Cavada <cavada@fbk.eu>
##
##  Copyright (c) 2006 by FBK 
##
##  This is free software; you can redistribute it and/or
##  modify it under the terms of the GNU Lesser General Public
##  License as published by the Free Software Foundation; either
##  version 2 of the License, or (at your option) any later version.
##
##  This software is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  Lesser General Public License for more details.
##
##  You should have received a copy of the GNU Lesser General Public
##  License along with this library; if not, write to the Free Software
##  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA.
##
##  For more information about this software see <http://rat.fbk.eu>
##  or email to the RAT Team <rat@fbk.eu>.
##  Please report bugs to <rat@fbk.eu>.
##
##  ===========================================================================

from gtkmvc.model import ListStoreModel
from models.check_panel import ReCheckingPanelModel
from models.synthesis_panel import SynthesisPanelModel
import utils.globals
from tools.stubs.nusmv import ReNuSMVStub
from tools.stubs.marduk_stub import MardukStub

import re
import os.path
import os
import gtk
import pango
import gobject


class RealizabilityModel (ListStoreModel):

    """Model for realizability. Contains the set of realizability
    problems"""

    # -------------------------------------- #
    #      Set of observable properties      #
    # -------------------------------------- #    
    __properties__ = {
        "changed" : False, # the status changed
        "busy_level" : 0     # some stub is in execution
        }

    def __init__(self, signals, requirements):
        """@param signals is the SignalsModel instance within the project model
        @param requirements is the RequirementsModel instance within the project model
        """
        
        ListStoreModel.__init__(self,
                                gobject.TYPE_INT,    # num
                                gobject.TYPE_STRING, # status
                                gobject.TYPE_PYOBJECT, # reqs (as strings)
                                gobject.TYPE_STRING, # flat model
                                gobject.TYPE_STRING, # synthesis file
                                gobject.TYPE_STRING  # synthesis options
                                )

        self.connect("row-changed", self.__on_row_changed)

        self.counter = 0        
        
        self.check_panel = ReCheckingPanelModel(self)
        self.synthesis_panel = SynthesisPanelModel(self)

        # this are used during model checking
        self.signals = signals
        self.requirements = requirements
        
        # ------------------------------
        #   Stubs and stub output      -
        # ------------------------------
        self.stub_output_buffer = gtk.TextBuffer()
        self.stub = ReNuSMVStub(utils.globals.NUSMV_PATH,
                                self.stub_output_buffer)

        workdirectory = utils.globals.MARDUK_SRC_DIR
        self.spec_file  = os.path.join(utils.globals.MARDUK_SRC_DIR, "_input.xml")
        self.marduk_stub = None #HIDE: MardukStub(utils.globals.MARDUK_PATH, self.stub_output_buffer, workdirectory, self.spec_file)

        # The stub that is currently running. (Either self.SYNTHESIS or self.REALIZABILITY)
        self.current_stub = None
        self.SYNTHESIS = 1
        self.REALIZABILITY = 2

        # Buffer for presenting results
        self.res_buffer = gtk.TextBuffer()
        self.res_buffer.create_tag("spacer", pixels_above_lines=12)        
        self.res_buffer.create_tag("title1", scale=pango.SCALE_LARGE,
                                   weight=pango.WEIGHT_BOLD,
                                   foreground="gray30", pixels_below_lines=8)
        self.res_buffer.create_tag("title2", scale=1.1,
                                   foreground="gray50", pixels_below_lines=5)
        self.res_buffer.create_tag("result_ok", family="Monospace",
                                   left_margin=12, foreground="dark green",
                                   weight=pango.WEIGHT_BOLD,
                                   underline=pango.UNDERLINE_SINGLE,
                                   pixels_below_lines=2)

        self.res_buffer.create_tag("result_ko", family="Monospace",
                                   left_margin=12, foreground="red",
                                   weight=pango.WEIGHT_BOLD,
                                   underline=pango.UNDERLINE_SINGLE,
                                   pixels_below_lines=2)

        self.res_buffer.create_tag("reqs", family="Monospace",
                                   left_margin=12, foreground="blue",
                                   pixels_below_lines=2)

        self.res_buffer.create_tag("flat-model", family="Monospace",
                                   left_margin=12, foreground="gray20",
                                   pixels_below_lines=2,
                                   scale=pango.SCALE_SMALL)

        self.res_buffer.create_tag("error-title", scale=pango.SCALE_LARGE,
                                   foreground="red", weight=pango.WEIGHT_BOLD,
                                   pixels_below_lines=6)
        self.res_buffer.create_tag("error-line", 
                                   foreground="darkred", weight=pango.WEIGHT_LIGHT)
        self.res_buffer.create_tag("error-text", 
                                   foreground="darkblue", family="Monospace")

        self.res_buffer.create_tag("operation", family="Monospace",
                                   foreground="gray40")

        self.res_buffer.create_tag("normal", family="Sans")
        self.res_buffer.create_tag("message", pixels_below_lines=0, foreground="gray30")

        return


    def append(self, status, reqs, flat_model, file=None, synthesis_options=None):
        assert (status in ("yes", "no"))
        self.counter += 1
        ListStoreModel.append(self, (self.counter, status, reqs, flat_model, file, synthesis_options))
        return

    def check_realizability(self, check_reqs_cons):
        assert (not check_reqs_cons) # not supported!

        # finds the list of names of selected requirements:
        reqs = map(lambda x: x[0].name,
                   self.requirements.get_checked_properties())

        if len(reqs) == 0:
            self.show_initial_message()
            return

        self.stub.set_end_session_callback(self.__realizability_stub_end,
                                           reqs)
                   
        self.__prepare_stub_session(self.stub)
        # launches the stub's session and exits (the result will arrive later
        # to __reqs_poss_stub_end):
        self.__insert_tag_text("Checking realizability...\n",
                               "operation", True)
        
        self.busy_level += 1
        self.current_stub = self.REALIZABILITY
        self.stub.run_session()
        return

    
    def do_synthesis(self, check_reqs_cons):
        assert (not check_reqs_cons) # not supported!

        if self.synthesis_panel.file_name == None: return
        
        # finds the list of names of selected requirements:
        reqs = map(lambda x: x[0].name,
                   self.requirements.get_checked_properties())

        if len(reqs) == 0:
            self.show_initial_message()
            return

        self.marduk_stub.set_end_session_callback(self.__synthesis_stub_end,
                                           reqs)
                   
        self.__prepare_synthesis_session(self.marduk_stub)
        # launches the stub's session and exits (the result will arrive later
        # to __reqs_poss_stub_end):
        self.__insert_tag_text("Performing synthesis...\n",
                               "operation", True)
        
        self.busy_level += 1
        self.current_stub = self.SYNTHESIS
        self.marduk_stub.run_session()
        return

    
    def stop_stub(self):
        if self.current_stub == self.model.SYNTHESIS: self.marduk_stub.stop_session()
        else: self.stub.stop_session()
        self.current_stub = None


    def show_realizable(self, iter): 
        num = self[iter][0]
        status = self[iter][1]
        reqs = self[iter][2]
        fmodel = self[iter][3]
        file = self[iter][4]
        options = self[iter][5]

        if reqs is None: return

        if file: # This means we are dealing with a synthesis result
            self.__insert_tag_text("Synthesis number %d\n" % num,
                                   "title1", True)
            if len(reqs) > 1: rtxt = "requirements were"
            else: rtxt = "requirement was"

            self.__insert_tag_text("Selected %s found " % rtxt, "normal")
            if status == "yes": self.__insert_tag_text("realizable\n", "result_ok")
            else: self.__insert_tag_text("not realizable\n", "result_ko")

            self.__insert_tag_text("%s: " % rtxt.capitalize(), "normal")
            self.__insert_tag_text(", ".join(reqs), "reqs")
            self.__insert_tag_text("\n", "reqs")
            self.__insert_tag_text("Synthesis result stored to %s.\n" % file, "normal")
            self.__insert_tag_text("Synthesis options were: %s." % options, "normal")
            if fmodel:
                self.__insert_tag_text("Flat Model was:\n", "spacer title2")
                self.__insert_tag_text(fmodel, "flat-model")
                pass
            
        else:       
            self.__insert_tag_text("Realizability number %d\n" % num,
                                   "title1", True)
            if len(reqs) > 1: rtxt = "requirements were"
            else: rtxt = "requirement was"

            self.__insert_tag_text("Selected %s found " % rtxt, "normal")
            if status == "yes": self.__insert_tag_text("realizable\n", "result_ok")
            else: self.__insert_tag_text("not realizable\n", "result_ko")

            self.__insert_tag_text("%s: " % rtxt.capitalize(), "normal")
            self.__insert_tag_text(", ".join(reqs), "reqs")
            self.__insert_tag_text("\n", "reqs")

            if fmodel:
                self.__insert_tag_text("Flat Model was:\n", "spacer title2")
                self.__insert_tag_text(fmodel, "flat-model")
                pass
        
        return

    def __prepare_synthesis_session(self, stub):

        """Prepares the synthesis stub for the session: submits the signals and
        requirments and options (getting them
        from the synthesis panel's model)"""

        # options
        file = self.synthesis_panel.file_name
        stub.add_opt("-i %s" % self.spec_file)
        stub.add_opt("-o %s" % file)
        opts = self.synthesis_panel.get_options()
        for opt in opts: stub.add_opt(opt)
        
        # signals and requirements setting:
        stub.set_signals_requirements(self.signals, self.requirements)
        return


    def __prepare_stub_session(self, stub):

        """Prepares the stub for the session: submits the signals and
        requirments and options and the checking command (getting them
        from the check panel's model)"""

        # options
        cmd, opts = self.check_panel.get_command_and_options()
        for opt in opts: stub.add_command(opt)
        
        # command for technology
        if self.check_panel.get_use_sat():
            stub.add_command("go_bmc")
            self.sat_length = self.check_panel.length
        else:
            stub.add_command("go")
            self.sat_length = None
            pass

        stub.add_command("write_flat_model")

        # checking command
        stub.set_checking_command(cmd)

        # signals and requirements setting:
        stub.set_signals_requirements(self.signals, self.requirements)
        return

    def __insert_tag_text(self, text, tags, clean=False):
        if clean: self.res_buffer.set_text("")
        self.res_buffer.insert_with_tags_by_name(self.res_buffer.get_end_iter(),
                                                 text, *tags.split())
        return
        

    def __show_error(self, errors):
        self.__insert_tag_text("Error\n", "error-title")
        for line, msg in errors:
            self.__insert_tag_text("At line %d: " % int(line), "error-line")
            self.__insert_tag_text("%s\n" % msg, "error-text")
            pass
        return            

    def __show_stopped(self):
        self.__insert_tag_text("Realizability Checking has been stopped\n",
                               "spacer error-title")
        return 

    def __show_synthesis_stopped(self):
        self.__insert_tag_text("Synthesis aborted!\n",
                               "spacer error-title", clean=True)
        self.__insert_tag_text("The synthesis process has been aborted, either manually or through an error.\n",
                               "error-text")
        self.__insert_tag_text("For the latter case check your stdout and stderr for more information on the error.\n",
                               "error-text")
        return 

    
    def show_initial_message(self):
        self.__insert_tag_text("""To check Realizability, select one or more
requirements, then press 'Check' button.""",
                               "message", True)
        return            

    def show_initial_synthesis_message(self):
        self.__insert_tag_text("""To synthesize, select one or more
requirements, then press 'Synthesize' button.""",
                               "message", True)
        return            


    def __process_results(self, ret, reqs):
        """Ret is the process returned value"""

        if ret is None:
            self.__show_stopped()
            return
        
        txt = self.stub_output_buffer.get_text(*self.stub_output_buffer.get_bounds())

        txt = txt[txt.rfind("# Starting a new session of "):]

        c = re.compile("^file .+: line (\d+): (.+)$", re.MULTILINE)
        errors = c.findall(txt)
        status = "error"

        idx = txt.find(": the strategy has been found")
        if idx >= 0:
            status = "yes"
            pass

        else:
            idx = txt.find(": no strategy exists")
            if idx >= 0:
                status = "no"
                pass
            pass

        # extracts the flat model
        _flat_begin='-- Begin MBP Flat Model'
        _flat_end='-- End MBP Flat Model'
        flat_model = ""
        idx2 = -1
        idx = txt.find(_flat_begin)
        if idx >=0: idx2 = txt[idx:].find(_flat_end)
        if idx2 >= 0:
            flat_model = txt[idx+len(_flat_begin)+1:idx+idx2].strip()
            flat_model = flat_model.replace("\n\n", "\n")
            pass

        if status == "error":
            self.res_buffer.set_text("")
            self.__show_error(errors)
            pass
        
        else: self.append(status, reqs, flat_model)     
        return

    
    def __process_synthesis_results(self, ret, reqs):
        """Ret is the process returned value"""
       
        if ret is None:
            self.__show_synthesis_stopped()
            return
        
        txt = self.stub_output_buffer.get_text(*self.stub_output_buffer.get_bounds())

        txt = txt[txt.rfind("# Starting a new session of "):]

        status = "error"

        idx = txt.find("FINISHED")
        if idx >= 0:
            status = "yes"
            pass
        else:
            idx = txt.find("NOT REALIZABLE")
            if idx >= 0:
                status = "no"
                pass
            pass
            
        if status == "error":
            self.__insert_tag_text("Error\n","error-title", clean=True)
            self.__insert_tag_text("Abnormal termination of synthesis process!", "error-text")
            #self.__show_error(errors)
            pass
        
        else: self.append(status, reqs, None, self.synthesis_panel.file_name, str( self.synthesis_panel.get_options()))     
        return

    
    # ---------------------------------------- #
    #   Notifications for observed properties  #
    # ---------------------------------------- #

    # --------------------------------------- #
    #              User callbacks             #
    # --------------------------------------- #
    def __realizability_stub_end(self, ret, res, reqs):
        """reqs is the list of selected requirement names"""
        self.__process_results(ret, reqs)
        self.busy_level -= 1
        self.current_stub = None
        return

    def __synthesis_stub_end(self, ret, res, reqs):
        """reqs is the list of selected requirement names"""
        self.__process_synthesis_results(ret, reqs)
        self.busy_level -= 1
        self.current_stub = None
        return

    
    def __on_row_changed(self, model, path, iter):
        self.show_realizable(iter)
        return

    pass # end of class

