##  ===========================================================================
##  Author: Roberto Cavada <cavada@fbk.eu>
##
##  Copyright (c) 2005 by FBK 
##
##  This is free software; you can redistribute it and/or
##  modify it under the terms of the GNU Lesser General Public
##  License as published by the Free Software Foundation; either
##  version 2 of the License, or (at your option) any later version.
##
##  This software is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  Lesser General Public License for more details.
##
##  You should have received a copy of the GNU Lesser General Public
##  License along with this library; if not, write to the Free Software
##  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA.
##
##  For more information about this software see <http://rat.fbk.eu>
##  or email to the RAT Team <rat@fbk.eu>.
##  Please report bugs to <rat@fbk.eu>.
##
##  ===========================================================================

from gtkmvc.model import Model, TreeStoreModel

import utils.globals

import xml.dom
import gtk
import gobject


# ----------------------------------------------------------------------
class TraceModel (TreeStoreModel):
    """The model of a single trace, intended as a set of stripes of
    signals or other.
    @see TraceStripeModel
    """

    # ----------------------------------------
    #  Observable properties
    # ----------------------------------------

    __properties__ = {
        "name" : None,
        "description_changed" : False,  # something changed in the description
        }

    def __init__(self, name=None, dependencies=None, formula=None):
        """@param dependencies is the list of the property names
        which self is built on.
        @param name if None, a name will be provided by the
        TracesModel this trace will be added to."""
        
        # set of TraceStripeModel's instances
        TreeStoreModel.__init__(self, gobject.TYPE_PYOBJECT) 

        if name: self.name = name

        self.history = [] # keeps track of what happend to this trace
        self.notes = ""
        self.loops = []
        self.max_loop = 0
        if dependencies is None: dependencies = []
        self.dependencies = dependencies

        self.creation_info = ""
        self.formula = formula        
        self.loops_sep = ", "
        return

    def add_event_to_history(self, text):
        self.history.append(text)
        self.description_changed = True
        return

    def set_dependencies(self, deps):
        if deps != self.dependencies:
            self.dependencies = deps
            self.description_changed = True
            pass
        return

    def get_dependencies(self): return self.dependencies

    def set_formula(self, formula): self.formula = formula
    def get_formula(self): return self.formula

    def get_creation_info(self): return self.creation_info
    def set_creation_info(self, info): self.creation_info = info

    def update_dependency(self, old_name, new_name):
        """Searches old_name in the dependencies, and if found
        substitutes it with new_name. @return True if substitution
        occurs"""
        try: idx = self.dependencies.index(old_name)
        except ValueError: return False

        self.dependencies[idx] = new_name
        self.description_changed = True
        return True
    

    def to_xml(self, root_doc, parent_node):
        top_node = root_doc.createElement("trace")

        # name
        name_node = root_doc.createElement("name")
        name_node.appendChild(root_doc.createTextNode(self.name))
        top_node.appendChild(name_node)

        # dependencies
        deps_node = root_doc.createElement("dependencies")
        
        for dep in self.dependencies:
            dep_node = root_doc.createElement("dependency")
            dep_node.appendChild(root_doc.createTextNode(dep))
            deps_node.appendChild(dep_node)
            pass
        top_node.appendChild(deps_node)

        # loops
        loops_node = root_doc.createElement("loops")
        loops_str = self.loops_sep.join(map(str, self.loops))
        loops_node.appendChild(root_doc.createTextNode(loops_str))
        top_node.appendChild(loops_node)

        # stripes:
        iter = self.get_iter_first() 
        self.to_xml_stripe(iter, root_doc, top_node)
                               
        # notes:
        notes_node = root_doc.createElement("notes")
        notes_node.appendChild(root_doc.createTextNode(self.notes))
        top_node.appendChild(notes_node)

        # creation_info:
        notes_node = root_doc.createElement("creation_info")
        notes_node.appendChild(root_doc.createTextNode(self.creation_info))
        top_node.appendChild(notes_node)       

        # originating property
        if self.formula: txt = self.formula
        else: txt = ""
        notes_node = root_doc.createElement("formula")
        notes_node.appendChild(root_doc.createTextNode(txt))
        top_node.appendChild(notes_node)

        # history
        notes_node = root_doc.createElement("history")
        notes_node.appendChild(root_doc.createTextNode("\\".join(self.history)))
        top_node.appendChild(notes_node)

        parent_node.appendChild(top_node)
        return top_node


    def to_xml_stripe(self, iter, root_doc, parent_node):
        """Recursive service for stripes"""
        while iter:
            stripe = self.get_value(iter, 0)
            stripe_node = stripe.to_xml(root_doc, parent_node)

            # recursive call
            iter_children = self.iter_children(iter)
            while iter_children:
                self.to_xml_stripe_aux(iter_children, root_doc, stripe_node)
                iter_children = self.iter_next(iter_children)
                pass

            iter = self.iter_next(iter)
            pass
        return
    

    def from_xml(self, root_doc, node):
        m = { "name" : None,
              "loops" : None,
              "notes" : None,
              "creation_info" : None,
              "formula" : None,
              "history" : None }
        
        for child in node.childNodes:
            if child.nodeType == xml.dom.Node.ELEMENT_NODE and m.has_key(child.tagName):
                m[child.tagName] = ""
                for text_child in child.childNodes:
                    if text_child.nodeType == xml.dom.Node.TEXT_NODE:
                        m[child.tagName] += text_child.data
                        pass
                    pass
                pass

            elif child.nodeType == xml.dom.Node.ELEMENT_NODE and child.tagName=="stripe":
                self.from_xml_stripe(root_doc, child, None)
                pass
            pass

        assert m["name"].strip()
        self.name = m["name"].strip()
        self.notes = m["notes"].strip()
        if m["creation_info"]: self.creation_info = m["creation_info"].strip()
        if m["formula"]: self.formula = m["formula"].strip()
        
        if m["history"]: self.history = m["history"].strip().split("\\")

        ls = m["loops"].strip() 
        if ls: self.loops = ls.split(self.loops_sep)

        # collects dependencies:
        deps = node.getElementsByTagName("dependency")
        for dep in deps:
            dep_str = ""
            for text_child in dep.childNodes:
                if text_child.nodeType == xml.dom.Node.TEXT_NODE: dep_str += text_child.data
                pass
            self.dependencies.append(dep_str.strip())
            pass
        
        return

    def from_xml_stripe(self, root_doc, node, iter):
        """Recursive service for nested stripes"""
        stripe = TraceStripeModel("")
        stripe.from_xml(root_doc, node)
        iter2 = self.append(iter)
        self.set_value(iter2, 0, stripe)

        for child in node.childNodes:
            if child.nodeType == xml.dom.Node.ELEMENT_NODE and child.tagName=="stripe":
                self.from_xml_stripe(root_doc, child, iter2)
                pass
            pass
        return


    def add_loop_at_step(self, step):
        """Adds a loop from the last state to the state at given step"""
        if step not in self.loops:
            self.loops.append(step)
            self.description_changed = True
            pass
        return

    def is_step_in_loop(self, step):
        """@return True if given step is within a loop"""
        for loop in self.loops:
            if step >= loop: return True
            pass
        return False

    def is_step_a_loop(self, step):
        """@return True if the given step is a loop step"""
        return step in self.loops
    
    def foreach_func(self, model, path, iter):
        self.max_loop = max(model.get_value(iter, 0).get_steps(), self.max_loop)
        return

    def get_max_steps(self):
        """Returns the maximum steps number in the trace. Traverses
        the entire model"""
        self.max_loop = 0
        self.foreach(self.foreach_func)
        return self.max_loop        

    pass # end of class


# ----------------------------------------------------------------------
class TraceStripeModel (Model):
    """A single labelled list of (implicitly timed) values."""

    def __init__(self, name):
        Model.__init__(self)

        self.name = name
        self.values = []
        self.__is_boolean = False
        return


    def to_xml(self, root_doc, parent_node):
        """@return the stripe node in the case sub stripes must be
        appended to it"""
        
        stripe_node = root_doc.createElement("stripe")

        # name
        name_node = root_doc.createElement("name")
        name_node.appendChild(root_doc.createTextNode(self.name))
        stripe_node.appendChild(name_node)

        # type
        type_node = root_doc.createElement("type")
        if self.__is_boolean: type_txt = "boolean"
        else: type_txt = "enumerative"
        type_node.appendChild(root_doc.createTextNode(type_txt))
        stripe_node.appendChild(type_node)

        # steps
        for idx in range(len(self.values)):
            step_node = root_doc.createElement("step")

            sval = self.values[idx]
            sname = "%d" % idx

            # step's name
            sname_node = root_doc.createElement("name")
            sname_node.appendChild(root_doc.createTextNode(sname))
            step_node.appendChild(sname_node)

            # step's value
            sval_node = root_doc.createElement("value")
            sval_node.appendChild(root_doc.createTextNode(sval))
            step_node.appendChild(sval_node)

            stripe_node.appendChild(step_node)
            pass
        
        # finally appends the stripe to the parent node. We return the stripe
        # node in the case this stripe contains sub stripes that have to be
        # appended to it
        parent_node.appendChild(stripe_node)     
        return stripe_node


    def from_xml(self, root_doc, node):
        m = { "name" : None,
              "type" : None}

        steps = {}

        for child in node.childNodes:
            if child.nodeType == xml.dom.Node.ELEMENT_NODE and m.has_key(child.tagName):
                m[child.tagName] = ""
                for text_child in child.childNodes:
                    if text_child.nodeType == xml.dom.Node.TEXT_NODE:
                        m[child.tagName] += text_child.data
                        pass
                    pass
                pass
            
            elif child.nodeType == xml.dom.Node.ELEMENT_NODE and child.tagName=="step":
                # retrieves a single (step_name, value) pair
                m2 = { "name" : None,
                       "value" : None}
                
                for child2 in child.childNodes:
                    if child2.nodeType == xml.dom.Node.ELEMENT_NODE and m2.has_key(child2.tagName):
                        m2[child2.tagName] = ""
                        for text_child in child2.childNodes:
                            if text_child.nodeType == xml.dom.Node.TEXT_NODE:
                                m2[child2.tagName] += text_child.data
                                pass
                            pass
                        pass
                    pass
                assert m2["name"].strip()
                steps[m2["name"].strip()]=m2["value"].strip()
                pass
            pass

        assert m["name"].strip()

        # sets the internals:
        self.name  = m["name"].strip()
        self.__is_boolean = (m["type"].strip().lower().rfind("boolean") >= 0)

        # reorders the steps
        step_nums = steps.keys()
        step_nums.sort()
        for step in step_nums: self.add_value(steps[step])    

        return
    

    def set_boolean(self, value=True): self.__is_boolean = value

    def is_boolean(self): return self.__is_boolean        

    def get_name(self):
        """@return the stripe's name"""
        return self.name

    def set_name(self, name):
        """Changes the stripe's name"""
        self.name = name
        return

    def get_steps(self):
        """@return the number of currently contained steps"""
        return len(self.values)

    def add_value(self, value):
        """Adds a new step, containing the given value"""
        self.values.append(value)
        return

    def remove_step(self, step):
        """Removes the value at the given step. Reduces the n,ber of steps."""
        del self.values[step]
        return
    
    def set_value_at_step(self, step, value):
        """Sets the value at the given step.
        @param step starts from 0"""
        self.values[step] = value
        return

    def get_value_at_step(self, step):
        """@return the value at given step (initial step is 0)"""
        return self.values[step]

    pass # end of class


