##  ===========================================================================
##  Author: Roberto Cavada <cavada@fbk.eu>
##
##  Copyright (c) 2005 by FBK 
##
##  This is free software; you can redistribute it and/or
##  modify it under the terms of the GNU Lesser General Public
##  License as published by the Free Software Foundation; either
##  version 2 of the License, or (at your option) any later version.
##
##  This software is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  Lesser General Public License for more details.
##
##  You should have received a copy of the GNU Lesser General Public
##  License along with this library; if not, write to the Free Software
##  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA.
##
##  For more information about this software see <http://rat.fbk.eu>
##  or email to the RAT Team <rat@fbk.eu>.
##  Please report bugs to <rat@fbk.eu>.
##
##  ===========================================================================

from gtkmvc.model import TreeStoreModel

from models.trace import TraceModel
import utils.globals

import xml.dom
import gtk
import gobject

# ----------------------------------------------------------------------
class TracesModel (TreeStoreModel):

    """The model of the set of categorized traces. Each category
    contains a set of traces that are associated to that category.

    Categories "Trash" and "New" are created by default.
    """

    # ---------------------------------------------------------- #
    #        Observable properties                               #
    # ---------------------------------------------------------- #
    __properties__ = {
        "user_category" : None, # name of a user-defined category
        "trace_edited" : False, # a trace has been changed
        "trace_added" : None,  # this is the iterator to the lastly added trace
        "trace_removed" : None, # this is the trace model that have been removed
        "has_trace_window" : False, # if this model has an associatd trace window
        }

    
    def __init__(self, parent_model=None):
        
        # set of TraceModel instances:
        TreeStoreModel.__init__(self,
                                gobject.TYPE_STRING,  # cat name
                                gobject.TYPE_PYOBJECT # trace
                                )

        self.cats = gtk.ListStore(gobject.TYPE_STRING,  # cat name
                                  gobject.TYPE_STRING,  # cat description
                                  gobject.TYPE_BOOLEAN, # show
                                  gobject.TYPE_BOOLEAN, # can edit
                                  gobject.TYPE_INT,     # num of traces  
                                  ) 

        self.selected_traces = gtk.ListStore(gobject.TYPE_PYOBJECT, # trace
                                             gobject.TYPE_BOOLEAN,  # selected
                                             gobject.TYPE_STRING,   # cat name
                                             )

        self.traces_counter = 0
        self.trace_names = []
        self.trace_names_cat = {} # associates category names and traces names
        
        self.add_standard_categories()
        return

    def add_standard_categories(self):
        # creates default category
        for cat in utils.globals.CATEGORIES:
            self.add_category(cat, utils.globals.CAT_DESCRIPTIONS[cat])
            pass
        return


    def is_there_category(self, name):
        """@return True is the given category exists, False otherwise"""
        return self.trace_names_cat.has_key(name)

    def get_cats_num(self):
        """@return the number of categories"""
        return len(self.trace_names_cat.keys())
    

    def change_category(self, old_name, new_name, new_description):
        """Modify information about an existing category. If the
        category is not editable, nothing will happen. Use
        add_category to add a new category if the category does not
        exists. new_name must contain a category name that does not exists"""

        assert self.is_there_category(old_name)
        if new_name != old_name: assert not self.is_there_category(new_name)
        
        traces = [] # the set of traces that belong to that category

        # possibly description and/or name
        iter = self.cats.get_iter_first()
        while iter:
            if self.cats.get_value(iter, 0) == old_name:
                if not self.cats.get_value(iter, 3): return # checks if can edit
                if old_name != new_name: self.cats.set_value(iter, 0, new_name)
                if self.cats.get_value(iter, 1) != new_description:
                    self.cats.set_value(iter, 1, new_description)
                    pass
                break
            iter = self.cats.iter_next(iter)
            pass
        
        if old_name != new_name:
            # name must change
            iter = self.get_iter_first()
            while iter:
                cat_name = self.get_value(iter, 0)
                if cat_name == old_name:
                    self.set_value(iter, 0, new_name)
                    triter = self.iter_children(iter)
                    while triter:
                        trace = self.get_value(triter, 1)
                        traces.append(trace)
                        triter = self.iter_next(triter)
                        pass
                    break
                iter = self.iter_next(iter)
                pass

            iter = self.selected_traces.get_iter_first()
            while iter:
                if self.selected_traces.get_value(iter, 2) == old_name:
                    self.selected_traces.set_value(iter, 2, new_name)
                    pass
                iter = self.selected_traces.iter_next(iter)
                pass

            # updates other internal structures:
            self.trace_names_cat[new_name] = self.trace_names_cat[old_name]
            del self.trace_names_cat[old_name]
            pass
        
        # touches the traces to obtain refresh of their description
        # (that contains category info):
        for trace in traces:
            if old_name != new_name: trace.add_event_to_history("Category '%s' has been renamed to '%s'" % (old_name, new_name))
            else: trace.description_changed = True
        self.user_category = new_name
        return

    def get_selected_traces_model(self): return self.selected_traces

    def get_trace_model(self, name):
        """@return a trace model intance from the trace name. If not found,
        None is returned"""
        iter = self.get_iter_first()
        while iter:
            triter = self.iter_children(iter)
            while triter:
                trace = self.get_value(triter, 1)
                if trace.name == name : return trace
                triter = self.iter_next(triter)
                pass
            
            iter = self.iter_next(iter)
            pass

        return None
    
    def get_selected_categories_names(self):
        """@return the list of the names of the currently selected categories"""

        sel_cats = []
        iter = self.cats.get_iter_first()
        while iter:
            if self.cats.get_value(iter, 2): sel_cats.append(self.cats.get_value(iter, 0))
            iter = self.cats.iter_next(iter)
            pass

        return sel_cats
        

    def get_trace_category(self, trace_model):
        """@return the category name whose the given trace belongs to,
        or None if not found"""
        iter = self.get_iter_first()
        while iter:
            cat_name = self.get_value(iter, 0)
            triter = self.iter_children(iter)
            while triter:
                if trace_model == self.get_value(triter, 1): return cat_name
                triter = self.iter_next(triter)
                pass
            
            iter = self.iter_next(iter)
            pass

        return None
        

    def get_new_trace_name(self):
        """Returns a suitable trace name for a new trace"""
        while True:
            name = "Trace_%d" % self.traces_counter
            if name not in self.trace_names: break
            self.traces_counter += 1
            pass
        
        return name

    def update_selected_traces(self):
        self.selected_traces.clear()

        sel_cats = self.get_selected_categories_names()
        
        # retrieves the traces set to be shown:
        iter = self.get_iter_first()
        while iter:
            cat_name = self.get_value(iter, 0)
            if cat_name in sel_cats:
                triter = self.iter_children(iter)
                while triter:
                    trace = self.get_value(triter, 1)
                    i = self.selected_traces.append()
                    self.selected_traces.set_value(i, 0, trace) # pushes the trace
                    self.selected_traces.set_value(i, 1, True)  # visible
                    self.selected_traces.set_value(i, 2, cat_name)  # cat_name
                    
                    triter = self.iter_next(triter)
                    pass
                pass
            iter = self.iter_next(iter)
            pass

        return                                                                 

    def get_categories_model(self):
        """@return the list of categories as ListStore object"""
        return self.cats

    def get_category_description(self, cat_name):
        """Returns None if the category does not exist"""
        iter = self.cats.get_iter_first()
        while iter:
            if self.cats.get_value(iter, 0) == cat_name: return self.cats.get_value(iter, 1)
            iter = self.cats.iter_next(iter)
            pass
        
        return None

    def add_category(self, name=None, desc=None, user_can_edit=False):
        """Creates a category and returns an iterator to it.
        If the category already exists return the iterator to it"""

        # resolves name and desc:
        if name is None:
            name = utils.globals.CAT_DEF_NAME
            desc = utils.globals.CAT_DESCRIPTIONS[name]
            pass

        # searches the already existing category
        iter = self.get_iter_first()
        while iter:
            if self.get_value(iter, 0) == name: return iter
            iter = self.iter_next(iter)
            pass

        # adds the new category
        iter = self.append(None)
        self.set_value(iter, 0, name)

        self.trace_names_cat[name] = []
        
        ic = self.cats.append(None)
        self.cats.set_value(ic, 0, name)
        self.cats.set_value(ic, 1, desc)
        self.cats.set_value(ic, 2, False)
        self.cats.set_value(ic, 3, user_can_edit)
        self.cats.set_value(ic, 4, 0)        

        if user_can_edit: self.user_category = name
        return iter

    def remove_category(self, name, trash_traces=True):

        """This applies only to those categories that are editable.
        @param trash_traces when True, forces to move all the
        contained traces to the trash"""
        assert self.is_there_category(name)
        
        traces = [] # the set of traces that belong to that category

        # possibly description and/or name
        iter = self.cats.get_iter_first()
        while iter:
            if self.cats.get_value(iter, 0) == name:
                if not self.cats.get_value(iter, 3): return # checks if can edit
                self.cats.remove(iter)
                break
            iter = self.cats.iter_next(iter)
            pass
        
        # name must change
        iter = self.get_iter_first()
        while iter:
            cat_name = self.get_value(iter, 0)
            if cat_name == name:
                triter = self.iter_children(iter)
                while triter:
                    trace = self.get_value(triter, 1)
                    traces.append(trace)
                    triter = self.iter_next(triter)
                    pass
                
                self.remove(iter)
                break
            
            iter = self.iter_next(iter)
            pass

        # updates other internal structures:
        del self.trace_names_cat[name]

        # recalculates selected traces
        self.update_selected_traces()
        
        # touches the traces to obtain refresh of their description
        # (that contains category info):
        for trace in traces:
            if trash_traces and name != utils.globals.CAT_TRA_NAME:
                self.remove_trace(trace, name)
                self.add_trace(trace, utils.globals.CAT_TRA_NAME)
            else: trace.description_changed = True # simple touch
            pass
        
        self.user_category = name
        return        
    
        
    def add_trace(self, trace_model, cat_name=None, track_event=True):
        """Adds a new trace to the given category. If the category does not exist,
        it creates a new category before adding the trace.
        If the trace is not assigned to a name, a new name will be chosen automatically.
        If the trace has a name that already exists in the category, an exception ValueError
        is raised.
        """
                
        if cat_name is None: cat_name = utils.globals.CAT_DEF_NAME
        if not trace_model.name: trace_model.name = self.get_new_trace_name()

        if trace_model.name in self.trace_names_cat[cat_name]:
            raise ValueError ("Category '%s' already contains trace '%s'" \
                              % (cat_name, trace_model.name))

        # self is an observer for this trace
        trace_model.registerObserver(self)

        cat_iter = self.add_category(cat_name)
        
        iter = self.append(cat_iter)
        self.set_value(iter, 1, trace_model)

        self.trace_names.append(trace_model.name)
        self.trace_names_cat[cat_name].append(trace_model.name)
        self.trace_added = iter

        # searches the category and updates the traces num:
        iter = self.cats.get_iter_first()
        while iter:
            if self.cats.get_value(iter, 0) == cat_name:
                num = self.cats.get_value(iter, 4)
                self.cats.set_value(iter, 4, num+1)
                break
            iter = self.cats.iter_next(iter)
            pass

        if track_event: trace_model.add_event_to_history("Added to category '%s'" % cat_name)      
        return
                
    def remove_trace(self, trace_model, cat_name=None, track_event=True):
        """Removes the given trace for the given category
        (default=default category). Nothing happens if the category or
        the trace do not exist"""

        if cat_name is None: cat_name = utils.globals.CAT_DEF_NAME

        # searches the category:
        iter = self.get_iter_first()
        while iter:
            if self.get_value(iter, 0) == cat_name: break
            iter = self.iter_next(iter)
            pass

        if iter is None: return # cat not found

        # now searches the trace
        triter = self.iter_children(iter)
        while triter:
            if self.get_value(triter, 1) == trace_model:

                self.remove(triter)
                self.trace_names.remove(trace_model.name)
                self.trace_names_cat[cat_name].remove(trace_model.name)
                self.trace_removed = trace_model
                if track_event: trace_model.add_event_to_history("Removed from category '%s'" % cat_name) 
                break
            triter = self.iter_next(triter)
            pass

        if triter:
            # searches the category and updates the traces num:
            iter = self.cats.get_iter_first()
            while iter:
                if self.cats.get_value(iter, 0) == cat_name:
                    num = self.cats.get_value(iter, 4)
                    self.cats.set_value(iter, 4, num-1)
                    break
                iter = self.cats.iter_next(iter)
                pass
            pass

        trace_model.unregisterObserver(self)
        return
    
        
    def to_xml(self, root_doc, parent_node):
        """See project.dtd to understand what happens here"""

        # Builds the xml DOM:
        top_node = root_doc.createElement("categories")
        
        iter = self.get_iter_first()
        while iter:
            # creates the category:
            cat_node = root_doc.createElement("category")

            cat_name = self.get_value(iter, 0)
            cat_can_edit = "yes"
            cat_desc = ""

            # searches the category info:
            iter_cat = self.cats.get_iter_first()
            while iter_cat:
                if self.cats.get_value(iter_cat, 0) == cat_name: # found
                    if self.cats.get_value(iter_cat, 3): cat_can_edit = "yes"
                    else: cat_can_edit = "no"
                    cat_desc = self.cats.get_value(iter_cat, 1)
                    pass
                iter_cat = self.cats.iter_next(iter_cat)
                pass
            
            # name
            name_node = root_doc.createElement("name")
            name_node.appendChild(root_doc.createTextNode(cat_name))
            cat_node.appendChild(name_node)

            # editable
            edit_node = root_doc.createElement("editable")
            edit_node.appendChild(root_doc.createTextNode(cat_can_edit))
            cat_node.appendChild(edit_node)

            # editable
            notes_node = root_doc.createElement("notes")
            notes_node.appendChild(root_doc.createTextNode(cat_desc))
            cat_node.appendChild(notes_node)
            
            # traces
            iter_trace = self.iter_children(iter)
            while iter_trace:
                trace = self.get_value(iter_trace, 1)
                # adds trace to the traces set:
                trace.to_xml(root_doc, cat_node)
                iter_trace = self.iter_next(iter_trace)
                pass

            top_node.appendChild(cat_node) # appends the category
            iter = self.iter_next(iter)
            pass
        
        parent_node.appendChild(top_node)      
        return

    def from_xml(self, root_doc, node):
        
        cats = node.getElementsByTagName("category")
        for cat in cats:

            # Collects category's info and creates it
            m = { "name" : None,
                  "editable" : None,
                  "notes" : None }

            for child in cat.childNodes:
                if child.nodeType == xml.dom.Node.ELEMENT_NODE and m.has_key(child.tagName):
                    m[child.tagName] = ""
                    for text_child in child.childNodes:
                        if text_child.nodeType == xml.dom.Node.TEXT_NODE:
                            m[child.tagName] += text_child.data
                            pass
                        pass
                    pass
                pass
            
            assert m['name'].strip()
            
            self.add_category(m['name'].strip(),
                              m['notes'].strip(),
                              m['editable'].strip() == "yes")

            # adds the traces
            traces = cat.getElementsByTagName("trace")
            for trace_node in traces:
                trace = TraceModel()
                trace.from_xml(root_doc, trace_node)
                try: self.add_trace(trace, m['name'].strip(), track_event=False)
                except ValueError: pass
                pass

            pass            
        
        return

    def changed_property_name(self, prop_model, old_name):
        """This is called when a property name has been changed.  The
        model must check if the property occurs into the traces, and
        update the dependencies accordingly"""

        updated = 0

        # searches the property within the traces' dependencies
        iter = self.get_iter_first()
        while iter:
            iter2 = self.iter_children(iter)
            while iter2:
                trace = self.get_value(iter2, 1)
                if trace.update_dependency(old_name, prop_model.name):
                    updated += 1
                    trace.add_event_to_history("Dependency '%s' has been renamed to '%s'" \
                                               % (old_name, prop_model.name))
                    pass
                
                iter2 = self.iter_next(iter2)
                pass                    
            
            iter = self.iter_next(iter)
            pass

        return

    def changed_property_specification(self, prop_model, old_spec):
        """Called when a property's specification is changed. This
        invalidates those traces that depend on that property."""
        updated = 0

        # this is to avoid that a moved trace can be tracked twice or more:
        changed_traces = []
 
        # searches the property within the traces' dependencies
        iter = self.get_iter_first()
        while iter:
            iter2 = self.iter_children(iter)
            cat_name = self.get_value(iter, 0)
            while iter2:
                trace = self.get_value(iter2, 1)
                if prop_model.name in trace.get_dependencies() and trace not in changed_traces:
                    updated += 1
                    changed_traces.append(trace)
                    if cat_name != utils.globals.CAT_OOD_NAME:
                        trace.add_event_to_history(
                            "Property '%s' has been changed, making this trace potentially invalid." \
                            % prop_model.name)
                        # moves the trace
                        iter2 = self.iter_next(iter2)
                        self.remove_trace(trace, cat_name)
                        self.add_trace(trace, utils.globals.CAT_OOD_NAME)
                        continue
                    else: trace.add_event_to_history("Property '%s' has been changed again." % prop_model.name)
                    pass                
                
                iter2 = self.iter_next(iter2)
                pass                    
            
            iter = self.iter_next(iter)
            pass

        return


    def deleted_property(self, prop_model):
        """Called when a property is removed. This invalidates
        those traces that depend on that property."""
        # this is to avoid that a moved trace can be tracked twice or more:
        changed_traces = []
 
        # searches the property within the traces' dependencies
        iter = self.get_iter_first()
        while iter:
            iter2 = self.iter_children(iter)
            cat_name = self.get_value(iter, 0)
            while iter2:
                trace = self.get_value(iter2, 1)
                if prop_model.name in trace.get_dependencies() and trace not in changed_traces:
                    changed_traces.append(trace)
                    if cat_name != utils.globals.CAT_OOD_NAME:
                        trace.add_event_to_history(
                            "Property '%s' has been deleted, making this trace invalid." \
                            % prop_model.name)
                        # moves the trace
                        iter2 = self.iter_next(iter2)
                        self.remove_trace(trace, cat_name)
                        self.add_trace(trace, utils.globals.CAT_OOD_NAME)
                        continue
                    pass                
                
                iter2 = self.iter_next(iter2)
                pass                    
            
            iter = self.iter_next(iter)
            pass

        return


    def invalidated_property(self, prop_model):
        """Called when a property is invalidated, due to failed
        syntax/type checking. This invalidates those traces that
        depend on that property."""
        # this is to avoid that a moved trace can be tracked twice or more:
        changed_traces = []
 
        # searches the property within the traces' dependencies
        iter = self.get_iter_first()
        while iter:
            iter2 = self.iter_children(iter)
            cat_name = self.get_value(iter, 0)
            while iter2:
                trace = self.get_value(iter2, 1)
                if prop_model.name in trace.get_dependencies() and trace not in changed_traces:
                    changed_traces.append(trace)
                    if cat_name != utils.globals.CAT_OOD_NAME:
                        trace.add_event_to_history(
                            "Property '%s' has been invalidated, making this trace invalid." \
                            % prop_model.name)
                        # moves the trace
                        iter2 = self.iter_next(iter2)
                        self.remove_trace(trace, cat_name)
                        self.add_trace(trace, utils.globals.CAT_OOD_NAME)
                        continue
                    pass                
                
                iter2 = self.iter_next(iter2)
                pass                    
            
            iter = self.iter_next(iter)
            pass

        return


    # ----------------------------------------------------------------------
    #               Observable properties
    # ----------------------------------------------------------------------
    def property_name_change_notification(self, model, old, new):

        if old in self.trace_names and old != new:
            cat_name = None
            
            # updates the lists to show the change, and searches the category
            go = True
            iter = self.get_iter_first()
            while iter and go:
                iter2 = self.iter_children(iter)
                while iter2:
                    trace = self.get_value(iter2, 1)
                    if trace == model:
                        cat_name = self.get_value(iter, 0)
                        self.row_changed(1, iter2)
                        go = False
                        break
                    iter2 = self.iter_next(iter2)
                    pass
                iter = self.iter_next(iter)
                pass

            assert cat_name

            # updates the internal state:
            self.trace_names[self.trace_names.index(old)] = new
            self.trace_names_cat[cat_name][self.trace_names_cat[cat_name].index(old)] = new
            
            # updates the lists to show the change:
            iter = self.selected_traces.get_iter_first()
            while iter:
                trace = self.selected_traces.get_value(iter, 0)
                if trace == model:
                    self.selected_traces.row_changed(0, iter)
                    break

                iter = self.selected_traces.iter_next(iter)
                pass

            pass

        return


    def property_description_changed_change_notification(self, model, old, new):
        """Trace description changed for saome reason"""
        self.trace_edited = True
        return

    pass # end of class
