##  ===========================================================================
##  Author: Roberto Cavada <cavada@fbk.eu>
##
##  Copyright (c) 2005 by FBK 
##
##  This is free software; you can redistribute it and/or
##  modify it under the terms of the GNU Lesser General Public
##  License as published by the Free Software Foundation; either
##  version 2 of the License, or (at your option) any later version.
##
##  This software is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  Lesser General Public License for more details.
##
##  You should have received a copy of the GNU Lesser General Public
##  License along with this library; if not, write to the Free Software
##  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA.
##
##  For more information about this software see <http://rat.fbk.eu>
##  or email to the RAT Team <rat@fbk.eu>.
##  Please report bugs to <rat@fbk.eu>.
##
##  ===========================================================================


import os, threading, locale
import gobject
import utils.compat

if hasattr(gobject, "threads_init"): gobject.threads_init()
else: import gtk; gtk.threads_init()

# ----------------------------------------------------------------------
class LauncherThread (threading.Thread):
    """This class implements a launcher for a shell command. The
    command is executed in the background and its output (both stdout
    and stderr) are optionally sent to a text buffer.

    After the initialization, run launch_command
    """
    
    def __init__(self, buffer=None, end_callback=None, user_arg=None):

        """
           @param buffer is the optional gtk.TextBuffer to send the command output
           @param end_callback is a function to be called just before the thread
                  ends. The callback receives an exit code and the provided optional
                  user_arg. The exit code is the process exit code, or None if the
                  thread has been killed by the user by using the method kill.
                  
           @param user_arg is the optional argument to be passed to end_callback
                  when called. 
                  """        

        threading.Thread.__init__(self)

        if hasattr(locale, "getpreferredencoding"): self.encoding = locale.getpreferredencoding()
        else: self.encoding = None

        self.buffer = buffer
        
        self.command = None
        self.pipe = None
        self.end_callback = end_callback
        self.user_arg = user_arg
        self.killed = False
        self.workdirectory = None
        return

    def idle_add_line(self, line):
        """This function MUST return False, as it is called by the
        IDLE function og gtk"""

        if self.buffer:
            if self.encoding: text = unicode(line, self.encoding).encode('utf8')
            else: text = unicode(line).encode('utf8')
            self.buffer.insert(self.buffer.get_end_iter(), text)
            pass
                
        return False # this is mandatory (idle function)

    def __call_end_callback(self):
        if self.end_callback is None: return

        if self.killed: ret_val = None
        else: ret_val = self.pipe.wait()

        self.end_callback(ret_val, self.user_arg)
        return 

    def launch_command(self, command):
        """@return True if started, False otherwise"""
        if self.isAlive(): return False
        self.command = command
#         if self.workdirectory != None:
#             print self.workdirectory
#         else:
#             print "No workdir!"
        self.start()
        return True

    def run(self):
        if self.command is None: return
        self.killed = False

        self.pipe = utils.compat.create_pipe(self.command, self.workdirectory)

        if self.pipe == None:
            self.killed = True
            self.__call_end_callback()
            return

        line = self.pipe.stdout.readline()
        while line:
            gobject.idle_add(self.idle_add_line, line)
            line = self.pipe.stdout.readline()
            pass

        self.__call_end_callback()
        return

    def kill(self):
        self.killed = True
        utils.compat.kill_pipe(self.pipe)        
        return

    pass # end of class


# ----------------------------------------------------------------------
class FunctionThread (threading.Thread):

    """This class executes a function in background. Then the
    function's execution is done, a notification function is called
    within the gtk mainloop."""

    def __init__(self, function, end_callback=None, user_arg=None):

        """
           @param function is the function to be executed
                  function receives the value provided by user_arg
           @param end_callback is a function to be called just before the thread
                  ends. The callback cb(res_val, user_arg, exit_code) receives:
                  - the value returned by the executed function
                  - the value provided by the user through user_arg
                  - an exit code:
                    0: function regularly executed
                    1: function thrown an exception
                                      
           @param user_arg is the optional argument to be passed to both the
           executed function and the end_callback when called. 
           """        

        threading.Thread.__init__(self)

        self.fun = function
        self.end_callback = end_callback
        self.user_arg = user_arg
        self.exit_code = 0
        self.fun_res = None
        return


    def start(self):
        """@return True if started, False otherwise"""
        if self.isAlive(): return False
        threading.Thread.start(self)
        return True

    def run(self):
        if self.fun is None: return

        self.fun_res = None
        self.exit_code = 0
        try:
            if self.user_arg is None: self.fun_res = self.fun()
            else: self.fun_res = self.fun(self.user_arg)
        except: self.exit_code = 1           

        # calls the callback from the main gtk loop 
        if self.end_callback: gobject.idle_add(self.__call_end_callback_wrapper)
        return

    def __call_end_callback_wrapper(self):
        """This method is always executed from the main gtk loop"""
        self.end_callback(self.fun_res, self.user_arg, self.exit_code)            
        return False # to close idle_add

    pass # end of class
