##  ===========================================================================
##  Author: Roberto Cavada <cavada@fbk.eu>
##
##  Copyright (c) 2005 by FBK 
##
##  This is free software; you can redistribute it and/or
##  modify it under the terms of the GNU Lesser General Public
##  License as published by the Free Software Foundation; either
##  version 2 of the License, or (at your option) any later version.
##
##  This software is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  Lesser General Public License for more details.
##
##  You should have received a copy of the GNU Lesser General Public
##  License along with this library; if not, write to the Free Software
##  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA.
##
##  For more information about this software see <http://rat.fbk.eu>
##  or email to the RAT Team <rat@fbk.eu>.
##  Please report bugs to <rat@fbk.eu>.
##
##  ===========================================================================


import gtk
import gobject
import pango

class ParenMatchTextBuffer (gtk.TextBuffer):
    """This is a TextBuffer with paren matching. The only public method
    is show_paren_match that shows paren matching. You should
    connect show_paren_matching to the textview signal move-cursor, normally
    as a IDLE function.
    """

    def __init__(self, table=None):
        gtk.TextBuffer.__init__(self, table)

        self.connect("changed", self.__on_bu_changed)

        self.create_tag("paren-match-found",
                        background="yellow")
        
        self.create_tag("paren-match-not-found",
                        background="red",
                        weight=pango.WEIGHT_BOLD)

        self.pmatch_tags = [] # list of tuples (tag_name, mark)
        return


    def show_paren_match(self, iter=None):
        """@param iter is normally the insertion iterator (default if
        None or not specified)"""
        
        opps = ["(", "[", "{"]
        clps = [")", "]", "}"]

        self.__hide_paren_match()

        if iter is None: iter = self.get_iter_at_mark(self.get_insert())

        c = iter.get_char()
        meth = None
        if c in opps:
            inc_p = c
            dec_p = clps[opps.index(inc_p)]
            meth = gtk.TextIter.forward_char
        
        elif iter.backward_char():
            c = iter.get_char()
            if c in clps:
                inc_p = c
                dec_p = opps[clps.index(inc_p)]
                meth = gtk.TextIter.backward_char
                pass            
            pass

        if meth is None: return False

        # finds the corresponding closing/opening paren
        found = False
        it = iter.copy()
        counter = 1
        while meth(it):
            c = it.get_char()
            if c == inc_p: counter += 1
            if c == dec_p: counter -= 1
            
            if counter == 0: found = True; break
            pass

        
        if found:
            self.__show_paren_match_at_iter(iter, "paren-match-found")
            self.__show_paren_match_at_iter(it, "paren-match-found")
        else: self.__show_paren_match_at_iter(iter, "paren-match-not-found")

        return False
    

    def __hide_paren_match(self):
        # removes previously applied taggings:
        for tag in self.pmatch_tags:
            it = self.get_iter_at_mark(tag[1])
            it2 = it.copy()
            it2.forward_char()
            self.remove_tag_by_name(tag[0], it, it2)
            self.delete_mark(tag[1])
            pass
        self.pmatch_tags = []
        return

    def __show_paren_match_at_iter(self, iter, tag_name):
        iter2 = iter.copy()
        iter2.forward_char()
        self.apply_tag_by_name(tag_name, iter, iter2)

        self.pmatch_tags.append((tag_name, self.create_mark(None, iter)))
        return    

    # --------------------------------------------------
    #   Signals
    # --------------------------------------------------

    def __on_bu_changed(self, buf):
        self.show_paren_match()
        return

    pass # end of class



def get_approx_chars_num(widget, pix_width):
    context = widget.get_pango_context()
    metrics = context.get_metrics(context.get_font_description())
    width = pango.PIXELS(metrics.get_approximate_char_width())
    if width != 0: return float(pix_width) / width
    else: return 0


def get_approx_chars_pixel(widget, chars):
    context = widget.get_pango_context()
    metrics = context.get_metrics(context.get_font_description())
    width = pango.PIXELS(metrics.get_approximate_char_width())
    return chars * width

