##  ===========================================================================
##  Author: Roberto Cavada <cavada@fbk.eu>
##
##  Copyright (c) 2005 by FBK 
##
##  This is free software; you can redistribute it and/or
##  modify it under the terms of the GNU Lesser General Public
##  License as published by the Free Software Foundation; either
##  version 2 of the License, or (at your option) any later version.
##
##  This software is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
##  Lesser General Public License for more details.
##
##  You should have received a copy of the GNU Lesser General Public
##  License along with this library; if not, write to the Free Software
##  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA.
##
##  For more information about this software see <http://rat.fbk.eu>
##  or email to the RAT Team <rat@fbk.eu>.
##  Please report bugs to <rat@fbk.eu>.
##
##  ===========================================================================

import gtk
import gobject
import pango
import math
import os
import os.path
import utils.globals

from gtkmvc.view import View 


# To load style for the stripe renderer
gtk.rc_parse(os.path.join(utils.globals.STYLES_DIR, "styles.rc"))


# ----------------------------------------------------------------------
class TraceView (View):
    """The view of a trace model"""

    GLADE_FILE = os.path.join(utils.globals.GLADE_DIR, "trace.glade")

    def __init__(self, ctrl):
        View.__init__(self, ctrl, self.GLADE_FILE, "sw_top")
        return

    def set_horizontal_scroll(self, enable):
        sw = self['sw_top']
        h, v = sw.get_policy()
        if enable: sw.set_policy(gtk.POLICY_AUTOMATIC, v)
        else: sw.set_policy(gtk.POLICY_NEVER, v)
        return

    def set_vertical_scroll(self, enable):
        sw = self['sw_top']
        h, v = sw.get_policy()
        if enable: sw.set_policy(h, gtk.POLICY_AUTOMATIC)
        else: sw.set_policy(h, gtk.POLICY_NEVER)
        return

    pass # end of class



# ----------------------------------------------------------------------
class CellRendererTrace (gtk.GenericCellRenderer):

    """Customed ad-hoc cell renderer for displaying traces"""
    
    __gproperties__ = {

        "value" : (gobject.TYPE_STRING, "The current value to show",
                   "The current value to show",
                   None, gobject.PARAM_READWRITE),
        
        "prev-value" : (gobject.TYPE_STRING, "The value shown on the previous (left) cell",
                        "This is the value shown by the previous cell. This can be None.",
                        None, gobject.PARAM_READWRITE),

        "next-value" : (gobject.TYPE_STRING, "The value shown on the next (right) cell",
                        "This is the value shown by the next cell. This can be None.",
                        None, gobject.PARAM_READWRITE),


        "is-boolean" :  (gobject.TYPE_BOOLEAN, "A flag that specify if the displayed value is boolean.",
                         "A flag that specify if the displayed value is boolean.",
                         False, gobject.PARAM_READWRITE ),

        "frame-fg-color" : (gobject.TYPE_STRING, "The colour of the line used to draw the cell frame",
                            "The colour of the line used to draw the cell frame",
                            None, gobject.PARAM_READWRITE),

        "waveform-fg-color" : (gobject.TYPE_STRING, "The colour of the line used to draw the waveform",
                               "The colour of the line used to draw the the waveform",
                               None, gobject.PARAM_READWRITE),
        "waveform-sel-fg-color" : (gobject.TYPE_STRING,
                                   "The colour of the line used to draw the waveform, when the row is selected",
                                   "The colour of the line used to draw the the waveform, when the row is selected",
                                   None, gobject.PARAM_READWRITE),

        "max-width" : (gobject.TYPE_UINT,
                       "The maximum cell width, in pixels",
                       "The maximum cell width, in pixels",
                       0, # min
                       1000, # max
                       100, # default
                       gobject.PARAM_READWRITE),

        "bool-waveform-width" : (gobject.TYPE_UINT,
                       "The waveform line width, in pixels",
                       "The waveform line width, in pixels",
                       0, # min
                       5, # max
                       2, # default
                       gobject.PARAM_READWRITE),
        
        }

    def __init__(self):
        self.__gobject_init__()
        self.is_boolean = False
        self.value = None
        self.prev_value = None
        self.next_value = None
        self.frame_fg_color = None
        self.waveform_fg_color = None
        self.waveform_sel_fg_color = None
        self.bool_waveform_width = 2

        self.max_width = 100 
        return
    
    def do_set_property(self, pspec, value):
        setattr(self, pspec.name.replace("-","_"), value)
        return
    
    def do_get_property(self, pspec):
        return getattr(self, pspec.name.replace("-","_"))

    def on_get_size(self, widget, cell_area):
        if self.value is None: return (0, 0, 0, 0)

        # calculates width and height
        ctx = widget.get_pango_context()
        metrics = ctx.get_metrics(ctx.get_font_description())
        width = pango.PIXELS(metrics.get_approximate_char_width()) * len(self.value)
        
        if cell_area:
            width = max(width, cell_area.width)
            height = cell_area.height
        else: height = 20
        width = min(width, self.max_width)
                
        xpad = self.get_property("xpad")
        ypad = self.get_property("ypad")
       
        calc_width = xpad*2 + width
        calc_height = ypad*2 + height
        if cell_area and width > 0 and height > 0:
            x_offset = self.get_property("xalign") * (cell_area.width - calc_width - xpad)
            y_offset = self.get_property("yalign") * (cell_area.height - calc_height - ypad)
        else:
            x_offset = 0
            y_offset = 0
            pass
        
        return (x_offset, y_offset, calc_width, calc_height)


    def create_text_layout(self, widget, width):
        """This function is called during rendering. The result will
        be a layout built and returned to contain the text of the
        current value, possibly truncated to fit in the maximum
        space. The layout will have a maximum size of the given width"""
     
        ctx = widget.get_pango_context()
        metrics = ctx.get_metrics(ctx.get_font_description())

        text = self.value

        approx_width = pango.PIXELS(metrics.get_approximate_char_width())
        max_chars = width / approx_width
        
        if len(text) > max_chars: text = "..." + text[-max_chars:] # truncates the text
        layout = widget.create_pango_layout(text)
        layout.set_alignment(pango.ALIGN_CENTER)

        return layout
        
    
    def on_render(self, window, widget, background_area, cell_area, expose_area, flags):
        if self.value is None: return        
        draw_rect = expose_area.intersect(cell_area)

        # Draws the area:
        self.draw_cell(window, widget, draw_rect, flags)
        return

    def draw_cell(self, drawable, widget, rect, flags):
        """Does the drawing work"""
        self.draw_frame(drawable, widget, rect, flags)
        self.draw_waveform(drawable, widget, rect, flags)
        return


    def draw_text(self, drawable, widget, rect, flags):
        """Draws the text within (at the centre) the given rect,
        returns the real rect used to draw the text"""
        
        gct = drawable.new_gc()
        gct.set_rgb_fg_color(self.get_best_text_color(widget, flags))

        layout = self.create_text_layout(widget, rect.width)
        text_w, text_h = layout.get_pixel_size()

        res = gtk.gdk.Rectangle(rect.x + (rect.width - text_w)/2,
                                rect.y + (rect.height - text_h)/2 - 1,
                                text_w, text_h)

        drawable.draw_layout(gct, res.x, res.y, layout)
        return res
        

    def draw_waveform(self, drawable, widget, rect, flags):
        wave_rect = rect.copy()        
        if self.is_boolean:
            delta = max(wave_rect.height / 4, 1)
            wave_rect.y += delta
            wave_rect.height -= delta
            self.draw_waveform_boolean(widget, drawable, wave_rect, flags)
            
        else: # an enumerative value:
            delta = max(wave_rect.height / 8, 1)
            wave_rect.y += delta
            wave_rect.height -= delta
            self.draw_waveform_enumerative(widget, drawable, wave_rect, flags)
            pass
        
        return

    def draw_waveform_boolean(self, widget, drawable, rect, flags):
        col = self.get_best_waveform_color(widget, flags)
        gc = drawable.new_gc(line_width=self.bool_waveform_width)
        gc.set_rgb_fg_color(col)

        hy = rect.y + self.bool_waveform_width / 2
        ly = rect.y + rect.height - self.bool_waveform_width / 2
        
        if "TRUE" == self.value or ("FALSE" != self.value and int(self.value)): 
            drawable.draw_line(gc, rect.x, hy, rect.x+rect.width, hy)
        else: drawable.draw_line(gc, rect.x, ly, rect.x+rect.width, ly)

        if self.prev_value is not None and self.prev_value != self.value:
            drawable.draw_line(gc, rect.x, hy, rect.x, ly+1)
            pass
            
        if self.next_value not in (self.value, None):
            x = rect.x+rect.width
            drawable.draw_line(gc, x, hy, x, ly+1)
            pass
        
        return


    def draw_waveform_enumerative(self, widget, drawable, rect, flags):
        col = self.get_best_waveform_color(widget, flags)
        gc = drawable.new_gc()
        gc.set_rgb_fg_color(col)
        
        ymid = rect.y + rect.height/2
        ybot = rect.y + rect.height - 1
        xr = rect.x+rect.width

        segms = []
        if self.prev_value is None or self.prev_value != self.value:
            x1 = rect.x+rect.height/2
            segms += [(x1, ybot, rect.x, ymid), (rect.x, ymid, x1, rect.y)]
            if self.next_value is None or self.next_value != self.value:
                x2 = rect.x*2 + rect.width - x1
                segms += [(x1, rect.y, x2, rect.y), (x2, rect.y, xr, ymid),
                          (xr, ymid, x2, ybot), (x2, ybot, x1, ybot)]
            else: segms += [(x1, rect.y, xr, rect.y), (xr, ybot, x1, ybot)]

        elif self.next_value is None or self.next_value != self.value:
            x2 = rect.x + rect.width - rect.height/2
            segms += [(rect.x, rect.y, x2, rect.y), (x2, rect.y, xr, ymid),
                      (xr, ymid, x2, ybot), (x2, ybot, rect.x, ybot)]
        else: segms += [(rect.x, rect.y, xr, rect.y), (xr, ybot, rect.x, ybot)]
            
        drawable.draw_segments(gc, segms)
        self.draw_text(drawable, widget, rect, flags)        
        return
        

    def draw_frame(self, drawable, widget, rect, flags):
        col = self.get_best_frame_color(widget, flags)
        gc = drawable.new_gc()
        gc.set_rgb_fg_color(col)
        drawable.draw_rectangle(gc, False, rect.x, rect.y, rect.width, rect.height)
        return

    def get_best_waveform_color(self, treeview, flags):
        """Returns the best foreground color for the waveform line
        depending on the current context"""

        if self.waveform_fg_color is None:
            st = treeview.get_style()
            if flags & gtk.CELL_RENDERER_SELECTED: col = st.fg[gtk.STATE_SELECTED]
            else: col = st.fg[gtk.STATE_NORMAL]
            return col.copy()

        if self.waveform_sel_fg_color is not None and flags & gtk.CELL_RENDERER_SELECTED:
            return gtk.gdk.color_parse(self.waveform_sel_fg_color)
        return gtk.gdk.color_parse(self.waveform_fg_color)


    def get_best_text_color(self, treeview, flags):
        st = treeview.get_style()
        if flags & gtk.CELL_RENDERER_SELECTED: col = st.fg[gtk.STATE_SELECTED]
        else: col = st.fg[gtk.STATE_NORMAL]
        return col.copy()

    def get_best_frame_color(self, treeview, flags):
        """Returns the best foreground color for the frame depending
        on the current context"""

        # if not set by the user through the property, calculates the
        # color of the frame, depending on the cell background and
        # selection status:
        if self.frame_fg_color is None:
            st = treeview.get_style()
            if flags & gtk.CELL_RENDERER_SELECTED: col = st.bg[gtk.STATE_SELECTED]
            elif self.get_property("cell-background-set"):
                col = self.get_property("cell-background-gdk")
            else: col = st.bg[gtk.STATE_NORMAL]
            col = col.copy()
            
        else: col = gtk.gdk.color_parse(self.frame_fg_color)

        # changes slightly the color, depending on the prelit as well:
        if flags & gtk.CELL_RENDERER_PRELIT: col_delta = 12000
        elif self.frame_fg_color is None: col_delta = 7000
        else: col_delta = 0

        light_threshold = 50000
        if col.red < light_threshold: col.red += col_delta
        else: col.red -= col_delta
        if col.green < light_threshold: col.green += col_delta
        else: col.green -= col_delta
        if col.blue < light_threshold: col.blue += col_delta
        else: col.blue -= col_delta
        
        return col
    
    
    pass # end of class

gobject.type_register(CellRendererTrace)
