#!/usr/bin/env python

# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================



"""Diagnosability Check

Check whether a property is diagnosable with respect to a
diagnosability property and zero or more path restrictions.

Inputs      IM : an integrated model.
            P : a diagnosability property.
            Obs : a set of observables.

Outputs     Yes if the property is diagnosable, No if not,
            and Unknown if no counterexample has been found
            (using BMC in hybrid models).


Processing  Using model checking techniques, the COMPASS toolset
            checks if a model is diagnosable with respect to a
            given diagnosability property and zero or more path
            restrictions."""


import booter # import this as first module in scripts

import property_management.src.constants as prop_const
import property_management.src.api as prop_api

from utils import fsap_process
from utils import exceptions, log
from utils.misc import relpath

import options
import slim_compiler.src.smv.api as smv_api
from slim_compiler.src.smv.walkers import InfiniteDomainFinderWalker
import subprocess
import os

from runscripts.basic import ToolScript
from runscripts.fault_tree_generation__srs_017 import FsapGeneralScript

import invokers.slim as slim


class FsapDiagnosability(FsapGeneralScript):
    """This is the script that can be used to run FSAP to perform
    a diagnosability check"""

    def register_options(self, opt_parser):
        super(FsapDiagnosability, self).register_options(opt_parser)

        opt_parser.add_option("--path-restrictions", dest="FSAP_RESTRICTIONS_FILE_NAME",
                              help="Path restrictions file")

        opt_parser.add_option("--property-context", dest="FSAP_CONTEXT_FILE_NAME",
                              help="Property context file")

        opt_parser.add_option("--property-num", dest="FSAP_PROPERTY_NUMBER",
                              action="store", type = "int", default=0,
                              help="Specifies the number of a property to deal with. Count begins with 0 which is the default value.")

        opt_parser.add_option("--bmc-length", dest="NUSMV_BMC_LENGTH",
                              action="store", type="int", default=10,
                              help="Maximum path lenght for BMC.")

        opt_parser.add_option("--diag-delay", dest="DIAG_DELAY",
                              action="store", type="string", default=None,
                              help="Specifies the detection delay (do not specify in case of discrete models)")

        return

    def check_options(self, opt_parser, options):
        super(FsapDiagnosability, self).check_options(opt_parser, options)
        
        # properties are mandatory
        if not options.PROPERTIES_FILE_NAME:
            opt_parser.error("The diagnosability properties file was not specified.")
            pass

        # the property number has to be >= 0.
        if options.FSAP_PROPERTY_NUMBER < 0:
            opt_parser.error("The number of a property has to be 0 or greater.")
            pass
        
        return

    def store_options(self, opt_parser, options):
        super(FsapDiagnosability, self).store_options(opt_parser, options)

        # path restrictions
        if opt_parser.values.FSAP_RESTRICTIONS_FILE_NAME:
            self.restrictions_filename = opt_parser.values.FSAP_RESTRICTIONS_FILE_NAME
        else:
            self.restrictions_filename = None

        # property context
        if opt_parser.values.FSAP_CONTEXT_FILE_NAME:
            self.context_filename = opt_parser.values.FSAP_CONTEXT_FILE_NAME
        else:
            self.context_filename = None

        assert opt_parser.values.NUSMV_BMC_LENGTH
        options.NUSMV_BMC_LENGTH = opt_parser.values.NUSMV_BMC_LENGTH

        if opt_parser.values.DIAG_DELAY:
            self.DIAG_DELAY = opt_parser.values.DIAG_DELAY
        else:
            self.DIAG_DELAY = -1.0

        return

    def action(self):
        """Builds FSAP and performs a diagnosability check for two given propositional
        properties and optional path restrictions."""
        super(FsapDiagnosability, self).action()

        # verify and prepare observables

        assert self._observables_num >= 0, "Tool chain is broken. The number of observable must have been calculated during failures/observables dumping."
        if self._observables_num == 0:
            raise exceptions.UserException("There are no observables in the model. A diagnosability check is meaningless in such a situation.")
        observables = smv_api.get_observations(self.root_impl)

        # verify diagnosability properties
        
        if self.formulae_doc is None:
            raise exceptions.UserException("Diagnosability properties file is empty.")

        if options.FSAP_PROPERTY_NUMBER < 0 or options.FSAP_PROPERTY_NUMBER >= len(self.formulae_doc):
            raise exceptions.UserException("Error: the specified number of the left property is outside of the range: %d is not in [0, %d)." %\
                             (options.FSAP_PROPERTY_NUMBER, len(self.formulae_doc)))

        log.user.info("Diagnosability Analysis is to be performed for property number %d..." %\
                      (options.FSAP_PROPERTY_NUMBER))

        # prepare restrictions and properties
  
        orig_prop = self.formulae_doc[options.FSAP_PROPERTY_NUMBER]
        orig_prop_text = prop_api.get_formula_as_smv_string(orig_prop)
        smv_prop = self.smv_formulae_doc[options.FSAP_PROPERTY_NUMBER]
        smv_prop_text = prop_api.get_formula_as_smv_string(smv_prop)

        if not prop_api.check_is_propositional_formula(smv_prop):
            raise exceptions.UserException("A diagnosability check cannot be performed for non-propositional formulas : %s"\
                             % orig_prop_text)

        if self.restrictions_filename is not None:
            restrictions_buffer = self.process_property_file(self.root_impl, self.restrictions_filename)
            if len(restrictions_buffer) < 1:
                log.user.warning('Attention: The file "%s" doesn\'t contain valid path restriction formulas.' % self.restrictions_filename)
        else:
            restrictions_buffer = []

        # prepare property context

        if self.context_filename is not None:
            property_context = self.process_property_file(self.root_impl, self.context_filename)
            if len(property_context) < 1:
                log.user.warning('Attention: The file "%s" doesn\'t contain a valid property context formula.' % self.context_filename)
            elif len(property_context) > 1:
                raise exceptions.UserException("Error: The document specifying the property context can contain at most one formula.")
            else:
                property_context = property_context[0]
        else:
            property_context = "TRUE"

        # set the delay

        if self.DIAG_DELAY == -1.0:
            options.DIAG_DELAY = self.DIAG_DELAY
        else:
            options.DIAG_DELAY = self.get_delay_value(self.DIAG_DELAY)
        log.user.info("Diagnosability Analysis is to be performed with delay: %s..." %options.DIAG_DELAY)

        # set delta_t_var_name
        walker = InfiniteDomainFinderWalker()
        walker.walk(self.root_impl)

        infinite_domain = walker.has_found_infinite_data()
        timed_domain = walker.has_found_clock_data()

        if timed_domain:
            delta_t_var_name, tu = smv_api.get_smv_delta_var()
        else:
            delta_t_var_name = None

        # send task to nusmv-fsap process

        (success, trace_A, trace_B) = self.fsap.check_diagnosability(smv_prop_text, property_context, observables, restrictions_buffer, delta_t_var_name=delta_t_var_name)

        # announce results

        log.user.info('A diagnosability check has been performed for the property (%s).' % (orig_prop_text))

        if success == "Error":
            log.user.info('An error has occured during Diagnosability Analysis (no result was returned).')
        elif success == "True":
            log.user.info('The property was found to be diagnosable.')
        elif success == "Unknown":
            log.user.info('No counterexample for diagnosability of the given property was found within the given bound.')
        else:
            log.user.info('The property was found to be not diagnosable (see counterexample in %s and %s).' % (trace_A, trace_B))

        return

    def process_property_file(self, root_impl, filename):
        """Parses a property file containing one or more propositional properties,
        and returns a string list of all formulas in smv format."""

        smv_prop_buffer = []

        if filename is None: return smv_prop_buffer

        prop_doc = slim.get_formulae_doc(root_impl, filename)

        if prop_doc is None or len(prop_doc) < 1:
            return smv_prop_buffer

        prop_api.convert_atomic_propositions_to_smv(prop_doc, root_impl)

        for prop in prop_doc:
            if not prop_api.check_is_propositional_formula(prop):
                raise exceptions.UserException("Diagnosability checks can work only with propositional formulas : %s" % prop_api.get_formula_as_smv_string(prop))

            smv_prop_buffer.append(prop_api.get_formula_as_smv_string(prop))

        return smv_prop_buffer


    def get_delay_value(self, delay_str):

        values = delay_str.split(" ")

        if len(values) == 1:
            # It must be an integer/float
            try:
                delay = float(delay_str)
                return delay
            except ValueError:
                raise exceptions.UserException("Error: the specified delay is not valid. %s is not a float or a "\
                                                   "valid slim expression" %delay_str)
        else:
            # It must be an expression of the type <numeric value> <time unit>
            if len(values) != 2:
                raise exceptions.UserException("Error: the specified delay is not valid. %s is not a float or a "\
                                                   "valid slim expression" %delay_str)
            time_val = values[0]
            #FIXME: using regular expressions to understand whether it is an integer or a float
            if "." not in time_val:
                time_val += ".0"
            time_unit = values[1]
            if time_unit not in ["msec", "sec", "min", "hour", "day"]:
                raise exceptions.UserException("Error: the time unit is not valid. %s must be one of the following "\
                                                   "values: ['msec', 'sec', 'min', 'hour', 'day']" %time_unit)
            delay_str = time_val + time_unit
            smv_delay = smv_api.parse_and_convert_numeric_expression(self.root_impl, delay_str)
            if "f'" in str(smv_delay):
                smv_delay = smv_delay.replace("f'", "")
            delay = eval(smv_delay)
            return delay


    pass # end of class
# ----------------------------------------------------------------------
        
if __name__ == "__main__":
    main = FsapDiagnosability()
    main.run()

    
