#!/usr/bin/env python
# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================



"""Fault Tree Verification

TODO: Complete this docstring.
"""

import booter # import this as first module in scripts

from utils.property_checker import DFTPropertyChecker, PropertyCheckingRecord
from runscripts.basic import BaseScript
from utils import misc, log
import property_management.src.properties as properties_module
import options
import os

#ERROR REPORT MODES
REPORT_VIA_LOG = 0
REPORT_VIA_RETURN = 1
ERROR_REPORT_MODES = [REPORT_VIA_LOG, REPORT_VIA_RETURN]

class FaultTreeVerification (BaseScript):
    """This script executes performance evaluation.
    It replaces performability_evaluation__srs_022.py"""

    _actions = ["to-sigref", "to-imc", "evaluation"]

    def __init__(self, derived_actions=[]):
        super(FaultTreeVerification, self).__init__(derived_actions=derived_actions+self._actions)
        
        self.fault_tree_filename = None
        self.basic_event_rates_filename = None
        
        return
    
    def get_usage(self):
        return "usage: %prog [options] fault_tree_filename basic_event_rates_filename"

    def register_options(self, opt_parser):
        super(FaultTreeVerification, self).register_options(opt_parser)

        opt_parser.add_option("-p", "--property", dest="PROPERTIES_FILE_NAME",
                              help="Properties file")

        opt_parser.add_option("--stepwise", dest="STEPWISE",
                              action="store_true", default=False,
                              help="Calculate probabilities stepwise")

        def __check_positive(option, opt_str, value, parser):
            if value < 1: parser.error("Granularity must be positive.")
            return

        opt_parser.add_option("--granularity", callback=__check_positive, 
                              action="callback", dest="STEPWISE_GRANULARITY", type="int",
                              default=10, help="Number of steps in stepwise calculation")

        def __check_tool(option, opt_str, value, parser):
            tools = ["imca", "mrmc"]
            if value == "auto":
                options.MODEL_CHECKERS = tools
            elif value in tools:
                options.MODEL_CHECKERS = [value]
            else:
                parser.error("Invalid tool selected: %s." % value)

            return
        
        opt_parser.add_option("--tool", callback=__check_tool, 
                              action="callback", type="string",
                              default="auto", help="Tool to use for performability analysis (imca, mrmc, auto)")

        def __check_error_bound(option, opt_str, value, parser):
            if value <= 0: parser.error("Error bound must be positive.")
            #TODO: no logging yet
            if value >= 1.0: print "Error bound possibly too high"
            return

        opt_parser.add_option("--error-bound", callback=__check_error_bound, 
                              action="callback", dest="ERROR_BOUND", type="float",
                              default=1e-3, help="Error bound for IMCA process")

        return

    def check_options(self, opt_parser, options):
        super(FaultTreeVerification, self).check_options(opt_parser, options)

        # properties are mandatory
        if not options.PROPERTIES_FILE_NAME:
            opt_parser.error("Properties file was not specified.")
            pass
        
        if len(options.POSITIONAL_ARGS) != 2: opt_parser.error("Incorrect number of arguments.")

        self.fault_tree_filename = options.POSITIONAL_ARGS[0]
        self.basic_event_rates_filename = options.POSITIONAL_ARGS[1]

        return
    

    def action(self):
        super(FaultTreeVerification, self).action()

        dft_checker = DFTPropertyChecker(self.fault_tree_filename, self.basic_event_rates_filename, options.PROPERTIES_FILE_NAME, options.MODEL_CHECKERS)
        dft_checker.debug_info = self.debug_info

        dft_checker.gen_model_files()

        properties_document = properties_module.parse_properties_document(options.PROPERTIES_FILE_NAME)
        for property in properties_document:
            (res, error_msg) = dft_checker.analyze(property, options.ERROR_BOUND, options.STEPWISE, options.STEPWISE_GRANULARITY)
            self.print_results(property, res)
            if error_msg:
                log.system.error("Possible error during DFT verification")

    def print_results(self, formula, probs):
        desc = formula.description
        if desc == '':
            desc = formula.name
        log.user.info("Fault tree verification for '%s' returns\n%s" % (desc, probs))

    pass # end of class

# ----------------------------------------------------------------------
        
if __name__ == "__main__":
    
    main = FaultTreeVerification()
    main.run()
