#!/usr/bin/env python

# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================



"""Failure Mode and Effects Analysis

Generate an FMEA table given an operational model, extended with a fault
model, optionally extended with an FDIR model, a set of fault configurations
and a set of properties.

Inputs      M : an operational model.
            FM : a fault model.
            FDM : an FDIR model (optional).
            FC : cardinality of a set of fault configurations.
            P : a set of properties.

Outputs     FMEA table.

Processing  Using model checking techniques, the COMPASS toolset
            generates an FMEA table starting from an operational model
            extended with a fault model, optionally extended with an
            FDIR model, cardinality of a set of fault configurations,
            and a set of properties. The user is allowed to specify
            the cardinality of the fault configurations, e.g.,
            "all the combinations of faults with cardinality at most
            two".
"""

import booter # import this as first module in scripts

import property_management.src.constants as prop_const
import property_management.src.api as prop_api
from utils import fsap_process
from utils import exceptions, log
from utils.misc import relpath, get_out_fname
import options
import slim_compiler.src.smv.api as smv_api

import subprocess
import os

from runscripts.basic import ToolScript
from runscripts.fault_tree_generation__srs_017 import FsapGeneralScript


class FsapFmeaTable(FsapGeneralScript):
    """This is the script that can be used to run FSAP to generate a FMEA
    (Fault Mode and Effect Analysis) table for a set of propositional
    properties"""

    def register_options(self, opt_parser):
        # Specific options for FMEA computation
        super(FsapFmeaTable, self).register_options(opt_parser)

        opt_parser.add_option("--faults-cardinality", dest="FSAP_FAULT_CONFIGURATION_CARDINALITY",
                              action="store", type = "int", default=1,
                              help="Specifies the cardinality of a set of fault configurations. Minimal allowed value is 1. Default value is 1.")

        opt_parser.add_option("--dynamic", dest="FSAP_USE_DYNAMIC",
                              action="store", default=False,
                              help="Enable the generation of Dynamic FMEA tables")
        return

    def check_options(self, opt_parser, options):
        super(FsapFmeaTable, self).check_options(opt_parser, options)
        
        # properties are mandatory
        if not options.PROPERTIES_FILE_NAME:
            opt_parser.error("Properties file was not specified.")
            pass

        if options.FSAP_FAULT_CONFIGURATION_CARDINALITY <= 0:
            opt_parser.error("Cardinality of fault configuration has to be 1 or greater")
            pass
        
        return


    def action(self):
        """Builds FSAP and performing FMEA table generation for
        propositional properties"""
        super(FsapFmeaTable, self).action()
        
        if self.formulae_doc is None:
            raise exceptions.UserException("Properties file is empty")

        log.user.info("FMEA table is to be generated for %d properties..." % 
                      len(self.formulae_doc))

        props = [];
        for orig_prop, smv_prop in zip(self.formulae_doc, self.smv_formulae_doc):
            
            orig_prop_text = prop_api.get_formula_as_smv_string(orig_prop)
            smv_prop_text = prop_api.get_formula_as_smv_string(smv_prop)

            if not prop_api.check_is_propositional_formula(smv_prop):
                raise exceptions.UserException("FMEA Table cannot be computed for non-propositional formulas : %s"\
                                 % orig_prop_text)
            props.append(smv_prop_text)
            pass

        # Generates the FMEA table
        if options.FSAP_USE_DYNAMIC:
            tmp_fmea_fname = self.fsap.compute_dyn_fmea_table(props,
                                            options.FSAP_FAULT_CONFIGURATION_CARDINALITY)
        else: tmp_fmea_fname = self.fsap.compute_fmea_table(props,
                                            options.FSAP_FAULT_CONFIGURATION_CARDINALITY)
        
        # rename the files according to the base name
        fmea_fname = get_out_fname(options.BASE_FILE_NAME + ".fmea")
  
        # If FSAP execution is unsuccessful then an exception is raise
        # and here we assume the run was successful and file has been generated.
        assert os.path.isfile(tmp_fmea_fname),\
               "FMEA Table file has not been generated by FSAP for unknown reason (advice: run with '-l debug' to see FSAP messages)"

        os.rename(tmp_fmea_fname, fmea_fname)
           
        if not options.FSAP_KEEP_SMV_IDS:
            fsap_process.convert_fmea_table(fmea_fname, fmea_fname,\
                                            smv_api.convert_smv_const_to_slim)
            pass

        log.user.info('%sFMEA Table has been generated and stored in file "%s"' %
                      ("Dynamic " if options.FSAP_USE_DYNAMIC else "", 
                       relpath(fmea_fname)))
        return
    pass # end of class
# ----------------------------------------------------------------------
        
if __name__ == "__main__":
    main = FsapFmeaTable()
    main.run()

    
