#!/usr/bin/env python

# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


"""Time Divergence Checking

For all clock data-subcomponents, verify whether a time-divergence
conditions exists.

TBD
"""

import booter  # import this as first module in scripts

import property_management.src.api as properties_api
import property_management.src.constants as properties_const
from property_management.src.properties import Property, PatternInstance, StateFormulaPlaceholderInstance
from property_management.src.formulae import AtomicProposition
import slim_compiler.src.smv.api as smv_api
import options
from utils import exceptions, log
from invokers import nusmv
from utils.main_extension_converter import MainExtensionConverter
from utils import fsap_process
from utils.misc import get_out_fname
import os
import subprocess


from runscripts.fault_tree_generation__srs_017 import FsapGeneralScript



class FsapFaultCoverageAnalysis(FsapGeneralScript):
    """This is the script that can be used to run FSAP to perform the Fault
    Coverage Analysis"""


    def register_options(self, opt_parser):
        super(FsapFaultCoverageAnalysis, self).register_options(opt_parser)

        opt_parser.add_option("--property-num", dest="FSAP_PROPERTY_NUMBER",
                              action="store", type="int", default=0,
                              help="Specifies the number of the property used for recoverability (default: 0)")

        opt_parser.add_option("--fault-str", dest="FAULT_STR",
                              action="store", type="string", default=None,
                              help="Specifies the fault to be detected (form: <component_path>.error = error:<error_state>")

        opt_parser.add_option("--detection-delay", dest="DETECTION_DELAY",
                              action="store", type="string", default=None,
                              help="Specifies the detection delay (do not specify in case of discrete model)")

        opt_parser.add_option("--bmc-length", dest="NUSMV_BMC_LENGTH",
                              action="store", type="int", default=10,
                              help="Maximum path length for BMC.")
        return



    def check_options(self, opt_parser, options):
        super(FsapFaultCoverageAnalysis, self).check_options(opt_parser, options)

        # properties are required since they are used for recoverability
        if not options.PROPERTIES_FILE_NAME:
            opt_parser.error("The recoverability properties file  was not specified.")
            pass

        # the property number has to be >= 0
        if options.FSAP_PROPERTY_NUMBER < 0:
            opt_parser.error("The number of a property has to be 0 or greater.")
            pass

        # the fault must be specified
        if opt_parser.values.FAULT_STR is None:
            opt_parser.error("The fault to be detected was not specified.")
            pass

        return



    def store_options(self, opt_parser, options):
        super(FsapFaultCoverageAnalysis, self).store_options(opt_parser, options)

        if opt_parser.values.FAULT_STR:
            self.FAULT_STR = opt_parser.values.FAULT_STR

        if opt_parser.values.DETECTION_DELAY:
            self.DETECTION_DELAY = opt_parser.values.DETECTION_DELAY
        else:
            self.DETECTION_DELAY = -1.0

        assert opt_parser.values.NUSMV_BMC_LENGTH
        options.NUSMV_BMC_LENGTH = opt_parser.values.NUSMV_BMC_LENGTH

        return



    def action(self):
        super(FsapFaultCoverageAnalysis, self).action()

        # Additional checks

        if options.FSAP_PROPERTY_NUMBER >= len(self.formulae_doc):
            raise exceptions.UserException("Error: the specified number of the property is outside of the range: " \
                                               "%d is not in [0, %d)." %(options.FSAP_PROPERTY_NUMBER, len(self.formulae_doc)))

        assert self.formulae_doc is not None
        assert self.FAULT_STR is not None

        # Run FCA Recovery

        log.user.info("Recoverability Analysis is to be performed for property number %d..." %options.FSAP_PROPERTY_NUMBER)

        smv_prop = self.smv_formulae_doc[options.FSAP_PROPERTY_NUMBER]
        smv_prop_text = properties_api.get_formula_as_smv_string(smv_prop)

        fca_recovery_result, fca_recovery_trace_fname = nusmv.model_checking(self.fsap, smv_prop)

        if fca_recovery_result:
            recovery_msg = "Recoverable"
        elif fca_recovery_result is None:
            # Msg should be 'May be Recoverable'
            recovery_msg = "Recoverable"
        else:
            recovery_msg = "Not Recoverable"

        log.user.info("Result of the Recoverability Analysis for properties %s: %s..." \
                          %(smv_prop_text, recovery_msg))

        if self.fsap.is_alive():
            self.fsap.kill()

        # Run FCA Detection

        self.fail_obs_alm_fname = get_out_fname(options.BASE_FILE_NAME + ".fail_obs_alm")
        logfname = os.path.splitext(self.smv_fname)[0] + ".fsapout"
        self.fsap = fsap_process.FsapProcess(self.smv_fname, self.fail_obs_alm_fname,
                                             need_msat=self.need_msat, 
                                             stdout=subprocess.PIPE, stderr=subprocess.STDOUT, 
                                             dumpers=[open(logfname, "w")])

        log.user.info("Detection Analysis is to be performed for fault: %s..." %self.FAULT_STR)


        # Creating a property corresponding to the specified fault
        pattern_descr = [p for p in properties_api.get_patterns_document(properties_const.Logic.LTL)]
        assert pattern_descr[0].id == "propositional"
        pattern_instance = PatternInstance.create(properties_const.Logic.LTL, pattern_descr[0].id)
        atomic_prop = AtomicProposition.create(properties_const.Logic.LTL, self.FAULT_STR)
        placeholder = StateFormulaPlaceholderInstance.create(properties_const.Logic.LTL, "P", atomic_prop)
        pattern_instance.add_placeholder_instance(placeholder)
        _property = Property.create(pattern_instance, "propositional 0", self.FAULT_STR)
       
        # Getting the delay
        if self.DETECTION_DELAY is None or self.DETECTION_DELAY == -1.0:
            delay = self.DETECTION_DELAY
        else:
            delay = self.get_delay_value(self.DETECTION_DELAY)

        formulae_doc = properties_api.create_formulae_document_for_properties([_property])
        smv_formulae_doc = properties_api.create_formulae_document_for_properties([_property])
        properties_api.convert_atomic_propositions_to_smv(smv_formulae_doc, self.root_impl)

        properties =  [(properties_api.get_formula_as_smv_string(slim),
                        properties_api.get_formula_as_smv_string(smv))
                       for slim, smv in zip(formulae_doc, smv_formulae_doc)]
        smv_prop_text = properties[0][1]

        if not properties_api.check_is_propositional_formula(smv_formulae_doc[0]):
            raise exceptions.UserException("Fault Detection cannot be computed for non-propositional formulae: %s" \
                                               % formulae_doc[0])

        if delay is not None:

            if delay == -1.0:
                delay = None
                delay_counter_var = None
            else:
                assert delay >= 0.0
                delay_counter_bound = None

                if delay >= 1:
                    delay_counter_bound = delay + 1
                elif delay == 0:
                    delay_counter_bound = 0.00001
                else:
                    # increase first non-zero digit by one, eg: 0.004823 -> 0.005823
                    factor = 10
                    while delay * factor < 1:
                        factor = factor * 10
                    delay_counter_bound = delay + (1.0 / factor)

                self.extend_smv_with_delay_counter(smv_prop_text, delay_counter_bound, "fsap_delay_counter",
                                                   self.smv_fname)
                delay_counter_var = "fsap_delay_counter"
                log.user.info("Detection Analysis is to be performed with delay: %s..." %str(delay))

        else:
            delay_counter_var = None

        detect_fname = self.fsap.compute_fault_detection(smv_prop_text, options.NUSMV_BMC_LENGTH, delay_counter_var, delay)

        if detect_fname is None:
            detect_msg = "Not Detectable" # No alarms in the model
        else:
            res = self.load_detection(detect_fname)
            if not res:
                detect_msg = "Not Detectable" # Empty alarms list
            else:
                detect_msg = "Detectable"

        log.user.info("Result of the Detection Analysis for fault %s: %s..." \
                          %(self.FAULT_STR, detect_msg))

        if detect_msg == "Not Detectable":
            classification = "Uncovered"
        else:
            if recovery_msg == "Recoverable":
                classification = "Covered"
            else:
                classification = "Uncovered"

        log.user.info("Fault Coverage Analysis has been performed for fault '%s' and the result is: %s" \
                          %(self.FAULT_STR, classification))

        return



    def get_delay_value(self, delay_str):

        values = delay_str.split(" ")

        if len(values) == 1:
            # It must be an integer/float
            try:
                delay = float(delay_str)
                return delay
            except ValueError:
                raise exceptions.UserException("Error: the specified delay is not valid. %s is not a float or a "\
                                                   "valid slim expression" %delay_str)
        else:
            # It must be an expression of the type <numeric value> <time unit>
            if len(values) != 2:
                raise exceptions.UserException("Error: the specified delay is not valid. %s is not a float or a "\
                                                   "valid slim expression" %delay_str)
            time_val = values[0]
            #FIXME: using regular expressions to understand whether it is an integer or a float
            if "." not in time_val:
                time_val += ".0"
            time_unit = values[1]
            if time_unit not in ["msec", "sec", "min", "hour", "day"]:
                raise exceptions.UserException("Error: the time unit is not valid. %s must be one of the following "\
                                                   "values: ['msec', 'sec', 'min', 'hour', 'day']" %time_unit)
            delay_str = time_val + time_unit
            smv_delay = smv_api.parse_and_convert_numeric_expression(self.root_impl, delay_str)
            if "f'" in str(smv_delay):
                smv_delay = smv_delay.replace("f'", "")
            delay = eval(smv_delay)
            return delay


    def extend_smv_with_delay_counter(self, smv_prop, delay_counter_bound, delay_counter_var, smv_fname):
        assert delay_counter_bound > 0
        assert smv_fname is not None

        conv = MainExtensionConverter()

        delta_var = None

        from slim_compiler.src.smv.walkers import InfiniteDomainFinderWalker
        walker = InfiniteDomainFinderWalker()
        walker.walk(self.root_impl)

        infinite_domain = walker.has_found_infinite_data()
        timed_domain = walker.has_found_clock_data()

        if timed_domain:
            # count time
            delta_var = conv.delta_var()
        else:
            # count steps
            delta_var = "1"

        delay_counter_bound_str = None
        delay_var_type = None

        if infinite_domain:
            # delay counter bound is float
            delay_counter_bound_str = "%f" % delay_counter_bound
            delay_var_type = "real"
        else:
            # delay bound is int
            assert delay_counter_bound % 1 == 0
            delay_counter_bound_str = "%d" % delay_counter_bound
            delay_var_type = "-1..%d" % delay_counter_bound

        rnd_choice_var = "%s_rnd" % delay_counter_var

        main_extension = \
            "  VAR %s : %s;\n" % (delay_counter_var, delay_var_type) + \
            "  VAR %s : boolean;\n" % rnd_choice_var + \
            "\n" + \
            "  ASSIGN init(%s) :=\n" % (delay_counter_var) + \
            "    case\n" + \
            "      %s & %s : 0;\n" % (smv_prop, rnd_choice_var) + \
            "      TRUE : -1;\n" + \
            "    esac;\n" + \
            "\n"

        main_extension += \
            "  ASSIGN next(%s) :=\n" % (delay_counter_var) + \
            "    case\n" + \
            "      %s = -1 & !next(%s) : -1;\n" % (delay_counter_var, smv_prop) + \
            "      %s = -1 &  next(%s) & !%s : -1;\n" % (delay_counter_var, smv_prop, rnd_choice_var) + \
            "      %s = -1 &  next(%s) &  %s :  0;\n" % (delay_counter_var, smv_prop, rnd_choice_var) + \
            "      %s + %s < %s : %s + %s;\n" % (delay_counter_var, delta_var, delay_counter_bound_str, delay_counter_var, delta_var) + \
            "      TRUE : %s;\n" % (delay_counter_bound_str) + \
            "    esac;\n" + \
            "\n"

        conv.set_main_extension(main_extension)

        out = open(smv_fname, 'w')
        conv.dump(self.root_impl, out)
        out.close()

        return


    def load_detection(self, fda_fname):
        """Function which tries to load the result of the fault detection and returns a boolean to indicate success or failure.

           @param fda_fname: the name of the file in which the result of fault detection was stored
           @type  fda_fname: string

           @return res: the result of loading the fault detection file
           @rtype  res: boolean
        """

        try:
            res = False

            _file = open(fda_fname, "r")
            line = _file.readline()
            while line:
                # the first line contains the headers and it is ignored
                line = _file.readline().strip()
                if line: 
                    fields = line.split()
                    assert len(fields) == 1
                    res = True
                    pass
                
                pass
            _file.close()

        except Exception:
            raise Exception("There was a problem while reading the FDA result")

        return res



    pass # end of class
# ----------------------------------------------------------------------
        
if __name__ == "__main__":
    main = FsapFaultCoverageAnalysis()
    main.run()

    
