#!/usr/bin/env python

# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================



"""Failure Isolation Analysis

Compute fault isolation measures for a model.

Inputs      M : an operational model.
            FM : a fault model.
            FDM : an FDIR model (optional).
            Alm : a set of alarms.

Outputs     a set of fault trees, one for each alarm.

Processing  Using model checking techniques, the COMPASS toolset
            computes fault isolation measures for an operational
            model, extended with a fault model and an FDIR model.
            Fault isolation measures is produced by computing
            for each input alarm the set of minimal explanations
            (single faults or combinations thereof) that are
            compatible with the alarm being true. The set of
            minimal explanations is presented as a set of fault
            trees, one for each alarm.
"""

import booter # import this as first module in scripts

from utils import fsap_process
from utils import exceptions, log
from utils.misc import relpath, get_out_fname
import options
import slim_compiler.src.smv.api as smv_api

import subprocess
import os

from runscripts.basic import ToolScript
from runscripts.fault_tree_generation__srs_017 import FsapGeneralScript


class FsapFaultIsolation(FsapGeneralScript):
    """This is the script that can be used to run FSAP to perform
    fault isolation analysis"""

    def register_options(self, opt_parser):
        # Specific options 
        super(FsapFaultIsolation, self).register_options(opt_parser)

        # remove --property options because Fault Isolation does no need properties
        assert opt_parser.has_option("--property")
        opt_parser.remove_option("--property")
        
        return

    def check_options(self, opt_parser, options):
        super(FsapFaultIsolation, self).check_options(opt_parser, options)
        
        # properties are mandatory
        assert not options.PROPERTIES_FILE_NAME # see in register_options function
        
        return


    def action(self):
        """Builds FSAP and performing Fault Isolation Analysis"""
        super(FsapFaultIsolation, self).action()

        assert self._alarms_num >= 0, "Tool chain is broken. The number of alarm must have been calculated during failures/observables/alarms dumping"

        if self._alarms_num == 0:
            raise exceptions.UserException("There are no alarms in model. Fault Isolation Analysis is meaningless in such situation")
        
        log.user.info("Fault Isolation Analysis is to be performed for %d alarms...", 
                      self._alarms_num)

        tmp_isolat_fname = self.fsap.compute_fault_isolation(self._alarms_num)

        # rename the files according to the base name
        out_files = []
        for files, num in zip(tmp_isolat_fname, range(len(tmp_isolat_fname))):
             # If FSAP execution is unsuccessful then an exception is raise
            # and here we assume the run was successful and file has been generated.
            assert os.path.isfile(files[0]) or os.path.isfile(files[1]),\
                   "For unknown reasons FSAP has not generated output files"

            ev_fname = get_out_fname("%s.%d.flt_events" % (options.BASE_FILE_NAME, num))
            gt_fname = get_out_fname("%s.%d.flt_gates" % (options.BASE_FILE_NAME, num))
            out_files.append((relpath(ev_fname), relpath(gt_fname)))

            os.rename(files[0], ev_fname)
            os.rename(files[1], gt_fname)
           
            if not options.FSAP_KEEP_SMV_IDS:
                fsap_process.convert_fault_tree_events(ev_fname, ev_fname,
                                                       smv_api.convert_smv_const_to_slim)
                fsap_process.convert_fault_tree_gates(gt_fname, gt_fname,
                                                      smv_api.convert_smv_const_to_slim)
                pass
            pass

        log.user.info('Fault Isolation has been performed and results are stored in files: %s' % 
                      ", ".join(("<%s,%s>" % p for p in out_files)))
        
        return out_files
 
    pass # end of class
# ----------------------------------------------------------------------
        
if __name__ == "__main__":
    main = FsapFaultIsolation()
    main.run()

    
