#!/usr/bin/env python

# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================



"""Fault Tolerance Evaluation

Compute fault tolerance measures given a set of properties and a
corresponding set of fault trees.

Inputs      Props: A set of properties.
	    FTs: a set of fault trees.

Outputs     FTM: The fault tolerance measure.

Processing  Given a component, a set of properties, and a set of corresponding fault trees
            generated for the component, the COMPASS toolset shall compute fault
            tolerance measures. Fault tolerance measures shall reflect the capability of
            the component to sustain single or multiple faults and shall be computed
            as follows: for each cardinality, a table shall report the number of unique
            minimal cut sets with that cardinality.
"""

import booter # import this as first module in scripts

import property_management.src.constants as prop_const
import property_management.src.api as prop_api
from utils import fsap_process
from utils import exceptions, log, cutsets
from utils.misc import relpath, get_out_fname
import options
import slim_compiler.src.smv.api as smv_api

import subprocess
import os

from runscripts.basic import ToolScript
from runscripts.model_checking__srs_015 import NuSMVBasedScript
from runscripts.fault_tree_generation__srs_017 import FsapGeneralScript


from runscripts.basic import ToolScript
class FaultToleranceEvaluation (ToolScript):
    """This is the script that can be used to generate the Fault
    Tolerance Measures for a set of properties and fault trees."""

    def register_options(self, opt_parser):
        # Specific options for Fault Detection
        super(FaultToleranceEvaluation, self).register_options(opt_parser)

        opt_parser.add_option("--dynamic", dest="FSAP_USE_DYNAMIC",
                              action="store", default=False,
                              help="Enable the generation of Dynamic Fault Trees")

	opt_parser.add_option("--fts-files", dest="FSAP_FTS_FILES",
			      help="Fault trees XML files")

        return


    def check_options(self, opt_parser, options):
        super(FaultToleranceEvaluation, self).check_options(opt_parser, options)
        
	# fault tree files are mandatory
        if opt_parser.values.FSAP_FTS_FILES is None:
	    opt_parser.error("No fault trees were specified.")
            
        return


    def store_options(self, opt_parser, options):
        super(FaultToleranceEvaluation, self).store_options(opt_parser, options)

        if opt_parser.values.FSAP_FTS_FILES:
            self.fts_files = opt_parser.values.FSAP_FTS_FILES
	else:
	    self.fts_files = None

        return


    def action(self):
        """Builds FSAP and performs Fault Tolerance Evaluation
	for the given fault trees and properties."""
        super(FaultToleranceEvaluation, self).action()
        
	# get ft files
	fts = self.fts_files.split(" ")

	# instantiate model

        log.system.debug("Running Fault Tolerance Evaluation, taking files: %s" % ", ".join(fts))

  	result_table = []
        csm = cutsets.get_unique_cut_sets_from_files(fts)
        for card in csm: result_table.append((card, len(csm[card])))

        log.system.debug("Done with Fault Tolerance Evaluation")      

	# generate output filename
        ftoe_fname = get_out_fname(options.BASE_FILE_NAME + ".ftoe")

	# print result
        out = open(ftoe_fname, "w")
        out.write("# %s\n" % "\t".join(("Cardinality","Cut-sets")))

        out.writelines(["\t".join(map(str,row)) for row in result_table] + ["\n"])
	out.close()

	# announce results
	log.user.info('Fault Tree Evaluation results are stored in "%s".' % ftoe_fname)

        return
    pass # end of class
# ----------------------------------------------------------------------
        
if __name__ == "__main__":
    main = FaultToleranceEvaluation()
    main.run()

