#!/usr/bin/env python

# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================



"""Model Simulation

Produce a simulation trace of an operational model (optionally
extended with a fault model and/or an FDIR model).

Inputs      M : an operational model to be simulated.
            FM : a fault model (optional).
            FDM : an FDIR model (optional).
            C : a set of user constrants on the simulation.

Outputs     Tr : a simulation trace.

Processing  Using model checking techniques, the COMPASS toolset
            shall produce a simulation trace of the given
            operational model (optionally extended with a fault
            model and an FDIR model). The trace shall conform to
            a set of simulation constraints. The output trace shall
            be generated in XML format.
"""

import booter # import this as first module in scripts

import property_management.src.api as properties_api
from invokers import slim, nusmv
from utils import exceptions, log
import utils.misc
import options

import subprocess

from runscripts.model_checking__srs_015 import NuSMVGeneral
class NuSMVSimulation (NuSMVGeneral):
    """This script can be used to perform model simulation via nusmv"""


    def register_options(self, opt_parser):
        # Specific options for nusmv
        super(NuSMVSimulation, self).register_options(opt_parser)

        def __check_positive(option, opt_str, value, parser):
            if value < 0: parser.error("Simulation length must be positive.")
            setattr(parser.values, option.dest, value)
            return
        opt_parser.add_option("--sim-length", 
                              action="callback", callback=__check_positive, 
                              dest="NUSMV_SIM_LENGTH", type="int",
                              help="Length of traces obtained with simulation")
        return

    def action(self):
        super(NuSMVSimulation, self).action()

        trace = nusmv.simulation_random(self.nusmv, None)
        if trace: log.user.info("Trace of simulation can be found in '%s'" % utils.misc.relpath(trace))
        else: log.user.info("No trace is available (too strong constraints?)")
        return

    pass # end of class
# ----------------------------------------------------------------------
        
if __name__ == "__main__":
    
    main = NuSMVSimulation()
    main.run()

    
