#!/usr/bin/env python

# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


""" TFPG Effectiveness Check """

import booter # import this as first module in scripts

#import property_management.src.constants as prop_const
#import property_management.src.api as prop_api

#from utils import exceptions, log
#from utils.misc import relpath, get_out_fname

from utils.misc import get_out_fname

#import slim_compiler.src.smv.api as smv_api
import subprocess
import os

#from runscripts.basic import ToolScript

#import invokers.slim as slim



import options

from utils import fame_tfpg_process
#from runscripts.model_checking__srs_015 import NuSMVBasedScript

from tfpg2smv.tfpg2smv import parse_to_internal_struct, dump_tfpg_as_smv, compute_smv_node_name

from utils.fame_tfpg_process import FameTfpgProcess

from utils import log
import optparse
import tempfile

import logging


def main():

    log.init()

    ###########################################################################
    #############################  DEFINE OPTIONS  ############################
    ###########################################################################

    usage = ("usage: %prog [options]")
    opt_parser = optparse.OptionParser(usage)

    # TFPG xml file
    opt_parser.add_option("--tfpg-file", dest="TFPG_FILE_NAME",
        help="XML file containing the TFPG model.")

    # target_fm_set (colon-separated list of failure modes)
    opt_parser.add_option("--target-fm-set", dest="TFPG_TARGET_FM_SET",
        help="Specifies the set of FMs that need to be diagnosed (as a group); \
        colon-separated.")

    # sampling_rate (larger than 0, can have decimal part)
    opt_parser.add_option("--sampling-rate", dest="TFPG_SAMPLING_RATE",
        action="store", type = "float", default=1.0, help="Specifies the \
        interval at which the monitored system is sampled.")

    # target_system_mode (None; or, system mode to be considered)
    opt_parser.add_option("--target-system-mode", dest="TFPG_TARGET_SYSTEM_MODE",
        help="Specifies the system mode for which effectiveness should be analyzed \
        (optional).")

    # free system mode
    opt_parser.add_option("--free-system-mode", dest="TFPG_FREE_SYS_MODE",
        action="store_true", default=False, help="Allow system mode to change over time.")

    # use BMC
    opt_parser.add_option("--use-bmc", dest="NUSMV_USE_BMC",
        action="store_true", default=False, help="Use bounded model checking.")

    # BMC bound
    opt_parser.add_option("--bmc-length", dest="NUSMV_BMC_LENGTH",
        action="store", type="int", default=10, help="Maximum path length for BMC.")

    ###########################################################################
    ##########################  PARSE/CHECK OPTIONS  ##########################
    ###########################################################################

    (opts, args) = opt_parser.parse_args()

    # tfpg file name
    if not opts.TFPG_FILE_NAME:
        opt_parser.error("The XML file describing the TFPG was not specified.")
        return
    else:
        options.TFPG_FILE_NAME = opts.TFPG_FILE_NAME
        log.system.debug("TFPG_FILE_NAME: %s" % options.TFPG_FILE_NAME)

    # target failuremode set
    if not opts.TFPG_TARGET_FM_SET:
        opt_parser.error("The target failuremode set was not specified.")
        return
    else:
        options.TFPG_TARGET_FM_SET = opts.TFPG_TARGET_FM_SET.split(':')
        log.system.debug("TFPG_TARGET_FM_SET: %s" % options.TFPG_TARGET_FM_SET)

    # sampling rate
    if opts.TFPG_SAMPLING_RATE:
        options.TFPG_SAMPLING_RATE = opts.TFPG_SAMPLING_RATE
    log.system.debug("TFPG_SAMPLING_RATE: %f" % options.TFPG_SAMPLING_RATE)

    # target system mode
    if opts.TFPG_TARGET_SYSTEM_MODE:
       options.TFPG_TARGET_SYSTEM_MODE = opts.TFPG_TARGET_SYSTEM_MODE
    log.system.debug("TFPG_TARGET_SYSTEM_MODE: %s" % options.TFPG_TARGET_SYSTEM_MODE)

    # free system mode
    if opts.TFPG_FREE_SYS_MODE:
       options.TFPG_FREE_SYS_MODE = opts.TFPG_FREE_SYS_MODE
    else:
       options.TFPG_FREE_SYS_MODE = False
    log.system.debug("TFPG_FREE_SYS_MODE: %s" % options.TFPG_FREE_SYS_MODE)

    # use BMC
    options.NUSMV_USE_BMC = opts.NUSMV_USE_BMC
    log.system.debug("NUSMV_USE_BMC: %s" % options.NUSMV_USE_BMC)

    if opts.NUSMV_BMC_LENGTH:
        options.NUSMV_BMC_LENGTH = opts.NUSMV_BMC_LENGTH
    log.system.debug("NUSMV_BMC_LENGTH: %s" % options.NUSMV_BMC_LENGTH)

    # set/enable other options

    options.BASE_FILE_NAME = os.path.splitext(options.TFPG_FILE_NAME)[0]

    ###########################################################################
    ##################################  MAIN  #################################
    ###########################################################################

    tfpg_smv_fname = os.path.splitext(options.TFPG_FILE_NAME)[0] + ".smv"

    log_out, logfname = tempfile.mkstemp(suffix='.fametfpgout',
        prefix=os.path.splitext(tfpg_smv_fname)[0], dir=".", text=True)

    log.system.debug("Invoking FAME-TFPG, with redirected to " + logfname)

    system_mode_var_name = "system_mode"
    time_tick = "time_tick"
    frozen_sys_mode = True
    dump_validation_props = False

    # dump TFPG as SMV

    tfpg_structural = parse_to_internal_struct(options.TFPG_FILE_NAME)
    dump_tfpg_as_smv(tfpg_structural, tfpg_smv_fname, time_tick, options.TFPG_SAMPLING_RATE,
        frozen_sys_mode, dump_validation_props)

    # get observables (monitored nodes, system mode, time tick)

    observables = tfpg_structural.get_monitored_discrepancies()

    # constructs a new instance of the tfpg_fame process

    fame_tfpg_proc = FameTfpgProcess(tfpg_smv_fname, ord_fname=None, \
        need_msat=False, args=[], stdout=subprocess.PIPE,            \
        stderr=subprocess.STDOUT, dumpers=[os.fdopen(log_out, "w")])

    fame_tfpg_proc.set_tmp_filename_prefix(get_out_fname(options.BASE_FILE_NAME + "__temp_fametfpg__"))

    # send task to fame-tfpg process

    result = fame_tfpg_proc.check_tfpg_diagnosability (observables, options.TFPG_TARGET_FM_SET,
        system_mode_var_name, time_tick, options.TFPG_TARGET_SYSTEM_MODE,
        options.NUSMV_USE_BMC, options.NUSMV_BMC_LENGTH)

    # announce results

    log.user.info('An effectiveness check has been performed for the TFPG in %s.' % options.TFPG_FILE_NAME)

    if result['result'] is None:
        log.user.info('An error has occurred during TFPG Effectiveness Analysis.')
    elif result['result'] == 'diagnosable':
        log.user.info('The TFPG is effective for diagnosis under the given constraints.')
    elif result['result'] == 'not_diagnosable':
        log.user.info('The TFPG is not effective for diagnosis under the given ' +
            'constraints (see counterexample in %s and %s).' % (result['critical_pair'][0], result['critical_pair'][1]))
    else:
        assert result['result'] == 'diagnosability_unknown'
        log.user.info('No counterexample has been found within bound %d.' % options.NUSMV_BMC_LENGTH)

    # close process

    fame_tfpg_proc.kill()

    return


# ----------------------------------------------------------------------
        
if __name__ == "__main__":
    main()
    
