#!/usr/bin/env python

# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================

""" TFPG Validation """

import booter  # import this as first module in scripts

import options, tempfile, os, subprocess
from utils import log

from runscripts.model_checking__srs_015 import NuSMVGeneral
from invokers import nusmv
from utils.misc import get_out_fname

from hasdel_tools.src.xml.tfpg import TfpgDocument, tfpg_doc_to_tfpg_internal_struct, \
     tfpg_internal_struct_to_tfpg_doc
from hasdel_tools.src.tfpg.tfpg import NodeType

class TfpgValidation (NuSMVGeneral):

    def get_usage(self):
        return "usage: %prog --tfpg_file <filename>"

    def register_options(self, opt_parser):
        # Specific options for nusmv
        super(TfpgValidation, self).register_options(opt_parser)
        opt_parser.add_option("--tfpg-file", dest="TFPG_FILE_NAME",
                              help="TFPG xml file")
        return

    def check_options(self, opt_parser, options):
        if not options.TFPG_FILE_NAME:
            opt_parser.error("The TFPG file was not specified.")
        return

    def store_options(self, opt_parser, options):
        assert opt_parser.values.TFPG_FILE_NAME
        options.TFPG_FILE_NAME = opt_parser.values.TFPG_FILE_NAME
        return

    def action(self):
        """Builds nusmv and performs model checking on all
        properties that is possible to model check"""

        filename = options.TFPG_FILE_NAME

        # Checks if the TFPG file exists.

        if not os.path.exists(filename):
            print "The file: " + relpath(filename) + " does not exist!"
            return

        print "Validating TFPG file '%s'... " % filename

        # Gets the TfpgDocument and any validation errors. Parses the TFPG file.

        tfpg_doc, validation_errors = TfpgDocument().parse(filename)

        if len(validation_errors) > 0:
            print "The file '"+filename+"' contains validation errors:"
            for err in validation_errors:
                print "   >", str(err)
            return 

        # Gets any syntax errors. 

        syntax_errors = TfpgDocument().get_syntax_errors_in_tfpg_document(tfpg_doc)

        if len(syntax_errors) > 0:
            print "The file '"+filename+"' contains syntax errors:"
            for err in syntax_errors:
                print "   >", str(err)
            return 

        # Converts the TFPG file from the TfpgDocument to the TfpgInternalStruct

        tfpg_internal_struct = tfpg_doc_to_tfpg_internal_struct(tfpg_doc)
       
        # Gets any errors from the Tfpg. 

        error_list = tfpg_internal_struct.get_errors_list()
        if len(error_list) > 0:
            print "The file '"+filename+"' contains syntax errors:"
            for err in error_list:
                print "   >", str(err)
            return
        
        # Gets any errors from the Tfpg Edges. 

        for node in tfpg_internal_struct.get_nodes_list():
            if len(node.get_node_errors_list()) > 0:
                print "The file '"+filename+"' contains syntax errors:"
                for err in node.get_node_errors_list():
                    print "   >", str(err)
                return

        # Gets any errors from the Tfpg Nodes. 

        for edge in tfpg_internal_struct.get_edges_list():
            if len(edge.get_edge_errors_list()) > 0:
                print "The file '"+filename+"' contains syntax errors:"
                for err in edge.get_edge_errors_list():
                    print "   >", str(err)
                return

        # Checks if a Discrepancy must have at least one incoming edge

        error = False
        discrepancy = ""
        nodes_name_type_list = []
        edges_destNode_list = []

        for node in tfpg_internal_struct.get_nodes_list():
            nodes_name_type_list.append([node.get_name(), node.get_type()])

        for edge in tfpg_internal_struct.get_edges_list():
            edges_destNode_list.append(edge.get_destNode().get_name())

        for nd in nodes_name_type_list:
            if (nd[1] != NodeType.FM) and (nd[0] not in edges_destNode_list):
                print "The file: " + relpath(filename) + " contains syntax errors:\n\t" + str(DiscrepancyWithoutIncomingEdge(nd[0]))
                return

        # all validation checks passed

        print "All checks passed."
        return

    pass  # end of class
# ----------------------------------------------------------------------

if __name__ == "__main__":

    main = TfpgValidation()
    main.run()

