#!/usr/bin/env python

# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


""" TFPG Synthesis """


import booter # import this as first module in scripts

import property_management.src.constants as prop_const
import property_management.src.api as prop_api

from utils import fsap_process
from utils import exceptions, log
from utils.misc import relpath

import options
import slim_compiler.src.smv.api as smv_api
import subprocess
import os

import tempfile
from utils.misc import get_out_fname

from runscripts.basic import ToolScript

import invokers.slim as slim

from runscripts.model_checking__srs_015 import NuSMVBasedScript

from utils.fame_tfpg_process import FameTfpgProcess


class TfpgSynthesis(NuSMVBasedScript):

    def __init__(self):
        super(TfpgSynthesis, self).__init__()
        return

    def register_options(self, opt_parser):
        super(TfpgSynthesis, self).register_options(opt_parser)

        opt_parser.add_option("--associations-file", dest="TFPG_ASSOC_FILE_NAME",
                              help="TFPG SLIM associations file")

        opt_parser.add_option("--bmc-length", dest="NUSMV_BMC_LENGTH",
            action="store", type="int", default=10, help="Maximum path length for BMC.")

        return

    def check_options(self, opt_parser, options):
        super(TfpgSynthesis, self).check_options(opt_parser, options)
        
        if not options.FAULT_INJECTION_FILE_NAME:
            opt_parser.error("Failure injection file was not specified.")

        if not options.TFPG_ASSOC_FILE_NAME:
            opt_parser.error("The TFPG SLIM associations file was not specified.")
 
        return

    def store_options(self, opt_parser, options):
        super(TfpgSynthesis, self).store_options(opt_parser, options)

        assert opt_parser.values.TFPG_ASSOC_FILE_NAME
        options.TFPG_ASSOC_FILE_NAME = opt_parser.values.TFPG_ASSOC_FILE_NAME

        assert opt_parser.values.NUSMV_BMC_LENGTH
        options.NUSMV_BMC_LENGTH = opt_parser.values.NUSMV_BMC_LENGTH

        return

    def action(self):
        super(TfpgSynthesis, self).action()

        log_out, logfname = tempfile.mkstemp(suffix='.fametfpgout',
            prefix=os.path.splitext(self.smv_fname)[0], dir=".", text=True)

        print "log_out:",log_out
        print "logfname:",logfname

        log.system.debug("Invoking FAME-TFPG, with redirected to " + logfname)

        print "--> parsing associations file (%s)" % options.TFPG_ASSOC_FILE_NAME
        from tfpg2smv.tfpg2smv import get_slim_associations_from_file
        associations = get_slim_associations_from_file(options.TFPG_ASSOC_FILE_NAME)

        print "--> dumping extended smv file (%s)" % self.smv_fname
        from tfpg2smv.tfpg2smv import dump_extension_for_synthesis_to_smv
        dump_extension_for_synthesis_to_smv(self.root_impl, associations, self.smv_fname)

        sa_fm_list_name = os.path.splitext(self.smv_fname)[0] + "_sa_fm_list.xml"
        print "--> dumping associations for cutset analysis (%s)" % sa_fm_list_name
        from tfpg2smv.tfpg2smv import dump_discrepancies_for_synthesis_to_xml
        dump_discrepancies_for_synthesis_to_xml (associations, sa_fm_list_name)

        print "--> instantiating tfpg fame process"
        nusmv_args = ["-sa_compass", "-sa_compass_task", sa_fm_list_name]
        fame_tfpg_proc = FameTfpgProcess(self.smv_fname, ord_fname=None, \
            need_msat=True, args=nusmv_args, stdout=subprocess.PIPE,     \
            stderr=subprocess.STDOUT, dumpers=[os.fdopen(log_out, "w")])
        fame_tfpg_proc.set_tmp_filename_prefix(get_out_fname(options.BASE_FILE_NAME + "__temp_fametfpg__"))

        print "--> calling synthesis function"
        tfpg_res_name = os.path.splitext(self.smv_fname)[0] + "_synth_tfpg.txml"

        system_modes = [d for d in associations if associations[d][2] == "TfpgMode"]
        failure_modes = [d for d in associations if associations[d][2] == "FailureMode"]
        
        discrepancies = []
        for d in associations:
            if associations[d][2] == "UnmonitoredDiscr":
                discrepancies.append((d,"not_monitored"))
            elif associations[d][2] == "MonitoredDiscr":
                discrepancies.append((d,"monitored"))
            else: continue

        res = fame_tfpg_proc.synthesize_tfpg(self.smv_fname, tfpg_res_name, system_modes, failure_modes, discrepancies, "", options.NUSMV_BMC_LENGTH)

        print "Result:", res
        if res == "ok":
		    print "Synthesized TFPG:", tfpg_res_name

        # close process
        fame_tfpg_proc.kill()

        return

    pass # end of class
# ----------------------------------------------------------------------
        

if __name__ == "__main__":
    main = TfpgSynthesis()
    main.run()

