#!/usr/bin/env python

# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


"""Time Divergence Checking

For all clock data-subcomponents, verify whether a time-divergence
conditions exists.

TBD
"""

import booter  # import this as first module in scripts

from slim_compiler.src.smv.walkers import ClockCollectorWalker
import slim_compiler.src.smv.api as smvapi
from slim_compiler.src.smv import support
from slim_compiler.src.dataStructure import nominalbehaviour

from utils import log

from invokers import nusmv
import options

from runscripts.model_checking__srs_015 import NuSMVGeneral

import itertools
import operator
import csv
import os.path


class NuXMVClockDivergenceChecking (NuSMVGeneral):
    """This is the script that can be used to perform
    clock-divergence checking"""

    def register_options(self, opt_parser):
        # Specific options for nusmv
        super(NuSMVGeneral, self).register_options(opt_parser)

        def __check_positive(option, opt_str, value, parser):
            if value < 0:
                parser.error("Value must be positive.")
            setattr(parser.values, option.dest, value)
            return

        opt_parser.add_option(
            "--bmc-length",
            action="callback", callback=__check_positive,
            dest="NUSMV_BMC_LENGTH", type="int",
            help="Problem max length used when performing BMC")

        opt_parser.add_option(
            "--clock-bound",
            action="callback", callback=__check_positive,
            dest="NUSMV_CLOCK_BOUND", type="string",
            help="Clock max bound for used when "
            "performing clock diverge checking (can be None). "
            "This is ignored when reading from clock bound file")

        opt_parser.add_option(
            "--gen-clock-bound",
            dest="GEN_CLOCK_BOUND",
            action="store_true", default=False,
            help="Generates a CSV file (taking filename from option "
            "clock-bound-file which is required) containing all clock "
            "data subcomponents, and "
            "associating each with a default clock bound read either from "
            "option clock-bound if specified, or from "
            "options.NUSMV_CLOCK_BOUND if not specified")

        opt_parser.add_option(
            "--clock-bound-file",
            dest="CLOCK_BOUND_FNAME", type="string",
            help="The clock bound file, written used when generating it, or "
            "read for later manual"
            "editing and customization. This may be handy when performing "
            "clock diverge checking to specify specific bounds for "
            "desidered clocks")

        return

    def check_options(self, opt_parser, options):
        super(NuXMVClockDivergenceChecking, self).check_options(
            opt_parser, options)

        # when generating clock bounds, file name must be provided
        if options.GEN_CLOCK_BOUND and not options.CLOCK_BOUND_FNAME:
            opt_parser.error("Fatal Error: when using --gen-clock-bound, "
                             "--clock-bound-file must also be used")

        # when specified for reading, clock bound file must exist
        if (not options.GEN_CLOCK_BOUND and
                options.CLOCK_BOUND_FNAME and
                not os.path.isfile(options.CLOCK_BOUND_FNAME)):

            opt_parser.error("Fatal Error: input file for clock bound not "
                             "found: '%s'" % options.CLOCK_BOUND_FNAME)

        return

    def action(self):
        """Builds nusmv and performs model checking on all
        properties that is possible to model check"""

        options.SMV_FAIR_TIMED = False
        options.SMV_GENERATE_TIME_VAR = True

        super(NuXMVClockDivergenceChecking, self).action()

        if not self.timed:
            log.user.info(
                '\nClock-Divergence Checking: Model is not timed, '
                'skipping checking')
            return

        if options.GEN_CLOCK_BOUND:
            self.gen_clock_bounds(options.CLOCK_BOUND_FNAME)
        else:
            clock_bounds = (
                self.read_clock_bounds_from_file(options.CLOCK_BOUND_FNAME)
                if options.CLOCK_BOUND_FNAME
                else self.read_clock_bounds_from_model())

            self.check_clock_bounds(clock_bounds)

    def check_clock_bounds(self, clock_bounds):
        """clock_bounds  is a list of pairs (clock_full_name, clock_bound)"""

        # split wrt impl_name, clock_name
        _tmp = ((clock_name.split("."), clock_bound)
                for clock_name, clock_bound in clock_bounds)

        # split to impl_fullname, clock_name, clock_full_name, clock_bound
        clock_list = [(".".join(name[:-1]),
                       name[-1],
                       ".".join(name),
                       clock_bound)
                      for name, clock_bound in _tmp]

        for k, g in itertools.groupby(
                sorted(clock_list), key=operator.itemgetter(0)):

            log.user.info(
                '\nClock-Divergence Checking in "{impl}":'.format(
                    impl=k if k else "(root)"))

            for impl_full_name, clock_name, clock_full_name, clock_bound in g:

                dsc = support.get_subcomponent_from_path(self.root_impl,
                                                         clock_full_name)
                assert isinstance(dsc, nominalbehaviour.DataSubcomponent)
                assert isinstance(dsc.data_type, nominalbehaviour.Clock)

                if clock_bound not in ("-1", "", None, "None"):
                    bound_smv = smvapi.parse_and_convert_numeric_expression(
                        self.root_impl, clock_bound,
                        dest_tu=dsc.data_type.time_unit)
                else:
                    bound_smv = None

                clock_smv = smvapi.parse_and_convert_id(
                    self.root_impl, clock_full_name)

                # invokes xsap
                _ret, _val = nusmv.check_clock_divergence(
                    self.nusmv, clock_smv, clock_bound=bound_smv)

                result = self.nusmv.clock_diverge_result_to_string(_ret)
                if self.nusmv.BOUNDED == _ret:
                    result += " (time: '%s')" % _val
                elif self.nusmv.UNBOUNDED == _ret:
                    result += " (trace in '%s')" % _val

                bound = ('set with bound "{0}"'.format(clock_bound)
                         if bound_smv
                         else 'set with no bound')
                msg = '    clock "{clock}" {bound} is {result}'.format(
                    clock=clock_name,
                    bound=bound,
                    result=result)
                log.user.info(msg)

    def read_clock_bounds_from_file(self, fname):
        with open(fname, 'rb') as csvfile:
            reader = csv.reader(csvfile, delimiter=' ', quotechar='"')
            return [row for row in reader]

    def read_clock_bounds_from_model(self):
        """Returns a list of pair (clock_full_name, bound) both strings"""

        res = []
        walker = ClockCollectorWalker()
        walker.walk(self.root_impl)

        for k, g in itertools.groupby(
                sorted(walker.get_found_clocks()),
                key=operator.itemgetter(2)):  # 2 is implementation fullname
            res.extend((clock_fullname, str(options.NUSMV_CLOCK_BOUND))
                       for clock_fullname, _, _, _ in g)
        return res

    def gen_clock_bounds(self, fname):
        with open(fname, 'wb') as csvfile:
            writer = csv.writer(csvfile, delimiter=' ',
                                quotechar='"', quoting=csv.QUOTE_MINIMAL)

            for row in self.read_clock_bounds_from_model():
                writer.writerow(row)

    pass  # end of class
# ----------------------------------------------------------------------

if __name__ == "__main__":

    main = NuXMVClockDivergenceChecking()
    main.run()
