# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================
from utils.py2xml import Py2XML
from slim_compiler.src.dataStructure.visitors import printVisitor
from slim_compiler.src.dataStructure.expressions import TimeUnitType


"""This module provides classes for storing a SLIM specification.
""" 

from slim_compiler.src.errorManagement.errors import *
from slim_compiler.src.dataStructure.errorbehaviour import ErrorModelType, ErrorModelImplementation
from slim_compiler.src.dataStructure.nominalbehaviour import ComponentType, ComponentImplementation
from slim_compiler.src.deprecated import *
from slim_compiler.src.dataStructure.located_info import LocatedInformation
from slim_compiler.src.dataStructure.visitors.compileVisitor import CheckVisitor
from utils import log

class CompileInformation(object):
    """This class is the top level data container for a SLIM specification.
    """
    
    def __init__(self):
        """Initializes a compile information object.
        """
        # stores the filenames that were parsed
        self.files = []
        # stores the current file
        self.current_file = ""
        # stores the packages and their locations
        self.packages = {}
        # stores the current package
        self.current_package = ""
        # stores the current component
        self.current_component = ""
        # stores the errors that were recognized during checking
        self.error_collector = ErrorCollector()
        # indicates whether model uses time scaling (i.e. time units for clocks)
        self.uses_time_scaling = False
        # when time scaling enabled, defines base unit for clocks
        self.base_unit = TimeUnitType.HOUR
    
    def to_XML(self):
        return Py2XML().convert(self)
    
    def to_SXML(self, name, assertNoErrors):
        return printVisitor.toSXMLVisitor().visit(self, name, assertNoErrors)
    
    @deprecated    
    def print_packages(self):
        """Prints a listing of all packages contained in the SLIM specification
        with their definition places.
        """
        for package in self.packages.keys():
            package_obj = self.packages[package]
            print "Package \"%s\" declared in %s at line %d:%d." \
            % (package, package_obj.file, package_obj.line, package_obj.pos)
    
    @deprecated
    def print_files(self):
        """Prints a listing of all files that were parsed.
        """
        for file in self.files:
            print file
        
    def get_root_components(self):
        """Retrieves all component implementation objects that qualify for being
        a root component, i.e. that are not referenced as a subcomponent type.
        """
        non_root_components = set()
        
        possible_root_components = set()
        for package in self.packages.values():
            for subpackage in [package.private_subpackage,
                               package.public_subpackage]:
                possible_root_components |= set(subpackage.implementations.values()) 
        
        for impl in possible_root_components:
            for subcomp in impl.non_data_subcomponents.values():
                ref_impl = subcomp.get_referenced_component_implementation()
                if ref_impl is not None:
                    non_root_components.add(ref_impl)
        
        return possible_root_components - non_root_components
    
    def get_error_root_components(self):
        """Retrieves all component implementation objects that qualify for being
        a root component, i.e. that are not referenced as a subcomponent type.
        """
        possible_root_components = set()
        for package in self.packages.values():
            for subpackage in [package.private_subpackage,
                               package.public_subpackage]:
                possible_root_components |= set(subpackage.error_implementations.values()) 
        
        return possible_root_components
    
    def get_fdir_components(self):
        """Retrieves all component implementation objects that are FDIR
        components
        
        @return:  A list FDIR component implementations,
        @rtype:   list of ComponentImplementation
        """
        fdir_components = []
        for package in self.packages.values():
            for subpackage in [package.private_subpackage,
                               package.public_subpackage]:
                for impl in subpackage.implementations.values():
                    if impl.fdir:
                        fdir_components.append(impl)
        return fdir_components
    
    def add_package(self, package_name, line, pos):
        """Adds a package with the given name and position information to the
        data container and returns the corresponding package object. If the
        package already exists, the existing package object is returned.
        
        @param    package_name:
                    The name of the package to add.
        @type     package_name:
                    string
        @param    line:
                    The line in which the package was defined.
        @type     line:
                    int
        @param    pos:
                    The column position at which the package was defined.
        @type     pos:
                    int
                    
        @return:  The package object corresponding to the given package name.
        @rtype:   A compile.Package object.
        """
        package = self.get_package(package_name) 
        if package is None:
            self.packages[package_name] = Package(self, package_name,
                                                  self.current_file, line, pos)
        else:
            # only add an error if it is not the default package, because
            # the default package can be used in multiple files 
            if package_name != "__default__":
                self.add_error(
                        PackageAlreadyDeclaredError(self.current_file, line,
                                                    pos, "A-2", package))
        return self.packages[package_name]
        
    def add_error(self, error):
        """Adds an error to the error collector object.
        
        @param    error:
                    The error to add.
        @type     error:
                    A subclass of errors.AbstractError.
        """
        self.error_collector.add(error)

    def get_package(self, package_name):
        """Returns the package object with the given name if it exists and None
        otherwise.
        
        @param    package_name:
                    The package name to be searched.
        @type     package_name:
                    string
        @return:  The package object with the given name.
        @rtype:   A compile.Package object.
        """
        if package_name in self.packages:
            return self.packages[package_name]       
        else:
            return None

    def get_total_number_of_transitions(self):
        """Returns the total number of transitions contained in all components
        contained in the compile information object.
        
        @return:    The total number of transitions.
        @rtype:     int
        """
        result = 0
        for package in self.packages.values():
            result += package.get_total_number_of_transitions()
        return result

    def get_total_number_of_modes(self):
        """Returns the total number of modes contained in all components
        contained in the compile information object.
        
        @return:    The total number of modes.
        @rtype:     int
        """
        result = 0
        for package in self.packages.values():
            result += package.get_total_number_of_modes()
        return result
    
    def get_total_number_of_component_types(self):
        """Returns the total number of component types contained in all
        components contained in the compile information object.
        
        @return:    The total number of component types.
        @rtype:     int
        """
        result = 0
        for package in self.packages.values():
            result += package.get_total_number_of_component_types()
        return result

    def get_total_number_of_component_implementations(self):
        """Returns the total number of component implementations contained in
        all components contained in the compile information object.
        
        @return:    The total number of component implementations.
        @rtype:     int
        """
        result = 0
        for package in self.packages.values():
            result += package.get_total_number_of_component_implementations()
        return result

    def get_total_number_of_nondata_component_instances(self):
        """Returns the total number of non-data component instances contained in
        all components contained in the compile information object.
        
        @return:    The total number of non-data component instances.
        @rtype:     int
        """
        result = 0
        for package in self.packages.values():
            result += package.get_total_number_of_nondata_component_instances()
        return result

    def get_total_number_of_data_component_instances(self):
        """Returns the total number of data component instances contained in all
        components contained in the compile information object.
        
        @return:    The total number of data component instances.
        @rtype:     int
        """
        result = 0
        for package in self.packages.values():
            result += package.get_total_number_of_data_component_instances()
        return result
    
    def get_total_number_of_outgoing_data_ports(self):
        """Returns the total number of outgoing data ports contained in all
        components contained in the compile information object.
        
        @return:    The total number of outgoing data ports.
        @rtype:     int
        """
        result = 0
        for package in self.packages.values():
            result += package.get_total_number_of_outgoing_data_ports()
        return result

    def get_total_number_of_incoming_data_ports(self):
        """Returns the total number of incoming data ports contained in all
        components contained in the compile information object.
        
        @return:    The total number of incoming data ports.
        @rtype:     int
        """
        result = 0
        for package in self.packages.values():
            result += package.get_total_number_of_incoming_data_ports()
        return result

    def get_total_number_of_incoming_event_ports(self):
        """Returns the total number of incoming event ports contained in all
        components contained in the compile information object.
        
        @return:    The total number of incoming event ports.
        @rtype:     int
        """
        result = 0
        for package in self.packages.values():
            result += package.get_total_number_of_incoming_event_ports()
        return result

    def get_total_number_of_outgoing_event_ports(self):
        """Returns the total number of outgoing event ports contained in all
        components contained in the compile information object.
        
        @return:    The total number of outgoing event ports.
        @rtype:     int
        """
        result = 0
        for package in self.packages.values():
            result += package.get_total_number_of_outgoing_event_ports()
        return result
    
    def get_total_number_of_data_port_connections(self):
        """Returns the total number of data port connections contained in all
        components contained in the compile information object.
        
        @return:    The total number of data port connections.
        @rtype:     int
        """
        result = 0
        for package in self.packages.values():
            result += package.get_total_number_of_data_port_connections()
        return result

    def get_total_number_of_flows(self):
        """Returns the total number of flows contained in all components
        contained in the compile information object.
        
        @return:    The total number of flows.
        @rtype:     int
        """
        result = 0
        for package in self.packages.values():
            result += package.get_total_number_of_flows()
        return result

    def get_total_number_of_event_port_connections(self):
        """Returns the total number of event port connections contained in all
        components contained in the compile information object.
        
        @return:    The total number of event port connections.
        @rtype:     int
        """
        result = 0
        for package in self.packages.values():
            result += package.get_total_number_of_event_port_connections()
        return result
    
    def log_metrics(self):
        """Prints all metrics to the metrics log"""

        log.metrics.info("Metrics for the compiled SLIM model:")
        log.metrics.info("    Total number of component types: %d" % self.get_total_number_of_component_types())
        log.metrics.info("    Total number of component implementations: %d" % self.get_total_number_of_component_implementations())
        log.metrics.info("    Total number of data component instances: %s" % self.get_total_number_of_data_component_instances())
        log.metrics.info("    Total number of other component instances: %s" % self.get_total_number_of_nondata_component_instances())

        log.metrics.info("    Total number of modes: %d" % self.get_total_number_of_modes())
        log.metrics.info("    Total number of transitions: %d" % self.get_total_number_of_transitions())

        log.metrics.info("    Total number of outgoing data ports: %d" % self.get_total_number_of_outgoing_data_ports())
        log.metrics.info("    Total number of incoming data ports: %d" % self.get_total_number_of_incoming_data_ports())
        log.metrics.info("    Total number of data port connections: %s" % self.get_total_number_of_data_port_connections())
        log.metrics.info("    Total number of flows: %s" % self.get_total_number_of_flows())
        log.metrics.info("    Total number of outgoing event ports: %d" % self.get_total_number_of_outgoing_event_ports())
        log.metrics.info("    Total number of incoming event ports: %d" % self.get_total_number_of_incoming_event_ports())
        log.metrics.info("    Total number of event port connections: %s" % self.get_total_number_of_event_port_connections())
        
        


class Package(LocatedInformation):
    """This class is the data container for a package in a SLIM specification.
    """
    
    def __init__(self, compile_information, name, file, line, pos):
        """Initializes a Package object.
        
        @param    compile_information:
                    The compile information object that is the parent of this
                    package object.
        @type     compile_information:
                    compile.CompileInformation
        @param    name:
                    The name of the package.
        @type     name:
                    string
        @param    file:
                    The file in which the package was defined.
        @type     file:
                    string
        @param    line:
                    The line in which the package was defined.
        @type     line:
                    int
        @param    pos:
                    The column position at which the package was defined.
        @type     pos:
                    int
        """
        LocatedInformation.__init__(self, compile_information, "package", file, line, pos)
        self.name = name
	# create a public and a private subpackage
        self.public_subpackage = Subpackage(self, file, line, pos)
        self.private_subpackage = Subpackage(self, file, line, pos)
        self.compile_information = compile_information

    def get_ref_identifier(self):
        """Retrieves an identifier that is used as a reference point for errors."""
        return self.name

    def get_component_type(self, type_name, public_only):
        """Retrieves the component type object with the given name if it exists.
        
        @param    type_name:
                    The name of the component type to be retrieved.
        @type     type_name:
                    string
        @param    public_only:
                    If set, the component type is only searched in the public
                    section of the package.
        @type     public_only:
                    bool
        @return:  The component type object with the given name if it exists and
                  None otherwise.
        @rtype:   A nominalbehaviour.ComponentType object.
        """
        type = self.public_subpackage.get_component_type(type_name)
        if type is None and not public_only:
            type = self.private_subpackage.get_component_type(type_name) 
        return type

    def get_component_implementation(self, impl_name, public_only):
        """Retrieves the component implementation object with the given name if
        it exists.
        
        @param    impl_name:
                    The name of the component implementation to be retrieved.
        @type     impl_name:
                    string
        @param    public_only:
                    If set, the component implementation is only searched in
                    the public section of the package.
        @type     public_only:
                    bool
        @return:  The component implementation object with the given name if it
                  exists and None otherwise.
        @rtype:   A nominalbehaviour.ComponentImplementation object.
        """
        impl = self.public_subpackage.get_component_implementation(impl_name)
        if impl is None and not public_only:
            impl = self.private_subpackage.get_component_implementation(impl_name) 
        return impl
        
    def get_error_type(self, type_name, public_only):
        """Retrieves the error type object with the given name if it exists.
        
        @param    type_name:
                    The name of the error type to be retrieved.
        @type     type_name:
                    string
        @param    public_only:
                    If set, the error type is only searched in the public
                    section of the package.
        @type     public_only:
                    bool
        @return:  The error type object with the given name if it exists and
                  None otherwise.
        @rtype:   A errorbehaviour.ErrorType object.
        """
        type = self.public_subpackage.get_error_type(type_name)
        if type is None and not public_only:
            type = self.private_subpackage.get_error_type(type_name) 
        return type
    
    def get_error_implementation(self, impl_name, public_only):
        """Retrieves the error implementation object with to the given name if
        it exists.
        
        @param    impl_name:
                    The name of the error implementation to be retrieved.
        @type     impl_name:
                    string
        @param    public_only:
                    If set, the error implementation is only searched in
                    the public section of the package.
        @type     public_only:
                    bool
        @return:  The error implementation object with the given name if it
                  exists and None otherwise.
        @rtype:   A errorbehaviour.ErrorImplementation object.
        """
        impl = self.public_subpackage.get_error_implementation(impl_name)
        if impl is None and not public_only:
            impl = self.private_subpackage.get_error_implementation(impl_name) 
        return impl
    
    def get_implementations_for_type(self, type_name, public_only):
        """Retrieves all component implementations referring to the given
        component type name.
        
        @param    type_name:
                    The name of the component type the implementations refer to.
        @type     type_name:
                    string
        @param    public_only:
                    If set, the implementations are only searched in the public
                    section of the package.
        @type     public_only:
                    bool
        @result:  All component implementations referring to the given component
                  type.
        @rtype:   A list of nominalbehaviour.ComponentImplementation objects.
        """
        result = []
        for impl in self.public_subpackage.implementations.values():
            if impl.name.startswith(type_name + "."):
                result.append(impl) 
        if not public_only:
            for impl in self.private_subpackage.implementations.values():
                if impl.name.startswith(type_name + "."):
                    result.append(impl)
        return result

    def get_error_implementations_for_type(self, type_name, public_only):
        """Retrieves all error model implementations referring to the given
        error model type name.
        
        @param    type_name:
                    The name of the error model type the implementations refer
                    to.
        @type     type_name:
                    string
        @param    public_only:
                    If set, the implementations are only searched in the public
                    section of the package.
        @type     public_only:
                    bool
        @result:  All error model implementations referring to the given
                  error model type.
        @rtype:   A list of errorbehaviour.ErrorModelImplementation objects.
        """
        result = []
        for impl in self.public_subpackage.error_implementations.values():
            if impl.name.startswith(type_name + "."):
                result.append(impl) 
        if not public_only:
            for impl in self.private_subpackage.error_implementations.values():
                if impl.name.startswith(type_name + "."):
                    result.append(impl)
        return result

    def get_public_subpackage(self):
        return self.public_subpackage

    def get_private_subpackage(self):
        return self.private_subpackage
    
    def add_error(self, error):
        """Adds an error to the error collector object.
        
        @param    error:
                    The error to add.
        @type     error:
                    A subclass of errors.AbstractError.
        """
        self.parent.add_error(error)

    def get_total_number_of_transitions(self):
        """Returns the total number of transitions contained in all components
        contained in the package.
        
        @return:    The total number of transitions.
        @rtype:     int
        """
        result = 0
        for subpackage in [self.public_subpackage, self.private_subpackage]:
            result += subpackage.get_total_number_of_transitions()
        return result

    def get_total_number_of_modes(self):
        """Returns the total number of modes contained in all components
        contained in the package.
        
        @return:    The total number of modes.
        @rtype:     int
        """
        result = 0
        for subpackage in [self.public_subpackage, self.private_subpackage]:
            result += subpackage.get_total_number_of_modes()
        return result
    
    def get_total_number_of_component_types(self):
        """Returns the total number of component types contained in all
        components contained in the package.
        
        @return:    The total number of component types.
        @rtype:     int
        """
        result = 0
        for subpackage in [self.public_subpackage, self.private_subpackage]:
            result += subpackage.get_total_number_of_component_types()
        return result

    def get_total_number_of_component_implementations(self):
        """Returns the total number of component implementations contained in
        all components contained in the package.
        
        @return:    The total number of component implementations.
        @rtype:     int
        """
        result = 0
        for subpackage in [self.public_subpackage, self.private_subpackage]:
            result += subpackage.get_total_number_of_component_implementations()
        return result

    def get_total_number_of_nondata_component_instances(self):
        """Returns the total number of non-data component instances contained in
        all components contained in the package.
        
        @return:    The total number of non-data component instances.
        @rtype:     int
        """
        result = 0
        for subpackage in [self.public_subpackage, self.private_subpackage]:
            result += subpackage.get_total_number_of_nondata_component_instances()
        return result

    def get_total_number_of_data_component_instances(self):
        """Returns the total number of data component instances contained in all
        components contained in the package.
        
        @return:    The total number of data component instances.
        @rtype:     int
        """
        result = 0
        for subpackage in [self.public_subpackage, self.private_subpackage]:
            result += subpackage.get_total_number_of_data_component_instances()
        return result
    
    def get_total_number_of_outgoing_data_ports(self):
        """Returns the total number of outgoing data ports contained in all
        components contained in the package.
        
        @return:    The total number of outgoing data ports.
        @rtype:     int
        """
        result = 0
        for subpackage in [self.public_subpackage, self.private_subpackage]:
            result += subpackage.get_total_number_of_outgoing_data_ports()
        return result

    def get_total_number_of_incoming_data_ports(self):
        """Returns the total number of incoming data ports contained in all
        components contained in the package.
        
        @return:    The total number of incoming data ports.
        @rtype:     int
        """
        result = 0
        for subpackage in [self.public_subpackage, self.private_subpackage]:
            result += subpackage.get_total_number_of_incoming_data_ports()
        return result

    def get_total_number_of_incoming_event_ports(self):
        """Returns the total number of incoming event ports contained in all
        components contained in the package.
        
        @return:    The total number of incoming event ports.
        @rtype:     int
        """
        result = 0
        for subpackage in [self.public_subpackage, self.private_subpackage]:
            result += subpackage.get_total_number_of_incoming_event_ports()
        return result

    def get_total_number_of_outgoing_event_ports(self):
        """Returns the total number of outgoing event ports contained in all
        components contained in the package.
        
        @return:    The total number of outgoing event ports.
        @rtype:     int
        """
        result = 0
        for subpackage in [self.public_subpackage, self.private_subpackage]:
            result += subpackage.get_total_number_of_outgoing_event_ports()
        return result
    
    def get_total_number_of_data_port_connections(self):
        """Returns the total number of data port connections contained in all
        components contained in the package.
        
        @return:    The total number of data port connections.
        @rtype:     int
        """
        result = 0
        for subpackage in [self.public_subpackage, self.private_subpackage]:
            result += subpackage.get_total_number_of_data_port_connections()
        return result

    def get_total_number_of_flows(self):
        """Returns the total number of flows contained in all components
        contained in the package.
        
        @return:    The total number of flows.
        @rtype:     int
        """
        result = 0
        for subpackage in [self.public_subpackage, self.private_subpackage]:
            result += subpackage.get_total_number_of_flows()
        return result

    def get_total_number_of_event_port_connections(self):
        """Returns the total number of event port connections contained in all
        components contained in the package.
        
        @return:    The total number of event port connections.
        @rtype:     int
        """
        result = 0
        for subpackage in [self.public_subpackage, self.private_subpackage]:
            result += subpackage.get_total_number_of_event_port_connections()
        return result

class Subpackage(LocatedInformation):
    """This class is the data container for a subpackage, i.e. the public
    or private section of a package, of a SLIM specification.
    """
    
    def __init__(self, package, file, line, pos):
        """Initializes a subpackage object.
        
        @param    package:
                    The package object that is the parent of this subpackage
                    object.
        @type     package:
                    A nominalbehaviour.Package object.
        @param    file:
                    The file in which the subpackage was defined.
        @type     file:
                    string
        @param    line:
                    The line in which the subpackage was defined.
        @type     line:
                    int
        @param    pos:
                    The column position at which the subpackage was defined.
        @type     pos:
                    int
        """
        LocatedInformation.__init__(self, package, "subpackage", file, line, pos)
        # stores the component types
        self.types = {}
        # stores the component implementations
        self.implementations = {}
        # stores the error types
        self.error_types = {}
        # stores the error implementations
        self.error_implementations = {}
        # stores the reference to the parent package
        self.package = package
        # stores the name of the package the subpackage is located in
        self.package_name = package.name

    def get_ref_identifier(self):
        """Retrieves an identifier that is used as a reference point for errors."""
        return self.parent.get_ref_identifier()
                
    def add_component_type(self, type_name, category, fdir, line, pos):
        """Adds a component type to the subpackage and returns the corresponding
        component type object.
        
        @param    type_name:
                    The name of the component type.
        @type     type_name:
                    string
        @param    category:
                    The category of the component type.
        @type     category:
                    A constant defined in nominalbehaviour.Category.
        @param    fdir:
                    Determines whether the component type is marked as FDIR.
        @type     fdir:
                    bool
        @param    line:
                    The line in which the component type was defined.
        @type     line:
                    int
        @param    pos:
                    The column position at which the component type was defined.
        @type     pos:
                    int
        @return:  The corresponding component type object.
        @rtype:   A nominalbehaviour.ComponentType object.
        """
        type = self.parent.get_component_type(type_name, False)
        
        if type is None:
            new_type = ComponentType(self, type_name, category,
                                     fdir, self.parent.parent.current_file,
                                     line, pos)
            self.types[type_name] = new_type
            type = self.types[type_name]
        else:
            self.add_error(
                    ComponentTypeAlreadyDeclaredError(
                                self.parent.parent.current_file,
                                line, pos, "B-1", type))
        return type 

    def add_component_implementation(self, impl_name, category, fdir, line, pos):
        """Adds a component implementation to the subpackage and returns the
        corresponding component implementation object.
        
        @param    impl_name:
                    The name of the component implementation.
        @type     impl_name:
                    string
        @param    category:
                    The category of the component implementation.
        @type     category:
                    A constant defined in nominalbehaviour.Category.
        @param    fdir:
                    Determines whether the component implementation is marked as
                    FDIR.
        @type     fdir:
                    bool
        @param    line:
                    The line in which the component implementation was defined.
        @type     line:
                    int
        @param    pos:
                    The column position at which the component implementation
                    was defined.
        @type     pos:
                    int
        @return:  The corresponding component implementation object.
        @rtype:   A nominalbehaviour.ComponentImplementation object.
        """
        impl = self.parent.get_component_implementation(impl_name, False)
        
        if impl is None:
            new_impl = ComponentImplementation(self, impl_name, category, fdir,
                                               self.parent.parent.current_file,
                                               line, pos)
            self.implementations[impl_name] = new_impl
            impl = self.implementations[impl_name]
        else:
            self.add_error(
                    ComponentImplementationAlreadyDeclaredError(
                                self.parent.parent.current_file,
                                line, pos, "E-1", impl))            
        return impl

    def add_error_type(self, error_type_name, line, pos):
        """Adds an error model type to the subpackage and returns the
        corresponding error model type object.
        
        @param    error_type_name:
                    The name of the error model type.
        @type     error_type_name:
                    string
        @param    line:
                    The line in which the error model type was defined.
        @type     line:
                    int
        @param    pos:
                    The column position at which the error model type was
                    defined.
        @type     pos:
                    int
        @return:  The corresponding error model type object.
        @rtype:   An errorbehaviour.ErrorModelType object.
        """
        type = self.parent.get_error_type(error_type_name, False)
        
        if type is None:
            new_type = ErrorModelType(self, error_type_name, 
                                      self.parent.parent.current_file,
                                      line, pos)
            self.error_types[error_type_name] = new_type
            type = self.error_types[error_type_name]
        else:
            self.add_error(
                    ErrorTypeAlreadyDeclaredError(
                            self.parent.parent.current_file, line,
                            pos, "J-1", type))
        return type
            
    def add_error_implementation(self, error_impl_name, line, pos):
        """Adds an error model implementation to the subpackage and returns the
        corresponding error model implementation object.
        
        @param    error_impl_name:
                    The name of the error model implementation.
        @type     error_impl_name:
                    string
        @param    line:
                    The line in which the error model implementation was defined.
        @type     line:
                    int
        @param    pos:
                    The column position at which the error model implementation
                    was defined.
        @type     pos:
                    int
        @return:  The corresponding error model implementation object.
        @rtype:   A errorbehaviour.ErrorModelImplementation object.
        """
        impl = self.parent.get_error_implementation(error_impl_name, False)
        
        if impl is None:
            new_impl = ErrorModelImplementation(self, error_impl_name,
                                                self.parent.parent.current_file,
                                                line, pos)
            self.error_implementations[error_impl_name] = new_impl
            impl = self.error_implementations[error_impl_name]
        else:
            self.add_error(
                    ErrorImplementationAlreadyDeclaredError(
                            self.parent.parent.current_file, line,
                            pos, "K-1", impl))
        return impl

    def add_error(self, error):
        """Adds an error to the error collector object.
        
        @param    error:
                    The error to add.
        @type     error:
                    A subclass of errors.AbstractError.
        """
        self.parent.add_error(error)

    def is_empty(self):
        """Returns True if there are no types and implementations defined in
        the subpackage and False otherwise.
        """
        return not (self.types or self.implementations or self.error_types or
                    self.error_implementations)

    def get_component_type(self, type_name):
        """Retrieves the component type object with the given name.
        
        @param    type_name:
                    The name of the component type object to search.
        @type     type_name:
                    string
        @result:  The component type object with the given name.
        @rtype:   A nominalbehaviour.ComponentType object.
        """
        if type_name in self.types.keys():
            type = self.types[type_name]
            return type
        else:
            return None

    def get_component_implementation(self, impl_name):
        """Retrieves the component implementation object with the given name.
        
        @param    impl_name:
                    The name of the component implementation object to search.
        @type     impl_name:
                    string
        @result:  The component implementation object with the given name.
        @rtype:   A nominalbehaviour.ComponentImplementation object.
        """
        if impl_name in self.implementations.keys():
            impl = self.implementations[impl_name]
            return impl
        else:
            return None

    def get_error_type(self, type_name):
        """Retrieves the error model type object with the given name.
        
        @param    type_name:
                    The name of the error model type object to search.
        @type     type_name:
                    string
        @result:  The error model type object with the given name.
        @rtype:   A errorbehaviour.ErrorModelType object.
        """
        if type_name in self.error_types.keys():
            type = self.error_types[type_name]
            return type
        else:
            return None
        
    def get_error_implementation(self, impl_name):
        """Retrieves the error model implementation object with the given name.
        
        @param    impl_name:
                    The name of the error model implementation object to search.
        @type     impl_name:
                    string
        @result:  The error model implementation object with the given name.
        @rtype:   A errorbehaviour.ErrorModelImplementation object.
        """
        if impl_name in self.error_implementations.keys():
            impl = self.error_implementations[impl_name]
            return impl
        else:
            return None

    def get_total_number_of_transitions(self):
        """Returns the total number of transitions contained in all components
        contained in the subpackage.
        
        @return:    The total number of transitions.
        @rtype:     int
        """
        result = 0
        for impl in self.implementations.values():
            result += impl.get_total_number_of_transitions()
        return result

    def get_total_number_of_modes(self):
        """Returns the total number of modes contained in all components
        contained in the subpackage.
        
        @return:    The total number of modes.
        @rtype:     int
        """
        result = 0
        for impl in self.implementations.values():
            result += impl.get_total_number_of_modes()
        return result
    
    def get_total_number_of_component_types(self):
        """Returns the total number of component types in the subpackage.
        
        @return:    The total number of component types.
        @rtype:     int
        """
        return len(self.types)
        
    def get_total_number_of_component_implementations(self):
        """Returns the total number of component implementations in the
        subpackage.
        
        @return:    The total number of component implementations.
        @rtype:     int
        """
        return len(self.implementations)

    def get_total_number_of_nondata_component_instances(self):
        """Returns the total number of non-data component instances contained in
        all components contained in the subpackage.
        
        @return:    The total number of non-data component instances.
        @rtype:     int
        """
        result = 0
        for impl in self.implementations.values():
            result += impl.get_total_number_of_nondata_component_instances()
        return result

    def get_total_number_of_data_component_instances(self):
        """Returns the total number of data component instances contained in all
        components contained in the subpackage.
        
        @return:    The total number of data component instances.
        @rtype:     int
        """
        result = 0
        for impl in self.implementations.values():
            result += impl.get_total_number_of_data_component_instances()
        return result
    
    def get_total_number_of_outgoing_data_ports(self):
        """Returns the total number of outgoing data ports contained in all
        components contained in the subpackage.
        
        @return:    The total number of outgoing data ports.
        @rtype:     int
        """
        result = 0
        for type in self.types.values():
            result += type.get_total_number_of_outgoing_data_ports()
        return result

    def get_total_number_of_incoming_data_ports(self):
        """Returns the total number of incoming data ports contained in all
        components contained in the subpackage.
        
        @return:    The total number of incoming data ports.
        @rtype:     int
        """
        result = 0
        for type in self.types.values():
            result += type.get_total_number_of_incoming_data_ports()
        return result

    def get_total_number_of_incoming_event_ports(self):
        """Returns the total number of incoming event ports contained in all
        components contained in the subpackage.
        
        @return:    The total number of incoming event ports.
        @rtype:     int
        """
        result = 0
        for type in self.types.values():
            result += type.get_total_number_of_incoming_event_ports()
        return result

    def get_total_number_of_outgoing_event_ports(self):
        """Returns the total number of outgoing event ports contained in all
        components contained in the subpackage.
        
        @return:    The total number of outgoing event ports.
        @rtype:     int
        """
        result = 0
        for type in self.types.values():
            result += type.get_total_number_of_outgoing_event_ports()
        return result
    
    def get_total_number_of_data_port_connections(self):
        """Returns the total number of data port connections contained in all
        components contained in the subpackage.
        
        @return:    The total number of data port connections.
        @rtype:     int
        """
        result = 0
        for impl in self.implementations.values():
            result += impl.get_total_number_of_data_port_connections()
        return result

    def get_total_number_of_flows(self):
        """Returns the total number of flows contained in all components
        contained in the subpackage.
        
        @return:    The total number of flows.
        @rtype:     int
        """
        result = 0
        for impl in self.implementations.values():
            result += impl.get_total_number_of_flows()
        return result

    def get_total_number_of_event_port_connections(self):
        """Returns the total number of event port connections contained in all
        components contained in the subpackage.
        
        @return:    The total number of event port connections.
        @rtype:     int
        """
        result = 0
        for impl in self.implementations.values():
            result += impl.get_total_number_of_event_port_connections()
        return result
