# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


import sys
from antlr3 import *
from antlr3.compat import set, frozenset

from antlr3.tree import *
from slim_compiler.src.generated.SlimParser import SlimParser

from slim_compiler.src.errorManagement.errors import *
from slim_compiler.src.errorManagement.exceptions import *
from slim_compiler.src import globals

class DerivedParser(SlimParser):
     
    def __init__(self, compile_information, input, state=None):
        SlimParser.__init__(self, input, state)
        self.compile_information = compile_information
        self.parsing_error = False
        self.allow_names = False

    def reportError(self, error):
        #SlimParser.reportError(self, error)
        if isinstance(error, UnwantedTokenException):
            tokenName = "<unknown>"
            if error.expecting == EOF:
                token_name = "EOF"

            else:
                token_name = self.tokenNames[error.expecting]
            
            self.add_error(UnwantedTokenError(
                                        self.compile_information.current_file,
                                        error.line,
                                        error.charPositionInLine + 1,
                                        self._get_cur_rule(),
                        self.getTokenErrorDisplay(error.getUnexpectedToken()),
                        token_name))
            
        elif isinstance(error, MissingTokenException):
            token_name = "<unknown>"
            if error.expecting == EOF:
                token_name = "EOF"
            else:
                token_name = self.tokenNames[error.expecting]
                
            self.add_error(MissingTokenError(
                                        self.compile_information.current_file,
                                        error.line,
                                        error.charPositionInLine + 1,
                                        self._get_cur_rule(),
                                        self.getTokenErrorDisplay(error.token),
                                        token_name))
        
        elif isinstance(error, MismatchedTokenException):

            token_name = "<unknown>"
            if error.expecting == EOF:
                token_name = "EOF"
            else:
                token_name = self.tokenNames[error.expecting]
                
            self.add_error(MismatchedTokenError(
                                        self.compile_information.current_file,
                                        error.line,
                                        error.charPositionInLine + 1,
                                        self._get_cur_rule(),
                                        self.getTokenErrorDisplay(error.token),
                                        token_name))        

        elif isinstance(error, NoViableAltException):
            self.add_error(NoViableAltError(
                                        self.compile_information.current_file,
                                        error.line,
                                        error.charPositionInLine + 1,
                                        self._get_cur_rule(),
                                        self.getTokenErrorDisplay(error.token)))
            
        elif isinstance(error, EarlyExitException):
            self.add_error(EarlyExitError(
                                        self.compile_information.current_file,
                                        error.line,
                                        error.charPositionInLine + 1,
                                        self._get_cur_rule(),
                                        self.getTokenErrorDisplay(error.token)))
            
        elif isinstance(error, MismatchedSetException):
            self.add_error(MismatchedSetError(
                                        self.compile_information.current_file,
                                        error.line,
                                        error.charPositionInLine + 1,
                                        self._get_cur_rule(),
                                        self.getTokenErrorDisplay(error.token),
                                        repr(error.expecting)))
            
        elif isinstance(error, MismatchedNotSetException):
            self.add_error(MismatchedNotSetError(
                                        self.compile_information.current_file,
                                        error.line,
                                        error.charPositionInLine + 1,
                                        self._get_cur_rule(),
                                        self.getTokenErrorDisplay(error.token),
                                        repr(error.expecting)))
            
        elif isinstance(error, FailedPredicateException):
            self.add_error(FailedPredicateError(
                                        self.compile_information.current_file,
                                        error.line,
                                        error.charPositionInLine + 1,
                                        error.ruleName,
                                        error.predicateText))

        elif isinstance(error, UnsupportedFeatureException):
            if (globals.PARSER_RAISE_UNSUPPORTED_FEATURE_EXCEPTIONS):
                self.add_error(UnsupportedFeatureError(
                                            self.compile_information.current_file,
                                            error.line,
                                            error.charPositionInLine + 1,
                                            error.feature))

        elif isinstance(error, SLIMParserException):
            self.add_error(SLIMParserError(
                                        self.compile_information.current_file,
                                        error.line,
                                        error.charPositionInLine + 1,
                                        self._get_cur_rule(),
                                        error.message))

        elif isinstance(error, MismatchedTreeNodeException):
            assert False
        elif isinstance(error, RecognitionException):
            assert False
        else:
            assert False

        self.parsing_error = True

    def specification(self, ):
        return SlimParser.specification(self) 
        
    def has_parsing_errors(self):
        return self.parsing_error
    
    def add_error(self, error):
        self.compile_information.add_error(error)

    def _get_cur_rule(self):
        if (len(self.rules) > 0):
            return self.rules[-1]
        else:
            return "<unknown>"
