# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================
from slim_compiler.src.dataStructure.nominalbehaviour import Clock,\
    TIME_DELAY_CLK_NAME, DataSubcomponent, Continuous
from slim_compiler.src.dataStructure.expressions import NodeType, \
    TimeUnitType, NumberTreeNode
import utils
from slim_compiler.src.errorManagement.errors import ErrorStateUrgentInTypeError,\
    ErrorStateInvariantInTypeError, ErrorStateAlreadyDeclaredInTypeError,\
    ErrorPropagationAlreadyDeclaredInImplementationError,\
    ErrorPropagationAlreadyDeclaredInTypeError,\
    ErrorStateAlreadyDeclaredInImplementationError,\
    ErrorClockAlreadyDeclaredError, ErrorEventAlreadyDeclaredError,\
    ErrorEventAlreadyDeclaredInTypeError,\
    ErrorTransitionIsProbabilisticButHasGuardError,\
    ReactivationTransitionHasGuardError, ReactivationTransitionHasTimeDelayError,\
    GuardRefersToUnknownClockError, EffectRefersToUnknownClockError,\
    StateInvariantRefersToUnknownClockError

"""This module provides classes for storing the error behaviour of the
components in a SLIM specification.
"""

from located_info import LocatedInformation

class ErrorModelType(LocatedInformation):
    """This class is the data container for an error model type in a SLIM
    specification.
    """
    
    def __init__(self, subpackage, name, file, line, pos):
        """Initializes an error model type object.
        
        @param    subpackage:
                    The subpackage in which the error model type is located.
        @type     subpackage:
                    compile.Subpackage
        @param    name:
                    The name of the error model type.
        @type     name:
                    string
        @param    file:
                    The file in which the error model type was defined.
        @type     file:
                    string
        @param    line:
                    The line in which the error model type was defined.
        @type     line:
                    int
        @param    pos:
                    The column position at which the error model type was
                    defined.
        @type     pos:
                    int                    
        """
        LocatedInformation.__init__(self, subpackage, "error model", file, line,
                                    pos)
        self.subpackage = subpackage
        self.name = name
        self.states             = {}
        self.error_propagations = {}

    def get_ref_identifier(self):
        """Retrieves an identifier that is used as a reference point for errors."""
        #return self.parent.get_ref_identifier() + "::" + self.name
        return self.name

    def get_starting_state_name(self):
        """Retrieves the name of the starting state.
        
        @return:    The name of the starting state.
        @rtype:     string
        """
        starting_state = self.get_starting_state()
        if starting_state is not None:
            return starting_state.name
        
    def add_error_state(self, name, invariant, urgent, line, pos):
        """Adds an error state to the error model type.
        
        @param    name:
                    The name of the state.
        @type     name:
                    string
        @param    invariant:
                    The invariant of the mode.
        @type     invariant:
                    An object of a subclass of expressions.ExpressionTreeNode.
        @param    urgent:
                    The upper bound for the maximal sojourn time in this state.
        @type     urgent:
                    An object of a subclass of expressions.NumberTreeNode.
        @param    line:
                    The line in which the state was defined.
        @type     line:
                    int
        @param    pos:
                    The column position at which the state was defined.
        @type     pos:
                    int                    
        """
        if urgent:
            self.add_error(ErrorStateUrgentInTypeError(self.file, line, pos, "J-5", name))
        self._add_state(name, ErrorStateType.ERROR, invariant, line, pos)
    
    def add_activation_state(self, name, invariant, urgent, line, pos):
        """Adds an activation state to the error model type.
        
        @param    name:
                    The name of the state.
        @type     name:
                    string
        @param    invariant:
                    The invariant of the mode.
        @type     invariant:
                    An object of a subclass of expressions.ExpressionTreeNode.
        @param    urgent:
                    The upper bound for the maximal sojourn time in this state.
        @type     urgent:
                    An object of a subclass of expressions.NumberTreeNode.
        @param    line:
                    The line in which the state was defined.
        @type     line:
                    int
        @param    pos:
                    The column position at which the state was defined.
        @type     pos:
                    int                    
        """
        if urgent:
            self.add_error(ErrorStateUrgentInTypeError(self.file, line, pos, "J-5", name))
        self._add_state(name, ErrorStateType.ACTIVATION, invariant, line, pos)
    
    def add_initial_state(self, name, invariant, urgent, line, pos):
        """Adds a state to the error model type.
        
        @param    name:
                    The name of the state.
        @type     name:
                    string
        @param    invariant:
                    The invariant of the mode.
        @type     invariant:
                    An object of a subclass of expressions.ExpressionTreeNode.
        @param    urgent:
                    The upper bound for the maximal sojourn time in this state.
        @type     urgent:
                    An object of a subclass of expressions.NumberTreeNode.
        @param    line:
                    The line in which the state was defined.
        @type     line:
                    int
        @param    pos:
                    The column position at which the state was defined.
        @type     pos:
                    int                    
        """
        if urgent:
            self.add_error(ErrorStateUrgentInTypeError(self.file, line, pos, "J-5", name))
        self._add_state(name, ErrorStateType.INITIAL, invariant, line, pos)
        
    def _add_state(self, name, type, invariant, line, pos):
        """Adds a state to the error model type.
        
        @param    name:
                    The name of the state.
        @type     name:
                    string
        @param    type:
                    The type of the state.
        @type     type:
                    A constant defined in errorbehaviour.ErrorStateType.
        @param    line:
                    The line in which the state was defined.
        @type     line:
                    int
        @param    pos:
                    The column position at which the state was defined.
        @type     pos:
                    int                    
        """
        
        if invariant != None:
            self.add_error(
                     ErrorStateInvariantInTypeError(self.file, line, pos,
                                                    "K-19", name))
        
        if name in self.states:
            state = self.states[name]
            self.add_error(ErrorStateAlreadyDeclaredInTypeError(self.file,
                                                                line, pos,
                                                                "J-2", state,
                                                                self))
        else:
            self.states[name] = ErrorState(self, name, type, invariant, None, self.file, line,
                                           pos)

    def add_in_propagation(self, name, line, pos):
        """Adds an incoming error propagation to the error model type.
        
        @param    name:
                    The name of the propagation.
        @type     name:
                    string
        @param    line:
                    The line in which the propagation was defined.
        @type     line:
                    int
        @param    pos:
                    The column position at which the propagation was defined.
        @type     pos:
                    int
        """
        self._add_propagation(name, ErrorPropagationType.IN, line, pos)
    
    def add_out_propagation(self, name, line, pos):
        """Adds an outgoing error propagation to the error model type.
        
        @param    name:
                    The name of the propagation.
        @type     name:
                    string
        @param    line:
                    The line in which the propagation was defined.
        @type     line:
                    int
        @param    pos:
                    The column position at which the propagation was defined.
        @type     pos:
                    int
        """
        self._add_propagation(name, ErrorPropagationType.OUT, line, pos)
    
    def _add_propagation(self, name, type, line, pos):
        """Adds an error propagation to the error model type.
        
        @param    name:
                    The name of the propagation.
        @type     name:
                    string
        @param    type:
                    The type of the propagation.
        @type     type:
                    A constant defined in ErrorPropagationType.
        @param    line:
                    The line in which the propagation was defined.
        @type     line:
                    int
        @param    pos:
                    The column position at which the propagation was defined.
        @type     pos:
                    int
        """
        if name in self.error_propagations:
            prop = self.error_propagations[name]
            self.add_error(ErrorPropagationAlreadyDeclaredInTypeError(self.file,
                                                                      line, pos,
                                                                      "J-2",
                                                                      self,
                                                                      prop))
        else:
            self.error_propagations[name] = ErrorPropagation(self, name, type,
                                                             self.file, line,
                                                             pos)
            
        #TODO: CONTINUE HERE
        implementations = self.get_referenced_error_implementations()
        
        for impl in implementations:
            if name in impl.error_events:
                event = impl.error_events[name]
                self.add_error(ErrorPropagationAlreadyDeclaredInImplementationError(self.file,line,pos,name, impl.name,event))     
        
    def get_referenced_error_implementations(self):
        """Returns set of error implementation objects the error model type
        refers to, if such exist. Result may be an empty set.
        
        @result:    Set of error implementation objects the error model type
                    refers to, if such exist, Result may be an empty set..
        @rtype:     A set(errorbehaviour.ErrorModelImplementation) object.
        """
        implementations = self.parent.error_implementations
        implementations = filter(lambda x: x.name.startswith(self.name+'.'), implementations.values())
        
        return set(implementations)
    
    def add_error(self, error):
        """Adds an error to the error collector object.
        
        @param    error:
                    The error to add.
        @type     error:
                    A subclass of errors.AbstractError.
        """
        self.parent.add_error(error)

    def get_starting_state(self):
        """Retrieves the starting state of the error model type.
        
        @return:    The starting state of the error model type.
        @rtype:     An ErrorState object.
        """
        starting_states = self.get_all_starting_states()
        if len(starting_states) > 0:
            return starting_states.pop()
        else:
            return None

    starting_state = property(get_starting_state)

    def get_all_starting_states(self):
        """Returns a set of all starting states of the error model type.
        
        @return:    A set of all starting states of the error model type.
        @rtype:     A set of ErrorStates objects.
        """
        result = set()
        for state in self.states.values():
            if state.type == ErrorStateType.INITIAL or \
               state.type == ErrorStateType.ACTIVATION:
                result.add(state)
                
        return result
    
class ErrorModelImplementation(LocatedInformation):
    """This class is the data container for an error model implementation in a
    SLIM specification.
    """
    
    nextErrorTransitionID = 0

    def __init__(self, subpackage, name, file, line, pos):
        """Initializes an error model implementation object.
        
        @param    subpackage:
                    The subpackage in which the error model implementation is
                    located.
        @type     subpackage:
                    compile.Subpackage
        @param    name:
                    The name of the error model implementation.
        @type     name:
                    string
        @param    file:
                    The file in which the error model implementation was
                    defined.
        @type     file:
                    string
        @param    line:
                    The line in which the error model implementation was
                    defined.
        @type     line:
                    int
        @param    pos:
                    The column position at which the error model implementation
                    was defined.
        @type     pos:
                    int                    
        """
        LocatedInformation.__init__(self, subpackage,
                                    "error model implementation", file, line,
                                    pos)
        self.name = name
        self.category = "error model implementation"
        self.subpackage = subpackage
        self.states            = {}
        # Note: In order to support stopwatches, clocks may also contain
        # continuous variables that are automatically generated
        # (user may not add these)
        self.clocks            = {}
        self.error_events      = {}
        self.error_transitions = []
        self.reactivation_transitions = []

    @classmethod
    def reset(cls):
        ErrorModelImplementation.nextErrorTransitionID = 0

    def get_ref_identifier(self):
        """Retrieves an identifier that is used as a reference point for errors."""
        #return self.parent.get_ref_identifier() + "::" + self.name
        return self.name

    def get_id(self):
        """
        Returns a unique identifier for this error model implementation
        represented as a tuple. Right now, the only entry in this tuple is the
        local name of the error model implementation.
        
        Note: When packages are supported, the package name also has to be
              considered.
        
        @return:    A unique identifier for this error model implementation.
        @rtype:     tuple
        """
        return (self.name,)

    def equals(self, ci):
        """
        Returns true iff the current error model implementation is equal to
        another given error model implementation.
        
        @param    ci:    The error model implementation object that is to be
                         compared to the current one.
        @type     ci:    ErrorModelImplementation
        @return:  True, if and only if the error model implementations are equal.
        @rtype:   boolean
        """
        if not isinstance(ci, ErrorModelImplementation): return False
        else: return (self.get_id() == ci.get_id())
    
    def get_all_starting_states(self):
        """Returns a set of all starting states of the error model implementation.
        
        @return:    A set of all starting states of the error model implementation.
        @rtype:     A set of ErrorStates objects.
        """
        result = set()
        for state in self.states.values():
            if state.type == ErrorStateType.INITIAL or \
               state.type == ErrorStateType.ACTIVATION:
                result.add(state)
                
        return result
    
    def create_time_delay_clock(self):
        """Adds a default clock to measure the time spent in the current state
        to the implementation. Doesn't do anything if the clock has already
        been added.
        
        @result:  The default clock object
        @rtype:   nominalbehaviour.DataSubcomponent
        """
        if not TIME_DELAY_CLK_NAME in self.clocks:
            # Issue 591: Use continuous variables for time delay clocks
            # Inserting directly, as opposed to self.add_clock, since there
            # is no proper interface for continuous variables
            self.clocks[TIME_DELAY_CLK_NAME] = DataSubcomponent(self, TIME_DELAY_CLK_NAME,
                    Continuous(self.file, 0, 0), [],
                    NumberTreeNode.createZero(self.file, 0, 0), self.file, 0, 0)

        time_delay_clock = self.clocks[TIME_DELAY_CLK_NAME]
        return time_delay_clock

    def add_error_state(self, name, invariant, urgent, line, pos):
        """Adds an error state to the error model implementation.
        
        @param    name:
                    The name of the state.
        @type     name:
                    string
        @param    invariant:
                    The invariant of the mode.
        @type     invariant:
                    An object of a subclass of expressions.ExpressionTreeNode.
        @param    urgent:
                    The upper bound for the maximal sojourn time in this state.
        @type     urgent:
                    An object of a subclass of expressions.NumberTreeNode.
        @param    line:
                    The line in which the state was defined.
        @type     line:
                    int
        @param    pos:
                    The column position at which the state was defined.
        @type     pos:
                    int                    
        """
        self._add_state(name, ErrorStateType.ERROR, invariant, urgent, line, pos)

    def add_activation_state(self, name, invariant, urgent, line, pos):
        """Adds an activation state to the error model implementation.
        
        @param    name:
                    The name of the state.
        @type     name:
                    string
        @param    invariant:
                    The invariant of the mode.
        @type     invariant:
                    An object of a subclass of expressions.ExpressionTreeNode.
        @param    urgent:
                    The upper bound for the maximal sojourn time in this state.
        @type     urgent:
                    An object of a subclass of expressions.NumberTreeNode.
        @param    line:
                    The line in which the state was defined.
        @type     line:
                    int
        @param    pos:
                    The column position at which the state was defined.
        @type     pos:
                    int                    
        """
        self._add_state(name, ErrorStateType.ACTIVATION, invariant, urgent, line, pos)
    
    def add_initial_state(self, name, invariant, urgent, line, pos):
        """Adds a state to the error model implementation.
        
        @param    name:
                    The name of the state.
        @type     name:
                    string
        @param    invariant:
                    The invariant of the mode.
        @type     invariant:
                    An object of a subclass of expressions.ExpressionTreeNode.
        @param    urgent:
                    The upper bound for the maximal sojourn time in this state.
        @type     urgent:
                    An object of a subclass of expressions.NumberTreeNode.
        @param    line:
                    The line in which the state was defined.
        @type     line:
                    int
        @param    pos:
                    The column position at which the state was defined.
        @type     pos:
                    int                    
        """
        self._add_state(name, ErrorStateType.INITIAL, invariant, urgent, line, pos)
        
    def _add_state(self, name, type, invariant, urgent, line, pos):
        """Adds a state to the error model implementation.
        
        @param    name:
                    The name of the state.
        @type     name:
                    string
        @param    type:
                    The type of the state.
        @type     type:
                    A constant defined in errorbehaviour.ErrorStateType.
        @param    line:
                    The line in which the state was defined.
        @type     line:
                    int
        @param    pos:
                    The column position at which the state was defined.
        @type     pos:
                    int                    
        """
        if name in self.states:
            state = self.states[name]
            self.add_error(
                    ErrorStateAlreadyDeclaredInImplementationError(self.file,
                                                                   line, pos,
                                                                   "K-14",
                                                                   self, state))
        else:
            self.states[name] = ErrorState(self, name, type, invariant, urgent, self.file, line,
                                           pos)

    def add_clock(self, clock_name, data_type, line, pos):
        """Adds a data subcomponent to the component implementation.
        
        @param    clock_name:
                    The name of the clock.
        @type     clock_name:
                    string
        @param    data_type:
                    The data type of the clock.
        @type     data_type:
                    An object of a subclass of
                    nominalbehaviour.AbstractDataType.
        @param    line:
                    The line in which the clock was defined.
        @type     line:
                    int
        @param    pos:
                    The column position at which the clock was defined.
        @type     pos:
                    int
        """
        new_subcomponent = DataSubcomponent(self, clock_name, data_type,
                                     [], None,
                                     self.file, line, pos)

        if not clock_name in self.clocks:
            self.clocks[clock_name] = new_subcomponent
        else:
            self.add_error(
                     ErrorClockAlreadyDeclaredError(self.file, line, pos,
                                                    "K-17", self.clocks[clock_name]))
    
    def add_event(self, name, distribution, rate, time_unit, line, pos):
        """Adds an error event to the error model implementation.

        @param    name:
                    The name of the event.
        @type     name:
                    string
        @param    distribution:
                    The occurrence distribution of the event.
        @type     distribution:
                    A constant defined in DistributionType.
        @param    rate:
                    The rate of the event.
        @type     rate:
                    int
        @param    time_unit:
                    The time unit of the distribution.
        @type     time_unit:
                    A constant defined in expressions.TimeUnitType.
        @param    line:
                    The line in which the event was defined.
        @type     line:
                    int
        @param    pos:
                    The column position at which the event was defined.
        @type     pos:
                    int                    
        """
        if name in self.error_events:
            event = self.error_events[name]
            self.add_error(ErrorEventAlreadyDeclaredError(self.file, line, pos,
                                                          "K-4", self, event))
        else:
            self.error_events[name] = ErrorEvent(self, name, distribution, rate,
                                                 time_unit, self.file, line,
                                                 pos)

    def add_transition(self, source, trigger, target, guard, time_delay, effects, line, pos):
        """Adds an error transition to the error model implementation.

        @param    source:
                    The source state of the transition.
        @type     source:
                    string
        @param    trigger:
                    The trigger of the transition.
        @type     trigger:
                    string
        @param    target:
                    The target state of the transition.
        @type     target:
                    string
        @param    guard:
                    The guard of the transition.
        @type     guard:
                    An object of a subclass of expression.ExpressionTreeNode.
        @param    time_delay:
                    The lower and upper bound to the time delay which will be
                    incorporated into the guard
        @type     time_delay:
                    A 2-tuple of a subclass of expression.NumberTreeNode.
        @param    effects:
                    A list of effects of the transition.
        @type     effects:
                    A list of objects of a subclass of
                    expression.ExpressionTreeNode.
        @param    line:
                    The line in which the transition was defined.
        @type     line:
                    int
        @param    pos:
                    The column position at which the transition was defined.
        @type     pos:
                    int                    
        """
        error_transition = ErrorTransition(self, ErrorModelImplementation.nextErrorTransitionID,
                                           source, False, trigger, target, guard, effects,
                                           time_delay,
                                           self.file, line, pos)
        ErrorModelImplementation.nextErrorTransitionID += 1

        self.error_transitions.append(error_transition)

    def add_reactivation_transition(self, source, target, guard, time_delay, effects, line, pos):
        """Adds an reactivation transition to the error model implementation.

        @param    source:
                    The source state of the transition.
        @type     source:
                    string
        @param    target:
                    The target state of the transition.
        @type     target:
                    string
        @param    guard:
                    The guard of the transition.
        @type     guard:
                    An object of a subclass of expression.ExpressionTreeNode.
        @param    time_delay:
                    The lower and upper bound to the time delay which will be
                    incorporated into the guard
        @type     time_delay:
                    A 2-tuple of a subclass of expression.NumberTreeNode.
        @param    effects:
                    A list of effects of the transition.
        @type     effects:
                    A list of objects of a subclass of
                    expression.ExpressionTreeNode.
        @param    line:
                    The line in which the transition was defined.
        @type     line:
                    int
        @param    pos:
                    The column position at which the transition was defined.
        @type     pos:
                    int                    
        """
        error_transition = ErrorTransition(self, ErrorModelImplementation.nextErrorTransitionID,
                                           source, True, None, target, guard, effects, time_delay,
                                           self.file, line, pos)
        ErrorModelImplementation.nextErrorTransitionID += 1

        if guard:
            self.add_error(ReactivationTransitionHasGuardError(self.file, line, pos, "K-31", error_transition))

        if time_delay:
            self.add_error(ReactivationTransitionHasTimeDelayError(self.file, line, pos, "K-32", error_transition))

        self.reactivation_transitions.append(error_transition)

    def add_error(self, error):
        """Adds an error to the error collector object.
        
        @param    error:
                    The error to add.
        @type     error:
                    A subclass of errors.AbstractError.
        """
        self.parent.add_error(error)

    def get_referenced_error_type(self):
        """Returns the error type object the error model implementation
        refers to, if it exists, None otherwise.
        
        @result:    The component type object the error model implementation
                    refers to, if it exists, None otherwise.
        @rtype:     A errorbehaviour.ErrorModelType object.
        """
        type = self.parent.get_error_type(self.name.split(".")[0])
        return type

    def get_transitions_leaving_state(self, source_state):
        """Retrieves all error transitions which emanate from the given state.
        
        @param    source_state:
                    The source state of the transitions.
        @type     source_state:
                    string
        @return:  All error transitions which emanate from the given state.
        @rtype:   A list of ErrorTransition objects.
        """
        result = []
        for trans in (self.error_transitions + self.reactivation_transitions):
            if trans.source_state == source_state:
                result.append(trans)
                
        return result

    def get_data_element_type_invariant(self, name, state, line, pos):
        """Returns the type of the data element found under the given name
        taking the rules of guards into consideration.
        
        Note that expressions of error implementations can only have clocks
        as data components.
        
        @param    name:
                    The name of the data element.
        @type     name:
                    string
        @param    mode:
                    The state to which errors should be associated.
        @type     mode
                    A ErrorState object.
        @param    line:
                    The line in which the guard is defined.
        @type     line:
                    int
        @param    pos:
                    The column position at which the guard is defined.
        @type     pos:
                    int
        @return:  The type of the data element found under the given name.
        @rtype:   A tuple consisting of a constant defined in
                  expressions.NodeType and additional information.
        """

        if name not in self.clocks:
            self.add_error(StateInvariantRefersToUnknownClockError(self.file, line, pos,
                                                          "K-18", name, state))
            return (NodeType.ERROR, None, None)
        else:
            clock = self.clocks[name]
            return (clock.data_type.get_node_type(), clock, None)

    def get_data_element_type_guard(self, name, trans, line, pos):
        """Returns the type of the data element found under the given name
        taking the rules of guards into consideration.
        
        Note that expressions of error implementations can only have clocks
        as data components.
        
        @param    name:
                    The name of the data element.
        @type     name:
                    string
        @param    trans:
                    The transition to which errors should be associated.
        @type     trans:
                    A ErrorTransition object.
        @param    line:
                    The line in which the guard is defined.
        @type     line:
                    int
        @param    pos:
                    The column position at which the guard is defined.
        @type     pos:
                    int
        @return:  The type of the data element found under the given name.
        @rtype:   A tuple consisting of a constant defined in
                  expressions.NodeType and additional information.
        """

        if name not in self.clocks:
            self.add_error(GuardRefersToUnknownClockError(self.file, line, pos,
                                                          "K-22", name, self, trans))
            return (NodeType.ERROR, None, None)
        else:
            clock = self.clocks[name]
            return (clock.data_type.get_node_type(), clock, None)

    def get_data_element_type_effect(self, name, trans, line, pos):
        """Returns the type of the data element found under the given name
        taking the rules of effects into consideration.
        
        @param    name:
                    The name of the data element.
        @type     name:
                    string
        @param    trans:
                    The transition to which errors should be associated.
        @type     trans:
                    A ErrorTransition object.
        @param    line:
                    The line in which the effect is defined.
        @type     line:
                    int
        @param    pos:
                    The column position at which the effect is defined.
        @type     pos:
                    int
        @return:  The type of the data element found under the given name.
        @rtype:   A tuple consisting of a constant defined in
                  expressions.NodeType and additional information.
        """
        if name not in self.clocks:
            self.add_error(EffectRefersToUnknownClockError(self.file, line, pos,
                                                          "K-26", name, self, trans))
            return (NodeType.ERROR, None, None)

        else:
            clock = self.clocks[name]
            return (clock.data_type.get_node_type(), clock, None)

class ErrorStateType(object):
    """Contains constants to represent the type of an error state.
    This class is used as an enum.
    """
    INITIAL    = 0        # an initial state
    ACTIVATION = 1        # an activation state
    ERROR      = 3        # an error state
    
    def str_to_err_state_type(cls, txt):
        """This method converts a given string into one of our pseudo enums
        """
        if ("initial" == txt):
            return cls.INITIAL
        elif("activation" == txt):
            return cls.ACTIVATION
        elif("error" == txt):
            return cls.ERROR
        else:
            assert False
            
    def err_state_type_to_str(cls, type):
        """This method converts a given state type to a textual string which is
        used for SLIM output
        """
        if cls.INITIAL == type:
            return "initial"
        elif cls.ACTIVATION == type:
            return "activation"
        #elif cls.NOMINAL == type:
         #   return "nominal"
        elif cls.ERROR == type:
            return "error"
        else:
            assert False
            
    str_to_err_state_type = classmethod(str_to_err_state_type)
    err_state_type_to_str = classmethod(err_state_type_to_str)

class ErrorState(LocatedInformation):
    """AThis class is the data container for an error state in a SLIM 
    specification.
    """
    
    def __init__(self, parent, name, type, invariant, urgent, file, line, pos):
        """Initializes an error state object.
        
        @param    parent:
                    The error model type or error model implementation object
                    this state belongs to.
        @type     parent:
                    An ErrorModelType or ErrorModelImplementation object.
        @param    name:
                    The name of the error state.
        @type     name:
                    string
        @param    type:
                    The type of the error state.
        @type     type:
                    A constant defined in ErrorStateType.
        @param    file:
                    The file in which the error state was defined.
        @type     file:
                    string
        @param    line:
                    The line in which the error state was defined.
        @type     line:
                    int
        @param    pos:
                    The column position at which the error state was
                    defined.
        @type     pos:
                    int                    
        """
        LocatedInformation.__init__(self, parent, "error state", file,
                                    line, pos)
        self.name = name
        self.type = type
        self.invariant = invariant
        self.urgent = urgent
        
    def get_ref_identifier(self):
        """Retrieves an identifier that is used as a reference point for errors."""
        return self.parent.get_ref_identifier() + "::state::" + self.name

class ErrorTransition(LocatedInformation):
    """This class is the data container for an error transition in a SLIM
    specification.
    """
    def __init__(self, error_impl, number, source_state, is_reactivation, trigger, target_state,
                 guard, effects, time_delay, file, line, pos):
        """Initializes an error transition object.
        
        @param    error_impl:
                    The error model implementation this transition belongs to.
        @type     error_impl:
                    An ErrorModelImplementation object.
        @param    source_state:
                    The source state of the transition.
        @type     source_state:
                    string
        @param    is_reactivation:
                    States whether this is a reactivation transition.
        @type     is_reactivation:
                    bool
        @param    trigger:
                    The trigger of the transition.
        @type     trigger:
                    string
        @param    target_state:
                    The target state of the transition.
        @type     target_state:
                    string        
        @param    file:
                    The file in which the transition was defined.
        @type     file:
                    string
        @param    line:
                    The line in which the transition was defined.
        @type     line:
                    int
        @param    pos:
                    The column position at which the transition was defined.
        @type     pos:
                    int                    
        """
        LocatedInformation.__init__(self, error_impl, "error state", file, line,
                                    pos)
        self.number = number
        self.source_state = source_state
        self.target_state = target_state
        self.is_reactivation = is_reactivation
        self.trigger      = trigger
        self.guard        = guard
        self.effects      = effects
        self.time_delay   = time_delay
        
    def get_ref_identifier(self):
        """Retrieves an identifier that is used as a reference point for errors."""
        return self.parent.get_ref_identifier() + "::trans::" + str(self.number)

class DistributionType(object):
    POISSON = 1
    
    def dist_type_to_str(cls, type):
        if type == cls.POISSON:
            return "poisson"
        else:
            assert False
            
    def str_to_dist_type(cls, text):
        if text == "poisson":
            return cls.POISSON
        else:
            assert False
    
    dist_type_to_str = classmethod(dist_type_to_str)
    str_to_dist_type = classmethod(str_to_dist_type)

class ErrorEvent(LocatedInformation):
    """This class is the data container for an error event in a SLIM
    specification.
    """
    
    def __init__(self, error_impl, name, distribution, parameter, time_unit,
                 file, line, pos):
        """Initializes an error event object.
        
        @param    error_impl:
                    The error model implementation this error event belongs to.
        @type     error_impl:
                    An ErrorModelImplementation object.
        @param    name:
                    The name of the error event.
        @type     name:
                    string
        @param    distribution:
                    The distribution of the error event.
        @type     type:
                    A constant defined in DistributionType.
        @param    parameter:
                    The parameter of the distribution.
        @type     parameter:
                    string
        @param    time_unit:
                    The time unit of the distribution.
        @type     time_unit:
                    A constant defined in TimeUnit.
        @param    file:
                    The file in which the error event was defined.
        @type     file:
                    string
        @param    line:
                    The line in which the error event was defined.
        @type     line:
                    int
        @param    pos:
                    The column position at which the error event was defined.
        @type     pos:
                    int                    
        """
        LocatedInformation.__init__(self, error_impl, "error event", file, line,
                                    pos)
        self.name = name
        self.distribution = distribution
        self.parameter = parameter
        self.time_unit = time_unit
        
    def get_ref_identifier(self):
        """Retrieves an identifier that is used as a reference point for errors."""
        return self.parent.get_ref_identifier() + "::event::" + self.name

class ErrorPropagationType(object):
    """Contains constants to represent the type of an error propagation.
    This class is used as an enum.
    """
    IN  = 0        # an incoming error propagation
    OUT = 1        # an outgoing error propagation
    
    def str_to_prop_type(cls, txt):
        """This method converts a given string into one of our pseudo enums
        """
        if "in" == txt:
            return cls.IN
        elif "out" == txt:
            return cls.OUT
        else:
            assert False
            
    def prop_type_to_str(cls, type):
        """This method converts a given propagation type to a textual string
        that is used for SLIM output
        """
        if cls.IN == type:
            return "in"
        elif cls.OUT == type:
            return "out"
        else:
            assert False
    
    str_to_prop_type = classmethod(str_to_prop_type)
    prop_type_to_str = classmethod(prop_type_to_str)    

class ErrorPropagation(LocatedInformation):
    """This class is the data container for an error propagation in a SLIM
    specification.
    """
    
    def __init__(self, error_type, name, type, file, line, pos):
        """Initializes an error propagation object.
        
        @param    error_type:
                    The error model type this propagation belongs to.
        @type     error_impl:
                    An ErrorModelType object.
        @param    name:
                    The name of the error propagation.
        @type     name:
                    string
        @param    type:
                    The type of the error propagation.
        @type     type:
                    A constant defined in ErrorPropagationType.
        @param    file:
                    The file in which the error propagation was defined.
        @type     file:
                    string
        @param    line:
                    The line in which the error propagation was defined.
        @type     line:
                    int
        @param    pos:
                    The column position at which the error propagation was
                    defined.
        @type     pos:
                    int                    
        """
        LocatedInformation.__init__(self, error_type, "error propagation", file,
                                    line, pos)
        self.name = name
        self.type = type
        
    def get_ref_identifier(self):
        """Retrieves an identifier that is used as a reference point for errors."""
        return self.parent.get_ref_identifier() + "::propagation::" + self.name
