# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================
from slim_compiler.src import dataStructure
from slim_compiler.src.dataStructure.located_info import LocatedInformation
from slim_compiler.src.errorManagement.errors import PortAlreadyDeclaredError,\
    InitializationValueTypeMismatchError, EnumValueIllegalError,\
    ComponentCategoryCannotDeclarePortsError, AnalogDataPortsNotSupportedError,\
    ImplementationCategoryDoesNotSupportSubcomponentsError,\
    SubcomponentDeclarationUsesUnsupportedBindingError,\
    SubcomponentAlreadyDeclaredError, ModeAlreadyDeclaredError,\
    ReactivationTransitionHasTimeDelayError, ReactivationTransitionHasGuardError,\
    GuardRefersToInactiveSubcomponentError,\
    GuardRefersToUnknownSubcomponentOrDataPortError,\
    DataSubcomponentSetAndNotActiveError,\
    EffectRefersToUnknownIdentifierLhsError,\
    EffectRefersToInactiveSubcomponentRhsError,\
    EffectRefersToUnknownIdentifierRhsError,\
    ModeInvariantRefersToIllegalDataTypeError,\
    ModeInvariantRefersToInactiveSubcomponentError,\
    ModeInvariantRefersToUnknownSubcomponentError,\
    PortFlowReferencesInactiveSubcomponentError,\
    FlowReferencesUnknownIncomingDataPortOfSubcompError,\
    FlowReferencesUnknownIncomingDataPortError,\
    FlowReferencesUnknownSubcomponentError,\
    FlowReferencesUnknownOutgoingDataPortError,\
    FlowReferencesUnknownOutgoingDataPortOfSubcompError
from slim_compiler.src.dataStructure.expressions import NodeType, NumberTreeNode,\
    IdentifierTreeNode, BooleanTreeNode, UnaryOperatorTreeNode,\
    BinaryOperatorTreeNode, TrajectoryEquationTreeNode,\
    TrajectoryExpressionTreeNode, CaseExpressionTreeNode, TimeUnitType

"""This module provides classes for storing the nominal behaviour of components
in a SLIM specification.
"""

DEFAULTINITMODE = "DefaultInitialMode"
TIME_DELAY_CLK_NAME = "_t"

class ComponentType(LocatedInformation):
    """This class is the data container for a component type in a SLIM
    specification.
    """
    
    def __init__(self, subpackage, name, category, fdir, file, line, pos):
        """Initializes a component type object.
        
        @param    subpackage:
                    The subpackage in which the component type is located.
        @type     subpackage:
                    compile.Subpackage
        @param    name:
                    The name of the component type.
        @type     name:
                    string
        @param    category:
                    The category of the component type.
        @type     category:
                    A constant defined in nominalbehaviour.Category.
        @param    fdir:
                    Determines whether the component type is marked as FDIR.
        @type     fdir:
                    bool
        @param    file:
                    The file in which the component type was defined.
        @type     file:
                    string
        @param    line:
                    The line in which the component type was defined.
        @type     line:
                    int
        @param    pos:
                    The column position at which the component type was defined.
        @type     pos:
                    int
        """
        LocatedInformation.__init__(self, subpackage, category, file, line, pos)
        self.name = name
        self.fdir = fdir
        self.category = category
        self.subpackage = subpackage
        self.in_data_ports = {}
        self.out_data_ports = {}
        self.in_event_ports = {}
        self.out_event_ports = {}
        
    def get_ref_identifier(self):
        """Retrieves an identifier that is used as a reference point for errors."""
        # return self.parent.get_ref_identifier() + "::" + self.name
        return self.name
        
    def add_port(self, file, port_name, in_port, event_port, nonblocking, data_type, default_value,
                 observable, alarm, line, pos):
        """Adds a port to the component type.
        
        @param    file:
                    The path to the currently processed file
        @type     file:
                    string
        @param    port_name:
                    The name of the port.
        @type     port_name:
                    string
        @param    in_port:
                    Determines whether the port is an incoming or an outgoing
                    port.
        @type     in_port:
                    bool
        @param    event_port:
                    Determines whether the port is an event or a data port.
        @type     event_port:
                    bool
        @param    nonblocking:
                    Determines whether the port is nonblocking.
        @type     nonblocking:
                    bool
        @param    data_type:
                    The data type of the port.
        @type     data_type:
                    An object of a subclass of
                    nominalbehaviour.AbstractDataType.
        @param    default_value:
                    The default value of the port.
        @type     default_value:
                    An object of a subclass of expressions.ExpressionTreeNode.
        @param    observable:
                    Determines whether the port is tagged as observable.
        @type     observable:
                    bool
        @param    alarm:
                    Determines whether the port is tagged as alarm
        @type     alarm:
                    bool
        @param    line:
                    The line in which the port was defined.
        @type     line:
                    int
        @param    pos:
                    The column position at which the port was defined.
        @type     pos:
                    int
        @result:  The port object that was added.
        @rtype:   An object of a subclass of nominalbehaviour.Port.
        """
        if not self.category_supports_ports():
            self.add_error(
                    ComponentCategoryCannotDeclarePortsError(
                            file, line, pos, "D-1", self,
                            Category.category_to_str(self.category)))

        if not event_port:
            if not self.port_type_is_supported(data_type):
                self.add_error(
                        AnalogDataPortsNotSupportedError(file, line, pos,
                        "D-3", Port(self, port_name, file, line, pos),
                        data_type.get_name()))
                
        (port, _, _) = self.get_port(port_name)    
        if not port:
            if event_port:
                if in_port:
                    self.in_event_ports[port_name] = InEventPort(self, 
                                                                 port_name,
                                                                 file,
                                                                 line, pos)
                    return self.in_event_ports[port_name]
                else:
                    self.out_event_ports[port_name] = OutEventPort(self,
                                                                   port_name,
                                                                   nonblocking,
                                                                   file,
                                                                   line, pos)
                    return self.out_event_ports[port_name]
            else:
                if in_port:
                    self.in_data_ports[port_name] = \
                            DataPort(self, port_name, data_type, default_value,
                                     file, line, pos)
                    return self.in_data_ports[port_name]
                else:
                    self.out_data_ports[port_name] = \
                            OutDataPort(self, port_name, data_type,
                                        default_value, observable, alarm, file,
                                        line, pos)
                    return self.out_data_ports[port_name]
        else:
            self.add_error(
                    PortAlreadyDeclaredError(
                            file, line, pos, "D-2", port))                    
    
    def add_error_event_port(self, port_name, distribution,
                             distribution_param, time_unit, line, pos):
        """Adds an error event port to the component type.
        See ErrorEventPort for details.
        
        The distribution must be specified as a string. It must be either
        "fixed" or poisson.
        If a time unit can be None (to use the default time unit) or one of
        the following strings:
        "msec", "sec", "min", "hour", "day"
        """ 
        # TODO: XXX This is just a quick hack to enable adding error event
        # ports when performing model extension. No checks are performed.
        # Note: This method will not be called by the parser.
        
        # The distribution and time unit are specified as strings (if they are
        # specified), so they must be converted to the corresponding
        # constants.
        # TODO: Should it really be allowed to specify no distribution?
        distribution_type = None
        if distribution:
            distribution_type = dataStructure.errorbehaviour.DistributionType.str_to_dist_type(
                                                                distribution)
        self.out_event_ports[port_name] = ErrorEventPort(self, port_name,
                                                         distribution_type,
                                                         distribution_param,
                                                         time_unit, self.file,
                                                         line, pos)
    
    def add_error(self, error):
        """Adds an error to the error collector object.
        
        @param    error:
                    The error to add.
        @type     error:
                    A subclass of errors.AbstractError.
        """
        self.parent.add_error(error)

    def category_supports_ports(self):
        """Returns true if the category of the component type supports the
        declaration of ports.
        
        @result:    True if the category supports the declaration of ports,
                    False otherwise.
        @rtype:     bool
        """
        if self.category == Category.MEMORY or self.category == Category.BUS:
            return False
        else:
            return True
        
    def port_type_is_supported(self, data_type):
        """Returns True if the data type is supported in an port declaration.
        
        @param    data_type:
                    The data type of the port.
        @type     data_type:
                    An object of a subclass of
                    nominalbehaviour.AbstractDataType.
        """
        if isinstance(data_type, (Clock, Continuous)):
            return False
        else:
            return True        
        
    def get_port(self, port_name):
        """Retrieves the port object with the given name.
        
        @param    port_name:
                    The name of the port.
        @type     port_name:
                    string
        @result:  A triple consisting of the port object with the given name and
                  information on the type of the port.
        @rtype:   A triple (port, in_port, event_port) where port is an object
                  of a subclass of nominalbehaviour.Port, in_port is True if
                  the port is an incoming port and False if it is an outgoing
                  port and event_port is True if the port is an event port and 
                  False if it is a data port. If the port does not exist
                  (None, None, None) is returned.
        """
        if port_name in self.in_event_ports:
            return (self.in_event_ports[port_name], True, True)
        elif port_name in self.out_event_ports:
            return (self.out_event_ports[port_name], False, True)
        elif port_name in self.in_data_ports:
            return (self.in_data_ports[port_name], True, False)
        elif port_name in self.out_data_ports:
            return (self.out_data_ports[port_name], False, False)
        else:
            return (None, None, None)
        
    def get_total_number_of_outgoing_data_ports(self):
        """Returns the total number of outgoing data ports in the component
        type.
        
        @return:    The total number of outgoing data ports.
        @rtype:     int
        """
        return len(self.out_data_ports)

    def get_total_number_of_incoming_data_ports(self):
        """Returns the total number of incoming data ports in the component
        type.
        
        @return:    The total number of incoming data ports.
        @rtype:     int
        """
        return len(self.in_data_ports)

    def get_total_number_of_incoming_event_ports(self):
        """Returns the total number of incoming event ports in the component
        type.
        
        @return:    The total number of incoming event ports.
        @rtype:     int
        """
        return len(self.in_event_ports)

    def get_total_number_of_outgoing_event_ports(self):
        """Returns the total number of outgoing event ports in the component
        type.
        
        @return:    The total number of outgoing event ports.
        @rtype:     int
        """
        return len(self.out_event_ports)

    def get_data_element_type_dataport_default_value(self, name, port, line, pos):
        # identifier has to be enum value belonging to the enum
        if not isinstance(port.data_type, Enum) :
            self.add_error(InitializationValueTypeMismatchError(self.file, line, pos, "D-4", port.data_type.get_name(), port))
            return (NodeType.ERROR, None, None)
        elif port.default_value.name not in port.data_type.values:
            self.add_error(EnumValueIllegalError(self.file, line, pos, "D-4", name, port))
            return (NodeType.ERROR, None, None)
        else:
            return (NodeType.ENUM_VALUE, port, [name])

class ComponentImplementation(LocatedInformation):
    """his class is the data container for a component implementation in a SLIM
    specification.
    """
    
    """Keep track of next IDs to create unique error reference identifiers"""
    nextPortConnectionID = 0
    nextFlowID = 0
    nextModeTransitionID = 0
    
    def __init__(self, subpackage, name, category, fdir, file, line, pos):
        """Initializes a component implementation object.
        
        @param    subpackage:
                    The subpackage in which the component implementation is
                    located.
        @type     subpackage:
                    compile.Subpackage
        @param    name:
                    The name of the component implementation.
        @type     name:
                    string
        @param    category:
                    The category of the component implementation.
        @type     category:
                    A constant defined in nominalbehaviour.Category.
        @param    fdir:
                    Determines whether the component implementation is marked as
                    FDIR.
        @type     fdir:
                    bool
        @param    file:
                    The file in which the component implementation was defined.
        @type     file:
                    string
        @param    line:
                    The line in which the component implementation was defined.
        @type     line:
                    int
        @param    pos:
                    The column position at which the component implementation
                    was defined.
        @type     pos:
                    int
        """
        LocatedInformation.__init__(self, subpackage, category, file, line, pos)
        self.name = name
        self.fdir = fdir
        self.subpackage = subpackage
        self.category = category
        self.non_data_subcomponents = {}
        self.data_subcomponents = {}
        self.data_port_connections = []
        self.event_port_connections = []
        self.flows = []
        self.modes = {}
        self.normal_mode_transitions = []
        self.reactivation_mode_transitions = []
        self.probability_mode_transitions = []
        # By default, the component implementation is not an error
        # subcomponent implementation.
        self.is_error_component_implementation = False
        return

    @classmethod
    def reset(cls):
        ComponentImplementation.nextPortConnectionID = 0
        ComponentImplementation.nextFlowID = 0
        ComponentImplementation.nextModeTransitionID = 0

    def get_ref_identifier(self):
        """Retrieves an identifier that is used as a reference point for errors."""
        #return self.parent.get_ref_identifier() + "::" + self.name
        return self.name

    def get_id(self):
        """Returns a unique identifier for this component implementation
        represented as a tuple.
                
        @return:  A unique identifier for this component implementation.
        @rtype:   tuple
        """
        return (self.parent.parent.name, self.name,)

    def equals(self, ci):
        """Returns True if the current component implementation is equal to
        another given component implementation, False otherwise.
        
        @param    ci:
                    The component implementation object that is to be compared
                    to the current one.
        @type     ci:
                    nominalbehaviour.ComponentImplementation
        @return:  True if the component implementations are equal, False
                  otherwise.
        @rtype:   bool
        """
        if not isinstance(ci, ComponentImplementation):
            return False
        else:
            return (self.get_id() == ci.get_id())

    def is_error_impl(self):
        """Returns True if the component implementation is an error
        implementation (i.e. contains at least one error mode).
        
        @result:  True if the component implementation is an error
                  implementation.
        @rtype:   bool
        """
        return self.is_error_component_implementation

    def add_non_data_subcomponent(self, subcomponent_name, category,
                                  implementation_ref, accesses, stored_in,
                                  running_on, in_modes, line, pos):
        """Adds a non-data subcomponent to the component implementation.
        
        @param    subcomponent_name:
                    The name of the subcomponent.
        @type     subcomponent_name:
                    string
        @param    category:
                    The category of the subcomponent.
        @type     category:
                    A constant defined in nominalbehaviour.Category.
        @param    implementation_ref:
                    The name of the component implementation that this
                    subcomponent declaration refers to.
        @type     implementation_ref:
                    string
        @param    accesses:
                    A list of other non-data subcomponents that this
                    subcomponent has an "accesses" binding with.
        @type     accesses:
                    A list of strings.
        @param    stored_in:
                    A list of other non-data subcomponents that this
                    subcomponent has a "stored in" binding with.
        @type     stored_in:
                    A list of strings.
        @param    running_on:
                    A list of other non-data subcomponents that this
                    subcomponent has a "running on" binding with.
        @type     running_on:
                    A list of strings.
        @param    in_modes:
                    A list of modes in which this subcomponent is active.
        @type     in_modes:
                    A list of strings.
        @param    line:
                    The line in which the subcomponent was declared.
        @type     line:
                    int
        @param    pos:
                    The column position at which the subcomponent was declared.
        @type     pos:
                    int
        """
        if in_modes is None: in_modes = []   
        
        new_subcomponent = NonDataSubcomponent(self, subcomponent_name, category,
                                        implementation_ref, accesses, stored_in,
                                        running_on, in_modes, self.file, line,
                                        pos)
        
        if not self.subcomponent_category_is_supported(category):
            self.add_error(
                    ImplementationCategoryDoesNotSupportSubcomponentsError(
                            self.file, line, pos, "F-6", self, 
                            Category.category_to_str(self.category),
                            Category.category_to_str(category)))
        if accesses:
            if not self.subcomponent_binding_is_supported(category,
                                                          Binding.ACCESSES):
                self.add_error(
                        SubcomponentDeclarationUsesUnsupportedBindingError(
                                self.file, line, pos, "F-9", new_subcomponent,
                                Category.category_to_str(category),
                                Binding.binding_to_str(Binding.ACCESSES)))
        if stored_in:
            if not self.subcomponent_binding_is_supported(category,
                                                          Binding.STOREDIN):
                self.add_error(
                        SubcomponentDeclarationUsesUnsupportedBindingError(
                                self.file, line, pos, "F-7", new_subcomponent,
                                Category.category_to_str(category), 
                                Binding.binding_to_str(Binding.STOREDIN)))
                
        if running_on:
            if not self.subcomponent_binding_is_supported(category,
                                                          Binding.RUNNINGON):
                self.add_error(
                        SubcomponentDeclarationUsesUnsupportedBindingError(
                                self.file, line, pos, "F-8", new_subcomponent,
                                Category.category_to_str(category),
                                Binding.binding_to_str(Binding.RUNNINGON)))

        subcomponent = self.get_subcomponent(subcomponent_name)

        if subcomponent is None:
            self.non_data_subcomponents[subcomponent_name] = new_subcomponent
        else:
            self.add_error(
                    SubcomponentAlreadyDeclaredError(self.file, line, pos,
                                                     "F-1", subcomponent))

    def add_data_subcomponent(self, subcomponent_name, data_type, default_value,
                              in_modes, line, pos):
        """Adds a data subcomponent to the component implementation.
        
        @param    subcomponent_name:
                    The name of the subcomponent.
        @type     subcomponent_name:
                    string
        @param    data_type:
                    The data type of the subcomponent.
        @type     data_type:
                    An object of a subclass of
                    nominalbehaviour.AbstractDataType.
        @param    default_value:
                    The default value of the subcomponent.
        @type     default_value:
                    An object of a subclass of expressions.ExpressionTreeNode.
        @param    in_modes:
                    A list of modes in which the subcomponent is active.
        @type     in_modes:
                    A list of strings.
        @param    line:
                    The line in which the subcomponent was defined.
        @type     line:
                    int
        @param    pos:
                    The column position at which the subcomponent was defined.
        @type     pos:
                    int
        """
        if in_modes is None: in_modes = []
        
        new_subcomponent = DataSubcomponent(self, subcomponent_name, data_type,
                                     in_modes, default_value,
                                     self.file, line, pos)
        
        if not self.subcomponent_category_is_supported(Category.DATA):
            self.add_error(
                    ImplementationCategoryDoesNotSupportSubcomponentsError(
                            self.file, line, pos, "F-6", self,
                            Category.category_to_str(self.category), "data"))
        
        subcomponent = self.get_subcomponent(subcomponent_name)
                
        if subcomponent is None:
            self.data_subcomponents[subcomponent_name] = new_subcomponent
        else:
            self.add_error(
                    SubcomponentAlreadyDeclaredError(self.file, line, pos,
                                                     "F-1", subcomponent))

    def add_data_port_connection(self, source_port, target_port, in_modes, line,
                                 pos):
        """Adds a data port connection to the component implementation.
        
        @param    source_port:
                    The source port of the connection.
        @type     source_port:
                    string
        @param    target_port:
                    The target port of the connection.
        @type     target_port:
                    string
        @param    in_modes:
                    A list of modes in which the connection is active.
        @type     in_modes:
                    A list of strings.
        @param    line:
                    The line in which the port connection was defined.
        @type     line:
                    int
        @param    pos:
                    The column position at which the port connection was
                    defined.
        @type     pos:
                    int
        """
        if in_modes is None: in_modes = []
        self.data_port_connections.append(
                PortConnection(self, ComponentImplementation.nextPortConnectionID, source_port,
                               target_port, in_modes, False, self.file, line, pos))
        ComponentImplementation.nextPortConnectionID += 1
    
    def add_event_port_connection(self, source_port, target_port, in_modes,
                                  passive, line, pos):
        """Adds an event port connection to the component implementation.
        
        @param    source_port:
                    The source port of the connection.
        @type     source_port:
                    string
        @param    target_port:
                    The target port of the connection.
        @type     target_port:
                    string
        @param    in_modes:
                    A list of modes in which the connection is active.
        @type     in_modes:
                    A list of strings.
        @param    passive:
                    A boolean flag indicating whether the connection is passive.
        @type     passive:
                    A boolean.
        @param    line:
                    The line in which the port connection was defined.
        @type     line:
                    int
        @param    pos:
                    The column position at which the port connection was
                    defined.
        @type     pos:
                    int
        """
        if in_modes is None: in_modes = []
        self.event_port_connections.append(
                PortConnection(self, ComponentImplementation.nextPortConnectionID, source_port,
                               target_port, in_modes, passive, self.file, line, pos))
        ComponentImplementation.nextPortConnectionID += 1

    def add_flow(self, expression, in_modes, line, pos):
        """Adds a flow to the component implementation.
        
        @param    expression:
                    The expression that defines the flow.
        @type     expression:
                    A expressions.BinaryOperatorTreeNode object.
        @param    in_modes:
                    A list of modes in which the flow is active.
        @type     in_modes:
                    A list of strings.
        @param    line:
                    The line in which the flow was defined.
        @type     line:
                    int
        @param    pos:
                    The column position at which the flow was defined.
        @type     pos:
                    int
        """
        self.flows.append(Flow(self, ComponentImplementation.nextFlowID, expression, in_modes, self.file, line, pos))
        ComponentImplementation.nextFlowID += 1

    def add_mode(self, name, invariant, urgent, line, pos):
        """Adds a nominal mode to the component implementation.
        
        @param    name:
                    The name of the mode.
        @type     name:
                    string
        @param    invariant:
                    The invariant of the mode.
        @type     invariant:
                    An object of a subclass of expressions.ExpressionTreeNode.
        @param    urgent:
                    The upper bound for the maximal sojourn time in this state.
        @type     urgent:
                    An object of a subclass of expressions.NumberTreeNode.
        @param    line:
                    The line in which the mode was defined.
        @type     line:
                    int
        @param    pos:
                    The column position at which the mode was defined.
        @type     pos:
                    int
        """
        self._addMode(name, invariant, ModeType.NOMINAL, urgent, line, pos)

    def create_time_delay_clock(self):
        """Adds a default clock to measure the time spent in the current mode
        to the implementation. Doesn't do anything if the clock has already
        been added.
        
        @result:  The default clock object
        @rtype:   nominalbehaviour.DataSubcomponent
        """
        if not TIME_DELAY_CLK_NAME in self.data_subcomponents:
            # Issue 591: Use continuous variables for time delay clocks
            self.data_subcomponents[TIME_DELAY_CLK_NAME] = DataSubcomponent(self, TIME_DELAY_CLK_NAME,
                                    Continuous(self.file, self.line, self.pos),
                                    [], NumberTreeNode.createZero(self.file, self.line, self.pos),
                                    self.file, self.line, self.pos)
        time_delay_clock = self.data_subcomponents[TIME_DELAY_CLK_NAME]
        return time_delay_clock

    def add_initial_mode(self, name, invariant, urgent, line, pos):
        """Adds an initial mode to the component implementation.
        
        @param    name:
                    The name of the mode.
        @type     name:
                    string
        @param    invariant:
                    The invariant of the mode.
        @type     invariant:
                    An object of a subclass of expressions.ExpressionTreeNode.
        @param    urgent:
                    The upper bound for the maximal sojourn time in this mode.
        @type     urgent:
                    An object of a subclass of expressions.NumberTreeNode.
        @param    line:
                    The line in which the mode was defined.
        @type     line:
                    int
        @param    pos:
                    The column position at which the mode was defined.
        @type     pos:
                    int
        """
        self._addMode(name, invariant, ModeType.INITIAL, urgent, line, pos)
    
    def add_activation_mode(self, name, invariant, urgent, line, pos):
        """Adds an activation mode to the component implementation.
        
        @param    name:
                    The name of the mode.
        @type     name:
                    string
        @param    invariant:
                    The invariant of the mode.
        @type     invariant:
                    An object of a subclass of expressions.ExpressionTreeNode.
        @param    urgent:
                    The upper bound for the maximal sojourn time in this mode.
        @type     urgent:
                    An object of a subclass of expressions.NumberTreeNode.
        @param    line:
                    The line in which the mode was defined.
        @type     line:
                    int
        @param    pos:
                    The column position at which the mode was defined.
        @type     pos:
                    int
        """
        self._addMode(name, invariant, ModeType.ACTIVATION, urgent, line, pos)

    def add_error_mode(self, name, invariant, urgent, line, pos):
        """Adds an error mode to the component implementation.
        
        @param    name:
                    The name of the mode.
        @type     name:
                    string
        @param    invariant:
                    The invariant of the mode.
        @type     invariant:
                    An object of a subclass of expressions.ExpressionTreeNode.
        @param    urgent:
                    The upper bound for the maximal sojourn time in this mode.
        @type     urgent:
                    An object of a subclass of expressions.NumberTreeNode.
        @param    line:
                    The line in which the mode was defined.
        @type     line:
                    int
        @param    pos:
                    The column position at which the mode was defined.
        @type     pos:
                    int
        """
        self._addMode(name, invariant, ModeType.ERROR, urgent, line, pos)

    def _addMode(self, name, invariant, type, urgent, line, pos):
        """Adds a mode of the given type to the component implementation.
        
        @param    name:
                    The name of the mode.
        @type     name:
                    string
        @param    invariant:
                    The invariant of the mode.
        @type     invariant:
                    An object of a subclass of expressions:ExpressionTreeNode.
        @param    type:
                    The type of the mode.
        @type     type:
                    A constant defined in nominalbehaviour.ModeType.
        @param    line:
                    The line in which the mode was defined.
        @type     line:
                    int
        @param    pos:
                    The column position at which the mode was defined.
        @type     pos:
                    int
        """
        
        if name not in self.modes:
            self.modes[name] = Mode(self, name, type, invariant, urgent, self.file,
                                    line, pos)
        else:
            mode = self.modes[name]
            if name != DEFAULTINITMODE:
                self.add_error(
                    ModeAlreadyDeclaredError(
                            self.parent.parent.parent.current_file,
                            line, pos, "I-2", mode))

    def add_normal_mode_transition(self, source_mode, target_mode, trigger,
                                   guard, time_delay, effects, line, pos):
        """Adds a normal (i.e. non-probability mode transition to the component
        implementation.
        
        @param    source_mode:
                    The source mode of the transition.
        @type     source_mode:
                    string
        @param    target_mode:
                    The target mode of the transition.
        @type     target_mode:
                    string
        @param    trigger:
                    The trigger of the transition.
        @type     trigger:
                    A list of strings.
        @param    guard:
                    The guard of the transition.
        @type     guard:
                    An object of a subclass of expression.ExpressionTreeNode.
        @param    time_delay:
                    The lower and upper bound to the time delay which will be
                    incorporated into the guard
        @type     time_delay:
                    A 2-tuple of a subclass of expression.NumberTreeNode.
        @param    effects:
                    A list of effects of the transition.
        @type     effects:
                    A list of objects of a subclass of
                    expression.ExpressionTreeNode.
        @param    line:
                    The line in which the transition was defined.
        @type     line:
                    int
        @param    pos:
                    The column position at which the transition was defined.
        @type     pos:
                    int                    
        """
        mode_transition=ModeTransition(self,
                               ComponentImplementation.nextModeTransitionID, 0,
                               source_mode, target_mode,
                               ModeTransitionLabel(trigger, guard, time_delay, effects),
                               self.file, line, pos)
        self.normal_mode_transitions.append(mode_transition)
        ComponentImplementation.nextModeTransitionID += 1

    def add_reactivation_mode_transition(self, source_mode, target_mode, guard, time_delay, effects, line, pos):
        """Adds a reactivaton transition
        
        @param    source_mode:
                    The source mode of the transition.
        @type     source_mode:
                    string
        @param    target_mode:
                    The target mode of the transition.
        @type     target_mode:
                    string
        @param    guard:
                    The guard of the transition.
        @type     guard:
                    An object of a subclass of expression.ExpressionTreeNode.
        @param    time_delay:
                    The lower and upper bound to the time delay which will be
                    incorporated into the guard
        @type     time_delay:
                    A 2-tuple of a subclass of expression.NumberTreeNode.
        @param    effects:
                    A list of effects of the transition.
        @type     effects:
                    A list of objects of a subclass of
                    expression.ExpressionTreeNode.
        @param    line:
                    The line in which the transition was defined.
        @type     line:
                    int
        @param    pos:
                    The column position at which the transition was defined.
        @type     pos:
                    int
        """
        mode_transition = ModeTransition(self,
                               ComponentImplementation.nextModeTransitionID, 0,
                               source_mode, target_mode,
                               ReactivationModeTransitionLabel(effects),
                               self.file, line, pos) 
        if guard:
            self.add_error(ReactivationTransitionHasGuardError(self.file, line, pos, "I-22", mode_transition))

        if time_delay:
            self.add_error(ReactivationTransitionHasTimeDelayError(self.file, line, pos, "I-23", mode_transition))

        self.reactivation_mode_transitions.append(mode_transition)
        ComponentImplementation.nextModeTransitionID += 1

    def add_probability_mode_transition(self, source_mode, target_mode,
                                        probability, line, pos):
        """Adds a probability mode transition to the component implementation.
        
        @param    source_mode:
                    The source mode of the transition.
        @type     source_mode:
                    string
        @param    target_mode:
                    The target mode of the transition
        @type     target_mode:
                    string
        @param    probability:
                    The probability of the transition.
        @type     probability:
                    An expressions.NumberTreeNode object.
        @param    line:
                    The line in which the transition was defined.
        @type     line:
                    int
        @param    pos:
                    The column position at which the transition was defined.
        @type     pos:
                    int                    
        """
        self.probability_mode_transitions.append(
                ModeTransition(self,
                               ComponentImplementation.nextModeTransitionID, 0,
                               source_mode, target_mode,
                               ProbabilityModeTransitionLabel(probability),
                               self.file, line, pos))
        ComponentImplementation.nextModeTransitionID += 1                    
        
    def add_error(self, error):
        """Adds an error to the error collector object.
        
        @param    error:
                    The error to add.
        @type     error:
                    A subclass of errors.AbstractError.
        """
        self.parent.add_error(error)
       
    def get_subcomponent(self, subcomponent_name):
        """Returns the subcomponent with the given name if one exists, None
        otherwise.
                
        @param    subcomponent_name:
                    The name of the subcomponent.
        @type     subcomponent_name:
                    string
        @result:  The subcomponent with the given name if one exists, None
                  otherwise.
        @rtype:   An object of nominalbehaviour.DataSubcomponent or
                  nominalbehaviour.NonDataSubcomponent.
        """
        subcomponent = None
        if subcomponent_name in self.non_data_subcomponents:
            subcomponent = self.non_data_subcomponents[subcomponent_name]
        elif subcomponent_name in self.data_subcomponents:
            subcomponent = self.data_subcomponents[subcomponent_name]
        return subcomponent
             
    def subcomponent_category_is_supported(self, category):
        """Returns whether the category of the component implementation
        supports the declaration of subcomponents of a given category.
                
        @param    category:
                    The category of a subcomponent.
        @type     category:
                    The string representation of a value defined in
                    nominalbehaviour.Category.
        @result:  True if the category of the component implementation
                  supports the declaration of subcomponents of a given category
                  and False otherwise.
        @rtype:   bool
        """
        if self.category == Category.PROCESS:
            if category != Category.THREAD_GROUP and \
               category != Category.THREAD and \
               category != Category.DATA:
                return False
            else:
                return True
        elif self.category == Category.THREAD_GROUP:
            if category != Category.THREAD_GROUP and \
               category != Category.THREAD and \
               category != Category.DATA:
                return False
            else:
                return True
        elif self.category == Category.THREAD:
            if category != Category.DATA:
                return False
            else:
                return True
        elif self.category == Category.MEMORY:
            if category != Category.MEMORY:
                return False
            else:
                return True
        elif self.category == Category.DEVICE:
            if category != Category.DATA:
                return False 
            else:
                return True
        elif self.category == Category.SYSTEM:
            if category != Category.PROCESS and \
               category != Category.DATA and \
               category != Category.PROCESSOR and \
               category != Category.MEMORY and \
               category != Category.BUS and \
               category != Category.DEVICE and \
               category != Category.SYSTEM:
                return False
            else:
                return True
        else:
            return False
        
    def subcomponent_binding_is_supported(self, category, binding):
        """Returns True if the given category supports the given binding and
        False otherwise.
        
        @param    category:
                    The category that is to be checked for support.
        @type     category:
                    A value defined in nominalbehaviour.Category.
        @param    binding:
                    The binding that is to be checked for support.
        @type     binding:
                    A value defined in nominalbehaviour.Binding.
        @return:  True if the given category supports the given binding and
                  False otherwise.
        @rtype:   bool
        """
        if category == Category.PROCESS:
            if binding != Binding.STOREDIN and \
               binding != Binding.RUNNINGON:
                return False
            else:
                return True
        elif category == Category.PROCESSOR or \
             category == Category.MEMORY or \
             category == Category.BUS or \
             category == Category.DEVICE or \
             category == Category.SYSTEM:
            if binding != Binding.ACCESSES:
                return False
            else:
                return True
        else:
            return False
        
    def get_referenced_component_type(self):
        """Returns the component type object the component implementation
        refers to, if it exists, None otherwise.
        
        @result:    The component type object the component implementation
                    refers to, if it exists, None otherwise.
        @rtype:     A nominalbehaviour.ComponentType object.
        """
        type = self.parent.get_component_type(self.name.split(".")[0])
        return type
    
    def get_mode(self, mode_name):
        """Returns a mode object with the given name if one exists, None
        otherwise.
        
        @param    mode_name:
                    The name of the mode to search.
        @type     mode_name:
                    string
        @result:  A mode object with the given name if one exists, None
                  otherwise.
        @rtype:   A nominalbehaviour.Mode object.
        """ 
        if mode_name in self.modes:
            return self.modes[mode_name]
        else:
            return None

    def port_is_target_of_connection_or_flow(self, port, mode=None):
        """Returns whether the given port is the target of a port connection
        or flow in the given mode.
        
        @param    port:
                    The name of the port.
        @type     port:
                    string
        @param    mode:
                    The name of the mode in which the connection or flow has to
                    be active. If not set or set to None, then it is determined
                    whether there exists a mode in which the port is the target
                    of a flow or connection.
        @type     mode:
                    string
        @return:  A set of connections that target the given port (in the given
                  mode, respectively).
        @rtype:   A set of PortConnection and Flow objects.
        """
        
        connectionsAndFlows = set()
        
        if mode is not None:
            for conn in self.data_port_connections:
                if conn.target_port == port:
                    if set(conn.in_modes) == set() or \
                       mode in conn.in_modes:
                        connectionsAndFlows.add(conn)
                        
            for flow in self.flows:
                if flow.expression.left.name == port:
                    if set(flow.in_modes) == set() or \
                       mode in flow.in_modes:
                        connectionsAndFlows.add(flow)
        else:
            for mode_name in self.modes.keys():
                connectionsAndFlows = connectionsAndFlows.union(self.port_is_target_of_connection_or_flow(port, mode_name)) 
                    
        return connectionsAndFlows
    
    def port_is_target_of_effect(self, port):
        """Returns whether the given port is the target of an effect of a
        mode transition.
        
        @param    port:
                    The name of the port.
        @type     port:
                    string
        @return:  A set of transitions that set the given port in one of their
                  effects.
        @rtype:   A set of ModeTransition objects.
        """
        
        transitions = set()
        
        for trans in self.normal_mode_transitions:
            effects = trans.label.effects
            for effect in effects:
                if effect.left.name == port:
                    transitions.add(trans)
                
        return transitions
    
    def get_all_trajectory_equation_variables(self, root_node):
        """Returns all trajectory equation variables contained in the given
        expression.
        
        @param    root_node:
                    The root node of the expression.
        @type     root_node:
                    An object of a subclass of expressions.ExpressionTreeNode.
        @return:  A list of all trajectory equation variables contained in the
                  given expression.
        @rtype:   A list of strings.
        """
        if isinstance(root_node, IdentifierTreeNode):
            return []
        elif isinstance(root_node, NumberTreeNode):
            return []
        elif isinstance(root_node, BooleanTreeNode):
            return []
        elif isinstance(root_node, UnaryOperatorTreeNode):
            return []
        elif isinstance(root_node, BinaryOperatorTreeNode):
            left = self.get_all_trajectory_equation_variables(root_node.left)
            right = self.get_all_trajectory_equation_variables(root_node.right)
            return left + right
        elif isinstance(root_node, TrajectoryEquationTreeNode):
            return [root_node.variable.name]
        elif isinstance(root_node, TrajectoryExpressionTreeNode):
            return []
        elif isinstance(root_node, CaseExpressionTreeNode):
            result = []
            for (_, expr) in root_node.cases:
                result += self.get_all_trajectory_equation_variables(expr)
            result += self.get_all_trajectory_equation_variables(root_node.default_expression)
            return result
        else:
            assert False

    def get_data_element_type_guard(self, name, source_mode, trans, line, pos):
        """Returns the type of the data element found under the given name
        taking the rules of guards into consideration.
        
        @param    name:
                    The name of the data element.
        @type     name:
                    string
        @param    source_mode:
                    The source mode of the transition of the guard.
        @type     source_mode:
                    string
        @param    trans:
                    The transition to which errors should be associated.
        @type     trans:
                    A ModeTransition object.
        @param    line:
                    The line in which the guard is defined.
        @type     line:
                    int
        @param    pos:
                    The column position at which the guard is defined.
        @type     pos:
                    int
        @return:  The type of the data element found under the given name.
        @rtype:   A tuple consisting of a constant defined in
                  expressions.NodeType and additional information.
        """
        is_in_enums = False
        enum_type = None
        for subcomp in self.data_subcomponents.values():
            if name == subcomp.name:
                if not subcomp.in_modes or source_mode in subcomp.in_modes:
                    if isinstance(subcomp.data_type, Int):
                        return (NodeType.INT, subcomp, None)
                    elif isinstance(subcomp.data_type, IntRange):
                        return (NodeType.INT_RANGE, subcomp, subcomp.data_type.get_bounds())
                    elif isinstance(subcomp.data_type, Real):
                        return (NodeType.REAL, subcomp, None)
                    elif isinstance(subcomp.data_type, Bool):
                        return (NodeType.BOOL, subcomp, None)
                    elif isinstance(subcomp.data_type, Continuous):
                        return (NodeType.CONTINUOUS, subcomp, None)
                    elif isinstance(subcomp.data_type, Clock):
                        return (NodeType.CLOCK, subcomp, None)
                    elif isinstance(subcomp.data_type, Enum):
                        return (NodeType.ENUM_FIELD, subcomp, subcomp.data_type.values)
                else:
                    self.add_error(
                            GuardRefersToInactiveSubcomponentError(self.file,
                                                                   line, pos,
                                                                   "I-12", name,
                                                                   trans))
                    return (NodeType.ERROR, None, None)
            else:
                if isinstance(subcomp.data_type, Enum):
                    if name in subcomp.data_type.values:
                        is_in_enums = True
                        
        # search the identifier in the ports of the type
        type = self.get_referenced_component_type()
        if type is not None:
            ports = type.in_data_ports.values() + type.out_data_ports.values()
            for port in ports:
                if name == port.name:
                    if isinstance(port.data_type, Int):
                        return (NodeType.INT, port, None)
                    elif isinstance(port.data_type, IntRange):
                        return (NodeType.INT_RANGE, port, port.data_type.get_bounds())
                    elif isinstance(port.data_type, Real):
                        return (NodeType.REAL, port, None)
                    elif isinstance(port.data_type, Bool):
                        return (NodeType.BOOL, port, None)
                    elif isinstance(port.data_type, Continuous):
                        return (NodeType.CONTINUOUS, port, None)
                    elif isinstance(port.data_type, Clock):
                        return (NodeType.CLOCK, port, None)
                    elif isinstance(port.data_type, Enum):
                        return (NodeType.ENUM_FIELD, port, port.data_type.values)
                else:
                    if isinstance(port.data_type, Enum):
                        if name in port.data_type.values:
                            is_in_enums = True
                            enum_type = port
                             
        if not is_in_enums:
            self.add_error(
                    GuardRefersToUnknownSubcomponentOrDataPortError(self.file,
                                                                    line, pos,
                                                                    "I-12",
                                                                    name,
                                                                    trans))
            return (NodeType.ERROR, None, None)
        else:
            return (NodeType.ENUM_VALUE, enum_type, [name])
    
    def get_data_element_type_lhs_effect(self, name, target_mode, trans, line,
                                         pos):
        """Returns the type of the data element found under the given name
        taking the rules of effects into consideration.
        
        @param    name:
                    The name of the data element.
        @type     name:
                    string
        @param    target_mode:
                    The target mode of the transition of the effect.
        @type     target_mode:
                    string
        @param    trans:
                    The transition to which errors should be associated.
        @type     trans:
                    A ModeTransition object.
        @param    line:
                    The line in which the effect is defined.
        @type     line:
                    int
        @param    pos:
                    The column position at which the effect is defined.
        @type     pos:
                    int
        @return:  The type of the data element found under the given name.
        @rtype:   A tuple consisting of a constant defined in
                  expressions.NodeType and additional information.
        """
        if name in self.data_subcomponents.keys():
            subcomp = self.data_subcomponents[name]
            if not subcomp.in_modes or target_mode in subcomp.in_modes:
                if isinstance(subcomp.data_type, Int):
                    return (NodeType.INT, subcomp, None)
                elif isinstance(subcomp.data_type, IntRange):
                    return (NodeType.INT_RANGE, subcomp, subcomp.data_type.get_bounds())
                elif isinstance(subcomp.data_type, Real):
                    return (NodeType.REAL, subcomp, None)
                elif isinstance(subcomp.data_type, Bool):
                    return (NodeType.BOOL, subcomp, None)
                elif isinstance(subcomp.data_type, Clock):
                    return (NodeType.CLOCK, subcomp, None)
                elif isinstance(subcomp.data_type, Continuous):
                    return (NodeType.CONTINUOUS, subcomp, None)
                elif isinstance(subcomp.data_type, Enum):
                    return (NodeType.ENUM_FIELD, subcomp, subcomp.data_type.values)
            else:
                self.add_error(
                        DataSubcomponentSetAndNotActiveError(self.file,
                                                             line, pos, "I-13",
                                                             name, target_mode,
                                                             trans))
                return (NodeType.ERROR, None, None)
        else:
            type = self.get_referenced_component_type()
            if type is not None:
                port = None
                if name in type.out_data_ports.keys():
                    port = type.out_data_ports[name]
                    
                if port is not None:
                    if isinstance(port.data_type, Int):
                        return (NodeType.INT, port, None)
                    elif isinstance(port.data_type, IntRange):
                        return (NodeType.INT_RANGE, port, port.data_type.get_bounds())
                    elif isinstance(port.data_type, Real):
                        return (NodeType.REAL, port, None)
                    elif isinstance(port.data_type, Bool):
                        return (NodeType.BOOL, port, None)
                    elif isinstance(port.data_type, Enum):
                        return (NodeType.ENUM_FIELD, port, port.data_type.values)
                        
        self.add_error(EffectRefersToUnknownIdentifierLhsError(self.file, line,
                                                               pos, "I-13",
                                                               name, trans))
        return (NodeType.ERROR, None, None)

    def get_data_element_type_rhs_effect(self, name, source_mode, trans, line,
                                         pos):
        """Returns the type of the data element found under the given name
        taking the rules of effects into consideration.
        
        @param    name:
                    The name of the data element.
        @type     name:
                    string
        @param    source_mode:
                    The source mode of the transition of the effect.
        @type     source_mode:
                    string
        @param    trans:
                    The transition to which errors should be associated.
        @type     trans:
                    A ModeTransition object.
        @param    line:
                    The line in which the effect is defined.
        @type     line:
                    int
        @param    pos:
                    The column position at which the effect is defined.
        @type     pos:
                    int
        @return:  The type of the data element found under the given name.
        @rtype:   A tuple consisting of a constant defined in
                  expressions.NodeType and additional information.
        """
        is_in_enums = False
        enum_type = None # type of enum if enum name
        for subcomp in self.data_subcomponents.values():
            if name == subcomp.name:
                if not subcomp.in_modes or source_mode in subcomp.in_modes:
                    if isinstance(subcomp.data_type, Int):
                        return (NodeType.INT, subcomp, None)
                    elif isinstance(subcomp.data_type, IntRange):
                        return (NodeType.INT_RANGE, subcomp, subcomp.data_type.get_bounds())
                    elif isinstance(subcomp.data_type, Real):
                        return (NodeType.REAL, subcomp, None)
                    elif isinstance(subcomp.data_type, Bool):
                        return (NodeType.BOOL, subcomp, None)
                    elif isinstance(subcomp.data_type, Clock):
                        return (NodeType.CLOCK, subcomp, None)
                    elif isinstance(subcomp.data_type, Continuous):
                        return (NodeType.CONTINUOUS, subcomp, None)
                    elif isinstance(subcomp.data_type, Enum):
                        return (NodeType.ENUM_FIELD, subcomp, subcomp.data_type.values)
                else:
                    self.add_error(
                        EffectRefersToInactiveSubcomponentRhsError(self.file,
                                                                   line, pos,
                                                                   "I-14", name,
                                                                   source_mode,
                                                                   trans))
                    return (NodeType.ERROR, None, None)
            else:
                if isinstance(subcomp.data_type, Enum):
                    if name in subcomp.data_type.values:
                        is_in_enums = True
                        enum_type = subcomp
                
        type = self.get_referenced_component_type()
        if type is not None:
            ports = type.in_data_ports.values() + type.out_data_ports.values()
            for port in ports:
                if name == port.name:
                    if isinstance(port.data_type, Int):
                        return (NodeType.INT, port, None)
                    elif isinstance(port.data_type, IntRange):
                        return (NodeType.INT_RANGE, port, port.data_type.get_bounds())
                    elif isinstance(port.data_type, Real):
                        return (NodeType.REAL, port, None)
                    elif isinstance(port.data_type, Bool):
                        return (NodeType.BOOL, port, None)
                    elif isinstance(port.data_type, Enum):
                        return (NodeType.ENUM_FIELD, port, port.data_type.values)
                else:
                    if isinstance(port.data_type, Enum):
                        if name in port.data_type.values:
                            is_in_enums = True
                            enum_type = port
                            
        if is_in_enums:
            return (NodeType.ENUM_VALUE, enum_type, [name])
        else:
            self.add_error(
                EffectRefersToUnknownIdentifierRhsError(self.file, line, pos,
                                                        "I-14", name, trans))
            return (NodeType.ERROR, None, None) 
    
    def get_data_element_type_invariant(self, name, mode, mode_obj, line, pos):
        """Returns the type of the data element found under the given name
        taking the rules of invariants into consideration.
        
        @param    name:
                    The name of the data element.
        @type     name:
                    string
        @param    mode:
                    The name of the mode with the invariant.
        @type     mode:
                    string
        @param    mode_obj:
                    The mode with which the invariant is associated.
        @type     mode:
                    A Mode object.
        @param    line:
                    The line in which the invariant is defined.
        @type     line:
                    int
        @param    pos:
                    The column position at which the invariant is defined.
        @type     pos:
                    int
        @return:  The type of the data element found under the given name.
        @rtype:   A tuple consisting of a constant defined in
                  expressions.NodeType and additional information.
        """
        if name in self.data_subcomponents:
            subcomp = self.data_subcomponents[name]
            if not subcomp.in_modes or mode in subcomp.in_modes:
                if isinstance(subcomp.data_type, Continuous):
                    return (NodeType.CONTINUOUS, subcomp, None)
                elif isinstance(subcomp.data_type, Clock):
                    return (NodeType.CLOCK, subcomp, None)
                else:
                    self.add_error(
                            ModeInvariantRefersToIllegalDataTypeError(self.file,
                                                                      line, pos,
                                                                      "I-7",
                                                                      name,
                                                  subcomp.data_type.get_name(),
                                                                    mode_obj))
            else:
                self.add_error(
                            ModeInvariantRefersToInactiveSubcomponentError(
                                                                    self.file,
                                                                    line, pos,
                                                                    "I-7",
                                                                    name,
                                                                    mode_obj))
        
        else:
            self.add_error(
                    ModeInvariantRefersToUnknownSubcomponentError(self.file,
                                                                  line, pos,
                                                                  "I-7",
                                                                  name, 
                                                                  mode_obj))
        return (NodeType.ERROR, None, None) 
        
    def get_data_element_type_lhs_flow(self, name, in_modes, flow, line, pos):
        """Returns the type of the data element found under the given name
        taking the rules of the left-hand side of flows into consideration.
        
        @param    name:
                    The name of the data element.
        @type     name:
                    string
        @param    in_modes:
                    The In-modes-list of the flow .
        @type     in_modes:
                    list of strings
        @param    flow:
                    The flow to which this lhs is associated.
        @type     flow:
                    A Flow object.
        @param    line:
                    The line in which the flow is defined.
        @type     line:
                    int
        @param    pos:
                    The column position at which the flow is defined.
        @type     pos:
                    int
        @return:  The type of the data element found under the given name.
        @rtype:   A tuple consisting of a constant defined in
                  expressions.NodeType and additional information.
        """        
        if name.find(".") == -1:
            type = self.get_referenced_component_type()
            if type is not None:
                if name in type.out_data_ports:
                    port = type.out_data_ports[name]
                    if isinstance(port.data_type, Int):
                        return (NodeType.INT, port, None)
                    elif isinstance(port.data_type, Real):
                        return (NodeType.REAL, port, None)
                    elif isinstance(port.data_type, Bool):
                        return (NodeType.BOOL, port, None)
                    elif isinstance(port.data_type, Enum):
                        return (NodeType.ENUM_FIELD, port, port.data_type.values)
                    elif isinstance(port.data_type, IntRange):
                        return (NodeType.INT_RANGE, port, port.data_type.bounds)
                else:
                    self.add_error(
                            FlowReferencesUnknownOutgoingDataPortError(self.file,
                                                                       line, pos,
                                                                       "H-1",
                                                                       name,
                                                                       flow))
        else:
            [subcomp_name, port_name] = name.split(".")
            if subcomp_name not in self.non_data_subcomponents.keys():
                self.add_error(
                            FlowReferencesUnknownSubcomponentError(self.file,
                                                                   line, pos,
                                                                   "H-1",
                                                                   subcomp_name,
                                                                   flow))
                
            else:
                subcomp = self.non_data_subcomponents[subcomp_name]
                
                # check that the subcomponent is active in all modes in which
                # the flow is active
                
                subcomp_modes = set(self.modes.keys() if subcomp.in_modes == []
                                else subcomp.in_modes)
                                
                flow_modes = set(self.modes.keys() if in_modes == []
                                 else in_modes)
                
                if not flow_modes.issubset(subcomp_modes):
                    illegal_modes = flow_modes - subcomp_modes
                    for mode in illegal_modes:
                        self.add_error(
                                PortFlowReferencesInactiveSubcomponentError(
                                        self.file, line, pos, "H-1",
                                        subcomp_name, mode, flow))
                else:
                    type = subcomp.get_referenced_component_type()
                    if type is not None:
                        if port_name in type.in_data_ports.keys():
                            port = type.in_data_ports[port_name]
                            if isinstance(port.data_type, Int):
                                return (NodeType.INT, port, None)
                            elif isinstance(port.data_type, Real):
                                return (NodeType.REAL, port, None)
                            elif isinstance(port.data_type, Bool):
                                return (NodeType.BOOL, port, None)
                            elif isinstance(port.data_type, Enum):
                                return (NodeType.ENUM_FIELD, port, port.data_type.values)
                            elif isinstance(port.data_type, IntRange):
                                return (NodeType.INT_RANGE, port, port.data_type.bounds)
                        else:
                            self.add_error(
                                    FlowReferencesUnknownIncomingDataPortOfSubcompError(
                                            self.file, line, pos, "H-1",
                                            subcomp_name, port_name, flow))  
        return (NodeType.ERROR, None, None)
    
    def get_data_element_type_rhs_flow(self, name, in_modes, flow, line, pos):
        """Returns the type of the data element found under the given name
        taking the rules of the right-hand side of flows into consideration.
        
        @param    name:
                    The name of the data element.
        @type     name:
                    string
        @param    in_modes:
                    The In-modes-list of the flow .
        @type     in_modes:
                    list of strings
        @param    flow:
                    The flow to which this lhs is associated.
        @type     flow:
                    A Flow object.
        @param    line:
                    The line in which the flow is defined.
        @type     line:
                    int
        @param    pos:
                    The column position at which the flow is defined.
        @type     pos:
                    int
        @return:  The type of the data element found under the given name.
        @rtype:   A tuple consisting of a constant defined in
                  expressions.NodeType and additional information.
        """
        enum_type = None # If ID is an enum value, this var contains the type that declared it
        if name.find(".") == -1:
            is_in_enums = False
            type = self.get_referenced_component_type()
            if type is not None:
                for port in type.in_data_ports.values():
                    if name == port.name:
                        if isinstance(port.data_type, Int):
                            return (NodeType.INT, port, None)
                        elif isinstance(port.data_type, Real):
                            return (NodeType.REAL, port, None)
                        elif isinstance(port.data_type, Bool):
                            return (NodeType.BOOL, port, None)
                        elif isinstance(port.data_type, Enum):
                            return (NodeType.ENUM_FIELD, port, port.data_type.values)
                        elif isinstance(port.data_type, IntRange):
                            return (NodeType.INT_RANGE, port, port.data_type.bounds)
                    else:
                        if isinstance(port.data_type, Enum):
                            if name in port.data_type.values:
                                is_in_enums = True
                                enum_type = port
            if type is not None:
                for port in type.out_data_ports.values():
                    if isinstance(port.data_type, Enum):
                            if name in port.data_type.values:
                                is_in_enums = True
                                enum_type = port
                
            for subcomp in self.non_data_subcomponents.values():
                type = subcomp.get_referenced_component_type()
                if type is not None:
                    for port in type.out_data_ports.values():
                        if isinstance(port.data_type, Enum):
                            if name in port.data_type.values:
                                is_in_enums = True
                                enum_type = port
                                
            if not is_in_enums:
                self.add_error(
                        FlowReferencesUnknownIncomingDataPortError(self.file,
                                                                   line, pos,
                                                                   "H-2", name,
                                                                   flow))
            else:
                return (NodeType.ENUM_VALUE, enum_type, [name])
        else:
            [subcomp_name, port_name] = name.split(".")
            if subcomp_name not in self.non_data_subcomponents.keys():
                self.add_error(
                            FlowReferencesUnknownSubcomponentError(self.file,
                                                                   line, pos,
                                                                   "H-2",
                                                                   subcomp_name,
                                                                   flow))
            else:
                subcomp = self.non_data_subcomponents[subcomp_name]
                
                # check that the subcomponent is active in all modes in which
                # the flow is active
                
                subcomp_modes = set(self.modes.keys() if subcomp.in_modes == []
                                else subcomp.in_modes)
                                
                flow_modes = set(self.modes.keys() if in_modes == []
                                 else in_modes)
                
                if not flow_modes.issubset(subcomp_modes):
                    illegal_modes = flow_modes - subcomp_modes
                    for mode in illegal_modes:
                        self.add_error(
                                PortFlowReferencesInactiveSubcomponentError(
                                        self.file, line, pos, "H-2",
                                        subcomp_name, mode, flow))
                else:
                    type = subcomp.get_referenced_component_type()
                    if type is not None:
                        if port_name in type.out_data_ports.keys():
                            port = type.out_data_ports[port_name]
                            if isinstance(port.data_type, Int):
                                return (NodeType.INT, port, None)
                            elif isinstance(port.data_type, Real):
                                return (NodeType.REAL, port, None)
                            elif isinstance(port.data_type, Bool):
                                return (NodeType.BOOL, port, None)
                            elif isinstance(port.data_type, Enum):
                                return (NodeType.ENUM_FIELD, port, port.data_type.values)
                            elif isinstance(port.data_type, IntRange):
                                return (NodeType.INT_RANGE, port, port.data_type.bounds)
                        else:
                            self.add_error(
                                    FlowReferencesUnknownOutgoingDataPortOfSubcompError(
                                            self.file, line, pos, "H-2", subcomp_name,
                                            port_name, flow))

        return (NodeType.ERROR, None, None)

    def get_data_element_type_datasubcomponent_default_value(self, name, data, line, pos):
        # identifier has to be enum value belonging to the enum
        if not isinstance(data.data_type, Enum) :
            self.add_error(InitializationValueTypeMismatchError(self.file, line, pos, "F-4", data.data_type.get_name(), data))
            return (NodeType.ERROR, None, None)
        elif data.default_value.name not in data.data_type.values:
            self.add_error(EnumValueIllegalError(self.file, line, pos, "F-4", name, data))
            return (NodeType.ERROR, None, None)
        else:
            return (NodeType.ENUM_VALUE, data, [name])

    def port_is_target_in_flow(self, out_port_name):
        for flow in self.flows:
            if flow.expression.left.name == out_port_name:
                return True
        return False
    
    def port_is_target_in_connection(self, out_port_name):
        for conn in self.data_port_connections:
            if conn.target_port == out_port_name:
                return True
        return False
                                                                
    def port_is_target_in_flow_or_connection(self, out_port_name):
        """Returns True if the given port name is target of a flow or port
        connection, False otherwise.
        
        @param    out_port_name:
                    The name of the port.
        @type     out_port_name:
                    string
        @return:  True if the given port name is target of a flow or port
                  connection, False otherwise.
        @rtype:   bool
        """
        return self.port_is_target_in_flow(out_port_name) or \
               self.port_is_target_in_connection(out_port_name)

    def get_fully_qualified_name(self):
        """Returns the fully qualified name of the component implementation.
        
        @return:    The fully qualified name of the component implemenation.
        @rtype:     string
        """
        return (self.parent.parent.name + "::" + self.name)
                
    def get_non_data_subcomponent(self, subcomponent_name):
        """Returns non-data subcomponent object with the given name if one
        exists and None otherwise.
        
        @param    subcomponent_name:
                    The name of the subcomponent object to search.
        @type     subcomponent_name:
                    string
        @return:  The non-data subcomponent object with the given name if one
                  exists and None otherwise.
        @rtype:   nominalbehaviour.NonDataSubcomponent
        """
        if subcomponent_name in self.non_data_subcomponents.keys():
            return self.non_data_subcomponents[subcomponent_name]
        else:
            return None
                      
    def get_starting_mode(self):
        """Returns the starting mode of the component implementation.
        
        Note: If the component implementation has multiple starting modes, only
        the first one found is returned.
        
        @return:  The starting mode of the component implementation.
        @rtype:   nominalbehaviour.Mode
        """
        for mode in self.modes.values():
            if mode.type == ModeType.ACTIVATION or \
               mode.type == ModeType.INITIAL:
                return mode
            
    def get_normal_transitions_leaving_mode(self, mode):
        """Get all non-probabilistic transitions leaving the given mode.
        
        @param    mode:
                    The mode which the transitions leave.
        @type     mode:
                    string
        @result:  A set of all transitions leaving the given mode.
        @rtype:   A set of nominalbehaviour.ModeTransition objects.
        """
        result = set()
        for transition in self.normal_mode_transitions:
            if transition.source_mode == mode:
                result.add(transition)
        return result
    
    def get_normal_transitions_targeting_mode(self, mode):
        """Get all non-probabilistic transitions targeting the given mode.
        
        @param    mode:
                    The mode which the transitions target.
        @type     mode:
                    string
        @result:  A set of all transitions targeting the given mode.
        @rtype:   A set of nominalbehaviour.ModeTransition objects.
        """
        result = set()
        for transition in self.normal_mode_transitions:
            if transition.target_mode == mode:
                result.add(transition)
        return result
    
    def get_transitions_targeting_mode(self, mode):
        """Get all transitions targeting the given mode.
        
        @param    mode:
                    The mode which the transitions target.
        @type     mode:
                    string
        @result:  A set of all transitions targeting the given mode.
        @rtype:   A set of nominalbehaviour.ModeTransition objects.
        """
        result = set()
        
        for transition in (self.normal_mode_transitions + self.probability_mode_transitions):
            if transition.target_mode == mode:
                result.add(transition)
        return result

    def get_total_number_of_transitions(self):
        """Returns the total number of transitions in the component
        implementation.
        
        @return:    The total number of transitions.
        @rtype:     int
        """
        return len(self.normal_mode_transitions)

    def get_total_number_of_modes(self):
        """Returns the total number of modes in the component implementation.
        
        @return:    The total number of modes.
        @rtype:     int
        """
        return len(self.modes)

    def get_total_number_of_nondata_component_instances(self):
        """Returns the total number of non-data component instances in the
        component implementation.
        
        @return:    The total number of non-data component instances.
        @rtype:     int
        """
        return len(self.non_data_subcomponents)

    def get_total_number_of_data_component_instances(self):
        """Returns the total number of data component instances in the component
        implementation.
        
        @return:    The total number of data component instances.
        @rtype:     int
        """
        return len(self.data_subcomponents)

    def get_total_number_of_data_port_connections(self):
        """Returns the total number of data port connections in the component
        implementation.
        
        @return:    The total number of data port connections.
        @rtype:     int
        """
        return len(self.data_port_connections)

    def get_total_number_of_flows(self):
        """Returns the total number of flows in the component implementation.
        
        @return:    The total number of flows.
        @rtype:     int
        """
        return len(self.flows)

    def get_total_number_of_event_port_connections(self):
        """Returns the total number of event port connections in the component
        implementation.
        
        @return:    The total number of event port connections.
        @rtype:     int
        """
        return len(self.event_port_connections)

        
class Port(LocatedInformation):
    """This class is the data container for a port in a SLIM specification.
    """
    
    def __init__(self, type, name, file, line, pos):
        """Initializes a port object.
        
        @param    type:
                    The component type the port belongs to.
        @type     type:
                    A ComponentType object.
        @param    name:
                    The name of the port.
        @type     name:
                    string
        @param    file:
                    The file in which the port is defined.
        @type     file:
                    string
        @param    line:
                    The line in which the port is defined.
        @type     line:
                    int
        @param    pos:
                    The column position at which the port was defined.
        @type     pos:
                    int
        """
        LocatedInformation.__init__(self, type, "port", file, line, pos)
        self.name = name
        
    def get_ref_identifier(self):
        """Retrieves an identifier that is used as a reference point for errors."""
        return self.parent.get_ref_identifier() + "::port::" + self.name
        
class InEventPort(Port):
    """This class is the data container for an incoming event port in a SLIM
    specification.
    """

    def __init__(self, type, name, file, line, pos):
        """Initializes an incoming event port object.

        @param    type:
                    The component type the port belongs to.
        @type     type:
                    A ComponentType object.
        @param    name:
                    The name of the port.
        @type     name:
                    string
        @param    file:
                    The file in which the port is defined.
        @type     file:
                    string
        @param    line:
                    The line in which the port is defined.
        @type     line:
                    int
        @param    pos:
                    The column position at which the port was defined.
        @type     pos:
                    int
        """
        Port.__init__(self, type, name, file, line, pos)

class OutEventPort(Port):
    """This class is the data container for an outgoing event port in a SLIM
    specification.
    """
    
    def __init__(self, type, name, nonblocking, file, line, pos):
        """Initializes an outgoing event port object.

        @param    type:
                    The component type the port belongs to.
        @type     type:
                    A ComponentType object.
        @param    name:
                    The name of the port.
        @type     name:
                    string
        @param    nonblocking:
                    Specifies whether the port is nonblocking.
        @type     nonblocking:
                    bool
        @param    file:
                    The file in which the port is defined.
        @type     file:
                    string
        @param    line:
                    The line in which the port is defined.
        @type     line:
                    int
        @param    pos:
                    The column position at which the port was defined.
        @type     pos:
                    int
        """
        Port.__init__(self, type, name, file, line, pos)
        self.nonblocking = nonblocking

class DataPort(Port):
    """This class is a data container for a data port in a SLIM specification.
    """
    
    def __init__(self, type, name, data_type, default_value, file, line, pos):
        """Initializes a data port object.
        
        @param    type:
                    The component type the port belongs to.
        @type     type:
                    A ComponentType object.
        @param    name:
                    The name of the port.
        @type     name:
                    string
        @param    data_type:
                    The data type of the port.
        @type     data_type:
                    An object of a subclass of 
                    nominalbehaviour.AbstractDataType.
        @param    default_value:
                    The default value of the port.
        @type     default_value:
                    An object of expressions.IdentifierTreeNode,
                    expressions.BooleanTreeNode or expressions.NumberTreeNode.
        @param    file:
                    The file in which the port was defined.
        @type     file:
                    string
        @param    line:
                    The line in which the port was defined.
        @type     line:
                    int
        @param    pos:
                    The column position at which the port was defined.
        @type     pos:
                    int
        """
        Port.__init__(self, type, name, file, line, pos)
        self.data_type = data_type
        self.name = name
        self.default_value = default_value

class OutDataPort(DataPort):
    """This class is a data container for an outgoing data port in a SLIM
    specification.
    """
    
    def __init__(self, type, name, data_type, default_value, observable, alarm, 
                 file, line, pos):
        """Initializes an outgoing data port object.
        
        @param    type:
                    The component type the port belongs to.
        @type     type:
                    A ComponentType object.
        @param    name:
                    The name of the port.
        @type     name:
                    string
        @param    data_type:
                    The data type of the port.
        @type     data_type:
                    An object of a subclass of 
                    nominalbehaviour.AbstractDataType.
        @param    default_value:
                    The default value of the port.
        @type     default_value:
                    An object of expressions.IdentifierTreeNode,
                    expressions.BooleanTreeNode or expressions.NumberTreeNode.
        @param    observable:
                    Indicates whether the port is observable.
        @type     observable:
                    bool
        @param    alarm:
                    Indicates whether the port is alarm.
        @type     alarm:
                    bool
        @param    file:
                    The file in which the port was defined.
        @type     file:
                    string
        @param    line:
                    The line in which the port was defined.
        @type     line:
                    int
        @param    pos:
                    The column position at which the port was defined.
        @type     pos:
                    int
        """
        DataPort.__init__(self, type, name, data_type, default_value, file, line, pos)
        self.name = name
        self.observable = observable
        self.alarm = alarm

class ErrorEventPort(OutEventPort):
    """An error event port which is used to represent probabilistic events.
    
    This is an extended out event port. It is associated with a probability
    distribution and a parameter for this distribution.
    The distribution can either be fixed or poisson, indicated by the
    corresponding constants defined in errorbehaviour.DistributionType.
    For a fixed distribution, the parameter is the probability of the event.
    For a poisson distribution, the parameter is the rate of the event.
    In case of poisson distributions, a time unit for the rate may be
    specified, represented by one of the constants defined in the class
    errorbehaviour.TimeUnit.
    
    Note: An error event port must never be connected with another port.
    """
    
    def __init__(self, type, name, distribution, distribution_param, time_unit,
                 file, line, pos):
        OutEventPort.__init__(self, type, name, False, file, line, pos)
        self.distribution = distribution
        self.parameter = distribution_param
        self.time_unit = time_unit
        
    def get_parameter(self):
        """Returns the normalized parameter of the distribution associated
        with this port or None if no distribution is associated.
        
        @return: The parameter of the distribution associated with this port.
                 This is the literal string from the SLIM input (format [0-9]+'.'[0-9]+).
        @rtype:  String
        """
        return self.parameter
        
class PortConnection(LocatedInformation):
    """This class is a data container for a port connection in a SLIM
    specification.
    """
    
    def __init__(self, impl, number, source_port, target_port, in_modes,
                 passive, file, line, pos):
        """Initializes a port connection object.
        
        @param    impl:
                    The component implementation the connection belongs to.
        @type     impl:
                    A ComponentImplementation object.
        @param    number:
                    The id of the port connection.
        @type     number:
                    A non-negative integer.
        @param    source_port:
                    The source port of the port connection.
        @type     source_port:
                    string
        @param    target_port:
                    The target port of the port connection.
        @type     target_port:
                    string
        @param    in_modes:
                    A list of modes in which the connection is active.
        @type     in_modes:
                    A list of strings.
        @param    passive:
                    A boolean flag indicating whether the connection is passive.
        @type     passive:
                    A boolean.
        @param    file:
                    The file in which the port connection is defined.
        @type     file:
                    string
        @param    line:
                    The line in which the port connection is defined.
        @type     line:
                    int
        @param    pos:
                    The column position at which the port connection is defined.
        @type     pos:
                    int
        """
        LocatedInformation.__init__(self, impl, "port connection", file, line, pos)
        self.source_port = source_port
        self.target_port = target_port
        self.in_modes = in_modes
        self.passive = passive
        self.number = number

    def get_ref_identifier(self):
        """Retrieves an identifier that is used as a reference point for errors."""
        return self.parent.get_ref_identifier() + "::conn::" + str(self.number) 

class Flow(LocatedInformation):
    """This class is a data container for a flow between ports in a SLIM
    specification.
    """

    def __init__(self, impl, number, expression, in_modes, file, line, pos):
        """Initializes a flow object.
         
        @param    impl:
                    The component implementation the flow belongs to.
        @type     impl:
                    A ComponentImplementation object.
        @param    number:
                    The id of the flow.
        @type     number:
                    A non-negative integer.
        @param    expression:
                    The assignment connecting the outgoing port with the
                    incoming ports.
        @type     expression:
                    An object of a subclass of expressions.ExpressionTreeNode.
        @param    in_modes:
                    A list of modes in which the flow is active.
        @type     in_modes:
                    A list of strings.
        @param    file:
                    The file in which the flow is defined.
        @type     file:
                    string
        @param    line:
                    The line in which the flow is defined.
        @type     line:
                    int
        @param    pos:
                    The column position at which the flow is defined. 
        """
        LocatedInformation.__init__(self, impl, "flow", file, line, pos)
        self.expression = expression
        self.in_modes = in_modes
        self.number = number
        
    def get_ref_identifier(self):
        """Retrieves an identifier that is used as a reference point for errors."""
        return self.parent.get_ref_identifier() + "::flow::" + str(self.number) 

class NonDataSubcomponent(LocatedInformation):
    """This class is a data container for a non-data subcomponent in a SLIM
    specification.
    """

    def __init__(self, implementation, name, category, implementation_ref,
                 accesses, stored_in, running_on, in_modes, file, line, pos):
        """Initializes a non-data subcomponent object.
        
        @param    implementation:
                    The implementation in which the subcomponent is defined.
        @type     implementation:
                    nominalbehaviour.ComponentImplementation
        @param    name:
                    The name of the subcomponent.
        @type     name:
                    string
        @param    category:
                    The category of the subcomponent.
        @type     category:
                    A constant defined in nominalbehaviour.Category.
        @param    implementation_ref:
                    The referenced component implementation.
        @type     implementation_ref:
                    string
        @param    accesses:
                    The names of all subcomponents to which the subcomponent
                    has an accesses binding.
        @type     accesses:
                    A list of strings.
        @param    stored_in:
                    The names of all subcomponents to which the subcomponent
                    has an stored-in binding.
        @type     stored_in:
                    A list of strings.
        @param    running_on:
                    The names of all subcomponents to which the subcomponent
                    has an running-on binding.
        @type     running_on:
                    A list of strings.
        @param    in_modes:
                    A list of modes in which the subcomponent is active.
        @type     in_modes:
                    A list of strings.
        @param    file:
                    The file in which the subcomponent is defined.
        @type     file:
                    string
        @param    line:
                    The line in which the subcomponent was defined.
        @type     line:
                    int
        @param    pos:
                    The column position at which the subcomponent was defined.
        @type     pos:
                    int
        """
        LocatedInformation.__init__(self, implementation, category, file, line,
                                    pos)
        self.implementation = implementation
        self.name = name
        self.implementation_ref = implementation_ref
        self.accesses = accesses
        self.stored_in = stored_in
        self.running_on = running_on
        self.in_modes = in_modes
        self.category = category

    def get_ref_identifier(self):
        """Retrieves an identifier that is used as a reference point for errors."""
        return self.parent.get_ref_identifier() + "::subcomp::" + self.name 
                    
    def add_error(self, error):
        """Adds an error to the error collector object.
        
        @param    error:
                    The error to add.
        @type     error:
                    A subclass of errors.AbstractError.
        """
        self.parent.add_error(error)
        
    def parse_implementation_reference(self):
        """Parses the implementation reference and partitions it.
        
        @return:  A tuple consisting of the package name, the name of the type
                  and the name of the implementation.
        @rtype:   A tuple of three strings.
        """
        package_name = None
        type_name = None
        implementation_name = None
        
        if self.implementation_ref.find("::") != -1:
            split = self.implementation_ref.split("::")
            package_name = split[0]
            if split[1].find(".") != -1:
                split = split[1].split(".")
                type_name = split[0]
                implementation_name = split[1]
            else:
                type_name = split[1]
        else:
            if self.implementation_ref.find(".") != -1:
                split = self.implementation_ref.split(".")
                type_name = split[0]
                implementation_name = split[1]
            else:
                type_name = self.implementation_ref
                
        return (package_name, type_name, implementation_name)
        
    def get_referenced_component_type(self):
        """Returns the component type referenced by the subcomponent.
        
        @return:  The component type referenced by the subcomponent.
        @rtype:   nominalbehaviour.ComponentType
        """
        (package_name, type_name, _) = \
                self.parse_implementation_reference()
        error = False
            
        if package_name:
            ci = self.parent.parent.parent.parent
            package = ci.get_package(package_name)
            if package is None:
                error = True
        else:
            package = self.parent.parent.parent
            
        if not error:
            if package_name:
                type = package.get_component_type(type_name, True)
            else:
                type = package.get_component_type(type_name, False)
                
            if type is None:
                error = True
                
        if not error:
            return type
        else:
            return None
            
    def get_referenced_component_implementation(self):
        """Returns the component implementation referenced by the subcomponent.
        
        @return:  The component implementation referenced by the subcomponent.
        @rtype:   nominalbehaviour.ComponentImplementation
        """
        (package_name, type_name, implementation_name) = \
                self.parse_implementation_reference()
        error = False
        impl = None
        
        if package_name:
            ci = self.parent.parent.parent.parent
            package = ci.get_package(package_name)
            if package is None:
                error = True
        else:
            package = self.parent.parent.parent
            
        if not error:
            if package_name:
                if implementation_name:
                    impl = package.get_component_implementation(
                                    type_name + "." + implementation_name, True)
                else:
                    impls = package.get_implementations_for_type(type_name, True)
                    if len(impls) == 1:
                        impl = impls[0]
                    else:
                        error = True
            else:
                if implementation_name:
                    impl = package.get_component_implementation(
                                    type_name + "." + implementation_name, False)
                else:
                    impls = package.get_implementations_for_type(type_name, False)
                    if len(impls) == 1:
                        impl = impls[0]
                    else:
                        error = True
                
            if impl is None:
                error = True
        
        if not error:
            return impl
        else:
            return None
        
class DataSubcomponent(LocatedInformation):
    """This class is a data container for a data subcomponent in a SLIM
    specification.
    """

    def __init__(self, implementation, name, data_type, in_modes, default_value,
                 file, line, pos):
        """Initializes a data subcomponent object.

        @param    implementation:
                    The implementation in which the subcomponent is defined.
        @type     implementation:
                    nominalbehaviour.ComponentImplementation
        @param    name:
                    The name of the subcomponent.
        @type     name:
                    string
        @param    data_type:
                    The data type of the subcomponent.
        @type     data_type:
                    An object of a subclass of
                    nominalbehaviour.AbstractDataType.
        @param    in_modes:
                    A list of modes in which the subcomponent is active.
        @type     in_modes:
                    A list of strings.
        @param    default_value:
                    The default value of the subcomponent.
        @type     default_value:
                    A expressions.NumberTreeNode, expressions.BooleanTreeNode
                    or expressions.IdentifierTreeNode object.
        @param    file:
                    The file in which the subcomponent is defined.
        @type     file:
                    string
        @param    line:
                    The line in which the subcomponent was defined.
        @type     line:
                    int
        @param    pos:
                    The column position at which the subcomponent was defined.
        @type     pos:
                    int        
        """
        LocatedInformation.__init__(self, implementation, "data", file, line,
                                    pos)
        self.implementation = implementation
        self.name = name
        self.data_type = data_type
        self.in_modes = in_modes
        self.default_value = default_value

    def get_ref_identifier(self):
        """Retrieves an identifier that is used as a reference point for errors."""
        return self.parent.get_ref_identifier() + "::subcomp::" + self.name 

    def add_error(self, error):
        """Adds an error to the error collector object.
        
        @param    error:
                    The error to add.
        @type     error:
                    A subclass of errors.AbstractError.
        """
        self.parent.add_error(error)

class Binding(object):
    ACCESSES = 0
    STOREDIN = 1
    RUNNINGON = 2
    
    def str_to_binding(cls, name):
        if (name == "accesses"):
            return cls.ACCESSES
        elif (name == "stored in"):
            return cls.STOREDIN
        elif (name == "running on"):
            return cls.RUNNINGON
        else:
            assert False
            
    def binding_to_str(cls, binding):
        if (binding == cls.ACCESSES):
            return "accesses"
        elif (binding == cls.STOREDIN):
            return "stored in"
        elif (binding == cls.RUNNINGON):
            return "running on"
        else:
            assert False

    str_to_binding = classmethod(str_to_binding)
    binding_to_str = classmethod(binding_to_str)

class Category(object):
    PROCESS = 0
    THREAD_GROUP = 1
    THREAD = 2
    DATA = 3
    PROCESSOR = 4
    MEMORY = 5
    DEVICE = 6
    BUS = 7
    SYSTEM = 8
    
    def str_to_category(cls, name):
        if (name == "process"):
            return cls.PROCESS
        elif (name == "thread group"):
            return cls.THREAD_GROUP
        elif (name == "thread"):
            return cls.THREAD
        elif (name == "data"):
            return cls.DATA
        elif (name == "processor"):
            return cls.PROCESSOR
        elif (name == "memory"):
            return cls.MEMORY
        elif (name == "device"):
            return cls.DEVICE
        elif (name == "bus"):
            return cls.BUS
        elif (name == "system"):
            return cls.SYSTEM
        else:
            assert False

    def category_to_str(cls, category):
        if (category == cls.PROCESS):
            return "process"
        elif (category == cls.THREAD_GROUP):
            return "thread group"
        elif (category == cls.THREAD):
            return "thread"
        elif (category == cls.DATA):
            return "data"
        elif (category == cls.PROCESSOR):
            return "processor"
        elif (category == cls.MEMORY):
            return "memory"
        elif (category == cls.DEVICE):
            return "device"
        elif (category == cls.BUS):
            return "bus"
        elif (category == cls.SYSTEM):
            return "system"
        else:
            assert False

    str_to_category = classmethod(str_to_category)
    category_to_str = classmethod(category_to_str)

class AbstractDataType(LocatedInformation):
    """This class is the base class for all data type classes.
    """
    
    def __init__(self, file, line, pos):
        LocatedInformation.__init__(self, None, "data type", file, line, pos)
    
    def get_name(self):
        return "abstract data type"
    
    def get_node_type(self):
        return NodeType.ERROR
    
class Bool(AbstractDataType):

    def __init__(self, file, line, pos):
        AbstractDataType.__init__(self, file, line, pos)

    def get_name(self):
        return "bool"

    def get_node_type(self):
        return NodeType.BOOL

class Clock(AbstractDataType):
    def __init__(self, time_unit, file, line, pos):
        AbstractDataType.__init__(self, file, line, pos)
        assert time_unit is None or isinstance(time_unit,int)
        self.time_unit = time_unit

    def get_name(self):
        return "clock"

    def get_node_type(self):
        return NodeType.CLOCK

class Continuous(AbstractDataType):

    def __init__(self, file, line, pos):
        AbstractDataType.__init__(self, file, line, pos)

    def get_name(self):
        return "continuous"

    def get_node_type(self):
        return NodeType.CONTINUOUS

class Enum(AbstractDataType):
    
    def __init__(self, values, file, line, pos):
        AbstractDataType.__init__(self, file, line, pos)
        self.values = values

    def get_duplicates(self):
        duplicates = []
        unique_set = set(self.values)
        for elem in unique_set:
            if self.values.count(elem) > 1:
                duplicates.append(elem)
        return duplicates

    def get_name(self):
        return "enum"

    def get_node_type(self):
        return NodeType.ENUM_FIELD

class Int(AbstractDataType):
    
    def __init__(self, file, line, pos):
        AbstractDataType.__init__(self, file, line, pos)

    def get_name(self):
        return "int"

    def get_node_type(self):
        return NodeType.INT

class IntRange(AbstractDataType):
    def __init__(self, lower, upper, file, line, pos):
        AbstractDataType.__init__(self, file, line, pos)
        self.lower = lower
        self.upper = upper

    def value_is_valid(self, value):
        return (self.lower <= value and value <= self.upper)
    
    def is_compatible_with(self, otherIntRange):
        return (self.lower == otherIntRange.lower and 
                self.upper == otherIntRange.upper)

    def get_name(self):
        return "int range"
    
    def bounds_are_valid(self):
        return (self.lower < self.upper)
    
    def get_bounds(self):
        return [self.lower, self.upper]

    bounds = property(get_bounds)

    def get_node_type(self):
        return NodeType.INT_RANGE

class Real(AbstractDataType):
    
    def __init__(self, file, line, pos):
        AbstractDataType.__init__(self, file, line, pos)

    def get_name(self):
        return "real"

    def get_node_type(self):
        return NodeType.REAL

class ModeType(object):
    """Contains constants to represent the type of a mode.
    This class is used as an enum.
    """
    INITIAL = 0    # an initial mode
    ACTIVATION = 1 # an activation mode
    ERROR = 2      # an error mode
    NOMINAL = 3    # a nominal mode

class Mode(LocatedInformation):
    """A mode of the nominal behaviour of a SLIM component.
    
    The type of a mode can be one of the following:
    initial, activation, nominal
    This type must be represented by one of the constants defined in
    ModeType.
    """
    def __init__(self, impl, name, type, invariant, urgent, file, line, pos):
        LocatedInformation.__init__(self, impl, type, file, line, pos)
        self.name = name
        self.type = type
        self.invariant = invariant
        self.urgent = urgent

    def get_ref_identifier(self):
        """Retrieves an identifier that is used as a reference point for errors."""
        return self.parent.get_ref_identifier() + "::mode::" + self.name 

class ModeTransition(LocatedInformation):
    """A transition between modes of a SLIM component.
    
    A transition consists of its source state, its destination state and a
    label.
    """
    def __init__(self, impl, number, internal_number, source_mode, target_mode, label, file,
                 line, pos):
        LocatedInformation.__init__(self, impl, "transition transition", file, line, pos)
        self.source_mode = source_mode
        self.target_mode = target_mode
        self.label = label
        self.number = number
        self.internal_number = internal_number
        
    def get_ref_identifier(self):
        """Retrieves an identifier that is used as a reference point for errors."""
        return self.parent.get_ref_identifier() + "::trans::" + str(self.number) 

class ModeTransitionLabel(object):
    """A label of a mode transition, consisting of a trigger event (either
    an incoming or outgoing event), a guard (defining when the transition can
    be taken) and an effect (assigning values to data subcomponents or outgoing
    data ports).
    """
    def __init__(self, trigger, guard, time_delay, effects):
        self.trigger    = trigger
        self.guard      = guard
        self.time_delay = time_delay
        self.effects    = effects

class ReactivationModeTransitionLabel(object):
    """A label of a reactivation mode transition, consisting only
    of effects (assigning values to data subcomponents or outgoing
    data ports).
    """
    def __init__(self, effects):
        self.effects = effects

class ProbabilityModeTransitionLabel(object):
    
    def __init__(self, probability):
        self.probability = probability
