# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================
"""This module provides visitors on the expression-structure
"""

from functools import partial

from slim_compiler.src.dataStructure.expressions import OperatorType, \
    BinaryOperatorTreeNode, UnaryOperatorTreeNode, \
    NumberTreeNode, NodeType, TimeUnitType, IdentifierTreeNode, BooleanTreeNode,\
    CaseExpressionTreeNode, ExpressionTreeNode, TrajectoryEquationTreeNode
from slim_compiler.src.dataStructure.visitors.baseVisitor import BaseVisitor, \
    CompileVisitor, TransformVisitor, EvalVisitor
from slim_compiler.src.errorManagement.errors import TrajectoryEquationReferencesIllegalType, \
    ExpressionTypeMismatchError, ContinuousNotSetToRealError, \
    ClockSetToNonZeroError, ErrorFieldsIncomparableError, \
    ModeFieldsIncomparableError, EnumsIncomparableError, \
    ErrorStateValueIllegalError, ModeValueIllegalError, EnumValueIllegalError, \
    UnaryExpressionTypeMismatchError, CaseConditionNotOfTypeBoolError, \
    CaseExpressionsHaveMultipleTypesError, \
    CaseExpressionsHaveDifferentValidEnumsError, \
    CaseExpressionsHaveDifferentValidEnumValuesError, ExpressionIsNonLinearError,\
    TrajectoryExpressionNotScaledInTimeScaledModel
from slim_compiler.src.dataStructure import expressions
from utils.exceptions import InternalException
from fractions import Fraction

class Sentinel(NumberTreeNode):
    """This class is a simple marker in an expression tree to replace
    case nodes for SimplifyCaseVisitor.
    Note: inherits from NumberTreeNode for debugging purposes"""
    def __init__(self, node_type):
        super(Sentinel, self).__init__("0", None, 0, 0)
        self.number = "#"
        self.node_type = node_type

    def clone(self):
        return Sentinel(self.node_type)

class SimplifyCaseVisitor(object):
    """This class simplifies case expressions to plain boolean
    expressions using implications and conjunctions. The expression
    to be simplified is required to be of Boolean type."""
    def __init__(self):
        super(SimplifyCaseVisitor, self).__init__()

        # These variables are markers in the expression tree to
        # note where to replace what
        self.top_node = None
        self.case_node = None
        self.lastboolnode = None
        self.binary_left = False

    def replace_case(self, expr):
        """Returns expr with replaced case nodes.
        Replacement happens inline, but expr may be new node
        if case was top node"""
        assert expr.node_type == NodeType.BOOL, repr(expr) + " is not boolean"

        replaced = True
        while replaced is True:
            self.top_node = None
            self.case_node = None
            self.lastboolnode = None
            self.binary_left = False
            (expr, replaced) = self.place_sentinel(expr)
            if replaced:
                if self.top_node is None:
                    # Possible if case node is top node
                    self.top_node = expr
                bool_case_node = self.generate_cases(self.top_node, self.case_node)
                # If case as subexpression of bool expression: place as child
                if self.lastboolnode is not None:
                    # Case was subexpression of boolean expression, simply insert
                    if isinstance(self.lastboolnode, UnaryOperatorTreeNode):
                        self.lastboolnode.subexpression = bool_case_node
                    elif isinstance(self.lastboolnode, BinaryOperatorTreeNode):
                        if self.binary_left:
                            self.lastboolnode.left = bool_case_node
                        else:
                            self.lastboolnode.right = bool_case_node
                else:
                    # Case was (sub) expression of arithmetic expression only, replace top node
                    expr = bool_case_node
        return expr

    def generate_cases(self, top_node, case_node):
        """When transforming case expressions, express case as follows:
          case b1 : e1; b2 : e2 otherwise e3 end
          ((b1 -> e1) AND (-b1 AND b2 -> e2) AND (-b1 AND -b2 -> e3))

        top_node is the (sub)expression which will be cloned for each case
        top_node MUST contain sentinel
        case_node is the node previously contained within top_node and replaced
        by sentinel

        Note: this significantly increases expression size
        """
        conjunction = None # Top level conjunction
        conditions = [cond for (cond, _) in case_node.cases]
        # build (a -> b) AND (c -> d)..
        for i, (_, expr) in enumerate(case_node.cases):
            # build (-b1 AND .. -bj AND bi) -> ei
            cond = conditions[i].clone()
            for j in range(0, i):
                not_cond = self.make_not(conditions[j].clone())
                cond = self.make_bin(cond, not_cond, OperatorType.AND)

            impl = self.make_bin(cond, top_node.clone(), OperatorType.IMP)
            conjunction = self.make_bin(conjunction, impl, OperatorType.AND)
            # Replace the sentinel with the current case expression
            impl.right = self.replace_sentinel(impl.right, expr)

        # Build final condition for otherwise
        other_cond = None
        for cond in conditions:
            neg_cond = self.make_not(cond.clone())
            other_cond = self.make_bin(other_cond, neg_cond, OperatorType.AND)
        # Note: Not cloning top_node, this is the final use
        other_impl = self.make_bin(other_cond, top_node, OperatorType.IMP)
        conjunction = self.make_bin(conjunction, other_impl, OperatorType.AND)
        # Replace the sentinel with the otherwise case expression
        other_impl.right = self.replace_sentinel(other_impl.right, case_node.default_expression)

        return conjunction

    def place_sentinel(self, expr):
        """Returns tuple with expr with optional replaced sentinel
        and boolean indicating if replacement occurred
        Places at most one sentinel"""
        if isinstance(expr, CaseExpressionTreeNode):
            self.case_node = expr
            return (Sentinel(expr.node_type), True)

        if isinstance(expr, BinaryOperatorTreeNode):
            if expr.node_type == NodeType.BOOL and expr.left.node_type == NodeType.BOOL:
                self.lastboolnode = expr
                self.binary_left = True
                self.top_node = expr.left
            (expr.left, placed) = self.place_sentinel(expr.left)
            if placed:
                return (expr, True)

            if expr.node_type == NodeType.BOOL and expr.right.node_type == NodeType.BOOL:
                self.lastboolnode = expr
                self.binary_left = False
                self.top_node = expr.right
            (expr.right, placed) = self.place_sentinel(expr.right)
            if placed:
                return (expr, True)
        elif isinstance(expr, UnaryOperatorTreeNode):
            if expr.node_type == NodeType.BOOL and expr.subexpression.node_type == NodeType.BOOL:
                self.lastboolnode = expr
                self.top_node = expr.subexpression
            (expr.subexpression, placed) = self.place_sentinel(expr.subexpression)
            if placed:
                return (expr, True)

        return (expr, False)

    def replace_sentinel(self, expr, value):
        """Looks for sentinel in expr and replaces it
        with value inline. Returns the new expression"""
        if isinstance(expr, Sentinel):
            return value

        if isinstance(expr, BinaryOperatorTreeNode):
            expr.left = self.replace_sentinel(expr.left, value)
            expr.left.parent = expr
            expr.right = self.replace_sentinel(expr.right, value)
            expr.right.parent = expr
        elif isinstance(expr, UnaryOperatorTreeNode):
            expr.subexpression = self.replace_sentinel(expr.subexpression, value)
            expr.subexpression.parent = expr

        return expr

    def make_not(self, subexpr):
        return UnaryOperatorTreeNode.createNode(subexpr, OperatorType.NOT, False)

    def make_bin(self, left, right, operator):
        return BinaryOperatorTreeNode.createNode(left, right, operator, False)

class InvariantToNNFVisitor(BaseVisitor):
    """This class provides visit-methods for the expression-structure.
    The visit-methods return an equivalent expression structure which
    uses a reduced set of boolean operators (NOT, AND, OR) and the
    formula is in Negated Normal Form.
    Operators are ordered in the result tree as
      Boolean <- Relational <- Arithmetic
    Note: Expressions are modified in place.
    Note2: Expression must be type checked and of boolean type.
    Note3: Expression may not contain case expressions, simplify the expression first
    """
    def __init__(self, to_cnf=True):
        super(InvariantToNNFVisitor, self).__init__()
        self.to_cnf = to_cnf

    def visit_ExpressionTreeNode(self, root_node):
        return root_node

    def visit_BinaryOperatorTreeNode(self, root_node):
        # A & B == simple(A) & simple(B)
        # DNF: A | B == simple(A) | simple(B)
        if (root_node.operator == OperatorType.AND and self.to_cnf) or \
           (root_node.operator == OperatorType.OR and not self.to_cnf):
            root_node.left = self.visit(root_node.left)
            root_node.right = self.visit(root_node.right)
            return root_node
        # A | B == and(simple(A) X simple(B))
        # DNF: A & B == or(simple(A) X simple(B))
        elif (root_node.operator == OperatorType.OR and self.to_cnf) or \
             (root_node.operator == OperatorType.AND and not self.to_cnf):
            left = self.visit(root_node.left)
            right = self.visit(root_node.right)

            if self.to_cnf:
                clauses_left = self.get_clauses(left, OperatorType.AND)
                clauses_right = self.get_clauses(right, OperatorType.AND)
            else:
                clauses_left = self.get_clauses(left, OperatorType.OR)
                clauses_right = self.get_clauses(right, OperatorType.OR)

            conjunction = None
            for left_clause in clauses_left:
                for right_clause in clauses_right:
                    if self.to_cnf:
                        left_or_right = BinaryOperatorTreeNode.create_or(left_clause, right_clause)
                        conjunction = BinaryOperatorTreeNode.create_and(conjunction, left_or_right)
                    else:
                        left_and_right = BinaryOperatorTreeNode.create_and(left_clause, right_clause)
                        conjunction = BinaryOperatorTreeNode.create_or(conjunction, left_and_right)

            return conjunction

        # Normalize other operators into AND, OR and NOT
        # A imp B == simple(~A or B)
        elif root_node.operator == OperatorType.IMP:
            root_node.left = self.make_not(root_node.left)
            root_node.left.parent = root_node
            root_node.operator = OperatorType.OR
            return self.visit(root_node)
        # A iff B == A xnor B == simple((~A & ~B) | (A & B))
        elif root_node.operator == OperatorType.IFF or root_node.operator == OperatorType.XNOR \
             or (root_node.operator == OperatorType.EQUAL and root_node.left.node_type == NodeType.BOOL):
            not_a = self.make_not(root_node.left.clone())
            not_b = self.make_not(root_node.right.clone())
            not_a_and_not_b = BinaryOperatorTreeNode.create_and(not_a, not_b)
            a_and_b = BinaryOperatorTreeNode.create_and(root_node.left, root_node.right)

            root_node.left = not_a_and_not_b
            root_node.right = a_and_b
            root_node.left.parent = root_node
            root_node.right.parent = root_node
            root_node.operator = OperatorType.OR

            return self.visit(root_node)
        # A xor B == simple((~A & B) | (A & ~B))
        elif root_node.operator == OperatorType.XOR \
             or (root_node.operator == OperatorType.NOTEQUAL and root_node.left.node_type == NodeType.BOOL):
            not_a = self.make_not(root_node.left.clone())
            not_b = self.make_not(root_node.right.clone())
            not_a_and_b = BinaryOperatorTreeNode.create_and(not_a, root_node.right)
            a_and_not_b = BinaryOperatorTreeNode.create_and(root_node.left, not_b)

            root_node.left = not_a_and_b
            root_node.right = a_and_not_b
            root_node.left.parent = root_node
            root_node.right.parent = root_node
            root_node.operator = OperatorType.OR

            return self.visit(root_node)

        # Otherwise just return the node
        else:
            return root_node

    def visit_UnaryOperatorTreeNode(self, root_node):
        if root_node.operator == OperatorType.NOT \
           and isinstance(root_node.subexpression, (BinaryOperatorTreeNode, UnaryOperatorTreeNode)):
            # Push negation inwards
            sub_op = root_node.subexpression.operator

            # ~(A & B) == simple(~A | ~B)
            if sub_op == OperatorType.AND:
                root_node.subexpression.left = self.make_not(root_node.subexpression.left)
                root_node.subexpression.right = self.make_not(root_node.subexpression.right)
                root_node.subexpression.left.parent = root_node.subexpression
                root_node.subexpression.right.parent = root_node.subexpression
                root_node.subexpression.operator = OperatorType.OR
                return self.visit(root_node.subexpression)
            # ~(A | B) == simple(~A & ~B)
            elif sub_op == OperatorType.OR:
                root_node.subexpression.left = self.make_not(root_node.subexpression.left)
                root_node.subexpression.right = self.make_not(root_node.subexpression.right)
                root_node.subexpression.left.parent = root_node.subexpression
                root_node.subexpression.right.parent = root_node.subexpression
                root_node.subexpression.operator = OperatorType.AND
                return self.visit(root_node.subexpression)
            # ~(A imp B) == simple(A & ~B)
            elif sub_op == OperatorType.IMP:
                root_node.subexpression.right = self.make_not(root_node.subexpression.right)
                root_node.subexpression.right.parent = root_node.subexpression
                root_node.subexpression.operator = OperatorType.AND
                return self.visit(root_node.subexpression)
            # ~(A iff B) == ~(A xnor B) == simple(A xor B)
            elif sub_op == OperatorType.IFF or sub_op == OperatorType.XNOR:
                root_node.subexpression.operator = OperatorType.XOR
                return self.visit(root_node.subexpression)
            # ~(A xor B) == simple(A xnor B)
            elif sub_op == OperatorType.XOR:
                root_node.subexpression.operator = OperatorType.XNOR
                return self.visit(root_node.subexpression)
            # ~(~A) == simple(A)
            elif sub_op == OperatorType.NOT:
                return self.visit(root_node.subexpression.subexpression)

            # Can both be comparing bool or numeric expressions
            elif sub_op == OperatorType.EQUAL:
                root_node.subexpression.operator = OperatorType.NOTEQUAL
                return self.visit(root_node.subexpression)
            elif sub_op == OperatorType.NOTEQUAL:
                root_node.subexpression.operator = OperatorType.EQUAL
                return self.visit(root_node.subexpression)

            # Comparing numeric expressions, flip operator
            elif sub_op == OperatorType.LESS:
                root_node.subexpression.operator = OperatorType.GREATEROREQUAL
                return self.visit(root_node.subexpression)
            elif sub_op == OperatorType.LESSOREQUAL:
                root_node.subexpression.operator = OperatorType.GREATER
                return self.visit(root_node.subexpression)
            elif sub_op == OperatorType.GREATER:
                root_node.subexpression.operator = OperatorType.LESSOREQUAL
                return self.visit(root_node.subexpression)
            elif sub_op == OperatorType.GREATEROREQUAL:
                root_node.subexpression.operator = OperatorType.LESS
                return self.visit(root_node.subexpression)

            # Ought not to occur
            else:
                assert False, "Invalid NOT expression in NNF visitor"
        else:
            # Either not literal or arithmetic minus
            return root_node

    def visit_CaseExpressionTreeNode(self, root_node):
        assert False, "Case expression not simplified for NNF"

    def make_not(self, subexpr):
        return UnaryOperatorTreeNode.createNode(subexpr, OperatorType.NOT, False)

    def get_clauses(self, root_node, operator):
        """Returns the list of clauses for conjunction specified by root_node.
        E.g. [a ^ b ^ c ^ d]
        For DNF it is disjunction"""
        if isinstance(root_node, BinaryOperatorTreeNode) and root_node.operator == operator:
            return self.get_clauses(root_node.left, operator) + self.get_clauses(root_node.right, operator)
        else:
            return [root_node]

class SimplifyExprVisitor(TransformVisitor):
    """Simplify trivial subexpressions"""
    def __init__(self):
        super(SimplifyExprVisitor, self).__init__()
        # Import here to avoid cyclic dependency
        from slim_compiler.src.dataStructure.compile import CompileInformation
        self.evaluate = ExpressionValueVisitor(CompileInformation(), False)

    def _make_result_node(self, root_node, result):
        # Either bool, int. float
        if result is True or result is False:
            bool_node = BooleanTreeNode(result, root_node.line, root_node.pos)
            bool_node.file = root_node.file
            return bool_node
        else:
            if isinstance(result, Fraction):
                num_node = NumberTreeNode(str(result.numerator), None, root_node.line, root_node.pos)
                num_node.file = root_node.file
                # Hard cast to real, as to avoid a real division being changed to integer
                num_node.node_type = NodeType.REAL_CONST
                
                denom_node = NumberTreeNode(str(result.denominator), None, root_node.line, root_node.pos)
                denom_node.file = root_node.file
                # Hard cast to real, as to avoid a real division being changed to integer
                denom_node.node_type = NodeType.REAL_CONST

                return BinaryOperatorTreeNode.createNode(num_node, denom_node, OperatorType.DIVIDE, False)
            # Must be numeric
            num_node = NumberTreeNode(repr(result), None, root_node.line, root_node.pos)
            num_node.file = root_node.file
            return num_node

        return root_node

    def action_UnaryOperatorTreeNode(self, root_node, visit_node):
        # First try to evaluate this node
        result = self.evaluate.visit(root_node)
        if result is not None:
            return self._make_result_node(root_node, result)

        if isinstance(root_node.subexpression, UnaryOperatorTreeNode) and root_node.subexpression.operator == root_node.operator:
            return root_node.subexpression.subexpression

        return root_node

    def action_BinaryOperatorTreeNode(self, root_node, visit_node):
        # First try to evaluate this node
        result = self.evaluate.visit(root_node)
        if result is not None:
            return self._make_result_node(root_node, result)

        result_l = self.evaluate.visit(root_node.left)
        result_r = self.evaluate.visit(root_node.right)

        if root_node.operator == OperatorType.AND:
            if result_l is False or result_r is False:
                return self._make_result_node(root_node, False)
            if result_l is True:
                return root_node.right
            if result_r is True:
                return root_node.left
        elif root_node.operator == OperatorType.OR:
            if result_l is True or result_r is True:
                return self._make_result_node(root_node, True)
            if result_l is False:
                return root_node.right
            if result_r is False:
                return root_node.left

        elif root_node.operator == OperatorType.PLUS:
            if result_l == 0:
                return root_node.right
            if result_r == 0:
                return root_node.left
            if isinstance(root_node.right, UnaryOperatorTreeNode) and root_node.right.operator == OperatorType.UNARY_MINUS:
                root_node.operator = OperatorType.MINUS
                root_node.right.subexpression.parent = root_node
                root_node.right = root_node.right.subexpression
            elif isinstance(root_node.left, UnaryOperatorTreeNode) and root_node.left.operator == OperatorType.UNARY_MINUS:
                root_node.operator = OperatorType.MINUS
                root_node.left.subexpression.parent = root_node
                root_node.left = root_node.left.subexpression
        elif root_node.operator == OperatorType.MINUS:
            if result_r == 0:
                return root_node.left
            if isinstance(root_node.right, UnaryOperatorTreeNode) and root_node.right.operator == OperatorType.UNARY_MINUS:
                root_node.operator = OperatorType.PLUS
                root_node.right.subexpression.parent = root_node
                root_node.right = root_node.right.subexpression
        elif root_node.operator == OperatorType.TIMES:
            if result_l == 0 or result_r == 0:
                return self._make_result_node(root_node, 0)
            if result_l == 1:
                return root_node.right
            if result_r == 1:
                return root_node.left
        elif root_node.operator == OperatorType.DIVIDE:
            if result_r == 0:
                # Division by zero, keep it so it gets caught in the future
                return root_node
            if result_l == 0:
                return self._make_result_node(root_node, 0)

        elif root_node.operator == OperatorType.MOD:
            if result_r == 1:
                return self._make_result_node(root_node, 1)

        return root_node


class SimplifyLinEq(BaseVisitor):
    """Builds linear equations in the form AX + BY ... CZ `op` constant
    Input must be a linear equation, where variables are clock or continuous
    and in NNF form.
    Trivial expressions must have been simplified
    Root node must be a relational or boolean operator
    Removes all trajectory equations as well"""

    def build(self, root_node):
        """Build the linear expression starting from this node"""
        if self.is_constant(root_node):
            return root_node

        if isinstance(root_node, BinaryOperatorTreeNode):
            if root_node.operator in [OperatorType.AND, OperatorType.OR]:
                # Recursive until a numeric expression is found
                left = self.build(root_node.left)
                right = self.build(root_node.right)
                if left is None:
                    return right
                if right is None:
                    return left
                root_node.left = left
                root_node.left.parent = root_node
                root_node.right = right
                root_node.right.parent = root_node
                return root_node
            else:
                # Build the lin exp for the relational operator
                return self.build_rel(root_node)
        elif isinstance(root_node, UnaryOperatorTreeNode):
            # This should not occur in NNF
            assert False, "Unary node in NNF expression"
        elif isinstance(root_node, TrajectoryEquationTreeNode):
            # Remove trajectory equations
            return None
        else:
            assert False, "Not NNF expression %s" % str(root_node)

    def build_rel(self, root_node):
        """Build linear expression over relational expression (e.g. left < right)"""
        if self.is_constant(root_node):
            return root_node

        # Map of variables names to their original nodes
        id_map = {}

        # Get the mapping of each variable to their factors
        lhs_eq = self.get_eq(root_node.left, id_map)
        rhs_eq = self.get_eq(root_node.right, id_map)

        # Process terms: move all variables to the LHS by means of subtraction
        # 1*x < 2*y == 1*x - 2*y < 0
        for key in rhs_eq:
            if key is None:
                continue
            if not key in lhs_eq:
                lhs_eq[key] = self.make_un(rhs_eq[key], OperatorType.UNARY_MINUS)
            else:
                lhs_eq[key] = self.make_bin(lhs_eq[key], rhs_eq[key], OperatorType.MINUS)

        # Move the constant on the LHS to the RHS by means of subtraction
        # 2*x + 3 < 0 == 2*x < -3
        if None in lhs_eq:
            if not None in rhs_eq:
                rhs_eq[None] = self.make_un(lhs_eq[None], OperatorType.UNARY_MINUS)
            else:
                rhs_eq[None] = self.make_bin(rhs_eq[None], lhs_eq[None], OperatorType.MINUS)
            lhs_eq.pop(None, None)

        # If there is no constant, insert '0'
        if None in rhs_eq:
            rhs_eq = {None: rhs_eq[None]}
        else:
            rhs_eq = {None: NumberTreeNode("0", None, 0, 0)}

        # Build the equations from the term map
        root_node.left = self.build_eq(lhs_eq, id_map)
        root_node.right = self.build_eq(rhs_eq, id_map)

        return root_node

    def build_eq(self, cur_eq, id_map):
        """Construct SLIM expression given the cur_eq term map
        For each key, construct a term"""
        expr = None
        for key in cur_eq:
            if key is None:
                continue
            id_node = id_map[key]
            factor = SimplifyExprVisitor().visit(cur_eq[key])
            factor_node = self.make_bin(factor, id_node, OperatorType.TIMES)
            expr = self.make_bin(expr, factor_node, OperatorType.PLUS)
        if None in cur_eq:
            constant = SimplifyExprVisitor().visit(cur_eq[None])
            expr = self.make_bin(expr, constant, OperatorType.PLUS)
        return expr

    def get_eq(self, root_node, id_map):
        """Construct a map of variables to factors, in the form
        {name : expr,
         name2: expr,
         None: constant_expr}, where name is a literal,
         which is added to id_map"""
        if self.is_constant(root_node):
            return self.constant(root_node, id_map)

        if isinstance(root_node, UnaryOperatorTreeNode):
            # Only unary minus is possible at this point
            # Negate all inner expressions
            return self.minus(root_node, id_map)
        if isinstance(root_node, BinaryOperatorTreeNode):
            if root_node.operator in [OperatorType.TIMES, OperatorType.DIVIDE]:
                # A division is multiplication with inverse, so scale can be used
                return self.scale(root_node, id_map)
            elif root_node.operator in [OperatorType.PLUS, OperatorType.MINUS]:
                # Minus is plus with negated value
                return self.offset(root_node, id_map)
            elif root_node.operator in [OperatorType.MOD]:
                # For constant expression, mod can be used
                return self.mod(root_node, id_map)
        if isinstance(root_node, IdentifierTreeNode):
            # Non-constant identifier must be literal
            return self.literal(root_node, id_map)
        assert False, "Linear expression contains unsupported non-constant expression: %s"%str(root_node)

    def literal(self, root_node, id_map):
        """Construct a literal in the equation map and insert into id_map"""
        if not root_node.name in id_map:
            id_map[root_node.name] = root_node
        return {root_node.name: NumberTreeNode("1", None, 0, 0)}

    def constant(self, root_node, id_map):
        """Insert as constant expression in the equation map"""
        return {None: root_node}

    def scale(self, root_node, id_map):
        """Scale all the items in the equation map by given (constant) expression"""
        # Operator TIMES and DIVIDE
        cur_eq_left = self.get_eq(root_node.left, id_map)
        cur_eq_right = self.get_eq(root_node.right, id_map)

        constant_left = self.is_constant(root_node.left)

        assert root_node.operator != OperatorType.DIVIDE or (not constant_left), "Nonlinear expression "%str(root_node)
        assert constant_left or self.is_constant(root_node.right), "Nonlinear expression "%str(root_node)

        # At least the left or right is constant
        if constant_left:
            for key in cur_eq_right:
                cur_eq_right[key] = self.make_bin(cur_eq_left[None], cur_eq_right[key], root_node.operator)
            return cur_eq_right
        else:
            for key in cur_eq_left:
                cur_eq_left[key] = self.make_bin(cur_eq_left[key], cur_eq_right[None], root_node.operator)
            return cur_eq_left

    def offset(self, root_node, id_map):
        """Add or subtract terms on the RHS to the LHS"""
        # Operator PLUS and MIN
        cur_eq_left = self.get_eq(root_node.left, id_map)
        cur_eq_right = self.get_eq(root_node.right, id_map)

        for key in cur_eq_right:
            if not key in cur_eq_left:
                # If only on the RHS, simply isert the (possibly negated) expression
                if root_node.operator == OperatorType.PLUS:
                    cur_eq_left[key] = cur_eq_right[key]
                else: # MIN
                    cur_eq_left[key] = self.make_un(cur_eq_right[key], OperatorType.UNARY_MINUS)
            else:
                # Else, merge using the given oeprator
                cur_eq_left[key] = self.make_bin(cur_eq_left[key], cur_eq_right[key], root_node.operator)

        return cur_eq_left

    def mod(self, root_node, id_map):
        """mod can only be performed on constant expression, so simply return as-is"""
        assert self.is_constant(root_node.left) and self.is_constant(root_node.right)
        return {None: root_node}

    def minus(self, root_node, id_map):
        """Unary minus operator: Negate all the terms"""
        # Operator MINUS
        cur_eq = self.get_eq(root_node.subexpression, id_map)

        for key in cur_eq:
            cur_eq[key] = self.make_un(cur_eq[key], OperatorType.UNARY_MINUS)

        return cur_eq

    def is_constant(self, root_node):
        """Checks if the expression is constant. That is, it does not contain
        any variables that are of hybrid type"""
        if isinstance(root_node, BinaryOperatorTreeNode):
            return self.is_constant(root_node.left) and self.is_constant(root_node.right)
        if isinstance(root_node, UnaryOperatorTreeNode):
            return self.is_constant(root_node.subexpression)
        if isinstance(root_node, IdentifierTreeNode):
            if root_node.node_type in [NodeType.CLOCK, NodeType.CONTINUOUS]:
                return False
            else:
                return True
        if isinstance(root_node, (NumberTreeNode, BooleanTreeNode)):
            return True

        assert False, "Unexpected node in expression tree: " + str(type(root_node))

    def make_un(self, subexpr, operator):
        return UnaryOperatorTreeNode.createNode(subexpr, operator, False)

    def make_bin(self, left, right, operator):
        return BinaryOperatorTreeNode.createNode(left, right, operator, False)

class Point(object):
    def __init__(self, value, inc=0):
        self.value = value
        self.inc = inc

    def __repr__(self, *args, **kwargs):
        return str(self.value)
    def __eq__(self, other):
        return (self.value, self.inc) == (other.value, other.inc)
    def __ne__(self, other):
        return (self.value, self.inc) != (other.value, other.inc)
    def __lt__(self, other):
        return (self.value, self.inc) < (other.value, other.inc)
    def __gt__(self, other):
        return (self.value, self.inc) > (other.value, other.inc)
    def __le__(self, other):
        return (self.value, self.inc) <= (other.value, other.inc)
    def __ge__(self, other):
        return (self.value, self.inc) >= (other.value, other.inc)
    def copy(self):
        return Point(self.value, self.inc)

class Range(object):
    def __init__(self, lower, upper):
        self.lower = lower
        self.upper = upper
    def __repr__(self, *args, **kwargs):
        result = "[" if self.lower.inc == 0 else "("
        result += str(self.lower) + ", " + str(self.upper)
        result += "]" if self.upper.inc == 0 else ")"
        return result

    def intersection(self, other, inline=False):
        assert self.intersects(other)
        if not inline:
            result = Range(max(self.lower, other.lower).copy(), min(self.upper, other.upper).copy)
            return result
        else:
            self.lower = max(self.lower, other.lower).copy()
            self.upper = min(self.upper, other.upper).copy()

    def union(self, other, inline=False):
        assert self.unifies(other)
        if not inline:
            result = Range(min(self.lower, other.lower).copy(), max(self.upper, other.upper).copy)
            return result
        else:
            self.lower = min(self.lower, other.lower).copy()
            self.upper = max(self.upper, other.upper).copy()

    def intersects(self, other):
        return (self.lower <= other.upper and self.upper >= other.lower) or (self.upper >= other.lower and self.lower <= other.upper)
    def unifies(self, other):
        return self.intersects(other) or ((self.lower.value == other.upper.value) and (self.lower.inc == 0 or other.upper.inc == 0)) \
            or ((self.upper.value == other.lower.value) and (self.lower.inc == 0 or other.upper.inc == 0))

    def copy(self):
        return Range(self.lower.copy(), self.upper.copy())

class Interval(object):
    def __init__(self, lower = float("-inf"), upper = float("inf"), lower_inc = True, upper_inc = True):
        self.ranges = [ Range(Point(lower, 0 if lower_inc else 1), Point(upper, 0 if upper_inc else -1)) ]

    def is_convex(self):
        return len(self.ranges) <= 1

    def __repr__(self, *args, **kwargs):
        return ' U '.join(map(str,self.ranges))

    def intersect(self, other):
        new_ranges = []
        i_self = 0
        i_other = 0
        
        while i_self < len(self.ranges) and i_other < len(other.ranges):
            if self.ranges[i_self].intersects(other.ranges[i_other]):
                new_ranges.append(self.ranges[i_self].intersection(other.ranges[i_other]))
            if self.ranges[i_self].upper <= other.ranges[i_other].upper:
                i_self += 1
            else:
                i_other += 1

        interval = Interval()
        interval.ranges = new_ranges
        return interval

    def union(self, other):
        new_ranges = []
        i_self = 0
        i_other = 0

        union = None
        while i_self < len(self.ranges) and i_other < len(other.ranges):
            if union is None:
                if self.ranges[i_self].lower <= other.ranges[i_other].lower:
                    union = self.ranges[i_self].copy()
                    i_self += 1
                else:
                    union = other.ranges[i_other].copy()
                    i_other += 1
                new_ranges.append(union)
                continue

            if self.ranges[i_self].lower <= union.upper:
                union.union(self.ranges[i_self], True)
                i_self += 1
            elif other.ranges[i_other].lower <= union.upper:
                union.union(self.other[i_other], True)
                i_other += 1
            else:
                union = None

        while i_self < len(self.ranges):
            if union.unifies(self.ranges[i_self]):
                union.union(self.ranges[i_self], True)
            else:
                union = self.ranges[i_self]
                new_ranges.append(union)
            i_self += 1

        while i_other < len(other.ranges):
            if union.unifies(other.ranges[i_other]):
                union.union(other.ranges[i_other], True)
            else:
                union = other.ranges[i_other]
                new_ranges.append(union)
            i_other += 1

        interval = Interval()
        interval.ranges = new_ranges
        return interval

class InvariantIsConvexVisitor(EvalVisitor):
    """This class provides visit-methods for the expression-structure.
    The visit-methods check whether the invariant (or guard) is convex.
    
    Note that this visitor assumes that the expression is given in DNF,
    and simplified to remove trivial expressions
    This can be achieved by invoking InvariantToNNFVisitor(False) prior.
    """
    def action_ExpressionTreeNode(self, root_node, data):
        return True

    def action_BinaryOperatorTreeNode(self, root_node, data):
        return data.is_convex

    def visit_BinaryOperatorTreeNode(self, root_node, data, *args, **kwargs):
        # Overriding visit, form here on check_clause does the work recursively
        #assert root_node.operator in [OperatorType.OR, OperatorType.AND], "Expression not in DNF"

        data.is_convex = True
        intervals = self.check_clause(root_node)
        for key in intervals:
            data.is_convex &= intervals[key].is_convex()

    def check_clause(self, root_expr):
        if isinstance(root_expr, BinaryOperatorTreeNode):
            if root_expr.operator == OperatorType.OR or root_expr.operator == OperatorType.AND:
                left_range = self.check_clause(root_expr.left)
                right_range = self.check_clause(root_expr.right)
                # Generate disjunction of ranges
                for key in right_range:
                    if key not in left_range:
                        left_range[key] = right_range[key]
                    else:
                        if root_expr.operator == OperatorType.OR:
                            left_range[key] = left_range[key].union(right_range[key])
                        else:
                            left_range[key] = left_range[key].intersect(right_range[key])
                return left_range
            else:
                # Must be relational operator, determine range
                linear_expr = SimplifyLinEq().build(root_expr.clone())
                return self.get_rel_range(linear_expr);
        else:
            # Boolean literal or not literal, no region
            return {}

    def get_rel_range(self, root_expr):
        # Must be (aX `op` Const)
        if not root_expr.left.node_type in [NodeType.CLOCK, NodeType.CONTINUOUS]:
            return {}
        if root_expr.left.operator != OperatorType.TIMES:
            return {}
        constant = ExpressionValueVisitor(None, False).visit(root_expr.right)
        factor = ExpressionValueVisitor(None, False).visit(root_expr.left.left)
        var = root_expr.left.right.name
        if constant is None or factor is None:
            return {}
        else:
            constant = float(constant) / factor
            op = root_expr.operator
            if factor < 0:
                if root_expr.operator == OperatorType.LESS:
                    op = OperatorType.GREATER
                elif root_expr.operator == OperatorType.LESSOREQUAL:
                    op = OperatorType.GREATEROREQUAL
                elif root_expr.operator == OperatorType.GREATER:
                    op = OperatorType.LESS
                elif root_expr.operator == OperatorType.GREATEROREQUAL:
                    op = OperatorType.LESSOREQUAL

            if op == OperatorType.EQUAL:
                return {var: Interval(constant, constant)}
            elif op == OperatorType.NOTEQUAL:
                return {var: Interval(upper=constant, upper_inc=False).union(Interval(lower=constant, lower_inc=False))}

            elif op == OperatorType.LESS:
                return {var: Interval(upper=constant, upper_inc=False)}
            elif op == OperatorType.LESSOREQUAL:
                return {var: Interval(upper=constant)}

            elif op == OperatorType.GREATER:
                return {var: Interval(lower=constant, lower_inc=False)}
            elif op == OperatorType.GREATEROREQUAL:
                return {var: Interval(lower=constant)}
        assert False

class AnnotateVisitor(CompileVisitor):
    """Visitor to annotate Identifier nodes with correct type and reference.
    Note: this class bypasses normal visit patterns, as only expressions
    are interesting"""
    def __init__(self, compile_info):
        super(AnnotateVisitor, self).__init__()

        # For error management
        self.compile_info = compile_info

        self.reset()

    def reset(self):
        self.annotation_meth = None

    def visit_ComponentType(self, node):
        for port in node.in_data_ports.values():
            if port.default_value is not None:
                self.annotation_meth = partial(node.get_data_element_type_dataport_default_value,
                                               port = port)
                self.visit(port.default_value)
        for port in node.out_data_ports.values():
            if port.default_value is not None:
                self.annotation_meth = partial(node.get_data_element_type_dataport_default_value,
                                               port = port)
                self.visit(port.default_value)

        self.reset()

    def visit_ComponentImplementation(self, node):
        # Import here to stop cyclic dependency
        from slim_compiler.src.dataStructure.nominalbehaviour import ModeTransitionLabel, ReactivationModeTransitionLabel

        # Data subcomponents
        for data in node.data_subcomponents.values():
            if data.default_value is not None:
                self.annotation_meth = partial(node.get_data_element_type_datasubcomponent_default_value,
                                               data = data)
                self.visit(data.default_value)

        self.reset()

        # Flows
        for flow in node.flows:
            self.annotation_meth = partial(node.get_data_element_type_lhs_flow,
                                           flow = flow, in_modes = flow.in_modes)
            self.visit(flow.expression.left)
            self.annotation_meth = partial(node.get_data_element_type_rhs_flow,
                                           flow = flow, in_modes = flow.in_modes)
            self.visit(flow.expression.right)

        self.reset()

        # Invariants
        for mode in node.modes.values():
            if mode.invariant is not None:
                self.annotation_meth = partial(node.get_data_element_type_invariant,
                                               mode = mode.name, mode_obj=mode)
                self.visit(mode.invariant)

        self.reset()

        # Guards and effects
        for trans in node.normal_mode_transitions + node.reactivation_mode_transitions:
            if isinstance(trans.label, ModeTransitionLabel):
                # guard
                if trans.label.guard is not None:
                    self.annotation_meth = partial(node.get_data_element_type_guard,
                                                   source_mode = trans.source_mode, trans = trans)
                    self.visit(trans.label.guard)

            if isinstance(trans.label, (ModeTransitionLabel, ReactivationModeTransitionLabel)):
                # effects
                effects = trans.label.effects
                for effect in effects:
                    self.annotation_meth = partial(node.get_data_element_type_lhs_effect,
                                                   target_mode = trans.target_mode, trans = trans)
                    self.visit(effect.left)

                    self.annotation_meth = partial(node.get_data_element_type_rhs_effect,
                                                   source_mode = trans.source_mode, trans = trans)
                    self.visit(effect.right)

        self.reset()
        # Ignoring other nodes under ComponentImplementation
        pass

    def visit_ErrorModelImplementation(self, node):
        # Invariants
        for state in node.states.values():
            if state.invariant is not None:
                self.annotation_meth = partial(node.get_data_element_type_invariant,
                                               state = state)
                self.visit(state.invariant)

        self.reset()

        # Guards and effects
        for trans in node.error_transitions + node.reactivation_transitions:
            if trans.guard is not None:
                self.annotation_meth = partial(node.get_data_element_type_guard,
                                               trans = trans)
                self.visit(trans.guard)

            for effect in trans.effects:
                self.annotation_meth = partial(node.get_data_element_type_effect,
                                               trans = trans)
                self.visit(effect.left)
                # Same annotation method for right, the scope is identical (just clocks)
                self.visit(effect.right)

        self.reset()
        # Ignoring other nodes under ErrorModelImplementation
        pass

    def action_IdentifierTreeNode(self, node, mode = None):
        assert self.annotation_meth != None, "Annotation method not set for annotation visitor for identifier %s"%node.name

        (type, ref, add_info) = self.annotation_meth(name = node.name, line = node.line, pos = node.pos)

        if type == NodeType.ERROR:
            node.node_type = NodeType.ERROR
            return

        node.ref = ref

        if type == NodeType.INT or type == NodeType.BOOL or \
           type == NodeType.REAL or type == NodeType.CONTINUOUS or \
           type == NodeType.CLOCK or type == NodeType.INT_CONST or \
           type == NodeType.REAL_CONST:
            node.node_type = type
        elif type == NodeType.ENUM_FIELD:
            node.node_type = NodeType.ENUM_FIELD
            node.enum_type = add_info
        elif type == NodeType.ENUM_VALUE:
            node.node_type = NodeType.ENUM_VALUE
            node.enum_values = add_info
        elif type == NodeType.INT_RANGE:
            node.node_type = NodeType.INT_RANGE
            node.bounds = add_info
        else:
            assert False, "Unknown node type"
            node.node_type = NodeType.ERROR

class TypeVisitor(CompileVisitor):
    """Determines type of (sub)expressions, and makes sure they are correct.
    Otherwise, adds error to compileinfo argument.
    Note: Assumes that leaf nodes already have types (eg number, boolean , id nodes)
    Note: specify action functions, need to type in post order"""
    def __init__(self, compile_info):
        super(TypeVisitor, self).__init__()

        self.compile_info = compile_info
        
        # Reference object is to keep track of objects that define expressions
        # allows error messages to indicate which part causes the error (for CGM)
        self.current_reference_obj = None
        self.current_reference_obj_stack = []
        
    def _is_ref_obj(self, node):
        # Bit of a hack: filter object types to see which ones are interesting for error messages
        from slim_compiler.src.dataStructure.compile import Package, Subpackage
        from slim_compiler.src.dataStructure.nominalbehaviour import ComponentImplementation,\
            ComponentType, DataSubcomponent, Flow, NonDataSubcomponent, Port, Mode,\
            ModeTransition
        from slim_compiler.src.dataStructure.errorbehaviour import ErrorEvent,\
            ErrorModelImplementation, ErrorModelType, ErrorPropagation, ErrorState,\
            ErrorTransition

        return isinstance(node, (Package, Subpackage,
                                 ComponentType, ComponentImplementation, Port, Flow,
                                 DataSubcomponent, NonDataSubcomponent, Mode, ModeTransition,
                                 ErrorModelType, ErrorModelImplementation, ErrorState,
                                 ErrorTransition, ErrorEvent, ErrorPropagation)
                          )

    def visit(self, node, *args):
        ref = False
        if self._is_ref_obj(node):
            self.current_reference_obj_stack.append(self.current_reference_obj)
            self.current_reference_obj = node
        res = super(TypeVisitor, self).visit(node, *args)
        if ref:
            self.current_reference_obj = self.current_reference_obj_stack[-1]
            self.current_reference_obj_stack.pop()
        return res

    def action_TrajectoryEquationTreeNode(self, root_node):
        if root_node.variable.node_type != NodeType.CONTINUOUS and \
           root_node.variable.node_type != NodeType.ERROR:
            self.compile_info.add_error(
                    TrajectoryEquationReferencesIllegalType(
                            root_node.file, root_node.line, root_node.pos, "I-5",
                            NodeType.type_to_str(root_node.variable.node_type),
                            self.current_reference_obj))
            root_node.node_type = NodeType.ERROR
        elif self.compile_info.uses_time_scaling and root_node.time_unit is None:
            self.compile_info.add_error(TrajectoryExpressionNotScaledInTimeScaledModel(root_node.file, root_node.line, root_node.pos,
                                                                                       "I-24", root_node, self.current_reference_obj))
            root_node.node_type = NodeType.ERROR
        else:
            root_node.node_type = NodeType.BOOL

    def action_BinaryOperatorTreeNode(self, root_node):
        result_table = expressions.get_result_table_for_operator(root_node.operator, self.compile_info.uses_time_scaling)
        left_type, right_type = root_node.left.node_type, root_node.right.node_type

        continuous_types = [NodeType.CONTINUOUS, NodeType.CLOCK]

        if (left_type, right_type) in result_table:
            root_node.node_type = result_table[(left_type, right_type)]

            if left_type in continuous_types and root_node.operator == OperatorType.ASSIGN:
                self._check_hybrid_assign(root_node)

            elif left_type in expressions.enum_types and right_type in expressions.enum_types:
                self._check_enum_operator(root_node)
        else:
            if left_type != NodeType.ERROR and right_type != NodeType.ERROR:
                self.compile_info.add_error(
                        ExpressionTypeMismatchError(root_node.file, root_node.line,
                                                    root_node.pos,
                                                    NodeType.type_to_str(left_type),
                                                    NodeType.type_to_str(right_type),
                                                    OperatorType.operator_to_str(root_node.operator),
                                                    self.current_reference_obj))
            root_node.node_type = NodeType.ERROR

        # Check if expressions with clocks are linear
        if root_node.node_type != NodeType.ERROR:
            extended_continuous_types = continuous_types + [NodeType.INT,
                                                            NodeType.REAL]
            nonlinear_operators = [OperatorType.TIMES, OperatorType.DIVIDE]
            constant_types = [NodeType.INT_CONST, NodeType.REAL_CONST]
            if root_node.node_type in extended_continuous_types and (
                   (root_node.operator in nonlinear_operators and\
                    root_node.left.node_type in extended_continuous_types and \
                    root_node.right.node_type in extended_continuous_types) or \
                   (root_node.operator == OperatorType.DIVIDE and \
                    root_node.left.node_type in constant_types and \
                    root_node.right.node_type in extended_continuous_types)
                                                                     ):
                    self.compile_info.add_error(
                            ExpressionIsNonLinearError(root_node.file, root_node.line,
                                                       root_node.pos,
                                                       NodeType.type_to_str(root_node.left.node_type),
                                                       NodeType.type_to_str(root_node.right.node_type),
                                                       OperatorType.operator_to_str(root_node.operator),
                                                       self.current_reference_obj))
    
                    root_node.node_type = NodeType.ERROR


    def action_UnaryOperatorTreeNode(self, root_node):
        result_table = expressions.get_result_table_for_operator(root_node.operator, self.compile_info.uses_time_scaling)
        subtype = root_node.subexpression.node_type

        if subtype in result_table:
            root_node.node_type = result_table[subtype]
            if OperatorType.MSEC <= root_node.operator <= OperatorType.DAY:
                if root_node.node_type == NodeType.REAL:
                    # See ticket 543. If the result is REAL type, use x_TOREAL operator type
                    root_node.operator = OperatorType.MSEC_TOREAL + (root_node.operator - OperatorType.MSEC)
                else:
                    root_node.operator = OperatorType.REAL_TOMSEC + (root_node.operator - OperatorType.MSEC)
        else:
            if subtype != NodeType.ERROR:
                self.compile_info.add_error(
                                            UnaryExpressionTypeMismatchError(root_node.file, root_node.line,
                                            root_node.pos,
                                            NodeType.type_to_str(root_node.subexpression.node_type),
                                            OperatorType.operator_to_str(root_node.operator),
                                            self.current_reference_obj))
            root_node.node_type = NodeType.ERROR

    def action_CaseExpressionTreeNode(self, root_node):
        result_types = []

        for (cond, expr) in root_node.cases:
            result_types.append(expr.node_type)
        result_types.append(root_node.default_expression.node_type)

        # check for boolean conditions
        for (cond, _) in root_node.cases:
            if cond.node_type != NodeType.BOOL and cond.node_type != NodeType.ERROR:
                self.compile_info.add_error(
                                            CaseConditionNotOfTypeBoolError(root_node.file,
                                                cond.line,
                                                cond.pos,
                                                NodeType.type_to_str(cond.node_type),
                                                self.current_reference_obj))

        # check for equality of result types
        if not NodeType.ERROR in result_types:  # otherwise do not continue checking
            result_types_set = set(result_types)

            if len(result_types_set) == 1:  # exactly one type found
                root_node.node_type = iter(result_types_set).next()

            if len(result_types_set) > 1 or NodeType.ENUM_FIELD in result_types or NodeType.ENUM_VALUE in result_types:
                # more than one type found, or enum type
                if result_types_set - expressions.enum_types_set == set():
                    value_indices = []
                    field_indices = []
                    for i in range(0, len(result_types)):
                        if  result_types[i] == NodeType.ENUM_FIELD:
                            field_indices.append(i)
                        else:
                            value_indices.append(i)

                    # check whether all enum values are contained in all enum fields
                    values_contained_in_fields = True
                    field_values = []
                    value_values = []
                    for i in field_indices:
                        # Note: Field values are overwritten as in a correct tree there can only be one
                        if i < len(result_types) - 1:
                            field_values = (root_node.cases[i])[1].enum_type
                        elif i == len(result_types) - 1:
                            field_values = root_node.default_expression.enum_type
                    for j in value_indices:
                        if j < len(result_types) - 1:
                            value_values += (root_node.cases[j])[1].enum_values
                        elif j == len(result_types) - 1:
                            value_values += root_node.default_expression.enum_values

                    # check whether all enum fields agree on valid values
                    enum_fields_agree_on_values = True
                    for i in field_indices:
                        for j in field_indices:
                            if i != j:
                                if i < len(result_types) - 1:
                                    values_field_1 = (root_node.cases[i])[1].enum_type
                                elif  i == len(result_types) - 1:
                                    values_field_1 = root_node.default_expression.enum_type
                                if j < len(result_types) - 1:
                                    values_field_2 = (root_node.cases[j])[1].enum_type
                                elif  j == len(result_types) - 1:
                                    values_field_2 = root_node.default_expression.enum_type

                                if set(values_field_1) != set(values_field_2):
                                    enum_fields_agree_on_values = False

                    # check whether fields and values agree on valid values
                    if len(field_values) > 0 and len(value_values) > 0:
                        if len(set(value_values) - set(field_values)) > 0:
                            values_contained_in_fields = False

                    # Note: ideally check if all values comes from one single enum,
                    # however this information is not available

                    if not enum_fields_agree_on_values:
                        self.compile_info.add_error(
                            CaseExpressionsHaveDifferentValidEnumValuesError(root_node.file,
                                                                    root_node.line,
                                                                    root_node.pos,
                                                                    self.current_reference_obj))
                        root_node.node_type = NodeType.ERROR

                    if not values_contained_in_fields:
                        self.compile_info.add_error(
                            CaseExpressionsHaveDifferentValidEnumsError(root_node.file,
                                                                      root_node.line,
                                                                      root_node.pos,
                                                                      self.current_reference_obj))
                        root_node.node_type = NodeType.ERROR

                    if enum_fields_agree_on_values and values_contained_in_fields:
                        if len(value_indices) > 0:
                            root_node.node_type = NodeType.ENUM_VALUE
                            root_node.enum_values = value_values
                        else:
                            root_node.node_type = NodeType.ENUM_FIELD
                            root_node.enum_type = field_values

                elif result_types_set - expressions.numeric_types_set == set():
                    # only numeric types, calculate least type that covers all
                    ordered_result_types = [val for val in expressions.numeric_types if val in result_types_set]
                    root_node.node_type = ordered_result_types[-1]
                else:
                    result_type_names_set = set()
                    for type in result_types_set:
                        result_type_names_set.add(NodeType.type_to_str(type))

                    self.compile_info.add_error(
                            CaseExpressionsHaveMultipleTypesError(
                                    root_node.file, root_node.line, root_node.pos,
                                    result_type_names_set,
                                    self.current_reference_obj))
                    root_node.node_type = NodeType.ERROR
        else:
            root_node.node_type = NodeType.ERROR

    def _check_hybrid_assign(self, root_node):
        # Check special assignment cases for hybrid types
        # Continuous may only be assigned constants, clock only zero
        # special for clock: there may be a time scale operator, which is removed
        assert root_node.operator == OperatorType.ASSIGN

        if root_node.left.node_type == NodeType.CONTINUOUS:
            if not isinstance(root_node.right, NumberTreeNode):
                self.compile_info.add_error(
                                            ContinuousNotSetToRealError(root_node.file,
                                                             root_node.line,
                                                             root_node.pos,
                                                             self.current_reference_obj))
                root_node.node_type = NodeType.ERROR
        else:
            # If time scaled zero -> transform to simple zero
            # Note: not checking for x_TOREAL nodes, as they should not occur here
            if isinstance(root_node.right, UnaryOperatorTreeNode) \
                and (OperatorType.MSEC <= root_node.right.operator <= OperatorType.DAY or \
                     OperatorType.REAL_TOMSEC <= root_node.right.operator <= OperatorType.REAL_TODAY) \
                and isinstance(root_node.right.subexpression, NumberTreeNode) \
                and int(root_node.right.subexpression.get_value()) == 0:
                    root_node.right = root_node.right.subexpression
                    root_node.right.subexpression.parent = root_node.right
            elif not isinstance(root_node.right, NumberTreeNode) or root_node.right.get_value() != 0:
                self.compile_info.add_error(
                                            ClockSetToNonZeroError(root_node.file,
                                                                   root_node.line,
                                                                   root_node.pos,
                                                                   self.current_reference_obj))
                root_node.node_type = NodeType.ERROR

    def _check_enum_operator(self, root_node):
        # Helper for binary operators involving enums
        # Make sure enum is compatible
        left_type, right_type = root_node.left.node_type, root_node.right.node_type

        if left_type == NodeType.ENUM_FIELD and right_type == NodeType.ENUM_VALUE:
            difference = set(root_node.right.enum_values) - set(root_node.left.enum_type)
            if len(difference) > 0:
                self.compile_info.add_error(
                                            EnumValueIllegalError(file, root_node.line,
                                                       root_node.pos, "A-0",
                                                       difference,
                                                       self.current_reference_obj))
                root_node.node_type = NodeType.ERROR
        elif left_type == NodeType.ENUM_VALUE and right_type == NodeType.ENUM_FIELD:
            difference = set(root_node.left.enum_values) - set(root_node.right.enum_type)
            if len(difference) > 0:
                self.compile_info.add_error(EnumValueIllegalError(root_node.file, root_node.line,
                                                       root_node.pos, "A-0",
                                                       difference,
                                                       self.current_reference_obj))
                root_node.node_type = NodeType.ERROR
        elif left_type == NodeType.MODE_FIELD and right_type == NodeType.MODE_VALUE:
            if root_node.right.name not in root_node.left.enum_type:
                self.compile_info.add_error(ModeValueIllegalError(root_node.right.name, self.current_reference_obj))
                root_node.node_type = NodeType.ERROR
        elif left_type == NodeType.MODE_VALUE and right_type == NodeType.MODE_FIELD:
            if root_node.left.name not in root_node.right.enum_type:
                self.compile_info.add_error(ModeValueIllegalError(root_node.left.name, self.current_reference_obj))
                root_node.node_type = NodeType.ERROR
        elif left_type == NodeType.ERROR_FIELD and right_type == NodeType.ERROR_VALUE:
            if root_node.right.name not in root_node.left.enum_type:
                self.compile_info.add_error(ErrorStateValueIllegalError(root_node.right.name, self.current_reference_obj))
                root_node.node_type = NodeType.ERROR
        elif left_type == NodeType.ERROR_VALUE and right_type == NodeType.ERROR_FIELD:
            if root_node.left.name not in root_node.right.enum_type:
                self.compile_info.add_error(ErrorStateValueIllegalError(root_node.left.name, self.current_reference_obj))
                root_node.node_type = NodeType.ERROR
        elif left_type == right_type:
            if not set(root_node.left.enum_type) == set(root_node.right.enum_type):
                root_node.node_type = NodeType.ERROR
                if left_type == NodeType.ENUM_FIELD:
                    self.compile_info.add_error(EnumsIncomparableError(root_node.file,
                                                            root_node.line,
                                                            root_node.pos,
                                                            root_node.left.name,
                                                            root_node.right.name,
                                                            self.current_reference_obj))
                elif left_type == NodeType.MODE_FIELD:
                    self.compile_info.add_error(ModeFieldsIncomparableError(
                                                            root_node.left.name,
                                                            root_node.right.name,
                                                            self.current_reference_obj))
                elif left_type == NodeType.ERROR_FIELD:
                    self.compile_info.add_error(ErrorFieldsIncomparableError(
                                                            root_node.left.name,
                                                            root_node.right.name,
                                                            self.current_reference_obj))
        else:
            assert False

class RangeTransformVisitor(TransformVisitor):
    """Transforms assignments to INT ranges s.t. they are performed
    modulo the range of the target.
    Note: no check is made for the source type, type checker deals with that"""
    def action_BinaryOperatorTreeNode(self, root_node, visit_result):
        if root_node.operator == OperatorType.ASSIGN and root_node.left.node_type == NodeType.INT_RANGE:
            lower_bound = NumberTreeNode(str(root_node.left.bounds[0]), None, 0, 0)
            right_minus_lower = BinaryOperatorTreeNode(root_node.right, lower_bound,
                                                       OperatorType.MINUS, 0, 0)
            right_minus_lower.node_type = NodeType.INT_RANGE
            precalculated_value = NumberTreeNode(str(root_node.left.bounds[1]
                                                 - root_node.left.bounds[0] + 1),
                                                 None, 0, 0)
            perform_modulo = BinaryOperatorTreeNode(right_minus_lower,
                                                    precalculated_value,
                                                    OperatorType.MOD, 0, 0)
            perform_modulo.node_type = NodeType.INT_RANGE
            final_node = BinaryOperatorTreeNode(perform_modulo, lower_bound,
                                                OperatorType.PLUS, 0, 0)

            root_node.right = final_node
            root_node.node_type = root_node.left.node_type
        return root_node

class ExpressionPostProcessor(CompileVisitor):
    """Parsers expression trees and upcasts constant
    types to their variable counterparts. Also ensures
    both types of binary operators are identical.
    Note: visit method overridden, as a pre-order is required
    Note2: MUST be called as the last step, as it invalidates
    the expression type hierarchy (ie assign clock to clock)"""
    def __init__(self):
        super(ExpressionPostProcessor, self).__init__()

    def action_DataPort(self, node):
        if node.default_value is not None:
            # Ensure the data type of the default value matches that of the component itself
            if node.data_type.get_node_type() in expressions.numeric_types:
                self.visit(node.default_value, node.data_type.get_node_type())

    def action_OutDataPort(self, node):
        if node.default_value is not None:
            # Ensure the data type of the default value matches that of the component itself
            if node.data_type.get_node_type() in expressions.numeric_types:
                self.visit(node.default_value, node.data_type.get_node_type())

    def action_DataSubcomponent(self, node):
        if node.default_value is not None:
            # Ensure the data type of the default value matches that of the component itself
            if node.data_type.get_node_type() in expressions.numeric_types:
                self.visit(node.default_value, node.data_type.get_node_type())

    def action_BinaryOperatorTreeNode(self, root_node, type=None):
        if root_node.node_type == NodeType.ERROR:
            return

        if type != None:
            # A type is to be applied
            assert root_node.node_type in expressions.numeric_types

            root_node.node_type = type
            self.visit(root_node.left, type)
            self.visit(root_node.right, type)
            return

        if root_node.operator == OperatorType.ASSIGN:
            # Case 1: Assignment of numeric values
            if root_node.left.node_type == NodeType.BOOL:
                # Simply recurse when not assigning something numeric
                self.visit(root_node.right)
            else:
                # Apply lhs type to rhs when assigning something numeric
                self.visit(root_node.right, root_node.left.node_type)
        elif root_node.left.node_type in expressions.numeric_types and \
            root_node.right.node_type in expressions.numeric_types:
            # Case 2: Any other binary operator involving numerics upcasts
            # both sides to highest type, and not constant

            type_ind = max(expressions.numeric_types.index(root_node.left.node_type),
                           expressions.numeric_types.index(root_node.right.node_type))
            type = expressions.numeric_types[type_ind]
            # Exception for INT_RANGE: In SMV, INT_RANGE is integer, whereas INT and INT_CONST are signed word
            # To make them compatible, force type to INT_RANGE if left or right is INT_RANGE.
            if root_node.left.node_type == NodeType.INT_RANGE or root_node.right.node_type == NodeType.INT_RANGE:
                type = NodeType.INT_RANGE
            if type == NodeType.INT_CONST: type = NodeType.INT
            elif type == NodeType.REAL_CONST: type = NodeType.REAL

            self.visit(root_node.left, type)
            self.visit(root_node.right, type)
        else:
            self.visit(root_node.left)
            self.visit(root_node.right)

    def action_UnaryOperatorTreeNode(self, root_node, type=None):
        if root_node.node_type == NodeType.ERROR:
            return

        # First check if the node is a timed operator, which may change the applied type
        if OperatorType.MSEC <= root_node.operator <= OperatorType.DAY or\
             OperatorType.MSEC_TOREAL <= root_node.operator <= OperatorType.DAY_TOREAL:
            # Timed expressions must be of REAL or CLOCK type
            self.visit(root_node.subexpression, NodeType.REAL)
            return

        if OperatorType.REAL_TOMSEC <= root_node.operator <= OperatorType.REAL_TODAY:
            # Timed expressions must be of REAL or CLOCK type
            self.visit(root_node.subexpression, NodeType.CLOCK)
            return

        if type != None:
            root_node.type = type
            self.visit(root_node.subexpression, type)
            return

        # No logic for unary expressions required
        self.visit(root_node.subexpression)

    def action_CaseExpressionTreeNode(self, root_node, type=None):
        if root_node.node_type == NodeType.ERROR:
            return

        if type != None:
            root_node.type = type
            for (cond, expr) in root_node.cases:
                self.visit(cond) # Condition is boolean
                self.visit(expr, type)
            self.visit(root_node.default_expression, type)
            return

        # Find max type and apply
        types = []
        for (cond, expr) in root_node.cases:
            types.append(expr.node_type)
        result_types_set = set(types)
        if result_types_set - expressions.numeric_types_set == set():
            # only numeric types, calculate least type that covers all
            ordered_result_types = [val for val in expressions.numeric_types if val in result_types_set]
            max_type = ordered_result_types[-1]
            if type == NodeType.INT_CONST: type = NodeType.INT
            elif type == NodeType.REAL_CONST: type = NodeType.REAL
            for (cond, expr) in root_node.cases:
                self.visit(cond) # Condition is boolean
                self.visit(expr, max_type)
            self.visit(root_node.default_expression, max_type)

    def action_TrajectoryEquationTreeNode(self, root_node, type=None):
        if root_node.node_type == NodeType.ERROR:
            return

        root_node.trajectory_expression.node_type = NodeType.CONTINUOUS

    def action_NumberTreeNode(self, root_node, type=None):
        if root_node.node_type == NodeType.ERROR:
            return

        if type != None:
            root_node.node_type = type
        else:
            # Force to non-const type (ex for default values)
            if root_node.node_type == NodeType.INT_CONST: root_node.node_type = NodeType.INT
            elif root_node.node_type == NodeType.REAL_CONST: root_node.node_type = NodeType.REAL

class ExpressionValueVisitor(EvalVisitor):
    """Visitor that evaluates expression based on available static information. Uses default
    values for identifier nodes where applicable. Returns None if expression cannot be evaluated,
    or argument is not an expression"""
    def __init__(self, compile_info, use_defaults=True):
        super(ExpressionValueVisitor, self).__init__()

        self.compile_info = compile_info
        self.use_defaults = use_defaults

    def action_IdentifierTreeNode(self, root_node, data):
        if self.use_defaults:
            return self._get_value_of_identifier_considering_default_values(root_node)
        else:
            return None

    def action_BooleanTreeNode(self, root_node, data):
        return root_node.value

    def action_NumberTreeNode(self, root_node, data):
        return root_node.get_value()

    def action_TrajectoryExpressionTreeNode(self, root_node, data):
        # deprecated
        return None

    def action_BinaryOperatorTreeNode(self, root_node, data):
        left_val = data.left
        right_val = data.right
        res = None
        if left_val is not None and right_val is not None:
            if root_node.operator == OperatorType.PLUS:
                res = left_val + right_val
            elif root_node.operator == OperatorType.MINUS:
                res = left_val - right_val
            elif root_node.operator == OperatorType.TIMES:
                res = left_val * right_val
            elif root_node.operator == OperatorType.DIVIDE:
                real_types = [NodeType.REAL_CONST, NodeType.REAL, NodeType.CONTINUOUS, NodeType.CLOCK]
                if root_node.left.node_type in real_types or root_node.right.node_type in real_types:
                    res = Fraction(Fraction(left_val), Fraction(right_val))
                else:
                    res = left_val / right_val
            elif root_node.operator == OperatorType.MOD:
                res = left_val % right_val

            elif root_node.operator == OperatorType.AND:
                res = left_val and right_val
            elif root_node.operator == OperatorType.OR:
                res = left_val or right_val
            elif root_node.operator == OperatorType.XOR:
                res = not (left_val == right_val)
            elif root_node.operator == OperatorType.XNOR:
                res = (left_val == right_val)
            elif root_node.operator == OperatorType.IFF:
                res = (left_val == right_val)
            elif root_node.operator == OperatorType.IMP:
                res = (not left_val) or right_val

            elif root_node.operator == OperatorType.LESS:
                res = left_val < right_val
            elif root_node.operator == OperatorType.LESSOREQUAL:
                res = left_val <= right_val
            elif root_node.operator == OperatorType.GREATER:
                res = left_val > right_val
            elif root_node.operator == OperatorType.GREATEROREQUAL:
                res = left_val >= right_val
            elif root_node.operator == OperatorType.EQUAL:
                res = left_val == right_val
            elif root_node.operator == OperatorType.NOTEQUAL:
                res = left_val != right_val

            elif root_node.operator == OperatorType.ASSIGN:
                # Cannot evaluate assignment
                return None

            else:
                assert False, "Unknown binary operator %d to evaluate"%root_node.operator
        return res

    def action_UnaryOperatorTreeNode(self, root_node, data):
        val = data.subexpression
        if val is None:
            return None

        if root_node.operator == OperatorType.UNARY_MINUS:
            return -val
        elif root_node.operator == OperatorType.NOT:
            return not val
        elif OperatorType.MSEC <= root_node.operator <= OperatorType.DAY or\
             OperatorType.MSEC_TOREAL <= root_node.operator <= OperatorType.DAY_TOREAL or\
             OperatorType.REAL_TOMSEC <= root_node.operator <= OperatorType.REAL_TODAY:
            # Same logic applies as for TimeScaleTransformVisitor
            factor = self._get_timescale_factor(root_node.operator)
            if root_node.node_type != NodeType.CLOCK:
                # If moving from time to discrete, invert
                factor = -factor
            if factor < 0:
                return float(val) / -factor
            else:
                return val * factor
        else:
            assert False, "Unknown unary operator %d to evaluate"%root_node.operator


    def action_TrajectoryEquationTreeNode(self, root_node, data):
        # Trajectory equations are part of invariants and simply evaluate to
        # True as they do not invalidate invariants
        return True

    def action_CaseExpressionTreeNode(self, root_node, data):
        # For cases, find first case with condition that has been evaluated
        # to True, otherwise return result of default case evaluation
        for (cond_eval, expr_eval) in data.cases:
            if cond_eval is True:
                return expr_eval
        return data.default_expression

    def _get_timescale_factor(self, operator):
        # TODO: This is copy/pasta from TimeScaleTransformVisitor,
        # may need to be put in utility class
        scale = {
            OperatorType.MSEC : 1,
            OperatorType.SEC : 1000,
            OperatorType.MIN : 60 * 1000,
            OperatorType.HOUR : 3600 * 1000,
            OperatorType.DAY : 24*3600*1000,
        }

        base_operator = TimeUnitType.to_operator(self.compile_info.base_unit)

        # The REAL_TOX operators are the same as X, so convert them back
        if OperatorType.REAL_TOMSEC <= operator <= OperatorType.REAL_TODAY:
            operator = OperatorType.MSEC + (operator - OperatorType.REAL_TOMSEC)
        # The X_TOREAL operators are the same as X, so convert them back
        elif OperatorType.MSEC_TOREAL <= operator <= OperatorType.DAY_TOREAL:
            operator = OperatorType.MSEC + (operator - OperatorType.MSEC_TOREAL)

        if not operator in scale:
            raise InternalException("Unable to get time scale factor for operator: " + str(operator))

        if base_operator < operator:
            factor = scale[operator] / scale[base_operator]
            # Negative result means should be inverted (ie division required)
            return factor
        else:
            factor = scale[base_operator] / scale[operator]
            return -factor

    def _get_value_of_identifier_considering_default_values(self, root_node):
        """Retrieves the value of the IdentifierTreeNode, considering the
        given default values.

        @param    root_node:
                    The IdentifierTreeNode to get value of.
        @type     root_node:
                    IdentifierTreeNode
        @return:  The value of the IdentifierTreeNode. None if it is unknown.
        """
        assert isinstance(root_node, IdentifierTreeNode)
        res = None
        if root_node.node_type == NodeType.ENUM_VALUE:
            # If the identifier node is an enum value, simply
            # return its name
            res = root_node.name
        elif root_node.node_type == NodeType.CLOCK:
            res = 0.0
        elif root_node.ref is not None:
            if root_node.ref.default_value is not None:
                def_val = root_node.ref.default_value
                res = self.visit(def_val)
            # TODO: it might be interesting to obtain value if target of
            # some port connection (else case, ref = port)
        return res
