# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================
from slim_compiler.src.dataStructure.visitors.baseVisitor import CompileVisitor,\
    TransformVisitor
from slim_compiler.src.dataStructure.located_info import LocatedInformation
from slim_compiler.src.dataStructure.nominalbehaviour import ComponentImplementation,\
    ModeTransitionLabel, ModeTransition, ReactivationModeTransitionLabel,\
    ModeType
from slim_compiler.src.dataStructure.errorbehaviour import ErrorTransition
from slim_compiler.src.dataStructure.visitors.expressionVisitor import InvariantToNNFVisitor,\
    SimplifyCaseVisitor, SimplifyLinEq, SimplifyExprVisitor
from slim_compiler.src.dataStructure.expressions import BinaryOperatorTreeNode,\
    TrajectoryEquationTreeNode, OperatorType, IdentifierTreeNode, NodeType,\
    NumberTreeNode, UnaryOperatorTreeNode
from slim_compiler.src.dataStructure.visitors.printVisitor import toSLIMVisitor

class WildcardTransformVisitor(TransformVisitor):
    """Transforms wildcard transitions into regular ones"""
    def action_ComponentImplementation(self, node, visit_result):
        normal_list = []
        for trans in node.normal_mode_transitions:
            if trans.source_mode == "*":
                self._expand_wildcard_normal(node, normal_list, trans, False)
            else:
                normal_list.append(trans)
        node.normal_mode_transitions = normal_list

        reactivation_list = []
        for trans in node.reactivation_mode_transitions:
            if trans.source_mode == "*":
                self._expand_wildcard_normal(node, reactivation_list, trans, True)
            else:
                reactivation_list.append(trans)
        node.reactivation_mode_transitions = reactivation_list

        return node

    def _expand_wildcard_normal(self, impl, list, trans, is_reactivate):
        #TODO: May need to update ComponentImplementation.nextModeTransitionID
        number = trans.number
        #TODO: transition_counter may be required,
        # increment for each new transition
        internal_number = trans.internal_number

        if impl.modes:
            transition_counter = 1
            for mode in impl.modes:
                effects = [effect.clone() for effect in trans.label.effects]
                if is_reactivate:
                    label = ReactivationModeTransitionLabel(effects)
                else:
                    label = ModeTransitionLabel(trans.label.trigger, trans.label.guard.clone() if trans.label.guard else None, trans.label.time_delay, effects)
                mode_transition=ModeTransition(impl,
                                       number,
                                       internal_number,
                                       mode, trans.target_mode,
                                       label, impl.file,
                                       trans.line, trans.pos)
                list.append(mode_transition)
            transition_counter += 1
            ComponentImplementation.nextModeTransitionID += 1

    def action_ErrorModelImplementation(self, node, visit_result):
        type = node.get_referenced_error_type()
        states = node.states.keys()
        if type is not None:
            states += type.states.keys()

        normal_list = []
        for trans in node.error_transitions:
            if trans.source_state == "*":
                self._expand_wildcard_error(node, normal_list, trans)
            else:
                normal_list.append(trans)
        node.error_transitions = normal_list

        reactivation_list = []
        for trans in node.reactivation_transitions:
            if trans.source_state == "*":
                self._expand_wildcard_error(node, reactivation_list, trans)
            else:
                reactivation_list.append(trans)
        node.reactivation_transitions = reactivation_list

        return node

    def _expand_wildcard_error(self, impl, list, trans):
        states = impl.states.keys()
        for state in states:
                effects = [effect.clone for effect in trans.effects]
                list.append(
                            ErrorTransition(impl,
                                            trans.number,
                                            state,
                                            trans.is_reactivation,
                                            trans.trigger,
                                            trans.target_state,
                                            trans.guard.clone(),
                                            effects,
                                            trans.time_delay,
                                            trans.file,
                                            trans.line,
                                            trans.pos))

class TriggerListTransformVisitor(TransformVisitor):
    """Transforms transitions with multiple triggers into regular ones"""
    def action_ComponentImplementation(self, node, visit_result):
        new_list = []
        for trans in node.normal_mode_transitions:
            if trans.label.trigger is not None and not isinstance(trans.label.trigger, basestring):
                for trigger in trans.label.trigger:
                    effects = [effect.clone() for effect in trans.label.effects]
                    label = ModeTransitionLabel(trigger, trans.label.guard.clone() if trans.label.guard else None, trans.label.time_delay, effects)
                    mode_transition=ModeTransition(node,
                                           ComponentImplementation.nextModeTransitionID, 0,
                                           trans.source_mode, trans.target_mode,
                                           label,
                                           trans.file, trans.line, trans.pos)
                    new_list.append(mode_transition)
                    ComponentImplementation.nextModeTransitionID += 1
            else:
                new_list.append(trans)

        node.normal_mode_transitions = new_list

        return node

class LocatedInfoVisitor(CompileVisitor):
    def __init__(self):
        super(LocatedInfoVisitor, self).__init__()

        self.current_parent_stack = [None]

    def _is_loc_obj(self, node):
        # Bit of a hack: filter object types to see which ones are or type LocatedInformation
        return isinstance(node, (LocatedInformation))

    def visit(self, node, *args):
        ref = self._is_loc_obj(node)
        if ref:
            self._set_parent(node, self.current_parent_stack[-1])
            self.current_parent_stack.append(node)
        res = super(LocatedInfoVisitor, self).visit(node, *args)
        if ref:
            self.current_parent_stack.pop()
        return res

    def _set_parent(self, node, parent):
        if parent is None:
            return

        # print "parent of " + type(node).__name__ + " is " + type(self.current_parent).__name__
        node.parent = parent
        if node.file is None:
            node.file = parent.file
        if node.line == 0:
            node.line = parent.line
        if node.pos == 0:
            node.pos = parent.pos

class Clause(object):
    def __init__(self):
        self.cond = None
        self.equations = []

    def add_expr(self, expr):
        if self._is_constant(expr):
            self.cond = BinaryOperatorTreeNode.create_and(self.cond, expr)
        else:
            self.equations.append(LinExp(expr))

    def _is_constant(self, expr):
        if isinstance(expr, BinaryOperatorTreeNode):
            return self._is_constant(expr.left) and \
                   self._is_constant(expr.right)
        elif isinstance(expr, UnaryOperatorTreeNode):
            return self._is_constant(expr.subexpression)
        elif isinstance(expr, IdentifierTreeNode):
            return expr.node_type not in [NodeType.CLOCK, NodeType.CONTINUOUS]
        else:
            return True

    def __repr__(self):
        strExpr = [repr(eq) for eq in self.equations]
        if self.cond:
            strExpr += [repr(self.cond)]
        result = ' AND '.join(strExpr)
        return result

class LinExp(object):
    def __init__(self, expr):
        assert isinstance(expr, BinaryOperatorTreeNode)
        self.terms = []
        self.operator = expr.operator
        self.constant = expr.right
        self._get_factors(expr.left)

    def _get_factors(self, expr):
        assert isinstance(expr, BinaryOperatorTreeNode)
        assert expr.operator in [OperatorType.PLUS, OperatorType.TIMES]
        if expr.operator == OperatorType.PLUS:
            self.get_factors(self, expr.left)
            self.get_factors(self, expr.right)
        else:
            assert isinstance(expr.right, IdentifierTreeNode)
            self.terms.append(Term(expr.left, expr.right))

    def __repr__(self):
        return "%s %s %s"%(' + '.join([str(term) for term in self.terms]), OperatorType.operator_to_str(self.operator), self.constant)

class Term(object):
    def __init__(self, factor, var):
        self.factor = factor
        self.var = var

    def __repr__(self):
        return "%s * %s"%(self.factor, self.var)

class SimulatorVisitor(CompileVisitor):
    """This visitor is used to modify the compile info s.t. the
    Monte Carlo simulator can handle it (a few things are simplified
    or rewritten into a better parseable format)"""
    def __init__(self):
        super(SimulatorVisitor, self).__init__()

    def _get_trajectories(self, inv, trajectories):
        if isinstance(inv, BinaryOperatorTreeNode):
            self._get_trajectories(inv.left, trajectories)
            self._get_trajectories(inv.right, trajectories)
        elif isinstance(inv, TrajectoryEquationTreeNode):
            trajectories.append(inv)

    def _get_clauses(self, inv, clauses):
        if isinstance(inv, BinaryOperatorTreeNode):
            if inv.operator == OperatorType.OR:
                self._get_clauses(inv.left, clauses)
                self._get_clauses(inv.right, clauses)
            else:
                newclause = Clause()
                self._get_clause(inv, newclause)
                clauses.append(newclause)
        else:
            newclause = Clause()
            self._get_clause(inv, newclause)
            clauses.append(newclause)

    def _get_clause(self, inv, clause):
        if isinstance(inv, BinaryOperatorTreeNode):
            if inv.operator == OperatorType.AND:
                self._get_clause(inv.left, clause)
                self._get_clause(inv.right, clause)
            else:
                clause.add_expr(inv)
        else:
            clause.add_expr(inv)

    def _transform_invariant(self, inv):
        trajectories = []
        invariants = []

        self._get_trajectories(inv, trajectories)

        inv = SimplifyCaseVisitor().replace_case(inv)
        inv = InvariantToNNFVisitor(False).visit(inv)
        inv = SimplifyExprVisitor().visit(inv)
        inv = SimplifyLinEq().build(inv)
        self._get_clauses(inv, invariants)

        return (invariants, trajectories)

    def action_Mode(self, node):
        if node.invariant:
            invariant = node.invariant.clone()
            (node.clauses, node.trajectories) = self._transform_invariant(invariant)
        else:
            (node.clauses, node.trajectories) = (None, None)

    def action_ModeTransitionLabel(self, node):
        if node.guard is not None:
            guard = node.guard.clone()
            (node.clauses, _) = self._transform_invariant(guard)

class SimulatorPrintVisitor(toSLIMVisitor):
    def action_Mode(self, node, data):
        result = node.name + " :"
        if node.type == ModeType.INITIAL:
            result += " initial"
        elif node.type == ModeType.ACTIVATION:
            result += " activation"
        elif node.type == ModeType.ERROR:
            result += " error"
        result += " mode"
        if node.urgent:
            result += " urgent in " + data.urgent
        if node.invariant:
            result += " while " + data.invariant
        if node.clauses:
            result += " while " + ' OR '.join([str(clause) for clause in node.clauses])
        if node.trajectories:
            result += " traj: " + str(node.trajectories)
        result += ";"
        return result
