# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================
"""This module provides visitors for dealing with timed specifications"""

from slim_compiler.src.dataStructure.visitors.baseVisitor import CompileVisitor,\
    TransformVisitor
from slim_compiler.src.dataStructure.expressions import BinaryOperatorTreeNode,\
    OperatorType, TimeUnitType, NumberTreeNode, NodeType, IdentifierTreeNode,\
    TrajectoryEquationTreeNode
from slim_compiler.src.dataStructure.nominalbehaviour import TIME_DELAY_CLK_NAME
from utils.exceptions import InternalException

class CheckTimeScaleUsageVisitor(CompileVisitor):
    """This class provides visit-methods for the CompileInformation-structure.
    The visit-methods check whether time scaling operators are present in the model
    and set the uses_time_scaling flag if so."""
    def __init__(self, compile_info):
        super(CheckTimeScaleUsageVisitor, self).__init__()

        self.compile_info = compile_info
        self.compile_info.uses_time_scaling = False
        self.compile_info.base_unit = TimeUnitType.SEC

    def action_Clock(self, node):
        if node.time_unit is not None:
            self.compile_info.uses_time_scaling = True

    def action_ErrorEvent(self, node):
        if node.time_unit is not None:
            self.compile_info.uses_time_scaling = True

    def action_UnaryOperatorTreeNode(self, node):
        if node.operator in [OperatorType.MSEC, OperatorType.SEC, OperatorType.MIN, OperatorType.HOUR, OperatorType.DAY]:
            self.compile_info.uses_time_scaling = True
        if node.operator in [OperatorType.REAL_TOMSEC, OperatorType.REAL_TOSEC, OperatorType.REAL_TOMIN, OperatorType.REAL_TOHOUR, OperatorType.REAL_TODAY]:
            self.compile_info.uses_time_scaling = True
        if node.operator in [OperatorType.MSEC_TOREAL, OperatorType.SEC_TOREAL, OperatorType.MIN_TOREAL, OperatorType.HOUR_TOREAL, OperatorType.DAY_TOREAL]:
            self.compile_info.uses_time_scaling = True

    def action_TrajectoryEquation(self, node):
        if node.time_unit is not None:
            self.compile_info.uses_time_scaling = True

class TimeDelayClockVisitor(CompileVisitor):
    """This class provides visit-methods for the CompileInformation-structure.
    The visit-methods modify the compile information in the following way:
    1.) Every implementation which uses the time delay clock (happens when
    the 'within' keyword is used on a transition and/or 'urgent' specified on mode)
    will have the clock added
    2.) Add guards to transitions with the 'within' specifier
    3.) Add invariants according to the 'urgent' parameter
    4.) Reset of the clock is added to outgoing transitions of delay modes
    """
    def __init__(self, compile_info):
        super(TimeDelayClockVisitor, self).__init__()

        self.compile_info = compile_info
        self.ts_transform = TimeScaleTransformVisitor(self.compile_info)

    def action_ComponentImplementation(self, node):
        for mode_name in node.modes:
            mode = node.modes[mode_name]
            if self._modeHasDelay(node, mode):
                self._createModeInvariant(mode, node)

                # Reset the delay clock for all transitions that leave the mode
                for t in (node.normal_mode_transitions + node.reactivation_mode_transitions):
                    if t.source_mode == mode_name:
                        self._createGuardEffect(t.label, node)

    def action_ErrorModelImplementation(self, node):
        for state_name in node.states:
            state = node.states[state_name]
            if self._stateHasDelay(node, state):
                self._createModeInvariant(state, node)

                # Reset the delay clock for all transitions that leave the mode
                for t in (node.error_transitions + node.reactivation_transitions):
                    if t.source_state == state_name:
                        self._createGuardEffect(t, node)

    def _createModeInvariant(self, node, impl):
        time_delay_clock = impl.create_time_delay_clock()
        if node.urgent is not None:
            # Create urgency invariant
            upper_bound = self._createDelayInvariant(time_delay_clock, node.urgent)
            node.invariant = BinaryOperatorTreeNode.create_and(upper_bound, node.invariant)

            node.urgent = None

        traj = self._createDelayTrajectory(time_delay_clock)
        node.invariant = BinaryOperatorTreeNode.create_and(node.invariant, traj)

    def _createDelayInvariant(self, clock, upper):
        id = self._createClockNode(clock)
        upper = self._convertTimeExpr(upper)
        return BinaryOperatorTreeNode.createNode(id, upper, OperatorType.LESSOREQUAL, False)

    def _createDelayTrajectory(self, clock):
        id = self._createClockNode(clock)
        traj = TrajectoryEquationTreeNode(id, NumberTreeNode("1", clock.file, clock.line, clock.pos), clock.line, clock.pos,
                                          self.compile_info.base_unit)
        traj.file = clock.file
        return traj

    def _createGuardEffect(self, node, impl):
        time_delay_clock = impl.create_time_delay_clock()

        if node.time_delay is not None:
            # Create time delay guard: time_delay[0] <= _t and _t <= time_delay[1]
            time_delay_guard = self._createDelayGuard(time_delay_clock, node.time_delay[0], node.time_delay[1])

            # Combine with existing guard if possible
            if node.guard:
                # Conjunction of actual guard and time_delay_guard  
                node.guard = BinaryOperatorTreeNode.create_and(node.guard, time_delay_guard)
            else:
                node.guard = time_delay_guard

            node.time_delay = None
        
        # Reset the clock always
        for effect in node.effects:
            if effect.left.name == time_delay_clock.name:
                break;
        else:
            reset_effect = self._createDelayReset(time_delay_clock)
            node.effects.append(reset_effect)

    def _createDelayGuard(self, clock, lower, upper):
        '''Create lower <= ident_name and ident_name <= upper'''
        lower = self._convertTimeExpr(lower)
        upper = self._convertTimeExpr(upper)

        id1 = self._createClockNode(clock)
        left = BinaryOperatorTreeNode.createNode(lower, id1, OperatorType.LESSOREQUAL, self.compile_info.uses_time_scaling)

        id2 = self._createClockNode(clock)
        right = BinaryOperatorTreeNode.createNode(id2, upper, OperatorType.LESSOREQUAL, self.compile_info.uses_time_scaling)

        and_node = BinaryOperatorTreeNode.create_and(left, right)
        return and_node

    def _createDelayReset(self, clock):
        id = self._createClockNode(clock)
        return BinaryOperatorTreeNode.createNode(id, NumberTreeNode.createZero(clock.file, clock.line, clock.pos), OperatorType.ASSIGN, False)

    def _convertTimeExpr(self, expr):
        expr = self.ts_transform.visit(expr)
        return expr
    
    def _modeHasDelay(self, impl, mode):
        if mode.urgent is not None:
            return True
        for trans in impl.normal_mode_transitions:
            if trans.source_mode == mode and trans.label.time_delay is not None:
                return True

    def _stateHasDelay(self, impl, state):
        if state.urgent is not None:
            return True
        for trans in impl.error_transitions:
            if trans.source_state == state and trans.time_delay is not None:
                return True

    def _createClockNode(self, clock):
        id = IdentifierTreeNode(clock.name, clock.line, clock.pos)
        id.file = clock.file
        id.node_type = clock.data_type.get_node_type()
        return id

class TimeScaleTransformVisitor(TransformVisitor):
    """This class provides visit-methods for the CompileInformation-structure.
    The visit-methods transform the time scales in the expressions, e.g. replace
    min(3.0) with 1/60 * 3.0.
    
    Note that all time units are transformed to the base unit of the compile info given.
    """

    def __init__(self, compile_info):
        super(TimeScaleTransformVisitor, self).__init__()

        self.compile_info = compile_info

        self._scale = {
                OperatorType.MSEC : 1.0,
                OperatorType.SEC : 1000.0,
                OperatorType.MIN : 60 * 1000.0,
                OperatorType.HOUR : 3600 * 1000.0,
                OperatorType.DAY : 24*3600*1000.0,
                }

    def _get_factor(self, operator):
        base_operator = TimeUnitType.to_operator(self.compile_info.base_unit)

        # The REAL_TOX operators are the same as X, so convert them back
        if OperatorType.REAL_TOMSEC <= operator <= OperatorType.REAL_TODAY:
            operator = OperatorType.MSEC + (operator - OperatorType.REAL_TOMSEC)
        # The X_TOREAL operators are the same as X, so convert them back
        elif OperatorType.MSEC_TOREAL <= operator <= OperatorType.DAY_TOREAL:
            operator = OperatorType.MSEC + (operator - OperatorType.MSEC_TOREAL)

        if not operator in self._scale:
            raise InternalException("Unable to get time scale factor for operator: " + str(operator))

        if base_operator < operator:
            factor = self._scale[operator] / self._scale[base_operator]
            # Negative result means should be inverted (ie division required)
            return -factor
        else:
            factor = self._scale[base_operator] / self._scale[operator]
            return factor

    def _get_scale_expr(self, expr, factor):
        inverse = False
        if factor < 0:
            inverse = True
            factor = -factor

        if factor == 1:
            return expr

        factor = repr(factor)

        if inverse:
            new_node = BinaryOperatorTreeNode.createDiv(expr, factor, self.compile_info.uses_time_scaling)
        else:
            new_node = BinaryOperatorTreeNode.createMul(expr, factor, self.compile_info.uses_time_scaling)

        return new_node

    def _get_scaled_value(self, number, factor):
        inverse = False
        if factor < 0:
            inverse = True
            factor = -factor

        if factor == 1:
            return number

        value = number.get_value()

        if inverse:
            value = value / factor
        else:
            value = value * factor

        new_node = NumberTreeNode(str(value), None, 0, 0)
        return new_node

    def action_Clock(self, node, visit_result):
        if node.time_unit is not None:
            # No-op for clocks
            # Do not filter out time unit, as still required by SMV translator
            pass

        return node

    def action_TrajectoryEquationTreeNode(self, root_node, visit_result):
        if root_node.time_unit is not None:
            factor = self._get_factor(TimeUnitType.to_operator(root_node.time_unit))
            new_node = self._get_scaled_value(root_node.trajectory_expression, factor)

            root_node.trajectory_expression = new_node
            root_node.time_unit = None

        return root_node

    def action_ErrorEvent(self, node, visit_result):
        if node.time_unit is not None:
            factor = self._get_factor(TimeUnitType.to_operator(node.time_unit))

            inverse = False
            if factor < 0:
                inverse = True
                factor = -factor

            param = float(node.parameter)
            if not inverse:
                node.parameter = repr(param*factor)
            else:
                node.parameter = repr(float(param)/factor)

            node.time_unit = None

        return node

    def action_ErrorEventPort(self, node, visit_result):
        if node.time_unit is not None:
            factor = self._get_factor(TimeUnitType.to_operator(node.time_unit))

            inverse = False
            if factor < 0:
                inverse = True
                factor = -factor

            param = float(node.parameter)
            if not inverse:
                node.parameter = repr(param*factor)
            else:
                node.parameter = repr(float(param)/factor)

            node.time_unit = None

        return node

    def action_UnaryOperatorTreeNode(self, root_node, visit_result):
        # The visitor should not generate a new node for TimeScaleVisitor, action does that
        assert root_node == visit_result
        if root_node.operator == OperatorType.NOT or root_node.operator == OperatorType.UNARY_MINUS:
            # Nothing to do
            return root_node

        factor = self._get_factor(root_node.operator)
        if root_node.node_type == NodeType.CLOCK:
            # Converting from Real to Clock means inverting the operation
            factor = -factor
        new_node = self._get_scale_expr(root_node.subexpression, factor)

        return new_node
