# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


"""This module provides functions to parse and validate fault injection
documents in the deprecated implementation-based format and convert them to
the component-based format.

In the implementation-based format, error models are assigned to component
implementations.
In the component-based format, error models are assigned to component
instances.
"""

import copy
import os.path
from lxml import etree

from slim_compiler.src.model_extension import fault_injection
from slim_compiler.src.model_extension.util import general


# The filename of the XML Schema document that defines the structure of valid
# fault injection documents in the implementation-based format:
IMPLEMENTATION_BASED_XSD_FILENAME = os.path.join(
        os.path.dirname(__file__), "xsd",
        "implementation_fault_injection.xsd")


def parse_implementation_based_injection_document(data_source):
    """Parses the implementation-based fault injection document from the
    specified data source and returns a tree representation of the document.
    
    Note: The document will not be validated. This must be done explicitly
    using the function validate_implementation_based_injection_document.
    
    @param   data_source: The source of the data to be parsed. This may be a
                          filename or a file-like object.
    @type    data_source: A string or a file-like object.
    @return: A tree representation of the document.
    @rtype:  etree._Element
    """
    # Parse the document with the specified filename.
    # Note: XML comments must be discarded. Otherwise the parser would yield
    # data structures cluttered with unwanted elements.
    # Unnecessary whitespace is discarded to ensure that pretty printing will
    # work properly, in case that the caller of this function wants to
    # serialize the document or parts thereof.
    # See http://codespeak.net/lxml/FAQ.html#why-doesn-t-the-pretty-print-option-reformat-my-xml-output
    # The result of the parsing process will be an etree._ElementTree object.
    # Return the root element of this tree.
    parser = etree.XMLParser(remove_comments=True, remove_blank_text=True)
    document_tree = etree.parse(data_source, parser)
    return document_tree.getroot()


def validate_implementation_based_injection_document(
                            fault_injection_document):
    """Validates the specified implementation-based fault injection document
    against the adequate XML Schema Definition and raises an exception if the
    document is not valid.
    
    Note: This is a purely syntactic which does not include a check of the
    error model associations with respect to a certain SLIM model.
    
    @param   fault_injection_document:
                The root element of what is supposed to be an
                implementation-based fault injection document.
                Note: The document must be specified in form of the root
                element.
    @type    fault_injection_document:
                etree._Element
    @raises  etree.DocumentInvalid: If the document is invalid.
    """
    # Parse the XML Schema document that defines the structure of valid fault
    # injection documents in the implementation-based format.
    schema_tree = etree.parse(IMPLEMENTATION_BASED_XSD_FILENAME)
    schema = etree.XMLSchema(schema_tree)
    # Validate the fault injection document.
    # If the document is invalid, an exception will be raised.
    schema.assertValid(fault_injection_document)


def create_component_based_injection_document(original_document,
                                              root_implementation):
    """Returns a new component-based fault injection document that is
    semantically equivalent to the specified implementation-based document.
    
    Note: The returned document will not be checked for validity with respect
    to the given SLIM model.
    
    @param   original_document:
                The implementation-based fault injection document to be
                considered. It must be a valid document according to the XML
                Schema Definition implementation_fault_injection.xsd.
                Note: The document must be specified in form of the root
                element. The XML tree rooted at that element must not contain
                comments or processing instructions.
    @type    original_document:
                etree._Element
    @param   root_implementation:
                The root implementation of the SLIM model that is the subject
                of the specified fault injection document.
    @type    root_implementation:
                nominalbehaviour.ComponentImplementation
    @return: A component-based fault injection document.
    @rtype:  fault_injection.FaultInjectionDocument
    @raises  etree.DocumentInvalid: If the specified document is invalid.
    """
    # Validate the specified fault injection document.
    # If the document is invalid, an exception will be raised.
    validate_implementation_based_injection_document(original_document)
    
    # Create a new fault injection document in the format that assigns error
    # models to component instances rather than to component implementations.
    new_document = fault_injection.FaultInjectionDocument.create()
    # The child elements of the root element of the specified document
    # represent associations between component implementations and error model
    # implementations. Compute corresponding associations for each instance of
    # the respective component implementation and add them to the newly
    # created fault injection document.
    for implementation_association in original_document:
        instance_associations = _create_component_associations(
                                    implementation_association,
                                    root_implementation)
        for instance_association in instance_associations:
            new_document.add_association(instance_association)
    
    return new_document


def _create_component_associations(implementation_association,
                                   root_implementation):
    """Returns a list of component-based error model associations that are
    semantically equivalent to the specified implementation-based association.
    
    An implementation-based error model association refers to a certain
    component implementation. This function will create a corresponding
    association for each instance of that component implementation.
    
    @param   implementation_association:
                The implementation-based association, represented by an XML
                element. The XML tree rooted at that element must not contain
                comments or processing instructions.
    @type    implementation_association:
                etree._Element
    @param   root_implementation:
                The root implementation of the SLIM model to be considered.
    @type    root_implementation:
                nominalbehaviour.ComponentImplementation
    @return: A list of component-based error model associations.
    @rtype:   fault_injection.ErrorModelAssociation
    """
    # An error model association for a component implementation is an XML
    # element with three child elements:
    # 1) The element that identifies the nominal component implementation.
    # 2) The element that identifies the error model implementation.
    # 3) The element that specifies the fault effects to be injected.
    # The first and second child elements each have two child elements
    # themselves:
    # a) The element whose text content is the name of the package containing
    #    the component implementation or error model implementation.
    # b) The element whose text content is the implementation name.
    nominal_package_name = implementation_association[0][0].text
    nominal_implementation_name = implementation_association[0][1].text
    error_model_package_name = implementation_association[1][0].text
    error_model_implementation_name = implementation_association[1][1].text
    fault_effects_specification = implementation_association[2]
    
    # Get the component implementation that is referenced by the association.
    compile_info = root_implementation.subpackage.package.compile_information
    try:
        package = compile_info.packages[nominal_package_name]
    except KeyError:
        # TODO: Use a custom exception class.
        raise Exception(("Invalid fault injection document: Package '%s' " +
                        "could not be found.") % nominal_package_name)
    # Note: False is passed as the second parameter of
    # get_component_implementation so that both the public and the private
    # subpackage are searched.
    component_implementation = package.get_component_implementation(
                                        nominal_implementation_name,
                                        False)
    
    # Create an appropriate error model association for each instance of that
    # component implementation and return them in a list.
    def component_has_right_implementation(component):
        return component.implementation == component_implementation
    components_iterator = general.instance_generator(
                                root_implementation,
                                component_has_right_implementation)
    
    component_associations = []
    for component in components_iterator:
        # The fault effects specification can be reused in the instance-based
        # association. However, it must be copied because otherwise it would
        # be removed from the original implementation-based association.
        component_association = fault_injection.ErrorModelAssociation.create(
                component.path,
                error_model_package_name,
                error_model_implementation_name,
                copy.deepcopy(fault_effects_specification))
        component_associations.append(component_association)
        
    return component_associations
