# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================
from slim_compiler.src.dataStructure.errorbehaviour import ErrorStateType


"""This module provides the function create_extended_model, which performs
model extension for a given SLIM model.
"""
# TODO: In this file, a maximum line length of 80 is used, instead of 79 as
# specified in PEP8.

# Note: The alias "slim_compile" is used instead of "compile" to avoid
# shadowing of the built-in function "compile".
# TODO: Some of these imports might not strictly be needed: They define classes
# and this module works with instances of these classes, but never instantiates
# these classes, so there is no strong connection to the imported modules.
import slim_compiler.src.dataStructure.compile as slim_compile
import slim_compiler.src.dataStructure.errorbehaviour as error
import slim_compiler.src.dataStructure.nominalbehaviour as nominal
from slim_compiler.src.dataStructure import expressions
import slim_compiler.src.errorManagement.errors as compile_errors
from slim_compiler.src.model_extension import constants
from slim_compiler.src.model_extension import fault_injection
from slim_compiler.src.model_extension import errors
from slim_compiler.src.model_extension.util import general
import slim_compiler.src.model_extension.util.expressions as util_expressions
from slim_compiler.src.dataStructure.visitors.printVisitor import toSLIMVisitor
from slim_compiler.src.dataStructure.visitors.time import TimeDelayClockVisitor
from slim_compiler.src.dataStructure.visitors.expressionVisitor import AnnotateVisitor, TypeVisitor

from utils import log

# Note: The data structures of the SLIM compiler are designed to be used by
# the SLIM parser, so many of their methods expect as parameters the line and
# position of a part of the SLIM specification. In this module, these data
# structures are used to create a SLIM model without parsing a specification
# file, so no meaningful values can be provided for these parameters. 0 will be
# passed instead. If a filename is expected as a parameter, None will be
# passed.

# IMPORTANT NOTE: This model extension mechanism relies on the assumption that
# all states of an error model are visible, i.e. declared in the error model
# type.


def create_extended_model(original_root_implementation,
                          fault_injection_document):
    """Creates an extended SLIM model for the original model that is determined
    by the specified root component implementation.
    
    The extended model is obtained from the original model by integrating the
    error models that are associated with the components of the original model
    as defined by the specified fault injection document.
    A new slim_compile.CompileInformation object will be created to represent
    the extended model.
    
    @param   original_root_implementation: The implementation of the root
                                           component of the original SLIM
                                           model.
    @type    original_root_implementation: nominal.ComponentImplementation
    @param   fault_injection_document:     A document that associates nominal
                                           components with error models.
    @type    fault_injection_document:     fault_injection.\
                                           FaultInjectionDocument
    @return: The implementation of the root component of the extended model.
    @rtype:  nominal.ComponentImplementation
    @raises  InvalidFaultInjectionDocumentError: If the specified fault
                injection document is not valid with respect to the specified
                SLIM model.
    """
    # Check the validity of the fault injection document with respect to the
    # original SLIM model.
    error_list = fault_injection.get_errors_in_injection_document(
                                    fault_injection_document,
                                    original_root_implementation)
    if len(error_list) > 0:
        raise InvalidFaultInjectionDocumentError(error_list)
    
    # Create the object that will represent the extended model and add the
    # default package to the model.
    extended_compile_info = slim_compile.CompileInformation()
    # Copy over properties of CompileInformation of the original root
    extended_compile_info.uses_time_scaling = original_root_implementation.subpackage.package.\
            compile_information.uses_time_scaling
    extended_compile_info.base_unit = original_root_implementation.subpackage.package.\
            compile_information.base_unit

    extended_compile_info.add_package(constants.DEFAULT_PACKAGE_NAME, 0, 0)
    # Create an object to represent the root component. A component is a
    # specific instance of a component implementation and can be identified by
    # a sequence of subcomponent identifiers (i.e. a path in the tree structure
    # of the SLIM model), starting at the root component.
    # For the root component, this sequence is an empty tuple.
    original_root_component = general.ComponentInstance(
                                    original_root_implementation, ())
    # Add the extended types and implementations to the extended model and
    # get the extended version of the specified root component implementation.
    extended_root_implementation = _add_extended_declarations(
                                        extended_compile_info,
                                        original_root_component,
                                        fault_injection_document)
    
    # Annotate the expressions in the extended model with type information.
    # Note that the expressions of the original model that were simply reused
    # in the extended model should already be annotated before, but some of the
    # expressions that were added during model extension are not.
    # Note: After the annotation, the error collector of the compile
    # information object might contain errors, because extended models can
    # violate some restrictions that are imposed on nominal models.
    # TODO: Check whether it is really possible that the error collector
    # contains errors.
    extended_compile_info = extended_root_implementation.subpackage.package.\
                                                        compile_information

    annotator = AnnotateVisitor(extended_compile_info)
    type_checker = TypeVisitor(extended_compile_info)

    annotator.visit(extended_compile_info)
    type_checker.visit(extended_compile_info)

    return extended_root_implementation


# TODO: Maybe the following class should be moved to the module "errors".
# In this case, the docstring has to be adapted.
class InvalidFaultInjectionDocumentError(Exception):
    """An object indicating that a fault injection document is incorrect.
    
    While the exceptions in the module "errors" are designed to represent only
    a single specific error at a time, InvalidFaultInjectionDocumentError
    objects provide a list of (possibly) more than one such specific errors.
    """
    
    def __init__(self, error_list):
        """Initializes this InvalidFaultInjectionDocumentError object.
        
        @param   error_list: A list containing the errors that were detected in
                             the fault injection document.
        @type    error_list: A list of errors.FaultInjectionDocumentBaseError
                             objects.
        """
        self.error_list = error_list
        
    def __str__(self):
        """Returns a human-readable string representation of the errors in the
        fault injection document.
        
        @return: A string containing the descriptions of all the errors that
                 were detected in the fault injection document.
        @rtype:  string
        """
        return ("The fault injection document is incorrect:\n" +
                "\n".join(map(str, self.error_list)))


def _add_extended_declarations(extended_compile_info,
                               original_component,
                               fault_injection_document):
    """Creates the extended declarations (component types and implementations)
    for all direct and indirect subcomponents of the specified component
    (including that component itself) and adds them to the specified extended
    SLIM model.
    
    @param   extended_compile_info:    The object representing the extended
                                       SLIM model to which the extended
                                       declarations will be added.
    @type    extended_compile_info:    slim_compile.CompileInformation
    @param   original_component:       A component of the original (not
                                       extended) SLIM model.
    @type    original_component:       general.ComponentInstance
    @param   fault_injection_document: A document that associates nominal
                                       components with error models.
    @type    fault_injection_document: fault_injection.FaultInjectionDocument
    @return: The extended implementation for the specified original component.
    @rtype:  nominal.ComponentImplementation
    """
    # Add the extended declarations recursively for all subcomponents of the
    # specified original component to the specified extended SLIM model.
    # While doing this, remember the names of the extended implementations for
    # these subcomponents. These extended implementations will be referenced
    # later, when the extended implementation for the specified component
    # itself is created.  
    subcomponent_id_to_implementation_name_dict = {}
    for subcomponent_id in original_component.implementation.\
                                        non_data_subcomponents:
        subcomponent = original_component.get_subcomponent_instance(
                                                    subcomponent_id)
        extended_implementation = _add_extended_declarations(
                                        extended_compile_info,
                                        subcomponent,
                                        fault_injection_document)
        subcomponent_id_to_implementation_name_dict[subcomponent_id] = \
                extended_implementation.name

    # Add the extended type and implementation for the specified component to
    # the public part of the default package of the specified extended SLIM
    # model. This requires the error model that is associated with the
    # component by means of the fault injection document.
    error_model = _get_associated_error_model(original_component,
                                              fault_injection_document)
    target_subpackage = general.get_public_default_subpackage(
                                        extended_compile_info)
    # It is necessary to generate a name for the extended type that is unique
    # in the default package of the extended model. Using the old name would
    # possible cause name collisions because
    # - all extended types are stored in the default package and
    # - multiple extended types might be created for a single original type,
    #   since there might be multiple components with the same type but
    #   different associated error models.
    # Note that there will be only one implementation for each type in the
    # extended model, so the extended implementation names are unique by
    # construction.
    extended_type_name = _generate_unique_extended_type_name(
                                    original_component,
                                    target_subpackage.package)
    extended_type = _add_extended_type(target_subpackage,
                                       original_component.type,
                                       error_model,
                                       extended_type_name)
    extended_implementation = _add_extended_implementation(
                                target_subpackage,
                                original_component,
                                error_model,
                                fault_injection_document,
                                extended_type_name,
                                subcomponent_id_to_implementation_name_dict)
    return extended_implementation


def _add_extended_type(target_subpackage,
                       original_type,
                       error_model,
                       extended_type_name):
    """Creates the extended component type for the specified original type,
    using the specified error model, and adds it to the specified extended SLIM
    model.
    
    @precondition:
        The specified subpackage must not yet contain a component type with the
        specified name. 
    
    @param   target_subpackage:  The subpackage to which the extended type
                                 shall be added.
    @type    target_subpackage:  slim_compile.Subpackage
    @param   original_type:      The original component type.
    @type    original_type:      nominal.ComponentType
    @param   error_model:        The error model that is associated with the
                                 original component.
    @type    error_model:        general.ErrorModel
    @param   extended_type_name: The name to be used for the extended type.
    @type    extended_type_name: string
    @return: The extended type for the specified original type.
    @rtype:  nominal.ComponentType
    """
    # Use a copy of the specified original type as the basis for the extended
    # type.
    extended_type = _add_copy_of_type(target_subpackage,
                                      original_type,
                                      extended_type_name)
    # If the component, whose type is considered here, has an associated error
    # model, then add the additional features to the extended type that are
    # needed to represent the error behavior.
    if error_model is not None:
        # Add for each error propagation of the associated error model a
        # corresponding event port to the extended component type.
        _add_error_propagation_ports(extended_type, error_model.type)
        # Add to the extended component type an out data port of type enum
        # whose possible values represent the states of the associated error
        # model.

        _add_error_states_port(extended_type, error_model)
        
    return extended_type
    
    
def _add_extended_implementation(
        target_subpackage,
        original_component,
        error_model,
        fault_injection_document,
        extended_type_name,
        subcomponent_id_to_implementation_name_dict):
    """Creates the extended component implementation for the specified original
    component, using the specified error model, and adds it to the specified
    extended SLIM model.
    
    @precondition:
        The specified subpackage must not yet contain an implementation for the
        component type with the specified name. 
    
    @param   target_subpackage:        The subpackage to which the extended
                                       implementation shall be added.
    @type    target_subpackage:        slim_compile.Subpackage
    @param   original_component:       The original component.
    @type    original_component:       general.ComponentInstance
    @param   error_model:              The error model that is associated with
                                       the original component.
    @type    error_model:              general.ErrorModel
    @param   fault_injection_document: A document that associates nominal
                                       components with error models.
    @type    fault_injection_document: fault_injection.FaultInjectionDocument
    @param   extended_type_name:       The name of the extended type that is
                                       associated with the extended
                                       implementation to be created.
    @type    extended_type_name:       string
    @param   subcomponent_id_to_implementation_name_dict:
                                       A dictionary mapping each of the
                                       identifiers of the subcomponents of the
                                       specified component to the name of the
                                       extended implementation for that
                                       subcomponent.
    @type    subcomponent_id_to_implementation_name_dict:
                                       dict (string -> string)
    @return: The extended implementation for the specified original component.
    @rtype:  nominal.ComponentImplementation
    """
    # TODO: Maybe the error model should not be passed as a parameter since it
    # can be determined using the specified fault injection document.
    
    # Compute the name of the extended implementation based on the name of the
    # corresponding extended type and the suffix of the original implementation
    # name.
    (original_type_name, suffix) = \
            original_component.implementation.name.split(".")
    extended_implementation_name = extended_type_name + "." + suffix
    # Add the extended implementation to the specified subpackage.
    # The category is the same as for the original component and the extended
    # component is an FDIR component if the original component was an FDIR
    # component.
    extended_implementation = target_subpackage.add_component_implementation(
                                    extended_implementation_name,
                                    original_component.implementation.category,
                                    original_component.implementation.fdir,
                                    0, 0)
    
    # Add the subcomponents of the original implementation to the extended one.
    # Use the dictionary providing the names of the extended implementations in
    # order to add the extended versions of the non-data subcomponents.
    _add_subcomponents(extended_implementation,
                       original_component.implementation,
                       subcomponent_id_to_implementation_name_dict)
    # Add the modes of the original implementation to the extended one.
    _add_modes(extended_implementation, original_component.implementation)
    # Add the event port connections of the original implementation to the
    # extended one.
    _add_original_event_port_connections(extended_implementation,
                                         original_component.implementation)
    
    # Add the port connections that are used to represent error propagation
    # between the subcomponents of the specified component.
    # Note that these connections must even be added if the specified component
    # does not have an associated error model itself.
    _add_subcomponents_propagation_connections(extended_implementation,
                                               original_component,
                                               fault_injection_document)
    
    # Copy the reactivation transitions of the original nominal model implementation
    # to the extended one
    _add_reactivation_transitions(extended_implementation,
                                  original_component.implementation)
    
    # If an error model is associated with the specified component, then some
    # additional parts have to be added to the extended component and some
    # parts of the original component have to be added in a modified way.
    if error_model is not None:
        # The component has an associated error model.  
        # Get the specification that defines the fault effects to be injected
        # into the component.
        fault_effects_specification = _get_fault_effects_specification(
                                                original_component,
                                                fault_injection_document)
        # Add the flows to the extended component, taking into account the
        # fault effects associated with the error states of the associated
        # error model.
        _add_extended_flows(extended_implementation,
                            original_component.implementation,
                            error_model,
                            fault_effects_specification)
        # Add additional flows to the extended component, representing the
        # data port connections of the original component and taking into
        # account the fault effects associated with the error states of the
        # associated error model.
        _add_extended_flows_for_port_connections(
                            extended_implementation,
                            original_component.implementation,
                            error_model,
                            fault_effects_specification)
        # Add an error component to the extended model to represent the
        # behavior of the associated error model.
        # TODO: Maybe only one error component type and implementation should
        # be added for each error model, even if it is associated with multiple
        # components.
        (error_component_type, error_component_implementation) = \
                _add_error_component(target_subpackage, error_model)
        # Add the error component as a subcomponent of the extended component.
        # Note that all other subcomponents have already been added.
        _add_as_error_subcomponent(extended_implementation,
                                   error_component_implementation)
        # Connect the enum port indicating the error state of the error
        # subcomponent with the corresponding enum port of its supercomponent.
        # TODO: Maybe a flow should be used instead of a port connection. 
        error_states_port_name = (constants.ERROR_SUBCOMPONENT_IDENTIFIER +
                                  "." + constants.ERROR_STATES_ENUM_PORT_NAME)
        extended_implementation.add_data_port_connection(
                error_states_port_name,
                constants.ERROR_STATES_ENUM_PORT_NAME,
                [], 0, 0)
        # Add the port connections that are used to forward error events from
        # the error subcomponent to its supercomponent and vice versa.
        _add_error_propagation_forwarding_connections(extended_implementation,
                                                      error_model.type)
        # Add the port connections that are used to represent error
        # propagation from the current component to its subcomponents.
        _add_super_to_subcomponent_propagation_connections(
                                    extended_implementation,
                                    original_component,
                                    error_model.type,
                                    fault_injection_document)
        # Add the port connections that are used to represent error propagation
        # from the subcomponents of the current component to the component
        # itself.
        _add_sub_to_supercomponent_propagation_connections(
                                    extended_implementation,
                                    original_component,
                                    error_model.type,
                                    fault_injection_document)
        # Add the transitions to the extended component, taking into account
        # the behavior of the associated error model and the fault effects
        # associated to its error states.
        # Note: this has to be done after event connections are made for
        # error propagations
        _add_extended_transitions(extended_implementation,
                                  original_component.implementation,
                                  error_model,
                                  fault_effects_specification)
    else:
        # The component has no associated error model.
        # Just add the flows, data port connections and transitions of the
        # original component to the extended one.
        _add_original_flows(extended_implementation,
                            original_component.implementation)
        _add_original_data_port_connections(extended_implementation,
                                            original_component.implementation)
        _add_original_transitions(extended_implementation,
                                  original_component.implementation)
    
    return extended_implementation


def _add_copy_of_type(target_subpackage,
                      original_component_type,
                      new_component_type_name):
    """Adds a copy of the specified component type to the specified subpackage,
    under the specified name, and returns the corresponding 
    nominal.ComponentType object.
    
    @param   target_subpackage:       The subpackage to which the copy of the
                                      component type shall be added.
    @type    target_subpackage:       slim_compile.Subpackage
    @param   original_component_type: The component type to be copied.
    @type    original_component_type: nominal.ComponentType
    @param   new_component_type_name: The name for the copy of the component
                                      type.
    @type    new_component_type_name: string
    @return: The copy of the specified component type that was added to the
             specified subpackage.
    @rtype:  nominal.ComponentType
    """
    # Add a copy of the specified component type to the specified subpackage.
    component_type_copy = target_subpackage.add_component_type(
                    new_component_type_name,
                    original_component_type.category,
                    original_component_type.fdir,
                    0, 0)
    # Add to the copied component type the data and event ports of the original
    # component type.
    for port in original_component_type.in_data_ports.itervalues():
        # TODO: Should a copy of the default value be created?
        component_type_copy.add_port(port.file, port.name, True, False, False, port.data_type,
                                     port.default_value, False, False, 0, 0)
    for port in original_component_type.out_data_ports.itervalues():
        # TODO: Should a copy of the default value be created?
        component_type_copy.add_port(port.file, port.name, False, False, False, port.data_type,
                                     port.default_value, port.observable,
                                     port.alarm, 0, 0)
    for port in original_component_type.in_event_ports.itervalues():
        component_type_copy.add_port(port.file, port.name, True, True, False, None, None, False,
                                     False, 0, 0)
    for port in original_component_type.out_event_ports.itervalues():
        component_type_copy.add_port(port.file, port.name, False, True, port.nonblocking, None, None, False,
                                     False, 0, 0)
    
    return component_type_copy


def _generate_unique_extended_type_name(original_component, package):
    """Returns a new component type name, for the extended version of the
    specified component, that is unique in the specified package.
    
    The generated name will be based on the subcomponent identifier path and
    the type name of the specified component.
    
    @param   original_component: The original component to be considered.
    @type    original_component: general.ComponentInstance
    @param   package: The package in which the new type name must be unique.
    @type    package: slim_compile.Package
    @return: A new component type name that is unique in the specified package.
    @rtype:  string
    """
    # TODO: Use constants instead of string literals.
    prefix = ("Extended_" + "_".join(original_component.path) + "_" +
              original_component.type.name)
    return _generate_unique_type_name(prefix, package)


# TODO: Maybe the following function should be moved to a utility module.
def _generate_unique_type_name(prefix, package):
    """Returns a new component type name that is unique in the specified
    package and starts with the specified prefix.
    
    The name will be generated by appending the smallest number
    n = 1, 2, 3, ... to the specified prefix such that the resulting name is
    unique in the specified package.
    
    @param   prefix:  A string to be used as the prefix of the new name.
    @type    prefix:  string
    @param   package: The package in which the new type name must be unique.
    @type    package: slim_compile.Package
    @return: A new component type name that is unique in the specified package
             and starts with the specified prefix.
    @rtype:  string
    """
    # Note: False is passed as the second argument of
    # package.get_component_type_ because the both the public and the private
    # subpackage must be considered.
    n = 1
    while package.get_component_type(prefix + str(n), False):
        n += 1
    return prefix + str(n)


def _add_error_propagation_ports(component_type, error_model_type):
    """Adds for each error propagation of the specified error model type a
    corresponding event port to the specified component type.
    
    @param   component_type:   The component type to be extended by error event
                               ports.
    @type    component_type:   nominal.ComponentType
    @param   error_model_type: The error model type whose error propagations
                               are to be represented in the specified component
                               type.
    @type    error_model_type: error.ErrorModelType
    """
    for propagation in error_model_type.error_propagations.itervalues():
        # Add an event port representing the current error propagation to the
        # specified component type.
        # Distinguish in and out error propagations.
        is_in_event = (propagation.type == error.ErrorPropagationType.IN)
        # Use an encoding of the propagation name as the name of the new event
        # port to avoid name clashes with existing ports.
        port_name = _get_event_port_name(propagation.name)
        # Error propagations which become out event ports should become nonblocking
        if is_in_event:
            component_type.add_port(component_type.file, port_name, is_in_event, True, False, None, None,
                                False, False, 0, 0)
        else:
            component_type.add_port(component_type.file, port_name, is_in_event, True, True, None, None,
                                False, False, 0, 0)
        

def _add_error_event_ports(component_type,
                           error_model_implementation):
    """Adds for each error event of the specified error model implementation a
    corresponding outgoing event port to the specified component type.
    
    @param   component_type:             The component type to be extended by
                                         error event ports.
    @type    component_type:             nominalbehaviour.ComponentType
    @param   error_model_implementation: The error model implementation whose
                                         error events are to be represented in
                                         the specified component type.
    @type    error_model_implementation: error.ErrorModelImplementation
    """
    for error_event in error_model_implementation.error_events.itervalues():
        # Add the outgoing event port representing the current error event to
        # the specified component type.
        # Note: Since the distribution and time unit of the error event are
        # expressed by constants (or are None if not specified), they have to
        # be converted to their string representation in order to pass them to
        # the method nominal.ComponentType.add_error_event_port.
        distribution = None
        if error_event.distribution is not None:
            distribution = error.DistributionType.dist_type_to_str(
                                            error_event.distribution)
        # Use an encoding of the error event name as the name of the new event
        # port to avoid name clashes with existing ports.
        # Note: In fact, there should be no existing ports because event ports
        # representing error events are only added to error subcomponents.
        port_name = _get_event_port_name(error_event.name) 
        component_type.add_error_event_port(port_name,
                                            distribution,
                                            error_event.parameter,
                                            error_event.time_unit,
                                            0, 0)


def _add_error_states_port(component_type, error_model):
    """Adds to the specified component type an out data port of type enum whose
    possible values represent the states of the error model with the specified
    type.
    
    Returns the nominal.OutDataPort object representing the new port.
    
    The enum values will be encoded versions of the names of the error states
    for two reasons:
      - To avoid name conflicts with enum values of existing ports in case the
        component type is the type of an extended component.
      - To avoid confusion with the mode names of error subcomponents, which
        are the same as the names of the error states represented by these
        modes.
    For details about the name encoding see
    _get_enum_value_for_error_state_name.
    
    If the error model type has no starting state (that means it has not states
    at all), then an additional value will be added to the enum to represent
    the starting state.
    
    @param   component_type:   The component type to be extended by an error
                               states enum port.
    @type    component_type:   nominal.ComponentType
    @param   error_model: The error model whose states are to be represented by the enum port.
    @type    error_model: util.general.ErrorModel
    @return: The nominal.OutDataPort object representing the added port.
    @rtype:  nominal.OutDataPort
    """
    
    # Create the list of possible enum values. These values must represent the
    # error states of the specified error model.
    # Instead of using the names of the error states directly as enum values,
    # they are encoded using _get_enum_value_for_error_state_name.
    error_states = _get_error_states(error_model)
    enum_values = map(_get_enum_value_for_error_state_name, error_states.iterkeys())

    # Check whether the list of enum values is empty, which would mean that the
    # error model has no states.
    if len(enum_values) == 0:
        # The error model has no states. For the rest of the model extension
        # construction, at least one error state is required, so add a value to
        # the enum to represent this artificial state.
        # Note that this value corresponds to the artificial mode that is added
        # in _add_modes_to_error_component if the error model has no states.
        # This artificial state is considered the starting error state and the
        # corresponding enum value is the initial value of the enum port.
        new_enum_value = _get_enum_value_for_error_state_name(
                            constants.DEFAULT_STARTING_STATE_NAME)
        enum_values.append(new_enum_value)
        initial_value = new_enum_value
    else:
        # The error model/implementation has at least one state. Thus it must also have a
        # starting state. Use the enum value representing this starting state
        # as the initial value of the enum port.
        start_states = [e for e in error_states.itervalues() if e.type in {ErrorStateType.ACTIVATION, ErrorStateType.INITIAL}]
        assert len(start_states) == 1
        initial_value = _get_enum_value_for_error_state_name(start_states[0].name)
    
    # The initial value must be specified as a SLIM expression in tree
    # representation.
    initial_value_tree = expressions.IdentifierTreeNode(initial_value, 0, 0)
    
    # Create the enum data type with the adequate possible values and add the
    # outgoing enum port with this type to the component type.
    # Note: The name of the port, constants.ERROR_STATES_ENUM_PORT_NAME, is
    # guaranteed to be unique among the port names of a component.
    enum_data_type = nominal.Enum(enum_values, None, 0, 0)
    return component_type.add_port(error_model.type.file,
                                   constants.ERROR_STATES_ENUM_PORT_NAME,
                                   False,
                                   False,
                                   False,
                                   enum_data_type,
                                   initial_value_tree,
                                   False,
                                   False,
                                   0,
                                   0)
    
    
def _get_enum_value_for_error_state_name(error_state_name):
    """Returns an encoding of the specified error state name which can be used
    as a value of the enum port of an extended component or error subcomponent
    that indicates the current error state of the component.
    
    Note: The returned string will not be a valid SLIM identifier with respect
    to the syntax for nominal SLIM models. This is useful to avoid name clashes
    with existing identifiers.
    
    The current implementation of the encoding just adds the prefix "_" to the
    error state name.
    
    @param   error_state_name: The name of the error state to be encoded.
    @type    error_state_name: string
    @return: An encoding of the specified error state name.
    @rtype:  string
    """
    return "_" + error_state_name


def _get_event_port_name(error_propagation_or_event_name):
    """Returns an encoding of the specified error propagation name or error
    event name, which can be used as the name of a corresponding event port of
    the extended model.
    
    Note: Error propagations and error events are parts of an error model.
    
    Note: The returned string will not be a valid SLIM identifier with respect
    to the syntax of valid nominal models. This is useful to avoid name clashes
    with existing event ports.
    
    The current implementation of the encoding just adds the prefix "#" to the
    error propagation name.
    
    @param   error_propagation_or_event_name: The name of the error
                                              propagation or error event.
    @type    error_propagation_or_event_name: string
    @return: An encoding of the specified error propagation name or error event
             name.
    @rtype:  string
    """
    return "#" + error_propagation_or_event_name


def _add_subcomponents(extended_implementation,
                       original_implementation,
                       subcomponent_id_to_implementation_name_dict):
    """Adds the subcomponents of the specified original component
    implementation to the specified extended one, considering the new names of
    the (extended) subcomponent implementations.
    
    Note: Data subcomponents as well as non-data subcomponents will be added.
    
    @param   extended_implementation: The extended implementation to which the
                                      subcomponents shall be added.
    @type    extended_implementation: nominal.ComponentImplementation
    @param   original_implementation: The original implementation.
    @type    original_implementation: nominal.ComponentImplementation
    @param   subcomponent_id_to_implementation_name_dict:
                                       A dictionary mapping each of the
                                       identifiers of the subcomponents of the
                                       specified component to the name of the
                                       extended implementation for that
                                       subcomponent.
    @type    subcomponent_id_to_implementation_name_dict:
                                       dict (string -> string)
    """
    # Add the non-data subcomponents of the original component to the extended
    # one, considering the new names of the (extended) subcomponent
    # implementations.
    for subcomponent_declaration in \
            original_implementation.non_data_subcomponents.itervalues():
        
        extended_implementation.add_non_data_subcomponent(
                subcomponent_declaration.name,
                subcomponent_declaration.category,
                subcomponent_id_to_implementation_name_dict[
                            subcomponent_declaration.name],
                subcomponent_declaration.accesses,
                subcomponent_declaration.stored_in,
                subcomponent_declaration.running_on,
                subcomponent_declaration.in_modes,
                0,
                0)
        
    # Add the data subcomponents of the original component to the extended one.
    for subcomponent_declaration in \
            original_implementation.data_subcomponents.itervalues():
        
        extended_implementation.add_data_subcomponent(
                subcomponent_declaration.name,
                subcomponent_declaration.data_type,  # TODO: Make a copy?
                subcomponent_declaration.default_value,  # TODO: Make a copy?
                subcomponent_declaration.in_modes,
                0,
                0)
    
    
def _add_modes(extended_implementation, original_implementation):
    """Adds the modes of the specified original component implementation to the
    specified extended one.
    
    @param   extended_implementation: The extended implementation to which the
                                      modes shall be added.
    @type    extended_implementation: nominal.ComponentImplementation
    @param   original_implementation: The original implementation.
    @type    original_implementation: nominal.ComponentImplementation
    """
    for mode in original_implementation.modes.itervalues():
        # TODO: It might be better to avoid the non-public method _addMode.
        # TODO: Should a copy of the invariant be created?
        extended_implementation._addMode(mode.name, mode.invariant, mode.type, None,
                                         0, 0)
    
    
def _add_original_event_port_connections(extended_implementation,
                                         original_implementation):
    """Adds the event port connections of the specified original component
    implementation to the specified extended one.
    
    @param   extended_implementation: The extended implementation to which the
                                      port connections shall be added.
    @type    extended_implementation: nominal.ComponentImplementation
    @param   original_implementation: The original implementation.
    @type    original_implementation: nominal.ComponentImplementation
    """
    for connection in original_implementation.event_port_connections:
        extended_implementation.add_event_port_connection(
                            connection.source_port,
                            connection.target_port,
                            connection.in_modes,
                            connection.passive,
                            0, 0)


def _add_original_data_port_connections(extended_implementation,
                                        original_implementation):
    """Adds the event port connections of the specified original component
    implementation to the specified extended one.
    
    @param   extended_implementation: The extended implementation to which the
                                      port connections shall be added.
    @type    extended_implementation: nominal.ComponentImplementation
    @param   original_implementation: The original implementation.
    @type    original_implementation: nominal.ComponentImplementation
    """
    for connection in original_implementation.data_port_connections:
        extended_implementation.add_data_port_connection(
                            connection.source_port,
                            connection.target_port,
                            connection.in_modes,
                            0, 0)


def _add_original_flows(extended_implementation, original_implementation):
    """Adds the flows of the specified original component implementation to the
    specified extended one.
    
    @precondition:
        This is only suitable for extended components without error behavior.
        Otherwise _add_extended_flows must be used.
    
    @param   extended_implementation:     The extended implementation to which
                                          the flows shall be added.
    @type    extended_implementation:     nominal.ComponentImplementation
    @param   original_implementation:     The original implementation.
    @type    original_implementation:     nominal.ComponentImplementation
    """
    for flow in original_implementation.flows:
        # TODO: Should a copy of the flow expression be created?
        extended_implementation.add_flow(flow.expression, flow.in_modes, 0, 0)


def _add_extended_flows(extended_implementation,
                        original_implementation,
                        error_model,
                        fault_effects_specification):
#TODO: modify docstring
    """Adds the flows of the specified original component implementation to the
    specified extended one, taking into account the fault effects associated
    with the error states of the specified associated error model.
    
    @param   extended_implementation:     The extended implementation to which
                                          the flows shall be added.
    @type    extended_implementation:     nominal.ComponentImplementation
    @param   original_implementation:     The original implementation.
    @type    original_implementation:     nominal.ComponentImplementation
    @param   error_model_type:            The type of the error model to be
                                          incorporated in the extended
                                          implementation.
    @type    error_model_type:            error.ErrorModelType
    @param   fault_effects_specification: A specification that assigns fault
                                          effects to the states of the
                                          specified error model.
    @type    fault_effects_specification: fault_injection.\
                                          FaultEffectsSpecification
    """
    for flow in original_implementation.flows:
        _add_extended_flow(extended_implementation,
                           original_implementation,
                           error_model,
                           fault_effects_specification,
                           flow)


def _add_extended_flows_for_port_connections(extended_implementation,
                                             original_implementation,
                                             error_model,
                                             fault_effects_specification):
    """For each data port connection of the specified original component
    implementation, adds a corresponding flow to the specified extended
    implementation, taking into account the fault effects associated with the
    error states of the specified associated error model.
    
    @param   extended_implementation:     The extended implementation to which
                                          the flows shall be added.
    @type    extended_implementation:     nominal.ComponentImplementation
    @param   original_implementation:     The original implementation.
    @type    original_implementation:     nominal.ComponentImplementation
    @param   error_model:                 The error model to be
                                          incorporated in the extended
                                          implementation.
    @type    error_model:                 general.ErrorModel
    @param   fault_effects_specification: A specification that assigns fault
                                          effects to the states of the
                                          specified error model.
    @type    fault_effects_specification: fault_injection.\
                                          FaultEffectsSpecification
    """
    for connection in original_implementation.data_port_connections:
        # To add an extended flow to the extended component implementation, an
        # "original" flow (nto including the fault effects) must be provided as
        # a basis for the extended one.
        # Therefore we have to create a flow that is equivalent to the current
        # port connection.
        flow = general.create_flow_for_port_connection(extended_implementation, connection)
        _add_extended_flow(extended_implementation,
                           original_implementation,
                           error_model,
                           fault_effects_specification,
                           flow)


def _add_extended_flow(extended_implementation,
                       original_implementation,
                       error_model,
                       fault_effects_specification,
                       original_flow):
    """Adds the flows of the specified original component implementation to the
    specified extended one, taking into account the fault effects associated
    with the error states of the specified associated error model.
    
    @param   extended_implementation:     The extended implementation to which
                                          the flow shall be added.
    @type    extended_implementation:     nominal.ComponentImplementation
    @param   original_implementation:     The original implementation.
    @type    original_implementation:     nominal.ComponentImplementation
    @param   error_model:                 The error model to be
                                          incorporated in the extended
                                          implementation.
    @type    error_model:                 general.ErrorModel
    @param   fault_effects_specification: A specification that assigns fault
                                          effects to the states of the
                                          specified error model.
    @type    fault_effects_specification: fault_injection.\
                                          FaultEffectsSpecification
    @param   original_flow:               The original flow that shall be
                                          used as a basis of the extended flow
                                          to be added.
    @type    original_flow:               nominal.Flow
    """
    # The expression on the right-hand side of the extended flow must be a
    # "case" expression, where each case corresponds to a certain error state.
    # A case is a pair consisting of its condition (here: the component is in a
    # certain error state) and the expression defining the value of the flow
    # (here: the corresponding fault effect for the target of the flow).
    # The expression on the right-hand side of the original flow will be used
    # in the default case of the extended flow.
    # First, compute all cases except for the default case. 
    cases = []
    for error_state_name in _get_error_states(error_model).iterkeys():
        # Create the tree representation of the condition that is true iff
        # the extended component is in the considered error state.
        # This condition uses the enum value representing the error state,
        # so this enum value must be determined first.
        error_state_enum_value = _get_enum_value_for_error_state_name(
                                                    error_state_name)
        # Note: The error state is determined by means of the error state enum
        # port of the error subcomponent, not the error state enum port of the
        # extended component itself. This is done to avoid a violation of the
        # restrictions for flows: Their right-hand sides are not allowed to
        # refer to outgoing data ports of the component itself, only of its
        # subcomponents.
        error_states_port_name = (constants.ERROR_SUBCOMPONENT_IDENTIFIER +
                                  "." + constants.ERROR_STATES_ENUM_PORT_NAME)
        state_condition_tree = \
                util_expressions.create_slim_enum_value_condition(
                        error_states_port_name, [error_state_enum_value])
        
        # Get the associated fault effects of the current error state.
        fault_effects = fault_effects_specification.get_effects(
                                                error_state_name)
        # Create a list containing the tree representations of the fault
        # effects.
        # TODO: Maybe the fault effects specification should provide the
        # tree representations directly.
        fault_effect_trees = map(util_expressions.parse_slim_assignment,
                                 fault_effects)
        
        # Store the error state condition and the the fault effect for the flow
        # target under consideration (both in tree representation) as a case,
        # if there is such a fault effect.
        flow_target_port_name = original_flow.expression.left.name
        effect_trees_for_flow_target = \
                [fault_effect_tree
                 for fault_effect_tree in fault_effect_trees
                 if fault_effect_tree.left.name == flow_target_port_name]
        assert len(effect_trees_for_flow_target) <= 1
        if len(effect_trees_for_flow_target) == 1:
            fault_effect_tree = effect_trees_for_flow_target[0]
            cases.append((state_condition_tree, fault_effect_tree.right))
    
    # Add an extended flow (including the fault effects) separately for each
    # mode in which the original flow was active. This is necessary as some
    # fault effects might only be applicable in certain modes because they
    # refer to outgoing data ports of subcomponents that are inactive in some
    # modes. In such cases the fault effect should still be applied in all
    # modes where it is possible.
    # Note: If no modes are given in the specification of the original flow,
    # then it is active in all modes.
    # Important: The original implementation must have at least one mode for
    # this to work.
    assert len(original_implementation.modes) > 0
    mode_names = original_flow.in_modes
    if not mode_names:
        mode_names = original_implementation.modes.keys()
    for mode_name in mode_names:
        # Consider only those cases (i.e. pairs of error state condition and
        # fault effect expression) for which the fault effect is applicable in
        # the currently considered mode.
        applicable_cases = []
        for case in cases:
            if general.check_expression_is_applicable_in_flow(
                        case[1], original_implementation, mode_name):
                # TODO: Maybe a copy of the expression for the respective case
                # should be used to avoid multiple references to the same
                # expression tree in different flows. This could be done by
                # converting the expression to a string an re-parsing it.
                applicable_cases.append(case)
            else:
                warning_template = ('The fault effect "%s := %s" can not be ' +
                                    'applied in mode "%s" of the component ' +
                                    'implementation "%s".')
                warning = warning_template % (flow_target_port_name,
                                              toSLIMVisitor().visit(case[1]),
                                              mode_name,
                                              original_implementation.name)
                log.system.warning(warning)
        
        # Create the expression to be used as the right-hand side of the
        # extended flow. If there are cases representing applicable fault
        # effects, then use the right-hand side of the original flow for the
        # default case. Otherwise, just reuse the original right-hand side as
        # the right-hand side of the extended flow. (No "case" expression is
        # needed in this case.)
        if len(applicable_cases) > 0:
            # TODO: Should a copy of the right-hand side of the original flow
            # expression be created?
            case_expression = expressions.CaseExpressionTreeNode(
                                    applicable_cases,
                                    original_flow.expression.right,
                                    0, 0)
            # TODO: Should a copy of the left-hand side of the original flow
            # expression be created?
            extended_flow_expression = expressions.BinaryOperatorTreeNode(
                                            original_flow.expression.left,
                                            case_expression,
                                            expressions.OperatorType.ASSIGN,
                                            0, 0)
        else:
            # TODO: Should a copy of the flow expression be created?
            extended_flow_expression = original_flow.expression
        
        # Add the extended flow for the currently considered mode. 
        extended_implementation.add_flow(extended_flow_expression,
                                         [mode_name], 0, 0)

def _add_reactivation_transitions(extended_implementation,
                                  original_implementation):
    """Adds the reactivation transitions of the specified original component implementation
    to the specified extended one.
    
    @param   extended_implementation: The extended implementation to which the
                                      transitions shall be added.
    @type    extended_implementation: nominal.ComponentImplementation
    @param   original_implementation: The original implementation.
    @type    original_implementation: nominal.ComponentImplementation
    """
    for transition in original_implementation.reactivation_mode_transitions:
        extended_implementation.add_reactivation_mode_transition(
                transition.source_mode,
                transition.target_mode,
                None,
                None,
                transition.label.effects,
                0,
                0)

def _add_original_transitions(extended_implementation,
                              original_implementation):
    """Adds the transitions of the specified original component implementation
    to the specified extended one.
    
    @precondition:
        This is only suitable for extended components without error behavior.
        Otherwise _add_extended_transitions must be used.
    
    @param   extended_implementation: The extended implementation to which the
                                      transitions shall be added.
    @type    extended_implementation: nominal.ComponentImplementation
    @param   original_implementation: The original implementation.
    @type    original_implementation: nominal.ComponentImplementation
    """
    for transition in original_implementation.normal_mode_transitions:
        extended_implementation.add_normal_mode_transition(
                transition.source_mode,
                transition.target_mode,
                transition.label.trigger,
                transition.label.guard,  # TODO: Should a copy be made?
                None,
                transition.label.effects[:],  # Copy the list, not the expressions
                0,
                0)

def _get_error_states(error_model):
    # States must be either in the type or the implementation
    assert len(error_model.type.states)>0 or len(error_model.implementation.states) > 0
    error_states = error_model.implementation.states if len(error_model.implementation.states)>0 else error_model.type.states
    return error_states

def _add_extended_transitions(extended_implementation,
                              original_implementation,
                              error_model,
                              fault_effects_specification):
    """Adds all necessary transitions to the specified extended component
    implementation, taking into account the behavior of the specified
    associated error model and the fault effects associated with its error
    states.
    
    @param   extended_implementation:     The extended implementation to which
                                          the transitions shall be added.
    @type    extended_implementation:     nominal.ComponentImplementation
    @param   original_implementation:     The original implementation.
    @type    original_implementation:     nominal.ComponentImplementation
    @param   error_model:                 The error model that is associated
                                          with the original component.
    @type    error_model:                 general.ErrorModel
    @param   fault_effects_specification: A specification that assigns fault
                                          effects to the states of the
                                          specified error model. These effects
                                          must be valid assignments for the
                                          specified original implementation.
    @type    fault_effects_specification: fault_injection.\
                                          FaultEffectsSpecification
    """
    # Add the non-reset transitions of the original component to the extended
    # one, considering fault effects.
    _add_non_reset_transitions(extended_implementation,
                               original_implementation,
                               error_model,
                               fault_effects_specification)
    # Add the reset transitions of the original component to the extended one,
    # considering fault effects.
    _add_reset_transitions(extended_implementation,
                           original_implementation,
                           error_model,
                           fault_effects_specification)
    
    # Add the transitions that are used to apply fault effects whenever the
    # error subcomponent takes a transition.
    _add_fault_effect_transitions(extended_implementation,
                                  original_implementation,
                                  error_model.implementation,
                                  fault_effects_specification)
    
    # Adapt reset transitions to reset the time delay clock (if necessary)
    TimeDelayClockVisitor(extended_implementation.parent.parent.parent).visit(extended_implementation)

    
def _add_non_reset_transitions(extended_implementation,
                               original_implementation,
                               error_model,
                               fault_effects_specification):
    """Adds the non-reset transitions of the specified original component
    implementation to the specified extended one, considering the fault
    effects that are associated with error states.
    
    @param   extended_implementation:     The extended implementation to which
                                          the transitions shall be added.
    @type    extended_implementation:     nominal.ComponentImplementation
    @param   original_implementation:     The original implementation.
    @type    original_implementation:     nominal.ComponentImplementation
    @param   error_model:                 The error model to be
                                          incorporated in the extended
                                          implementation.
    @type    error_model:                 general.ErrorModel
    @param   fault_effects_specification: A specification that assigns fault
                                          effects to the states of the
                                          specified error model. These effects
                                          must be valid assignments for the
                                          specified original implementation.
    @type    fault_effects_specification: fault_injection.\
                                          FaultEffectsSpecification
    """
    # Construct a list containing the non-reset transitions of the original
    # component implementation.
    non_reset_transitions = \
            [transition for transition in
             original_implementation.normal_mode_transitions
             if transition.label.trigger != constants.RESET_TRIGGER]
    # For each non-reset transition of the original component, add the
    # corresponding transitions to the extended component.
    for transition in non_reset_transitions:
        # The original transition translates to multiple transitions (in
        # general) in the extended component:
        # For each error state of the extended component there must be a copy
        # of the original transition that is enabled in this state and includes
        # the fault effects for this state.
        for error_state_name in _get_error_states(error_model).iterkeys():
            # Create the tree representation of the additional guard condition
            # that is true iff the extended component is in the considered
            # error state.
            # This guard uses the enum value representing the error state, so
            # this enum value must be determined first.
            error_state_enum_value = _get_enum_value_for_error_state_name(
                                                        error_state_name)
            state_condition_tree = \
                    util_expressions.create_slim_enum_value_condition(
                                constants.ERROR_STATES_ENUM_PORT_NAME,
                                [error_state_enum_value])
            # The new transition must only be enabled if the original
            # transition would be enabled *and* the component is in the
            # considered error state.
            new_guard = \
                    util_expressions.create_conjunction_of_slim_expressions(
                                                    transition.label.guard,
                                                    state_condition_tree)
                    
            # Get the associated fault effects of the current error state.
            fault_effects = fault_effects_specification.get_effects(error_state_name)
            
            # Create a list containing the tree representations of the fault
            # effects.
            # TODO: Maybe the fault effects specification should provide the
            # tree representations directly.
            fault_effect_trees = map(util_expressions.parse_slim_assignment,
                                     fault_effects)
            # Fault effects for data ports that are targets of flows or port
            # connections must not be applied as transition effects but in a
            # flow expression, so we don't consider them in this function.
            chk_target = original_implementation.\
                            port_is_target_of_connection_or_flow
            fault_effect_trees = filter(
                    lambda effect_tree: not chk_target(effect_tree.left.name),
                    fault_effect_trees)
            # In general, only a subset of the fault effects can be applied:
            # The data elements which they refer to on the left- and right-hand
            # side of the assignment must be active in the target or source
            # mode respectively.
            applicable_effect_trees = []
            for effect_tree in fault_effect_trees:
                effect_is_applicable = general.check_effect_is_applicable(
                                                effect_tree,
                                                original_implementation,
                                                transition.source_mode,
                                                transition.target_mode)
                if effect_is_applicable:
                    applicable_effect_trees.append(effect_tree)
                else:
                    warning_template = ('The fault effect "%s" can not be ' +
                                        'applied on the following ' +
                                        'transition of the component ' +
                                        'implementation "%s": "%s".')
                    warning = (warning_template %
                               (toSLIMVisitor().visit(effect_tree),
                                original_implementation.name,
                                toSLIMVisitor().visit(transition)))
                    log.system.warning(warning)
            
            # The effects of the new transitions must be the effects of
            # original transition combined with the fault effects.
            # The fault effects must override the original effects in case of
            # conflicts.
            new_effects = util_expressions.get_slim_assignments_union(
                                            applicable_effect_trees,
                                            transition.label.effects)
            
            # Add the new transition to the extended implementation.
            extended_implementation.add_normal_mode_transition(
                                    transition.source_mode,
                                    transition.target_mode,
                                    transition.label.trigger,
                                    new_guard,
                                    None,
                                    new_effects,
                                    0,
                                    0)
            
            
def _add_reset_transitions(extended_implementation,
                           original_implementation,
                           error_model,
                           fault_effects_specification):
    """Adds the reset transitions of the specified original component
    implementation to the specified extended one, considering the fault effects
    that are associated with error states.
    
    @precondition:
        There must not be more than one reset transition leaving the same
        state in the specified error model implementation, because
        otherwise the effect of reset transitions in the extended
        component can not be determined unambiguously.
    
    @param   extended_implementation:     The extended implementation to which
                                          the transitions shall be added.
    @type    extended_implementation:     nominal.ComponentImplementation
    @param   original_implementation:     The original implementation.
    @type    original_implementation:     nominal.ComponentImplementation
    @param   error_model:                 The error model to be incorporated in
                                          the extended implementation.
    @type    error_model:                 general.ErrorModel
    @param   fault_effects_specification: A specification that assigns fault
                                          effects to the states of the
                                          specified error model. These effects
                                          must be valid assignments for the
                                          specified original implementation.
    @type    fault_effects_specification: fault_injection.\
                                          FaultEffectsSpecification
    """
    # TODO: Maybe this method should be merged with _add_non_reset_transitions.

    # Construct a list containing the reset transitions of the original
    # component implementation.
    reset_transitions = \
            [transition for transition in
             original_implementation.normal_mode_transitions
             if transition.label.trigger == constants.RESET_TRIGGER]
    # For each reset transition of the original component, add the
    # corresponding transitions to the extended component.
    for transition in reset_transitions:
        # The original transition translates to multiple transitions (in
        # general) in the extended component:
        # For each error state of the extended component there must be a copy
        # of the original transition that is enabled in this state.
        # Note that in the error subcomponent of the extended component, there
        # will be an outgoing reset transition for each error state (see
        # _add_transitions_to_error_subcomponents for details). We will call
        # these transitions "reaction transitions". (They will be synchronized
        # with the reset transitions in the extended component.)
        # The copy of the original transition (of the nominal component) that
        # is added to the extended component must include the fault effect of
        # the target state of the reaction transition leaving the considered
        # error state.   
        # Important: This means that there must not be more than one reaction
        # transition leaving a single error state, because otherwise the fault
        # effect could not be determined unambiguously.
        for error_state_name in _get_error_states(error_model).iterkeys():
            # Create the tree representation of the additional guard condition
            # that is true iff the extended component is in the considered
            # error state.
            # This guard uses the enum value representing the error state, so
            # this enum value must be determined first.
            error_state_enum_value = _get_enum_value_for_error_state_name(
                                                        error_state_name)
            state_condition_tree = \
                    util_expressions.create_slim_enum_value_condition(
                                constants.ERROR_STATES_ENUM_PORT_NAME,
                                [error_state_enum_value])
            # The new transition must only be enabled if the original
            # transition would be enabled *and* the component is in the
            # considered error state.
            new_guard = \
                    util_expressions.create_conjunction_of_slim_expressions(
                                                    transition.label.guard,
                                                    state_condition_tree)
            
            # Get a list containing the reset transition of the error model
            # that leaves the considered error state. The list will be empty if
            # no such transition exists.
            # Note: There must be at most one such transition.
            reaction_transitions = \
                   [error_transition for error_transition in
                    error_model.implementation.error_transitions
                    if (error_transition.source_state == error_state_name and
                        error_transition.trigger == constants.RESET_TRIGGER)]
            assert len(reaction_transitions) <= 1
            # Check whether the error state has an outgoing reset transition.
            if len(reaction_transitions) == 0:
                # The error state has no outgoing reset transition.
                # Note: Thus the error subcomponent will have a reset
                # transition loop for the mode that corresponds to this error
                # state. So the fault effects of this error state itself have
                # to be applied.
                fault_effects = fault_effects_specification.get_effects(
                                                        error_state_name)
            else:
                # The error state has an outgoing reset transition.
                # The fault effects of its target state have to be applied.
                fault_effects = fault_effects_specification.get_effects(
                                    reaction_transitions[0].target_state)

            # Create a list containing the tree representations of the fault
            # effects.
            # TODO: Maybe the fault injection specification should provide the
            # tree representations directly.
            fault_effect_trees = map(util_expressions.parse_slim_assignment,
                                     fault_effects)
            # Fault effects for data ports that are targets of flows or port
            # connections must not be applied as transition effects but in a
            # flow expression, so we don't consider them in this function.
            chk_target = original_implementation.\
                            port_is_target_of_connection_or_flow
            fault_effect_trees = filter(
                    lambda effect_tree: not chk_target(effect_tree.left.name),
                    fault_effect_trees)
            # In general, only a subset of the fault effects can be applied:
            # The data elements which they refer to on the left- and right-hand
            # side of the assignment must be active in the target or source
            # mode respectively.
            applicable_effect_trees = []
            for effect_tree in fault_effect_trees:
                effect_is_applicable = general.check_effect_is_applicable(
                                                effect_tree,
                                                original_implementation,
                                                transition.source_mode,
                                                transition.target_mode)
                if effect_is_applicable:
                     applicable_effect_trees.append(effect_tree)
                else:
                    warning_template = ('The fault effect "%s" can not be ' +
                                        'applied on the following ' +
                                        'transition of the component ' +
                                        'implementation "%s": "%s".')
                    warning = (warning_template %
                               (toSLIMVisitor().visit(effect_tree),
                                original_implementation.name,
                                toSLIMVisitor().visit(transition)))
                    log.system.warning(warning)

            # The effects of the new transitions must be the effects of
            # original transition combined with the fault effects.
            # The fault effects must override the original effects in case of
            # conflicts.
            new_effects = util_expressions.get_slim_assignments_union(
                                            applicable_effect_trees,
                                            transition.label.effects)
            
            # The trigger of the new mode transition to be added must be the
            # name of the reset event port of the error subcomponent (so that
            # the transition can be synchronized with the corresponding
            # reaction transition in the error subcomponent).
            trigger = (constants.ERROR_SUBCOMPONENT_IDENTIFIER + "." +
                       constants.RESET_PORT_NAME)
            
            # Add the new transition to the extended implementation.
            extended_implementation.add_normal_mode_transition(
                                    transition.source_mode,
                                    transition.target_mode,
                                    trigger,
                                    new_guard,
                                    None,
                                    new_effects,
                                    0,
                                    0)

            
def _add_fault_effect_transitions(extended_implementation,
                                  original_implementation,
                                  error_model_implementation,
                                  fault_effects_specification):
    """Add to the specified extended component implementation the
    transitions that are used to apply fault effects whenever the error
    subcomponent of the extended component takes a non-reset transition.
    
    Note: For each non-reset transition of the associated error model, a
    corresponding transition will be added to the extended component, even
    if the target error state of the error model transition has no
    associated fault effects. For error event transitions, this is
    necessary because otherwise the corresponding transition of the error
    subcomponent would be ignored when creating the configuration LTS
    since its trigger (an error event port) is not connected to another
    port. For error propagation transitions this is not required because
    they don't need to be represented in the configuration LTS if they
    don't affect the supercomponent or neighbor subcomponents.
    Nevertheless, corresponding transitions will also be added for error
    propagation transitions even if their target states have no associated
    fault effects, for the sake of simplicity. 
    
    @precondition:
        There must not be more than one transition with the same trigger
        leaving the same state in the specified error model
        implementation, because otherwise the transitions that apply the fault
        effects in the extended component will not be synchronized with one
        unique transition of the extended component.
    
    @param   extended_implementation:     The extended implementation to which
                                          the transitions shall be added.
    @type    extended_implementation:     nominal.ComponentImplementation
    @param   original_implementation:     The original implementation.
    @type    original_implementation:     nominal.ComponentImplementation
    @param   error_model_implementation:  The error model implementation to be
                                          incorporated in the extended
                                          implementation.
    @type    error_model_implementation:  error.ErrorModelImplementation
    @param   fault_effects_specification: A specification that assigns fault
                                          effects to the states of the
                                          specified error model. These effects
                                          must be valid assignments for the
                                          specified original implementation.
    @type    fault_effects_specification: fault_injection.\
                                          FaultEffectsSpecification
    """
    # For each error transition in the associated error model implementation
    # that is not a reset transition, and each mode of the extended component,
    # add a corresponding transition to the extended component.
    # - This transition must lead back to its source mode (it is a loop).
    # - The guard of this transition must be true iff the extended component is
    #   in the error state that is the source state of the error transition.
    # - The trigger must be the error event that is triggered by the
    #   correspondent of the error transition in the error subcomponent.
    #   (Note: The corresponding transition of the error subcomponent is not
    #   added here, but in _add_transitions_to_error_component.)
    # - The effects of the transition must be the fault effects of the target
    #   state of the error transition. (This is the main purpose of the added
    #   transition: to apply the fault effect.)
    # 
    # (Note that reset transitions are handled in _add_reset_transitions
    # because they are synchronized with the corresponding transitions of the
    # error subcomponent.)
    # First, construct a list containing the non-reset transitions of the
    # associated error model.
    non_reset_transitions = [transition for transition in
                             error_model_implementation.error_transitions
                             if transition.trigger != constants.RESET_TRIGGER]
    # Now add the corresponding transitions to the extended component.
    for error_transition in non_reset_transitions:
        # Get the associated fault effects of the target state of the error
        # transition.
        fault_effects = fault_effects_specification.get_effects(
                                    error_transition.target_state)
        # The target state has associated fault effects. Create a list
        # containing the tree representations of these effects.
        # TODO: Maybe the fault injection specification should provide the tree
        # representations directly.
        fault_effect_trees = map(util_expressions.parse_slim_assignment,
                                 fault_effects)
        # Fault effects for data ports that are targets of flows or port
        # connections must not be applied as transition effects but in a flow
        # expression, so we don't consider them in this function.
        chk_target = original_implementation.\
                            port_is_target_of_connection_or_flow
        fault_effect_trees = filter(
                lambda effect_tree: not chk_target(effect_tree.left.name),
                fault_effect_trees)
        
        # Create the tree representation of the guard condition, that is true
        # iff the extended component is in the source error state of the error
        # transition.
        # This guard uses the enum value representing the error state, so this
        # enum value must be determined first.
        source_state_enum_value = _get_enum_value_for_error_state_name(
                                        error_transition.source_state)
        guard = util_expressions.create_slim_enum_value_condition(
                            constants.ERROR_STATES_ENUM_PORT_NAME,
                            [source_state_enum_value])
        
        # Construct the trigger of the transitions to be added. It must be the
        # name of the outgoing error event port representing the error event or
        # error propagation that is the trigger of the original error
        # transition.
        port_name = _get_event_port_name(error_transition.trigger)
        error_trigger = constants.ERROR_SUBCOMPONENT_IDENTIFIER + "." + port_name
        # List of error triggers found
        triggers = []
        # First search for incoming error events. These are connected to the error component
        # if error_trigger is an error propagation (and the target of the connection)
        for conn in extended_implementation.event_port_connections:
            if conn.target_port == error_trigger:
                triggers.append(conn.source_port)
        # No incoming error events, the transition must originate from the error model itself
        # error_trigger is either an error event or an outgoing error propagation
        if triggers == []:
            # Error event from error subcomponent: either port or event
            triggers.append(error_trigger)
        # For each mode of the extended model, add a transition (leading to the
        # same mode) with the computed parameters.
        # Note that the modes of the extended model are the same as in the
        # original model. The modes of the original model are used here so that
        # it is not required that the modes have already be added to the
        # extended component.
        for mode_name in original_implementation.modes.iterkeys():
            # In general, only a subset of the fault effects can be applied:
            # The data elements which they refer to on the left- and right-hand
            # side of the assignment must be active in the target or source
            # mode respectively (which are identical here).
            applicable_effect_trees = []
            for effect_tree in fault_effect_trees:
                effect_is_applicable = general.check_effect_is_applicable(
                                                effect_tree,
                                                original_implementation,
                                                mode_name,
                                                mode_name)
                if effect_is_applicable:
                    applicable_effect_trees.append(effect_tree)
                else:
                    warning_template = ('The fault effect "%s" can not be ' +
                                        'applied in mode "%s" of the ' +
                                        'component implementation "%s".')
                    warning = (warning_template %
                               (toSLIMVisitor().visit(effect_tree),
                                mode_name,
                                original_implementation.name))
                    log.system.warning(warning)
            
            # For each trigger, add the transition
            for trigger in triggers:
                # TODO: Should copies of the guard and effects be created?
                extended_implementation.add_normal_mode_transition(
                        mode_name, mode_name, trigger, guard, None,
                        applicable_effect_trees, 0, 0)


def _add_error_component(subpackage, error_model):
    # TODO: Maybe this method should be split: One method to add the error
    # model type, one to add the implementaton.
    """Adds an error component to the specified subpackage representing
    the behavior of the specified error model and returns the created
    error component.
    
    @param   subpackage:  The subpackage to which the error component shall be
                          added.
    @type    subpackage:  slim_compile.Subpackage
    @param   error_model: The error model to be represented by the error
                          component.
    @type    error_model: general.ErrorModel
    @return: A pair consisting of the type and the implementation of the
             created error component.
    @rtype:  pair: (nominal.ComponentType, nominal.ComponentImplementation)
    """
    # 1) Add the error component type.
    # --------------------------------
    # Generate a name for the error component type to be created, based on the
    # name of the specified error model type, that is unique in the package
    # containing the specified subpackage.
    error_component_type_name = _generate_unique_type_name(
                                    error_model.type.name,
                                    subpackage.package)
    # Add a new error component type to the specified subpackage.
    error_component_type = subpackage.add_component_type(
            error_component_type_name,
            nominal.Category.str_to_category(
                    constants.ERROR_COMPONENT_CATEGORY),
            False,
            0,
            0)
    # Add for each error propagation of the error model a corresponding event
    # port to the error component type.
    _add_error_propagation_ports(error_component_type,
                                 error_model.type)
    # Add for each error event of the error model a corresponding outgoing
    # event port to the error component type.
    _add_error_event_ports(error_component_type,
                           error_model.implementation)
    # Add to the error component type an out data port of type enum whose
    # possible values represent the states of the error model.
    _add_error_states_port(error_component_type, error_model)
    # Add to the error component an in event port that is used to trigger reset
    # reset transitions.
    error_component_type.add_port(error_model.implementation.file, constants.RESET_PORT_NAME, True, True, False,
                                  None, None, False, False, 0, 0)
    
    # 2) Add the error component implementation.
    # ------------------------------------------
    # Compute the name of the error component implementation, using the name of
    # its type.
    # TODO: Use a constant for the implementation name suffix.
    error_component_implementation_name =  (error_component_type_name + 
                                            ".Implementation")
    # Add a new error component implementation to the specified subpackage.
    # TODO: Maybe a special category "error" should be used instead of
    # "system".
    error_component_implementation = subpackage.add_component_implementation(
            error_component_implementation_name,
            nominal.Category.str_to_category(
                    constants.ERROR_COMPONENT_CATEGORY),
            False,
            0,
            0)
    # Mark the newly created implementation as an error component
    # implementation.
    error_component_implementation.is_error_component_implementation = True
    # Add the states of the error model as modes to the error component.
    _add_modes_to_error_component(error_component_implementation, error_model)
    
    # Add clocks of the error model implementation to the error component
    _add_clocks_to_error_component(error_component_implementation, error_model.implementation)
    
    # Add the error transitions of the error model as mode transitions to the
    # error component.
    _add_transitions_to_error_component(error_component_implementation,
                                        error_model.implementation)      
    
    # Adapt reset transitions to reset the time delay clock (if necessary)
    TimeDelayClockVisitor(error_component_implementation.parent.parent.parent).visit(error_component_implementation)
    
    # Return the error component as a pair of type and implementation
    return (error_component_type, error_component_implementation)

def _add_clocks_to_error_component(error_component_implementation, error_model_impl):
    """Adds the clocks of the specified error model implementation as to the specified
    error component implementation.
    
    @param   error_component_implementation: The error component implementation
                                             to which the modes shall be added.
    @type    error_component_implementation: nominal.ComponentImplementation
    @param   error_model_implementation:     The error model implementation to be
                                             incorporated in the extended
                                             implementation.
    @type    error_model_implementation:     error.ErrorModelImplementation
    """
    if len(error_model_impl.clocks)>0:
        for clock in error_model_impl.clocks.values():
            error_component_implementation.add_data_subcomponent(clock.name,
                                                                 clock.data_type,
                                                                 clock.default_value,
                                                                 clock.in_modes,
                                                                 0,0)

def _add_modes_to_error_component(error_component_implementation, error_model):
    """Adds the states of the specified error model as modes to the specified
    error component implementation.
    
    If the specified error model type has no states then an artificial starting
    mode will be added to the error component implementation.
    
    @param   error_component_implementation: The error component implementation
                                             to which the modes shall be added.
    @type    error_component_implementation: nominal.ComponentImplementation
    @param   error_model:                 The error model to be
                                          incorporated in the extended
                                          implementation.
    @type    error_model:                 general.ErrorModel
    @raises  Exception: If one of the states of error states has an unknown
                        type.
    """
    # States must be either in the type or the implementation
    assert len(error_model.type.states.items() + error_model.implementation.states.items()) > 0
    error_states = dict(error_model.type.states.items() + error_model.implementation.states.items())
    
    # Add the states of the specified error model as modes to the specified
    # error component implementation. These modes have no invariants.
    for state in error_states.itervalues():
        if state.type == error.ErrorStateType.INITIAL:
            error_component_implementation.add_initial_mode(
                                            state.name, state.invariant, None, 0, 0)
        elif state.type == error.ErrorStateType.ACTIVATION:
            error_component_implementation.add_activation_mode(
                                            state.name, state.invariant, None, 0, 0)
        elif state.type == error.ErrorStateType.ERROR:
            error_component_implementation.add_error_mode(
                                            state.name, state.invariant, None, 0, 0)
        else:
            # TODO: Use a custom exception class.
            raise Exception("FATAL ERROR: Unknown error state type")
        
    # Check whether the list of modes of the error component is empty, which
    # would mean that the specified error model has no states.
    # In this case, add an artificial starting mode (with a special name) to
    # the error component.
    if len(error_component_implementation.modes) == 0:
        error_component_implementation.add_initial_mode(
                constants.DEFAULT_STARTING_STATE_NAME, None, None, 0, 0)
        
        
def _add_transitions_to_error_component(error_component_implementation,
                                        error_model_implementation):
    """Adds the error transitions of the specified error model
    implementation as mode transitions to the specified error component
    implementation.
    
    The resulting mode transitions will have effects to indicate the
    current error state as the value of the error states enum port of the
    error component (see _add_error_states_port).
    
    Note: For all error states of the error model represented by the
    specified error component that have no outgoing reset transition,
    additional reset transitions (loops) will be added for the corresponding
    modes of the specified error component.
    This is necessary because it must be guaranteed that the
    supercomponent of the specified error component can always take its
    reset transitions, even if the error subcomponent does not really
    react on them.
    
    @param   error_component_implementation: The implementation of the error
                                             component to which the transitions
                                             shall be added.
    @type    error_component_implementation: nominal.ComponentImplementation
    @param   error_model_implementation:     The error model implementation
                                             whose transitions are to be
                                             transformed to mode transitions.
    @type    error_model_implementation:     error.ErrorModelImplementation
    """
    for error_transition in (error_model_implementation.error_transitions +
                             error_model_implementation.reactivation_transitions):
        # Transform the transition of the error model implementation to a mode
        # transitions of the specified error component implementation. As an
        # effect of this mode transition, the value of the enum port indicating
        # the current error state must be changed to represent the target state
        # of the original error transition.
        # Create the tree representation of the assignment that sets the value
        # of the error states enum port so that it represents the target state.
        # The corresponding enum value must be determined first.
        target_state_enum_value = _get_enum_value_for_error_state_name(
                                        error_transition.target_state)
        effect = util_expressions.create_slim_assignment(
                    constants.ERROR_STATES_ENUM_PORT_NAME,
                    target_state_enum_value)
            
        # Compute the trigger of the mode transition to be added.
        # If the original error transition is a reset transitions, then the
        # name of the reset event port of the error component will be used as
        # the trigger of the mode transition.
        # Otherwise, the trigger of the mode transition is the name of the
        # error event port representing the error event (the trigger) of the
        # original error transition.
        if not error_transition.is_reactivation:
            if error_transition.trigger == constants.RESET_TRIGGER:
                mode_transition_trigger = constants.RESET_PORT_NAME
            else:
                mode_transition_trigger = _get_event_port_name(
                                            error_transition.trigger)
        
        # Add the error transition to the error component implementation.
        if error_transition.is_reactivation:
            error_component_implementation.add_reactivation_mode_transition(
                            error_transition.source_state,
                            error_transition.target_state,
                            error_transition.guard,
                            None,
                            error_transition.effects + [effect],
                            0,
                            0)
        else:
            error_component_implementation.add_normal_mode_transition(
                            error_transition.source_state,
                            error_transition.target_state,
                            mode_transition_trigger,
                            error_transition.guard,
                            None,
                            error_transition.effects + [effect],
                            0,
                            0)
    
    # The supercomponent of the error subcomponent should be able to perform a
    # reset transition even if, in the error model, no reset reaction is
    # specified for the respective error state. Since reset transitions must be
    # synchronized with the error subcomponent, they can only be performed if
    # the error subcomponent offers an outgoing reset transition in its
    # respective mode. Therefore it is necessary to add an outgoing reset
    # transition (a loop) for each mode of the error subcomponent that has no
    # outgoing reset transition yet.
    _add_missing_reset_transitions(error_component_implementation)
    
    
def _add_missing_reset_transitions(error_component_implementation):
    """Adds a reset transition (loop) for each mode of the specified error
    component that does not have an outgoing reset transition yet.
    
    @precondition:
        The error transitions of the error model represented by the error
        component with the specified implementation must already be added to
        the error component.
    
    @param   error_component_implementation: The implementation of the error
                                             component to which the transitions
                                             shall be added.
    @type    error_component_implementation: nominal.ComponentImplementation
    """
    for mode_name in error_component_implementation.modes:
        # Get the set of reset transitions leaving the current mode.
        # Note: There must be at most one such transition.
        # Note: The trigger of a reset transition in an error component is not
        # the special reset keyword (as used in the original SLIM model) but
        # the name of the reset event port.
        reset_transitions = \
                [transition for transition in
                 error_component_implementation.normal_mode_transitions
                 if transition.source_mode == mode_name and
                    transition.label.trigger == constants.RESET_PORT_NAME]
        assert len(reset_transitions) <= 1
        # If there is no reset transition leaving the current mode, add a reset
        # transition loop for the mode (without an effect).
        if len(reset_transitions) == 0:
            error_component_implementation.add_normal_mode_transition(
                    mode_name, mode_name, constants.RESET_PORT_NAME,
                    None, None, [], 0, 0)
 
        
def _add_error_propagation_forwarding_connections(extended_implementation,
                                                  error_model_type):
    """Adds to the specified extended component implementation the port
    connections that are used to forward error events from its error
    subcomponent to the component itself and vice versa.
               
    @param   extended_implementation: The component implementation to which the
                                      connections shall be added.
    @type    extended_implementation: nominal.ComponentImplementation
    @param   error_model_type:        The error model type whose propagations
                                      are represented by the error events to be
                                      forwarded.
    @type    error_model_type:        error.ErrorModelType
    """
    # Connect all error event ports of the specified extended component (which
    # represent error propagations of the specified error model type) with the
    # corresponding ports of its error subcomponent.
    for propagation in error_model_type.error_propagations.itervalues():
        # The direction of the connection depends on the direction of the error
        # propagation that is represented by the event ports.
        if propagation.type == error.ErrorPropagationType.IN:
            # Establish a connection from the extended component to its error
            # subcomponent (incoming error propagation).
            # Note that the names of the event ports representing the error
            # propagations are are encoded versions of the corresponding
            # propagation names.
            port_name = _get_event_port_name(propagation.name)
            extended_implementation.add_event_port_connection(
                    port_name,
                    constants.ERROR_SUBCOMPONENT_IDENTIFIER + "." + port_name,
                    [], False, 0, 0)
        else:
            # Establish a connection from the error subcomponent to its
            # supercomponent (outgoing error propagation).
            # Note that the names of the event ports representing the error
            # propagations are are encoded versions of the corresponding
            # propagation names.
            port_name = _get_event_port_name(propagation.name)
            extended_implementation.add_event_port_connection(
                    constants.ERROR_SUBCOMPONENT_IDENTIFIER + "." + port_name,
                    port_name,
                    [], False, 0, 0)
            
            
def _add_super_to_subcomponent_propagation_connections(
        extended_implementation,
        original_component,
        error_model_type,
        fault_injection_document):
    """Adds to the specified extended component implementation the port
    connections that are used to represent error propagation from the extended
    component to its subcomponents.
               
    @param   extended_implementation:  The component implementation to which
                                       the connections shall be added.
    @type    extended_implementation:  nominal.ComponentImplementation
    @param   original_component:       The original component whose extended
                                       implementation is specified.
    @type    original_component:       general.ComponentInstance
    @param   error_model_type:         The type of the error model to be
                                       incorporated in the extended component.
    @type    error_model_type:         error.ErrorModelType
    @param   fault_injection_document: A document that associates nominal
                                       components with error models.
    @type    fault_injection_document: fault_injection.FaultInjectionDocument
    """
    # TODO: Maybe the error model type should not be passed as a parameter
    # since it can be determined using the specified fault injection document.
    
    for subcomponent_id in \
            original_component.implementation.non_data_subcomponents:
        # Get the error model associated with the current subcomponent.
        subcomponent = original_component.get_subcomponent_instance(
                                                    subcomponent_id)
        sub_error_model = _get_associated_error_model(subcomponent,
                                                      fault_injection_document)
        # Subcomponents without an associated error model can not send error
        # propagations.
        if sub_error_model is None:
            continue
        # Get the names of the outgoing error propagations of the specified
        # original component (more precisely: of its associated error model)
        # for which there is a matching incoming error propagation in the
        # error model associated with the current subcomponent.
        out_propagation_names = \
                [propagation.name for propagation in
                 error_model_type.error_propagations.itervalues()
                 if propagation.type == error.ErrorPropagationType.OUT]
        in_propagation_names = \
                [propagation.name for propagation in
                 sub_error_model.type.error_propagations.itervalues()
                 if propagation.type == error.ErrorPropagationType.IN]
        matching_propagation_names = filter(
                lambda x: x in in_propagation_names,
                out_propagation_names)
        # In the extended component, add a connection between the error event
        # ports representing the matching error propagations.
        # These connections are active in all modes of the extended component,
        # so an empty list is passed as the "in modes" parameter.
        # TODO: Check if this is correct.
        # Note that the identifier of the subcomponent in the extended model is
        # the same as in the original model (in contrast to its implementation
        # name, which might differ).
        # The outgoing error propagation of the extended component is
        # represented by a corresponding outgoing event port of its special
        # error subcomponent.
        for propagation_name in matching_propagation_names:
            # Note that the names of the event ports representing the error
            # propagations are encoded versions of the corresponding
            # propagation names.
            port_name = _get_event_port_name(propagation_name)
            extended_implementation.add_event_port_connection(
                    constants.ERROR_SUBCOMPONENT_IDENTIFIER + "." + port_name,
                    subcomponent.path[-1] + "." + port_name,
                    [], False, 0, 0)
            

def _add_sub_to_supercomponent_propagation_connections(
        extended_implementation,
        original_component,
        error_model_type,
        fault_injection_document):
    """Adds to the specified extended component implementation the port
    connections that are used to represent error propagation from the
    subcomponents of the extended component to the extended component itself.
    
    @param   extended_implementation:  The component implementation to which
                                       the connections shall be added.
    @type    extended_implementation:  nominal.ComponentImplementation
    @param   original_component:       The original component whose extended
                                       implementation is specified.
    @type    original_component:       general.ComponentInstance
    @param   error_model_type:         The type of the error model to be
                                       incorporated in the extended component.
    @type    error_model_type:         error.ErrorModelType
    @param   fault_injection_document: A document that associates nominal
                                       components with error models.
    @type    fault_injection_document: fault_injection.FaultInjectionDocument
    """
    # TODO: Maybe the error model type should not be passed as a parameter
    # since it can be determined using the specified fault injection document.
    
    for subcomponent_id in \
            original_component.implementation.non_data_subcomponents:
        # Get the error model associated with the current subcomponent.
        subcomponent = original_component.get_subcomponent_instance(
                                                    subcomponent_id)
        sub_error_model = _get_associated_error_model(subcomponent,
                                                      fault_injection_document)
        # Subcomponents without an associated error model can not receive error
        # propagations.
        if sub_error_model is None:
            continue
        # Get the names of the outgoing error propagations of the current
        # subcomponent (more precisely: of its associated error model) for
        # which there is a matching incoming error propagation in the error
        # model associated with its supercomponent.
        out_propagation_names = \
                [propagation.name for propagation in
                 sub_error_model.type.error_propagations.itervalues()
                 if propagation.type == error.ErrorPropagationType.OUT]
        in_propagation_names = \
                [propagation.name for propagation in
                 error_model_type.error_propagations.itervalues()
                 if propagation.type == error.ErrorPropagationType.IN]
        matching_propagation_names = filter(
                lambda x: x in in_propagation_names,
                out_propagation_names)
        # In the extended component, add a connection between the error event
        # ports representing the matching error propagations.
        # These connections are active in all modes of the extended component,
        # so an empty list is passed as the "in modes" parameter.
        # TODO: Check if this is correct.
        # Note that the identifier of the subcomponent in the extended model is
        # the same as in the original model (in contrast to its implementation
        # name, which might differ).
        # The incoming error propagation of the extended component is
        # represented by a corresponding incoming event port of its special
        # error subcomponent.
        for propagation_name in matching_propagation_names:
            # Note that the names of the event ports representing the error
            # propagations are encoded versions of the corresponding
            # propagation names.
            port_name = _get_event_port_name(propagation_name)
            extended_implementation.add_event_port_connection(
                    subcomponent_id + "." + port_name,
                    constants.ERROR_SUBCOMPONENT_IDENTIFIER + "." + port_name,
                    [], False, 0, 0)
            
            
def _add_subcomponents_propagation_connections(extended_implementation,
                                               original_component,
                                               fault_injection_document):
    """Adds to the specified extended component implementation the port
    connections that are used to represent error propagation between its
    subcomponents.
    
    @param   extended_implementation:  The component implementation to which
                                       the connections shall be added.
    @type    extended_implementation:  nominal.ComponentImplementation
    @param   original_component:       The original component whose extended
                                       implementation is specified.
    @type    original_component:       general.ComponentInstance
    @param   fault_injection_document: A document that associates nominal
                                       components with error models.
    @type    fault_injection_document: fault_injection.FaultInjectionDocument
    """
    # Store all non-data subcomponent declarations ("scd") that have an
    # associated error model together with that error model ("em").
    # Subcomponents without an associated error model do not have to be
    # considered in the following.
    scd_and_em_pairs = []
    for subcomponent_declaration in original_component.implementation.\
                                        non_data_subcomponents.values():
        subcomponent = original_component.get_subcomponent_instance(
                                        subcomponent_declaration.name)
        error_model = _get_associated_error_model(subcomponent,
                                                  fault_injection_document)
        if error_model is not None:
            scd_and_em_pairs.append((subcomponent_declaration, error_model))
    
    # For each pair of subcomponents with an error model, add the corresponding
    # error propagation connections, if necessary.
    for (subcomponent_declaration1, error_model1) in scd_and_em_pairs:
        for (subcomponent_declaration2, error_model2) in scd_and_em_pairs:
            # Only physically connected pairs of subcomponents can communicate
            # via error propagations.
            if not general.check_physical_connection(
                    subcomponent_declaration1, subcomponent_declaration2):
                continue

            # Get the names of the outgoing error propagations of the first
            # subcomponent (more precisely: of its associated error model) for
            # which there is a matching incoming error propagation in the error
            # model associated with the second subcomponent.
            out_propagation_names = \
                    [propagation.name for propagation in
                     error_model1.type.error_propagations.itervalues()
                     if propagation.type == error.ErrorPropagationType.OUT]
            in_propagation_names = \
                    [propagation.name for propagation in
                     error_model2.type.error_propagations.itervalues()
                     if propagation.type == error.ErrorPropagationType.IN]
            matching_propagation_names = filter(
                    lambda x: x in in_propagation_names,
                    out_propagation_names)
            # In the extended component, add a connection between the error
            # event ports representing the matching error propagations.
            # These connections are active in all modes of the extended
            # component, so an empty list is passed as the "in modes"
            # parameter.
            # TODO: Check if this is correct.
            # Note that the name of the subcomponents in the extended model is
            # the same as in the original model (in contrast to its
            # implementation name, which might differ).
            for propagation_name in matching_propagation_names:
                # Note that the names of the event ports representing the error
                # propagations are encoded versions of the corresponding
                # propagation names.
                port_name = _get_event_port_name(propagation_name)
                extended_implementation.add_event_port_connection(
                        subcomponent_declaration1.name + "." + port_name,
                        subcomponent_declaration2.name + "." + port_name,
                        [], False, 0, 0)
                
                
def _add_as_error_subcomponent(extended_implementation,
                               error_component_implementation):
    """Adds the error component, whose implementation is specified, as a
    subcomponent of the extended component and returns the resulting
    nominal.NonDataSubcomponent object.
    
    @precondition:
        The error subcomponent must be in the same package as the extended
        component.
               
    @param   extended_implementation:        The extended component
                                             implementation to which the
                                             subcomponent shall be added.
    @type    extended_implementation:        nominal.ComponentImplementation
    @param   error_component_implementation: The implementation name of the
                                             error component.
    @type    error_component_implementation: nominal.ComponentImplementation
    @return: The nominal.NonDataSubcomponent object representing the error
             subcomponent.
    @rtype:  nominal.NonDataSubcomponent
    """
    # Note: constants.ERROR_SUBCOMPONENT_IDENTIFIER is used as the identifier
    # for the error subcomponent. This name is guaranteed to be unique among
    # the subcomponent identifiers of any given supercomponent.
    return extended_implementation.add_non_data_subcomponent(
            constants.ERROR_SUBCOMPONENT_IDENTIFIER,
            error_component_implementation.category,
            error_component_implementation.name,
            [], [], [], [], 0, 0)
    

# The following function is cached because it might be called very often with
# the same arguments, which was identified as a major performance bottleneck.
# TODO: Importing the caching decorator from the property_management package
# introduces unnecessary coupling. Maybe the caching decorator should be copied
# or moved to a common module.
from property_management.src import util
@util.caching_function
def _get_error_model_association(component, fault_injection_document):
    """Returns the error model association object for the specified SLIM
    component.
    
    @param   component:                The component to be considered.
    @type    component:                general.ComponentInstance
    @param   fault_injection_document: A document that associates nominal
                                       components with error models.
    @type    fault_injection_document: fault_injection.FaultInjectionDocument
    @return: The error model association for the specified SLIM component, or
             None, if there is no such association.
    @rtype:  fault_injection.ErrorModelAssociation
    """
    return fault_injection_document.get_by_component_path(component.path)


def _get_associated_error_model(component, fault_injection_document):
    """Returns the object representing the error model that is associated with
    with the specified SLIM component.
    
    @param   component:                The component to be considered.
    @type    component:                general.ComponentInstance
    @param   fault_injection_document: A document that associates nominal
                                       components with error models.
    @type    fault_injection_document: fault_injection.FaultInjectionDocument
    @return: The error model (consisting of a type and an implementation) that
             is associated with the specified component, or None if the
             component has no associated error model.
    @rtype:  general.ErrorModel or NoneType.
    """
    error_model_association = _get_error_model_association(
                                component, fault_injection_document)
    if error_model_association is None:
        return None
    else:
        compile_info = component.implementation.subpackage.package.\
                                                compile_information
        return fault_injection.get_error_model_for_association(
                                error_model_association, compile_info)


def _get_fault_effects_specification(component, fault_injection_document):
    """Returns the specification that defines the fault effects to be injected
    into the specified SLIM component.
    
    @param   component:                The component to be considered.
    @type    component:                general.ComponentInstance
    @param   fault_injection_document: A document that associates nominal
                                       components with error models.
    @type    fault_injection_document: fault_injection.FaultInjectionDocument
    @return: The specification that defines the fault effects to be injected
             into the specified SLIM component.
    @rtype:  fault_injection.FaultEffectsSpecification
    """
    error_model_association = _get_error_model_association(
                                component, fault_injection_document)
    if error_model_association is None:
        return None
    else:
        return error_model_association.effects_specification
