# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


"""This module provides the classes to represent errors in fault injection
documents.
"""
# TODO: In this file, a maximum line length of 80 is used, instead of 79 as
# specified in PEP8.


class FaultInjectionDocumentBaseError(Exception):
    """The base class for errors related to fault injection."""
    pass


class NonWellFormedXmlError(FaultInjectionDocumentBaseError):
    """An object indicating that a certain fault injection document is not a
    well-formed XML document.
    """
    
    def __init__(self, error_description):
        """Initializes the NonWellFormedXmlError object.
        
        @param   error_description: The description of the syntax error.
        @type    error_description: string
        """
        self.description = error_description
        
    def __str__(self):
        """Returns a string representation of this syntax error.
        
        @return: The description of this syntax error.
        @rtype:  string
        """
        return self.description


class DocumentSyntaxError(FaultInjectionDocumentBaseError):
    """An object representing a syntax error in a fault injection document.
    
    Syntax errors are detected by validating the document against an XML
    Schema Definition.
    """
    
    def __init__(self, error_description):
        """Initializes the DocumentSyntaxError object.
        
        @param   error_description: The description of the syntax error.
        @type    error_description: string
        """
        self.description = error_description
        
    def __str__(self):
        """Returns a string representation of this syntax error.
        
        @return: The description of this syntax error.
        @rtype:  string
        """
        return self.description
    

class ErrorModelAssociationErrorWrapper(FaultInjectionDocumentBaseError):
    """An object indicating that a fault injection document contains an
    invalid error model association.
    
    This is a wrapper for objects representing a specific error model
    association error. It provides additional information about which
    association contains the error.
    """
    
    def __init__(self, original_error, component_path,
                 error_model_package_name, error_model_implementation_name):
        """Initializes the ErrorModelAssociationErrorWrapper object.
        
        @param   original_error:
                    The object representing the specific error.
        @type    original_error:
                    FaultInjectionDocumentBaseError
        @param   component_path:
                    The sequence of identifiers that is used in the incorrect
                    error model association to reference a certain component in
                    the SLIM model.
        @type    component_path:
                    A tuple of strings.
        @param   error_model_package_name:
                    The package name of the error model implementation that is
                    referenced in the incorrect error model association.
        @type    error_model_package_name:
                    string
        @param   error_model_implementation_name:
                    The error model implementation name that is referenced in
                    the incorrect error model association.
        @type    error_model_implementation_name:
                    string
        """
        self.original_error = original_error
        self.component_path = component_path
        self.error_model_package_name = error_model_package_name
        self.error_model_implementation_name = error_model_implementation_name
        
    def __str__(self):
        """Returns a human-readable string representation of this error.
        
        @return: A human-readable error message.
        @rtype:  string
        """
        intro_template = ('The error model association associating the ' +
                          'component "%s" with the error model ' +
                          'implementation "%s" (in package "%s") is ' +
                          'incorrect:\n')
        component_name = _get_readable_full_component_name(
                                        self.component_path)
        intro = intro_template % (component_name,
                                  self.error_model_implementation_name,
                                  self.error_model_package_name)      
        return intro + str(self.original_error)


class InvalidComponentPathError(FaultInjectionDocumentBaseError):
    """An object indicating that a certain sequence of identifiers is not a
    valid path to a component in a SLIM model.
    """

    def __init__(self, root_implementation, valid_path_segment, rest_of_path):
        """Initializes the InvalidComponentPathError object.
        
        @param   root_implementation:  The implementation of the root component
                                       of the SLIM model.
        @type    root_implementation:  nominal.ComponentImplementation
        @param   valid_path_segment:   The sequence of subcomponent identifiers
                                       at the beginning of the path, that is
                                       still a valid path.
        @type    valid_path_segment:   A tuple of strings.
        @param   rest_of_path:         The rest of the path, starting with the
                                       first identifier that makes the path
                                       invalid. This sequence must not be
                                       empty.
        @type    rest_of_path:         A tuple of strings.
        """
        self.root_implementation = root_implementation
        self.valid_path_segment = valid_path_segment
        self.rest_of_path = rest_of_path
        
    path = property(lambda self: self.valid_path_segment + self.rest_of_path,
                    doc="The complete (invalid) path.")
    
    def __str__(self):
        """Returns a human-readable string representation of this error.
        
        @return: A human-readable error message.
        @rtype:  string
        """
        template = ('The specified component path ("%s") is invalid: There ' +
                    'is no subcomponent "%s" in the component "%s".')
        path_as_string = _get_readable_full_component_name(self.path)
        valid_component_name = _get_readable_full_component_name(
                                        self.valid_path_segment)
        return template % (path_as_string, self.rest_of_path[0],
                           valid_component_name)


class NonexistentPackageError(FaultInjectionDocumentBaseError):
    """An object indicating that a fault injection specification refers to a
    package that does not exist in the considered SLIM model.
    """
    
    def __init__(self, package_name):
        """Initializes the NonexistentPackageError object.
        
        @param   package_name: The name of the nonexistent package.
        @type    package_name: string
        """
        self.package_name = package_name
        
    def __str__(self):
        """Returns a human-readable string representation of this error.
        
        @return: A human-readable error message.
        @rtype:  string
        """
        return ('The specified package does not exist: "%s"' %
                self.package_name)


class NonexistentErrorModelImplementationError(
                FaultInjectionDocumentBaseError):
    """An object indicating that a fault injection specification refers to an
    error model implementation that does not exist in the considered SLIM
    model.
    """
    
    def __init__(self, implementation_name, filePaths):
        """Initializes the NonexistentErrorModelImplementationError object.
        
        @param   implementation_name: The name of the nonexistent error model
                                      implementation.
        @type    implementation_name: string
        @param   filePaths:           The paths to the provided SLIM files
        @type    filePaths:           list
        """
        self.implementation_name = implementation_name
        self.filePaths = filePaths
        
    def __str__(self):
        """Returns a human-readable string representation of this error.
        
        @return: A human-readable error message.
        @rtype:  string
        """
        template = 'Nominal component "%s" is not declared in %s'
        return template % (self.implementation_name, ", ".join(('"'+f+'"') for f in self.filePaths))


class NonexistentErrorStateError(FaultInjectionDocumentBaseError):
    """An object indicating that a fault injection specification associates
    a fault effect with an error state that does not exist in the respective
    error model.
    """
    
    def __init__(self, error_state_name):
        """Initializes the NonexistentErrorStateError object.
        
        @param   error_state_name:      The name of the nonexistent error
                                        state.
        @type    error_state_name:      string
        """
        self.error_state_name = error_state_name
        
    def __str__(self):
        """Returns a human-readable string representation of this error.
        
        @return: A human-readable error message.
        @rtype:  string
        """
        return ('A fault effect has been associated with the error state ' +
                '"%s", but there is no such error state in the associated ' +
                'error model.') % self.error_state_name
        
        
class InvalidFaultEffectError(FaultInjectionDocumentBaseError):
    """An object representing an error in a fault effect."""
    
    def __init__(self, error_description):
        """Initializes the InvalidFaultEffectError object.
        
        @param   error_description: The description of the error.
        @type    error_description: string
        """
        self.description = error_description
        
    def __str__(self):
        """Returns a string representation of this error.
        
        @return: The description of this error.
        @rtype:  string
        """
        return self.description


# TODO: Review this function and its use cases. Maybe it is not really needed.
def _get_readable_full_component_name(path):
    """Returns the full name of a component as a human readable string.
    
    This does not return the name of the component implementation, but a string
    representing the path from the root component of the SLIM model.
    Example:
    >>> _get_readable_full_component_name(('sub1', 'sub2', 'sub3'))
    'sub1.sub2.sub3'
    
    If the specified path is empty (which stands for the root component), the
    string '<root>' will be returned.
    
    Important: The resulting string is not intended to be used for further
    computations, but only for user information. In general, it is not a valid
    component path!
    
    @param   path: A sequence of subcomponent identifiers, forming a path from
                   the root component of the SLIM model.
    @type    path: A tuple of strings.
    @return: A human readable component name based on the specified path.
    """
    if not path:
        return "<root>"
    else:
        return '.'.join(path)
