# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================
from slim_compiler.src.dataStructure.visitors.expressionVisitor import AnnotateVisitor,\
    TypeVisitor
from functools import partial


"""This module provides features to handle fault injection documents.

Fault injection documents define associations between nominal components and
error models of a SLIM model.

There are two main use cases for this module:
  - If you want to parse an existing fault injection document file (in XML
    format) and obtain the corresponding data structure, use the parse-method
    provided by this module. You will get a FaultInjectionDocument object.
    You can then inspect and/or modify this data structure.
  - If you want to create a new fault injection document, use the factory
    methods ("create") of the classes in this module to create and compose the
    parts of the new document. To create a complete document, the root element
    of the data structure that you construct must be a FaultInjectionDocument
    object.
"""
# TODO: In this file, a maximum line length of 80 is used, instead of 79 as
# specified in PEP8.

import os.path
import copy

from lxml import etree

from slim_compiler.src.model_extension import errors
from slim_compiler.src.model_extension.util import general
import slim_compiler.src.model_extension.util.expressions as util_expressions
from slim_compiler.src import testexpr
from slim_compiler.src.dataStructure import expressions
from slim_compiler.src.errorManagement import errors as compile_errors
# TODO: The following imports might are not strictly needed: They define
# classes and this module works with instances of these classes, but never
# instantiates these classes, so there is no strong connection to the imported
# modules.
from slim_compiler.src.dataStructure import nominalbehaviour
from slim_compiler.src.dataStructure import errorbehaviour


def parse(data_source):
    """Parses the fault injection document from the specified data source and
    returns a representation of the document using the data structures defined
    in this module.
    
    The representation of the document will be in tree structure, the root
    element being a FaultInjectionDocument object.
    
    Note: The document will not be validated. This must be done explicitly
    using the function get_errors_in_injection_document.
    
    @param   data_source: The source of the data to be parsed. This may be a
                          filename or a file-like object.
    @type    data_source: A string or a file-like object.
    @return: A tree representation of the document.
    @rtype:  FaultInjectionDocument
    """
    # TODO: Maybe this functjion should also validate the document with respect
    # to the adequate XML Schema Definition.
    # See get_syntax_errors_in_injection_document.
    
    # The parser that is used is created when this module is imported (see the
    # end of this file).
    # Note that the parser uses a lookup scheme in order to automatically map
    # certain XML elements to instances of the adequate classes which are
    # defined in this module.
    # The result of the parsing process will be an etree._ElementTree object.
    # Return the root element of this tree (a FaultInjectionDocument object).
    try:
        document_tree = etree.parse(data_source, _parser)
    except etree.XMLSyntaxError, e:
        raise errors.NonWellFormedXmlError(str(e))
    return document_tree.getroot()


# TODO: Maybe the following function should be moved to another module in order
# to separate the data structures for fault injection documents from their
# interpretation in the context of the SLIM compiler data structures.
def get_errors_in_fault_effect(fault_effect, component_implementation):
    """Validates the specified fault effect (in SLIM syntax) with respect to
    the specified nominal component implementation and returns a list
    containing the detected errors.
    
    A fault effect is invalid if it contains syntactic errors or if it is not
    applicable to the specified nominal component implementation.
    
    @param   fault_effect: The fault effect to be checked.
    @type    fault_effect: string
    @param   component_implementation: The object representing the nominal
                                       component implementation into which the
                                       fault effect is intended to be injected.
    @type    component_implementation: slim_compiler.src.dataStructure.\
                                       nominalbehaviour.ComponentImplementation
    @return: A list containing the errors in the fault effect. This list will
             be empty if the fault effect applicable to the specified nominal
             component implementation.
    @rtype:  A list of errors.InvalidFaultEffectError objects. 
    """
    # Initialize the list of errors to be returned.
    error_list = []
    # Validate the fault effect. Add an adequate error object to the list if
    # the fault effect is invalid.
    try:
        # Get a tree representation for the specified fault effect using the
        # data structures of the SLIM compiler. If the fault effect is
        # syntactically incorrect, an exception will be raised.
        effect_tree = util_expressions.parse_slim_assignment(fault_effect)
    except testexpr.InvalidExpressionException, e:
        error_list.append(errors.InvalidFaultEffectError(str(e)))
        # No further checks can be performed on syntactically invalid effects.
        return error_list
    
    # No syntax errors have been found so far in the fault effect.
    # There are two types of fault effects:
    # 1) The left-hand side of the effect is a data element that is target of a
    #    flow or port connection.
    # 2) The left-hand side of the effect is a data element that is not target
    #    of a flow or port connection.
    # The first type of fault effect must be applied using flow expressions
    # while the second type must be applied using transition effects.
    # Check the type of the fault effect and validate it with respect to the
    # rules for flow expressions (for the first type) or transition effects
    # (for the second type).
    # 
    # The validation is done as a side effect of the annotation methods in
    # nominalbehaviour.ComponentImplementation. They will attach error objects
    # to the error collector of the compile information object containing the
    # component implementation if the fault effect is not valid.
    # First, get the compile information object containing the component
    # implementation.
    compile_info = component_implementation.subpackage.package.\
                                            compile_information
    # To make sure that there are no error objects in the error collector that
    # stem from other parts of the model, replace the error collector with a
    # new one which does not contain errors.
    # Remember the old error collector in order to plug it back in after the
    # check.
    original_error_collector = compile_info.error_collector
    compile_info.error_collector = compile_errors.ErrorCollector()
    # The errors will be stored in the collector along with information about
    # the file in which they occurred and their position within that file.
    # In our case, the errors are not really located in a file. Remove the
    # misleading file information from the component implementation to avoid
    # problems. Remember this information in order to restore it after the
    # check.
    original_filename = component_implementation.file
    component_implementation.file = None
    
    # The validation of the fault effect is not specific to a certain mode of
    # the component. Thus, the status of the subcomponents referenced in the
    # fault effect (active or inactive) shall not be taken into account.
    # In order to do circumvent such checks, we replace the value of the
    # attribute "in_modes" of all subcomponent declarations in the specified
    # component implementation with an empty list, indicating that the
    # subcomponent is always active. The original values will be restored after
    # the check.
    # XXX: This is a very dirty hack!
    original_mode_names_dict = {}
    for sc_dec in component_implementation.non_data_subcomponents.values():
        original_mode_names_dict[sc_dec.name] = sc_dec.in_modes
        sc_dec.in_modes = []
    for sc_dec in component_implementation.data_subcomponents.values():
        original_mode_names_dict[sc_dec.name] = sc_dec.in_modes
        sc_dec.in_modes = []
    
    # Now perform the checks. Do this in a "try" block so that the clean-up
    # work can be done in the "finally" part, even if something goes wrong.
    try:
        # Setup annotation and type visitors
        annotator = AnnotateVisitor(compile_info)
        type_checker = TypeVisitor(compile_info)

        # Perform the checks depending on the type of the fault effect.
        # Note: The top level operator of the fault effect must be the binary
        # assignment operator and on the left-hand side, there must be a SLIM
        # identifier. This is guaranteed by util_expressions.parse_slim_assignment.
        assert isinstance(effect_tree, expressions.BinaryOperatorTreeNode)
        assert isinstance(effect_tree.left, expressions.IdentifierTreeNode)
        if component_implementation.port_is_target_of_connection_or_flow(
                                                    effect_tree.left.name):
            # The fault effect shall be applied using a flow expression.
            # Perform the checks with respect to an arbitrary mode (the first
            # one).
            assert len(component_implementation.modes) > 0
            mode_name = component_implementation.modes.keys()[0]

            annotator.annotation_meth = partial(component_implementation.get_data_element_type_lhs_flow,
                                                flow = None, in_modes = None)
            annotator.visit(effect_tree.left)
            annotator.annotation_meth = partial(component_implementation.get_data_element_type_rhs_flow,
                                                flow = None, in_modes = None)
            annotator.visit(effect_tree.right)
        else:
            # The fault effect shall be applied using a transition effects.
            # Perform the checks with respect to an arbitrary source and target
            # mode (the first one).
            assert len(component_implementation.modes) > 0
            mode_name = component_implementation.modes.keys()[0]
            annotator.annotation_meth = partial(component_implementation.get_data_element_type_lhs_effect,
                    target_mode = mode_name, trans = None)
            annotator.visit(effect_tree.left)
            annotator.annotation_meth = partial(component_implementation.get_data_element_type_rhs_effect,
                    source_mode = mode_name, trans = None)
            annotator.visit(effect_tree.right)

        # Type check the expression
        type_checker.visit(effect_tree)
        # Check whether the data element is non-continuous
        type = effect_tree.left.node_type
        if type == expressions.NodeType.CLOCK or type == expressions.NodeType.CONTINUOUS:
            error_list.append(errors.InvalidFaultEffectError(
                    "Clock or continuous data-subcomponent cannot be target of a fault injection"))
    finally:
        # Clean up: Restore the original "in_modes" attributes of the
        # subcomponent declaration objects.
        for sc_dec in component_implementation.non_data_subcomponents.values():
            sc_dec.in_modes = original_mode_names_dict[sc_dec.name]
        for sc_dec in component_implementation.data_subcomponents.values():
            sc_dec.in_modes = original_mode_names_dict[sc_dec.name]
        # Store the errors.
        # TODO: Store the errors as separate objects.
        if not compile_info.error_collector.has_no_errors():
            # Each error object has an attribute to indicate the SLIM file in which
            # the error occurred. However, in this case, the error is not really in
            # a SLIM file. Remove the file information.
            # This is a workaround to avoid runtime errors, which can occur in
            # compile_errors.ErrorCollector._calculate_line_offset if the error
            # is not really located in the SLIM file that is indicated by the error
            # object. A cleaner solution should be found for this problem.
            for error in compile_info.error_collector.errors:
                error.file = None
            error_message = str(compile_info.error_collector)
            error_list.append(errors.InvalidFaultEffectError(error_message))
        # Clean up: Plug the original error collector back in to restore the
        # original state of the compile information object, and reset the
        # filename of the component implementation.
        compile_info.error_collector = original_error_collector
        component_implementation.file = original_filename

    return error_list


# TODO: Maybe the following function should be moved to another module in order
# to separate the data structures for fault injection documents from their
# interpretation in the context of the SLIM compiler data structures.
def get_errors_in_error_model_association(error_model_association,
                                          root_implementation):
    """Validates the specified error model association with respect to the SLIM
    model that is determined by the specified root component implementation and
    returns a list containing the detected errors.
    
    @param   error_model_association: The error model association.
    @type    error_model_association: ErrorModelAssociation
    @param   root_implementation:     The implementation of the root component
                                      of the SLIM model to be considered.
    @type    root_implementation:     nominalbehaviour.ComponentImplementation
    @return: A list containing the errors in the error model association. This
             list will be empty if the association is valid.
    @rtype:  A list of instances of subclasses of
             errors.FaultInjectionDocumentBaseError. This may be one of the
             following:
                 - errors.InvalidComponentPathError
                 - errors.NonexistentPackageError
                 - errors.NonexistentErrorModelImplementationError
                 - errors.NonexistentErrorStateError
                 - errors.InvalidFaultEffectError
    """
    # Initialize the list of errors which will be returned.
    error_list = []
    
    # An error model association refers to a specific component of a SLIM
    # model. The component is determined by a sequence of subcomponent
    # identifiers, forming a path in the component hierarchy (which is a tree)
    # from the root component.
    path = error_model_association.component_path
    # Try to get an object representing the component that is determined by
    # the given path.
    # If this succeeds, this means that the component exists. In this case,
    # check also the validity of all fault effects specified in the error model
    # association with respect to the implementation of that component.
    try:
        component = general.ComponentInstance(root_implementation, path)
    except errors.InvalidComponentPathError, error:
        error_list.append(error)
    else:
        effects_specification = error_model_association.effects_specification
        for error_state in effects_specification.get_effective_states():
            for effect in effects_specification.get_effects(error_state):
                # Add all detected errors to the error list.
                error_list += get_errors_in_fault_effect(
                                    effect, component.implementation) 
    
    # Try to retrieve the error model whose implementation is referenced by the
    # specified error model association.
    # If the error model could be found, check also whether the error states,
    # for which there are fault effects specified, exist in that error model.
    # The object that represents the whole SLIM model is needed for this.
    compile_info = root_implementation.subpackage.package.compile_information
    try:
        error_model = get_error_model_for_association(error_model_association,
                                                      compile_info)
    except errors.NonexistentPackageError, error:
        error_list.append(error)
    except errors.NonexistentErrorModelImplementationError, error:
        error_list.append(error)
    else:
        # The error model could be found.
        # Check the existence of the error states for which there are fault
        # effects specified.
        # TODO: Check that the error states in the fault effects specification
        # are not the starting state of the error model.
        effects_specification = error_model_association.effects_specification
        for state in effects_specification.get_effective_states():
            if state not in error_model.type.states and state not in error_model.implementation.states:
                error_list.append(errors.NonexistentErrorStateError(state))
                    
    return error_list


def get_syntax_errors_in_injection_document(fault_injection_document):
    """Validates the specified fault injection document against the adequate
    XML Schema Definition and returns a list containing the syntactic errors.
    
    Note: Syntactic validation does not include a check of the error model
    associations with respect to a certain SLIM model.
    
    @param   fault_injection_document: The root element of what is supposed to
                                       be a fault injection document.
    @type    fault_injection_document: etree._Element
    @return: A list containing the syntactic errors in the specified document.
             This list will be empty if the document is valid with respect to
             the adequate XML Schema Definition.
    @rtype:  A list of errors.DocumentSyntaxError objects. 
    """
    # Initialize the list of errors to be returned.
    error_list = []
    # Validate the document. If it is invalid, an exception will be raised
    # which can be used to create an error object with an adequate error
    # description.
    # Note: The XML Schema object that is used for validation is created when
    # this module is imported (see the end of this file).
    try:
        _fault_injection_schema.assertValid(fault_injection_document)
    except etree.DocumentInvalid, e:
        error_list.append(errors.DocumentSyntaxError(str(e)))
    
    return error_list


# TODO: Maybe the following function should be moved to another module in order
# to separate the data structures for fault injection documents from their
# interpretation in the context of the SLIM compiler data structures.
def get_errors_in_injection_document(fault_injection_document,
                                     root_implementation):
    """Checks the specified fault injection document and returns a list
    containing the detected errors.
    
    This is a complete check of the fault injection document. The syntax of
    the document with respect to the adequate XML Schema Definition will be
    checked as well as its applicability with respect to the SLIM model that
    is determined by the specified root component implementation.
    
    @param   fault_injection_document: The fault injection document.
    @type    fault_injection_document: FaultInjectionDocument
    @param   root_implementation:      The implementation of the root component
                                       of the SLIM model to be considered.
    @type    root_implementation:      nominalbehaviourComponentImplementation
    @return: A list containing the errors in the fault injection document.
             This list will be empty if the document is valid.
    @rtype:  A list of objects of the following classes:
                - errors.DocumentSyntaxError for errors with respect to the
                    XML Schema Definition for fault injection documents
                - errors.ErrorModelAssociationErrorWrapper for errors in error
                    model associations.
    """
    # Check for syntax errors in the specified document with respect to the
    # adequate XML Schema definition. If such errors exist, abort the checking
    # process because the other checks can not be performed on a syntactically
    # invalid document.
    # TODO: If the rest of this function was wrapped in a try/catch block, it
    # might be possible to continue even after a syntax error is found.
    syntax_errors = get_syntax_errors_in_injection_document(
                                    fault_injection_document)
    if len(syntax_errors) > 0:
        return syntax_errors
    
    # The document is valid with respect to the XML Schema definition. Check
    # for errors with respect to the specified SLIM model.
    # Initialize the list of errors which will be returned.
    error_list = []
    # Check each error model association in the document. If errors are
    # detected, wrap each of them in an error object that provides additional
    # information about which error model association contains the error.
    for error_model_association in fault_injection_document:
        association_errors = get_errors_in_error_model_association(
                                            error_model_association,
                                            root_implementation)
        for original_error in association_errors:
            error_wrapper = errors.ErrorModelAssociationErrorWrapper(
                original_error,
                error_model_association.component_path,
                error_model_association.error_model_package_name,
                error_model_association.error_model_implementation_name)
            error_list.append(error_wrapper)

    return error_list


# TODO: Maybe the following function should be moved to another module in order
# to separate the data structures for fault injection documents from their
# interpretation in the context of the SLIM compiler data structures.
def get_error_model_for_association(error_model_association, compile_info):
    """Returns the error model whose implementation is referenced by the
    specified error model association.
    
    @param   error_model_association: The error model association.
    @type    error_model_association: ErrorModelAssociation
    @param   compile_info:            The object representing the SLIM model.
    @type    compile_info:            slim_compile.CompileInformation
    @return: The error model whose implementation is referenced by the
             specified error model association. An error model consists of its
             type and its implementation.
    @rtype:  general.ErrorModel
    @raise   errors.NonexistentPackageError:
                If the error model association refers to a package that does
                not exist in the SLIM model.
    @raise   errors.NonexistentErrorModelImplementationError:
                If the error model association refers to an error model
                implementation that does not exist in the SLIM model.
    """
    # First find the package that is supposed to contain the error model
    # implementation.
    package_name = error_model_association.error_model_package_name
    try:
        package = compile_info.packages[package_name]
    except KeyError:
        raise errors.NonexistentPackageError(package_name)
    # Now find the error model implementation in the package.
    # False is passed as the second parameter of get_error_implementation so
    # that both the public and the private subpackage are searched.
    implem_name = error_model_association.error_model_implementation_name
    error_model_implementation = package.get_error_implementation(implem_name,
                                                                  False)
    if not error_model_implementation:
        raise errors.NonexistentErrorModelImplementationError(implem_name,compile_info.files)
    # The error model implementation is found.
    # Get the corresponding error model type. The type of an error model is
    # always in the same package as its implementation.
    # Note: The type name must be extracted from the implementation name.
    # The implementation name of the error model is always of the form
    # "Type.SomeSuffix", where "Type" is the type name.
    assert error_model_implementation.name.find(".") != -1
    (type_name, suffix) = error_model_implementation.name.split(".")
    package = error_model_implementation.subpackage.package
    # False is passed as the second parameter of get_error_type so that both
    # the public and the private subpackage are searched.
    error_model_type = package.get_error_type(type_name, False)
    assert error_model_type is not None
    return general.ErrorModel(error_model_type, error_model_implementation)


# ----------------------------------------------------------------------------
# Data structures for the representation of fault injection documents:
# ----------------------------------------------------------------------------

class FaultInjectionDocument(etree.ElementBase):
    """A list of error model associations.
    
    Each error model association assigns an error model to a nominal component
    and specifies the effects of the error states of the error model on the
    data elements of the nominal component.
    
    Note: There must be at most one error model associated with a single
    nominal component. To ensure this, it is strongly recommended to use the
    method add_association and instead of the method append (or other list
    operations that add elements) to add associations to the document.
    
    Note: str(some_document) for a FaultInjectionDocument object some_document
    will return the string representation of the document in XML format.
    
    Important: FaultInjectionDocument objects must not be created like this:
    x = FaultInjectionDocument()
    Instead, the static factory method create must be used.
    This is because this class is a subclass of etree.ElementBase, which must
    not be instantiated directly. Refer to the lxml documentation for details.
    """
    
    @staticmethod
    def create():
        """Creates and returns a new FaultInjectionDocument object.
        
        @return: A new FaultInjectionDocument object.
        @rtype:  FaultInjectionDocument.
        """
        # Note: The parser that is used to create the object is created when
        # this module is imported (see the end of this file).
        return _parser.makeelement("{" + _XML_NAMESPACE_URI + "}" +
                                   _FAULT_INJECTION_DOCUMENT_TAG)
    
    def get_by_component_path(self, path):
        """Returns the error model association for the component that is
        determined by the specified path.
        
        Note that there can be at most one error model association for a single
        component.
        
        @param   path: A sequence of subcomponent identifiers, forming a path
                       from the root component of the SLIM model to a certain
                       component.
        @type    path: A tuple of strings.
        @return: The error model association for the specified component or
                 None if no error model is assigned to that component.
        @rtype:  ErrorModelAssociation
        """
        for association in self:
            if association.component_path == path:
                return association
        return None
    
    def add_association(self, error_model_association):
        """Adds the specified error model association to this document.
        
        @param   error_model_association:
                    The error model association to be added. Note: A deep copy
                    of the association will be created.
        @type    error_model_association:
                    ErrorModelAssociation
        @raises  ErrorModelAssociationAlreadyExistsError:
                    If the document does already contain an error model
                    association for the referenced component implementation. 
        """
        if self.get_by_component_path(error_model_association.component_path):
            raise ErrorModelAssociationAlreadyExistsError()
        else:
            # Don't add the specified error model association object itself but
            # rather a deep copy. This is necessary for the following reason:
            # The association object is an XML element, which might already be
            # used as a child element of some other parent element (e.g.
            # another fault injection document). Without copying, it would be
            # removed from that parent. See the lxml documentation.
            self.append(copy.deepcopy(error_model_association))
            
    def __str__(self):
        """Returns a string representation of this document in XML format.
        
        UTF-8 encoding will be used and an XML header will be included.
        
        @return: A string representation of this document in XML format.
        @rtype:  string
        """
        return etree.tostring(self, xml_declaration=True, encoding="UTF-8",
                              pretty_print=True)


class ErrorModelAssociation(etree.ElementBase):
    """An error model association for a specific components in a SLIM model.
    
    It assigns an error model to a specific component of a SLIM model, which is
    identified by a sequence of subcomponent identifiers forming a path form
    the root component of the model. It also specified the effects of the error
    states of that error model on the data elements of the nominal component.
    
    Important: ErrorModelAssociation objects must not be created like this:
    x = ErrorModelAssociation()
    Instead, the static factory method create must be used.
    This is because this class is a subclass of etree.ElementBase, which must
    not be instantiated directly. Refer to the lxml documentation for details.
    """
    
    @staticmethod
    def create(component_path,
               error_model_package_name,
               error_model_implementation_name,
               fault_effects_specification=None):
        """Creates and returns a new ErrorModelAssociation object.
        
        @param   component_path:
                    A sequence of subcomponent identifiers, which determines a
                    specific component of a SLIM model.
        @type    component_path:
                    A tuple of strings.
        @param   error_model_package_name:
                    The name of the package that contains the error model.
        @type    error_model_package_name:
                    string
        @param   error_model_implementation_name:
                    The implementation name of the error model.
        @type    error_model_implementation_name:
                    string
        @param   fault_effects_specification: 
                    (Optional) A specification of the fault effects to be
                    injected. Note: A deep copy of that specification will be
                    created.
        @type    fault_effects_specification:
                    FaultEffectsSpecification
        @return: A new ErrorModelAssociation object.
        @rtype:  ErrorModelAssociation
        """
        # This error model association is an XML element with three child
        # elements:
        # 1) The element that contains the path to the component as a string,
        #    where the single subcomponent identifiers are separated by dots.
        # 2) The element that identifies the error model implementation.
        #    It has two child elements:
        #    a) The element whose text content is the name of the package
        #       containing the error model implementation.
        #    b) The element whose text content is the implementation name of
        #       the error model.
        # 3) The element that specifies the fault effects to be injected.
        # Note: The parser that is used to create the elements is created when
        # this module is imported (see the end of this file).
        association = _parser.makeelement(
                "{" + _XML_NAMESPACE_URI + "}" + _ERROR_MODEL_ASSOCIATION_TAG)
        # Create the first child element.
        component_path_element = _parser.makeelement(
                "{" + _XML_NAMESPACE_URI + "}" + _COMPONENT_PATH_TAG)
        component_path_element.text = ".".join(component_path)
        association.append(component_path_element)
        # Create the second child element, including its children.
        error_model_element = _parser.makeelement(
                "{" + _XML_NAMESPACE_URI + "}" + _ERROR_MODEL_TAG)
        package_name_element = _parser.makeelement(
                "{" + _XML_NAMESPACE_URI + "}" + _PACKAGE_NAME_TAG)
        package_name_element.text = error_model_package_name
        error_model_element.append(package_name_element)
        implementation_name_element = _parser.makeelement(
                "{" + _XML_NAMESPACE_URI + "}" + _IMPLEMENTATION_NAME_TAG)
        implementation_name_element.text = error_model_implementation_name
        error_model_element.append(implementation_name_element)
        association.append(error_model_element)
        # The third child element is the specified fault effects specification.
        # Don't add the specified specification object itself but rather a deep
        # copy. This is necessary for the following reason: The specification
        # object is an XML element, which might already be used as a child
        # element of some other parent element (e.g. another error model
        # association). Without copying, it would be removed from that parent.
        # See the lxml documentation.
        # If no fault effects specification is specified, an empty
        # specification must be added to make the XML fragment complete.
        if fault_effects_specification is not None:
            new_effects_spec = copy.deepcopy(fault_effects_specification)
        else:
            new_effects_spec = FaultEffectsSpecification.create()
        association.append(new_effects_spec)
        return association
    
    # Note for the properties defined below:
    # Leading and trailing whitespace is removed when getting the component
    # path and the package name and implementation name of the error model, so
    # that whitespace can be used to beautify the XML document without altering
    # its meaning.
    
    # The component path is stored as a string where the single subcomponent
    # identifiers are separated by dots. The split-method is used to convert
    # the string to a list, which is finally converted to a tuple.
    # However, there is one special case: If the component path is empty, then
    # the "text" attribute of the underlying XML element will be None or solely
    # consist of whitespace. In this case, the path will be returned as an
    # empty tuple.
    def _get_component_path_as_tuple(self):
        if self[0].text is None or self[0].text.strip() == "":
            return ()
        else:
            return tuple(self[0].text.strip().split("."))
        
    component_path = property(
            _get_component_path_as_tuple,
            doc="The sequence of subcomponent identifiers that determines " +
                "the nominal component.\n" +
                "@type: A tuple of strings.")
    
    
    error_model_package_name = property(
            lambda self: self[1][0].text.strip(),
            doc="The name of the package that contains the error model " +
                "implementation.\n" +
                "@type: string")
    
    error_model_implementation_name = property(
            lambda self: self[1][1].text.strip(),
            doc="The name of the package that contains the error model " +
                "implementation.\n" +
                 "@type: string")
    
    effects_specification = property(
            lambda self: self[2],
            doc="The specification of the fault effects to be injected.\n" +
                 "@type: FaultEffectsSpecification")


class FaultEffectsSpecification(etree.ElementBase):
    """A specification that defines the fault effects for a certain error model
    association.
    
    Important: FaultEffectsSpecification objects must not be created like 
    this: x = FaultEffectsSpecification()
    Instead, the static factory method create must be used.
    This is because this class is a subclass of etree.ElementBase, which must
    not be instantiated directly. Refer to the lxml documentation for details.
    """
    
    @staticmethod
    def create():
        """Creates and returns a new FaultEffectsSpecification object.
        
        The new fault effects specification will contain no fault effects.
        They can be added later.
        
        @return: A new FaultEffectsSpecification object.
        @rtype:  FaultEffectsSpecification.
        """
        # Note: The parser that is used to create the object is created when
        # this module is imported (see the end of this file).
        return _parser.makeelement("{" + _XML_NAMESPACE_URI + "}" +
                                   _FAULT_EFFECTS_SPECIFICATION_TAG)
    
    def get_effective_states(self):
        """Returns a list containing the states of the associated error model
        that have fault effects.
        
        @return: A list containing the error states with fault effects.
        @rtype:  A list of strings.
        """
        # The XML element representing this fault effects specifications has
        # one child element for each error state with associated fault
        # effects. Collect the names of these error states.
        return [child.get("errorState") for child in self]
    
    def get_effects(self, error_state_name):
        """Returns a list containing the fault effects associated with the
        error state with the specified name.
        
        @param   error_state_name: The name of the error state whose fault
                                   effects shall be returned.
        @type    error_state_name: string
        @return: A list containing the fault effects for the specified state.
                 The fault effects are given as strings in SLIM syntax.
        @rtype:  A list of strings.
        """
        # The XML element representing this fault effects specification has
        # one child element for each error state with associated fault
        # effects. Find the one for the specified state.
        # If the specified state has no associated fault effects, an empty
        # list will be returned.
        for child in self:
            if child.get("errorState") == error_state_name:
                # The child element that corresponds to the specified error
                # state is found. The effects of this state are the text
                # contents of the elements contained in this child element.
                # Note: Leading and trailing whitespace is removed.
                return [effect_elem.text.strip() for effect_elem in child]
        else:
            # No fault effects exist for the specified error state.
            return []
            
    def add_effect(self, error_state_name, fault_effect):
        """Adds the specified effect to the list of fault effects of the
        specified error state.
        
        @param   error_state_name: The name of the error state which the effect
                                   shall be attached to.
        @type    error_state_name: string
        @param   fault_effect:     The fault effect to be added. It must be
                                   given as a string in SLIM syntax.
        @type    fault_effect:     string
        """
        # TODO: Check whether the specified effect is already associated with
        # the error state.
        # Create the XML element to represent the specified fault effect.
        effect_element = _parser.makeelement("{" + _XML_NAMESPACE_URI + "}" +
                                             _FAULT_EFFECT_TAG)
        effect_element.text = fault_effect
        # The XML element representing this fault effects specification has
        # one child element for each error state with associated fault
        # effects. If there is one for the specified state, then the XML
        # element representing the fault effect must be added as a child of
        # this child element. Otherwise, a new element must be created for the
        # specified error state.
        # Note: The parser that is used to create the XML elements is
        # initialized when this module is imported (see the end of this file).
        for child in self:
            if child.get("errorState") == error_state_name:
                # The child element that corresponds to the specified error
                # state is found. Add the effect.
                child.append(effect_element)
                break # Important to skip the "else" clause.
        else:
            # No fault effects exist for the specified error state. Create a
            # new child element to contain the fault effects of the error
            # state and add the new fault effect.
            new_child = _parser.makeelement("{" + _XML_NAMESPACE_URI + "}" +
                                            _FAULT_EFFECTS_FOR_STATE_TAG)
            new_child.set("errorState", error_state_name)
            self.append(new_child)
            new_child.append(effect_element)


class ErrorModelAssociationAlreadyExistsError(Exception):
    """An object indicating that an error model association could not be added
    to a certain fault injection document because it refers to a component for
    which there is already an association in the document.
    """
    # TODO: Maybe the path to the particular component should be stored in this
    # object.
    # TODO: Maybe the method __str__ should be implemented.
    pass


# ----------------------------------------------------------------------------
# Note: The following code will be executed once when this module is imported.
# ----------------------------------------------------------------------------

# Define some constants which are used in this module.
# The XML namespace that is used for fault injection documents:
_XML_NAMESPACE_URI = "http://compass.informatik.rwth-aachen.de/fault"
# The filename of the XML Schema document that defines the structure of valid
# fault injection documents.
_XSD_FILENAME = os.path.join(os.path.dirname(__file__),
                             "xsd", "fault_injection.xsd")

# Define constants for the XML tag names that are used in fault injection
# documents.
_FAULT_INJECTION_DOCUMENT_TAG = "faultInjection"
_ERROR_MODEL_ASSOCIATION_TAG = "association"
_COMPONENT_PATH_TAG = "componentPath"
_ERROR_MODEL_TAG = "errorModel"
_PACKAGE_NAME_TAG = "packageName"
_IMPLEMENTATION_NAME_TAG = "implementationName"
_FAULT_EFFECTS_SPECIFICATION_TAG = "faultEffectsSpecification"
_FAULT_EFFECTS_FOR_STATE_TAG = "faultEffects"
_FAULT_EFFECT_TAG = "effect"

# Initialize the parser that is used to parse and to create fault injection
# specification documents.
# XML comments must be discarded. Otherwise the parser would yield data
# structures cluttered with unwanted elements.
# A lookup scheme is used in order to automatically map certain XML elements
# to instances of the adequate classes which are defined in this module.
_parser = etree.XMLParser(remove_comments=True)
_lookup = etree.ElementNamespaceClassLookup()
_parser.set_element_class_lookup(_lookup)
# Associate the various XML element tags with the corresponding classes. 
# Note that all elements are in the same namespace.
_namespace = _lookup.get_namespace(_XML_NAMESPACE_URI)
_namespace[_FAULT_INJECTION_DOCUMENT_TAG] = FaultInjectionDocument
_namespace[_ERROR_MODEL_ASSOCIATION_TAG] = ErrorModelAssociation
_namespace[_FAULT_EFFECTS_SPECIFICATION_TAG] = FaultEffectsSpecification

# Parse the XML Schema document that defines the structure of valid fault
# injection documents. This is needed for validation.
_fault_injection_schema_tree = etree.parse(_XSD_FILENAME)
_fault_injection_schema = etree.XMLSchema(_fault_injection_schema_tree)
