#!/usr/bin/env python

# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


"""This module contains tests for the module fault_injection. It can be
executed directly.
"""

# Add the directory "tools" to the module search path. The ANTLR directory
# must be added as well.
import sys
import os.path
TOP_DIR = os.path.abspath(
            os.path.join(os.path.dirname(__file__), "..", "..", "..")) 
sys.path.append(os.path.abspath(TOP_DIR))
sys.path.append(os.path.abspath(os.path.join(TOP_DIR,
                                             "..", "resources", "antlr",
                                             "antlr-python-runtime-3.1.1")))

import unittest
from lxml import etree
from slim_compiler.src.model_extension import fault_injection
from slim_compiler.src.model_extension import errors


_VALID_DOCUMENTS_PATH = os.path.join(TOP_DIR, "slim_compiler", "src",
                                     "model_extension", "test_documents",
                                     "valid")
_INVALID_DOCUMENTS_PATH = os.path.join(TOP_DIR, "slim_compiler", "src",
                                     "model_extension", "test_documents",
                                     "invalid")


class TestInspectInvalidDocuments(unittest.TestCase):
    """A test case for parsing and validating invalid fault injection
    documents.
    """
        
    def test_validate_two_error_models_for_same_component(self):
        """Validates a document with two error models for one component."""
        filename = os.path.join(_INVALID_DOCUMENTS_PATH,
                                "two_error_models_for_same_component.xml")
        injection_doc = fault_injection.parse(filename)
        error_list = fault_injection.get_syntax_errors_in_injection_document(
                                                                injection_doc)
        assert len(error_list) == 1
        assert isinstance(error_list[0], errors.DocumentSyntaxError)
        
    def test_validate_fault_effects_defined_twice(self):
        """Validates a document which defines the effects of a state twice."""
        filename = os.path.join(_INVALID_DOCUMENTS_PATH,
                                "fault_effects_defined_twice.xml")
        injection_doc = fault_injection.parse(filename)
        error_list = fault_injection.get_syntax_errors_in_injection_document(
                                                                injection_doc)
        assert len(error_list) == 1
        assert isinstance(error_list[0], errors.DocumentSyntaxError)


class TestInspectValidDocument(unittest.TestCase):
    """A test case for parsing and inspecting a valid fault injection
    document.
    """
    
    def setUp(self):
        """Parses the test document.
        
        The content of the document will then be represented by the data
        structures defined in the module fault_injection.
        """
        filename = os.path.join(_VALID_DOCUMENTS_PATH, "small_example.xml")
        self.injection_document = fault_injection.parse(filename)
        self.association = self.injection_document.get_by_component_path(
                                                ("sub1", "sub2", "sub3"))
        assert self.association is not None
    
    def test_validate(self):
        """Tests that the parsed document is valid."""
        error_list = fault_injection.get_syntax_errors_in_injection_document(
                                                    self.injection_document)
        assert len(error_list) == 0
            
    def test_get_by_component_implementation(self):
        """Tests FaultInjectionDocument.get_by_nominal_implementation."""
        # None must be returned when requesting an error model association for
        # a nominal component without an associated error model.
        assoc = self.injection_document.get_by_component_path(("s", "t"))
        assert assoc is None
        
    def test_error_model(self):
        """Tests that the error model association provides the correct package
        name and implementation name for the error model."""
        assert self.association.error_model_package_name == "__default__"
        assert self.association.error_model_implementation_name == "Type.I"
        
    def test_fault_effects_specification(self):
        """Tests the class FaultEffectsSpecification."""
        effects_spec = self.association.effects_specification
        assert effects_spec is not None
        assert effects_spec.get_effective_states() == ["state1", "state2"]
        assert effects_spec.get_effects("stateWithoutEffect") == []
        assert effects_spec.get_effects("state1") == \
                ["someDataElement := 0", "anotherDataElement := 5"]
        assert effects_spec.get_effects("state2") == \
                ["someDataElement := 1", "yetAnotherDataElement := 6"]
        

class TestCreateAndInspectDocument(unittest.TestCase):
    """A test case for creating and inspecting a fault injection document."""
    # TODO: Implement more test methods for this test case.
    
    def test_create_simple_document(self):
        """Tests the construction of a simple fault injection document."""
        x = fault_injection.FaultInjectionDocument.create()
        assoc = fault_injection.ErrorModelAssociation.create(("s", "t", "u"),
                                                             "package", "E.I")
        assoc.effects_specification.add_effect("stateY", "out := 1")
        assoc.effects_specification.add_effect("stateY", "out2 := 2")
        x.add_association(assoc)


if __name__ == "__main__":
    # Run all tests in this module.
    unittest.main()
