# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================
from slim_compiler.src.dataStructure.visitors.expressionVisitor import AnnotateVisitor,\
    TypeVisitor
from functools import partial


"""This module provides utility classes and functions for handling SLIM models.
"""
# TODO: In this file, a maximum line length of 80 is used, instead of 79 as
# specified in PEP8.

from slim_compiler.src.model_extension import constants
from slim_compiler.src.model_extension import errors
import slim_compiler.src.model_extension.util.expressions as util_expressions
from slim_compiler.src.errorManagement import errors as compile_errors
# TODO: The following imports might not strictly be needed: They define classes
# and this module works with instances of these classes, but never instantiates
# these classes, so there is no strong connection to the imported modules.
import slim_compiler.src.dataStructure.compile as slim_compile
import slim_compiler.src.dataStructure.errorbehaviour as error
import slim_compiler.src.dataStructure.nominalbehaviour as nominal


# TODO: Maybe the following class should be moved to a more general module
# outside the model extension package, because it might be useful for other
# parts of the toolset as well.
class ComponentInstance(object):
    """An object representing a component of a SLIM model.
    
    A component is an instance of a component implementation. It is determined
    by a sequence of subcomponent identifiers, forming a path in the component
    hierarchy (which is a tree) from a specific root component.
    """
    
    def __init__(self, root_implementation, path):
        """Initializes this component object.
        
        @param   root_implementation: The implementation of the root component
                                      of the SLIM model.
        @type    root_implementation: nominal.ComponentImplementation
        @param   path:                The sequence of subcomponent identifiers
                                      leading from the root component to the
                                      component that shall be represented by
                                      this object. For the root component, this
                                      sequence is empty.
        @type    path:                A tuple of strings.
        @raise   errors.InvalidComponentPathError: If the specified sequence of
                    subcomponent identifiers is not a valid component path.
        """
        self.root_implementation = root_implementation
        self.path = path
        # Initialize the dictionary that is used to cache subcomponent
        # instances.
        self._subcomponents_cache = {}
        
        # Find the implementation of the component that shall be represented by
        # this object, following the specified path.
        # Keep track of the current position in the path, so that the first
        # invalid subcomponent identifier can be determined in case of failure.
        implementation = root_implementation
        for (path_position, subcomponent_id) in enumerate(path):
            try:
                subcomponent_declaration = \
                        implementation.non_data_subcomponents[subcomponent_id]
            except KeyError:
                # The path is valid before the current subcomponent identifier,
                # but invalid from that point.
                valid_path_segment = path[:path_position]
                rest_of_path = path[path_position:]
                raise errors.InvalidComponentPathError(root_implementation,
                                                       valid_path_segment,
                                                       rest_of_path)
            implementation = subcomponent_declaration.\
                                get_referenced_component_implementation()
        
        self.implementation = implementation
        self.type = implementation.get_referenced_component_type()
    
    def get_subcomponent_instance(self, subcomponent_id):
        """Returns a ComponentInstance object representing the subcomponent of
        of this component that has the specified subcomponent identifier.
        
        The returned objects are cached, so if this method is called multiple
        times with the same subcomponent identifier, the same object will be
        returned every time.
        
        @param   subcomponent_id: The identifier of the desired subcomponent.
        @type    subcomponent_id: string
        @return: An object representing the specified subcomponent of this
                 component.
        @rtype:  ComponentInstance
        """
        try:
            return self._subcomponents_cache[subcomponent_id]
        except KeyError:
            subcomponent_path = self.path + (subcomponent_id,)
            subcomponent_instance = ComponentInstance(self.root_implementation,
                                                      subcomponent_path)
            self._subcomponents_cache[subcomponent_id] = subcomponent_instance
            return subcomponent_instance


# TODO: Maybe the following function should be moved to a more general module
# outside the model extension package, because it might be useful for other
# parts of the toolset as well.
def instance_generator(root_implementation, filter_function=None):
    """Returns a generator-iterator that generates the component instances for
    the SLIM model with the specified root implementation.
    
    If a filter function is specified, the generator-iterator will only yield
    component instances for which the filter function returns True.
    
    @param   root_implementation: The implementation of the root component of
                                  the SLIM model.
    @type    root_implementation: nominal.ComponentImplementation
    @param   filter_function:     (Optional) A filter function or None.
    @type    filter_function:     A callable that takes a ComponentInstance
                                  object and returns a bool. Or NoneType.
    @return: A generator-iterator that generates the component instances for
             the specified SLIM model.
    @rtype:  A generator-iterator for ComponentInstance objects.
    """
    root_component = ComponentInstance(root_implementation, ())
    return instance_generator_for_root_component(root_component,
                                                 filter_function)


# TODO: Maybe the following function should be moved to a more general module
# outside the model extension package, because it might be useful for other
# parts of the toolset as well.
def instance_generator_for_root_component(root_component,
                                          filter_function=None):
    """Returns a generator-iterator that generates the component instances for
    the SLIM model with the specified root component (not root implementation).
    
    If a filter function is specified, the generator-iterator will only yield
    component instances for which the filter function returns True.
    
    @param   root_implementation: The root component of the SLIM model.
    @type    root_implementation: ComponentInstance
    @param   filter_function:     (Optional) A filter function or None.
    @type    filter_function:     A callable that takes a ComponentInstance
                                  object and returns a bool. Or NoneType.
    @return: A generator-iterator that generates the component instances for
             the specified SLIM model.
    @rtype:  A generator-iterator for ComponentInstance objects.
    """
    if (filter_function is None) or filter_function(root_component):
        yield root_component
    for subcomp_id in root_component.implementation.non_data_subcomponents:
        subcomponent = root_component.get_subcomponent_instance(subcomp_id)
        for component in instance_generator_for_root_component(
                                subcomponent, filter_function):
            yield component


class ErrorModel(object):
    """An object representing an error model of a SLIM model, consisting of a
    type and an implementation.
    """
    
    def __init__(self, type, implementation):
        """Initializes this ErrorModel object.
        
        @param   type:           The type of the error model.
        @type    type:           error.ErrorModelType
        @param   implementation: The implementation of the error model.
        @type    implementation: error.ErrorModelImplementation
        """
        self.type = type
        self.implementation = implementation


def get_public_default_subpackage(compile_info):
    """Returns the public subpackage of the default package of the specified
    SLIM model.
    
    @param   compile_info: An object representing the SLIM model.
    @type    compile_info: slim_compile.CompileInformation
    @return: The public subpackage of the default package of the SLIM model.
    @rtype:  slim_compile.Subpackage
    """
    default_package = compile_info.get_package(constants.DEFAULT_PACKAGE_NAME)
    return default_package.get_public_subpackage()


def check_effect_is_applicable(effect_tree, component_implementation,
                               source_mode_name, target_mode_name):
    """Returns a boolean value indicating whether the specified effect is
    applicable for a transition with the specified source and target mode
    names.
    
    The effect is not applicable in any of the following cases:
      - The data element on the left-hand side of the assignment is not active
        in the target mode of the transition.
      - Some data elements on the right-hand side of the assignment are not
        active in the source mode of the transition.
    
    @param   effect_tree:              The tree representation of the effect to
                                       be checked.
    @type    effect_tree:              expressions.BinaryOperatorTreeNode
    @param   component_implementation: The implementation of the component to
                                       be considered.
    @type    component_implementation: nominal.ComponentImplementation
    @param   source_mode_name:         The name of the source mode of the
                                       transition to be considered.
    @type    source_mode_name:         string
    @param   target_mode_name:         The name of the target mode of the
                                       transition to be considered.
    @type    target_mode_name:         string
    @return: A boolean value indicating whether the effect is applicable.
    @rtype:  bool
    """
    # TODO: This function is a hack and is heavily coupled with the compiler.
    # Maybe this can be improved.
    # TODO: Write tests for this function.

    # The check is performed as a side effect of the annotation methods in
    # nominal.ComponentImplementation. They will attach error objects to the
    # error collector of the compile information object containing the
    # component implementation if the effect is invalid.
    # First, get the compile information object containing the component
    # implementation.
    compile_info = component_implementation.subpackage.package.\
                                            compile_information
    # To make sure that there are no error objects in the error collector that
    # stem from other parts of the model, replace the error collector with a
    # new one which does not contain errors.
    # Remember the old error collector in order to plug it back in after the
    # check.
    original_error_collector = compile_info.error_collector
    compile_info.error_collector = compile_errors.ErrorCollector()
    # The errors will be stored in the collector along with information about
    # the file in which they occurred and their position within that file.
    # In our case, the errors are not really located in a file. Remove the
    # misleading file information from the component implementation to avoid
    # problems. Remember this information in order to restore it after the
    # check.
    original_filename = component_implementation.file
    component_implementation.file = None
    
    # Perform the check. Do this in a "try" block so that the clean-up work can
    # be done in the "finally" part, even if something goes wrong.
    try:
        # Setup annotation and type visitors
        annotator = AnnotateVisitor(compile_info)
        type_checker = TypeVisitor(compile_info)

        assert len(component_implementation.modes) > 0
        mode_name = component_implementation.modes.keys()[0]
        annotator.annotation_meth = partial(component_implementation.get_data_element_type_lhs_effect,
                target_mode = target_mode_name, trans = None)
        annotator.visit(effect_tree.left)
        annotator.annotation_meth = partial(component_implementation.get_data_element_type_rhs_effect,
                source_mode = target_mode_name, trans = None)
        annotator.visit(effect_tree.right)

        #TODO: this may trigger errors when otherwise fine after postprocess
        type_checker.visit(effect_tree)

        # Remember if there were errors.
        no_errors_found = compile_info.error_collector.has_no_errors()
    finally:
        # Clean up: Plug the original error collector back in to restore the
        # original state of the compile information object, and reset the
        # filename of the component implementation.
        compile_info.error_collector = original_error_collector
        component_implementation.file = original_filename
    
    return no_errors_found


def check_expression_is_applicable_in_flow(expression_tree,
                                           component_implementation,
                                           mode_name):
    """Returns a boolean value indicating whether the specified expression is
    applicable as a right-hand side of a flow in the specified mode.
    
    The expression is not applicable if it refers to an outgoing data port of a
    subcomponent that is not active in the specified mode.
    
    Note: The target of the flow is irrelevant for this function.
    
    @param   expression_tree:          The tree representation of the
                                       expression to be checked.
    @type    expression_tree:          expressions.ExpressionTreeNode
    @param   component_implementation: The implementation of the component to
                                       be considered.
    @type    component_implementation: nominal.ComponentImplementation
    @param   mode_name:                The name of the mode to be considered.
    @type    mode_name:                string
    @return: A boolean value indicating whether the effect is applicable.
    @rtype:  bool
    """
    # TODO: This function is a hack and is heavily coupled with the compiler.
    # Maybe this can be improved.
    # TODO: Write tests for this function.

    # The check is performed as a side effect of the annotation methods in
    # nominal.ComponentImplementation. They will attach error objects to the
    # error collector of the compile information object containing the
    # component implementation if the expression is not applicable.
    # First, get the compile information object containing the component
    # implementation.
    compile_info = component_implementation.subpackage.package.\
                                            compile_information
    # To make sure that there are no error objects in the error collector that
    # stem from other parts of the model, replace the error collector with a
    # new one which does not contain errors.
    # Remember the old error collector in order to plug it back in after the
    # check.
    original_error_collector = compile_info.error_collector
    compile_info.error_collector = compile_errors.ErrorCollector()
    # The errors will be stored in the collector along with information about
    # the file in which they occurred and their position within that file.
    # In our case, the errors are not really located in a file. Remove the
    # misleading file information from the component implementation to avoid
    # problems. Remember this information in order to restore it after the
    # check.
    original_filename = component_implementation.file
    component_implementation.file = None
    
    # Perform the check. Do this in a "try" block so that the clean-up work can
    # be done in the "finally" part, even if something goes wrong.
    try:
        # Setup annotation and type visitors
        annotator = AnnotateVisitor(compile_info)
        type_checker = TypeVisitor(compile_info)

        assert len(component_implementation.modes) > 0
        mode_name = component_implementation.modes.keys()[0]
        annotator.annotation_meth = partial(component_implementation.get_data_element_type_rhs_flow,
                mode = mode_name, trans = None)
        annotator.visit(expression_tree)

        #TODO: this may trigger errors when otherwise fine after postprocess
        type_checker.visit(expression_tree)

        # Remember if there were errors.
        no_errors_found = compile_info.error_collector.has_no_errors()
    finally:
        # Clean up: Plug the original error collector back in to restore the
        # original state of the compile information object, and reset the
        # filename of the component implementation.
        compile_info.error_collector = original_error_collector
        component_implementation.file = original_filename
    
    return no_errors_found


def check_physical_connection(subcomponent_declaration1,
                              subcomponent_declaration2):
    """Returns a boolean value indicating whether the specified subcomponents
    are physically connect such that outgoing error propagations of the first
    subcomponent are connected with corresponding incoming error propagations
    of the second subcomponent.
    
    Such a physical connection between two subcomponents subcomponent1 and
    subcomponent 2 exists if and only if one of the following is true:
      - subcomponent1 is a bus and subcomponent2 accesses subcomponent1
      - subcomponent1 is a process and subcomponent2 runs on subcomponent1
      - subcomponent1 is a memory and subcomponent2 is stored on subcomponent1
      - subcomponent1 and subcomponent2 both access a common bus
    
    @param   subcomponent_declaration1: The first subcomponent.
    @type    subcomponent_declaration1: nominal.NonDataSubcomponent
    @param   subcomponent_declaration2: The second subcomponent.
    @type    subcomponent_declaration2: nominal.NonDataSubcomponent
    @return: A boolean value indicating whether the specified subcomponents are
             physically connected.
    @rtype:  bool
    """
    # Components can only be physically connected if they are subcomponents of
    # the same supercomponent. This check is required because the following
    # checks are based on the subcomponent names and thus might yield positive
    # results even if the subcomponents reside in different supercomponents.
    super_implementation1 = subcomponent_declaration1.implementation
    super_implementation2 = subcomponent_declaration2.implementation
    if super_implementation1 is not super_implementation2:
        return False
    
    accesses  = (subcomponent_declaration1.name in
                 subcomponent_declaration2.accesses)
    runs_on   = (subcomponent_declaration1.name in
                 subcomponent_declaration2.running_on)
    stored_in = (subcomponent_declaration1.name in
                 subcomponent_declaration2.stored_in)
    
    common_busses = (set(subcomponent_declaration1.accesses) &
                     set(subcomponent_declaration2.accesses))
    access_common_bus = (len(common_busses) > 0)

    return (accesses or runs_on or stored_in or access_common_bus)


def create_flow_for_port_connection(extended_impl, data_port_connection):
    """Creates and returns an object representing a SLIM flow that is
    equivalent to the specified data port connection.
    
    For a port connection "source_port -> target_port", the equivalent flow
    expression in SLIM syntax is "target_port := source_port".
    
    @param   data_port_connection: A data port connection.
    @type    data_port_connection: nominal.PortConnection
    @return: A new SLIM flow that is equivalent to the specified data port
             connection.
    @rtype:  nominal.Flow
    """
    # Create a tree representation of the adequate flow expression (which is of
    # the form "target_port := source_port").
    flow_expression_tree = util_expressions.create_slim_assignment(
                                   data_port_connection.target_port,
                                   data_port_connection.source_port)
    # Note: The flow shall be active in the same modes as the port connection.
    # No useful information about the file and position of the flow can be
    # provided, so the corresponding parameters are set to None or 0
    # respectively.
    return nominal.Flow(extended_impl, 0, flow_expression_tree, data_port_connection.in_modes,
                        None, 0, 0)
