#!/usr/bin/python 

# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================



import sys
import os
import os.path
from slim_compiler.src.dataStructure.visitors.expressionVisitor import TypeVisitor

sys.path.append('../../../resources/antlr/antlr-python-runtime-3.1.1/')
sys.path.append('../../../resources/pychecker-0.8.18/')

import antlr3
import antlr3.tree
import logging
import functools

from generated.SlimLexer import SlimLexer
from generated.SlimWalkerIdentCheck import SlimWalkerIdentCheck
from errorManagement.errors import *
from errorManagement.exceptions import *
from dataStructure.compile import CompileInformation
from dataStructure.nominalbehaviour import *
from dataStructure.derivedparser import DerivedParser
import dataStructure.expressions
from slim_compiler.src import deprecated

def parse_identifier(name):
    qualifications = []
    while name.find(".") != -1:
        qualifications.append(name[0:name.find(".")])
        name = name[name.find(".") + 1:]
        
    return (qualifications, name)

def follow_identifier_path(cinfo, root_impl, name, error_collector):
    (qualifications, data_field) = parse_identifier(name)
    impl = root_impl
    
    error = False
    for qualification in qualifications:
        if not error:
            if qualification in impl.non_data_subcomponents.keys():
                subcomp = impl.non_data_subcomponents[qualification]
                impl = subcomp.get_referenced_component_implementation()                
            else:
                error_collector.add(WrongQualificationError(name, qualification))
                error = True
    
    if not error:
        if data_field == "mode":
            return impl
        elif data_field == "error":
            if "_errorSubcomponent" in impl.non_data_subcomponents.keys():
                subcomp = impl.non_data_subcomponents["_errorSubcomponent"]
                return subcomp.get_referenced_component_implementation()
            else:
                return None
        else:
            if data_field in impl.data_subcomponents.keys():
                subcomp = impl.data_subcomponents[data_field]
                return subcomp
            else:
                _type = impl.get_referenced_component_type()
                if _type is not None:
                    if data_field in _type.in_data_ports.keys():
                        port = _type.in_data_ports[data_field]
                        return port
                    elif data_field in _type.out_data_ports.keys():
                        port = _type.out_data_ports[data_field]
                        return port
                    else:
                        error_collector.add(WrongQualificationError(name,
                                                                    data_field))
                        return None
    else:
        return None

def get_type(data_type):
    if isinstance(data_type, Int):
        return NodeType.INT
    elif isinstance(data_type, Bool):
        return NodeType.BOOL
    elif isinstance(data_type, Real):
        return NodeType.REAL
    elif isinstance(data_type, Clock):
        return NodeType.CLOCK
    elif isinstance(data_type, Continuous):
        return NodeType.CONTINUOUS
    elif isinstance(data_type, Enum):
        return NodeType.ENUM_FIELD
    elif isinstance(data_type, IntRange):
        return NodeType.INT_RANGE
    else:
        assert False

def annotate_property_tree(cinfo, root_impl, root_node, error_collector):
    if isinstance(root_node, IdentifierTreeNode):
        if root_node.name.find(":") != -1:
            (qualifier, name) = root_node.name.split(":")
            root_node.name = name
            if qualifier == "enum":
                root_node.node_type = NodeType.ENUM_VALUE
            elif qualifier == "mode":
                root_node.node_type = NodeType.MODE_VALUE
            elif qualifier == "error":
                root_node.node_type = NodeType.ERROR_VALUE
        else:
            data_elem = follow_identifier_path(cinfo, root_impl, root_node.name,
                                               error_collector)
            if data_elem is not None:
                if isinstance(data_elem, ComponentImplementation):
                    if root_node.name.endswith("mode"):
                        valid = []
                        for mode in data_elem.modes.values():
#                            if mode.type != ModeType.ERROR:
                            valid.append(mode.name)
                        root_node.enum_type = valid
                        root_node.node_type = NodeType.MODE_FIELD
                    elif root_node.name.endswith("error"):
                        if (root_node.name.count(".") > 0):
                            name = root_node.name.rsplit(".", 1)[0]
                            name = name + "._errorSubcomponent.error"
                        else:
                            name = "_errorSubcomponent.error"
                        root_node.name = name
                        valid = []
                        for mode in data_elem.modes.values():
#                            if mode.type == ModeType.ERROR:
                            valid.append(mode.name)
                        root_node.enum_type = valid
                        root_node.node_type = NodeType.ERROR_FIELD
                elif isinstance(data_elem.data_type, Enum):
                    root_node.enum_type = data_elem.data_type.values
                    root_node.node_type = NodeType.ENUM_FIELD
                else:
                    root_node.node_type = get_type(data_elem.data_type)
            else:
                root_node.node_type = NodeType.ERROR
    elif isinstance(root_node, NumberTreeNode):
        pass
    elif isinstance(root_node, BooleanTreeNode):
        pass
    elif isinstance(root_node, UnaryOperatorTreeNode):
        annotate_property_tree(cinfo, root_impl, root_node.subexpression,
                               error_collector)
    elif isinstance(root_node, BinaryOperatorTreeNode):
        annotate_property_tree(cinfo, root_impl, root_node.left,
                               error_collector)
        annotate_property_tree(cinfo, root_impl, root_node.right,
                               error_collector)
    elif isinstance(root_node, CaseExpressionTreeNode):
        pass
    else:
        assert False
    

def check_property_tree(cinfo, property_tree, root_impl):
    error_collector = ErrorCollector()
    annotate_property_tree(cinfo, root_impl, property_tree, error_collector)
    # Temporarily switch error collector of cinfo
    collector_old = cinfo.error_collector
    cinfo.error_collector = error_collector
    TypeVisitor(cinfo).visit(property_tree)
    cinfo.error_collector = collector_old
    type = property_tree.node_type
    if type != NodeType.BOOL:
        raise InvalidExpressionException(error_collector)
    else:
        # expression was correct
        pass

@deprecated.deprecated
def annotate_lhs_effect_expression(root_node, component_implementation,
                                   error_collector):
    if isinstance(root_node, IdentifierTreeNode):
        type = get_type_lhs_effect_expression(root_node.name,
                                              component_implementation,
                                              error_collector)
        if type == NodeType.INT or type == NodeType.BOOL or \
           type == NodeType.REAL or type == NodeType.CONTINUOUS or \
           type == NodeType.CLOCK:
            root_node.node_type = type
        elif type != NodeType.ERROR:
            (subtype, add_info) = type
            if subtype == NodeType.ENUM_FIELD:
                root_node.enum_values = add_info
                root_node.node_type = NodeType.ENUM_FIELD
            elif subtype == NodeType.ENUM_VALUE:
                root_node.valid_enums = add_info
                root_node.node_type = NodeType.ENUM_VALUE
            elif subtype == NodeType.INT_RANGE:
                root_node.node_type = NodeType.INT_RANGE
                root_node.bounds = add_info
        else:
            root_node.node_type = NodeType.ERROR
                
    else:
        assert False

@deprecated.deprecated
def annotate_rhs_effect_expression(root_node, component_implementation,
                                   error_collector):
    if isinstance(root_node, IdentifierTreeNode):
        type = get_type_rhs_effect_expression(root_node.name,
                                              component_implementation,
                                              error_collector)
        if type == NodeType.INT or type == NodeType.BOOL or \
           type == NodeType.REAL or type == NodeType.CONTINUOUS or \
           type == NodeType.CLOCK or type == NodeType.INT_CONST or \
           type == NodeType.REAL_CONST:
            root_node.node_type = type
        elif type != NodeType.ERROR:
            (subtype, add_info) = type
            if subtype == NodeType.ENUM_FIELD:
                root_node.enum_values = add_info
                root_node.node_type = NodeType.ENUM_FIELD
            elif subtype == NodeType.ENUM_VALUE:
                root_node.valid_enums = add_info
                root_node.node_type = NodeType.ENUM_VALUE
            elif subtype == NodeType.INT_RANGE:
                root_node.node_type = NodeType.INT_RANGE
                root_node.bounds = add_info
        else:
            root_node.node_type = NodeType.ERROR
    elif isinstance(root_node, NumberTreeNode):
        pass
    elif isinstance(root_node, BooleanTreeNode):
        pass
    elif isinstance(root_node, UnaryOperatorTreeNode):
        annotate_rhs_effect_expression(root_node.subexpression,
                                       component_implementation,
                                       error_collector)
    elif isinstance(root_node, BinaryOperatorTreeNode):
        annotate_rhs_effect_expression(root_node.left,
                                       component_implementation,
                                       error_collector)
        annotate_rhs_effect_expression(root_node.right,
                                       component_implementation,
                                       error_collector)
    elif isinstance(root_node, TrajectoryEquationTreeNode):
        assert False
    elif isinstance(root_node, TrajectoryExpressionTreeNode):
        assert False
    elif isinstance(root_node, CaseExpressionTreeNode):
        for (cond, expr) in root_node.cases:
            annotate_rhs_effect_expression(cond, component_implementation,
                                           error_collector)
            annotate_rhs_effect_expression(expr, component_implementation,
                                           error_collector)
        annotate_rhs_effect_expression(root_node.default_expression,
                                       component_implementation,
                                       error_collector)
    else:
        assert False

@deprecated.deprecated
def get_type_lhs_effect_expression(name, impl, error_collector):
    if name in impl.data_subcomponents.keys():
        subcomp = impl.data_subcomponents[name]
        if isinstance(subcomp.data_type, Int):
            return NodeType.INT
        elif isinstance(subcomp.data_type, Real):
            return NodeType.REAL
        elif isinstance(subcomp.data_type, Bool):
            return NodeType.BOOL
        elif isinstance(subcomp.data_type, Clock):
            return NodeType.CLOCK
        elif isinstance(subcomp.data_type, Continuous):
            return NodeType.CONTINUOUS
        elif isinstance(subcomp.data_type, Enum):
            return (NodeType.ENUM_FIELD, subcomp.data_type.values)
        elif isinstance(subcomp.data_type, IntRange):
            return (NodeType.INT_RANGE, subcomp.data_type.get_bounds())
    else:
        type = impl.get_referenced_component_type()
        if type is not None:
            port = None
            if name in type.out_data_ports.keys():
                port = type.out_data_ports[name]
                    
            if port is not None:
                if isinstance(port.data_type, Int):
                    return NodeType.INT
                elif isinstance(port.data_type, Real):
                    return NodeType.REAL
                elif isinstance(port.data_type, Bool):
                    return NodeType.BOOL
                elif isinstance(port.data_type, Enum):
                    return (NodeType.ENUM_FIELD, port.data_type.values)
                elif isinstance(port.data_type, IntRange):
                    return (NodeType.INT_RANGE, port.data_type.get_bounds())
                        
    error_collector.add(InjectionRefersToUnknownIdentifierLhsError(name))
    return NodeType.ERROR

@deprecated.deprecated
def get_type_rhs_effect_expression(name, impl, error_collector):
    is_in_enums = []
    for subcomp in impl.data_subcomponents.values():
        if name == subcomp.name:
            if isinstance(subcomp.data_type, Int):
                return NodeType.INT
            elif isinstance(subcomp.data_type, Real):
                return NodeType.REAL
            elif isinstance(subcomp.data_type, Bool):
                return NodeType.BOOL
            elif isinstance(subcomp.data_type, Clock):
                return NodeType.CLOCK
            elif isinstance(subcomp.data_type, Continuous):
                return NodeType.CONTINUOUS
            elif isinstance(subcomp.data_type, Enum):
                return (NodeType.ENUM_FIELD, subcomp.data_type.values)
            elif isinstance(subcomp.data_type, IntRange):
                return (NodeType.INT_RANGE, subcomp.data_type.get_bounds())
        else:
            if isinstance(subcomp.data_type, Enum):
                if name in subcomp.data_type.values:
                    is_in_enums.append(subcomp.name)
                
    type = impl.get_referenced_component_type()
    if type is not None:
        ports = type.in_data_ports.values() + type.out_data_ports.values()
        for port in ports:
            if name == port.name:
                if isinstance(port.data_type, Int):
                    return NodeType.INT
                elif isinstance(port.data_type, Real):
                    return NodeType.REAL
                elif isinstance(port.data_type, Bool):
                    return NodeType.BOOL
                elif isinstance(port.data_type, Enum):
                    return (NodeType.ENUM_FIELD, port.data_type.values)
                elif isinstance(port.data_type, IntRange):
                    return (NodeType.INT_RANGE, port.data_type.get_bounds())
            else:
                if isinstance(port.data_type, Enum):
                    if name in port.data_type.values:
                        is_in_enums.append(port.data_type)
                            
    if is_in_enums:
        return (NodeType.ENUM_VALUE, is_in_enums)
    else:
        error_collector.add(InjectionRefersToUnknownIdentifierRhsError(name))
        return NodeType.ERROR 


@deprecated.deprecated
def check_effect_tree(effect, component_implementation):
    error_collector = ErrorCollector()
    annotate_lhs_effect_expression(effect.left, component_implementation,
                                   error_collector)
    annotate_rhs_effect_expression(effect.right, component_implementation,
                                   error_collector)
    type = expressions.get_type(effect, error_collector.add, "effect")
    if type == NodeType.ERROR or not isinstance(effect, BinaryOperatorTreeNode) or effect.operator_type != OperatorType.ASSIGN:
        raise InvalidExpressionException(error_collector)
    else:
        # effect was correct
        pass
