# =============================================================================
#  This file is part of the ``COMPASS'' Toolset.
#  Copyright (C) 2010 by Fondazione Bruno Kessler
#  Copyright (C) 2010 by RWTH Aachen University
#
#
#  The COMPASS Toolset can be copied, modified and/or redistributed
#  under the terms of the COMPASS General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the COMPASS General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The COMPASS Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  COMPASS General Public License for more details.
#
#  For more information on the COMPASS Project see the Project Web
#  site at <http://compass.informatik.rwth-aachen.de/> or email to
#  <compass-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <compass-support@lists.rwth-aachen.de>.
# =============================================================================


import sys
import os
from invokers import slim
if os.name == "nt":
    sys.path.append(".\\..\\..\\..") # 3 levels up to tools directory (WIN)
else: 
    sys.path.append("./../../../") # 3 levels up to tools directory (UNIX)

import invokers.slim
import slim_compiler.src.dataStructure.compile
from slim_compiler.src.dataStructure.nominalbehaviour import *
from slim_compiler.src.dataStructure.expressions import *

from slim_compiler.src.dataStructure.visitors.printVisitor import toSLIMVisitor

import string
import copy





# ==================================================================================================
# class ProctypeInfo
# ==================================================================================================

class ProctypeInfo:
    """This class collects all the information necessary to construct one proctype and do the translation."""



    # ----------------------------------------------------------------------------------------------



    def __init__(self):
        self.packagename = ""           # a string
        self.componentname = ""         # a string
        self.implementationname = ""    # a string

        self.proctypename = None

        self.translator = None          # reference to Slim2Spin_Translator - instance (for .enumNames)
        
        self.in_data_ports = []         # contains tuples:   (Name: string, Type: Bool/Int/Real/..., Default: *TreeNode)
        self.out_data_ports = []        # contains tuples:   (Name: string, Type: Bool/Int/Real/..., Default: *TreeNode)
        self.in_event_ports = []        # contains strings:   names of in event ports
        self.out_event_ports = []       # contains strings:   names of out event ports

        self.subcomponents = []         # contains tuples:   (Name: string, Implementation: string, InDataPorts: list, OutDataPorts: list, InEventPorts: list, OutEventPorts: list, InModes: list)
        self.datasubcomponents = []     # contains tuples:   (Name: string, Type: Bool/Int/Real/..., Default: *TreeNode, InModes: list)

        self.data_port_connections = [] # contains tuples:   (SourcePort: string, DestinationPort: string, InModes: list)
        self.event_port_connections = []# contains tuples:   (SourcePort: string, DestinationPort: string, InModes: list)
        self.flows = []                 # contains tuples:   (OutputPort: string, Expression: *TreeNode, InModes: list)
        
        self.modes = []                 # contains tuples:   (ModeName: string, ModeType: int (nominalbehaviour.ModeType.INITIAL/ACTIVATION/...))
        self.transitions = []           # contains tuples:   (StartMode: string, Trigger: string, Guard: *TreeNode, Effects: list of BinaryOperatorTreeNodes, DestinationMode: string)
        self.probabilistictransitions = [] # contains tuples:(StartMode: string, Rate: string, DestinationMode: string)



    # ----------------------------------------------------------------------------------------------



    def _getOwnInDataPortByName(self, aName):
        matchings = filter(lambda (name, typ, default): name == aName, self.in_data_ports)
        if len(matchings) > 0:
            return matchings[0]
        else:
            return None
    

    def _isOwnInDataPort(self, aName):
        return self._getOwnInDataPortByName(aName) != None
    

    def _getOwnOutDataPortByName(self, aName):
        matchings = filter(lambda (name, typ, default): name == aName, self.out_data_ports)
        if len(matchings) > 0:
            return matchings[0]
        else:
            return None
        

    def _isOwnOutDataPort(self, aName):
        return self._getOwnOutDataPortByName(aName) != None


    def _getOwnDataSubcomponentByName(self, aName):
        matchings = filter(lambda (name, typ, default, ims): name == aName, self.datasubcomponents)
        if len(matchings) > 0:
            return matchings[0]
        else:
            return None


    def _isOwnDataSubcomponent(self, aName):
        return self._getOwnDataSubcomponentByName(aName) != None


    def _getSubcomponentsInDataPortByName(self, aName):
        parts = aName.split(".")
        if (len(parts) == 2):
            for (scname, scimp, scidp, scodp, sciep, scoep, ims) in filter(lambda (scname, scimp, scidp, scodp, sciep, scoep, ims): scname == parts[0], self.subcomponents):
                matchings = filter(lambda (name, typ, default): name == parts[1], scidp)
                if len(matchings) > 0:
                    (name, typ, default) = matchings[0]
                    return (scname, scimp, name, typ, default)
        else:
            return None


    def _isSubcomponentsInDataPort(self, aName):
        return self._getSubcomponentsInDataPortByName(aName) != None


    def _getSubcomponentsOutDataPortByName(self, aName):
        parts = aName.split(".")
        if (len(parts) == 2):
            for (scname, scimp, scidp, scodp, sciep, scoep, ims) in filter(lambda (scname, scimp, scidp, scodp, sciep, scoep, ims): scname == parts[0], self.subcomponents):
                matchings = filter(lambda (name, typ, default): name == parts[1], scodp)
                if len(matchings) > 0:
                    (name, typ, default) = matchings[0]
                    return (scname, scimp, name, typ, default)
        else:
            return None


    def _isSubcomponentsOutDataPort(self, aName):
        return self._getSubcomponentsOutDataPortByName(aName) != None


    def _isSubcomponentsInEventPort(self, aTrigger):
        parts = aTrigger.split(".")
        if len(parts) == 2:
            for (scname, scimp, scidp, scodp, sciep, scoep, ims) in filter(lambda (scname, scimp, scidp, scodp, sciep, scoep, ims): scname == parts[0], self.subcomponents):
                if parts[1] in sciep:
                    return True
        return False
        
        
    def _asSubcomponentsInPort(self, aTrigger):
        return self._isSubcomponentsInEventPort(aTrigger)

        
    def _isSubcomponentsOutEventPort(self, aName):
        parts = aName.split(".")
        if len(parts) == 2:
            for (scname, scimp, scidp, scodp, sciep, scoep, ims) in filter(lambda (scname, scimp, scidp, scodp, sciep, scoep, ims): scname == parts[0], self.subcomponents):
                if parts[1] in scoep:
                    return True
        return False


    def _asSubcomponentsOutPort(self, aTrigger):
        return self._isSubcomponentsOutEventPort(aTrigger)


    def _getSubcomponentByName(self, aName):
        aName = aName.split(".")[0]    # If aName is <scname>.<subname>, e.g. for SC's data port, only take scname
        for sc in self.subcomponents:
            if sc[0] == aName:
                return sc
        return None


    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


    def _getTypeOfVariable(self, aName):
        # Own In Data Port
        if self._isOwnInDataPort(aName):
            return self._getOwnInDataPortByName(aName)[1]
        # Own Out Data Port
        elif self._isOwnOutDataPort(aName):
            return self._getOwnOutDataPortByName(aName)[1]
        # Own Data Subcomponent
        elif self._isOwnDataSubcomponent(aName):
            return self._getOwnDataSubcomponentByName(aName)[1]
        # Subcomponent's In Data Port
        elif self._isSubcomponentsInDataPort(aName):
            return self._getSubcomponentsInDataPortByName(aName)[3]
        # Subcomponent's Out Data Port
        elif self._isSubcomponentsOutDataPort(aName):
            return self._getSubcomponentsOutDataPortByName(aName)[3]
        # None of all (enum constant or invalid)
        else:
            return None


    def _isVariableOfTypeReal(self, aName):
        typ = self._getTypeOfVariable(aName)
        if typ != None:
            return isinstance(typ, Real)


    def _isVariableOfTypeIntRange(self, aName):
        typ = self._getTypeOfVariable(aName)
        if typ != None:
            return isinstance(typ, IntRange)

        
    def _doesExpressionRequireEmbeddedCCode(self, anExpressionTreeNode):
        if anExpressionTreeNode == None:
            result = None
            
        elif isinstance(anExpressionTreeNode, BinaryOperatorTreeNode):
            # For boolean opearators (and, iff, imp, or, xnor, xor): always false
            if anExpressionTreeNode.operator in [OperatorType.AND, OperatorType.IFF, OperatorType.IMP, OperatorType.OR, OperatorType.XNOR, OperatorType.XOR]:
                result = False
            # For all others (+, -, *, ...; =, !=, <, ...): check subexpressions
            else:
                result = self._doesExpressionRequireEmbeddedCCode(anExpressionTreeNode.left) or self._doesExpressionRequireEmbeddedCCode(anExpressionTreeNode.right)
                
        elif isinstance(anExpressionTreeNode, UnaryOperatorTreeNode):
            result = self._doesExpressionRequireEmbeddedCCode(anExpressionTreeNode.subexpression)
            
        elif isinstance(anExpressionTreeNode, IdentifierTreeNode):
            result = self._isVariableOfTypeReal(anExpressionTreeNode.name)
            
        elif isinstance(anExpressionTreeNode, NumberTreeNode):
            result = anExpressionTreeNode.get_type() == NodeType.REAL

        elif isinstance(anExpressionTreeNode, CaseExpressionTreeNode):
            result = reduce(lambda b1, b2: b1 or b2, map(lambda (b, e): self._doesExpressionRequireEmbeddedCCode(e), anExpressionTreeNode.cases))
            result = result or self._doesExpressionRequireEmbeddedCCode(anExpressionTreeNode.default_expression)
            
        elif isinstance(anExpressionTreeNode, BooleanTreeNode) or isinstance(anExpressionTreeNode, CaseExpressionTreeNode) or \
             isinstance(anExpressionTreeNode, FlowEquationTreeNode) or isinstance(anExpressionTreeNode, FlowExpressionTreeNode):
            result = False
            
        else:
            result = None
            
        return result


    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


    @staticmethod
    def _datatypeToSPIN(aDatatype):
        # SLIM          Object      SPIN/C
        # ----          ------      ------
        # bool          Bool        bool
        # clock         Clock       ---
        # continuous    Continues   ---
        # enum          Enum        translate names to ints
        # int           Int         int
        # real          Real        int + c_code{float} / c_code{double}
        # [l..u]        IntRange    int + modulo
        if isinstance(aDatatype, Bool):
            txt = "bool"
        elif isinstance(aDatatype, Enum) or isinstance(aDatatype, Int) or isinstance(aDatatype, Real) or isinstance(aDatatype, IntRange):
            txt = "int"
        else:
            txt = "ERR: unsupported type '" + aDatatype.get_name() + "'"
        return txt


    @staticmethod
    def _implementationNameToSPIN(anImplementationName):
        return (anImplementationName.replace("::", "_")).replace(".", "_")

    
    def _EventToNumber(self, anEventPort, inEvents, outEvents):
        if anEventPort in inEvents:
            return 1 + inEvents.index(anEventPort)
        elif anEventPort in outEvents:
            return len(inEvents) + 1 + outEvents.index(anEventPort)
        else:
            return 0


    def _guardToSPIN(self, aGuard):
        if aGuard == None:
            return "true"
        else:
            return self._expressionToSPIN(aGuard)


    def _effectsToSPIN(self, anExpressionList, aPrefix = "\t\t"):
        assignToNew = ""
        copyNewToStore = ""
                
        for anExpressionTreeNode in anExpressionList:
            if isinstance(anExpressionTreeNode, BinaryOperatorTreeNode):
                if anExpressionTreeNode.operator == OperatorType.ASSIGN:
                    if isinstance(anExpressionTreeNode.left, IdentifierTreeNode):
                        if self._isOwnOutDataPort(anExpressionTreeNode.left.name) or self._isOwnDataSubcomponent(anExpressionTreeNode.left.name):
                            assignToNew += aPrefix + self._assignmentToSPIN(anExpressionTreeNode.left.name, anExpressionTreeNode.right, True) + ";\n"
                            copyNewToStore += aPrefix + self._identifierToSPIN(anExpressionTreeNode.left.name, False) + " = " + self._identifierToSPIN(anExpressionTreeNode.left.name, True) + ";\n"
                        else:
                            assignToNew += aPrefix +  "ERR: effect's left hand side '" + anExpressionTreeNode.left.name + "' does not identify out data port or data subcomponent." + ";\n"
                    else:
                        assignToNew += aPrefix + "ERR: effect's left hand side is no identifier" + ";\n"
                else:
                    assignToNew += aPrefix + "ERR: effect is no ASSIGN" + ";\n"
            else:
                assignToNew += aPrefix + "ERR: effect is no binop" + ";\n"
            
        return assignToNew + copyNewToStore


    def _assignmentToSPIN(self, target, expression, buffered = False):
        """Translates target = expression to SPIN. If buffered, its not stored to 'store_' but to 'new_'"""
        txt = ""

        # Possible Targets of Assignments and when they are allowed:
        #
        #                          Init    Reset   Effect
        #                          ----    -----   ------
        #
        # d) own in data port      (+) ~> Value will always be received from parent via !update.
        # a) own out data port     +       +       +
        #
        # b) own data subcomponent +       +       +
        #
        # c) sc's in data port     +       +
        # e) sc's out data port    (+) ~> Value will always be received from subcomponent via ?update.
        
        # Target is a real variable => start embedded c-code
        if self._isVariableOfTypeReal(target):
            txt = "c_code { *((float *)&(P" + self.proctypename + "->"   #Abuse int as float:   *((float *)&(PSender->myIntVar)) = ...
            txt += self._identifierToSPIN(target, buffered) + ")) = " + self._expressionToSPIN(expression, True) + self.exprt(expression) + "; }"

        # Target is a intrange variable => modulo calculation
        elif self._isVariableOfTypeIntRange(target):
            typ = self._getTypeOfVariable(target)
            txt = self._identifierToSPIN(target, buffered) + " = ((" + self._expressionToSPIN(expression) + self.exprt(expression)
            txt +=  ") - " + str(typ.lower) + ") % " + str(typ.upper - typ.lower + 1) + " + " + str(typ.lower)

        # Target is a "normal" variable => just assign
        else:
            txt = self._identifierToSPIN(target, buffered) + " = " + self._expressionToSPIN(expression) + self.exprt(expression)

        return txt


    def _identifierToSPIN(self, anIdentifier, buffered = False):
        """Translates a SLIM identifier to the corresponding SPIN variable."""

        if buffered:
            where = "new_"
            storageSuffix = "_new"
        else:
            where = "store_"
            storageSuffix = "_store"

        # a) Assign to own out data port
        if self._isOwnOutDataPort(anIdentifier):
            return where + "self_" + anIdentifier
        # b) Assign to own data subcomponent
        elif self._isOwnDataSubcomponent(anIdentifier):
            return "self_" + anIdentifier + storageSuffix
        # c) Assign to subcomponent's in data port -- only allowed in initialization and reset
        elif self._isSubcomponentsInDataPort(anIdentifier):
            return where + anIdentifier.split(".")[0] + "_" + anIdentifier.split(".")[1]
        # d) Assign to own in data port -- only allowed in initialization
        elif self._isOwnInDataPort(anIdentifier):
            return where + "self_" + anIdentifier
        # e) Assign to subcomponent's out data port -- only allowed in initialization
        elif self._isSubcomponentsOutDataPort(anIdentifier):
            return where + anIdentifier.split(".")[0] + "_" + anIdentifier.split(".")[1]
        # f) Invalid assignment
        else:
            return "ERR: '" + anIdentifier + "' is invalid identifier and cannot be translated to SPIN variable"


    def exprt(self, anExpressionTreeNode):
        return "" # "[" + str(self._doesExpressionRequireEmbeddedCCode(anExpressionTreeNode)) + "]"
        #txt +=  "["+NodeType.type_to_str(Expressions.type_check(anExpressionTreeNode, None, None))+"]"
    
    def _expressionToSPIN(self, anExpressionTreeNode, asEmbeddedCCode = False):
        """Translates SLIM-expression stored in anExpressionTreeNode to SPIN. Assumes, it is not None (otherwise this Expression should not be used!). Does not handle assignments (use _effectsToSPIN instead)."""

        # First check, whether embedded C-Code has to be opened for subexpression (which must be a comparison)
        if (not asEmbeddedCCode) and self._doesExpressionRequireEmbeddedCCode(anExpressionTreeNode):
            return "c_expr { " + self._expressionToSPIN(anExpressionTreeNode, True) + " }"
            
        # Distinguish the different Node Types and generate appropriate SPIN expressions
        if isinstance(anExpressionTreeNode, BinaryOperatorTreeNode):
            left  = "(" + self._expressionToSPIN(anExpressionTreeNode.left, asEmbeddedCCode)  + ")"
            right = "(" + self._expressionToSPIN(anExpressionTreeNode.right, asEmbeddedCCode) + ")"

            # SLIM      OperatorType       SPIN (int)   C (real)
            # ----      ------------       ----         -
            # +         .PLUS              +            +
            # -b        .MINUS             -            -
            # *         .TIMES             *            *
            # /         .DIVIDE            /            /
            # mod       .MOD             ! %          ! %
            #
            # =         .EQUAL           ! ==           ==
            # !=        .NOTEQUAL          !=           !=
            # <         .LESS              <            <
            # >         .GREATER           >            >
            # <=        .LESSOREQUAL       <=           <=
            # >=        .GREATEROREQUAL    >=           >=
            #
            # and       .AND             ! &&         ! &&
            # iff       .IFF            !! ...       !! ...
            # imp       .IMP            !! ...       !! ...
            # or        .OR              ! ||         ! ||
            # xnor      .XNOR           !! ...       !! ...
            # xor       .XOR            !! ...       !! ...
            if anExpressionTreeNode.operator == OperatorType.MOD:
                txt = left + " %" + self.exprt(anExpressionTreeNode) + " " + right
            elif anExpressionTreeNode.operator == OperatorType.EQUAL:
                txt = left + " ==" + self.exprt(anExpressionTreeNode) + " " + right
            elif anExpressionTreeNode.operator == OperatorType.AND:
                txt = left + " &&" + self.exprt(anExpressionTreeNode) + " " + right
            elif anExpressionTreeNode.operator == OperatorType.OR:
                txt = left + " ||" + self.exprt(anExpressionTreeNode) + " " + right
            elif anExpressionTreeNode.operator in [OperatorType.IFF, OperatorType.XNOR]:
                # a <=> b   <---->   (a and b) or (!a and !b)  /  (a => b) and (b => a)  <->  (b or !a) and (a or !b)
                # a XNOR b   <---->   a <=> b   <---->   ...
                txt = "(" + left + " && " + right + ") or (!" + left + " && !" + right + ")"
            elif anExpressionTreeNode.operator == OperatorType.IMP:
                # a => b   <---->   b or !a
                txt = right + " || !" + left
            elif anExpressionTreeNode.operator == OperatorType.XOR:
                # a XOR b   <---->   (a and !b) or (!a and b)
                txt = "(" + left + " && !" + right + ") || (!" + left + " && " + right + ")"
            elif anExpressionTreeNode.operator in [OperatorType.PLUS, OperatorType.MINUS, OperatorType.TIMES, OperatorType.DIVIDE, OperatorType.NOTEQUAL,
                                                   OperatorType.LESS, OperatorType.GREATER, OperatorType.LESSOREQUAL, OperatorType.GREATEROREQUAL]:
                txt = left + " " + OperatorType.operator_to_str(anExpressionTreeNode.operator) + self.exprt(anExpressionTreeNode) + " " + right
            else:
                txt = left + "ERR: '" + OperatorType.operator_to_str(anExpressionTreeNode.operator) + "' unsupported as binary operator" + right

        elif isinstance(anExpressionTreeNode, UnaryOperatorTreeNode):
            right = "(" + self._expressionToSPIN(anExpressionTreeNode.subexpression, asEmbeddedCCode) + ")" + self.exprt(anExpressionTreeNode.subexpression)
            
            # SLIM      OperatorType       SPIN (int)
            # ----      ------------       ----
            # -u        .UNARY_MINUS       -
            # not       .NOT             ! !
            if anExpressionTreeNode.operator == OperatorType.UNARY_MINUS:
                txt = "-" + right
            elif anExpressionTreeNode.operator == OperatorType.NOT:
                txt = "!" + right
            else:
                txt = "ERR: '" + OperatorType.operator_to_str(anExpressionTreeNode.operator) + "' unsupported as unary operator" + right
            txt += self.exprt(anExpressionTreeNode)
            
        elif isinstance(anExpressionTreeNode, IdentifierTreeNode):
            # a/b/c) Read own in/out data port/data subcomponent
            if self._isOwnInDataPort(anExpressionTreeNode.name) or self._isOwnOutDataPort(anExpressionTreeNode.name) or self._isOwnDataSubcomponent(anExpressionTreeNode.name):
                if asEmbeddedCCode:
                    if self._isVariableOfTypeReal(anExpressionTreeNode.name):
                        # Typecast-Trick for abusing Promela-ints as floats:   *((float *)&(PSender->myIntVar))
                        txt = "*((float *)&(P" + self.proctypename + "->" + self._identifierToSPIN(anExpressionTreeNode.name) + "))"
                    else:
                        txt = "P" + self.proctypename + "->" + self._identifierToSPIN(anExpressionTreeNode.name)
                else:
                    txt = self._identifierToSPIN(anExpressionTreeNode.name)
            # d) Identifier is from ENUM data type
            elif (self.translator != None) and (anExpressionTreeNode.name in self.translator.enumNames):
                txt = "enum_" + anExpressionTreeNode.name
            # e) Invalid Read   (e.g., read a subcomponent's outgoing data port)
            else:
                txt = "ERR: invalid identifier '" + anExpressionTreeNode.name + "' on effect's right hand side"
            txt += self.exprt(anExpressionTreeNode)

        elif isinstance(anExpressionTreeNode, BooleanTreeNode):
            if anExpressionTreeNode.value:
                txt = "true"
            else:
                txt = "false"
            txt += self.exprt(anExpressionTreeNode)

        elif isinstance(anExpressionTreeNode, NumberTreeNode):
            txt = str(anExpressionTreeNode.number)
            txt += self.exprt(anExpressionTreeNode)

        elif isinstance(anExpressionTreeNode, CaseExpressionTreeNode):
            if asEmbeddedCCode:
                sign = " ? "
            else:
                sign = " -> "
            
            # Translate first case
            txt = "(" + self._expressionToSPIN(anExpressionTreeNode.cases[0][0], asEmbeddedCCode) + self.exprt(anExpressionTreeNode.cases[0][0])
            txt += sign + self._expressionToSPIN(anExpressionTreeNode.cases[0][1], asEmbeddedCCode) + self.exprt(anExpressionTreeNode.cases[0][1]) + " : "
            # Only one case => translate "otherwise"
            if len(anExpressionTreeNode.cases) == 1:
                txt += self._expressionToSPIN(anExpressionTreeNode.default_expression, asEmbeddedCCode) + self.exprt(anExpressionTreeNode.default_expression)
            # Not yet "otherwise" => translate the remaining cases
            else:
                smallerCaseExpression = copy.copy(anExpressionTreeNode)
                smallerCaseExpression.cases = smallerCaseExpression.cases[1:] # ignore first case
                txt = self._expressionToSPIN(smallerCaseExpression, asEmbeddedCCode)
            txt += ")" + self.exprt(anExpressionTreeNode)

        elif isinstance(anExpressionTreeNode, FlowEquationTreeNode) or isinstance(anExpressionTreeNode, FlowExpressionTreeNode):
            txt = "ERR: Flow Equations for continuous variables are not supported"

        else:
            txt = "ERR: Unknown ExpressionTreeNode"

        return txt


    def expressionsToString(self, anExpressionList):
        if len(anExpressionList) > 0:
            txt = ""
            for expression in anExpressionList:
                txt += self.expressionToString(expression) + "; "
            return txt
        else:
            return "EmptyExpressionList"


    def expressionToString(self, anExpressionTreeNode):
        if anExpressionTreeNode == None:
            txt = "EmptyExpression"   # Alternatively, use: "else: txt = toSLIMVisitor(anExpressionTreeNode)", but that sometimes misses s.th.
        elif isinstance(anExpressionTreeNode, BinaryOperatorTreeNode):
            txt = "(" + self.expressionToString(anExpressionTreeNode.left) + ") " + str(OperatorType.operator_to_str(anExpressionTreeNode.operator)) + " (" + self.expressionToString(anExpressionTreeNode.right) + ")"
        elif isinstance(anExpressionTreeNode, UnaryOperatorTreeNode):
            txt = OperatorType.operator_to_str(anExpressionTreeNode.operator) +  "(" + self.expressionToString(anExpressionTreeNode.subexpression) + ")"
        elif isinstance(anExpressionTreeNode, CaseExpressionTreeNode):
            txt = "case "
            for (b, e) in anExpressionTreeNode.cases:
                txt += self.expressionToString(b) + " : " + self.expressionToString(e) + "; "
            txt += "otherwise " + self.expressionToString(anExpressionTreeNode.default_expression) + " end"
        elif isinstance(anExpressionTreeNode, IdentifierTreeNode):
            txt = str(anExpressionTreeNode.name)
        elif isinstance(anExpressionTreeNode, BooleanTreeNode):
            txt = str(anExpressionTreeNode.value)
        elif isinstance(anExpressionTreeNode, NumberTreeNode):
            txt = str(anExpressionTreeNode.number)
        elif isinstance(anExpressionTreeNode, FlowEquationTreeNode):
            txt = anExpressionTreeNode.variable + "' = " + self.expressionToString(anExpressionTreeNode.flow_expression)
        elif isinstance(anExpressionTreeNode, FlowExpressionTreeNode):
            txt = ""
            if anExpressionTreeNode.times != None:
                txt += str(anExpressionTreeNode.times) + " * "
            txt += self.expressionToString(anExpressionTreeNode.variable)
            if anExpressionTreeNode.offset != None:
                txt += " + " + str(anExpressionTreeNode.offset)
        else:
            txt = "ERR: Unknown Expression Construct"
        return txt


    def _resetReactivatedDataSubcomponents(self, SourceMode, DestinationMode, aPrefix = "\t\t"):
        txt = ""

        # Reset Data Subcomponents which are reactivated
        for (name, typ, default, ims) in filter(lambda (name, typ, default, ims): (SourceMode not in ims) and (DestinationMode in ims), self.datasubcomponents):
            txt += aPrefix + self._assignmentToSPIN(name, default) + ";\n"

        return txt


    def _resetDisconnectedDataPorts(self, SourceMode, DestinationMode, aPrefix = "\t\t"):
        txt = ""

        # Disconnected from Data Port Connection
        for (sp, dp, ims) in filter(lambda (sp, dp, ims): (SourceMode in ims) and (DestinationMode not in ims), self.data_port_connections):
            # Own out data port (out - out connection)
            if self._isOwnOutDataPort(dp):
                (name, typ, default) = self._getOwnOutDataPortByName(dp)
                txt += aPrefix + self._assignmentToSPIN(dp, default) + ";\n"
            # Subcomponents' in data port (in - in / out - in connection)
            elif self._isSubcomponentsInDataPort(dp):
                (scname, scimp, name, typ, default) = self._getSubcomponentsInDataPortByName(dp)
                txt += aPrefix + self._assignmentToSPIN(dp, default) + ";\n"
            # Everything else in invalid!
            else:
                txt += aPrefix + "// ERR: invalid data port identifier '" + dp + "' disconnected from data port connection\n"
            
        # Disconnected from Flow
        for (op, expr, ims) in filter(lambda (op, expr, ims): (SourceMode in ims) and (DestinationMode not in ims), self.flows):
            # Own out data port (out - out connection)
            if self._isOwnOutDataPort(op):
                (name, typ, default) = self._getOwnOutDataPortByName(op)
                txt += aPrefix + self._assignmentToSPIN(op, default) + ";\n"
            # Everything else in invalid!
            else:
                txt += aPrefix + "// ERR: invalid data port identifier '" + op + "' disconnected from flow\n"

        return txt


    def _adjustSubcomponentsActivation(self, SourceMode, DestinationMode, aPrefix = "\t\t"):
        txt = ""
        
        # Deactivate those, active in SourceMode but inactive in DestinationMode
        for (scname, scimp, scidp, scodp, sciep, scoep, ims) in filter(lambda (scname, scimp, scidp, scodp, sciep, scoep, ims): (SourceMode in ims) and (DestinationMode not in ims), self.subcomponents):
            txt += aPrefix + scname + "!deactivate" + (len(scidp)+len(scodp)+1) * ", 0" + ";\n"
            txt += aPrefix + scname + "?deactivate" + (len(scidp)+len(scodp)+1) * ", _" + ";\n"

        # Activate those, inactive in SourceMode but active in DestinationMode
        for (scname, scimp, scidp, scodp, sciep, scoep, ims) in filter(lambda (scname, scimp, scidp, scodp, sciep, scoep, ims): (SourceMode not in ims) and (DestinationMode in ims), self.subcomponents):
            txt += aPrefix + scname + "!activate" + (len(scidp)+len(scodp)+1) * ", 0" + ";\n"
            txt += aPrefix + scname + "?activate" + (len(scidp)+len(scodp)+1) * ", _" + ";\n"

        return txt


    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


    def _prepare(self):
        """Some preparation on the collected data to simplify translation to SPIN."""
        
        # Proctype name consists of (Package + Component + Implementation) name
        if self.proctypename == None:
            self.proctypename = self.packagename + "_" + self.componentname + "_" + self.implementationname

        # We often need to know how many parameters (in + out data ports, + 1 for event communication) this component has
        self.nrparams = len(self.in_data_ports) + len(self.out_data_ports) + 1

        # Find out whether this component supports mode history and in which mode to start
        for (mname, mtype) in self.modes:
            if mtype == slim_compiler.src.dataStructure.nominalbehaviour.ModeType.INITIAL:
                self.startmode = mname
                self.modehistory = True
            if mtype == slim_compiler.src.dataStructure.nominalbehaviour.ModeType.ACTIVATION:
                self.startmode = mname
                self.modehistory = False

        # Just to make sure, that event ports always get the same number, sort the list (which was created from a dictionary, which is not guaranteed to be ordered!)
        self.in_event_ports.sort()
        self.out_event_ports.sort()

        for (scname, scimp, scidp, scodp, sciep, scoep, ims) in self.subcomponents:
            sciep.sort()
            scoep.sort()      



    # ----------------------------------------------------------------------------------------------    


        
    def to_SPIN(self):

        self._prepare()

        # TODO:
        # * Code Clean-Up:
        #   - Set root inputs to defaults. (Not in decl, since real!)
        #   - Perform renaming of variable indentifiers. Make more use of "_identifierToSPIN"/_ownDataSubcomponentToSPIN... which maps dataports and so on to the corresponding variable?!?
        #   - Check that always appropriate subset of subcomponents (all vs. only active) are used.
        #   - Try to extract identical/similar code blocks and abstract them to a helper function.
        #   - Introduce raising errors
        # * Look for Optimizations
        #   - Avoid "empty" inline functions.
        #   - No invalidate when nothing has changed? Problem: Supercomponent might wait on that "invalidate"? Possibly, use timeout.
        #   - Shortcut update when input values have not changed? Problem: Output values might change due to mode changes & consequently by changed connections/flows!
        #   - Transition effects: distinguish old/new only, when new is set before old is read the last time! [_
        #   - Only reset reactivated data subcomponents to defaults when no assignment to them follows! [_
        #   - Only reset disconnected ports when no new connection available.
        #   - Activate to reset only subcomponents with mode history!
        #   - Fixpoing Iteration: When avoidable?
        # * Make a real skript out of it.
        #   - Read parameters from command line, store output to file, ...
        # * Others
        #   - Check that by definition no further checks are necessary.
        #   - What about case sensitivity?

        # NOTE:
        # - Datasubcomponents may not be called ...                         ===> Solution: "self_" . <dsc-name> . ("_store"|"_new")
        #   - "eventInt", "eventSum"
        #   - as any other subcomponent (by definition fulfilled)
        #   - "pid_" + subcomponent
        #   - "store_..._...", "new_..._..."
        #   - as any Macro/inline function defined
        # - Modes may not be called ...                                     ===> Solution: modes prefixed mit "mode_"
        #   - "FirstActivationOrReactivationWithoutModeHistory"
        #   - "init"

        # YET UNCONSIDERED IN SPIN-SOLUTION:
        # - input from environment (or should this be modelled as a component?)
        # - probabilistic transitions
        # - time (Datatypes: Clock, Continues; Differential Equations)

        # Questions/Hints to Parser:
        #   ? How to invoke most cleverly?
        #   ? Why does "type_check" not work?
        #   ! Before "otherwise" in case-Expressions, no ";" is required!
        #   ! In ProbabilityModeTransitionLabel you call the property "probabilty", not "probability".
        #   ? Why are diverse things not checked? Do I have to call a check-method after parsing?
        #   ? File "C:\Users\M & M\Desktop\compass-tools-20090504\tools\slim_compiler\src\dataStructure\expressions.py", line 704, in _type_check_binary_operator     file)   TypeError: _type_check_binary_operator_int() takes exactly 5 arguments (6 given)


        
        # *** Start with head ***
        prelude = "// ----- " + self.packagename + "::" + self.componentname + "." + self.implementationname + " --------------------------------------------------\n\n"
        result = "proctype " + self.proctypename + "(chan parent)\n"
        result += "{\n"
        result += "\tprintf(\"### " + self.proctypename + "(%d) ### started\\n\", _pid);\n"



        # *** Variables for storing Inputs, Outputs, Subcomponents ***

        # Naming convention:   <Whose?>_<ElementName>_<store/new>
        # => Own In Data Port:  self_<idp-name>_<store/new>           rename!
        # => Own Out Data Port: self_<odp-name>_<store/new>           rename!
        # => Own Data Subcomp.: self_<dsc-name>_<store/new>           (renamed from: <dsc-name>)
        # => SC's In Data Port: <sc-name>_<idp-name>_<store/new>      rename!
        # => SC's Out Data P.:  <sc-name>_<odp-name>_<store/new>      rename!
        #
        # * PID of SCs:         pid_<sc-name>                         (already initial choice)
        # * CHAN of SCs:        chan_<sc-name>                        rename!
        #
        # * Enum Names:         enum_<enum-name>

        result += "\n\n\n\t// *** Variables for storing Inputs, Outputs, Subcomponents ***\n"
        result += "\tint eventInt;\n"
        result += "\tint eventSum;\n\n"

        # Own Data Ports
        for (name, typ, default) in self.in_data_ports:
            result += "\t" + self._datatypeToSPIN(typ) + " store_self_" + name + ";\n"
        for (name, typ, default) in self.out_data_ports:
            result += "\t" + self._datatypeToSPIN(typ) + " store_self_" + name + "; "
            result += self._datatypeToSPIN(typ) + " new_self_" + name + ";\n"

        # Own Data Subcomponents
        for (name, typ, default, ims) in self.datasubcomponents:
            result += "\n\t" + self._datatypeToSPIN(typ) + " self_" + name + "_store; "
            result += self._datatypeToSPIN(typ) + " self_" + name + "_new;\n"

        # Subcomponents' Data Ports
        for (subcomponentname, subcomponentimplementation, scidp, scodp, sciep, scoep, ims) in self.subcomponents:
            result += "\n\tpid pid_" + subcomponentname + ";\n"
            for (name, typ, default) in scidp:
                result += "\t" + self._datatypeToSPIN(typ) + " store_" + subcomponentname + "_" + name + ";\n"
            for (name, typ, default) in scodp:
                result += "\t" + self._datatypeToSPIN(typ) + " store_" + subcomponentname + "_" + name + "; "
                result += self._datatypeToSPIN(typ) + " new_" + subcomponentname + "_" + name + ";\n"


        
        # *** Create Subcomponents ***
        result += "\n\n\n\t// *** Create Subcomponents ***\n"
        for (subcomponentname, subcomponentimplementation, scidp, scodp, sciep, scoep, ims) in self.subcomponents:
            result += "\n\t// " + subcomponentname + " : " + subcomponentimplementation + "\n"
            result += "\tchan " + subcomponentname + " = [0] of {mtype"
            for (name, typ, default) in scidp:
                result += ", " + self._datatypeToSPIN(typ)
            for (name, typ, default) in scodp:
                result += ", " + self._datatypeToSPIN(typ)
            result += ", int};\n"
            result += "\tpid_" + subcomponentname + " = run " + self._implementationNameToSPIN(subcomponentimplementation) + "(" + subcomponentname + ");\n"

        

        # *** Wait for first activation ***
        result += "\n\n\n\t// *** Wait for first activation ***\n"
        result += "FirstActivationOrReactivationWithoutModeHistory:\n"
        result += "\tparent?activate" + self.nrparams * ", _" + ";\n"

        # Initialize Own Out Data Ports
        result += "\n\t// Initialize Own Out Data Ports\n"
        # No need to initialize own in data ports, as their values will be received from parent via !update.
        #for (name, typ, default) in self.in_data_ports:
        #    if default != None:
        #        result += "\t" + self._assignmentToSPIN(name, default) + ";\n"
        for (name, typ, default) in self.out_data_ports:
            if default != None:
                result += "\t" + self._assignmentToSPIN(name, default) + ";\n"

        # Initialize Data Subcomponents
        result += "\n\t// Initialize Data Subcomponents (only those active in start mode)\n"
        for (name, typ, default, ims) in filter(lambda (name, typ, default, ims): ims == [] or self.startmode in ims, self.datasubcomponents):
            if default != None:
                result += "\t" + self._assignmentToSPIN(name, default) + ";\n"

        # Initialize Subcomponents' In Data Ports
        # -> This *MUST* be done, since those values are sent to the subcomponent on updates, even if no connection exists, i.e. there is no "no input delievered" info.
        # -> On deactivation of data port connections/flows, those defaults are restored.
        # -> On deactivation of subcomponent, connections to it must be deactivated to => defaults restored (or no connection, thus already the defaults)
        result += "\n\t// Initialize Non-Data Subcomponents' In Data Ports (must be done since sent on updates)\n"
        for (subcomponentname, subcomponentimplementation, scidp, scodp, sciep, scoep, ims) in self.subcomponents:
            for (name, typ, default) in scidp:
                if default != None:
                    result += "\t" + self._assignmentToSPIN(subcomponentname + "." + name, default) + ";\n"
            # No need to initialize subcomponents' out data ports, as their values will be received from subcomponents via ?update
            #for (name, typ, default) in scodp:
            #    if default != None:
            #        result += "\t" + self._assignmentToSPIN(subcomponentname + "." + name, default) + ";\n"

        # Activate Non-Data Subcomponents
        result += "\n\t// Activate Non-Data Subcomponents (only those active in start mode)\n"
        for (scname, scimp, scidp, scodp, sciep, scoep, ims) in filter(lambda (scname, scimp, scidp, scodp, sciep, scoep, ims): ims == [] or self.startmode in ims, self.subcomponents):
            result += "\t" + scname + "!activate" + (len(scidp) + len(scodp) + 1) * ", 0" + ";\n"
            result += "\t" + scname + "?activate" + (len(scidp) + len(scodp) + 1) * ", _" + ";\n"

        # Activation completed
        result += "\n\t// Activation completed\n"
        result += "\tparent!activate" + self.nrparams * ", 0" + ";\n"
        result += "\tgoto " + self.startmode + ";\n"



        # *** Run as automaton & React to messages ***
        result += "\n\n\n\t// *** Run as automaton & React to messages ***\n"

        # Iterate over all modes
        for (mname, mtype) in self.modes:


            # Label
            result += "\n\n" + mname + ":\n"
            result += "\tprintf(\"### " + self.proctypename + "(%d) ### enters state '" + mname + "'\\n\", _pid);\n"


            # Start mode's do-loop
            result += "\tdo\n"


            # Transitions
            for (sm, tr, gu, ef, dm) in filter(lambda (sm, tr, gu, ef, dm): sm == mname, self.transitions):

                # Proactive Transition (w/o event)
                if tr == None:
                    result += "\n\t// Proactive Transition (w/o event): %s -[when %s then %s]-> %s\n" % (sm, self.expressionToString(gu), self.expressionsToString(ef), dm)
                    result += "\t:: nT(" + self._guardToSPIN(gu) + ") ->\n"
                    result += self._resetReactivatedDataSubcomponents(sm, dm) # before effects to make sure that assignments overwrite default
                    result += self._effectsToSPIN(ef)
                    result += self._resetDisconnectedDataPorts(sm, dm)        # after effects, since old values can be read in assignments; no overwrite problem, since no assignment to data ports receiving values from connections allowed!
                    result += self._adjustSubcomponentsActivation(sm, dm)
                    result += "\t\tparent!invalidate" + self.nrparams * ", 0" + ";  // propagate changes; fT by root!\n"
                    result += "\t\tgoto " + dm + ";\n"

                # Master Transition (own out event)
                elif (tr.find(".") == -1) and (tr in self.out_event_ports):
                    result += "\n\t// Master Transition (own out event): %s -[%s when %s then %s]-> %s\n" % (sm, tr, self.expressionToString(gu), self.expressionsToString(ef), dm)
                    result += "\t:: nT(" + self._guardToSPIN(gu) + ") ->\n"
                    result += "\t\tparent!trymultiway" + (self.nrparams-1)* ", 0" +  ", " + str(self._EventToNumber(tr, self.in_event_ports, self.out_event_ports)) + ";\n"
                    result += "\t\tparent?trymultiway" + (self.nrparams-1)* ", _" + ", eventInt;\n"
                    result += "\t\tif\n"
                    result += "\t\t:: eventInt > 0 ->\n"
                    result += self._resetReactivatedDataSubcomponents(sm, dm, "\t\t\t")
                    result += self._effectsToSPIN(ef, "\t\t\t")
                    result += self._resetDisconnectedDataPorts(sm, dm, "\t\t\t")
                    result += self._adjustSubcomponentsActivation(sm, dm, "\t\t\t")
                    result += "\t\t\tparent!invalidate" + self.nrparams * ", 0" + "; // Not necessary if no effects, but parent may expect to retrieve this message!; fT by root!\n"
                    result += "\t\t\tgoto " + dm + ";\n"
                    result += "\t\t:: else -> fT; // no communication partner => continue with loop (no jump/reentrance of state)\n"
                    result += "\t\tfi;\n"

                # Master Transition (subcomponent's in event)
                elif (tr.find(".") != -1) and (self._asSubcomponentsInPort(tr)):
                    (scname, scimp, scidp, scodp, sciep, scoep, ims) = self._getSubcomponentByName(tr)
                    result += "\n\t// Master Transition (subcomponent's in event): %s -[%s when %s then %s]-> %s\n" % (sm, tr, self.expressionToString(gu), self.expressionsToString(ef), dm)
                    result += "\t:: nT(" + self._guardToSPIN(gu) + ") ->\n"
                    result += "\t\t" + scname + "!trymultiway" + (len(scidp) + len(scodp))* ", 0" +  ", " + str(self._EventToNumber(tr.split(".")[1], sciep, scoep)) + ";\n"
                    result += "\t\t" + scname + "?trymultiway" + (len(scidp) + len(scodp))* ", _" + ", eventInt;\n"
                    result += "\t\tif\n"
                    result += "\t\t:: eventInt > 0 ->\n"
                    result += self._resetReactivatedDataSubcomponents(sm, dm, "\t\t\t")
                    result += self._effectsToSPIN(ef, "\t\t\t")
                    result += self._resetDisconnectedDataPorts(sm, dm, "\t\t\t")
                    result += self._adjustSubcomponentsActivation(sm, dm, "\t\t\t")
                    result += "\t\t\tparent!invalidate" + self.nrparams * ", 0" + "; // Not necessary if no effects, but parent may expect to retrieve this message!; fT by root!\n"
                    result += "\t\t\tgoto " + dm + ";\n"
                    result += "\t\t:: else -> fT; // no communication partner => continue with loop (no jump/reentrance of state)\n"
                    result += "\t\tfi;\n"

                # Reactive Transition => to be treated in "Incomming Message: trymultiway"-section!
                else:
                    result += "\n\t// No proactive/master transition: %s -[%s when %s then %s]-> %s\n" % (sm, tr, gu, self.expressionsToString(ef), dm)


            # Probabilistic Transitions
            for (sm, rate, dm) in filter(lambda (sm, rate, dm): sm == mname, self.probabilistictransitions):
                result += "\n\t// Probabilistic Transition: %s -[prob %s]-> %s\n" % (sm, rate, dm)
                result += "\t// ERR: Warning: Probabilistic Transitions are not fully supported. They are transformed into tau-Transitions!!!\n"
                result += "\t:: nT(true) ->\n"
                result += self._resetReactivatedDataSubcomponents(sm, dm)
                # result += self._effectsToSPIN(ef)
                result += self._resetDisconnectedDataPorts(sm, dm)
                result += self._adjustSubcomponentsActivation(sm, dm)
                result += "\t\tparent!invalidate" + self.nrparams * ", 0" + ";  // propagate changes; fT by root!\n"
                result += "\t\tgoto " + dm + ";\n"
                

            # Incomming Message: trymultiway (from parent)
            for inEventPort in self.in_event_ports:
                result += "\n\t// Incomming Message: trymultiway (from parent: " + inEventPort + " (" + str(self._EventToNumber(inEventPort, self.in_event_ports, self.out_event_ports)) + ") )\n"

                # Receive incomming trymultiway
                result += "\t:: parent?trymultiway" + (self.nrparams-1) * ", _" + ", " + str(self._EventToNumber(inEventPort, self.in_event_ports, self.out_event_ports)) + " ->\n"
                result += "\t\teventSum = 0;\n"

                # Ask subcomponents
                forwardingConnections = filter(lambda (sp, dp, ims): sp == inEventPort and self._asSubcomponentsInPort(dp) and (ims == [] or mname in ims), self.event_port_connections) 
                result += "\n\t\t// Ask subcomponents via Event Port Connections\n"
                for (sp, dp, ims) in forwardingConnections:
                    (scname, scimp, scidp, scodp, sciep, scoep, ims) = self._getSubcomponentByName(dp)
                    result += "\n\t\t\t// Event Port Connection: %s -> %s (and current mode %s is in modes %s)\n" % (sp, dp, mname, ims)
                    result += "\t\t\t" + scname + "!trymultiway" + (len(scidp)+len(scodp)) * ", 0" + ", " + str(self._EventToNumber(dp.split(".")[1], sciep, scoep)) + ";\n"
                    result += "\t\t\t" + scname + "?trymultiway" + (len(scidp)+len(scodp)) * ", _" + ", eventInt;\n"
                    result += "\t\t\teventSum = eventSum + eventInt;\n"

                # Reactive Transitions (own in event)
                reactiveTransitions = filter(lambda (sm, tr, gu, ef, dm): sm == mname and tr == inEventPort, self.transitions)
                if len(reactiveTransitions) > 0:
                    result += "\n\t\t// Reactive Transitions\n"
                    result += "\t\tif\n";
                    for (sm, tr, gu, ef, dm) in reactiveTransitions:
                        result += "\n\t\t\t// Reactive Transition (own in event): %s -[%s when %s then %s]-> %s\n" % (sm, tr, self.expressionToString(gu), self.expressionsToString(ef), dm)
                        result += "\t\t\t:: " + self._guardToSPIN(gu) + " ->\n"
                        result += self._resetReactivatedDataSubcomponents(sm, dm, "\t\t\t\t")
                        result += self._effectsToSPIN(ef, "\t\t\t\t")
                        result += self._resetDisconnectedDataPorts(sm, dm, "\t\t\t\t")
                        result += self._adjustSubcomponentsActivation(sm, dm, "\t\t\t\t")
                        result += "\t\t\t\t// Answer to trymultiway, no invalidate necessary since sended by Master Transition\n"
                        result += "\t\t\t\tparent!trymultiway" + (self.nrparams-1) * ", 0" + ", eventSum+1;\n"
                        result += "\t\t\t\tgoto " + dm + ";\n"
                    result += "\n\t\t\t// No Reactive Transition enabled => just answer to trymultiway\n"
                    result += "\t\t\t:: else ->\n"
                    result += "\t\t\t\tparent!trymultiway" + (self.nrparams-1) * ", 0" + ", eventSum;\n"
                    result += "\n\t\tfi;\n"
                
                # Answer incomming trymultiway (if no reactive transition)
                else:
                    result += "\n\t\t// No Reactive Transitions => just answer to trymultiway\n"
                    result += "\t\tparent!trymultiway" + (self.nrparams-1) * ", 0" + ", eventSum;\n"


            # Incomming Message: trymultiway (from subcomponent)
            for (scname, scimp, scidp, scodp, sciep, scoep, ims) in self.subcomponents:
                for outEventPort in scoep:
                    result += "\n\t// Incomming Message: trymultiway (from subcomponent " + scname + ": " + outEventPort + " (" + str(self._EventToNumber(outEventPort, sciep, scoep)) + ") )\n"

                    # Receive incomming trymultiway
                    result += "\t:: " + scname + "?trymultiway" + (len(scidp) + len(scodp)) * ", _" + ", " + str(self._EventToNumber(outEventPort, sciep, scoep)) + " ->\n"
                    result += "\t\teventSum = 0;\n"

                    # Ask parent
                    forwardingConnections = filter(lambda (sp, dp, ims): sp == scname + "." + outEventPort and (dp in self.out_event_ports) and (ims == [] or mname in ims), self.event_port_connections) 
                    result += "\n\t\t// Ask parent via Event Port Connections\n"
                    for (sp, dp, ims) in forwardingConnections:
                        result += "\n\t\t\t// Event Port Connection: %s -> %s (and current mode %s is in modes %s)\n" % (sp, dp, mname, ims)
                        result += "\t\t\tparent!trymultiway" + (self.nrparams-1) * ", 0" + ", " + str(self._EventToNumber(dp, self.in_event_ports, self.out_event_ports)) + ";\n"
                        result += "\t\t\tparent?trymultiway" + (self.nrparams-1) * ", _" + ", eventInt;\n"
                        result += "\t\t\teventSum = eventSum + eventInt;\n"
                        
                    # Ask other subcomponents
                    forwardingConnections = filter(lambda (sp, dp, ims): sp == scname + "." + outEventPort and self._asSubcomponentsInPort(dp) and (ims == [] or mname in ims), self.event_port_connections) 
                    result += "\n\t\t// Ask subcomponents via Event Port Connections\n"
                    for (sp, dp, ims) in forwardingConnections:
                        (scname2, scimp2, scidp2, scodp2, sciep2, scoep2, ims2) = self._getSubcomponentByName(dp)
                        result += "\n\t\t\t// Event Port Connection: %s -> %s (and current mode %s is in modes %s)\n" % (sp, dp, mname, ims)
                        result += "\t\t\t" + scname2 + "!trymultiway" + (len(scidp2)+len(scodp2)) * ", 0" + ", " + str(self._EventToNumber(dp.split(".")[1], sciep2, scoep2)) + ";\n"
                        result += "\t\t\t" + scname2 + "?trymultiway" + (len(scidp2)+len(scodp2)) * ", _" + ", eventInt;\n"
                        result += "\t\t\teventSum = eventSum + eventInt;\n"

                    # Reactive Transitions (own in event)
                    reactiveTransitions = filter(lambda (sm, tr, gu, ef, dm): sm == mname and tr == scname + "." + outEventPort, self.transitions)
                    if len(reactiveTransitions) > 0:
                        result += "\n\t\t// Reactive Transitions\n"
                        result += "\t\tif\n";
                        for (sm, tr, gu, ef, dm) in reactiveTransitions:
                            result += "\n\t\t\t// Reactive Transition (subcomponent's out event): %s -[%s when %s then %s]-> %s\n" % (sm, tr, self.expressionToString(gu), self.expressionsToString(ef), dm)
                            result += "\t\t\t:: " + self._guardToSPIN(gu) + " ->\n"
                            result += self._resetReactivatedDataSubcomponents(sm, dm, "\t\t\t\t")
                            result += self._effectsToSPIN(ef, "\t\t\t\t")
                            result += self._resetDisconnectedDataPorts(sm, dm, "\t\t\t\t")
                            result += "\t\t\t\t// Answer to trymultiway\n"
                            result += "\t\t\t\t" + scname + "!trymultiway" + (len(scidp)+len(scodp)) * ", 0" + ", eventSum+1;\n"
                            result += "\t\t\t\t// Receive invalidate from subcomponent containing the master transition (before possible deactivating it!)\n"
                            result += "\t\t\t\t" + scname + "?invalidate" + (len(scidp)+len(scodp)+1) * ", _" + ";\n"
                            result += "\t\t\t\t// Adjust Subcomponents' Activation (now it is even safe to deactivate the subcomponent containing the master)\n"
                            result += self._adjustSubcomponentsActivation(sm, dm, "\t\t\t\t")
                            result += "\t\t\t\t// Forward invalidate\n"
                            result += "\t\t\t\tparent!invalidate" + self.nrparams * ", 0" + ";\n"
                            result += "\t\t\t\tgoto " + dm + ";\n"
                        result += "\n\t\t\t// No Reactive Transition enabled => just answer to trymultiway\n"
                        result += "\t\t\t:: else ->\n"
                        result += "\t\t\t\t" + scname + "!trymultiway" + (len(scidp)+len(scodp)) * ", 0" + ", eventSum;\n"
                        result += "\n\t\tfi;\n"
                    
                    # Answer incomming trymultiway (if no reactive transition)
                    else:
                        result += "\n\t\t// No Reactive Transitions => just answer to trymultiway\n"
                        result += "\t\t" + scname + "!trymultiway" + (len(scidp)+len(scodp)) * ", 0" + ", eventSum;\n"


            # Incomming Message: update (from parent)
            result += "\n\t// Incomming Message: update (from parent)\n"

            # - Receive own input values
            result += "\t:: parent?update"
            for (name, typ, default) in self.in_data_ports:
                result += ", store_self_" + name
            result += (len(self.out_data_ports) + 1) * ", _" + " ->\n"

            result += "\t\tprintf(\"### " + self.proctypename + "(%d) ### update with new input values: "
            for (name, typ, default) in self.in_data_ports:
                result += name + "=%d, "
            result += "#=%d\\n\", _pid, "
            for (name, typ, default) in self.in_data_ports:
                result += "store_self_" + name + ", "
            result += str(len(self.in_data_ports)) + ");\n"

            # - Update Subcomponents: First Try
            result += "\n\t\t// Update Subcomponents: First Iteration\n"
            result += "\t\t" + self.proctypename + "_internalUpdate_" + mname + "();\n"
            
            # - Update Subcomponents: Fixpoint Iteration
            doNonFixpointCheck = ""
            doNonFixpointCopy = ""
            for (scname, scimp, scidp, scodp, sciep, scoep, ims) in self.subcomponents:
                for (name, typ, default) in scodp:
                    doNonFixpointCheck += "store_" + scname + "_" + name + " == new_" + scname + "_" + name + " && "
                    doNonFixpointCopy  += "store_" + scname + "_" + name +  " = new_" + scname + "_" + name + "; "
            if doNonFixpointCheck != "":
                result += "\n\t\t// Update Subcomponents: Fixpoint Iteration\n"
                result += "\t\tdo\n"
                result += "\t\t:: !(" + doNonFixpointCheck + "true) ->\n"
                result += "\t\t\t" + doNonFixpointCopy + "\n"
                result += "\t\t\t" + self.proctypename + "_internalUpdate_" + mname + "();\n"
                result += "\t\t:: else -> break;\n"
                result += "\t\tod;\n"
            else:
                result += "\n\t\t// Update Subcomponents: No Fixpoint Iteration needed!\n"

            # - Mode dependend internal update (inline function)
            prelude += "inline " + self.proctypename + "_internalUpdate_" + mname + "()\n"
            prelude += "{\n"

            for (scname, scimp, scidp, scodp, sciep, scoep, ims) in filter(lambda (scname, scimp, scidp, scodp, sciep, scoep, ims): ims == [] or mname in ims, self.subcomponents):
                prelude += "\t// Update subcomponent " + scname + "\n"
                
                forwardingConnections = filter(lambda (sp, dp, ims): dp.split(".")[0] == scname and (ims == [] or mname in ims), self.data_port_connections) 
                for (sp, dp, ims) in forwardingConnections:
                    prelude += "\tstore_" + scname + "_" + dp.split(".")[1] + " = store_"   # this also works for reals which are stored in PROMELA ints!
                    if sp.find(".") == -1:
                        prelude += "self_" + sp
                    else:
                        prelude += sp.split(".")[0] + "_" + sp.split(".")[1]
                    prelude += ";\n"

                prelude += "\t" + scname + "!update"
                for scInDataPort in scidp:
                    prelude += ", store_" + scname + "_" + scInDataPort[0]
                prelude += (len(scodp) + 1) * ", 0" + ";\n"

                prelude += "\t" + scname + "?update" + len(scidp) * ", _"
                for scOutDataPort in scodp:
                    prelude += ", new_" + scname + "_" + scOutDataPort[0]
                prelude += ", _;\n"
            
            prelude += "\tprintf(\".\\n\");\n}\n\n"

            # - Update own output values
            result += "\n\t\t// Update own output values\n"
            
            # - a) Via Data Port Connections from subcomponents' out data ports
            result += "\t\t// a) Via Data Port Connections from subcomponents' out data ports\n"
            forwardingConnections = filter(lambda (sp, dp, ims): self._isOwnOutDataPort(dp) and (ims == [] or mname in ims), self.data_port_connections) 
            for (sp, dp, ims) in forwardingConnections:
                result += "\t\tstore_self_" + dp + " = store_" + sp.split(".")[0] + "_" + sp.split(".")[1] + ";\n"   # this also works for reals which are stored in PROMELA ints!

            # - b) Via Flows from own in data ports
            result += "\t\t// b) Via Flows from own in data ports\n"
            for (op, exp, ims) in filter(lambda (op, exp, ims): (ims == []) or (mname in ims), self.flows):
                result += "\t\t" + self._assignmentToSPIN(op, exp) + ";\n"

            # - Respond with own output values
            result += "\n\t\t// Respond with own output values\n"
            result += "\t\tprintf(\"### " + self.proctypename + "(%d) ### update finally lead to new output values: "
            for (name, typ, default) in self.out_data_ports:
                result += name + "=%d, "
            result += "#=%d\\n\", _pid, "
            for (name, typ, default) in self.out_data_ports:
                result += "store_self_" + name + ", "
            result += str(len(self.out_data_ports)) + ");\n"
            
            result += "\t\tparent!update" + len(self.in_data_ports) * ", 0"
            for (name, typ, default) in self.out_data_ports:
                result += ", store_self_" + name
            result += ", 0;\n"

                    
            # Incomming Message: invalidate (from subcomponents only)
            result += "\n\t// Incomming Message: invalidate (from subcomponents only)\n"
            for (scname, scimp, scidp, scodp, sciep, scoep, ims) in self.subcomponents:
                result += "\t:: " + scname + "?invalidate" + (len(scidp) + len(scodp) +1) * ", _" + " ->\n"
                result += "\t\tparent!invalidate" + self.nrparams * ", 0" + ";\n"

                    
            # Incomming Message: deactivate (from parent)
            if self.modehistory:
                result += "\n\t// Incomming Message: deactivate (from parent) - keep mode history\n"
                
                result += "\t:: parent?deactivate" + self.nrparams * ", _" + " ->\n"
                result += "\t\t// deactivate active subcomponents\n"
                for (scname, scimp, scidp, scodp, sciep, scoep, ims) in filter(lambda (scname, scimp, scidp, scodp, sciep, scoep, ims): (ims == []) or (mname in ims), self.subcomponents):
                    result += "\t\t" + scname + "!deactivate" + (len(scidp)+len(scodp)+1) * ", 0" + ";\n"
                    result += "\t\t" + scname + "?deactivate" + (len(scidp)+len(scodp)+1) * ", _" + ";\n"
                result += "\t\tparent!deactivate" + self.nrparams * ", 0" + ";\n"
                
                result += "\t\t// reactivate\n"
                result += "\t\tparent?activate" + self.nrparams * ", _" + ";\n"
                for (scname, scimp, scidp, scodp, sciep, scoep, ims) in filter(lambda (scname, scimp, scidp, scodp, sciep, scoep, ims): (ims == []) or (mname in ims), self.subcomponents):
                    result += "\t\t" + scname + "!activate" + (len(scidp)+len(scodp)+1) * ", 0" + ";\n"
                    result += "\t\t" + scname + "?activate" + (len(scidp)+len(scodp)+1) * ", _" + ";\n"
                result += "\t\tparent!activate" + self.nrparams * ", 0" + ";\n"
                
            else:
                result += "\n\t// Incomming Message: deactivate (from parent) - reset mode history\n"       
                result += "\t:: parent?deactivate" + self.nrparams * ", _" + " ->\n"

                result += "\t\t// reset active subcomponents\n"
                for (scname, scimp, scidp, scodp, sciep, scoep, ims) in filter(lambda (scname, scimp, scidp, scodp, sciep, scoep, ims): (ims == []) or (mname in ims), self.subcomponents):
                    result += "\t\t" + scname + "!reset" + (len(scidp)+len(scodp)+1) * ", 0" + ";\n"
                    result += "\t\t" + scname + "?reset" + (len(scidp)+len(scodp)+1) * ", _" + ";\n"

                result += "\t\t// activate already deactivated subcomponents to reset them\n"
                for (scname, scimp, scidp, scodp, sciep, scoep, ims) in filter(lambda (scname, scimp, scidp, scodp, sciep, scoep, ims): (ims != []) and (mname not in ims), self.subcomponents):
                    result += "\t\t" + scname + "!activate" + (len(scidp)+len(scodp)+1) * ", 0" + ";\n"
                    result += "\t\t" + scname + "?activate" + (len(scidp)+len(scodp)+1) * ", _" + ";\n"
                    result += "\t\t" + scname + "!reset" + (len(scidp)+len(scodp)+1) * ", 0" + ";\n"
                    result += "\t\t" + scname + "?reset" + (len(scidp)+len(scodp)+1) * ", _" + ";\n"

                result += "\t\t// confirm deactivation\n"
                result += "\t\tparent!deactivate" + self.nrparams * ", 0" + ";\n"
                result += "\t\tgoto FirstActivationOrReactivationWithoutModeHistory;\n"
                
            
            # Incomming Message: reset (from parent)   ---   nearly identical to deactivate w/o mode history
            result += "\n\t// Incomming Message: reset (from parent)\n"       
            result += "\t:: parent?reset" + self.nrparams * ", _" + " ->\n"
            
            result += "\t\t// reset active subcomponents\n"
            for (scname, scimp, scidp, scodp, sciep, scoep, ims) in filter(lambda (scname, scimp, scidp, scodp, sciep, scoep, ims): (ims == []) or (mname in ims), self.subcomponents):
                result += "\t\t" + scname + "!reset" + (len(scidp)+len(scodp)+1) * ", 0" + ";\n"
                result += "\t\t" + scname + "?reset" + (len(scidp)+len(scodp)+1) * ", _" + ";\n"

            result += "\t\t// activate already deactivated subcomponents to reset them\n"
            for (scname, scimp, scidp, scodp, sciep, scoep, ims) in filter(lambda (scname, scimp, scidp, scodp, sciep, scoep, ims): (ims != []) and (mname not in ims), self.subcomponents):
                result += "\t\t" + scname + "!activate" + (len(scidp)+len(scodp)+1) * ", 0" + ";\n"
                result += "\t\t" + scname + "?activate" + (len(scidp)+len(scodp)+1) * ", _" + ";\n"
                result += "\t\t" + scname + "!reset" + (len(scidp)+len(scodp)+1) * ", 0" + ";\n"
                result += "\t\t" + scname + "?reset" + (len(scidp)+len(scodp)+1) * ", _" + ";\n"

            result += "\t\t// confirm reset\n"
            result += "\t\tparent!reset" + self.nrparams * ", 0" + ";\n"
            result += "\t\tgoto FirstActivationOrReactivationWithoutModeHistory;\n"


            # End of mode's do-loop
            result += "\n\tod; // --- " + mname + " --- \n"

        # </for (mname, mtype) in self.modes: ...>



        # *** Return completed proctype definition ***
        result += "\n}\n"
        result = prelude + result
        return result
    # </def to_SPIN()>:

    
# ==================================================================================================
# </class ProctypeInfo: ...>
# ==================================================================================================










# ==================================================================================================
# class Slim2Spin_Translator
# ==================================================================================================

class Slim2Spin_Translator:
    """This class contains the framework for Translation SLIM2SPIN.

    Usage: <Example>"""

    # ----------------------------------------------------------------------------------------------



    def __init__(self):
        self.theFileName = None
        self.theCompileInfo = None
        self.enumNames = {}
        self.root = None



    # ----------------------------------------------------------------------------------------------



    def parseFile(self, fileName):
        """Parses the file given by path/name and returns the number of parsing errors."""
        self.theFileName = fileName
        self.theCompileInfo = slim_compiler.src.dataStructure.compile.CompileInformation()
        invokers.slim.parse(self.theCompileInfo, [fileName])
        return len(self.theCompileInfo.error_collector.errors)

    
    def printParserErrors(self):
        print " "
        if len(self.theCompileInfo.error_collector.errors) > 0:
            print "The   P A R S I N G   E R R O R S   are:"
            print "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"
            print self.theCompileInfo.error_collector
            print "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"
        else:
            print "No parsing errors."
        print " "


    def printCompileInfo(self):
        print " "
        print "The Compile Info is:"
        print "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"
        print toSLIMVisitor().visit(self.theCompileInfo)
        print "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~"
        print " "


    def checkCompileInfo(self):
        slim.check(self.theCompileInfo)


    # ----------------------------------------------------------------------------------------------



    def _addEnumNames(self, aList):
        for enumName in aList:
            if enumName not in self.enumNames:
                self.enumNames.update({enumName: len(self.enumNames)+1})
    

    def _translateAllImplementations(self):
        result = ""
        
        # Walk through packages, collect type & implementation info, invoke translate for each implementation
        for packagename in self.theCompileInfo.packages.keys():

            #print "Translate package: %s" % (packagename)
            packageobject = self.theCompileInfo.packages[packagename]

            for subpackageobject in [packageobject.public_subpackage, packageobject.private_subpackage]: # for translation, public/private does not matter!

                for implementationname in subpackageobject.implementations.keys():
                    compname = string.split(implementationname, ".")[0]
                    implname = string.split(implementationname, ".")[1]

                    #print "\n\n\n\n\n\tTranslate implementation: %s::%s.%s" % (packagename, compname, implname)

                    pi = ProctypeInfo()
                    pi.packagename = packagename
                    pi.componentname = compname
                    pi.implementationname = implname
                    pi.translator = self

                    typeobject = subpackageobject.types[compname]
                    implementationobject = subpackageobject.implementations[implementationname]
                    #implementationobject.check()
                    
                    #print "\t\tIn Data Ports:"
                    for portname in typeobject.in_data_ports.keys():
                        portobject = typeobject.in_data_ports[portname]
                        #print "\t\t\t%s/%s : %s = %s" % (portname, portobject.name, portobject.data_type.get_name(), portobject.init_value)
                        pi.in_data_ports.append((portname, portobject.data_type, portobject.init_value))
                        if isinstance(portobject.data_type, Enum):
                            #print "\t\t\t\t=> EnumNames: %s" % portobject.data_type.values
                            self._addEnumNames(portobject.data_type.values)

                    #print "\t\tOut Data Ports:"
                    for portname in typeobject.out_data_ports.keys():
                        portobject = typeobject.out_data_ports[portname]
                        #print "\t\t\t%s/%s : %s = %s" % (portname, portobject.name, portobject.data_type.get_name(), portobject.init_value)
                        pi.out_data_ports.append((portname, portobject.data_type, portobject.init_value))
                        if isinstance(portobject.data_type, Enum):
                            #print "\t\t\t\t=> EnumNames: %s" % portobject.data_type.values
                            self._addEnumNames(portobject.data_type.values)

                    #print "\t\tIn Event Ports:"
                    for portname in typeobject.in_event_ports.keys():
                        #print "\t\t\t%s" % (portname)
                        pi.in_event_ports.append(portname)

                    #print "\t\tOut Event Ports:"
                    for portname in typeobject.out_event_ports.keys():
                        #print "\t\t\t%s" % (portname)
                        pi.out_event_ports.append(portname)

                    #print "\t\tNon-Data Subcomponents:"
                    for subcomponentname in implementationobject.non_data_subcomponents.keys():
                        subcomponentobject = implementationobject.non_data_subcomponents[subcomponentname]
                        subcomponenttypeobject = subcomponentobject.get_referenced_component_type()
                        subcomponentimplementationobject = subcomponentobject.get_referenced_component_implementation()            # Attention: check whether this really exists (not done by parser)
                        #subcomponentimplementationobject.check()
                        #print "\t\t\t%s/%s : %s/%s/%s from %s (used in %s)" % (subcomponentname, subcomponentobject.name, subcomponentobject.implementation_ref, subcomponenttypeobject.name, subcomponentimplementationobject.name, subcomponentimplementationobject.subpackage.package_name, subcomponentobject.implementation.name)

                        #print "\t\t\t\t|- In Data Ports:"
                        scidp = []
                        for portname in subcomponenttypeobject.in_data_ports.keys():
                            portobject = subcomponenttypeobject.in_data_ports[portname]
                            #print "\t\t\t\t\t%s/%s : %s = %s" % (portname, portobject.name, portobject.data_type.get_name(), portobject.init_value)
                            scidp.append((portname, portobject.data_type, portobject.init_value))
                            if isinstance(portobject.data_type, Enum):
                                #print "\t\t\t\t\t\t=> EnumNames: %s" % portobject.data_type.values
                                self._addEnumNames(portobject.data_type.values)

                        #print "\t\t\t\t|- Out Data Ports:"
                        scodp = []
                        for portname in subcomponenttypeobject.out_data_ports.keys():
                            portobject = subcomponenttypeobject.out_data_ports[portname]
                            #print "\t\t\t\t\t%s/%s : %s = %s" % (portname, portobject.name, portobject.data_type.get_name(), portobject.init_value)
                            scodp.append((portname, portobject.data_type, portobject.init_value))
                            if isinstance(portobject.data_type, Enum):
                                #print "\t\t\t\t\t\t=> EnumNames: %s" % portobject.data_type.values
                                self._addEnumNames(portobject.data_type.values)

                        #print "\t\t\t\t|- In Event Ports:"
                        sciep = []
                        for portname in subcomponenttypeobject.in_event_ports.keys():
                            #print "\t\t\t\t\t%s" % (portname)
                            sciep.append(portname)

                        #print "\t\t\t\t|- Out Event Ports:"
                        scoep = []
                        for portname in subcomponenttypeobject.out_event_ports.keys():
                            #print "\t\t\t\t\t%s" % (portname)
                            scoep.append(portname)

                        pi.subcomponents.append((subcomponentname, subcomponentimplementationobject.subpackage.package_name + "::" + subcomponentimplementationobject.name, scidp, scodp, sciep, scoep, subcomponentobject.in_modes))

                    #print "\t\tData Subcomponents:"
                    for subcomponentname in implementationobject.data_subcomponents.keys():
                        subcomponentobject = implementationobject.data_subcomponents[subcomponentname]
                        #print "\t\t\t%s/%s : %s = %s" % (subcomponentname, subcomponentobject.name, subcomponentobject.data_type.get_name(), subcomponentobject.init_value)
                        pi.datasubcomponents.append((subcomponentname, subcomponentobject.data_type, subcomponentobject.init_value, subcomponentobject.in_modes))
                        if isinstance(subcomponentobject.data_type, Enum):
                            #print "\t\t\t\t=> EnumNames: %s" % subcomponentobject.data_type.values
                            self._addEnumNames(subcomponentobject.data_type.values)

                    #print "\t\tData Port Connections:"
                    for connection in implementationobject.data_port_connections:
                        #print "\t\t\t%s -> %s, in modes %s" % (connection.source_port, connection.destination_port, connection.in_modes)
                        pi.data_port_connections.append((connection.source_port, connection.destination_port, connection.in_modes))

                    #print "\t\tEvent Port Connections:"
                    for connection in implementationobject.event_port_connections:
                        #print "\t\t\t%s -> %s, in modes %s" % (connection.source_port, connection.destination_port, connection.in_modes)
                        pi.event_port_connections.append((connection.source_port, connection.destination_port, connection.in_modes))

                    #print "\t\tFlows:"
                    for flow in implementationobject.flows:
                        #print "\t\t\t%s <-flow-- %s   (in modes: %s)" % (flow.port_name, flow.expression, flow.in_modes)
                        pi.flows.append((flow.port_name, flow.expression, flow.in_modes))

                    #print "\t\tModes:"
                    for modename in implementationobject.modes.keys():
                        modeobject = implementationobject.modes[modename]
                        #print "\t\t\t%s/%s (%s)" % (modename, modeobject.name, modeobject.type)
                        pi.modes.append((modename, modeobject.type))
                    
                    #print "\t\tNormal Mode Transitions:"
                    for transition in implementationobject.normal_mode_transitions:
                        #print "\t\t\t%s -[%s when %s then %s]-> %s" % (transition.source_mode, transition.label.trigger, transition.label.guard, transition.label.effects, transition.destination_mode)
                        #for effect in transition.label.effects:
                        #    print "\t\t\t\t Effect as Expression: " + expressionToString(effect) + "\n" 
                        pi.transitions.append((transition.source_mode, transition.label.trigger, transition.label.guard, transition.label.effects, transition.destination_mode))

                    #print "\t\tProbability Mode Transitions:"
                    for transition in implementationobject.probability_mode_transitions:
                        #print "\t\t\t%s -[prob %s]-> %s" % (transition.source_mode, transition.label.probabilty, transition.destination_mode)
                        pi.probabilistictransitions.append((transition.source_mode, transition.label.probabilty, transition.destination_mode))

                    #print "\t===> To SPIN:"
                    #print "\n\n\n\n" + pi.to_SPIN()
                    result += "\n\n\n\n\n" + pi.to_SPIN()

        return result


    def _translateHeader(self):
        """Important: to be called after (!) _translateAllImplementations(), since that collects some required info about enums."""
        
        # Message Types & Macros
        result = """// ----- Message Types ----------------------------------------------------------------------------

mtype = { update, invalidate, trymultiway, activate, deactivate, reset };



// ----- MACROs for ensuring atomicity of transitions ---------------------------------------------

bool gflagGoNextTransition = false;

#define nT(guard) d_step { \\
\tguard && !gflagGoNextTransition; \\
\tgflagGoNextTransition = true; \\
\tprintf("### (%d) ### Process %d is allowed to take its next transition!\\n", _pid, _pid) }

#define fT d_step { \\
\tgflagGoNextTransition = false; \\
\tprintf("### (%d) ### Process %d returned right to take next transition!\\n", _pid, _pid) }



// ----- DEFINE integer constants for enum data types ---------------------------------------------

"""
        
        # Integer Constants for ENUMs
        for enumName in self.enumNames.keys():
            result += "#define enum_%s %s\n" % (enumName, self.enumNames[enumName])

        # Return result
        return result
        

    def _translateEnvironment(self):
        
        # *** Head ***
        txt = """// ----- Environment ------------------------------------------------------------------------------

active proctype Environment()
{
\tprintf("### ENV(%d) ### Environment started.\\n", _pid);
"""

        # *** Find root implementation ***
        self.root = invokers.slim.get_root_impl(self.theCompileInfo)

        if self.root == None:
            txt += "\tERR: No root implementation found!\n}\n"
            return txt

        rootImplementation = self.root
        rootType = rootImplementation.get_referenced_component_type()
        #txt += "\tROOT: %s from package %s\n" % (self.root.name, self.root.subpackage.package_name)

        # *** Extract infos from Root Implementation ***
        rootsInDataPorts = [(portName, rootType.in_data_ports[portName].data_type, rootType.in_data_ports[portName].init_value) for portName in rootType.in_data_ports.keys()]
        rootsInDataPorts.sort()
        #txt += "\t- In Data Ports: %s \n" % rootType.in_data_ports.keys()
        #txt += "\t- In Data Ports: %s \n" % rootsInDataPorts
        rootsOutDataPorts = [(portName, rootType.out_data_ports[portName].data_type, rootType.out_data_ports[portName].init_value) for portName in rootType.out_data_ports.keys()]
        rootsOutDataPorts.sort()
        #txt += "\t- Out Data Ports: %s \n" % rootType.out_data_ports.keys()
        #txt += "\t- Out Data Ports: %s \n" % rootsOutDataPorts
        rootsNrParams = len(rootsInDataPorts) + len(rootsOutDataPorts) + 1
                            
        # *** Variables for storing Root Component ***
        txt += "\n\t// *** Variables for storing Root Component ***\n"
        
        txt += "\tpid pid_root;\n"
        txt += "\tchan chan_root = [0] of {mtype"
        for (name, typ, default) in rootsInDataPorts:
            txt += ", " + ProctypeInfo._datatypeToSPIN(typ)
        for (name, typ, default) in rootsOutDataPorts:
            txt += ", " + ProctypeInfo._datatypeToSPIN(typ)
        txt += ", int};\n\n"    

        for (name, typ, default) in rootsInDataPorts:
            txt += "\t" + ProctypeInfo._datatypeToSPIN(typ) + " root_" + name + "_store;\n"
        for (name, typ, default) in rootsOutDataPorts:
            txt += "\t" + ProctypeInfo._datatypeToSPIN(typ) + " root_" + name + "_store;\n"

        # *** Start & Activate Root Component ***
        txt += "\n\t// *** Start & Activate Root Component ***\n"
        txt += "\tnT(true);\n"

        txt += "\tpid_root = run " + ProctypeInfo._implementationNameToSPIN(self.root.subpackage.package_name + "::" + self.root.name) + "(chan_root);\n"

        txt += "\tchan_root!activate" + rootsNrParams * ", 0" + ";\n"
        txt += "\tchan_root?activate" + rootsNrParams * ", _" + ";\n"
        
        txt += "\tchan_root!update"
        for (name, typ, default) in rootsInDataPorts:
            txt += ", root_" + name + "_store"
        txt += (len(rootsOutDataPorts)+1) * ", 0" + ";\n"
        txt += "\tchan_root?update" + len(rootsInDataPorts) * ", _"
        for (name, typ, default) in rootsOutDataPorts:
            txt += ", root_" + name + "_store"
        txt += ", _;\n"

        txt += "\tfT;\n"
        
        # // *** React to Messages from Root ***
        txt += "\n\t// *** React to Messages from Root ***\n"
        txt += "\tdo\n"

        txt += "\t// Invalidate: react with update\n"
        txt += "\t:: chan_root?invalidate" + rootsNrParams * ", _" + " ->\n"
        txt += "\t\tchan_root!update"
        for (name, typ, default) in rootsInDataPorts:
            txt += ", root_" + name + "_store"
        txt += (len(rootsOutDataPorts)+1) * ", 0" + ";\n"
        txt += "\t\tchan_root?update" + len(rootsInDataPorts) * ", _"
        for (name, typ, default) in rootsOutDataPorts:
            txt += ", root_" + name + "_store"
        txt += ", _;\n"
        txt += "\t\tfT;  // root responsible for fT after invalidate-update-cycle!\n"

        txt += "\t// Trymultiway: respond 0 partners\n"
        txt += "\t:: chan_root?trymultiway" + rootsNrParams * ", _" + " ->\n"
        txt += "\t\tchan_root!trymultiway" + rootsNrParams * ", 0" + ";\n"

        txt += "\tod;\n"

        # *** Return result ***
        txt += "}\n"
        return txt


    def to_SPIN(self):
        # First, translate the implementations (which walks through the CompileInfo and collects all the info)
        implementations = self._translateAllImplementations()
        #implementations = "\t< i m p l e m e n t a t i o n s > \n" 

        # Header
        result = "// This file has been generated by slim2promela, based on SLIM-Input: " + self.theFileName + "\n\n\n\n"
        result += self._translateHeader()
        result += "\n\n\n\n\n\n// ================================================================================================\n\n"

        # Implementations
        result += implementations

        # Environment
        result += "\n\n\n\n\n\n// ================================================================================================\n\n\n\n\n\n\n"
        result += self._translateEnvironment()

        # Return result
        return result  



# ==================================================================================================
# </class Slim2Spin_Translator: ...>
# ==================================================================================================










# ==================================================================================================
# Main Program
# ==================================================================================================

# Print Start-of-Program Message
print "START OF SLIM2PROMELA\n\n\n"


# Create Translator Instance
myTranslator = Slim2Spin_Translator()


# Parse SLIM File
#myTranslator.parseFile("slim2promela.slim")
#myTranslator.parseFile("HalfAdder.slim")
#myTranslator.parseFile("FullAdder.slim")
#myTranslator.parseFile("examples\\RedundantBatterySystem2.slim")
#myTranslator.parseFile("examples\\TransitionChangesDataPort.slim")
#myTranslator.parseFile("examples\\Adder.slim"
myTranslator.parseFile("examples\\Negator.slim")

myTranslator.printParserErrors()
#myTranslator.printCompileInfo()

#myTranslator.checkCompileInfo()
#myTranslator.printParserErrors()


# Translate to SPIN
print myTranslator.to_SPIN()


# Print End-of-Program Message
print "\n\nEND OF SLIM2PROMELA\n"

# ==================================================================================================
# </Main Program>
# ==================================================================================================
