# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


"""This module contains the functionality of Slicing SLIM Models wrt a set of atomic propositions.

Use the convenience function doSlicing to obtain a sliced compile info from an existing one.

The algorithm has been designed and implemented by Maximilian R. Odenbrett as a
part of his diploma thesis, supervised by Thomas Noll, Viet Yen Nguyen, J.-P. Katoen.

RWTH Aachen University
"""





# ==================================================================================================
# Imports: adapt path, import reused parts of SPIN2Promela-Translation
# ==================================================================================================

import sys
import os

if os.name == "nt":
    sys.path.append(".\\..\\..\\..") # 3 levels up to tools directory (WIN)
else: 
    sys.path.append("./../../../")   # 3 levels up to tools directory (UNIX)

from slim_compiler.src.promela.shared import *
from slim_compiler.src.promela.renaming import *
from slim_compiler.src.dataStructure.visitors.printVisitor import toSLIMVisitor

# ==================================================================================================
# /Imports
# ==================================================================================================





# ==================================================================================================
# Classes to store an explicit component tree (ComponentTreeNode, Transition, PortConnection, etc.)
# ==================================================================================================

class ComponentTreeNode:

    # Constants for node types
    UNKNOWN          = 0
    NONDATACOMPONENT = 1
    DATACOMPONENT    = 2
    INDATAPORT       = 3
    OUTDATAPORT      = 4
    INEVENTPORT      = 5
    OUTEVENTPORT     = 6
    MODE             = 7

    # Constants for categories of node types
    NONDATASC        = 11 # = Non Data Subcomponent
    DATAELEMENT      = 12 # = In Data Port, Out Data Port, Data Subcomponent
    EVENTPORT        = 13 # = In Event Port, Out Event Port
    # MODE           = 7  # = Nominal Mode (Hint: constant from element type reused)

    # ~ ~ ~ ~ ~ ~ ~ ~ ~ ~

    def __init__(self, name, typ, accessPath, inCompileInfo, parent, inModes = None):
        # Basic infos
        self.name               = name
        self.typ                = typ
        self.accessPath         = accessPath
        self.inCompileInfo      = inCompileInfo
        self.parent             = parent
        self.subnodes           = {}

        # Additional info for Non Data Subcomponents, Data Subcomponents, Data Port Connections, Flows, Event Port Connections
        self.inModes            = inModes

        # Additional info for Modes
        self.invariantReadsFrom = []
        self.invariantWritesTo  = []

        # Flags for interesting/weak interesting
        self.isInteresting     = False
        self.isWeakInteresting = False

    # ~ ~ ~ ~ ~ ~ ~ ~ ~ ~

    def getSubnodesOfType(self, typ):
        return filter(lambda subnode: subnode.typ == typ, self.subnodes.values())
    
    def getModes(self):
        return self.getSubnodesOfType(ComponentTreeNode.MODE)

    def getNonDataSubcomponents(self):
        return self.getSubnodesOfType(ComponentTreeNode.NONDATACOMPONENT)

    # ~ ~ ~ ~ ~ ~ ~ ~ ~ ~

    def getSubnodeByNameAndCategory(self, name, category):
        """Note: Name can be an accessPath from this node to nested subnodes."""

        # If no name given, abort
        if name == None or name == "" or name == False:
            return None

        # If "." found in name, follow access path to next non data subcomponent
        elif name.find(".") != -1:
            (scName, tail) = name.split(".", 1)
            if (scName, ComponentTreeNode.NONDATASC) in self.subnodes:
                subNode = self.subnodes[(scName, ComponentTreeNode.NONDATASC)]
                return subNode.getSubnodeByNameAndCategory(tail, category)
            else:
                return None

        # Reached end of access path: return requested subnode
        else:
            return self.subnodes[(name, category)]

    # ~ ~ ~ ~ ~ ~ ~ ~ ~ ~

    def listContainsAllModes(self, modeList):
        return set(self.getModes()) == set(modeList)

    # ~ ~ ~ ~ ~ ~ ~ ~ ~ ~

    def hasModeHistory(self):
        if self.typ == ComponentTreeNode.NONDATACOMPONENT:
            for modeNode in self.getModes():
                if modeNode.inCompileInfo.type == ModeType.INITIAL:
                    return True
                elif modeNode.inCompileInfo.type == ModeType.ACTIVATION:
                    return False
        return False

    # ~ ~ ~ ~ ~ ~ ~ ~ ~ ~

    def hasInterestingResetableElementBelow(self, interestingDataElements, interestingModes):
        if self.typ == ComponentTreeNode.NONDATACOMPONENT:
            return checkExistence(lambda subnode: subnode.hasInterestingResetableElementBelow(interestingDataElements, interestingModes), self.subnodes.values())
        elif self.typ in [ComponentTreeNode.DATACOMPONENT, ComponentTreeNode.INDATAPORT, ComponentTreeNode.OUTDATAPORT]:
            return (self in interestingDataElements)
        elif self.typ in [ComponentTreeNode.MODE]:
            return (self in interestingModes)
        else:
            return False

    def hasInterestingResetableElementWithoutModeHistoryBelow(self, interestingDataElements, interestingModes):
        if self.typ == ComponentTreeNode.NONDATACOMPONENT:
            if not self.hasModeHistory():
                return self.hasInterestingResetableElementBelow(interestingDataElements, interestingModes)
            else:
                return checkExistence(lambda subnode: subnode.hasInterestingResetableElementWithoutModeHistoryBelow(interestingDataElements, interestingModes), self.subnodes.values())
        else:
            return False

    # ~ ~ ~ ~ ~ ~ ~ ~ ~ ~

    def bottomUpPropagateBeingInteresting(self):
        for subnode in self.subnodes.values():
            subnode.bottomUpPropagateBeingInteresting()

        self.isInteresting = checkExistence(lambda subnode: subnode.isInteresting, self.subnodes.values()) or  self.isInteresting
    
    # ~ ~ ~ ~ ~ ~ ~ ~ ~ ~

    def _typeToString(self, typ):
        if typ == ComponentTreeNode.UNKNOWN:
            return "Unknown"
        elif typ == ComponentTreeNode.NONDATACOMPONENT:
            return "Non Data Component"
        elif typ == ComponentTreeNode.DATACOMPONENT:
            return "Data Component"
        elif typ == ComponentTreeNode.INDATAPORT:
            return "In Data Port"
        elif typ == ComponentTreeNode.OUTDATAPORT:
            return "Out Data Port"
        elif typ == ComponentTreeNode.INEVENTPORT:
            return "In Event Port"
        elif typ == ComponentTreeNode.OUTEVENTPORT:
            return "Out Event Port"
        elif typ == ComponentTreeNode.MODE:
            return "Mode"
        else:
            return "(Undefined Type)"

    def toString(self, indent = 0):
        s = indent * "\t" + "- " + self._typeToString(self.typ) + ": " + self.name + " (" + self.accessPath + ")"
        if self.isInteresting:
            s += " ***!!!***"
        elif self.isWeakInteresting:
            s += " (((!!!)))"
        for subnode in self.subnodes.values():
            s += "\n" + subnode.toString(indent + 1)
        return s

# </class ComponentTreeNode: ...>



class Transition:
    
    def __init__(self, sourceMode, targetMode, trigger, guardReads, effects, reactivates, definingComponent, transition):
        self.sourceMode        = sourceMode
        self.targetMode        = targetMode
        self.trigger           = trigger
        self.guardReads        = guardReads
        self.effects           = effects
        self.reactivates       = reactivates
        self.definingComponent = definingComponent
        self.inCompileInfo     = transition
        self.isInteresting     = False
        
    def toString(self):
        s = self.sourceMode.accessPath + " -["
        if self.trigger != None:
            s += self.trigger.accessPath
        else:
            s += "<no trigger>"
        s += " when ("
        for read in self.guardReads:
            s += read.accessPath
        s += ") then ("
        for effect in self.effects:
            s += "(" + effect.writesTo.accessPath + "="
            for read in effect.readsFrom:
                s += read.accessPath + ","
            s += "<const>)"
        s += ") reactivates ("
        for reactivate in self.reactivates:
            s += reactivate.accessPath
        s += ")]-> " + self.targetMode.accessPath
        if self.isInteresting:
            s += " ***!!!***"
        return s



class TransitionEffect:
    
    def __init__(self, writesTo, readsFrom, effect):
        self.writesTo      = writesTo
        self.readsFrom     = readsFrom
        self.inCompileInfo = effect
        self.isInteresting = False



class PortConnection:
    
    def __init__(self, sourcePort, targetPort, inModes, definingComponent, connection):
        self.sourcePort        = sourcePort
        self.targetPort        = targetPort
        self.inModes           = inModes
        self.definingComponent = definingComponent
        self.inCompileInfo     = connection
        self.isNeeded          = False

    def toString(self):
        s = self.sourcePort.accessPath + " ---> " + self.targetPort.accessPath
        if self.isNeeded:
            s += " -Needed!-" 
        return s



class Flow:

    def __init__(self, sourcePorts, targetPort, inModes, definingComponent, flow):
        self.sourcePorts       = sourcePorts
        self.targetPort        = targetPort
        self.inModes           = inModes
        self.definingComponent = definingComponent
        self.inCompileInfo     = flow
        self.isNeeded          = False

    def toString(self):
        s = ""
        for sourcePort in self.sourcePorts:
            s += sourcePort.accessPath + ", "
        s += " ~~~> " + self.targetPort.accessPath
        if self.isNeeded:
            s += " -Needed!-"
        return s
        
# ==================================================================================================
# /Classes to store an explicit component tree (ComponentTreeNode, Transition, PortConnection, etc.)
# ==================================================================================================





# ==================================================================================================
# class SlimSlicer
# ==================================================================================================

class SlimSlicer:
    """This class stores the data and implements the algorithm used for slicing.

    Usage: 1.) Create SlimSlicer instance:
               >>> mySlimSlicer = SlimSlicer(verbose)

           2.) Generate explicit representation of component (instance) tree
               which is implicitly given by the specification and the root component:
               >>> mySlimSlicer.generateComponentTree(originalRootComponent)

           3.) Perform slicing based on a list of atomic propositions:
               >>> mySlimSlicer.performSlicing(atomicPropositions, enforceKeepingCycles)

           4.) Retrieve the sliced compile info and its root component:
               >>> (slicedCompileInfo, slicedRootComponent) = mySlimSlicer.createSlicedCompileInfo()
    """


    def __init__(self, verbose = False):
        """Initializes the data fields of a new SlimSlicer instance.

        @param    verbose:
                    Iff true, verbose output is printed to screen
                    (using shared.printOutputBlock_...).
        @type     verbose:
                    boolean, default False
        """

        # Store parameters
        self.verbose = verbose

        # Initialize storage for explicit representation of component (instance) tree
        self.rootComponentTreeNode = None       # Root of component (instance) tree: ComponentTreeNode
        self.modes = []                         # Global list of all modes: list(ComponentTreeNode with .typ = ComponentTreeNode.MODE)
        self.transitions = []                   # Global list of all transitions: list(Transition)
        self.dataPortConnections = []           # Global list of all data port connections: list(PortConnection)
        self.flows = []                         # Global list of all data flows: list(Flow)
        self.eventPortConnections = []          # Global list of all event port connections: list(PortConnection)

        # Initialize lists collecting elements found to be (weak) interesting during slicing
        self.interestingDataElements = []       # List of interesting data elements: list(ComponentTreeNode)
        self.weakInterestingDataElements = []   # List of weak interesting data elements: list(ComponentTreeNode)
        self.interestingEvents = []             # List of interesting events: list(ComponentTreeNode)
        self.interestingModes = []              # List of interesting modes: list(ComponentTreeNode)

        # Initialize storage for sliced compile info
        self.slicedCompileInfo = None
        self.slicedRootComponent = None


    # ----------------------------------------------------------------------------------------------

      
    def generateComponentTree(self, componentImplementation, name = "root", accessPath = "root", parent = None, inModes = None):
        """Recursively generates the explicit representation of the component
        (instance) tree which is implicitly given by the SLIM specification/
        compileInfo and a (root) component implementation.

        The SLIM specification/compileInfo contains every component type and
        implementation only once, independent from how often a component of that
        type/implementation is declared.
        However, for slicing it is important to consider component instances and
        not just their type and implementation since it might be that for some
        instances a lot can be sliced away while for other instances of the same
        type/implementation nothing can be sliced away.
        Thus, we make the tree of nested component instances explicit before
        slicing.

        To generate the component (instance) tree for a rootComponent, call:
        >>> aSlimSlicer.generateComponentTree(rootComponent)
        The root of the tree will be stored in aSlimcSlicer.rootComponentTreeNode.

        @precondition: - In componentImplementation and all its subcomponents the
                         underscores must already have been eliminated (since
                         they are used to transform the access path into the new
                         component type name).
                       - In every occuring componentImplementation empty in modes
                         lists (which implicitly contain every mode) must have been
                         filled up so that they now explicitly contain every mode.
        @param    componentImplementation:
                    The component (implementation) for which the component tree
                    is to be generated.
        @type     componentImplementation:
                    nominalbehaviour.ComponentImplementation
        @param    name:
                    Name of the component as a subcomponent in its supercomponent.
        @type     name:
                    string, default "root"
        @param    accessPath:
                    Access path from root to this component.
        @type     accessPath:
                    string, default "root"
        @param    parent:
                    Reference to node of the parent component (instance) in component (instance) tree.
        @type     parent:
                    ComponentTreeNode, default None
        @param    inModes:
                    List of modes of the supercomponent in which this subcomponent is active.
        @type     inModes:
                    list(ComponentTreeNode), default None
        @return:  Node representing the given component in the component (instance) tree.
        @rtype:   ComponentTreeNode
        """

        # *** Find corresponding componentType ***
        componentType = componentImplementation.get_referenced_component_type()


        # *** Create tree node for this component instance ***
        thisTreeNode = ComponentTreeNode(name,
                                         ComponentTreeNode.NONDATACOMPONENT,
                                         accessPath,
                                         componentImplementation,
                                         parent,
                                         inModes)


        # *** Create subnodes ***
                    
        # - Modes   (HINT: Do them first so that they are available for in modes lists.)
        for modeObject in componentImplementation.modes.values():
            if modeObject.type == ModeType.ERROR:
                modeTypeInfix = ".error:"
            else:
                modeTypeInfix = ".mode:"
            # Generate node representing mode
            modeNode = ComponentTreeNode(modeObject.name,
                                         ComponentTreeNode.MODE,
                                         accessPath + modeTypeInfix + modeObject.name,
                                         modeObject,
                                         thisTreeNode)
            # HINT: Invariant cannot be handeled here as data elements have not yet been created. They can't be done first as data subcomponents need the modes be created for the in_modes list.
            # Add node to tree
            thisTreeNode.subnodes.update({ (modeObject.name, ComponentTreeNode.MODE): modeNode })
            # Additionally, add node to global list
            self.modes.append(modeNode)
        
        # - Non data subcomponents   =>   recursive call
        for scObject in componentImplementation.non_data_subcomponents.values():
            scImpl = scObject.get_referenced_component_implementation()
            thisTreeNode.subnodes.update( { (scObject.name, ComponentTreeNode.NONDATASC):
                                            self.generateComponentTree(scImpl, scObject.name, accessPath + "." + scObject.name, thisTreeNode,
                                                                       [thisTreeNode.getSubnodeByNameAndCategory(modeName, ComponentTreeNode.MODE) for modeName in scObject.in_modes])
                                          } )
            
        # - Data subcomponents
        for scObject in componentImplementation.data_subcomponents.values():
            thisTreeNode.subnodes.update( { (scObject.name, ComponentTreeNode.DATAELEMENT):
                                            ComponentTreeNode(scObject.name,
                                                              ComponentTreeNode.DATACOMPONENT,
                                                              accessPath + "." + scObject.name,
                                                              scObject,
                                                              thisTreeNode,
                                                              [thisTreeNode.getSubnodeByNameAndCategory(modeName, ComponentTreeNode.MODE) for modeName in scObject.in_modes])
                                          } )
            
        # - In data ports
        for portObject in componentType.in_data_ports.values():
            thisTreeNode.subnodes.update( { (portObject.name, ComponentTreeNode.DATAELEMENT):
                                            ComponentTreeNode(portObject.name,
                                                              ComponentTreeNode.INDATAPORT,
                                                              accessPath + "." + portObject.name,
                                                              portObject,
                                                              thisTreeNode)
                                          } )
            
        # - Out data ports
        for portObject in componentType.out_data_ports.values():
            thisTreeNode.subnodes.update( { (portObject.name, ComponentTreeNode.DATAELEMENT):
                                            ComponentTreeNode(portObject.name,
                                                              ComponentTreeNode.OUTDATAPORT,
                                                              accessPath + "." + portObject.name,
                                                              portObject,
                                                              thisTreeNode)
                                          } )

        # - EXTENSION: Mode Invariants
        # HINT: Contained in mode-subnodes. The invariants couldn't be analyzed when the mode-subnodes were created as to this time the nodes for the data elements were not yet created.
        for modeNode in thisTreeNode.getModes():
            modeObject = modeNode.inCompileInfo
            modeNode.invariantReadsFrom = [thisTreeNode.getSubnodeByNameAndCategory(read,  ComponentTreeNode.DATAELEMENT) for read  in expressionReadsFrom(modeObject.invariant)]
            modeNode.invariantWritesTo  = [thisTreeNode.getSubnodeByNameAndCategory(write, ComponentTreeNode.DATAELEMENT) for write in trajectoryEquationsWriteTo(modeObject.invariant)]
            
        # - In event ports
        for portObject in componentType.in_event_ports.values():
            thisTreeNode.subnodes.update( { (portObject.name, ComponentTreeNode.EVENTPORT):
                                            ComponentTreeNode(portObject.name,
                                                              ComponentTreeNode.INEVENTPORT,
                                                              accessPath + "." + portObject.name,
                                                              portObject,
                                                              thisTreeNode)
                                          } )
            
        # - Out event ports
        for portObject in componentType.out_event_ports.values():
            thisTreeNode.subnodes.update( { (portObject.name, ComponentTreeNode.EVENTPORT):
                                            ComponentTreeNode(portObject.name,
                                                              ComponentTreeNode.OUTEVENTPORT,
                                                              accessPath + "." + portObject.name,
                                                              portObject,
                                                              thisTreeNode)
                                          } )


        # *** Establish relations ***
        # HINT: As we need to reason about all relations from a global
        #       perspective (e.g., find all transitions that influence an
        #       interesting data element) they are not stored nested in the
        #       tree but in global lists.
        #       The connection to the ComponentTreeNode of the component in
        #       which the relation is defined is established by the
        #       definingComponent-link.
        # HINT: We do not have to distinguish between own ports and those of a
        #       non data subcomponent as getSubnodeByNameAndCategory resolves the
        #       names of subcomponents' ports (sc.port returns port of sc).
        #       However, it is important that the subcomponent and its data
        #       elements already have been added to the tree.
        
        # - Data Port Connections
        for connection in componentImplementation.data_port_connections:
            sourcePort = thisTreeNode.getSubnodeByNameAndCategory(connection.source_port, ComponentTreeNode.DATAELEMENT)
            targetPort = thisTreeNode.getSubnodeByNameAndCategory(connection.target_port, ComponentTreeNode.DATAELEMENT)
            inModes = [thisTreeNode.getSubnodeByNameAndCategory(modeName, ComponentTreeNode.MODE) for modeName in connection.in_modes]
            self.dataPortConnections.append(PortConnection(sourcePort, targetPort, inModes, thisTreeNode, connection))

        # - Flows
        for flow in componentImplementation.flows:
            sourcePorts = [thisTreeNode.getSubnodeByNameAndCategory(read, ComponentTreeNode.DATAELEMENT) for read in expressionReadsFrom(flow.expression)]
            targetPort = thisTreeNode.getSubnodeByNameAndCategory(expressionWritesTo(flow.expression), ComponentTreeNode.DATAELEMENT)
            inModes = [thisTreeNode.getSubnodeByNameAndCategory(modeName, ComponentTreeNode.MODE) for modeName in flow.in_modes]
            self.flows.append(Flow(sourcePorts, targetPort, inModes, thisTreeNode, flow))

        # - Event Port Connections
        for connection in componentImplementation.event_port_connections:
            sourcePort = thisTreeNode.getSubnodeByNameAndCategory(connection.source_port, ComponentTreeNode.EVENTPORT)
            targetPort = thisTreeNode.getSubnodeByNameAndCategory(connection.target_port, ComponentTreeNode.EVENTPORT)
            inModes = [thisTreeNode.getSubnodeByNameAndCategory(modeName, ComponentTreeNode.MODE) for modeName in connection.in_modes]
            self.eventPortConnections.append(PortConnection(sourcePort, targetPort, inModes, thisTreeNode, connection))

        # - Transitions
        for transition in componentImplementation.normal_mode_transitions:
            sourceMode = thisTreeNode.getSubnodeByNameAndCategory(transition.source_mode, ComponentTreeNode.MODE)
            targetMode = thisTreeNode.getSubnodeByNameAndCategory(transition.target_mode, ComponentTreeNode.MODE)
            # HINT: Trigger is exactly one name of an event port or ""/None/False. In the latter case getSubnodeByNameAndCategory returns None as well.
            trigger = thisTreeNode.getSubnodeByNameAndCategory(transition.label.trigger, ComponentTreeNode.EVENTPORT)
            guardReads = [thisTreeNode.getSubnodeByNameAndCategory(read, ComponentTreeNode.DATAELEMENT) for read in expressionReadsFrom(transition.label.guard)]
            effects = [TransitionEffect(thisTreeNode.getSubnodeByNameAndCategory(expressionWritesTo(effect), ComponentTreeNode.DATAELEMENT),
                                        [thisTreeNode.getSubnodeByNameAndCategory(read, ComponentTreeNode.DATAELEMENT) for read in expressionReadsFrom(effect)],
                                        effect)
                       for effect in transition.label.effects]
            reactivates = filter(lambda ndsc: sourceMode not in ndsc.inModes and targetMode in ndsc.inModes, thisTreeNode.getNonDataSubcomponents())
            self.transitions.append(Transition(sourceMode, targetMode, trigger, guardReads, effects, reactivates, thisTreeNode, transition))
            
        # DEPRECATED: Probabilistic transitions are now stored like and along with normal transitions (Their trigger just refers to an ErrorEventPort).
        #for transition in componentImplementation.probability_mode_transitions:
        #    sourceMode = thisTreeNode.getSubnodeByNameAndCategory(transition.source_mode, ComponentTreeNode.MODE)
        #    targetMode = thisTreeNode.getSubnodeByNameAndCategory(transition.target_mode, ComponentTreeNode.MODE)
        #    trigger = None
        #    guardReads = []
        #    effects = []
        #    self.transitions.append(Transition(sourceMode, targetMode, trigger, guardReads, effects, reactivates, thisTreeNode, transition))


        # *** Print result ***
        if self.verbose and parent == None:
            printOutputBlock_Header("Results after:   SlimSlicer.generateComponentTree")
            
            printOutputBlock_Content("THE COMPONENT TREE:")
            printOutputBlock_Content(thisTreeNode.toString())

            printOutputBlock_Content("\nDATA PORT CONNECTIONS:")
            for portConnection in self.dataPortConnections:
                printOutputBlock_Content(portConnection.toString())

            printOutputBlock_Content("\nFLOWS:")
            for flow in self.flows:
                printOutputBlock_Content(flow.toString())

            printOutputBlock_Content("\nEVENT PORT CONNECTIONS:")
            for portConnection in self.eventPortConnections:
                printOutputBlock_Content(portConnection.toString())

            printOutputBlock_Content("\nTRANSITIONS:")
            for transition in self.transitions:
                printOutputBlock_Content(transition.toString())

            printOutputBlock_Footer()

        
        # *** Store root of component (instance) tree ***
        if parent == None:
            self.rootComponentTreeNode = thisTreeNode


        # *** Return result ***
        return thisTreeNode

    # </generateComponentTree(.): ...>
        

    # ----------------------------------------------------------------------------------------------


    @staticmethod
    def _extractAccessesFromAtomicProposition(anExpressionTreeNode):
        """
        >>> import slim_compiler.src.testexpr as testexpr
        
        >>> expr = testexpr.parseExpression("a.b.c = 1 and x.y")
        >>> _extractAccessesFromAtomicProposition(expr)
        ([u'a.b.c', u'x.y'], [])

        >>> ap = "x.y.z = enum:feuer and x.bol and a.b = a.c and u.v.mode = mode:m1 and v.u.error = error:felix"
        >>> expr = testexpr.parseExpression(ap)
        >>> _extractAccessesFromAtomicProposition(expr)
        ([u'x.y.z', u'x.bol', u'a.b', u'a.c'], [u'u.v.mode:m1', u'v.u.error:felix'])
        """
        dataElements = []
        nominalModes = []
        
        if isinstance(anExpressionTreeNode, BinaryOperatorTreeNode):
            # Attention: With = or !=, this might be a mode/error comparison!
            if anExpressionTreeNode.operator in [OperatorType.EQUAL, OperatorType.NOTEQUAL]:
                if isinstance(anExpressionTreeNode.left, IdentifierTreeNode) and \
                   isinstance(anExpressionTreeNode.right, IdentifierTreeNode):
                    # left to right:   <accessPath>.mode == <modeName> / <accessPath>.error == <errorName>
                    if (anExpressionTreeNode.left.node_type == NodeType.MODE_FIELD and anExpressionTreeNode.right.node_type == NodeType.MODE_VALUE) or \
                       (anExpressionTreeNode.left.node_type == NodeType.ERROR_FIELD and anExpressionTreeNode.right.node_type == NodeType.ERROR_VALUE):
                        nominalModes.append(anExpressionTreeNode.left.name + ":" + anExpressionTreeNode.right.name)
                        return (dataElements, nominalModes)
                    # right to left:   <modeName> == <accessPath>.mode / <errorName> = <accessPath>.error
                    elif (anExpressionTreeNode.left.node_type == NodeType.MODE_VALUE and anExpressionTreeNode.right.node_type == NodeType.MODE_FIELD) or \
                         (anExpressionTreeNode.left.node_type == NodeType.ERROR_VALUE and anExpressionTreeNode.right.node_type == NodeType.ERROR_FIELD):
                        nominalModes.append(anExpressionTreeNode.right.name + ":" + anExpressionTreeNode.left.name)
                        return (dataElements, nominalModes)
                        
            # No mode comparison found => treat left/right subexpressions normally
            (newDataElements, newNominalModes) = SlimSlicer._extractAccessesFromAtomicProposition(anExpressionTreeNode.left)
            dataElements.extend(newDataElements)
            nominalModes.extend(newNominalModes)
            (newDataElements, newNominalModes) = SlimSlicer._extractAccessesFromAtomicProposition(anExpressionTreeNode.right)
            dataElements.extend(newDataElements)
            nominalModes.extend(newNominalModes)

        elif isinstance(anExpressionTreeNode, UnaryOperatorTreeNode):
            (newDataElements, newNominalModes) = SlimSlicer._extractAccessesFromAtomicProposition(anExpressionTreeNode.subexpression)
            dataElements.extend(newDataElements)
            nominalModes.extend(newNominalModes)
            
        elif isinstance(anExpressionTreeNode, IdentifierTreeNode):
            # Only "normal" access paths are used (i.e., no enum/mode/error constants, no access paths to modes)
            if anExpressionTreeNode.node_type not in [NodeType.ENUM_VALUE, NodeType.MODE_FIELD, NodeType.MODE_VALUE, NodeType.ERROR_FIELD, NodeType.ERROR_VALUE]:
                dataElements.append(anExpressionTreeNode.name)

        elif isinstance(anExpressionTreeNode, CaseExpressionTreeNode):
            for (b, e) in anExpressionTreeNode.cases:
                (newDataElements, newNominalModes) = SlimSlicer._extractAccessesFromAtomicProposition(b)
                dataElements.extend(newDataElements)
                nominalModes.extend(newNominalModes)
                (newDataElements, newNominalModes) = SlimSlicer._extractAccessesFromAtomicProposition(e)
                dataElements.extend(newDataElements)
                nominalModes.extend(newNominalModes)
            (newDataElements, newNominalModes) = SlimSlicer._extractAccessesFromAtomicProposition(anExpressionTreeNode.default_expression)
            dataElements.extend(newDataElements)
            nominalModes.extend(newNominalModes)

        else: # Irrelevant: BooleanTreeNode, NumberTreeNode, FlowEquationTreeNode, FlowExpressionTreeNode are not considered!
            pass

        return (dataElements, nominalModes)


    def _tarjan(self, mode):
        # Explore this node
        self.tarjanStack.append(mode)
        self.tarjanNodeIndex[mode] = self.tarjanIndex
        self.tarjanNodeLowlink[mode] = self.tarjanIndex
        self.tarjanIndex = self.tarjanIndex + 1

        # Explore all successor nodes & adapt own lowlink
        for successor in [transition.targetMode for transition in filter(lambda transition: transition.sourceMode == mode, self.transitions)]:
            # ... yet unexplored
            if successor not in self.tarjanNodeIndex:
                self._tarjan(successor)
                self.tarjanNodeLowlink[mode] = min(self.tarjanNodeLowlink[mode], self.tarjanNodeLowlink[successor])
            # ... already explored but still on stack (could belong to this SCC)
            elif successor in self.tarjanStack:
                self.tarjanNodeLowlink[mode] = min(self.tarjanNodeLowlink[mode], self.tarjanNodeIndex[successor])

        # Check if current mode is root of a SCC
        if self.tarjanNodeLowlink[mode] == self.tarjanNodeIndex[mode]:
            # Collect all members of that SCC & remove them from stack
            membersOfSCC = []
            newMember = None
            while newMember != mode:
                newMember = self.tarjanStack.pop()
                membersOfSCC.append(newMember)
            # Store for each member the size of the SCC
            for memberOfSCC in membersOfSCC:
                self.tarjanNodeSCC[memberOfSCC] = len(membersOfSCC)
            
    
    def _recursivelyMakeModesOnCyclesInteresting(self, componentTreeNode):
        
        # *** Handle this component ***
        
        # Initialize values for Tarjan's SCC algorithm
        self.tarjanIndex = 0
        self.tarjanStack = []
        self.tarjanNodeIndex = {}
        self.tarjanNodeLowlink = {}
        self.tarjanNodeSCC = {}

        # Start Tarjan's SCC algorithm as long as unvisited modes exist
        for mode in componentTreeNode.getModes():
            if mode not in self.tarjanNodeIndex:
                self._tarjan(mode)

        # Make modes on cycles interesting
        #      a) mode in SCC with others (self.tarjanNodeSCC[mode] > 1)
        #      b) mode has self-loop
        for mode in componentTreeNode.getModes():
            if self.tarjanNodeSCC[mode] > 1 or \
               checkExistence(lambda transition: transition.sourceMode == mode and
                                                 transition.targetMode == mode
                              , self.transitions):
                self.interestingModes.append(mode)

        
        # *** Recursively handle non data subcomponents ***
        for ndsc in componentTreeNode.getNonDataSubcomponents():
            self._recursivelyMakeModesOnCyclesInteresting(ndsc)


    def _accessPathToComponentTreeNode(self, accessPath):
        if accessPath.find("mode:") != -1:
            accessPath = accessPath.replace("mode:", "")
            return self.rootComponentTreeNode.getSubnodeByNameAndCategory(accessPath, ComponentTreeNode.MODE)
        elif accessPath.find("error:") != -1:
            accessPath = accessPath.replace("error:", "")
            return self.rootComponentTreeNode.getSubnodeByNameAndCategory(accessPath, ComponentTreeNode.MODE)
        else:
            return self.rootComponentTreeNode.getSubnodeByNameAndCategory(accessPath, ComponentTreeNode.DATAELEMENT)
        

    def performSlicing(self, atomicPropositions, enforceKeepingCycles):
        """Performs the core slicing algorithm to identify interesting elements.

        The initial interesting elements are the data elements and modes used in
        the atomic proposition.
        In a fixpoint iteration all other elements which influence interesting
        elements are made interesting.
        
        @precondition: - The component (instance) tree has been generated by
                         SlimSlicer.generateComponentTree
        @param    atomicPropositions:
                    A list of atomic propositions from the property for which
                    slicing is to be done. The data elements and modes used in
                    the atomic propositions are the initial interesting elements.
        @type     atomicPropositions:
                    list(ExpressionTreeNode)
        @param    enforceKeepingCycles:
                    Iff true, all modes on syntactical cycles are kept in the
                              sliced specification.
                    Iff false, only those modes are kept which became
                               interesting wrt to the atomic propositions.
        @type     enforceKeepingCycles:
                    boolean
        """

        # *** Extract initially interesting data elements and modes from atomic propositions ***
        for atomicProposition in atomicPropositions:
            (newDataElements, newModes) = self._extractAccessesFromAtomicProposition(atomicProposition)
            self.interestingDataElements.extend([self._accessPathToComponentTreeNode(dataElement) for dataElement in newDataElements])
            self.interestingModes.extend([self._accessPathToComponentTreeNode(mode) for mode in newModes])


        # *** EXTENSION: If required, make all modes on syntactical cycles interesting ***
        if enforceKeepingCycles:
            self._recursivelyMakeModesOnCyclesInteresting(self.rootComponentTreeNode)


        # *** Fixpoint Iteration ***
        changed = True
        while changed:
            changed = False

            # ===== Transitions modifying interesting data elements =====
            # - Find transitions that modify interesting data elements (HINT: map -> is LHS of effect interesting?)
            interestingTransitions = filter(lambda transition: \
                                                   checkExistence(lambda effect: effect.writesTo in self.interestingDataElements, transition.effects)
                                            , self.transitions)
            # - Source modes of those transitions become interesting (HINT: target modes need not be interesting)
            newInterestingModes = [transition.sourceMode for transition in interestingTransitions]
            # - HINT: RHSs of effects, guards and triggers are treated in the "transitions from interesting to still uninteresting modes" or "transitions into interesting modes" section!
            # - Combine with previous results
            changed = strictOr([includeInList(self.interestingModes, newInterestingModes)], changed)

            # ===== Connections/Flows to interesting data ports =====
            # - Find connections/flows to interesting target data ports
            interestingConnections = filter(lambda dpc: dpc.targetPort in self.interestingDataElements, self.dataPortConnections)
            interestingFlows = filter(lambda flow: flow.targetPort in self.interestingDataElements, self.flows)
            # - All source data ports become interesting
            newInterestingDataElements = [dpc.sourcePort for dpc in interestingConnections]
            for flow in interestingFlows:
                newInterestingDataElements += flow.sourcePorts
            # - Modes where an interesting connection/flow is active become interesting
            newInterestingModes = []
            for conflow in interestingConnections + interestingFlows:
                if not conflow.definingComponent.listContainsAllModes(conflow.inModes):
                    newInterestingModes += conflow.inModes
            # - Combine with previous results
            changed = strictOr([includeInList(self.interestingDataElements, newInterestingDataElements),
                                includeInList(self.interestingModes, newInterestingModes)], changed)
            
            # ===== Connections to/from interesting event ports =====
            # - Find connections to/from interesting event ports
            interestingConnections = filter(lambda epc: (epc.targetPort in self.interestingEvents) or (epc.sourcePort in self.interestingEvents),
                                            self.eventPortConnections)
            # - The (other) event port becomes interesting (HINT: In both directions, because we do not distinguish master vs. reactive transitions here.)
            newInterestingEvents = [epc.sourcePort for epc in interestingConnections] + [epc.targetPort for epc in interestingConnections]
            # - Modes where an interesting connection (that is not active in every mode) is active become interesting
            newInterestingModes = []
            for epc in interestingConnections: 
                if not epc.definingComponent.listContainsAllModes(epc.inModes):
                    newInterestingModes += epc.inModes
            # - Combine with previous results
            changed = strictOr([includeInList(self.interestingEvents, newInterestingEvents),
                                includeInList(self.interestingModes, newInterestingModes)], changed)
            
            # ===== Transitions into interesting modes =====
            # - Find transitions into interesting modes   
            interestingTransitions = filter(lambda transition: transition.targetMode in self.interestingModes, self.transitions)
            # - Data elements on the RHS of effect assignments to EXTENSION: (weak or normal) interesting data elements become EXTENSION: weak interesting
            # - Data elements in the guard become EXTENSION: weak interesting
            newWeakInterestingDataElements = []
            for transition in interestingTransitions:
                for effect in filter(lambda effect: effect.writesTo in (self.interestingDataElements + self.weakInterestingDataElements), transition.effects):
                    newWeakInterestingDataElements += effect.readsFrom
                newWeakInterestingDataElements += transition.guardReads
            # - Source modes become interesting
            newInterestingModes = [transition.sourceMode for transition in interestingTransitions]
            # - Triggers become interesting
            newInterestingEvents = []
            for transition in interestingTransitions:
                if transition.trigger != None:
                    newInterestingEvents.append(transition.trigger)
            # - Combine with previous results
            changed = strictOr([includeInList(self.weakInterestingDataElements, newWeakInterestingDataElements),
                                includeInList(self.interestingModes, newInterestingModes),
                                includeInList(self.interestingEvents, newInterestingEvents)], changed)

            # ===== EXTENSION: Transitions from interesting to still uninteresting modes =====
            # - Find transitions from interesting (to uninteresting) modes
            interestingTransitions = filter(lambda transition: transition.sourceMode in self.interestingModes and
                                                               transition.targetMode not in self.interestingModes, self.transitions)
            # - EXTENSION: Data elements on the RHS of effect assignments to (strict) interesting data elements become weak interesting
            # - Data elements in their guards become weak interesting
            newWeakInterestingDataElements = []
            for transition in interestingTransitions:
                for effect in filter(lambda effect: effect.writesTo in self.interestingDataElements, transition.effects):
                    newWeakInterestingDataElements += effect.readsFrom
                newWeakInterestingDataElements += transition.guardReads
            # - Triggers become interesting
            newInterestingEvents = []
            for transition in interestingTransitions:
                if transition.trigger != None:
                    newInterestingEvents.append(transition.trigger)
            # - Combine with previous results
            changed = strictOr([includeInList(self.weakInterestingDataElements, newWeakInterestingDataElements),
                                includeInList(self.interestingEvents, newInterestingEvents)], changed)

            # ===== EXTENSION: Transitions with interesting triggers =====
            # - Find transitions with interesting triggers
            interestingTransitions = filter(lambda transition: transition.trigger in self.interestingEvents, self.transitions)
            # - Source modes become interesting
            newInterestingModes = [transition.sourceMode for transition in interestingTransitions]
            # HINT: Guards are handled in the section "transitions from interesting to still uninteresting modes" or "transitions into interesting modes"
            # - Combine with previous results
            changed = strictOr([includeInList(self.interestingModes, newInterestingModes)], changed)

            # ===== EXTENSION: Wrap-Up: Each weak interesting in data port is interesting =====
            # (Could be done directly when set to weak interesting, but then we would have to distinguish there [transition effects RHS, guards] between the different types of data elements)
            # - Find weak interesting in data ports            
            weakInterestingInDataPorts = filter(lambda dataElement: dataElement.typ == ComponentTreeNode.INDATAPORT, self.weakInterestingDataElements)
            # - Combine with previous results
            changed = strictOr([includeInList(self.interestingDataElements, weakInterestingInDataPorts)], changed)

            # ===== EXTENSION: For each weak interesting out data port, every source port of connections/flows active in *interesting* modes is interesting =====
            # (Cannot be done directly when adding it as interesting modes may be added. Different to handling of connections/flows to interesting ports!)
            # - Find connections/flows to weak interesting target data ports
            interestingConnections = filter(lambda dpc: dpc.targetPort in self.weakInterestingDataElements and \
                                                        isIntersectionNotEmpty(dpc.inModes, self.interestingModes)
                                            , self.dataPortConnections)
            interestingFlows = filter(lambda flow: flow.targetPort in self.weakInterestingDataElements and \
                                                   isIntersectionNotEmpty(dpc.inModes, self.interestingModes)
                                      , self.flows)
            # - All source data ports become interesting
            newInterestingDataElements = [dpc.sourcePort for dpc in interestingConnections]
            for flow in interestingFlows:
                newInterestingDataElements += flow.sourcePorts
            # - Combine with previous results
            changed = strictOr([includeInList(self.interestingDataElements, newInterestingDataElements)], changed)

            # ===== EXTENSION: Transitions reactivating an interesting data element or reactivating an interesting transitive control subcomponent without transitive mode history =====
            # - Find transitions reactivating an interesting data component
            interestingDataComponents = filter(lambda dataElement: dataElement.typ == ComponentTreeNode.DATACOMPONENT, self.interestingDataElements)            
            interestingTransitions = filter(lambda transition: \
                                                   checkExistence(lambda interestingDataComponent: transition.sourceMode not in interestingDataComponent.inModes and \
                                                                                                   transition.targetMode in interestingDataComponent.inModes
                                                                  , interestingDataComponents)
                                            , self.transitions)
            # - Find transitions reactivating a non-data subcomponent transitively containing an interesting data element or interesting mode w/o transitive mode history
            interestingTransitions += filter(lambda transition: \
                                                    checkExistence(lambda reactivated: reactivated.hasInterestingResetableElementWithoutModeHistoryBelow(self.interestingDataElements, self.interestingModes)
                                                                   , transition.reactivates)
                                             , self.transitions)
            # - Extract the source modes of those transitions to make them interesting
            #   HINT: That suffices since all transitions from interesting modes are kept.
            newInterestingModes = map(lambda transition: transition.sourceMode, interestingTransitions)
            # - Combine with previous results
            changed = strictOr([includeInList(self.interestingModes, newInterestingModes)], changed)

            # ===== EXTENSION: Modes changing an interesting data element by trajectory equation =====
            # - Find modes changing an interesting data element by trajectory equation
            newInterestingModes = filter(lambda mode: isIntersectionNotEmpty(mode.invariantWritesTo, self.interestingDataElements)
                                         , self.modes)
            # - Combine with previous results
            changed = strictOr([includeInList(self.interestingModes, newInterestingModes)], changed)

            # ===== EXTENSIONS: Data elements read in mode invariants =====
            # - Find
            newWeakInterestingDataElements = []
            for mode in self.interestingModes:
                newWeakInterestingDataElements += mode.invariantReadsFrom
            # - Combine with previous results
            changed = strictOr([includeInList(self.weakInterestingDataElements, newWeakInterestingDataElements)], changed)
            
        # </while changed: ... [Fixpoint Iteration] ...>


        # *** Set isInteresting-Flag of elements ***
        for dataElement in self.interestingDataElements: dataElement.isInteresting = True
        for dataElement in self.weakInterestingDataElements: dataElement.isWeakInteresting = True
        for event in self.interestingEvents: event.isInteresting = True
        for mode in self.interestingModes: mode.isInteresting = True

        self.rootComponentTreeNode.bottomUpPropagateBeingInteresting()
        for portConnection in self.dataPortConnections:
            portConnection.isNeeded = (portConnection.targetPort in self.interestingDataElements) or ((portConnection.targetPort in self.weakInterestingDataElements) and isIntersectionNotEmpty(portConnection.inModes, self.interestingModes)) # EXTENSION: +weak
        for flow in self.flows:
            flow.isNeeded = (flow.targetPort in self.interestingDataElements) or ((flow.targetPort in self.weakInterestingDataElements) and isIntersectionNotEmpty(flow.inModes, self.interestingModes)) # EXTENSION: +weak
        for portConnection in self.eventPortConnections:
            portConnection.isNeeded = portConnection.targetPort in self.interestingEvents
        for transition in self.transitions:
            for effect in transition.effects:                                   
                effect.isInteresting = (effect.writesTo in self.interestingDataElements) or \
                                       (effect.writesTo in self.weakInterestingDataElements and transition.targetMode in self.interestingModes) # EXTENSION: Keep assignments to weak interesting
            transition.isInteresting = (transition.sourceMode in self.interestingModes)


        # *** Print result ***
        if self.verbose:
            printOutputBlock_Header("Results after:   SlimSlicer.performSlicing")
            
            printOutputBlock_Content("INTERESTING DATA ELEMENTS:")
            for dataElement in self.interestingDataElements:
                printOutputBlock_Content(dataElement.toString())
            printOutputBlock_Content("\nWEAK INTERESTING DATA ELEMENTS:")
            for dataElement in self.weakInterestingDataElements:
                printOutputBlock_Content(dataElement.toString())
            printOutputBlock_Content("\nINTERESTING EVENTS:")
            for event in self.interestingEvents:
                printOutputBlock_Content(event.toString())
            printOutputBlock_Content("\nINTERESTING MODES:")
            for mode in self.interestingModes:
                printOutputBlock_Content(mode.toString())
            printOutputBlock_Content("\n[Based on this, we can conclude that the following things are interesting/needed, too:]")
            printOutputBlock_Content("\n=> INTERESTING COMPONENT TREE:")
            printOutputBlock_Content(self.rootComponentTreeNode.toString())
            printOutputBlock_Content("\n=> NEEDED DATA PORT CONNECTIONS:")
            for portConnection in self.dataPortConnections:
                printOutputBlock_Content(portConnection.toString())
            printOutputBlock_Content("\n=> NEEDED FLOWS:")
            for flow in self.flows:
                printOutputBlock_Content(flow.toString())
            printOutputBlock_Content("\n=> NEEDED EVENT PORT CONNECTIONS:")
            for portConnection in self.eventPortConnections:
                printOutputBlock_Content(portConnection.toString())
            printOutputBlock_Content("\n=> INTERESTING TRANSITIONS:")
            for transition in self.transitions:
                printOutputBlock_Content(transition.toString())
            printOutputBlock_Footer()
        
    # </performSlicing(.): ...>


    # ----------------------------------------------------------------------------------------------
    

    @staticmethod
    def _adaptInModesList(oldInModes, addSinkMode, componentTreeNode):
        """Adapts an "in_modes" list of modes from the unsliced compile info to a list for the sliced compile info.

        Basically, all uniteresting modes are removed.
        If addSinkMode is True and at least one uninteresting mode was contained in the modes list then "_SlicedSinkMode" is added.

        How to use this function for the different kind of elements that have an an "in_modes" list:
        - Non Data Subcomponents:
          >>> self._adaptInModesList(originalSubcomponent.in_modes, scTreeNode.isInteresting (always true when called), componentTreeNode)
              ... cf. diploma thesis ...
        - Data Subcomponents:
          >>> self._adaptInModesList(originalSubcomponent.in_modes, scTreeNode.isInteresting, componentTreeNode)
              ... cf. diploma thesis ...
        - Data Port Connections / Flows / Event Port Connections:
          >>> self._adaptInModesList(originalDataPortConnection.in_modes, dataPortConnection.targetPort.isInteresting, componentTreeNode)
          >>> self._adaptInModesList(originalFlow.in_modes, flow.targetPort.isInteresting, componentTreeNode)
          >>> self._adaptInModesList(originalEventPortConnection.in_modes, eventPortConnection.targetPort.isInteresting (always true when called), componentTreeNode)
              - We only consider connections which are needed, i.e. their targetPorts are interesting or at least weak interesting.
              - For a connection to an interesting targetPort:
                    - If it is not active in every mode, any mode in which it is active became interesting.
                      Thus, either in_modes contains all modes or all modes in in_modes are interesting.
                    - In the first case we have to ensure that the new in_modes list still contains every mode. So we have to add "_SlicedSinkMode" if any mode was sliced away, i.e. uninteresting modes exist.
                      In the second case we know that the connection is not active in any uninteresting mode. So "_SlicedSinkMode" must not be added. Independent of addSinkMode it won't be added since the in_modes list does not contain any uninteresting mode.
                    => addSinkMode should be set to True
              - For a connection to a non interesting but weak interesting targetPort:
                    - The targetPort's value is only needed in interesting modes. Thus, the same applies for the connection itself.
                      Consequently, "_SlicedSinkMode" need not (and must not) be added.
                    => addSinkMode should be set to False
              => addSinkMode should be set to targetPort.isInteresting



        @param    oldInModes:
                    The original "in_modes" list from the original compile info.
        @type     oldInModes:
                    list(string)
        @param    addSinkMode:
                    Iff True, "_SlicedSinkMode" is added when some uninteresting
                    modes have been removed from the list.
        @type     addSinkMode:
                    boolean
        @param    componentTreeNode:
                    Node of the component (instance) tree representing the non
                    data component containing the modes in question.
        @type     componentTreeNode:
                    ComponentTreeNode with .typ = ComponentTreeNode.NONDATACOMPONENT
        @return:  The new "in_modes" list for the sliced compile info.
        @rtype:   list(string)
        """

        # Translate the old list of mode names to mode objects
        oldInModesObjects = [componentTreeNode.getSubnodeByNameAndCategory(modeName, ComponentTreeNode.MODE) for modeName in oldInModes]

        # Create the new list of modes by filtering the interesting ones from the old list
        newInModesObjects = filter(lambda mode: mode.isInteresting, oldInModesObjects)

        # Retranslate the new list of mode objects to a list of mode names
        newInModeStrings = [mode.name for mode in newInModesObjects]

        # Add _SlicedSinkMode
        if addSinkMode   and   len( filter(lambda mode: not mode.isInteresting, oldInModesObjects) ):
            # If at least one uninteresting mode is included, ...
            # a) IF interesting: we know that every mode is included, thus we just have to add the _SlicedSinkMode
            # b) IF only weak interesting: we do not have to care about uninteresting modes
            newInModeStrings += ["_SlicedSinkMode"]

        # Return result
        return newInModeStrings
    
    # </_adaptInModesList(.): ...>


    def _adaptModeInvariant(self, oldInvariant, componentTreeNode):

        # Distinguish by ExpressionTreeNode types
        if oldInvariant == None:
            return None
        
        elif isinstance(oldInvariant, BinaryOperatorTreeNode):
            if oldInvariant.operator == OperatorType.AND:
                newLeftInvariant = self._adaptModeInvariant(oldInvariant.left, componentTreeNode)
                newRightInvariant = self._adaptModeInvariant(oldInvariant.right, componentTreeNode)
                if newLeftInvariant == None and newRightInvariant == None:
                    return None
                elif newLeftInvariant == None:
                    return newRightInvariant
                elif newRightInvariant == None:
                    return newLeftInvariant
                else:
                    return BinaryOperatorTreeNode(newLeftInvariant, newRightInvariant, OperatorType.AND, oldInvariant.line, oldInvariant.pos)
            else:
                return oldInvariant

        elif isinstance(oldInvariant, TrajectoryEquationTreeNode):
            if isinstance(oldInvariant.variable, IdentifierTreeNode):
                if componentTreeNode.getSubnodeByNameAndCategory(oldInvariant.variable.name, ComponentTreeNode.DATAELEMENT) in \
                   (self.interestingDataElements + self.weakInterestingDataElements):
                    return oldInvariant
                else:
                    return None
            else:
                return None
        
        else: 
            return oldInvariant
      

    
    def _recursivelyWriteSlicedComponentTreeToSlicedPublicPackage(self, slicedPublicPackageObject, componentTreeNode):
        """This function does the real work for createSlicedCompileInfo."""

        # *** Stop, if given tree node does not represent a non data component ***
        if componentTreeNode == None or componentTreeNode.typ != ComponentTreeNode.NONDATACOMPONENT:
            return


        # HINT: In the following we use the data from the original compile info to create elements in the sliced compile info accordingly.
        #       As the sliced compile is based on an explicit representation of the component tree, some component types/implementations might occure several times.
        #       To ensure that the different occurences based on the same type/implementation are independent, be aware that:
        #       - Original Strings, Integers, Lists of ~, ... (elementary data types) can be reused as parameters directly, as they are copied.
        #       - However, objects (esp. ExpressionTreeNodes storing default values and expressions) have to be deepcopied manually!
            

        # *** Create sliced Component Type ***
        slicedType = slicedPublicPackageObject.add_component_type(componentTreeNode.accessPath.replace(".", "_"), Category.SYSTEM, False, 0, 0)

        # - In Data Ports
        for portTreeNode in filter(lambda idp: idp.isInteresting, componentTreeNode.getSubnodesOfType(ComponentTreeNode.INDATAPORT)):
            originalPort = portTreeNode.inCompileInfo
            slicedType.add_port(originalPort.name, True, False, originalPort.data_type, copy.deepcopy(originalPort.default_value), False, False, originalPort.line, originalPort.pos)
        # - Out Data Ports   (ATTENTION: Some might indeed be only week interesting!)
        for portTreeNode in filter(lambda odp: odp.isInteresting or odp.isWeakInteresting, componentTreeNode.getSubnodesOfType(ComponentTreeNode.OUTDATAPORT)):
            originalPort = portTreeNode.inCompileInfo
            slicedType.add_port(originalPort.name, False, False, originalPort.data_type, copy.deepcopy(originalPort.default_value), originalPort.observable, originalPort.alarm, originalPort.line, originalPort.pos)
        # - In Event Ports
        for portTreeNode in filter(lambda iep: iep.isInteresting, componentTreeNode.getSubnodesOfType(ComponentTreeNode.INEVENTPORT)):
            originalPort = portTreeNode.inCompileInfo
            slicedType.add_port(originalPort.name, True, True, None, None, False, False, originalPort.line, originalPort.pos)
        # - Out Event Ports
        for portTreeNode in filter(lambda oep: oep.isInteresting, componentTreeNode.getSubnodesOfType(ComponentTreeNode.OUTEVENTPORT)):
            originalPort = portTreeNode.inCompileInfo
            # Keep ErrorEventPorts as ErrorEventPorts
            if isinstance(originalPort, ErrorEventPort):
                # HINT: slicedType.add_error_event_port(...) requires information as strings.
                #       But in ErrorEventPort they are not stored as strings any more.
                #       To avoid retransformation to string as done in toSLIMVisitor().visit(ErrorEventPort),
                #       we directly add a copy of the ErrorEventPort.
                slicedType.out_event_ports[originalPort.name] = copy.deepcopy(originalPort)
                slicedType.out_event_ports[originalPort.name].file = "<generated-by-slicer>"
            # Keep normal OutEventPorts as normal ...
            else:
                slicedType.add_port(originalPort.name, False, True, None, None, False, False, originalPort.line, originalPort.pos)
    

        # *** Create sliced Component Implementation ***
        oldImpl = componentTreeNode.inCompileInfo
        slicedImpl = slicedPublicPackageObject.add_component_implementation(componentTreeNode.accessPath.replace(".", "_") + ".Implementation", Category.SYSTEM, False, 0, 0)

        # - Non Data Subcomponents     (+ RECURSIVE CALL)
        for scTreeNode in filter(lambda ndsc: ndsc.isInteresting, componentTreeNode.getNonDataSubcomponents()):
            self._recursivelyWriteSlicedComponentTreeToSlicedPublicPackage(slicedPublicPackageObject, scTreeNode)
            originalSubcomponent = oldImpl.non_data_subcomponents[scTreeNode.name]
            slicedImpl.add_non_data_subcomponent(originalSubcomponent.name, originalSubcomponent.category,
                                                 scTreeNode.accessPath.replace(".", "_") + ".Implementation",
                                                 "", "", "",
                                                 self._adaptInModesList(originalSubcomponent.in_modes, True, componentTreeNode),
                                                 originalSubcomponent.line, originalSubcomponent.pos)
            
        # - Data Subcomponents
        for scTreeNode in filter(lambda dc: dc.isInteresting or dc.isWeakInteresting, componentTreeNode.getSubnodesOfType(ComponentTreeNode.DATACOMPONENT)):
            originalSubcomponent = oldImpl.data_subcomponents[scTreeNode.name]
            slicedImpl.add_data_subcomponent(originalSubcomponent.name, originalSubcomponent.data_type, copy.deepcopy(originalSubcomponent.default_value),
                                             self._adaptInModesList(originalSubcomponent.in_modes, scTreeNode.isInteresting, componentTreeNode),
                                             originalSubcomponent.line, originalSubcomponent.pos)
        
        # - Modes
        interestingModes = filter(lambda mode: mode.isInteresting, componentTreeNode.getModes())
        uninterestingModes = filter(lambda mode: not mode.isInteresting, componentTreeNode.getModes())
        for modeTreeNode in interestingModes:
            # Add every interesting mode (originalMode.type = initial/activation/nominal/error is preserved!)
            originalMode = modeTreeNode.inCompileInfo
            slicedImpl._addMode(originalMode.name,
                                self._adaptModeInvariant(originalMode.invariant, componentTreeNode),
                                originalMode.type, originalMode.line, originalMode.pos)
        if len(uninterestingModes) > 0:
            # If at least one mode is interesting, the initial/activation mode must be interesting as well => add "normal" sink state
            if len(interestingModes) > 0:
                slicedImpl._addMode("_SlicedSinkMode", None, ModeType.NOMINAL, 0, 0)
            # If no interesting mode at all, add "initial"/"activation" sink state as the only state
            else:
                if componentTreeNode.hasModeHistory():
                    slicedImpl._addMode("_SlicedSinkMode", None, ModeType.INITIAL, 0, 0)
                else:
                    slicedImpl._addMode("_SlicedSinkMode", None, ModeType.ACTIVATION, 0, 0)
                # HINT: As long as we do not store the mode name, they have to be identical. Not for redirecting transitions (in the second case there are no transitions)
                #       but for adaption of inModes lists where we might have to add the sink state as well.

        # ^- stored as subnode of componentTreeNode
        # v- stored in global list, but have definingComponent-link to componentTreeNode
        
        # - Data Port Connections
        for portConnection in filter(lambda portConnection: portConnection.definingComponent == componentTreeNode and portConnection.isNeeded, self.dataPortConnections):
            originalPortConnection = portConnection.inCompileInfo
            slicedImpl.add_data_port_connection(originalPortConnection.source_port, originalPortConnection.target_port,
                                                self._adaptInModesList(originalPortConnection.in_modes, portConnection.targetPort.isInteresting, componentTreeNode),
                                                originalPortConnection.line, originalPortConnection.pos) 

        # - Flows
        for flow in filter(lambda flow: flow.definingComponent == componentTreeNode and flow.isNeeded, self.flows):
            originalFlow = flow.inCompileInfo
            slicedImpl.add_flow(copy.deepcopy(originalFlow.expression),
                                self._adaptInModesList(originalFlow.in_modes, flow.targetPort.isInteresting, componentTreeNode),
                                originalFlow.line, originalFlow.pos)

        # - Event Port Connections
        for portConnection in filter(lambda portConnection: portConnection.definingComponent == componentTreeNode and portConnection.isNeeded, self.eventPortConnections):
            originalPortConnection = portConnection.inCompileInfo
            slicedImpl.add_event_port_connection(originalPortConnection.source_port, originalPortConnection.target_port,
                                                 self._adaptInModesList(originalPortConnection.in_modes, True, componentTreeNode),
                                                 originalPortConnection.line, originalPortConnection.pos)

        # - Transitions
        for transition in filter(lambda transition: transition.definingComponent == componentTreeNode and transition.isInteresting, self.transitions):
            # Retrieve all infos about the original transition
            originalTransition = transition.inCompileInfo
            # The source mode exists because the transition is interesting. But what about the target mode?
            if transition.targetMode.isInteresting:
                targetMode = originalTransition.target_mode
            else:
                targetMode = "_SlicedSinkMode"
            # Copy interesting effects
            interestingEffectsCopy = [copy.deepcopy(effect.inCompileInfo) for effect in filter(lambda effect: effect.isInteresting, transition.effects)]
            # Does identical sliced transition already exist?
            skipTransition = checkExistence(lambda exTrans: exTrans.source_mode == originalTransition.source_mode and \
                                                   exTrans.target_mode == targetMode and \
                                                   exTrans.label.trigger == originalTransition.label.trigger and \
                                                   checkSyntacticalIdentityOfExpressions(exTrans.label.guard, originalTransition.label.guard) and \
                                                   identicalLists(exTrans.label.effects, interestingEffectsCopy,
                                                                  lambda ef1, ef2: checkSyntacticalIdentityOfExpressions(ef1, ef2))
                                           , slicedImpl.normal_mode_transitions)
            # Add the transition
            if not skipTransition:
                slicedImpl.add_normal_mode_transition(originalTransition.source_mode, targetMode,
                                                      originalTransition.label.trigger, copy.deepcopy(originalTransition.label.guard), interestingEffectsCopy,
                                                      originalTransition.line, originalTransition.pos)
            
            # DEPRECATED: Probabilistic transitions are now stored along with normal transitions. Their trigger just refers to an ErrorEventPort.
            #if isinstance(originalTransition.label, ModeTransitionLabel):
            #    slicedImpl.add_normal_mode_transition(originalTransition.source_mode, targetMode,
            #                                          copy.deepcopy(originalTransition.label.trigger), copy.deepcopy(originalTransition.label.guard),
            #                                          [copy.deepcopy(effect.inCompileInfo) for effect in filter(lambda effect: effect.isInteresting, transition.effects)],
            #                                          originalTransition.line, originalTransition.pos)
            #elif isinstance(originalTransition.label, ProbabilityModeTransitionLabel):
            #    slicedImpl.add_probability_mode_transition(originalTransition.source_mode, targetMode,
            #                                               originalTransition.label.probability,
            #                                               originalTransition.line, originalTransition.pos)


        # *** Return created implementation ***
        return slicedImpl
    
    # </_recursivelyWriteSlicedComponentTreeToSlicedPublicPackage(.): ...>


    def createSlicedCompileInfo(self):
        """Creates a compile info/SLIM specification containing the slicing result.

        To this end, the interesting elements of the sliced Component (Instance)
        Tree are written to the new compile info. As different component instances
        of the same original component type/implementation might be sliced
        differently (cf. SlimSlicer.generateComponentTree), for each component
        instance exactly one type and one implementation is written to the
        new sliced compile info. As name for those the access path to the
        corresponding component instance is used (with "." replaced by "_", that
        is the reason why "_" must be eliminated from access paths before
        slicing). Note: Fortunately, the access path to a component instance is
        not affected by renaming type and implementation of that component and
        so it is consequently not necessary to adapt the atomic propositions.

        Although some aspects in the new sliced compile info are directly taken
        from the original compile info (e.g., expressions) the new compile info
        is totally independent from the old one since those aspects are copied
        using copy.deepcopy.

        Actually, createSlicedCompileInfo only sets up the new sliced compile
        info. The real work of filling it with the interesting parts is done
        in _recursivelyWriteSlicedComponentTreeToSlicedPublickPackage(...).

        The new sliced compile info and its root are returned and stored in
        SlimSlicer.slicedCompileInfo and SlimSlicer.slicedRootComponent.

        @precondition: No existing mode may be called "_SlicedSinkMode".

        @return: A tuple containing the sliced compile info and its root component.
        @rtype:  (compile.CompileInformation, nominalbehaviour.ComponentImplementation)
        """

        # *** Check that "_SlicedSinkMode" is not used as mode name ***
        if checkExistence(lambda mode: mode.name == "_SlicedSinkMode", self.modes):
            raise Exception("Name '_SlicedSinkMode' already used as a mode name in the original specification.")
        
        # *** Create new Compile Info ***
        self.slicedCompileInfo = slim_compiler.src.dataStructure.compile.CompileInformation()
        self.slicedCompileInfo.current_file = "<generated-by-slicer>"
        slicedPackageObject = self.slicedCompileInfo.add_package("SlicedSlimSpecification", 0, 0)
        slicedPublicPackageObject = slicedPackageObject.get_public_subpackage()

        # *** Write sliced Component Tree to new Compile Info ***
        self.slicedRootComponent = self._recursivelyWriteSlicedComponentTreeToSlicedPublicPackage(slicedPublicPackageObject, self.rootComponentTreeNode)

        # *** Print result ***
        if self.verbose:
            printOutputBlock("Results after:   SlimSlicer.createSlicedCompileInfo",
                             "ROOT: " + self.slicedRootComponent.subpackage.package.name + "::" + self.slicedRootComponent.name + "\n" + \
                             toSLIMVisitor().visit(self.slicedCompileInfo))

        # *** Return result ***
        return (self.slicedCompileInfo, self.slicedRootComponent)

# ==================================================================================================
# </class SlimSlicer: ...>
# ==================================================================================================





# ==================================================================================================
# Convenience Function for invoking the SLIM Slicer
# ==================================================================================================

def doSlicing(originalRootComponent, atomicPropositions, enforceKeepingCycles, verbose = False):
    """Convenience function for getting the Slicing of your compile info/SLIM
    specification wrt the atomic propositions done without the need of
    knowing the details of the class SlimSlicer.

    @precondition: - In componentImplementation and all its subcomponents the
                     underscores must already have been eliminated (since
                     they are used to transform the access path into the new
                     component type name).
                   - In every occuring componentImplementation empty in modes
                     lists (which implicitly contain every mode) must have been
                     filled up so that they now explicitly contain every mode.
                   - No existing mode may be called "_SlicedSinkMode".

    @param    originalRootComponent:
                The root component (implementation) of the specification which
                is to be sliced.
    @type     originalRootComponent:
                nominalbehaviour.ComponentImplementation
    @param    atomicPropositions:
                A list of atomic propositions from the property for which
                slicing is to be done. The data elements and modes used in
                the atomic propositions are the initial interesting elements.
    @type     atomicPropositions:
                list(ExpressionTreeNode)
    @param    enforceKeepingCycles:
                Iff true, all modes on syntactical cycles are kept in the
                          sliced specification.
                Iff false, only those modes are kept which became
                           interesting wrt to the atomic propositions.
    @type     enforceKeepingCycles:
                boolean
    @param    verbose:
                Iff true, verbose output is printed to screen
                (using shared.printOutputBlock_...).
    @type     verbose:
                boolean, default False
    @return: A tuple containing the sliced compile info and its root component.
             Note: Although the new compile info is based on the original one,
                   their data structures are totally independent as the
                   identical parts were copied using copy.deepcopy.
    @rtype:  (compile.CompileInformation, nominalbehaviour.ComponentImplementation)
    """

    # 0.) Create SlimSlicer instance
    mySlimSlicer = SlimSlicer(verbose)

    # 1.) Generate Component (Instance) Tree
    mySlimSlicer.generateComponentTree(originalRootComponent)

    # 2.) Do Slicing
    mySlimSlicer.performSlicing(atomicPropositions, enforceKeepingCycles)

    # 3.) Create sliced Compile Info
    (slicedCompileInfo, slicedRootComponent) = mySlimSlicer.createSlicedCompileInfo()

    # Finally, return result
    return (slicedCompileInfo, slicedRootComponent)

# ==================================================================================================
# /Convenience Function for invoking the SLIM Slicer
# ==================================================================================================





# ==================================================================================================
# Main Program
# ==================================================================================================

if __name__ == "__main__":
    
    # Print Start-of-Program Message
    print "START OF SLIM-SLICER\n"



    # ~~~~~ SetUp example ~~~~~
    
    # - "FullAdder.slim"
    #slimFile = "..\\promela\\examples\\FullAdder.slim"
    #atomicPropositions = []
    #atomicPropositions += ["ha1.s"]
    #atomicPropositions += ["ha1.s and ha2.s"]
    #atomicPropositions += ["s"]
    #rootComponentName = ""
    #enforceKeepingCycles = False
    
    # - "Adder_X.slim"
    #slimFile = "..\\promela\\examples\\Adder_X.slim"
    #atomicPropositions = []
    #atomicPropositions += ["bit1.output"]
    #atomicPropositions += ["adder.input1"]
    #atomicPropositions += ["adder.output"]
    #atomicPropositions += ["output"]
    #rootComponentName = ""
    #enforceKeepingCycles = False

    # - Test: "reactivationOfInterestingDataElements"
    #slimFile = "tests\\reactivationOfInterestingDataElements.slim"
    #atomicPropositions = []
    #atomicPropositions += ["dsc"]
    #atomicPropositions += ["A.x"]
    #atomicPropositions += ["A.B.y"]
    #atomicPropositions += ["A.B.C.z"]
    #rootComponentName = ""
    #enforceKeepingCycles = False

    # - Test: "connectionsActiveInModes"
##    slimFile = "tests\\connectionsActiveInModes\\connectionsActiveInModes.slim"
##    atomicPropositions = []
##    
##    atomicPropositions += ["explicitAlwaysActive.odp_conn = 0"]
##    atomicPropositions += ["explicitAlwaysActive.odp_flow = 0"]
##    #atomicPropositions += ["mode = mode:makeEAsIEPInteresting"]
##    atomicPropositions += ["mode = mode:makeEAsOEPInteresting"]
##    atomicPropositions += ["explicitAlwaysActive.left.idp = 0"]
##    atomicPropositions += ["explicitAlwaysActive.left.mode = mode:makeIEPInteresting"]
##    atomicPropositions += ["explicitAlwaysActive.left.mode = mode:makeOEPInteresting"]
##    atomicPropositions += ["explicitAlwaysActive.right.idp = 0"]
##    #atomicPropositions += ["explicitAlwaysActive.right.mode = mode:makeIEPInteresting"]
##    #atomicPropositions += ["explicitAlwaysActive.right.mode = mode:makeOEPInteresting"]
##
##    atomicPropositions += ["implicitAlwaysActive.odp_conn = 0"]
##    atomicPropositions += ["implicitAlwaysActive.odp_flow = 0"]
##    #atomicPropositions += ["mode = mode:makeIAsIEPInteresting"]
##    atomicPropositions += ["mode = mode:makeIAsOEPInteresting"]
##    atomicPropositions += ["implicitAlwaysActive.left.idp = 0"]
##    atomicPropositions += ["implicitAlwaysActive.left.mode = mode:makeIEPInteresting"]
##    atomicPropositions += ["implicitAlwaysActive.left.mode = mode:makeOEPInteresting"]
##    atomicPropositions += ["implicitAlwaysActive.right.idp = 0"]
##    #atomicPropositions += ["implicitAlwaysActive.right.mode = mode:makeIEPInteresting"]
##    #atomicPropositions += ["implicitAlwaysActive.right.mode = mode:makeOEPInteresting"]
##
##    atomicPropositions += ["explicitPartialActive.odp_conn = 0"]
##    atomicPropositions += ["explicitPartialActive.odp_flow = 0"]
##    #atomicPropositions += ["mode = mode:makeEPsIEPInteresting"]
##    atomicPropositions += ["mode = mode:makeEPsOEPInteresting"]
##    atomicPropositions += ["explicitPartialActive.left.idp = 0"]
##    atomicPropositions += ["explicitPartialActive.left.mode = mode:makeIEPInteresting"]
##    atomicPropositions += ["explicitPartialActive.left.mode = mode:makeOEPInteresting"]
##    atomicPropositions += ["explicitPartialActive.right.idp = 0"]
##    #atomicPropositions += ["explicitPartialActive.right.mode = mode:makeIEPInteresting"]
##    #atomicPropositions += ["explicitPartialActive.right.mode = mode:makeOEPInteresting"]
##    
##    rootComponentName = ""
##    enforceKeepingCycles = False
    
    # - Test:
    slimFile = "tests\\enforceKeepingCycles.slim"
    atomicPropositions = []
    atomicPropositions += ["c.mode = mode:c1"]
    rootComponentName = ""
    enforceKeepingCycles = False

    # ~~~~~ /SetUp example ~~~~~

    

    # Print the used example to screen
    print "Run Slicer on example:"
    print "- SLIM file: " + slimFile
    print "- Atomic Prop.: " + str(atomicPropositions)
    print "- Root Comp.: " + rootComponentName
    print "\n"

    # Parse SLIM file
    theCompileInfo = slim_compiler.src.dataStructure.compile.CompileInformation()
    invokers.slim.parse(theCompileInfo, [slimFile])
    print theCompileInfo.error_collector
    invokers.slim.check(theCompileInfo)
    print theCompileInfo.error_collector

    # Find root
    options.ROOT_NAME = rootComponentName
    theRootComponent = invokers.slim.get_root_impl(theCompileInfo, rootComponentName)

    # Parse Atomic Propositions
    atomicPropositionsAsTrees = []
    for atomicProposition in atomicPropositions:
        newAPTree = testexpr.parseExpression(atomicProposition)
        parse_property.check_property_tree(theCompileInfo, newAPTree, theRootComponent)
        atomicPropositionsAsTrees.append(newAPTree)

    # Preprocess
    recursivelyEliminateUnderscoresInComponentsDataElementsAndSubcomponents(theRootComponent)
    for atomicProposition in atomicPropositionsAsTrees:
        performUnderscoreEliminationInAtomicProposition(atomicProposition, theRootComponent)
    recursivelyFillUpEmptyInModesLists(theRootComponent)
    #printOutputBlock("Input:   the parsed, checked & preprocessed Compile Info & Atomic Propositions",
    #                 toSLIMVisitor(theCompileInfo) + "\n- - - - -\n" + \
    #                 "\n".join([expressionToString(atomicProposition) for atomicProposition in atomicPropositionsAsTrees]))

    # Verbose execute Slicer on example
    doSlicing(theRootComponent, atomicPropositionsAsTrees, enforceKeepingCycles, verbose=True)



    # Print End-of-Program Message
    print "\n\nEND OF SLIM-SLICER"

# ==================================================================================================
# /Main Program
# ==================================================================================================

