# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


# This is a high-level API to be used when converting slim hierachies
# or expressions to smv

from slim_compiler.src.smv import converter
import options
import os
from slim_compiler.src.dataStructure.visitors.expressionVisitor \
    import ExpressionPostProcessor
import copy


def convert_to_smv(root_impl, smv_file,
                   generate_time_var=options.SMV_GENERATE_TIME_VAR,
                   generate_time_fairness_alt=options.SMV_FAIR_TIMED):
    """Converts all the components starting from given root
    implementation. smv_file can be None to convert to the standard
    output. A walker is used to check if an hibrid-automata has to
    be generated.

    Parameters generate_time_var and generate_time_fairness_alt
    control if additional constraints for timed fsm have to be added,
    and will be ignored for discrete (non-timed) FSMs.

    Returns a pair of booleans (infinite, timed). Infinite is True
    if the domain is infinite (requires mathsat), or else
    False. Timed is True if the domain contains time (requires
    mathsat).

    """

    # Postprocess expression types for SMV
    ExpressionPostProcessor().visit(
        root_impl.subpackage.package.compile_information)

    from slim_compiler.src.smv.walkers import InfiniteDomainFinderWalker

    walker = InfiniteDomainFinderWalker()
    walker.walk(root_impl)

    is_timed = walker.has_found_clock_data()
    smvconvert = converter.SmvConverter(
        generate_timed_fsm=is_timed,
        generate_time_var=is_timed and generate_time_var,
        generate_time_fairness_alt=generate_time_fairness_alt)

    smvconvert.dump(root_impl, smv_file)

    return (walker.has_found_infinite_data(), is_timed)


def convert_expression(root_impl, slim_expression,
                       time_unit_dest=None,
                       id_prefix=""):
    """Converts a slim expression (see module
    dataStructures.expressions) to a smv expression represented by a
    string. Gets a ComponentImplementation and an expression tree, and
    returns a string containing the converted smv expression.

    @param time_unit_dest specifies the time unit (TU) to be used when
    converting clocks values to/from reals. THe value can be None or
    one value in
    `slim_compiler.smv.dataStructure.expressions.TimeUnitType`.

    If None (recommended), the SmvConverter will decide a best TU for
    encoding the expression. If `TimeUnitType.NONE`, the default TU
    will be used, otherwise the specified TU will be used for the
    conversion.

    For example, if using None::

      c1 : data clock min;
      c2 : data clock day;
      c3 : data clock min;

      -- the expression:
      (c1 + c2 >= c3)


    The SmvConverted chooses min as destination TU to optimize the
    encoding, and the expression gets converted to::

      (sc_c1 + sc_c2 * 1440) >= sc_c3
    """
    smvconvert = converter.SmvConverter()

    conv_id_prefix = (smvconvert.sc_name(id_prefix) + "."
                      if id_prefix
                      else "")
    smvconvert.set_id_prefix(
        smvconvert.root_var() + "." + conv_id_prefix)

    # Postprocess expression types for SMV
    # A copy is made so the original expression can be reused
    slim_expression = copy.deepcopy(slim_expression)
    ExpressionPostProcessor().visit(slim_expression)

    smvexpr = smvconvert.expression(root_impl, slim_expression,
                                    time_unit_dest=time_unit_dest)
    return smvexpr.replace(os.linesep, " ")


def parse_and_convert_numeric_expression(impl, expr_string,
                                         dest_type=None,
                                         dest_tu=None,
                                         id_prefix=""):
    """Converts a numeric slim expression (as string) into a smv expression.

    @param dest_type: literal in expressions.NodeType to decide the
    destination type when not clear from the context (e.g. 100 should
    be considered a clock). Default is `expressions.NodeType.REAL`.

    @param dest_tu: decides the destination
    time unit, which is handly if converting timed values for clock
    comparison. If None, the default time unit will be used, which is
    suitable for encoded delta and time variables
    """
    from slim_compiler.src.testexpr import parseExpression
    from slim_compiler.src.dataStructure.visitors.expressionVisitor import (
        TypeVisitor, ExpressionPostProcessor)
    from slim_compiler.src.dataStructure import expressions
    import slim_compiler.src.globals

    cinfo = impl.subpackage.package.compile_information
    dest_type = dest_type or expressions.NodeType.REAL

    # hack for issue 571
    slim_compiler.src.globals.PARSER_ACCEPT_EXTENDED_MODEL = True
    try:
        ast = parseExpression(expr_string)
        TypeVisitor(cinfo).visit(ast)
        ExpressionPostProcessor().visit(ast, dest_type)
        smv = convert_expression(impl, ast, time_unit_dest=dest_tu,
                                 id_prefix=id_prefix)
        return smv

    finally:
        slim_compiler.src.globals.PARSER_ACCEPT_EXTENDED_MODEL = False


def parse_and_convert_predicate(impl, pred_string, id_prefix=""):
    """Converts to smv given predicate expression as string"""
    from slim_compiler.src.testexpr import parseExpression
    from slim_compiler.src.parse_property import check_property_tree
    import slim_compiler.src.globals

    cinfo = impl.subpackage.package.compile_information

    # hack for issue 571
    slim_compiler.src.globals.PARSER_ACCEPT_EXTENDED_MODEL = True
    try:
        expr_ast = parseExpression(pred_string)
        check_property_tree(cinfo, expr_ast, impl)
        smv = convert_expression(impl, expr_ast, id_prefix=id_prefix)
        return smv

    finally:
        slim_compiler.src.globals.PARSER_ACCEPT_EXTENDED_MODEL = False


def parse_and_convert_id(impl, id_string, id_prefix=""):
    """Converts an identifier (as string) to the corresponding
    identified in smv"""
    from slim_compiler.src.testexpr import parseExpression
    from slim_compiler.src.dataStructure.visitors.expressionVisitor import (
        TypeVisitor,)
    import slim_compiler.src.globals

    cinfo = impl.subpackage.package.compile_information

    # hack for issue 571
    slim_compiler.src.globals.PARSER_ACCEPT_EXTENDED_MODEL = True
    try:
        id_ast = parseExpression(id_string)
        TypeVisitor(cinfo).visit(id_ast)
        smv = convert_expression(impl, id_ast, id_prefix=id_prefix)
        return smv

    finally:
        slim_compiler.src.globals.PARSER_ACCEPT_EXTENDED_MODEL = False


def get_smv_timed_var():
    """Returns a pair (name, time_unit) where name is the string name of
    the timed variable used when encoding smv timed fsm, and time_unit
    is the `slim_compiler.src.dataStructure.expressions.TimeUnitType`
    enumeration value of the time unit used to encode time
    """
    smv_conv = converter.SmvConverter()
    return (smv_conv.time_var(), smv_conv.time_var_tu())


def get_smv_delta_var():
    """Returns a pair (name, time_unit) where name is the string name of
    the time-delta variable used when encoding smv timed fsm, and time_unit
    is the `slim_compiler.src.dataStructure.expressions.TimeUnitType`
    enumeration value of the time unit used to encode time
    """
    smv_conv = converter.SmvConverter()
    return (smv_conv.delta_var(), smv_conv.time_var_tu())


def dump_error_event_probability_map(root_impl, prob_file):

    """Dumps to the given file the map between error events and the
    corresponding probability"""
    from slim_compiler.src.smv.walkers import ErrorEventProbWalker
    smvconvert = converter.SmvConverter()
    smvconvert.set_id_prefix(smvconvert.root_var() + ".")

    walker = ErrorEventProbWalker(smvconvert)
    walker.walk(root_impl)
    for ev, prob in walker.result:
        prob_file.write("%s = TRUE : %s;%s" % (ev, prob, os.linesep))

    return


def get_smv_language(root_impl):
    """Given an implementation, returns a map { smv_var_name : smv_var_type }
       where smv_var_name is a data port/components or variable for
       component mode, and smv_var_type is the type of the variable in
       smv syntax."""
    from slim_compiler.src.smv.walkers import SmvDataTypeWalker
    smvconvert = converter.SmvConverter()
    smvconvert.set_id_prefix(smvconvert.root_var() + ".")

    walker = SmvDataTypeWalker(smvconvert)
    walker.walk(root_impl)
    return walker.result


def get_observables(root_impl):
    """Returns the list of observables."""
    from slim_compiler.src.smv.walkers import \
        FailureObservationAlarmControllableWalker

    smvconvert = converter.SmvConverter()
    smvconvert.set_id_prefix(smvconvert.root_var() + ".")
    walker = FailureObservationAlarmControllableWalker(smvconvert)
    walker.walk(root_impl)
    return walker.observables


def get_observations(root_impl):
    """Calls the walker on the original SLIM model, retrieves then
    all the variables marked with 'observable' and returns them."""
    from slim_compiler.src.smv.walkers import \
        FailureObservationAlarmControllableWalker

    smvconvert = converter.SmvConverter()
    smvconvert.set_id_prefix(smvconvert.root_var() + ".")
    walker = FailureObservationAlarmControllableWalker(smvconvert)
    walker.walk(root_impl)
    return walker.get_observables()


def dump_failures_observations(root_impl, file):
    """Dumps to the given file XML text with a list of failures
    and observations. This file is required by FSAP.
    Returns a pair with the number of failures and observables."""
    from slim_compiler.src.smv.walkers import \
        FailureObservationAlarmControllableWalker

    smvconvert = converter.SmvConverter()
    smvconvert.set_id_prefix(smvconvert.root_var() + ".")
    walker = FailureObservationAlarmControllableWalker(smvconvert)
    walker.walk(root_impl)
    walker.dump_in_file_obs(file)
    return (walker.get_failures_num(), walker.get_observables_num())


def get_alarms(root_impl):
    """Returns the list of alarms."""
    from slim_compiler.src.smv.walkers import \
        FailureObservationAlarmControllableWalker

    smvconvert = converter.SmvConverter()
    smvconvert.set_id_prefix(smvconvert.root_var() + ".")
    walker = FailureObservationAlarmControllableWalker(smvconvert)
    walker.walk(root_impl)
    return walker.alarms


def get_alarmations(root_impl):
    """Calls the walker on the original SLIM model, retrieves then
    all the variables marked with 'alarm' and returns them."""
    from slim_compiler.src.smv.walkers import \
        FailureObservationAlarmControllableWalker

    smvconvert = converter.SmvConverter()
    smvconvert.set_id_prefix(smvconvert.root_var() + ".")
    walker = FailureObservationAlarmControllableWalker(smvconvert)
    walker.walk(root_impl)
    return walker.get_alarms()


def dump_failures_alarms(root_impl, file):
    """Dumps to the given file XML text with a list of failures
    and alarms. This file is required by FSAP.
    Returns a pair with the number of failures and alarms."""
    from slim_compiler.src.smv.walkers import \
        FailureObservationAlarmControllableWalker

    smvconvert = converter.SmvConverter()
    smvconvert.set_id_prefix(smvconvert.root_var() + ".")
    walker = FailureObservationAlarmControllableWalker(smvconvert)
    walker.walk(root_impl)
    walker.dump_in_file_alm(file)
    return (walker.get_failures_num(), walker.get_alarms_num())


def dump_failures_observations_alarms(root_impl, file):
    """Dumps to the given file XML text with a list of failures,
    observations and alarms. This file is required by FSAP.
    Returns a pair with the number of failures, observables and alarms."""
    # from slim_compiler.src.smv.walkers import FailureObservationAlarmWalker
    from slim_compiler.src.smv.walkers import \
        FailureObservationAlarmControllableWalker

    smvconvert = converter.SmvConverter()
    smvconvert.set_id_prefix(smvconvert.root_var() + ".")
    walker = FailureObservationAlarmControllableWalker(smvconvert)
    walker.walk(root_impl)
    walker.dump_in_file(file)
    return (walker.get_failures_num(), walker.get_observables_num(),
            walker.get_alarms_num())


def get_top_tau_event():
    """Returns an SMV identifier of the top-level tau event"""
    smvconvert = converter.SmvConverter()
    smvconvert.set_id_prefix("")
    return smvconvert.event_name(smvconvert.tau_event())


def convert_smv_id_to_slim(name, can_be_enum_literal=True):
    """Takes an identifier SMV and returns its original SLIM form.
    If an identifier has been introduced artificially during
    slim2smv conversion then None is returned.
     If can_be_enum_literal is False, name is expected to be the
     name of a variable or define (must begin with root name)"""
    smvconvert = converter.SmvConverter()
    return smvconvert.smv_id_to_slim_id(name, can_be_enum_literal)


def convert_smv_const_to_slim(const):
    """Takes the string of a SMV constant (an identifier, or a
    number) and returns its original SLIM form. If there is no
    correspondence in SLIM, None is returned"""
    smvconvert = converter.SmvConverter()
    return smvconvert.smv_const_to_slim_const(const)
