# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


import copy

# API for slim2smv process
from slim_compiler.src.dataStructure import nominalbehaviour, expressions

# ----------------------------------------------------------------------
#   High level function that can be used to easily navigate components
# ----------------------------------------------------------------------


def get_assigned_odp(impl):
    """Returns the set of assigned output data ports (as names)"""
    res = set()
    _type = impl.get_referenced_component_type()

    for trans in impl.normal_mode_transitions:
        for name in get_trans_assigns(impl, trans, False).keys():
            if name in _type.out_data_ports:
                res.add(name)

    return res


def get_trans_assigns(impl, trans, complete=True):
    """Returns a map { name : value } where the key is the name
    (string) of data subcomponent (DataSubcomponent) or the output
    data port (OutDataPort) assigned in the given ModeTransition,
    and value is the expression of the value that is being
    assigned.  If complete is True (default), all the data
    subcomponents and output data ports declared within impl will
    occur in the map, even if not explicitly assigned in the given
    mode transition. In this case the value is the initial value of
    the subcomponent (if any is declared) or else the identifier
    expression (the name) of the subcomponent itself."""

    comp_type = impl.get_referenced_component_type()

    result = (get_assignable_data(impl)
              if complete   # default for unspecified assignments
              else {})

    if isinstance(trans.label,
                  (nominalbehaviour.ModeTransitionLabel,
                   nominalbehaviour.ReactivationModeTransitionLabel)):
        # update the result with the assignments in the mode transition
        for e in trans.label.effects:
            assert expressions.OperatorType.ASSIGN == e.operator
            # retrieves the port or the subcompoenent from the name
            assert isinstance(e.left, expressions.IdentifierTreeNode)
            if e.left.name in comp_type.out_data_ports:
                k = e.left.name
            else:
                # otherwise it must be a data subcomponent
                assert e.left.name in impl.data_subcomponents
                k = e.left.name

            result[k] = e.right

    return result


def get_assignable_data(impl):
    """Returns a map { name : value } where the key is the name (string)
    of data subcomponent (DataSubcomponent) or the output data port
    (OutDataPort) declared within impl. The value is the initial value
    of the subcomponent (if any is declared) or else the identifier
    expression (the name) of the subcomponent itself.
    """

    result = {}
    comp_type = impl.get_referenced_component_type()

    # collects in data all out data ports and data components
    data = comp_type.out_data_ports.copy()
    data.update(impl.data_subcomponents)

    for k in sorted(data):
        d = data[k]
        if d.default_value:
            result[k] = d.default_value
        else:
            _id = expressions.IdentifierTreeNode(d.name, -1, -1)
            _id.node_type = _data_type_to_node_type(d.data_type)
            result[k] = _id

    return result


__trans_id_cache = {}


def get_trans_id(impl, trans):
    """Returns a unique identifier (a string) associated with a given
    ModeTransition.
    """

    idx = -1
    cache = None
    if impl in __trans_id_cache:
        cache = __trans_id_cache[impl]
    else:
        cache = {}
        for num, t in enumerate(impl.normal_mode_transitions):
            cache[t] = num
        __trans_id_cache[impl] = cache

    if trans in cache:
        idx = cache[trans]
    else:
        idx = impl.normal_mode_transitions.index(trans)
        cache[trans] = idx

    # index of given transition
    # idx = impl.normal_mode_transitions.index(trans)
    tau = ("tau_"
           if not trans.label.trigger
           else "")
    return "%s_%s_%s%d" % (trans.source_mode, trans.target_mode, tau, idx)


def get_trans_with_event(impl, event_name):
    """Returns the set of ModeTransition which are triggered by or
    trigger a given event. event_name is the event name to be
    searched among transitions of component implementation
    impl. To use tau, see function get_trans_with_tau"""
    return [t for t in impl.normal_mode_transitions
            if (isinstance(t.label, nominalbehaviour.ModeTransitionLabel) and
                event_name == t.label.trigger)]


def get_trans_with_tau(impl):
    """Returns the set of ModeTransition which are tau transition, i.e.
    which do not triggered by or trigger an in/out event."""
    return [t for t in impl.normal_mode_transitions
            if (isinstance(t.label, nominalbehaviour.ModeTransitionLabel) and
                not t.label.trigger)]


def get_constraints_list_from_invariants(invariants):
    """Returns a sequence of expressions, each being an invariant
    (not a trajectory)"""
    res = []
    queue = []
    queue.append(invariants)

    while len(queue) > 0:
        expr = queue.pop()
        if expr is None:
            continue
        if (isinstance(expr, expressions.BinaryOperatorTreeNode) and
                expressions.OperatorType.AND == expr.operator):
            queue += [expr.left, expr.right]

        elif not isinstance(expr, expressions.TrajectoryEquationTreeNode):
            res.append(expr)

    return res


def get_trajectories_list_from_invariants(invariants):
    """Returns a sequence of expressions, each being a trajectory
    (not a constraint). Returned expressions are
    TrajectoryEquationTreeNode"""
    res = []
    queue = []
    queue.append(invariants)

    while len(queue) > 0:
        expr = queue.pop()
        if expr is None:
            continue
        if (isinstance(expr, expressions.BinaryOperatorTreeNode) and
                expressions.OperatorType.AND == expr.operator):
            queue += [expr.left, expr.right]

        elif isinstance(expr, expressions.TrajectoryEquationTreeNode):
            res.append(expr)

    return res


def get_trajectories_map_from_invariants(invariants):
    """Returns a map {dsc_name : expr }, each being a trajectory
    (not a constraint)"""

    res = {}
    for traj in get_trajectories_list_from_invariants(invariants):
        assert isinstance(traj, expressions.TrajectoryEquationTreeNode)
        name = traj.variable.name
        assert not res.has_key(name), \
            "Not multiple derivatives for the same var"
        res[name] = traj

    return res


def get_connections_with_source_event(impl, event_name):
    """Returns the set of connections PortConnection which have
    event_name as source event port."""
    return [c for c in impl.event_port_connections
            if event_name == c.source_port]


def get_connections_with_target_event(impl, event_name):
    """Returns the set of connections PortConnection which have
    event_name as target event port."""
    return [c for c in impl.event_port_connections
            if event_name == c.target_port]


def get_connections_with_event(impl, event_name):
    """Returns the set of connections PortConnection which have
    event_name as source or target event port."""
    return [c for c in impl.event_port_connections
            if event_name in (c.source_port, c.target_port)]


def get_connections_with_events(impl, sevent_name, tevent_name):
    """Given a source event port and a target event port, returns the set of
    connections PortConnection that have those events as source and target
    respectively."""
    return [c for c in impl.event_port_connections
            if (sevent_name == c.source_port
                and tevent_name == c.target_port)]


def get_connections_with_source_data(impl, data_name):
    """Returns the set of connections PortConnection which have
    data_name as source data port."""
    return [c for c in impl.data_port_connections
            if data_name == c.source_port]


def get_connections_with_target_data(impl, data_name):
    """Returns the set of connections PortConnection which have
    data_name as target data port."""
    return [c for c in impl.data_port_connections
            if data_name == c.target_port]


def get_flows_with_target_data(impl, data_name):
    """Returns the set of Flow instances which have
    data_name as target data port."""
    res = []
    for flow in impl.flows:
        assert (isinstance(flow.expression,
                           expressions.BinaryOperatorTreeNode) and
                isinstance(flow.expression.left,
                           expressions.IdentifierTreeNode)), \
            "Flow has unexpected structure"

        if flow.expression.left.name == data_name:
            res.append(flow)

    return res


def get_connections_with_subcomponent_connected_event(impl, event_name):
    """Returns the set of connections which connect
    given event with sub-component events, i.e. (ev, sc.evi) or (sc.evi, ev)"""
    return ([c for c in get_connections_with_source_event(impl, event_name)
             if get_subcomponent_from_port(impl, c.target_port)]
            +
            [c for c in get_connections_with_target_event(impl, event_name)
             if get_subcomponent_from_port(impl, c.source_port)])


def get_connection_modes(impl, connection):
    """Given a PortConnection instance, returns the set
    of modes (string) in which the connection is active."""
    result = set(connection.in_modes)
    if not result:
        result = get_component_impl_modes(impl)
    return result


def get_flow_modes(impl, flow):
    """Given a Flow instance, returns the set
    of modes (string) in which the flow is active."""
    result = set(flow.in_modes)
    if not result:
        result = get_component_impl_modes(impl)
    return result


def get_connections_modes(impl, connections):
    """Given a set of PortConnection connections, returns the set
    of modes (string) in which at least one of those connections is
    active. Returned set can be empty if connections is empty"""
    result = set()
    for c in connections:
        result |= get_connection_modes(impl, c)
    return result


def get_subcomponent_modes(impl, sc):
    """Returns the set of modes (string) which the given
    subcomponent (NonDataSubcomponent or DataSubcomponent) is
    declared to be active in. If the subcomponent is always active,
    then the set of modes of the containing component is returned."""
    if sc.in_modes:
        return set(sc.in_modes)
    return get_component_impl_modes(impl)


def get_component_impl_modes(impl):
    """Returns the set of modes (strings) in which the given
    component implementation is declared to be active. If the
    component has no modes specified, the default mode name is
    returned as a singleton."""
    if impl.modes:
        return set(impl.modes.keys())
    return set([get_default_mode_name()])


def get_out_to_out_connections_with_source(impl, sc_out_event):
    """Gets connections (sc.OUT,OUT) given sc.OUT"""

    assert get_subcomponent_from_port(impl, sc_out_event)

    impl_type = impl.get_referenced_component_type()

    return [c
            for c in get_connections_with_source_event(impl,
                                                       sc_out_event)
            if c.target_port in impl_type.out_event_ports]


def get_out_to_out_connections_with_target(impl, out_event):
    """Gets connections (sc.OUT,OUT) given OUT"""

    impl_type = impl.get_referenced_component_type()

    assert out_event in impl_type.out_event_ports

    return [c
            for c in get_connections_with_target_event(impl, out_event)
            if get_subcomponent_from_port(impl, c.source_port)]


def get_in_to_in_connections_with_source(impl, in_event):
    """Gets connections (IN,sc.IN) given IN"""

    impl_type = impl.get_referenced_component_type()

    assert in_event in impl_type.in_event_ports

    return [c
            for c in get_connections_with_source_event(impl, in_event)
            if get_subcomponent_from_port(impl, c.target_port)]


def get_in_to_in_connections_with_target(impl, sc_in_event):
    """Gets connections (IN,sc.IN) given sc.IN"""

    assert get_subcomponent_from_port(impl, sc_in_event)

    impl_type = impl.get_referenced_component_type()

    return [c
            for c in get_connections_with_target_event(impl,
                                                       sc_in_event)
            if c.source_port in impl_type.in_event_ports]


def get_out_to_in_connections_with_source(impl, sc_out_event):
    """Gets connections (sc.OUT,sc.IN) given sc.OUT"""

    assert get_subcomponent_from_port(impl, sc_out_event)

    return [c
            for c in get_connections_with_source_event(impl, sc_out_event)
            if get_subcomponent_from_port(impl, c.target_port)]


def get_out_to_in_connections_with_target(impl, sc_in_event):
    """Gets connections (sc.OUT,sc.IN) given sc.IN"""

    assert get_subcomponent_from_port(impl, sc_in_event)

    return [c
            for c in get_connections_with_target_event(impl, sc_in_event)
            if get_subcomponent_from_port(impl, c.source_port)]


def get_out_to_out_connections_with_source_data(impl, sc_out_data):
    """Gets connections (sc.OUT,OUT) given sc.OUT"""

    assert get_subcomponent_from_port(impl, sc_out_data)

    impl_type = impl.get_referenced_component_type()

    return [c
            for c in get_connections_with_source_data(impl,
                                                      sc_out_data)
            if c.target_port in impl_type.out_data_ports]


def get_out_to_out_connections_with_target_data(impl, out_data):
    """Gets connections (sc.OUT,OUT) given OUT"""

    impl_type = impl.get_referenced_component_type()

    assert out_data in impl_type.out_data_ports

    return [c
            for c in get_connections_with_target_data(impl, out_data)
            if get_subcomponent_from_port(impl, c.source_port)]


def get_in_to_in_connections_with_source_data(impl, in_data):
    """Gets connections (IN,sc.IN) given IN"""

    impl_type = impl.get_referenced_component_type()

    assert in_data in impl_type.in_data_ports

    return [c
            for c in get_connections_with_source_data(impl, in_data)
            if get_subcomponent_from_port(impl, c.target_port)]


def get_in_to_in_connections_with_target_data(impl, sc_in_data):
    """Gets connections (IN,sc.IN) given sc.IN"""

    assert get_subcomponent_from_port(impl, sc_in_data)

    impl_type = impl.get_referenced_component_type()

    return [c
            for c in get_connections_with_target_data(impl,
                                                      sc_in_data)
            if c.source_port in impl_type.in_data_ports]


def get_out_to_in_connections_with_source_data(impl, sc_out_data):
    """Gets connections (sc.OUT,sc.IN) given sc.OUT"""

    assert get_subcomponent_from_port(impl, sc_out_data)

    return [c
            for c in get_connections_with_source_data(impl, sc_out_data)
            if get_subcomponent_from_port(impl, c.target_port)]


def get_out_to_in_connections_with_target_data(impl, sc_in_data):
    """Gets connections (sc.OUT,sc.IN) given sc.IN"""

    assert get_subcomponent_from_port(impl, sc_in_data)

    return [c
            for c in get_connections_with_target_data(impl, sc_in_data)
            if get_subcomponent_from_port(impl, c.source_port)]


def get_subcomponent_events_with_event(impl, event_name):
    """Returns the set of sub-component events such that there
    exists connection (ev, sc.evi) or (sc.evi, ev)"""
    return ([c.target_port
             for c in get_connections_with_source_event(impl, event_name)
             if get_subcomponent_from_port(impl, c.target_port)]
            +
            [c.source_port
             for c in get_connections_with_target_event(impl, event_name)
             if get_subcomponent_from_port(impl, c.source_port)])


def get_subcomponent_from_port(impl, port_name):
    """Returns a NonDataSubcomponent instance of the subcomponent
    that port_name refers to. port_name can be either an event or
    data port. If port_name does not refer a subcomponent, then
    None is returned. If port_name refers a non-existing
    subcomponent a IndexError exception is raised."""
    ndsc = None
    sc, ev = split_port(port_name)
    if sc:
        ndsc = get_non_data_subcomponent_from_name(impl, sc)
    return ndsc


def get_subcomponent_impl_from_port(impl, port_name):
    """Returns a ComponentImplementation of the subcomponent that
    port_name refers to. port_name can be either an event or data
    port. If port_name does not refer a subcomponent, then None is
    returned. If port_name refers a non-existing subcomponent a
    KeyError exception is raised."""
    sci = None
    sc, ev = split_port(port_name)
    if sc:
        sci = get_subcomponent_impl_from_name(impl, sc)
    return sci


def get_non_data_subcomponent_from_name(impl, sc_name):
    """Given a subcomponent instance name, returns the
    corresponding subcomponent (NonDataSubcomponent). If sc_name
    refers to a non-existing subcomponent, a KeyError exception is
    raised.  sc_name can be dotted notation (e.g. "a.b.c") in which
    case the NonDataSubcomponent representing the right-most
    non-data subcomponent is returned (NonDataSubcomponent of "c"
    in the example).
    """

    scs = sc_name.split(".")
    assert len(scs) > 0

    while True:
        ndsc = impl.non_data_subcomponents[scs[0]]
        if len(scs) == 1:
            return ndsc
        impl = ndsc.get_referenced_component_implementation()
        del scs[0]

    assert False, "This place cannot be reached"
    return


def get_subcomponent_impl_from_name(impl, sc_name):
    """Given a subcomponent instance name, returns the
    corresponding subcomponent implementation. If sc_name refers to
    a non-existing subcomponent, a KeyError exception is
    raised. If sc_name is the empty string, impl is returned."""

    if sc_name == "":
        return impl
    ndsc = get_non_data_subcomponent_from_name(impl, sc_name)
    return ndsc.get_referenced_component_implementation()


def has_history(impl):
    """Returns True if given implementation does not have an
    activation mode"""
    for k, m in impl.modes.iteritems():
        if nominalbehaviour.ModeType.ACTIVATION == m.type:
            return False
        if nominalbehaviour.ModeType.INITIAL == m.type:
            return True
    return True


def get_initial_or_activation_mode(impl):
    """Returns the name (string) of the mode that is marked as
    either initial or activation. An assertion occurs if
    not found."""
    for k, m in impl.modes.iteritems():
        if m.type in (nominalbehaviour.ModeType.ACTIVATION,
                      nominalbehaviour.ModeType.INITIAL):
            return k

    if not impl.modes:
        return get_default_mode_name()
    assert False, "Component %s has no activation or initial state" % impl.name
    return


def is_input_event(impl, event_name):
    """Returns True if given event name is an input event. Event
    can be a component's event or a sub-component's event"""
    sc, ev = split_port(event_name)
    ci = (get_subcomponent_impl_from_name(impl, sc)
          if sc
          else impl)

    # Access the component type to check the port
    ct = ci.get_referenced_component_type()
    return ev in ct.in_event_ports


def is_output_event(impl, event_name):
    """Returns True if given event name is an output event. Event
    can be a component's event or a sub-component's event"""
    sc, ev = split_port(event_name)
    ci = (get_subcomponent_impl_from_name(impl, sc)
          if sc
          else impl)

    # Access the component type to check the port
    ct = ci.get_referenced_component_type()
    return ev in ct.out_event_ports


def get_input_event(impl, event_name):
    """Returns the input port given the event name. Event
    can be a component's event or a sub-component's event"""
    sc, ev = split_port(event_name)

    ci = (get_subcomponent_impl_from_name(impl, sc)
          if sc
          else impl)

    # Access the component type to check the port
    ct = ci.get_referenced_component_type()
    return ct.in_event_ports[ev]


def get_output_event(impl, event_name):
    """Returns the output port given the event name. Event
    can be a component's event or a sub-component's event"""
    sc, ev = split_port(event_name)
    ci = (get_subcomponent_impl_from_name(impl, sc)
          if sc
          else impl)

    # Access the component type to check the port
    ct = ci.get_referenced_component_type()
    return ct.out_event_ports[ev]


def are_modes_complete(impl, modes):
    """Returns True if the given set of modes (string) includes all
    the possible modes of the given component"""
    return get_component_impl_modes(impl) <= modes


def get_data_subcomponents_in_expr(impl, expr):
    """Given an expression, searches and returns the set of data
    subcomponents (DataSubcomponent) occurring within it. This can be
    used to extract the language of an expression. The expression's
    context is the given component implementation."""
    nodes = map(lambda id: __expr_id_to_data(impl, id),
                __expr_search(expr,
                              search_types=(expressions.IdentifierTreeNode,)))
    return set(filter(lambda n:
                      isinstance(n, nominalbehaviour.DataSubcomponent),
                      nodes))


def get_continuous_data_subcomponent(impl):
    """Returns the set of continuos data sc in impl as names"""
    return set((n for n, dsc in
                filter(lambda ni: isinstance(ni[1].data_type,
                                             (nominalbehaviour.Clock,
                                              nominalbehaviour.Continuous)),
                       impl.data_subcomponents.iteritems())))


def get_discrete_data_subcomponents(impl):
    """Returns the set of discrete data subcompoenents"""
    return set(n for n, dsc in
               filter(lambda ni: not isinstance(ni[1].data_type,
                                                (nominalbehaviour.Clock,
                                                 nominalbehaviour.Continuous)),
                      impl.data_subcomponents.iteritems()))


def get_continuos_trajectories(impl, mode_name):
    """Given an implementation and a mode, returns a map {dsc:expr}
    where dsc is the name of DataSubCompoenent, and expr can be:
    1. the TrajectoryEquationTreeNode of dsc trajectory (if there is any and
    the dsc is active in that mode)
    2. the dsc (as an ID expression) if the dsc does not change.
    3. None for clocks

    Notice that clocks are incremented (if active in mode) as well
    (also if not occurring in the mode invars."""

    res = {}

    for name in sorted(get_continuous_data_subcomponent(impl)):
        dsc = impl.data_subcomponents[name]
        expr = None

        # is it not active in mode?
        if mode_name not in get_subcomponent_modes(impl, dsc):
            expr = expressions.IdentifierTreeNode(name, -1, -1)
            expr.node_type = _data_type_to_node_type(dsc.data_type)
        else:
            # here dsc is active
            traj = get_trajectories_map_from_invariants(
                impl.modes[mode_name].invariant)
            if name in traj:
                # explicitly assigned in trajectory
                expr = traj[name]
            else:
                # not assigned. Is it a clock?
                if isinstance(dsc.data_type, nominalbehaviour.Clock):
                    expr = None
                else:
                    # not changed: this is is not changed
                    expr = expressions.IdentifierTreeNode(name, -1, -1)
                    expr.node_type = _data_type_to_node_type(dsc.data_type)

        assert name not in res, "No multiple assignments in trajectories"
        res[name] = expr

    return res


def get_data_ports_in_expr(impl, expr):
    """Given an expression, searches and returns the set of data output ports
    (OutDataPort) occurring within it. This can be
    used to extract the language of an expression. The expression's
    context is the given component implementation."""
    nodes = map(lambda id: __expr_id_to_data(impl, id),
                __expr_search(expr,
                              search_types=(expressions.IdentifierTreeNode,)))
    return set(filter(lambda n: isinstance(n, nominalbehaviour.DataPort),
                      nodes))


def substitute_ids_in_expr(expr, mapping):
    """Returns a new expression which is created by substituting
    all the identifiers found in expr and in mapping with the
    corresponding expressions given by the mapping.
    mapping is a map name -> expression like
    the one returned by get_trans_assigns."""
    leaf_types = (expressions.BooleanTreeNode,
                  expressions.NumberTreeNode,
                  expressions.IdentifierTreeNode,)

    unsupported_types = (
        #expressions.TrajectoryEquationTreeNode,  # THIS is not supported
        expressions.TrajectoryExpressionTreeNode, )

    return __subst_ids_in_expr(expr, mapping, leaf_types, unsupported_types)


def get_compile_information_from_impl(impl):
    """Given a component implementation, returns the corresponding
    CompileInformation it is contained into"""
    return impl.subpackage.package.compile_information


# ----------------------------------------------------------------------
#        Support for names
# ----------------------------------------------------------------------
def get_default_mode_name():
    """When a component implementation has no modes, a default mode
    is created. The returned string is the name of that default
    mode"""
    return "#mode"  # this is tricky as slim syntax does not allow
                    # for names prefixed with a #


def split_port(port_name):
    """Splits a port name into a pair <sc, data>. sc can be None"""
    assert port_name.count(".") <= 1  # at most one dot is allowed
    return ([None]+port_name.split("."))[-2:]

# ----------------------------------------------------------------------
#        Internal services
# ----------------------------------------------------------------------


def __expr_search(expr, search_nodes=[], search_types=[]):
    """This is an internal function used to search for either nodes
    or nodes types within an expression. Returns the set of found nodes
    """
    leaf_types = (expressions.BooleanTreeNode,
                  expressions.NumberTreeNode,
                  expressions.IdentifierTreeNode,)

    unsupported_types = (expressions.TrajectoryEquationTreeNode,
                         expressions.TrajectoryExpressionTreeNode,)

    result = set()
    __expr_dfs(expr, search_nodes, search_types,
               leaf_types, unsupported_types,
               lambda node, res: res.add(node), result)

    return result


def __expr_id_to_data(impl, expr):
    """Given an IdentifierTreeNode, returns the corresponding
    DataComponent or Port it refers within given component
    implementation impl. KeyError is raised if expr is found within
    the component."""
    assert isinstance(expr, expressions.IdentifierTreeNode)

    # search among ports
    ctype = impl.get_referenced_component_type()
    for cm in (ctype.in_data_ports,
               ctype.out_data_ports,
               ctype.in_event_ports,
               ctype.out_event_ports, ):
        if expr.name in cm:
            return cm[expr.name]

    # search among subcomponents
    return impl.data_subcomponents[expr.name]


def __subst_ids_in_expr(expr, mapping, leaf_types, unsupported_types):
    """Private service of substitute_ids_in_expr. Mapping is a
    str:expr dictionary"""

    # leaves for base case
    if expr is None:
        return None

    # not supported:
    if isinstance(expr, unsupported_types):
        raise NotImplementedError("Explicitly unsupported node "
                                  "type in expression: %s" % expr)

    if (isinstance(expr, expressions.IdentifierTreeNode) and
            expr.name in mapping):
        return mapping[expr.name]  # apply substitution

    # generic leaves handling
    if isinstance(expr, leaf_types):
        return __expr_copy(expr)

    # induction step
    assert isinstance(expr, expressions.ExpressionTreeNode), expr
    expr_copy = __expr_copy(expr)

    # specific cases have to be handled here

    # NOTE: trajectories are kept as they are.
    if isinstance(expr_copy, expressions.TrajectoryEquationTreeNode):
        return __expr_copy(expr)

    # here generic cases are handled (unary and bynary operators)
    assert isinstance(expr_copy, expressions.OperatorTreeNode), expr_copy
    if isinstance(expr_copy, expressions.UnaryOperatorTreeNode):
        expr_copy.subexpression = __subst_ids_in_expr(expr.subexpression,
                                                      mapping,
                                                      leaf_types,
                                                      unsupported_types)
    elif isinstance(expr_copy, expressions.BinaryOperatorTreeNode):
        expr_copy.left = __subst_ids_in_expr(expr.left, mapping,
                                             leaf_types, unsupported_types)
        expr_copy.right = __subst_ids_in_expr(expr.right, mapping,
                                              leaf_types, unsupported_types)
    else:
        raise NotImplementedError("Unsupported  node type in expression: %s"
                                  % expr_copy)

    return expr_copy


def __expr_copy(expr):
    """Returns a copy of given node. No recursion is applied"""
    return copy.copy(expr)  # shallow copy only here


# ------------------------------------------------------------
# This function is the generalization of DFS in a binary tree
# For the sake of efficiency, the work is done within this
# internal recursive function.
def __expr_dfs(expr, search_nodes, search_types,
               leaf_types, unsupported_types,
               action, action_arg):
    """action is a function(node, arg) that is invoked when a node
    or node type matches"""

    # action can be applied
    if expr in search_nodes:
        action(expr, action_arg)
    if isinstance(expr, search_types):
        action(expr, action_arg)

    # not supported:
    if isinstance(expr, unsupported_types):
        raise NotImplementedError("Unsupported node type in expression: %s"
                                  % expr)

    # leaves for base case
    if expr is None or isinstance(expr, leaf_types):
        return

    # induction case
    assert isinstance(expr, expressions.ExpressionTreeNode), expr
    __expr_dfs(expr.left, search_nodes, search_types,
               leaf_types, unsupported_types, action, action_arg)
    __expr_dfs(expr.right, search_nodes, search_types,
               leaf_types, unsupported_types, action, action_arg)
    return
# ------------------------------------------------------------


# ----------------------------------------------------------------------
#        Code that might be moved to compile-information
# ----------------------------------------------------------------------
def get_root_components(cinfo):
    """Returns the set of component implementations which are the
    top-level components within the cinfo, i.e. component
    implementations which are NOT used to declare sub-components."""

    non_roots = set()  # set of non-roots
    candidates = set()  # set of possible (temporary) roots

    cts = get_component_types_with_impl(cinfo)
    for ct, cis in cts.iteritems():
        for ci in cis:
            if ci not in non_roots:
                non_roots |= get_all_children_implementations(ci)
                candidates.add(ci)

    # only those candidates which are not non-roots are actual roots
    return candidates - non_roots


def get_all_children_implementations(impl):
    """Given a component implementation, returns the set of all
    control sub-components implementation, including
    sub-components of sub-components"""
    result = set()
    for k, sc2 in [(k, sc.get_referenced_component_implementation())
                   for k, sc in impl.non_data_subcomponents.iteritems()]:
        result |= set([sc2]) | get_all_children_implementations(sc2)

    return result


def get_component_types_with_impl(cinfo):
    """Returns all the component types and associated
    implementations found within the compileinfo
    structure. Returned map has component types as keys, and the
    list of associated implementations as values."""

    result = {}
    for k, p in cinfo.packages.iteritems():
        psp = p.get_public_subpackage()
        for k, pt in psp.types.iteritems():
            result[pt] = p.get_implementations_for_type(k, True)

    return result


def get_implementation_from_name(cinfo, impl_name):
    """Searches among all component implementations the component
    with given name, and returns it. Returns None if not found"""
    for k, p in cinfo.packages.iteritems():
        psp = p.get_public_subpackage()
        if impl_name in psp.implementations:
            return psp.implementations[impl_name]

    return None  # not found


def get_error_component_types_with_impl(cinfo):
    """Returns all the error component types and associated error
    implementations found within the compileinfo
    structure. Returned map has component error types as keys, and
    the list of associated error implementations as values."""

    result = {}
    for k, p in cinfo.packages.iteritems():
        psp = p.get_public_subpackage()
        for k, pt in psp.error_types.iteritems():
            errs = p.get_error_implementations_for_type(k, True)
            result[pt] = errs if errs is not None else []

    return result


def get_subcomponent_from_path(impl, fully_qualified_name):

    """Given a root implementation, returns the corresponding sub
    component (either data or non-data), which can be reached from it
    through given path. ValueError is raised if the path cannot be
    found in the hierarchy of impl. fully_qualified_name is a string
    in the form name1.name2.[...].sc_name
    """

    res = None
    path = fully_qualified_name.split(".")

    curr_impl = impl
    while path:
        curr_sc = path.pop(0)
        if curr_sc in curr_impl.non_data_subcomponents:
            res = curr_impl.non_data_subcomponents[curr_sc]
            curr_impl = res.get_referenced_component_implementation()
        elif curr_sc in curr_impl.data_subcomponents:
            if path:
                raise ValueError("Invalid path: " + fully_qualified_name)
            res = curr_impl.data_subcomponents[curr_sc]
        else:
            raise ValueError("Invalid path: " +
                             fully_qualified_name +
                             " does not refer a subcomponent: " +
                             curr_sc)

    if res is None:
        ValueError("Invalid empty path")

    return res


def _data_type_to_node_type(data_type):
    """Converts given AbstractDataType to the corresponding
    expressions.NodeType. This can be used to create for example
    identifier expressions out of data components and ports"""

    assert isinstance(data_type, nominalbehaviour.AbstractDataType)
    return {nominalbehaviour.Int: expressions.NodeType.INT,
            nominalbehaviour.IntRange: expressions.NodeType.INT_RANGE,
            nominalbehaviour.Real: expressions.NodeType.REAL,
            nominalbehaviour.Bool: expressions.NodeType.BOOL,
            nominalbehaviour.Continuous: expressions.NodeType.CONTINUOUS,
            nominalbehaviour.Clock: expressions.NodeType.CLOCK,
            nominalbehaviour.Enum: expressions.NodeType.ENUM_FIELD,
            }[data_type.__class__]


def f7(seq):
    seen = set()
    seen_add = seen.add
    return [x for x in seq if x not in seen and not seen_add(x)]


__tu_operators = frozenset((expressions.OperatorType.MSEC,
                            expressions.OperatorType.REAL_TOMSEC,
                            expressions.OperatorType.MSEC_TOREAL,

                            expressions.OperatorType.SEC,
                            expressions.OperatorType.REAL_TOSEC,
                            expressions.OperatorType.SEC_TOREAL,

                            expressions.OperatorType.MIN,
                            expressions.OperatorType.REAL_TOMIN,
                            expressions.OperatorType.MIN_TOREAL,

                            expressions.OperatorType.HOUR,
                            expressions.OperatorType.REAL_TOHOUR,
                            expressions.OperatorType.HOUR_TOREAL,

                            expressions.OperatorType.DAY,
                            expressions.OperatorType.REAL_TODAY,
                            expressions.OperatorType.DAY_TOREAL, ))


def is_expression_timed_unit_op(expr):
    """Returns True iff given expression is a time-unit operator"""
    return (isinstance(expr, expressions.UnaryOperatorTreeNode) and
            expr.operator in __tu_operators)
