# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================

import sys, os, math, re, numbers
from decimal import Decimal

from property_management.src.api import get_atomic_proposition_in_smv_format as get_smv_format
from utils.main_extension_converter import MainExtensionConverter

################################################################# SMV CODE STUBS

def get_discrepancy_node_validation_props (time_trans_ivar_name):
    return """

    -- a node shouldn't be able to fire if it is already active
    INVARSPEC is_active -> (!must_fire & !can_fire);

    -- nodes should stay active once activated
    INVARSPEC is_active -> next(is_active);

    -- can- and must-firing signals must be mutually exclusive
    INVARSPEC (can_fire -> !must_fire) & (must_fire -> !can_fire);

    -- a node activation that must occur, must indeed occur
    INVARSPEC (instant_trans_type = NODE_ACTIVATION & must_fire) ->
        (!is_active & next(is_active));

    -- if a node activation occurs, it must be allowed or required to occur
    INVARSPEC (!is_active & next(is_active)) ->
        (instant_trans_type = NODE_ACTIVATION & (must_fire | can_fire));

    -- a node activation that cannot occur, must not occur
    INVARSPEC (!can_fire & !must_fire) -> is_active = next(is_active);

    -- if the current transition is not a node activation, the status
    -- of the node must not change
    INVARSPEC instant_trans_type != NODE_ACTIVATION -> is_active = next(is_active);

    -- time tick transitions must not be possible if a node must be activated
    INVARSPEC must_fire -> !%s;

    -- it should always be possible to activate an inactive node
    -- if this property fails, there might be a bug in the original tfpg
    -- (e.g. non-intersecting modes of incoming edges of an AND node) or
    -- in the translation to smv
    CTLSPEC (AG (EF (is_active)));""" % time_trans_ivar_name

discrepancy_node_common = """
    VAR is_active: boolean;
    ASSIGN init(is_active) := FALSE;
    ASSIGN next(is_active) := case
        instant_trans_type != NODE_ACTIVATION : is_active;
        can_fire : {FALSE, TRUE};
        must_fire : TRUE;
        TRUE : is_active;
        esac;"""

failure_mode_common = """
MODULE failuremode (instant_trans_type)

    VAR is_active : boolean;
    ASSIGN init(is_active) := {FALSE, TRUE};
    ASSIGN next(is_active) := case
        instant_trans_type != NODE_ACTIVATION : is_active;
        !is_active : {FALSE, TRUE};
        TRUE : is_active;
        esac;"""

failure_mode_validation_props = """

    -- nodes should stay active once activated
    INVARSPEC is_active -> next(is_active);

    -- if the current transition is not a node activation, the status
    -- of the node must not change
    INVARSPEC instant_trans_type != NODE_ACTIVATION -> is_active = next(is_active);

    -- it should always be possible to activate an inactive node
    -- if this property fails, there might be a bug in the original tfpg
    -- (e.g. non-intersecting modes of incoming edges of an AND node) or
    -- in the translation to smv
    CTLSPEC (AG (EF (is_active)));"""

instant_edge_common = """
MODULE instant_edge (source, target, tmax_is_infinity, system_mode_is_compatible)

    DEFINE is_active :=
        source.is_active &
        !target.is_active &
        system_mode_is_compatible;
    DEFINE can_fire := is_active & tmax_is_infinity;
    DEFINE must_fire := is_active & !tmax_is_infinity;"""

instant_edge_validation_props = """

    -- edges shouldn't allow or require firing if their target node is already
    -- active
    INVARSPEC target.is_active -> ! (can_fire | must_fire);
    INVARSPEC !is_active -> ! (can_fire | must_fire);

    -- can- and must-firing signals must be mutually exclusive
    INVARSPEC (can_fire -> !must_fire) & (must_fire -> !can_fire);"""

def get_edge_common (time_trans_ivar_name):
    return """
MODULE edge (tmin, tmax, tmax_is_infinity, source, target, instant_trans_type, %s, system_mode_is_compatible)

    VAR counter : 0..tmax;

    DEFINE is_active :=
        -- failure has propagated up to source node
        source.is_active &
        -- failure has not propagated to target node yet
        !target.is_active &
        -- system is in a mode where the edge is enabled
        system_mode_is_compatible;
    DEFINE can_fire :=
        is_active &
        counter >= tmin &
        (counter < tmax | tmax_is_infinity);
    DEFINE must_fire :=
        is_active &
        !tmax_is_infinity &
        counter = tmax;

    ASSIGN init(counter) := 0;
    ASSIGN next(counter) := case
        -- edge is not active
        next(!is_active) : 0;
        -- increment timer by one delta unit
        counter < tmax & %s :
            counter + 1;
        -- keep counter as-is
        TRUE : counter;
        esac;""" % (time_trans_ivar_name, time_trans_ivar_name)

def get_edge_validation_props (time_trans_ivar_name):
    return """

    -- the following is used to catch tmin/tmax specification errors
    INVARSPEC tmax_is_infinity -> (tmin = tmax);
    INVARSPEC 0 <= tmin;
    INVARSPEC tmin <= tmax;

    -- edges shouldn't allow or require firing if their target node is
    -- already active or if the edge itself is not active
    INVARSPEC target.is_active -> (!can_fire & !must_fire);
    INVARSPEC !is_active -> (!can_fire & !must_fire);

    -- can- and must-firing signals must be mutually exclusive
    INVARSPEC (can_fire -> !must_fire) & (must_fire -> !can_fire);

    -- the counter can be greater than zero only if the edge is active
    INVARSPEC counter > 0 -> is_active;

    -- counters can be increased only in a time_tick transition
    INVARSPEC counter < next(counter) -> (%s &
        is_active & next(is_active));

    -- during a time_tick transition, all counters that can be incremented
    -- must be incremented by one time unit
    INVARSPEC
        (%s & is_active & counter < tmax)
        ->
        ((counter + 1) = next(counter));""" % (time_trans_ivar_name, time_trans_ivar_name)

edge_constraint_A = """

    -- counters should not be reset if a mode switch changes to a mode where
    -- the edge is still active
    INVARSPEC instant_trans_type = MODE_CHANGE & is_active & next(is_active) ->
        counter = next(counter);"""

instant_trans_type_constraints_A = """
-- mode_change transitions occur iff the system mode is effectively changed
TRANS instant_trans_type = MODE_CHANGE <-> system_mode != next(system_mode);"""

instant_trans_type_constraints_B = """
-- node_activation transitions occur iff a node activation is changed
-- or: during a node activation, at least one node must be activated
TRANS instant_trans_type = NODE_ACTIVATION <->"""

global_prop_1_header = """
-- node_activation transition shouldn't be possible if no node can
-- be activated
INVARSPEC instant_trans_type = NODE_ACTIVATION ->"""

global_prop_2_header = """
-- when all nodes are activated, it must not be possible to do a
-- node_activation transition
INVARSPEC instant_trans_type = NODE_ACTIVATION ->"""

alarm_file_head = """<?xml version="1.0"?>
<fame>
<fmlist>
</fmlist>
<obslist>"""

alarm_file_tail = """</obslist>
<alarmlist>
</alarmlist>
<eventlist>
</eventlist>
</fame>
"""


########################################################################## UTILS

def is_divisible_by(dividend, divisor):

    return (Decimal('%s'%dividend) % Decimal('%s'%divisor)) == Decimal('0.0')

def tfpg_timings_are_divisible_by_sampling_rate (tfpg_file, sampling_rate):
    error_lst = []
    assert_sampling_rate_is_valid(sampling_rate)

    # extract tfpg timings from xml file

    tfpg = parse_to_internal_struct(tfpg_file)
    edges = tfpg.get_edges_list()

    for e in edges:

        tmin = float(e.get_tMin())
        assert_timing_is_valid(tmin)

        if not is_divisible_by(tmin, sampling_rate):
            error_lst.append("ERROR: tMin %f / sampling rate '%f' reminder must be 0.\n" % (tmin, sampling_rate))
            return False, error_lst

        tmax = float(e.get_tMax())
        assert_timing_is_valid(tmax)
        # ignore infinity
        if tmax == -1:
            continue

        if not is_divisible_by(float(e.get_tMax()), sampling_rate):
            error_lst.append("ERROR: tMax %f / sampling rate '%f' reminder must be 0.\n" % (tmax, sampling_rate))
            return False, error_lst

    return True, []

def assert_timing_is_valid (t):
    """ Verifies that a TFPG edge timing constant
    fulfills our assumptions. """
    assert isinstance(t, numbers.Real)
    #assert not has_decimal_part(t)
    assert (t >= 0 or t == -1)

def assert_sampling_rate_is_valid (sampling_rate):
    assert isinstance(sampling_rate, numbers.Real)
    assert sampling_rate > 0

def has_decimal_part (f):
    return (Decimal('%f' % f) % 1) != Decimal('0.0')


################################# HELPERS FOR STANDALONE DIAGNOSABILITY ANALYSIS

def write_obs_file (observables, obs_file, time_trans_ivar_name, system_mode_var_name):

    f = open(obs_file, 'w')

    for n in observables:
        f.write("%s.is_active\n" % compute_smv_node_name(n))
   
    f.write("%s\n" % time_trans_ivar_name)
    f.write("%s\n" % system_mode_var_name)

    f.close()

    print "written %s" % obs_file

def write_cond_and_cmd_files (tfpg, files_prefix, clear_existing_files, dump_cmd_files):

    counter = 0

    # write diagnosability conditions
    for cur_fm in tfpg.get_fm_nodes():

        counter += 1
        cond_file = files_prefix + ("_%d.cond" % counter)

        if os.path.exists(cond_file):
            if clear_existing_files:
                os.remove(cond_file)
                print "removed %s" % cond_file
            else:
                print "ERROR: The conditions file '" + cond_file + "' already exists.\n"
                sys.exit(1)

        f = open(cond_file, 'w')

        # compute smv name
        smv_name = compute_smv_node_name(cur_fm)

        # write condition to file

        f.write("CONDITION_L %s.is_active\n" % smv_name)
        f.write("CONDITION_R TRUE\n")
        f.write("CONTEXT_R !%s.is_active\n" % smv_name)

        f.close()

        print "written %s" % cond_file

        if dump_cmd_files:

            cmd_file = files_prefix + ("_%d.cmd" % counter)

            tmp_files_prefix = files_prefix[files_prefix.rfind('/')+1:]
            tmp_cond_file = cond_file[cond_file.rfind('/')+1:]

            if os.path.exists(cmd_file):
                if clear_existing_files:
                    os.remove(cmd_file)
                    print "removed %s" % cmd_file
                else:
                    print "ERROR: The conditions file '" + cmd_file + "' already exists.\n"
                    sys.exit(1) 

            f = open(cmd_file, 'w')

            f.write("set on_failure_script_quits\n")
            f.write("build_twin_plant -o %s -n %s\n" % (tmp_files_prefix+".obs", "\"!time_tick & instant_trans_type=NONE\""))
            f.write("load_diag_conditions -c %s -p -d -1\n" % tmp_cond_file)
            f.write("go\n")
            f.write("check_property\n")
            f.write("quit\n")

            f.close()

            print "written %s" % cmd_file

def create_diagnosability_files (tfpg, smv_file, obs_file, files_prefix, clear_existing_files, dump_cmd_files, time_trans_ivar_name, system_mode_var_name, sampling_rate, frozen_sys_mode):

    # create obs file (all observable nodes, plus observable transitions: timed transitions and system mode transitions)
    observables = tfpg.get_monitored_discrepancies()
    write_obs_file(observables, obs_file, time_trans_ivar_name, system_mode_var_name)

    # create cond file (each failure mode)
    write_cond_and_cmd_files(tfpg, files_prefix, clear_existing_files, dump_cmd_files)

    # create smv file
    dump_tfpg_as_smv(tfpg, smv_file, time_trans_ivar_name, sampling_rate, frozen_sys_mode, dump_validation_props=False)


############################################ VARIOUS PARSING / DUMPING FUNCTIONS

def parse_to_internal_struct (input_file):

    from hasdel_tools.src.xml.tfpg import TfpgDocument
    from hasdel_tools.src.xml.tfpg import tfpg_doc_to_tfpg_internal_struct

    # sanity checks

    if not os.path.exists(input_file):
        print "ERROR: The input file '" + input_file + "' does not exist.\n"
        return

    # parse xml document

    tfpg_doc, validation_errors = TfpgDocument().parse(input_file)

    if len(validation_errors) > 0:
        print "VALIDATION ERRORS: The file '" + input_file + "' contains validation errors.\n"

    if len(TfpgDocument().get_syntax_errors_in_tfpg_document(tfpg_doc)) > 0:
        print "SYNTAX ERRORS: The file '" + input_file + "' contains syntax errors.\n"

    # compute internal tfpg representation from xml parse tree

    return tfpg_doc_to_tfpg_internal_struct(tfpg_doc)


def get_slim_associations_from_file(input_file):

    from hasdel_tools.src.xml.tfpg_slim_associations import AssociationsDocument

    if not os.path.exists(input_file):
        print "ERROR: The input file '" + input_file + "' does not exist.\n"
        return

    assoc_doc, validation_errors = AssociationsDocument().parse(input_file)

    if len(validation_errors) > 0:
        print "VALIDATION ERRORS: The file '" + input_file + "' contains validation errors.\n"

    if len(AssociationsDocument.get_syntax_errors_in_assoc_document(assoc_doc)) > 0:
        print "SYNTAX ERRORS: The file '" + input_file + "' contains syntax errors.\n"

    assoc_dic = dict()

    for assoc in assoc_doc.failureModes_assoc_ls:
        _id = assoc.get('id')
        expr = assoc.get('expr')
        assoc_dic[_id] = ('OK', expr, 'FailureMode')

    for assoc in assoc_doc.monitoredDiscr_assoc_ls:
        _id = assoc.get('id')
        expr = assoc.get('expr')
        assoc_dic[_id] = ('OK', expr, 'MonitoredDiscr')

    for assoc in assoc_doc.unmonitoredDiscr_assoc_ls:
        _id = assoc.get('id')
        expr = assoc.get('expr')
        assoc_dic[_id] = ('OK', expr, 'UnmonitoredDiscr')

    for assoc in assoc_doc.tfpgModes_assoc_ls:
        _id = assoc.get('id')
        expr = assoc.get('expr')
        assoc_dic[_id] = ('OK', expr, 'TfpgMode')

    return assoc_dic


def dump_discrepancy_xml (tfpg, output_file, time_trans_ivar_name):

    """
    tfpg: internal tfpg struct
    output file: file where to dump the xml with the smv names of observables
    time trans ivar name: name for Boolean input variable indicating time ticks
    """

    fout = open(output_file, 'w')
    fout.write("%s\n" % alarm_file_head)

    # discrepancy nodes
    for obs in tfpg.get_monitored_discrepancies():
        node_name = obs.get_name()
        node_type = obs.get_type().lower()
        smv_name = "%s_node_%s" % (node_type, node_name)
        fout.write('  <obs name="%s.is_active"/>\n' % smv_name)

    # timed transition
    fout.write('  <obs name="%s"/>\n' % time_trans_ivar_name)
    # system mode
    fout.write('  <obs name="system_mode"/>\n')

    fout.write("%s\n" % alarm_file_tail)
    fout.close()


###################################################### MAIN CONVERSION FUNCTIONS

def compute_smv_node_name (n):
    t = n.get_type().lower()
    if "fm" == t:
        return "failuremode_%s" % n.get_name()
    else:
        return "%s_node_%s" % (t, n.get_name())

def compute_smv_edge_name (e):
    return "edge_%s_to_%s_%s" % (e.get_srcNode().get_name(), e.get_destNode().get_name(), e.get_name())

def compute_smv_edge_state_name (e):
    return "EState_%s" % e.get_name()

def compute_smv_edge_clock_name (e):
    return "ETime_%s" % e.get_name()

def compute_smv_mode_name (m):
    return "Mode_%s" % m

def read_tfpg_and_dump_discrepancy_xml (tfpg_file, output_file, time_trans_ivar_name):

    """
    tfpg file: xml file specifying the tfpg
    output file: file where to dump the xml with the smv names of observables
    time trans ivar name: name for Boolean input variable indicating time ticks
    """

    tfpg = parse_to_internal_struct(tfpg_file)
    dump_discrepancy_xml(tfpg, output_file, time_trans_ivar_name)

def read_tfpg_and_dump_smv (tfpg_file, output_file, time_trans_ivar_name, sampling_rate, frozen_sys_mode, dump_validation_props):

    """
    tfpg file: xml file specifying the tfpg
    output file: file where to dump smv code
    time trans ivar name: name for Boolean input variable indicating time ticks
    sampling rate: modifier for tmin/tmax values; set to <= 0 to disable
    """

    tfpg = parse_to_internal_struct(tfpg_file)
    dump_tfpg_as_smv(tfpg, output_file, time_trans_ivar_name, sampling_rate, frozen_sys_mode, dump_validation_props)

def dump_tfpg_as_smv (tfpg, output_file, time_trans_ivar_name, sampling_rate, frozen_sys_mode, dump_validation_props):

    """
    tfpg: internal tfpg struct
    output file:           file where to dump smv code
    time trans ivar name:  name for Boolean input variable indicating time ticks
    sampling rate:         modifier for tmin/tmax values; set to <= 0 to disable
    frozen_system_mode:    system mode behaves like a frozen variable
    dump_validation_props: Boolean variable specifying whether it is requested to 
                           add validation Axioms in the SMV file
    """

    # sanity check on parameters

    assert_sampling_rate_is_valid(sampling_rate)

    # translate internal tfpg representation to SMV and dump to file

    smv_code = "" # this string will contain the whole SMV code

    # system modes
    system_modes = tfpg.get_modes_list()
    single_mode_system = len(system_modes) == 1

    # transition types in addition to time_tick
    additional_transition_types = None
    if single_mode_system:
        additional_transition_types = "{NODE_ACTIVATION, NONE}"
    else:
        additional_transition_types = "{NODE_ACTIVATION, MODE_CHANGE, NONE}"

    # get nodes, edges, and maps between them

    discrepancy_nodes = []
    failuremode_nodes = []

    # key: node id; value: node object
    node_map = {}

    # key: node id; value: list of objects of incoming edges
    incoming_edges = {}

    for n in tfpg.get_nodes_list():

        nname = n.get_name()
        assert nname not in node_map
        assert nname not in incoming_edges
        node_map[nname] = n

        if n.get_type() == "FM":
            failuremode_nodes.append(n)
        else:
            discrepancy_nodes.append(n)
            incoming_edges[nname] = []

    assert len(incoming_edges) == len(discrepancy_nodes)

    node_ids = node_map.keys()

    edges = tfpg.get_edges_list()

    for e in edges:
        incoming_edges[e.get_destNode().get_name()].append(e)

    # compute existing combinations of node types and
    # number of incoming edges

    or_node_incoming_edges_count = set()
    and_node_incoming_edges_count = set()

    for n in tfpg.get_or_nodes():
        or_node_incoming_edges_count.add(len(incoming_edges[n.get_name()]))

    for n in tfpg.get_and_nodes():
        and_node_incoming_edges_count.add(len(incoming_edges[n.get_name()]))

    # dump discrepancy node modules

    for num_edges in or_node_incoming_edges_count:

        # header

        smv_code += "\nMODULE or_node_%d (" % num_edges
        for n in range(1, num_edges+1):
            smv_code += "edge%d, " % n
        smv_code += "time_tick, instant_trans_type)\n\n"

        # can fire

        smv_code += "    DEFINE can_fire  :=\n"
        for n in range(1, num_edges+1):
            if n == 1:
                smv_code += "        (edge1.can_fire"
            else:
                smv_code += " | edge%d.can_fire" % n
        smv_code += ") & !must_fire;\n"

        # must fire

        smv_code += "    DEFINE must_fire :=\n"
        for n in range(1, num_edges+1):
            if n == 1:
                smv_code += "        (edge1.must_fire"
            else:
                smv_code += " | edge%d.must_fire" % n
        smv_code += ");\n"

        # rest
        smv_code += discrepancy_node_common

        if dump_validation_props:
            smv_code += get_discrepancy_node_validation_props(time_trans_ivar_name)

        smv_code += "\n"

    for num_edges in and_node_incoming_edges_count:

        smv_code += "\nMODULE and_node_%d (" % num_edges
        for n in range(1, num_edges+1):
            smv_code += "edge%d, " % n
        smv_code += "time_tick, instant_trans_type)\n\n"

        # can fire

        smv_code += "    DEFINE can_fire  :=\n"
        for n in range(1, num_edges+1):
            smv_code += "        (edge%d.can_fire | edge%d.must_fire) &\n" % (n, n)

        smv_code += "        ("
        for n in range(1, num_edges+1):
            if n == 1:
                smv_code += "edge1.can_fire"
            else:
                smv_code += " | edge%d.can_fire" % n
            if n == num_edges:
                smv_code += ");\n"

        # must fire

        smv_code += "    DEFINE must_fire :=\n"
        for n in range(1, num_edges+1):
            if n < num_edges:
                smv_code += "        edge%d.must_fire &\n" % n
            else:
                smv_code += "        edge%d.must_fire;\n" % n

        # rest
        smv_code += discrepancy_node_common

        if dump_validation_props:
            smv_code += get_discrepancy_node_validation_props(time_trans_ivar_name)

        smv_code += "\n"

    # failuremode nodes

    smv_code += failure_mode_common

    if dump_validation_props:
        smv_code += failure_mode_validation_props

    # instant edges

    smv_code += "\n"
    smv_code += instant_edge_common

    if dump_validation_props:
        smv_code += instant_edge_validation_props

    # normal edges

    smv_code += "\n"
    smv_code += get_edge_common(time_trans_ivar_name)

    if dump_validation_props:

        smv_code += get_edge_validation_props(time_trans_ivar_name)

        if not single_mode_system:
            smv_code += edge_constraint_A

    smv_code += "\n\n"

    # main module

    smv_code += "--------------------------------------------------------------------------- MAIN\n\n"
    smv_code += "MODULE main\n\n"

    smv_code += "-- system modes and transition types\n\n"

    # input signal for time_tick
    smv_code += "IVAR %s : boolean;\n" % time_trans_ivar_name

    # input signal for other transitions
    smv_code += "IVAR instant_trans_type : %s;\n" % additional_transition_types

    # time ticks and other transitions have to be mutually exclusive (XOR)
    smv_code += "TRANS !%s | instant_trans_type=NONE;\n" % time_trans_ivar_name
    smv_code += "TRANS %s | instant_trans_type!=NONE;\n\n" % time_trans_ivar_name

    smv_code += "VAR system_mode : {%s};\n\n" % ', '.join(system_modes)
    
    if frozen_sys_mode:
        smv_code += "TRANS system_mode = next(system_mode);\n\n"

    smv_code += "-- failure modes\n\n"

    for n in failuremode_nodes:
        smv_code += "VAR failuremode_%s : failuremode (instant_trans_type);\n" % n.get_name()

    smv_code += "\n-- discrepancy nodes\n\n"

    for n in discrepancy_nodes:

        node_name = n.get_name()
        node_type = n.get_type().lower()

        smv_code += "VAR %s_node_%s : %s_node_%d (" % (node_type, node_name, node_type, len(incoming_edges[node_name]))

        for e in incoming_edges[node_name]:
            smv_code += "%s, " % compute_smv_edge_name(e)

        smv_code += "%s, instant_trans_type);\n" % time_trans_ivar_name

    smv_code += "\n-- edges\n\n"

    for e in tfpg.get_edges_list():

        # retrieve tmin/tmax and, if necessary, adjust wrt sampling rate

        tmin = float(e.get_tMin())
        tmax = float(e.get_tMax())

        assert_timing_is_valid(tmin)
        assert_timing_is_valid(tmax)

        assert is_divisible_by(tmin, sampling_rate)
        tmin = tmin / sampling_rate

        if tmax == -1:
            # tmax is infinity, nothing to be done
            pass
        else:
            assert is_divisible_by(tmax, sampling_rate)
            tmax = tmax / sampling_rate

        assert_timing_is_valid(tmin)
        assert_timing_is_valid(tmax)

        tmin = int(tmin)
        tmax = int(tmax)

        # declare edge

        tmax_is_infinity = -1 == tmax
        is_instant_edge = (0 == tmax) or (0 == tmin and tmax_is_infinity)

        if tmax_is_infinity:
            tmax = tmin

        source = e.get_srcNode().get_name()
        target = e.get_destNode().get_name()

        edge_modes = e.get_modes()
        mode_expression = None

        if len(system_modes) == 1 or len(system_modes) == len(edge_modes):
            mode_expression = "TRUE"
        elif len(edge_modes) == 1:
            mode_expression = "system_mode = %s" % edge_modes[0]
        else:
            assert len(edge_modes) > 1
            mode_expression = "system_mode in {%s}" % ', '.join(edge_modes)

        if is_instant_edge:
            smv_code += "VAR %s : instant_edge (%s, %s, %s, %s);\n" % \
                (compute_smv_edge_name(e),                            \
                compute_smv_node_name(node_map[source]),              \
                compute_smv_node_name(node_map[target]),              \
                ("TRUE" if tmax_is_infinity else "FALSE"),            \
                mode_expression)

        else:
            smv_code += "VAR %s : edge (%d, %d, %s, %s, %s, instant_trans_type, %s, %s);\n" % \
                (compute_smv_edge_name(e),                                        \
                tmin,                                                             \
                tmax,                                                             \
                ("TRUE" if tmax_is_infinity else "FALSE"),                        \
                compute_smv_node_name(node_map[source]),                          \
                compute_smv_node_name(node_map[target]),                          \
                time_trans_ivar_name,                                             \
                mode_expression)

    # constraints

    smv_code += "\n-- trans type constraints\n"

    if len(system_modes) > 1:
        smv_code += instant_trans_type_constraints_A

    smv_code += "\n"
    smv_code += instant_trans_type_constraints_B
    smv_code += "\n"

    for n in range(0,len(node_ids)):
        smv_name = compute_smv_node_name(node_map[node_ids[n]])
        if 0 == n:
            smv_code += "      (%s.is_active != next(%s.is_active)" % (smv_name, smv_name)
        else:
            smv_code += "       %s.is_active != next(%s.is_active)" % (smv_name, smv_name)
        if n < len(node_ids)-1:
            smv_code += " |\n"
        else:
            smv_code += ");\n"

    smv_code += "\n-- time tick is not possible if any node must fire\n"

    for n in range(0,len(discrepancy_nodes)):
        smv_name = compute_smv_node_name(discrepancy_nodes[n])
        if 0 == n:
            smv_code += "TRANS (%s.must_fire" % smv_name
        else:
            smv_code += "       %s.must_fire" % smv_name
        if n < len(discrepancy_nodes)-1:
            smv_code += " |\n"
        else:
            smv_code += ")\n"
    smv_code += "      -> !%s;\n" % time_trans_ivar_name

    # verification properties

    if dump_validation_props:

        smv_code += "\n-- global verification properties\n\n"

        # trans types must be XOR

        smv_code += "-- when a time tick occurs, no instant transition can occur\n"
        smv_code += "INVARSPEC %s -> instant_trans_type=NONE\n" % time_trans_ivar_name

        smv_code += "-- when an instant transition occurs, no time tick can occur\n"
        smv_code += "INVARSPEC instant_trans_type != NONE -> !%s\n" % time_trans_ivar_name

        smv_code += "-- at least one transition type must occur\n"
        smv_code += "INVARSPEC instant_trans_type != NONE | %s\n" % time_trans_ivar_name

        smv_code += "-- at most one transition type must occur\n"
        smv_code += "INVARSPEC instant_trans_type = NONE | !%s\n" % time_trans_ivar_name

        smv_code += global_prop_1_header
        smv_code += "\n"

        for n in range(0,len(node_ids)):
            smv_name = compute_smv_node_name(node_map[node_ids[n]])
            if 0 == n:
                smv_code += "     ("
            else:
                smv_code += "      "
            if "FM" == node_map[node_ids[n]].get_type():
                smv_code += "!%s.is_active" % smv_name
            else:
                smv_code += "%s.can_fire  |\n" % smv_name
                smv_code += "      %s.must_fire" % smv_name
            if n < len(node_ids)-1:
                smv_code += " |\n"
            else:
                smv_code += ");\n"

        smv_code += global_prop_2_header
        smv_code += "\n"

        node_ids = node_map.keys()
        for n in range(0,len(node_ids)):
            smv_name = compute_smv_node_name(node_map[node_ids[n]])
            if 0 == n:
                smv_code += "      (!%s.is_active" % smv_name
            else:
                smv_code += "       !%s.is_active" % smv_name
            if n < len(node_ids)-1:
                smv_code += " |\n"
            else:
                smv_code += ");\n"

    smv_code += "\n"

    # write SMV string to file

    f = open(output_file, 'w')
    f.write(smv_code)
    f.close()

    print "written %s" % output_file


def dump_extended_sys_as_smv (root_impl, tfpg, associations, output_fname):
    """ Dumps a SLIM model, extended with variables for mapping system
    traces to TFPG traces, to SMV. This is used in TFPG behavioral analysis
    (completeness).

    root_impl:      root implementation of the SLIM system model
    tfpg:           internal tfpg struct
    associations:   the TFPG SLIM associations
    output fname:   file where to dump smv code
    """

    # we assume that the system model contains clock data

    from slim_compiler.src.smv.walkers import InfiniteDomainFinderWalker

    walker = InfiniteDomainFinderWalker()
    walker.walk(root_impl)

    assert walker.has_found_clock_data()

    conv = MainExtensionConverter()

    discrepancy_nodes = tfpg.get_nodes_list()

    edges = tfpg.get_edges_list()
    delta_var = conv.delta_var()

    main_extension = "\n"

    # compute history variables for discrepancies

    main_extension += \
        "    -- discrepancy node definitions\n\n"

    for d in discrepancy_nodes:

        name = compute_smv_node_name(d)
        slim_expr = associations[d.get_name()][1]
        assoc = get_smv_format(slim_expr, root_impl)

        main_extension +=                           \
            "    VAR %s : boolean;\n"                 \
            "    INIT %s <-> (%s);\n"                 \
            "    TRANS next(%s) <-> (%s | next(%s));\n\n" \
            % (name, name, assoc, name, name, assoc)

    # compute mode definitions

    main_extension += "    -- Mode definitions\n\n"

    for m in tfpg.get_modes_list():
        mode_expr = get_smv_format(associations[m][1], root_impl)
        main_extension += "    DEFINE %s := %s;\n" % (compute_smv_mode_name(m), mode_expr)

    main_extension += "\n"

    # compute edge status and edge time definitions

    main_extension += \
        "    -- EState/ETime definitions\n"

    for e in edges:

        main_extension += "\n"

        source = compute_smv_node_name(e.get_srcNode())
        target = compute_smv_node_name(e.get_destNode())

        tmin = float(e.get_tMin())

        ################################################################# EState

        estate = compute_smv_edge_state_name(e)

        mode_expr = ' | '.join([compute_smv_mode_name(m) for m in e.get_modes()])
        estate_expr = "%s & (%s | (%s))" % (source, target, mode_expr)

        main_extension += \
            "    DEFINE %s := %s;\n" % (estate, estate_expr)

        ################################################################## ETime

        etime = compute_smv_edge_clock_name(e)

        main_extension += "    VAR %s : real;\n" % etime

        # init constraint
        main_extension += "    INIT (%s -> %s = 0) & (!%s -> %s = -1);\n" \
            % (estate, etime, estate, etime)

        # timed-trans constraint
        main_extension += "    TRANS %s > 0 -> (\n" \
            % (delta_var)
        # edge is active and propagation hasn't occurred yet
        main_extension += "      ((%s & !%s) -> (next(%s) = %s + %s)) &\n" \
            % (estate, target, etime, etime, delta_var)
        # edge is not active or propagation has already occurred
        main_extension += "      ((!%s | %s) -> (next(%s) = %s)));\n" \
            % (estate, target, etime, etime)

        # untimed-trans constraint
        main_extension += "    TRANS %s = 0 -> (\n" \
            % (delta_var)
        # !target' & !estate & estate' -> etime' = 0
        main_extension += "      ((next(!%s) & !%s & next(%s)) -> next(%s = 0)) &\n" \
            % (target, estate, estate, etime)
        # !target' & !estate' -> etime' = -1
        main_extension += "      ((next(!%s) & next(!%s)) -> (next(%s = -1))) &\n" \
            % (target, estate, etime)
        # !target' & estate & estate' -> etime = etime'
        main_extension += "      ((next(!%s) & %s & next(%s)) -> (%s = next(%s))) &\n" \
            % (target, estate, estate, etime, etime)

        if tmin == 0:
            # target' & (!source & source' & !target & mode_expr) -> etime' = 0
            main_extension += "      ((next(%s) &  (!%s & next(%s) & !%s & %s)) -> (next(%s) = 0)) &\n" \
                % (target, source, source, target, mode_expr, etime)
            # target' & $INVERT$ -> etime = etime'
            main_extension += "      ((next(%s) & !(!%s & next(%s) & !%s & %s)) -> (next(%s) = %s)));\n" \
                % (target, source, source, target, mode_expr, etime, etime)

        else:
            # target' -> etime = etime'
            main_extension += "      (next(%s) -> (%s = next(%s))));\n" \
                % (target, etime, etime)

        main_extension += "\n"

    # dump smv to disk

    conv.set_main_extension(main_extension)

    out = open(output_fname,'w')
    conv.dump(root_impl, out)
    out.close()

    return


def dump_extension_for_synthesis_to_smv (root_impl, associations, output_fname):
    """ Dumps a SLIM model, extended with variables defining discrepancies, to SMV.
    This is used in TFPG synthesis.

    root_impl:      root implementation of the SLIM system model
    associations:   the TFPG SLIM associations
    output fname:   file where to dump smv code
    """

    conv = MainExtensionConverter()

    main_extension = "\n"
    main_extension += \
        "    -- discrepancy node definitions\n\n"

    for (k, v) in associations.items():
        (val, error, _type) = v
        if _type == "TfpgMode" : continue

        name = "tfpgsynth_%s" % k
        slim_expr = error
        assoc = get_smv_format(slim_expr, root_impl)

        main_extension +=                     \
            "    VAR %s : boolean;\n"         \
            "    INVAR %s <-> (%s);\n\n" \
            % (name, name, assoc)
            
    # dump smv to disk

    conv.set_main_extension(main_extension)

    out = open(output_fname,'w')
    conv.dump(root_impl, out)
    out.close()

    return


def dump_discrepancies_for_synthesis_to_xml (associations, output_fname):
    """ Dumps the discrepancies to an XML file. This is used in TFPG synthesis,
    specifically in the minimal cutset computation.

    associations:   the TFPG SLIM associations
    output fname:   file where to dump smv code
    """

    f = open(output_fname,'w')

    f.write('<?xml version="1.0"?>\n')
    f.write('<compass>\n')
    f.write('<fmlist>\n')

    for (k, v) in associations.items():
        (val, error, _type) = v
        if _type != "TfpgMode" : 
            f.write('<fm name="tfpgsynth_%s" nominal_value="FALSE"/>\n' % k)
        
    f.write('</fmlist>\n')
    f.write('<obslist>\n')
    f.write('</obslist>\n')
    f.write('</compass>\n')

    f.close()

    return

