# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================

from gtkmvc import Controller
import gtk
import sys
import src.utils.log
from src.models.tfpg_errors import FMNotCorrectError, TimeNotCorrectError, NodeNotInEdge


class TfpgCtrl(Controller):
    """Contains the controller for the TFPG Viewer"""


    def __init__(self, model, view, structural_file, trace_file):
        """Class constructor

        It is possible to specify from the command line
        a structural file and a trace file
        """

        Controller.__init__(self, model, view)

        # represents the state of the trace the user wants to see.
        # it increments when the user executes a trace or wants to see
        # the next state
        self.trace_step = 0

        if trace_file:

            self.tfpg_graph = model.build_graph(structural_file)
            src.utils.log.system.info("Sructural model properly built")

            self.tfpg_trace = model.build_trace(trace_file)
            src.utils.log.system.info("Trace model properly built")

            view.pyplot_view_trace(self.tfpg_graph, self.tfpg_trace)
            self.trace_step = len(self.tfpg_trace) + 1

        elif structural_file:
            self.tfpg_trace = None
            self.tfpg_graph = model.build_graph(structural_file)
            src.utils.log.system.info("Sructural model properly built")
            view.pyplot_view_structural(self.tfpg_graph)

        else:
            self.tfpg_graph = None
            self.tfpg_trace = None
 
        return


    # ------------------------------------------------------------------------
    # Signals handling
    # ------------------------------------------------------------------------

    def on_main_window_destroy(self, action):
        """Function called when the user wants to quit the viewer"""
        sys.exit(0)
        return


    # Functions of File menu

    def on_file_new_tfpg_activate(self, action):
        """Function called when the user wants to create a new structural
           tfpg from scratch"""

        import os

        dialog = gtk.FileChooserDialog("New TFPG",
                                       None,
                                       gtk.FILE_CHOOSER_ACTION_SAVE,
                                       (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                                       gtk.STOCK_OPEN, gtk.RESPONSE_OK))
        dialog.set_default_response(gtk.RESPONSE_OK)
        dialog.set_do_overwrite_confirmation(True)

        filter = gtk.FileFilter()
        filter.set_name("FAME TFPG")
        filter.add_pattern("*.tfpg")
        dialog.add_filter(filter)
        response = dialog.run()        

        if response == gtk.RESPONSE_OK:
            name = dialog.get_filename()
            file_name, file_extension = os.path.splitext(name)
            name = file_name + ".tfpg"
            dialog.destroy()
            self.on_edit_menu_edit_activate(action, modified_filename=name, from_scratch=True)
        else:
            dialog.destroy()

        return


    def on_file_load_structural_activate(self, action):
        """Function called when the user wants to load a structural tfpg
           from the file menu"""
        filename = self.view.ui_load_structural_file()

        if filename != None:
            src.utils.log.system.debug("Loaded '%s' as structural file" % filename)

            try:
                self.tfpg_graph = self.model.build_graph(filename)
                src.utils.log.system.info("Structural model properly built")
                self.view.pyplot_view_structural(tfpg_graph=self.tfpg_graph)
 
            except Exception as ex:
                self.view.ui_show_message_dialog("error", str(ex), parent_name='window_top')
                return self.on_edit_menu_edit_activate(action, filename)

        return


    def on_file_load_trace_activate(self, action):
        """Function called when the user wants to load a tfpg trace
           from the file menu"""
        if self.tfpg_trace is None:
            tfpg_trace_filename = self.view.ui_load_trace_file()
            src.utils.log.system.debug("Loaded '%s' as trace file" % tfpg_trace_filename)

        if self.tfpg_graph is None:
            tfpg_graph_filename = self.view.ui_load_structural_file()
            src.utils.log.system.debug("Loaded '%s' as structural file" % tfpg_graph_filename)
            self.tfpg_graph = self.model.build_graph(tfpg_graph_filename)
            src.utils.log.system.info("Structural model properly built")

        try:
            if self.tfpg_trace is None:
                self.tfpg_trace = self.model.build_trace(tfpg_trace_filename)
                src.utils.log.system.info("Trace model properly built")
            self.view.pyplot_view_trace(self.tfpg_graph, self.tfpg_trace, write_info=True)

            # all steps have been performed
            # TODO: check trace_step is lower than trace.tfpg_states()
            self.trace_step = len(self.tfpg_trace)

        except Exception as ex:
            self.view.ui_show_message_dialog("error", str(ex),
                                             secondary_text="Please modify the file and retry.",
                                             parent_name='main_window')
            self.tfpg_trace = None

        return


    def on_file_save_as_concrete_activate(self, action):
        """Function called when the user wants to save a structural tfpg
           in human-readable format"""
        out_filename = self.view.ui_save_structural_file("tfpg")
        src.utils.log.system.debug("Saving structural model to '%s'" % out_filename)
        self.tfpg_graph.dump_to_concrete_syntax_file(out_filename)
        src.utils.log.system.info("Structural model properly saved")
        return


    def on_file_save_as_xml_activate(self, action):
        """Function called when the user wants to save a structural tfpg
        in xml format"""
        out_filename = self.view.ui_save_structural_file("txml")
        src.utils.log.system.debug("Saving structural model to '%s'" % out_filename)

        from src.models.tfpg_document import tfpg_internal_struct_to_tfpg_doc

        with open(out_filename, "w") as f:
            f.write(str(tfpg_internal_struct_to_tfpg_doc(self.tfpg_graph)))

        src.utils.log.system.info("Structural model properly saved")
        return


    # Functions of Edit menu

    def on_edit_menu_edit_activate(self, action, modified_filename=None, from_scratch=False):
        """Function called when the user wants to edit, using a textual editor,
           the visualized structural

           The second parameter (modified_filename) is used in case the user
           has saved some modifications, but there was a parsing error. In this
           case, an error dialog is shown and the text editor is re-opened
           loading data from modified_filename

           :param modified_filename: the temporal filename of the unsuccessfully modified file
           :type modified_filename: string
        """

        # TODO: disable button when viewing a trace?
        
        if modified_filename is None:
            tmp_filename = "/tmp/tmp_structural.tfpg"
            self.tfpg_graph.dump_to_concrete_syntax_file(tmp_filename)
            src.utils.log.system.debug("File '%s' created" % (tmp_filename))
        else:
            tmp_filename = modified_filename

        ret_code = self.view.ui_run_text_editor(tmp_filename, from_scratch=from_scratch)

        if ret_code == 0:
            if self.tfpg_graph != None: self.tfpg_graph.clear()

            try:
                self.tfpg_graph = self.model.build_graph(tmp_filename)
                src.utils.log.system.info("Structural model properly built")
                self.view.pyplot_view_structural(self.tfpg_graph)

            except Exception as ex:
                self.view.ui_show_message_dialog("error", str(ex), parent_name='window_top')
                return self.on_edit_menu_edit_activate(action, tmp_filename)

        else:
            src.utils.log.system.error("Error while using the external editor")
            raise Exception("Error while using the external editor")

        return


    # Functions of the button bar which controls the trace execution

    def on_btn_run_trace_clicked(self, action):
        """Function called when the user wants to execute a trace"""

        self.on_file_load_trace_activate(action)

        return



    def on_btn_next_state_clicked(self, action):
        """Function called when the user wants to see the next state of a trace"""

        if self.tfpg_graph is None:
            self.on_file_load_structural_activate(action)

        if self.tfpg_trace is None:
            tfpg_trace_filename = self.view.ui_load_trace_file()
            src.utils.log.system.debug("Loaded '%s' as trace file" % tfpg_trace_filename)
            self.tfpg_trace = self.model.build_trace(tfpg_trace_filename)

        total_steps = len(self.tfpg_trace)
        
        if self.trace_step < total_steps:
            self.view.pyplot_view_state(self.tfpg_graph, self.tfpg_trace[self.trace_step])
            self.trace_step = self.trace_step + 1

        else:
            self.view.ui_show_message_dialog("info", "End of trace reached!", parent_name='main_window')

        return


    def on_btn_prev_state_clicked(self, action):
        """Function called when the user wants to see the previous state of a trace"""

        if self.tfpg_graph is None:
            self.on_file_load_structura_activate(action)

        if self.tfpg_trace is None:
            tfpg_trace_filename = self.view.ui_load_trace_file()
            src.utils.log.system.debug("Loaded '%s' as trace file" % tfpg_trace_filename)
            self.tfpg_trace = self.model.build_trace(tfpg_trace_filename)

        else:

            if self.trace_step > 0:
                self.trace_step = self.trace_step - 1

                if self.trace_step == 0:
                    self.view.pyplot_view_structural(self.tfpg_graph)
                else:
                    self.view.pyplot_view_state(self.tfpg_graph,
                                                self.tfpg_trace[self.trace_step-1])

            else:
                self.view.ui_show_message_dialog("info", "Start of trace reached!",
                                                 parent_name='main_window')

        return


    def on_btn_move_left_clicked(self, action):
        """Function called when the user wants to move the graph left"""

        if self.tfpg_graph is None:
            self.view.ui_show_message_dialog("error", "You must load a structural tfpg",
                                             parent_name="main_window")

        else:
            self.view.ui_move_figure("left")
 
        return


    def on_btn_move_right_clicked(self, action):
        """Function called when the user wants to move the graph right"""

        if self.tfpg_graph is None:
            self.view.ui_show_message_dialog("error", "You must load a structural tfpg",
                                             parent_name="main_widow")
        
        else:
            self.view.ui_move_figure("right")
        
        return


    def on_btn_move_top_clicked(self, action):
        """Function called when the user wants to move the graph top"""

        if self.tfpg_graph is None:
            self.view.ui_show_message_dialog("error", "You must load a structural tfpg",
                                             parent_name="main_window")

        else:
            self.view.ui_move_figure("top")

        return


    def on_btn_move_down_clicked(self, action):
        """Function called when the user wants to move the graph down"""

        if self.tfpg_graph is None:
            self.view.ui_show_message_dialog("error", "You must load a structural tfpg",
                                             parent_name="main_window")

        else:
            self.view.ui_move_figure("bottom")

        return


    def on_btn_zoom_in_clicked(self, action):
        """Function called when the user wants to zoom in"""

        if self.tfpg_graph is None:
            self.view.ui_show_message_dialog("error", "You must load a structural tfpg",
                                             parent_name="main_window")

        else:
            self.view.ui_zoom_in()
        
        return


    def on_btn_zoom_out_clicked(self, action):
        """Function called when the user wants to zoom out"""

        if self.tfpg_graph is None:
            self.view.ui_show_message_dialog("error", "You must load a structural tfpg",
                                             parent_name="main_window")

        else:
            self.view.ui_zoom_out()
  
        return


    def on_btn_caption_clicked(self, action):
        """Function called when the user wants to display the caption"""

        popup = gtk.Window()
        popup.set_position(gtk.WIN_POS_MOUSE)
        popup.set_title( "Caption" )

        hbox = gtk.HBox()
        for ch in hbox.get_children():
            hbox.remove(ch)

        caption = self.view.create_caption()
        hbox.add(caption)
        popup.add(hbox)
        popup.set_default_size(400, 1400)
        popup.show_all()

        return

