# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================

from xml.dom import minidom
import networkx as nx
import matplotlib.pyplot as plt
from src.models.tfpg_structural import Tfpg as TfpgStructural
from src.models.tfpg_trace import TFPGState
from src.models.tfpg_trace import TFPGTrace


class XMLParser():
    """Class that converts the TFPG described by an xml file
    into an internal tfpg_graph"""

    def __init__(self):
        """Class constructor"""
        self.tfpg_graph = TfpgStructural()
        return


    def build_graph(self, xml_file):
        """Parses the XML file and builds the graph

        :param xml_file: the name of the input xml file
        "type xml_file: str

        """

        xml = minidom.parse(xml_file)

        tfpg_names = xml.getElementsByTagName("tfpg")
        for n in tfpg_names: name = n.getAttribute("name")
        self.tfpg_graph.set_name(name)
        
        nodes = xml.getElementsByTagName("node");
        edges = xml.getElementsByTagName("edge");
        modes = xml.getElementsByTagName("mode");

        for nd in nodes:
            nd_name = nd.getAttribute('name')
            if nd.getAttribute('isMonitored').lower() == "true":
                nd_monitored = True
            else:
                nd_monitored = False
            nd_type = nd.getElementsByTagName("type")[0].firstChild.nodeValue

            if nd_type == "FM":
                if nd_monitored is True:
                    raise FMNotCorrectError(nd_name)
                else:
                    node = self.tfpg_graph.add_fm_node(nd_name)

            elif nd_type == "AND":
                self.tfpg_graph.add_and_node(nd_name, nd_monitored)

            elif nd_type == "OR":
                self.tfpg_graph.add_or_node(nd_name, nd_monitored)

        modes_list = []
        for md in modes:
            if md.parentNode.parentNode.tagName == "tfpg":
                modes_list.append(md.firstChild.nodeValue)
        self.tfpg_graph.set_modes_list(modes_list)

        for eg in edges:
            eg_name = eg.getAttribute('name')
            eg_src = eg.getElementsByTagName("srcNode")[0].firstChild.nodeValue
            eg_dst = eg.getElementsByTagName("destNode")[0].firstChild.nodeValue
            eg_min_time = eg.getElementsByTagName("tMin")[0].firstChild.nodeValue
            eg_max_time = eg.getElementsByTagName("tMax")[0].firstChild.nodeValue
            eg_modes = eg.getElementsByTagName("mode")
            eg_modes_list = []
            for m in eg_modes:
                eg_modes_list.append(m.firstChild.nodeValue)

            edge = self.tfpg_graph.add_edge(self.tfpg_graph.get_node_by_name(eg_src), 
                                            eg_min_time, eg_max_time,
                                            eg_modes_list,
                                            self.tfpg_graph.get_node_by_name(eg_dst),
                                            eg_name)

        errors = self.tfpg_graph.get_errors_list()
        if errors != []:
            for err in errors:
                raise err

        return self.tfpg_graph


    def build_trace(self, xml_file):
        """Parses the xml file and builds the trace

        :param xml_file: the input xml file containing the tfpg trace
        :type xml_file: string

        """

        xml = minidom.parse(xml_file)
        states = xml.getElementsByTagName("tfpgState")
        tfpg_trace = TFPGTrace()

        for st in states:
            tfpg_st = TFPGState()

            global_time = st.getElementsByTagName("globalTime")[0].firstChild.nodeValue
            tfpg_st.set_global_time(global_time)

            mode = st.getElementsByTagName("mode")[0].firstChild.nodeValue
            tfpg_st.set_mode(mode)

            nodes = st.getElementsByTagName("node")
            edges = st.getElementsByTagName("edge")

            for nd in nodes:
                if nd.getAttribute('isActive') == 'true':
                    tfpg_st.set_node_state(nd.getAttribute('name'), 
                                           nd.getAttribute('activationTime'))
                else:
                    tfpg_st.set_node_state(nd.getAttribute('name'), "0.0")

            for eg in edges:
                if eg.getAttribute('isActive') == 'true':
                    tfpg_st.set_edge_state(eg.getAttribute('name'), True)
                else:
                    tfpg_st.set_edge_state(eg.getAttribute('name'), False)

            tfpg_trace.add_tfpg_state(tfpg_st)

        return tfpg_trace

pass # end of class
