# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================

import logging
import os
import functools
import re
import gobject

from tfpg_viewer.src.utils import options
import tfpg_viewer.src.utils.misc

__all__ = ["system", "user", "testing", "metrics", "init", 
           "get_level_from_string"]

# Initially, a simple logging is offered for system
system = logging.getLogger("System")  # global system logger
testing = logging.getLogger("Testing")  # global logger used for testing
metrics = logging.getLogger("Metrics")  # global logger used for metrics
user = None  # global user logger


def init():
    """Initialises the logging system according to data in
    options. Note: options must be already set before calling it"""
    global system
    global testing
    global metrics
    global user
    
    # ----------------------------------------
    #    System logger 
    # ----------------------------------------
    system = _SystemLogger()

    # ----------------------------------------
    #    Testing logger 
    # ----------------------------------------
    testing = _TestingLogger()
    
    # ----------------------------------------
    #    Metrics logger 
    # ----------------------------------------
    metrics = _MetricsLogger()

    # ----------------------------------------
    #    User logger 
    # ----------------------------------------
    user = _UserLogger()    
    return


def get_level_from_string(level_name):
    """Given a string, returns the corresponding logging
    level. Raise ValueError if bad string is passed"""
    if not level_name: return logging.INFO  # default

    LEVELS = {'debug': logging.DEBUG,
              'info': logging.INFO,
              'warning': logging.WARNING,
              'error': logging.ERROR,
              'critical': logging.CRITICAL}

    if level_name not in LEVELS: 
        raise ValueError("Logging level must be in '%s'" % \
                         ", ".join(LEVELS.keys()))

    return LEVELS[level_name]


def log_icd(func):
    """This is a decorator which can be used to mark functions
    as logged for the ICD compliance testing. It will result 
    in a message to the testing logger when the function is used."""
    
    @functools.wraps(func)
    def new_func(*args, **kwargs):
        testing.info(
            "Call to ICD function %(funcname)s." % {
            'funcname': func.__module__ + "." + func.__name__, }
        )
        return func(*args, **kwargs)
    return new_func	


# -----------------------------------------------------------
class _Logger (object):
    def __init__(self, logger_name):
        self.logger = logging.getLogger(logger_name)
        return
    
    pass  # end of class


# -----------------------------------------------------------
class _SystemLogger (_Logger):
    def __init__(self, suffix_name=""):
        _Logger.__init__(self, "System" + suffix_name)

        self.logger.setLevel(logging.DEBUG)
        # console handler
        ch = logging.StreamHandler()
        ch.setLevel(logging.ERROR)
        formatter = logging.Formatter("%(asctime)s - %(name)s - "
                                      "%(levelname)s - %(message)s")
        ch.setFormatter(formatter)        
        self.logger.addHandler(ch)

        if options.LOGGING_FNAME: logfname = options.LOGGING_FNAME
        else: logfname = tfpg_viewer.src.utils.misc.get_out_fname(
                                            options.BASE_FILE_NAME + ".log")
        
        # file handler (always)
        fch = logging.FileHandler(logfname, "w")
        fch.setLevel(get_level_from_string(options.LOGGING_SYS_LEVEL))
        formatter = logging.Formatter("%(asctime)s - %(name)s - "
                                      "%(levelname)s - %(message)s")
        fch.setFormatter(formatter)        
        self.logger.addHandler(fch)
        return

    def debug(self, msg, *args, **kwargs): 
        self.logger.debug(msg, *args, **kwargs)
        return
    
    def info(self, msg, *args, **kwargs): 
        self.logger.info(msg, *args, **kwargs)
        return
    
    def warning(self, msg, *args, **kwargs): 
        self.logger.warning(msg, *args, **kwargs)
        return
    
    def error(self, msg, *args, **kwargs): 
        self.logger.error(msg, *args, **kwargs)
        return
    
    def critical(self, msg, *args, **kwargs): 
        self.logger.critical(msg, *args, **kwargs)
        return
    
    pass  # end of class


# -----------------------------------------------------------
class _TestingLogger (_Logger):
    def __init__(self, suffix_name=""):
        _Logger.__init__(self, "Testing" + suffix_name)

        self.logger.setLevel(logging.DEBUG)
        # console handler
        ch = logging.StreamHandler()
        ch.setLevel(logging.ERROR)
        formatter = logging.Formatter("%(asctime)s - %(name)s - "
                                      "%(levelname)s - %(message)s")
        ch.setFormatter(formatter)        
        self.logger.addHandler(ch)

        if options.LOGGING_FNAME: logfname = options.LOGGING_FNAME
        else: logfname = tfpg_viewer.src.utils.misc.get_out_fname(
                                            options.BASE_FILE_NAME + ".log")
        
        # file handler (always)
        fch = logging.FileHandler(logfname, "w")
        fch.setLevel(get_level_from_string(options.LOGGING_TEST_LEVEL))
        formatter = logging.Formatter("%(asctime)s - %(name)s - "
                                      "%(levelname)s - %(message)s")
        fch.setFormatter(formatter)        
        self.logger.addHandler(fch)
        return

    def debug(self, msg, *args, **kwargs): 
        self.logger.debug(msg, *args, **kwargs)
        return
    
    def info(self, msg, *args, **kwargs): 
        self.logger.info(msg, *args, **kwargs)
        return
    
    def warning(self, msg, *args, **kwargs): 
        self.logger.warning(msg, *args, **kwargs)
        return
    
    def error(self, msg, *args, **kwargs): 
        self.logger.error(msg, *args, **kwargs)
        return
    
    def critical(self, msg, *args, **kwargs): 
        self.logger.critical(msg, *args, **kwargs)
        return
    
    pass  # end of class


# -----------------------------------------------------------
class _MetricsLogger (_Logger):
    def __init__(self, suffix_name=""):
        _Logger.__init__(self, "Metrics" + suffix_name)

        self.logger.setLevel(logging.DEBUG)
        # console handler
        ch = logging.StreamHandler()
        ch.setLevel(logging.ERROR)
        formatter = logging.Formatter("%(asctime)s - %(name)s - "
                                      "%(levelname)s - %(message)s")
        ch.setFormatter(formatter)        
        self.logger.addHandler(ch)

        if options.LOGGING_FNAME: logfname = options.LOGGING_FNAME
        else: logfname = tfpg_viewer.src.utils.misc.get_out_fname(
                                            options.BASE_FILE_NAME + ".log")
        
        # file handler (always)
        fch = logging.FileHandler(logfname, "w")
        fch.setLevel(logging.INFO)
        formatter = logging.Formatter("%(asctime)s - %(name)s - "
                                      "%(levelname)s - %(message)s")
        fch.setFormatter(formatter)        
        self.logger.addHandler(fch)
        return

    def debug(self, msg, *args, **kwargs): 
        self.logger.debug(msg, *args, **kwargs)
        return
        
    def info(self, msg, *args, **kwargs): 
        self.logger.info(msg, *args, **kwargs)
        return
        
    def warning(self, msg, *args, **kwargs): 
        self.logger.warning(msg, *args, **kwargs)
        return
        
    def error(self, msg, *args, **kwargs): 
        self.logger.error(msg, *args, **kwargs)
        return
        
    def critical(self, msg, *args, **kwargs): 
        self.logger.critical(msg, *args, **kwargs)
        return
    
    pass  # end of class

# -----------------------------------------------------------


class _UserLogger (_Logger):
    def __init__(self):
        _Logger.__init__(self, "User")

        # sets up the logger
        self.logger.setLevel(logging.INFO)

        # console handler
        ch = logging.StreamHandler()
        ch.setLevel(get_level_from_string(options.LOGGING_USER_LEVEL))
        formatter = logging.Formatter("%(message)s")
        ch.setFormatter(formatter)
        self.logger.addHandler(ch)

        # file handler (only when required)
        if options.LOGGING_FNAME:
            fch = logging.FileHandler(options.LOGGING_FNAME, "w")
            fch.setLevel(get_level_from_string(options.LOGGING_SYS_LEVEL))
            formatter = logging.Formatter("%(asctime)s - %(name)s - "
                                          "%(levelname)s - %(message)s")
            fch.setFormatter(formatter)
            self.logger.addHandler(fch)
            pass        

        return

    def info(self, msg, *args, **kwargs): 
        self.logger.info(msg, *args, **kwargs)
        return
    
    def warning(self, msg, *args, **kwargs): 
        self.logger.warning(msg, *args, **kwargs)
        return
        
    def error(self, msg, *args, **kwargs): 
        self.logger.error(msg, *args, **kwargs)
        return
        
    pass  # end of class


# -----------------------------------------------------------
class GuiHandler (logging.Handler):
    """This class is used to send logging info to a gtk.textview
    (in a GUI)"""

    def __init__(self, text_buffer, level=logging.NOTSET):
        """text)buffer is a gtk.TextBuffer instance to be used to
        send log records. It must support tag named 'timestamp' and
        'level'"""
        logging.Handler.__init__(self, level)
        formatter = logging.Formatter("[%(asctime)s] %(levelname)s: "
                                      "\t%(message)s")
        self.setFormatter(formatter)

        self.textbuf = text_buffer

        self.re_stdout = re.compile("([\w\s()\d-]+\sstdout\s:)\s+(.*)$")
        self.re_stderr = re.compile("([\w\d-]+\sstderr\s:)\s+(.*)$")
        return

    def emit(self, record):
        def callback(buf, msg): 
            # separates the timestamp and the level
            
            # parts is a list of couple (text, []), tags can be empty
            parts = []

            if msg[0] == '[':
                idx = msg.find(']') + 1
                if idx > 0: 
                    parts.append((msg[0:idx], ["timestamp"]))
                    idx2 = msg.find(':', idx) + 1
                    if idx2 > 0:
                        parts.append((msg[idx:idx2], ["level"]))
                        parts.append((msg[idx2:], []))
                    else:  # no match for the level
                        parts.append((msg[idx:], []))
                        pass
                    pass
                
                # no match for timestamp
                else: parts.append((msg, []))
                
            else:  # no match for timestamp
                parts.append((msg, []))
                pass
            
            # search for paticular patterns
            _parts = []
            for _msg, tags in parts:
                if "Sending command" in _msg: 
                    _parts.append((_msg, tags + ["command"]))
                else:
                    for _re, tagpref in ((self.re_stdout, "stdout"), 
                                         (self.re_stderr, "stderr"),
                                         ):                        
                        m = _re.search(_msg)
                        if m:
                            _parts.append((m.group(1), 
                                           tags + [tagpref, tagpref + "msg"]))
                            _parts.append((m.group(2) + "\n", 
                                           tags + [tagpref + "msg"]))
                            break                 
                        pass
                    if m is None: _parts.append((_msg, tags))  # no match
                    pass
                pass

            for _msg, tags in _parts:
                if tags: buf.insert_with_tags_by_name(buf.get_end_iter(), 
                                                      _msg, *tags)
                else: buf.insert(buf.get_end_iter(), _msg)
                pass
            
            return
        
        # this is done to protect from multithreading calls
        msg = self.format(record)
        gobject.idle_add(callback, self.textbuf, msg + os.linesep)
        return

    pass  # end of class
