# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================

from __future__ import division
from gtkmvc import View
import subprocess
import time
import os.path
import gtk
import networkx as nx
import pkg_resources
pkg_resources.require("matplotlib>=1.x.x")
import matplotlib.pyplot as plt
from matplotlib.backends.backend_gtkagg import FigureCanvasGTKAgg as FigureCanvas
from matplotlib.backends.backend_gtkagg import NavigationToolbar2GTKAgg as NavigationToolbar
import matplotlib.gridspec as gridspec
import tfpg_viewer.src.utils.misc
import tfpg_viewer.src.utils.tfpg_trace_description
from hasdel_tools.src.tfpg.tfpg import Edge, Node, NodeType
from hasdel_tools.src.tfpg.tfpg_errors import NotExistingNode


class TFPGView(View):
    """View for the TFPG graph visualization"""

    from tfpg_viewer.src.utils.globals import GLADE_DIR

    builder = os.path.join(GLADE_DIR, "tfpg_glade.glade")
    top = "window_top"


    def __init__(self, standalone=False):
        ## TODO: check comment
        """Class constructor

        There are some class variables wich are used.
        - nx_graph: represents the networkx graph structure which will be
          graphically represented. In this way, it can be easily updated
          when redrawings are required
        - pos: it is a dictionary containing the positions of every nx_graph
          node; in this way, it is not required to get it when updating
          the nx_graph
        - component_clicked: as the user can interact with the graph by clicking
          over one of its components, it is important that all the functions are aware of
          the last selected component

        The class variables dpi, axis_x_left_margin, axis_x_right_margin, axis_y_top_margin 
        and axis_y_bottom_margin are used in order to provide a better graphical represenation
        of the graph, allowing the user to zoom in/out, or to move it.

        """

        View.__init__(self)
        self.standalone = standalone

        if self.standalone:
            self.init_graphical_components()


        ## GZ: I think that it is important also to keep a global internal struct representation,
        ##     since we want that there is always a 1-1 correspondence between the internal
        ##     struct and the graphical representation
        ##     TODO: maybe, it could be useful to have two function to convert a tfpg_struc
        ##     to a graphical_struct and viceversa
        self.tfpg_graph = None

        ## GZ: not sure that nx_graph must be a global variable; I think it would be better
        ##     to modify (if needed) the internal tfpg structure and get the corresponding
        ##     graphical representation from it
        # self.nx_graph = None

        ## GZ: self.pos need to be global, since it is used in the zooming functions
        self.pos = None
        ## GZ: sure we need it? It is used only by the onClick() function
        self.component_clicked = None

        ## Variables used for the zooming function
        ## TODO: when zooming also node size/distance and font/size should be in(de)creased 
        self.dpi = None

        self.axis_x_left_margin = None
        self.axis_x_right_margin = None
        self.axis_y_bottom_margin = None
        self.axis_y_top_margin = None

        self.zoom_offset = 1
        
        ## GZ: added by me on 19.12.2013 -> notice that probably all the values used by the
        ## zooming functions could be passed as parameters
        self.node_size = 3000

        return



    def run(self):
        res = self.get_top_widget().run()
        self.get_top_widget().destroy()
        return res



    def init_graphical_components(self):
        """Function used to initialize some graphical components; it does not seem possible
        to do the same uperation using the glade editor.

        For instance, from glade it is not possible to create a button without
        label, add some system configurations need to be changed.

        
        Notice that this function is called only in case the viewer is run in standalone
        mode.
        """

        image = gtk.Image()
        image.set_from_stock(gtk.STOCK_MEDIA_PLAY, gtk.ICON_SIZE_SMALL_TOOLBAR)
        btn_play = self['btn_run_trace']
        btn_play.set_label("")
        btn_play.set_image(image)

        image2 = gtk.Image()
        image2.set_from_stock(gtk.STOCK_MEDIA_NEXT, gtk.ICON_SIZE_SMALL_TOOLBAR)
        btn_forward = self['btn_next_state']
        btn_forward.set_label("")
        btn_forward.set_image(image2)

        image3 = gtk.Image()
        image3.set_from_stock(gtk.STOCK_MEDIA_PREVIOUS, gtk.ICON_SIZE_SMALL_TOOLBAR)
        btn_prev = self['btn_prev_state']
        btn_prev.set_label("")
        btn_prev.set_image(image3)

        ## TODO: functions for moving the graph should be retrieved from the NavigationToolbar
        ##       provided by networkx like we have done in the GUI
        image4 = gtk.Image()
        image4.set_from_stock(gtk.STOCK_GO_BACK, gtk.ICON_SIZE_SMALL_TOOLBAR)
        btn_move_left = self['btn_move_left']
        btn_move_left.set_label("")
        btn_move_left.set_image(image4)

        image5 = gtk.Image()
        image5.set_from_stock(gtk.STOCK_GO_FORWARD, gtk.ICON_SIZE_SMALL_TOOLBAR)
        btn_move_right = self['btn_move_right']
        btn_move_right.set_label("")
        btn_move_right.set_image(image5)

        image6 = gtk.Image()
        image6.set_from_stock(gtk.STOCK_GO_UP, gtk.ICON_SIZE_SMALL_TOOLBAR)
        btn_move_up = self['btn_move_top']
        btn_move_up.set_label("")
        btn_move_up.set_image(image6)

        image7 = gtk.Image()
        image7.set_from_stock(gtk.STOCK_GO_DOWN, gtk.ICON_SIZE_SMALL_TOOLBAR)
        btn_move_down = self['btn_move_down']
        btn_move_down.set_label("")
        btn_move_down.set_image(image7)

        image8 = gtk.Image()
        image8.set_from_stock(gtk.STOCK_ZOOM_IN, gtk.ICON_SIZE_SMALL_TOOLBAR)
        btn_zoom_in = self['btn_zoom_in']
        btn_zoom_in.set_label("")
        btn_zoom_in.set_image(image8)

        image9 = gtk.Image()
        image9.set_from_stock(gtk.STOCK_ZOOM_OUT, gtk.ICON_SIZE_SMALL_TOOLBAR)
        btn_zoom_out = self['btn_zoom_out']
        btn_zoom_out.set_label("")
        btn_zoom_out.set_image(image9)

        # system configurations required to let gtk show images on buttons
        settings = gtk.settings_get_default()
        settings.props.gtk_button_images = True

        return



    ## GZ: this function could be named convert internal_struct to networkx_graph
    # def populate_networkx_graph(self, tfpg_graph):
    def get_updated_networkx_graph(self):
        # """Creates the networkx graph structure corresponding to tfpg_graph

        # Creates a nx.MultiDiGraph corresponding to tfpg_graph. It 
        # basically intializes the nx_graph class variable.

        # :param tfpg_graph: the graph to be represented
        # :type tfpg_graph: TFPGStructural

        # """
        """It returns the graphical representation of the internal representation
           of the TFPG. Notice that for seeing graphical changes, you need first to
           update the internal struct, and then to all this function and use the
           returned value.
        """

        nx_graph = nx.MultiDiGraph(name="graph")

        for nd in self.tfpg_graph.get_nodes_list():

            if nd.get_type() == "FM":
                shape = 's'
                nd_type = "Failure Mode"
            elif nd.get_type() == "AND":
                shape = 's'
                nd_type = "And Discrepancy"
            elif nd.get_type() == "OR":
                shape = 'o'
                nd_type = "Or Discrepancy"

            if nd.is_monitored():
                width = 6.0
            else:
                width = 1.0

            nx_graph.add_node(nd.get_name(), node_type=nd_type,
                              node_shape=shape, node_width=width,
                              node_label=nd.get_name(), node_color='white',
                              node_act_time="")
            
        for eg in self.tfpg_graph.get_edges_list():

            tMin = eg.get_tMin()
            tMax = eg.get_tMax()
            if float(tMax) == -1.0:
                tMax = r"$\infty$"
            eg_lbl = "[%s, %s] {%s}" %(tMin, tMax, ', '.join(eg.get_modes()))

            nx_graph.add_edge(eg.get_srcNode().get_name(), eg.get_destNode().get_name(), 
                              key=eg.get_name(),
                              edg_label=eg_lbl, edg_color='black', edg_act_time="",
                              edg_width=1.0)

        if self.pos == None:
            self.pos = nx.graphviz_layout(nx_graph,prog='dot')
        
        # for each node and each edge, a label is added to contain its
        # activation time
        for nd in self.tfpg_graph.get_nodes_list():
            nx_graph.add_node(nd.get_name() + "_nd_lbl") 
            nd_pos = self.pos[nd.get_name()]
            self.pos[nd.get_name() + "_nd_lbl"] = (nd_pos[0] + 15, nd_pos[1])

        for eg in self.tfpg_graph.get_edges_list():
            nx_graph.add_node(eg.get_name() + "_edg_lbl")
            eg_src_pos = self.pos[eg.get_srcNode().get_name()]
            eg_dst_pos = self.pos[eg.get_destNode().get_name()]
            eg_lbl_pos = (((abs(eg_src_pos[0]+eg_dst_pos[0]))/2)+10,
                          ((abs(eg_src_pos[1]+eg_dst_pos[1]))/2)-5)
            self.pos[eg.get_name() + "_edg_lbl"] = eg_lbl_pos

        return nx_graph



    def set_limits(self):
        """"Sets the bounds of the figure which contains the graphical representation.
        Once the nx_graph internal structure has been populated, this function reads 
        the positions of all the nodes, and takes the minumum and maximum values for
        the x and the y ranges.
        This is useful to properly center the figure in the window.

        """

        ymin = 1000
        ymax = None
        xmin = 1000
        xmax = None

        for (x, y) in self.pos.values():

            if x < xmin: xmin = x
            if x > xmax: xmax = x
            if y < ymin: ymin = y
            if y > ymax: ymax = y

        self.axis_x_left_margin = xmin-20
        self.axis_x_right_margin = xmax+20
        self.axis_y_bottom_margin = ymin-20
        self.axis_y_top_margin = ymax+20
        
        return


    
    ## GZ: sure parameter 'update' is required??
    # def create_tfpg_structural_figure(self, tfpg_graph):
    def create_tfpg_canvas_figure(self, tfpg_graph=None, update=False):
        # """It creates the structural figure for a given tfpg.

        # Basically, the main window is divided in three zones, one for the 
        # real representation of the graph, one for a simple caption and 
        # another one for the (possible) trace informations.

        # Moreover, the figure is done interactive, i.e. it is possible to click
        # on nodes and edges in order to display some related info.

        # Notice that when we want to update the shown graph (i.e. when zooming, moving it...)
        # we call this function with tfpg_graph = None; in this way, the positions of the
        # figure are not recomputed and the already populated internatl structure
        # is not recreated.

        # :param tfpg_graph: the structural tfpg to be represented
        # :type tfpg_graph: TfpgStructural
        # """
        """It creates the canvas figure (to be added to a GUI element) corresponding to
           the global TFPG internal representation
        """
        
        ## tfpg_graph is not None only the first time when a TFPG is read from file
        ## and it must be initialized
        if tfpg_graph != None:
            self.tfpg_graph = tfpg_graph

        # NB: the following dpi number (used to find a proper scale value for the 
        # representation) is the result of some (little) empirical evaluation
        # TODO: test it!
        # if self.tfpg_graph != None and self.tfpg_graph.get_nodes_list() != []:
        if self.dpi == None:
            node_nbr = len(self.tfpg_graph.get_nodes_list())
            self.dpi = ((1/node_nbr)*428)
        # if self.dpi > 60:
        #     self.dpi = 60

        fig = plt.figure(1, figsize=(15,12))
        fig.clf()
        fig.set_dpi(self.dpi)

        if self.standalone:
            gs = gridspec.GridSpec(2, 3,
                                   width_ratios=[1,4,1],
                                   height_ratios=[5,1]
                                   )

            graphical_rep = plt.subplot(gs[0:2],autoscalex_on=True,autoscaley_on=True)

        else:
            graphical_rep = plt.subplot(111)

        # if tfpg_graph != None:
        # if update is False:
        # self.populate_networkx_graph(tfpg_graph)
        nx_graph = self.get_updated_networkx_graph()
        ## GZ: be sure that tfpg_graph != None only in case it is the first time
        if tfpg_graph != None: self.set_limits()

        graphical_rep.set_xlim(self.axis_x_left_margin, self.axis_x_right_margin)
        graphical_rep.set_ylim(self.axis_y_bottom_margin, self.axis_y_top_margin)

        graphical_rep = self.draw_graph(nx_graph)

        if self.standalone:
            el_overview = plt.subplot(gs[2])
            plt.axis('off')
            if self.component_clicked != None:
                el_overview.text(0,1,"Element Info:",
                                 fontsize=2*len(self.tfpg_graph.get_nodes_list()))
                el_overview.text(0,0.8,self.print_component_info(self.component_clicked),
                                 fontsize=2*len(self.tfpg_graph.get_nodes_list()))
            else:
                el_overview.text(0,1,"")
                el_overview.text(0,0.8,"")

        canvas = FigureCanvas(fig)

        if self.standalone:
            fig.canvas.mpl_connect('motion_notify_event', self.onmotion)
            fig.canvas.mpl_connect('button_press_event', self.onclick)

        return canvas


    
    def create_tfpg_state_figure(self, tfpg_graph, tfpg_state, nx_graph, write_info=False):
        """Adds the state information to the networkx graph structure.

        The class variable nx_graph is updated with the information read
        from tfpg_state, such as the activation time or the flag to 
        understand whether the node is activer or not.

        Notice that the parameter tfpg_graph is passed as it contains the
        structural information for the state

        :param tfpg_graph: the structural graph 
        :type tfpg_graph:  TFPGStructural

        :param tfpg_state: the state to be shown
        :type tfpg_state: TFPGState

        """
        
        ## TODO: check the changes done to this function
        
        ## For now, states come into consideration only if the tfpg_viewer is run in
        ## standalone mode
        if self.standalone:

            if write_info:
                tfpg_viewer.src.utils.tfpg_trace_description.tfpg_trace_writer.write(
                    "====================== beginning of state ====================\n")

            nodes_state = tfpg_state.get_nodes_state()
            nodes_act_time = tfpg_state.get_nodes_act_time()
            edges_state = tfpg_state.get_edges_state()
            edges_act_time = tfpg_state.get_edges_act_time()

            for nd in tfpg_graph.get_nodes_list():

                if nodes_state[nd.get_name()] == True:
                    color = "red"
                    act_time = "t = " + str(nodes_act_time[nd.get_name()])
                    if write_info:
                        tfpg_viewer.src.utils.tfpg_trace_description.tfpg_trace_writer.node_activated(nd.get_name(), act_time)
                else:
                    color = "white"
                    act_time = ""

                nx_graph.node[nd.get_name()]['node_color'] = color
                nx_graph.node[nd.get_name()]['node_act_time'] = act_time

            for eg in tfpg_graph.get_edges_list():

                if edges_state[eg.get_name()] == True:
                    color = 'yellow'
                    width = 3.0
                    act_time = "t = " + str(edges_act_time[eg.get_name()])
                    if write_info:
                        tfpg_viewer.src.utils.tfpg_trace_description.tfpg_trace_writer.edge_activated(eg.get_name(), act_time)
                else:
                    color = 'black'
                    width = 1.0
                    act_time = ""

                src_nd = eg.get_srcNode().get_name()
                dst_nd = eg.get_destNode().get_name()
                name = eg.get_name()
                nx_graph.edge[src_nd][dst_nd][name]['edg_color'] = color
                nx_graph.edge[src_nd][dst_nd][name]['edg_width'] = width
                nx_graph.edge[src_nd][dst_nd][name]['edg_act_time'] = act_time

                if write_info:
                    tfpg_viewer.src.utils.tfpg_trace_description.tfpg_trace_writer.write(
                        "====================== end of state ==========================\n\n")

        return nx_graph


    ## TODO: sure we need tfpg_graph as parameter? I would say no
    def pyplot_view_structural(self, tfpg_graph=None, component_name='hbox_viewing_area'):
        """Shows a structural tfpg using pyplot

        The figure is plotted to the 'hbox_viewing_area' component
        (if no other component is specified)
        of the user interface.
        Basically, the following steps are performed:
        - creation and/or drawing of the internal nx_graph structure
        - creation of the caption
        - division of the output figure such that it is able to
          display both information
        - addition of some mouse events which let the user
          interactively operate with the representation

        :param tfpg_graph: the structural tfpg to be shown
        :type tfpg_graph: TFPGStructural

        """

        # canvas = self.create_tfpg_structural_figure(tfpg_graph)
        canvas = self.create_tfpg_canvas_figure(tfpg_graph=tfpg_graph)

        hbox = self[component_name]

        for ch in hbox.get_children():
            hbox.remove(ch)

        hbox.add(canvas)
        hbox.show_all()

        return



    ## TODO: check which are the needed parameters
    def pyplot_view_state(self, tfpg_graph, tfpg_state, write_info=False):
        """"Shows a tfpg state using pyplot

        The figure is plotted to the 'hbox_viewing_area' component
        of the user interface.

        Additionally to pyplot_view_structural, it outputs also
        trace-related information, like global_time and mode.

        Moreover, it write_info is set to True (not by default), trace
        information is written to file in a textual description 

        :param tfpg_graph: the structural tfpg
        :type tfpg_graph: TFPGStructural

        :param tfpg_state: the tfpg state
        :type tfpg_state: TFPGState


        """

        node_nbr = len(tfpg_graph.get_nodes_list())
        dpi = ((1/node_nbr)*428)

        fig = plt.figure(1, figsize=(15,12),dpi=dpi)
        fig.clf()

        if self.standalone:
            gs = gridspec.GridSpec(2, 3,
                                   width_ratios=[1,4,1],
                                   height_ratios=[5,1]
                                   )
            graphical_rep = plt.subplot(gs[0:2],autoscalex_on=True,autoscaley_on=True)

        else:
            graphical_rep = plt.subplot(111)

        ## TODO: it is dummy to call a function with a global variable as parameter
        nx_graph = self.get_updated_networkx_graph()
        nx_graph = self.create_tfpg_state_figure(self.tfpg_graph, tfpg_state, nx_graph, write_info=write_info)
        graphical_rep = self.draw_graph(nx_graph)

        info_fontsize = 2*len(self.tfpg_graph.get_nodes_list())
        info = plt.subplot(gs[3], autoscalex_on=True, autoscaley_on=True)
        plt.axis('off')
        info.set_title('Trace Information', fontsize=info_fontsize)
        info.text(0, 0.6, "Global time: " + str(tfpg_state.get_global_time()),
                  fontsize=info_fontsize)
        info.text(0, 0.4, "Mode: " + tfpg_state.get_mode(), fontsize=info_fontsize)
     
        canvas = FigureCanvas(fig)

        hbox = self['hbox_viewing_area']

        for ch in hbox.get_children():
            hbox.remove(ch)

        hbox.add(canvas)
        fig.canvas.mpl_connect('motion_notify_event', self.onmotion)
        fig.canvas.mpl_connect('button_press_event', self.onclick)
        hbox.show_all()

        return



    ## TODO: check which are the needed parameters
    def pyplot_view_trace(self, tfpg_graph, tfpg_trace, write_info=False):
        """"Shows a tfpg trace using pyplot.

        The figure is plotted to the hbox_viewing_area component
        of the user interface.

        If write_info is set to True (not the default), a file called
        trace_execution_flow.txt is written. This file contains a
        textual description of trace execution.

        :param tfpg_graph: the structural tfpg
        :type tfpg_graph: TFPGStructural

        :param tfpg_trace: the tfpg trace
        :type tfpg_trace: TFPGTrace


        """
 
        if write_info:
            tfpg_viewer.src.utils.tfpg_trace_description.init()
            tfpg_viewer.src.utils.tfpg_trace_description.tfpg_trace_writer.start()

        for tfpg_state in tfpg_trace.get_tfpg_states():
           
            if tfpg_state == tfpg_trace[0]:
                ## TODO: check whether this call is correct
                self.pyplot_view_structural(tfpg_graph=self.tfpg_graph)
                while gtk.events_pending():
                    gtk.main_iteration_do(block=False)
                time.sleep(2)

            self.pyplot_view_state(tfpg_graph, tfpg_state, write_info=write_info)

            while gtk.events_pending():
                gtk.main_iteration_do(block=False)
            time.sleep(2)

        if write_info: tfpg_viewer.src.utils.tfpg_trace_description.tfpg_trace_writer.terminate()

        return




    ## GZ: this function should be called with the graph
    # def draw_graph(self):
    def draw_graph(self, nx_graph):
        """Draws the internal graph using networkx library.

        It draws the nx_graph class variable. In this way, when
        this structure is modified, to graphically see the changes
        it is possible to call this function.
        """

        fig = plt.figure(1)

        nodes_lbl_list = dict()
        edges_lbl_list = dict()


        for nd in nx_graph.nodes(data=True):
            
            if "lbl" in nd[0]:
                nx.draw_networkx_nodes(nx_graph, self.pos, 
                                       nodelist=[nd[0]], 
                                       node_shape='s',
                                       # node_size=3000*self.zoom_offset,
                                       node_size = self.node_size,
                                       alpha=0.0)

            # tmp_annotation is the node which is shown in case the user clicks over a node
            # or an edge in order to display its information
            # elif "tmp_annotation" == nd[0]:

            #     nx.draw_networkx_nodes(nx_graph, self.pos, nodelist=[nd[0]],
            #                            node_shape='s',
            #                            node_size=(len(self.tfpg_graph.get_nodes_list())*758)*self.zoom_offset,
            #                            node_color='white')
            #     nx.draw_networkx_labels(nx_graph, self.pos, labels={nd[0]:nd[1]['text']},
            #                             font_size=len(self.tfpg_graph.get_nodes_list()))

            else:
                nx.draw_networkx_nodes(nx_graph, self.pos, nodelist=[nd[0]],
                                       node_shape=nd[1]['node_shape'],
                                       # node_size=3000*self.zoom_offset,
                                       node_size = self.node_size,
                                       node_color=nd[1]['node_color'],
                                       linewidths=nd[1]['node_width'])
                nodes_lbl_list[nd[0]] = nd[1]['node_label']
                nodes_lbl_list[nd[0] + "_nd_lbl"] = nd[1]['node_act_time']

        for edg in nx_graph.edges(data=True,keys=True):

            nx.draw_networkx_edges(nx_graph, self.pos, 
                                   edgelist=[(edg[0], edg[1])], 
                                   edge_color=edg[3]['edg_color'], 
                                   width=edg[3]['edg_width'])
            edges_lbl_list[(edg[0], edg[1])] = edg[3]['edg_label']
            nodes_lbl_list[edg[2] + "_edg_lbl"] = edg[3]['edg_act_time']

        # for k,v in nodes_lbl_list.iteritems():
        #     if "lbl" not in k:
        #         if len(v) > 6:
        #             nodes_lbl_list[k] = v[:6] + "..."

        nx.draw_networkx_labels(nx_graph, self.pos, labels=nodes_lbl_list,
                                font_size=10*self.zoom_offset)
    
        nx.draw_networkx_edge_labels(nx_graph,self.pos,edge_labels=edges_lbl_list,
                                     bbox=dict(boxstyle='round',ec='black',fc='white'),
                                     font_size=10*self.zoom_offset)
        plt.axis('off')
        return fig




    def create_caption(self):
        """Creates the caption for a given trace and/or graph"""

        fig = plt.figure(2)
        label_list = dict()
 
        graph = nx.MultiDiGraph(name="graph")

        # Adding failure mode/And discrepancy node
        graph.add_node("failure_mode", node_shape='s', node_alpha=1.0,
                       node_width=1.0, node_color='w')

        pos = nx.graphviz_layout(graph,prog='dot')
        fm_pos = pos["failure_mode"]

        # Adding an OR discrepancy node
        pos["or_discrepancy"] = (fm_pos[0], fm_pos[1]-1)
        graph.add_node("or_discrepancy", node_shape='o', node_alpha=1.0,
                       node_width=1.0, node_color='w')

        # Adding a monitored node
        pos["monitored_node"] = (fm_pos[0], fm_pos[1]-2)
        graph.add_node("monitored_node", node_shape='s', node_alpha=1.0,
                       node_width=3.0, node_color='w')

        # Adding failure mode/And discrepancy label node
        pos["failure_mode_lbl"] = (fm_pos[0]+3, fm_pos[1])
        graph.add_node("failure_mode_lbl", node_shape='s', node_alpha=0.0,
                       node_width=0.0, node_color='w')
        label_list["failure_mode_lbl"] = "Failure Mode\nAnd Discrepancy"

        #Adding or discrepancy label node
        pos["or_discrepancy_lbl"] = (fm_pos[0]+3, fm_pos[1]-1)
        graph.add_node("or_discrepancy_lbl", node_shape='s', node_alpha=0.0,
                       node_width=0.0, node_color='w')
        label_list["or_discrepancy_lbl"] = "Or Discrepancy"

        # Adding monitored node label
        pos["monitored_node_lbl"] = (fm_pos[0]+3, fm_pos[1]-2)
        graph.add_node("monitored_node_lbl", node_shape='s', node_alpha=0.0,
                       node_width=0.0, node_color='w')
        label_list["monitored_node_lbl"] = "Monitored node\n(general)"

        #Adding an edge
        pos["tmp_node_A"] = (fm_pos[0]-1, fm_pos[1]-4)
        graph.add_node("tmp_node_A", node_shape='o', node_alpha=1.0,
                       node_width=1.0, node_color='w')
        label_list["tmp_node_A"] = "A"
        pos["tmp_node_B"] = (fm_pos[0]+4, fm_pos[1]-4)
        graph.add_node("tmp_node_B", node_shape='o', node_alpha=1.0,
                       node_width=1.0, node_color='w')
        label_list["tmp_node_B"] = "B"
        graph.add_edge("tmp_node_A", "tmp_node_B")

        pos["edge_lbl"] = (fm_pos[0]+1.5, fm_pos[1]-5)
        graph.add_node("edge_lbl", node_shape='s', node_alpha=0.0,
                       node_width=0.0, node_color='w')
        label_list["edge_lbl"] = "Directed edge (from A to B)"

        if self.standalone:

            # Traces info
            pos["active_failure_mode"] = (fm_pos[0], fm_pos[1]-8)
            graph.add_node("active_failure_mode", node_shape='s', node_alpha=1.0,
                           node_width=1.0, node_color='r')

            pos["active_and_discrepancy"] = (fm_pos[0], fm_pos[1]-9)
            graph.add_node("active_and_discrepancy", node_shape='s', node_alpha=1.0,
                           node_width=1.0, node_color='r')

            pos["active_or_discrepancy"] = (fm_pos[0], fm_pos[1]-10)
            graph.add_node("active_or_discrepancy", node_shape='o', node_alpha=1.0,
                           node_width=1.0, node_color='r')

            pos["active_failure_mode_lbl"] = (fm_pos[0]+3, fm_pos[1]-8)
            graph.add_node("active_failure_mode_lbl", node_shape='s', node_alpha=0.0,
                           node_width=0.0, node_color='w')
            label_list["active_failure_mode_lbl"] = "Active FM"

            pos["active_and_discrepancy_lbl"] = (fm_pos[0]+3, fm_pos[1]-9)
            graph.add_node("active_and_discrepancy_lbl", node_shape='s', node_alpha=0.0,
                           node_width=0.0, node_color='w')
            label_list["active_and_discrepancy_lbl"] = "Active AND"
        
            pos["active_or_discrepancy_lbl"] = (fm_pos[0]+3, fm_pos[1]-10)
            graph.add_node("active_or_discrepancy_lbl", node_shape='s', node_alpha=0.0,
                           node_width=0.0, node_color='w')
            label_list["active_or_discrepancy_lbl"] = "Active OR"

        for nd in graph.nodes(data=True):

            shape = nd[1]['node_shape']
            color = nd[1]['node_color']
            width = nd[1]['node_width']
            alpha = nd[1]['node_alpha']
 
            nx.draw_networkx_nodes(graph, pos, nodelist=[nd[0]],
                                   node_shape=shape,
                                   node_size=1800,
                                   node_color=color,
                                   linewidths=width, alpha=alpha)

        for edg in graph.edges(data=True):
            nx.draw_networkx_edges(graph, pos, 
                                   edgelist=[(edg[0], edg[1])], 
                                   edge_color='black')
           
        nx.draw_networkx_labels(graph, pos, labels=label_list,
                                font_size=12)

        plt.axis("off")
        
        canvas = FigureCanvas(fig)

        return canvas



    # ------------------------------------------------------------------------
    # Mouse Listener Function
    # ------------------------------------------------------------------------



    def onclick(self, event):
        """Function called when the user presses the mouse button onto
        the graphical representation.

        Basically, if the user presses over a node or an edge, its information
        is shown by setting the global variable 'component_clicked' value to be the 
        selected component.
        """

        self.component_clicked = self.get_mouse_position(event)
        self.pyplot_view_structural()
            
        return



    def onmotion(self, event):
        """Function called when the user moves the mouse.
        
        It is used to check whether the user clicks over a node and/or 
        edge. Also the mouse icon is changed.
        """

        position = self.get_mouse_position(event)

        if position != None:
            self[self.top].get_window().set_cursor(gtk.gdk.Cursor(gtk.gdk.HAND2))
            
        else:
            self[self.top].get_window().set_cursor(gtk.gdk.Cursor(gtk.gdk.ARROW))

        return


    def get_mouse_position(self, event):
        """It checks whether the mouse is over a node or an edge.

        In case the mouse is over a node or an edge, it returns its name.
        Otherwise, it returns None.
        """

        for nd in self.pos.keys():

            (nd_x, nd_y) = self.pos[nd]

            if "edg" in nd:
                if event.xdata >= nd_x - 13 and event.xdata <= nd_x -8:
                    if event.ydata >= nd_y - 8 and event.ydata <= nd_y + 8:
                        return nd
            elif "nd_lbl" not in nd:
                if event.xdata >= nd_x - 5 and event.xdata <= nd_x + 5:
                    if event.ydata >= nd_y -15 and event.ydata <= nd_y + 15:
                        return nd

        return None

            
    def print_component_info(self, component, pretty_print=True):
        """Given a component, it returns its related information.

        This function is called exclusively in combination with the
        onclick method.

        If the var pretty_print is set to True (by default), a string
        containing the information in a pretty-print form is returned, 
        otherwise a dict is returned.

        For instance, in case component is an edge, the following dict
        is returned:
        {('name': <name>), ('src': <src>), ('dst': <dst>), ('tMin': <tMin>), ('tMax': <tMax>), 
         ('modes': <modes>)}

        """

        info = ""

        if "_edg_lbl" in component:
            edg_name = component[:component.find("_edg_lbl")]
            for edg in self.tfpg_graph.get_edges_list():
                ## TODO: check how it is contained tMax = INFTY
                if edg.get_name() == edg_name:

                    if pretty_print is True:
                        info = "   * Name: %s\n   * Src: %s\n   * Dest: %s\n   " \
                            "* tMin: %s\n   * tMax: %s\n   * Modes: %s" \
                            % (edg.get_name(), edg.get_srcNode().get_name(), 
                               edg.get_destNode().get_name(), 
                               edg.get_tMin(), edg.get_tMax(), ', '.join(edg.get_modes()))
                    else:
                        info = {'name': edg.get_name(), 'src': edg.get_srcNode().get_name(),
                                'dst': edg.get_destNode().get_name(), 'tMin': edg.get_tMin(),
                                'tMax': edg.get_tMax(), 'modes': ', '.join(edg.get_modes())}

                    
        else:
            for nd in self.tfpg_graph.get_nodes_list():
                if nd.get_name() == component:
                    
                    if pretty_print is True:
                        info = "   * Name: %s\n   * Type: %s\n   * Monitored: %s\n" \
                            %(nd.get_name(), nd.get_type(), nd.is_monitored())

                    else:
                        info = {'name': nd.get_name(), 'type': nd.get_type(),
                                'isMonitored': nd.is_monitored()}

        return info

        

    # ------------------------------------------------------------------------
    # UI Functions
    # ------------------------------------------------------------------------

    def ui_load_structural_file(self):
        """Function called when the user wants to load a structural file

        Essentially, it shows a FileChooserDialog in which choose the selected
        structural file
        """

        dialog = gtk.FileChooserDialog("Open..", #title
                                       None, #parent
                                       gtk.FILE_CHOOSER_ACTION_OPEN, #action
                                       (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                                        gtk.STOCK_OPEN, gtk.RESPONSE_OK) #buttons
                                       )

        dialog.set_default_response(gtk.RESPONSE_OK)


        filter = gtk.FileFilter()
        filter.set_name("structural files")
        filter.add_pattern("*.txml")
        filter.add_pattern("*.tfpg")
        dialog.add_filter(filter)

        response = dialog.run()
        if response == gtk.RESPONSE_OK:
            filename = dialog.get_filename()
            dialog.destroy()
            return filename

        dialog.destroy()
        return


    def ui_load_trace_file(self):
        """Function called when the user wants to load a trace file

        Essentially, it shows a FileChooserDialog in which choose the selected
        structural file
        """

        dialog = gtk.FileChooserDialog("Open..", #title
                                       None, #parent
                                       gtk.FILE_CHOOSER_ACTION_OPEN, #action
                                       (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                                        gtk.STOCK_OPEN, gtk.RESPONSE_OK) #buttons
                                       )

        dialog.set_default_response(gtk.RESPONSE_OK)


        filter = gtk.FileFilter()
        filter.set_name("trace files")
        filter.add_pattern("*.xml")
        dialog.add_filter(filter)

        response = dialog.run()
        if response == gtk.RESPONSE_OK:
            filename = dialog.get_filename()
        dialog.destroy()

        return filename


    def ui_save_structural_file(self, format):
        """Function called when the user wants to save a structural tfpg

        :param format: the format in which to save the structural tfpg
        :type format: {tfpg, xml}

        """

        if format not in "tfpg txml".split():
            raise Exception

        dialog = gtk.FileChooserDialog("Save as tfpg", #title
                                       None, #parent
                                       gtk.FILE_CHOOSER_ACTION_SAVE, #action
                                       (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                                        gtk.STOCK_SAVE, gtk.RESPONSE_OK) #buttons
                                       )
        dialog.set_default_response(gtk.RESPONSE_OK)
        dialog.set_do_overwrite_confirmation(True)

        filter = gtk.FileFilter()
        filter.set_name("tfpg files")
        filter.add_pattern("*.%s" % format)
        dialog.add_filter(filter)

        response = dialog.run()
        if response == gtk.RESPONSE_OK:
            out = dialog.get_filename()
            ext = ".%s" % format
            if not out.endswith(ext): out += ext
        else:
            out = None
        dialog.destroy()
        return out


    def ui_run_text_editor(self, filename, from_scratch=False):
        """Runs a text editor, if found, and returns its exit code

           If the parameter from_scratch is True, that means that the user is
           creating a new structural tfpg from scratch, i.e. we must show an
           empty file.

           :param filename: the name of the file to view in the text editor
           :type filename: string

           :param from_scratch: value specifying whether it is a new file
           :type from_scratch: boolean
        """

        for program in "emacs gedit".split():
            editor_name = tfpg_viewer.src.utils.misc.which(program)
            if editor_name: break

        if editor_name is None: raise Exception("Valid textual editor not found in the system")

        if from_scratch:
            with open(filename, "w") as f:
                f.write("\nNAME empty_tfpg\n\n")
                f.write("-- Definition of nodes\n")
                f.write("-- FM fm_node             -- A Failure Mode\n")
                f.write("-- AND and_node MONITORED -- An And Discrepancy\n")
                f.write("-- OR or_node             -- An Or Discrepancy\n\n")
                f.write("-- Definition of the edge modes\n")
                f.write("-- MODES a, b\n\n")
                f.write("-- Definition of edges\n")
                f.write("-- EDGE edge1 src_node dst_node min_time max_time (a,b)")

        proc = subprocess.Popen([editor_name, filename])
        proc.wait()

        return proc.returncode



    def ui_show_message_dialog(self, msg_type, msg_text, secondary_text=None, parent_name=None):
        """Function which simply shows a simple message dialog.

        :param msg_type: the type of the message to be shown
        :type msg_type: {info, error, warning}

        :param msg_text: the message text to be shown
        :type msg_text: string

        :param secondary_text: the (optional) secondary message text to be shown
        :type: string (default = None)

        :param parent_name: the name of the container of the dialog
        :type: string

        """

        msg_types = { 'info'    : gtk.MESSAGE_INFO,
                      'error'   : gtk.MESSAGE_ERROR, 
                      'warning' : gtk.MESSAGE_WARNING
                      }

        if msg_type not in msg_types.keys(): raise Exception("Invalid message type")
        else:
            if parent_name != None: parent = self[parent_name]
            else: parent = None
            dialog = gtk.MessageDialog(parent, 0, msg_types[msg_type],
                                       gtk.BUTTONS_OK, msg_text)
            if secondary_text != None:
                dialog.format_secondary_text(secondary_text)
            
            response = dialog.run()
                
        dialog.destroy()
        return



    ## GZ: not more needed right??
    def ui_move_figure(self, direction):
        """Moves the graph in the selected direction

        :param direction: the direction where to move the graph
        :type direction: {right, left, bottom, top}

        """

        offset = 20

        if direction == "right":
            self.axis_x_right_margin = self.axis_x_right_margin-offset
            self.axis_x_left_margin = self.axis_x_left_margin-offset

        elif direction == "left":
            self.axis_x_left_margin = self.axis_x_left_margin+offset
            self.axis_x_right_margin = self.axis_x_right_margin+offset

        elif direction == "bottom":
            self.axis_y_bottom_margin = self.axis_y_bottom_margin+offset
            self.axis_y_top_margin = self.axis_y_top_margin+offset

        elif direction == "top":
            self.axis_y_top_margin = self.axis_y_top_margin-20
            self.axis_y_bottom_margin = self.axis_y_bottom_margin-20

        # If the viewer is used in standalone mode, we need to call the
        # pyplot_view_structural function, which recomputes the figure and
        # shows it.
        # Otherwise, if the viewer is used integrated in FAME, we need simply
        # to recompute the figure and return it to FAME, which will show it
        if self.standalone:
            self.pyplot_view_structural()
        else:
            canvas = self.create_tfpg_structural_figure(None)
            return canvas

        return



    def ui_zoom_in(self):
        """Function which implements the increasing zoom feature

        """

        # self.zoom_offset = self.zoom_offset+0.5

        # for (k, (v1, v2)) in self.pos.items():
        #     self.pos[k] = (v1*self.zoom_offset, v2*self.zoom_offset)

        self.axis_x_left_margin = self.axis_x_left_margin + 30
        self.axis_x_right_margin = self.axis_x_right_margin - 30
        self.axis_y_top_margin = self.axis_y_top_margin - 30
        self.axix_y_bottom_margin = self.axis_y_bottom_margin + 30

        # self.node_size = self.node_size + 200
        self.dpi = self.dpi + 10

        if self.standalone:
            self.pyplot_view_structural()
        else:
            canvas = self.create_tfpg_canvas_figure(tfpg_graph=None)
            return canvas

        return


    def ui_zoom_out(self):
        """Function which implements the decreasing zoom feature

        """

        # for (k, (v1, v2)) in self.pos.items():
        #     self.pos[k] = (v1/self.zoom_offset, v2/self.zoom_offset)

        # self.zoom_offset = self.zoom_offset-0.5

        self.axis_x_left_margin = self.axis_x_left_margin - 20
        self.axis_x_right_margin = self.axis_x_right_margin + 20
        self.axis_y_top_margin = self.axis_y_top_margin + 20
        self.axis_y_bottom_margin = self.axis_y_bottom_margin - 20

        self.dpi = self.dpi - 10

        if self.standalone:
            self.pyplot_view_structural()
        else:
            canvas = self.create_tfpg_canvas_figure(tfpg_graph=None)
            return canvas

        return


    def export_to_file(self, filename, format):
        """Saves the current tfpg in an output file

           Possible formats are: png, pdf, dot
        """

        if format == 'dot':
            # In case of dot format, the graph must be changed a bit
            cleaned_graph = self.clean_graph_for_dot_format()
            nx.write_dot(cleaned_graph, str(filename)+"."+str(format))

        else:
            plt.savefig(str(filename)+"."+str(format))

        return


    def clean_graph_for_dot_format(self):
        """Cleans the graph in order to be exported in dot format

           Essentially, nodes containing node/edge labels (used for better
           show the graph in the GUI) are removed because in dot format
           they behave as different nodes
        """

        cleaned_graph = self.get_updated_networkx_graph()

        for nd in cleaned_graph.nodes(data=True):
            if "_lbl" in nd[0] :
                cleaned_graph.remove_node(nd[0])

        for edg in cleaned_graph.edges(data=True,keys=True):
            cleaned_graph[edg[0]][edg[1]][edg[2]]['label'] = edg[3]['edg_label']

        return cleaned_graph



    ## GZ: probably this function should be inserted into the fame_tools/tfpg struct, 
    ##     and there should be a function which updates the global variable in here
    ##     The view should simply say how to draw a TFPG, not modifying an internal struct
    def update_graph(self, component, values):
        """Updates the internal graph structure by setting 'values' for the selected 'component'


           If the internal graph structure contains a node or an edge named 'component', its 
           attributes (e.g. name, tMin, tMax, src, dst, modes in case of an edge) are modified
           accordingly to the values contained in the dict 'values'

           :param component: the component we want to modify
           :type  component: string

           :param values: the dict of pairs <attribute, value> corresponding to the updated 
                          values for 'component'
           :type  values: dict
        """

        ## TODO: notice that in case we want to update the edge/node name, we must also 
        ##       update the 'pos' dict

        error_list = []

        if "_edg_lbl" in component:

            # It is an edge
            component_name = component[:component.index("_edg_lbl")]
            
            if self.tfpg_graph.has_edge_by_name(str(component_name)):
                new_edge = self.tfpg_graph.get_edge_by_name(str(component_name))

                try:
                    if not self.tfpg_graph.has_node_by_name(values['src']):
                        raise NotExistingNode(values['src'], 'src')
                    if not self.tfpg_graph.has_node_by_name(values['dst']):
                        raise NotExistingNode(values['dst'], 'dst')
                    tmp_edge = Edge(self.tfpg_graph.get_node_by_name(values['src']),
                                    values['tMin'],
                                    values['tMax'],
                                    values['modes'].split(", "),
                                    self.tfpg_graph.get_node_by_name(values['dst']),
                                    values['name'])
                    if self.tfpg_graph.check_edge(tmp_edge) == []:
                        ## GZ: notice that in this situation I cannot simply use the set_name(),
                        ##     set_srcNode(), ... functions.
                        ##     Consider the following scenario:
                        ##     - I load the tfpg and I select the edge E1 which has tMin = 1 and tMax = 3
                        ##     - I want to change the tMin and set its value to 10
                        ##     - No changes are done to the structure, since tMin > tMax
                        ##     - In the info, are anyway reported the wrong values (i.e. tMin and TMax)
                        ##       in red color
                        ##     - The user understands the error and changes tMax = 3 to tMax = 30
                        ##     - He saves the changes but they cannot be done as when setting tMin tMax must
                        ##       still be updated
                        ##     The same would happen in case we would set tMax before setting tMin
                        new_edge.name = tmp_edge.get_name()
                        new_edge.srcNode = tmp_edge.get_srcNode()
                        new_edge.destNode = tmp_edge.get_destNode()
                        new_edge.tMin = tmp_edge.get_tMin()
                        new_edge.tMax = tmp_edge.get_tMax()
                        new_edge.modes = tmp_edge.get_modes()

                except Exception, e:
                    error_list.append(e)
            
        else:

            # It is a node
            if self.tfpg_graph.has_node_by_name(str(component)):
                new_node = self.tfpg_graph.get_node_by_name(str(component))

                try:
                    node_type = None

                    if values['type'] == "FM":
                        node_type = NodeType.FM
                    elif values['type'] == "AND":
                        node_type = NodeType.AND
                    elif values['type'] == "OR":
                        node_type = NodeType.OR

                    new_node.set_name(values['name'])
                    new_node.set_type(node_type)
                    new_node.set_isMonitored(values['isMonitored'])

                except Exception, e:
                    error_list.append(e)

        return error_list


pass  # end of class
