# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


# Module containing all the exceptions allowed in COMPASS code.
# Use only these exceptions, please.
#
# There are two main classes of exceptions: user ones UserException
# and internal ones InternalException.
# User exceptions are exceptions which are caused by errors in
# the input given to the tool.
# Internal exceptions are caused by bugs in the implementation code,
# by incorrect interaction with external tools or by external
# tools themselves.
#
# All the other exceptions can be derived only from these two.
#
# For example, incorrectly typed property will cause a user exception
# MistypeUserException derived from UserException.
# A segmentation fault caused by NuSMV will cause ProcessInternalException
# derived from InternalException.


################ USER EXCEPTIONS ########################

class UserException(Exception):
    """All user related exceptions have to be derived from this class"""
    def __init__(self, error_message):
        super(UserException, self).__init__(error_message)
        pass
    pass # end of class


class MistypeUserException(UserException):
    """An exception to deal with type errors in user specified expressions

     Attributes:
        file_name -- (string) the name of the file a type error is found
        line_number -- (int) the line number a type error is found
        error_message -- (string) a message explaining a type error
     """
    def __init__(self, file_name, line_number, error_message):
        
        msg = "Type error:%s line %d : Type Error : %s" %\
        ((file_name+",") if file_name else "", line_number, error_message)
        super(MistypeUserException, self).__init__(msg)
        pass
    pass # end of class
        
################ INTERNAL EXCEPTIONS ########################

class InternalException(Exception):
    """All tools related exception have to be derived from this class"""
    def __init__(self, error_message):
        super(InternalException, self).__init__(error_message)
        pass
    pass # end of class


class ProcessInternalException (InternalException):
    """This exception is raised when external tool behaves
    unexpectedly, i.e. dies, for example."""
    def __init__(self, title, details):
        super(ProcessInternalException, self).__init__("Error in %s: %s" % \
                                      (title, details))
        self.title = title
        self.description = details
        pass

    def get_title(self): return self.title
    def get_description(self): return self.description
    pass # end of class


class NusmvControlledException (ProcessInternalException):
    def __init__(self, details):
        super(NusmvControlledException, self).__init__("NuSMV", details)
        pass
    

class TraceException (NusmvControlledException):
    def __init__(self, details):
        super(TraceException, self).__init__(details)
        self.title = "NuSMV Trace"
        pass
        
