# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================

# In Python 2.5, the "with" statement needs to be imported.
# (In Python 2.6, it is enabled by default.)
from __future__ import with_statement

from utils import log
from utils.nusmv_process import NuSMVProcess
from utils.process import TextLineGrepper

from tfpg2smv.tfpg2smv import write_obs_file
from utils.tfpg_synthesizer import TfpgSynthesizer, CutSet

import re


############################# CLASS FameTfpgProcess ############################


class FameTfpgProcess (NuSMVProcess):
    """Process handler for running TFPG-related analyses"""

    # prefix for generated files
    _fname_prefix = "__fametfpg_prefix__"

    # algorithm for invariant checking with mathsat
    MSAT_INVAR_ALG = "een-sorensson"

    def __init__ (self, smv_fname, ord_fname=None, need_msat=False, args=[],
        stdout=None, stderr=None, dumpers=[]):

        # TODO: [BB] check if this import is really needed somewhere, else remove
        #import subprocess

        # Initialize the superclass (NuSMVProcess)
        super(FameTfpgProcess, self).__init__(smv_fname, ord_fname, need_msat,
            args, stdout=stdout, stderr=stderr, dumpers=dumpers)

        # we require the NuSMV version extended with addons
        assert not self.is_degraded_mode()

    def set_tmp_filename_prefix(self, prefix):
        """Sets the prefix of temporary generated files to the provided one."""
        self._fname_prefix = prefix
        pass

    ############################################################################
    ##########################  Behavioral Validation  #########################
    ############################################################################

    def check_tfpg_completeness (self, tfpg_structural, bmc_k):

        """ This function performs the completeness check of behavioral
        validation for TFPGs. It tries to find a system trace for which no
        valid corresponding TFPG trace exists, ie which causes the TFPG to
        violate some of its constraints when observing that trace.

        The function returns a tuple containing the analysis result and a
        completeness counterexample if one has been found. A completeness
        counterexample consists of a system trace and a corresponding TFPG
        trace that violates some TFPG constraints.

        The tfpg passed as a parameter is assumed to be the same used for
        creating the extended SMV file.        

        @param tfpg_structural: internal TFPG representation
        @type tfpg_structural: Tfpg object
        @param bmc_k: maximum depth for bounded model checking
        @type bmc_k: integer """

        # result can be: {error, complete, incomplete, bounded_completeness}
        # counterexample can be: {None, pair of traces}
        # counterexample_description is a string, containing a high-level description of the
        # incompleteness cause (eg. "missed activation of some_tfpg_node")
        res = {'result': None, 'counterexample': None, 'counterexample_description': None}

        # we start by assuming completeness; if a counterexample is found or
        # model checking can prove absence of counterexamples only up to the
        # given bound, this value is changed.
        res['result'] = 'complete'

        # check sanity of parameters
        assert isinstance(bmc_k, int)
        assert bmc_k >= 0

        log.system.debug("Sending 'set on_failure_script_quits' command to NuSMV.")
        self.send_cmd('set on_failure_script_quits')

        log.system.debug("Sending 'go_msat' command to NuSMV.")
        self.send_cmd('go_msat')

        # generate proof obligations (INVAR expressions as strings)

        discrepancies = tfpg_structural.get_or_nodes() + tfpg_structural.get_and_nodes()
        edges = tfpg_structural.get_edges_list()

        proof_obligations = self._get_completeness_proof_obligations(discrepancies, edges)

        # model check both generated properties

        for proof_obl in proof_obligations:

            prop = proof_obl['prop']
            print "checking '%s' ..." % prop,

            # send model checking command

            log.system.debug("Sending 'msat_check_invar_bmc' command to NuSMV.")
            self.send_cmd('msat_check_invar_bmc -a %s -k %d -p \"%s\"' % (self.MSAT_INVAR_ALG, bmc_k, prop))

            # retrieve results

            grep = TextLineGrepper(".*\s+(is false|is true|no counterexample or proof found for invariant .* up to %d|no counterexample found with bound %d|no proof or counterexample found with bound %d)" % (bmc_k, bmc_k, bmc_k))
            (m, txt) = grep.match(self)

            if m is None:
                res['result'] = 'error'
                print "error"
                break
            
            elif m.groups()[0] == 'is true':
                print "true"
                pass

            elif m.groups()[0] == 'is false':
                res['result'] = 'incomplete'
                orig_trace_fname = self._create_trace()
                res['counterexample'] = self.split_behval_completeness_tracefile(orig_trace_fname)
                res['counterexample_description'] = "%s of %s" % (proof_obl['type'], proof_obl['tfpg_node'])
                print "incomplete (%s)" % res['counterexample_description']
                break

            else:
                assert ('no counterexample or proof found for invariant' in m.groups()[0] or 'no counterexample found with bound' in m.groups()[0] or 'no proof or counterexample found with bound' in m.groups()[0])
                print "bounded complete"
                res['result'] = 'bounded_completeness'

        # 4. done ##############################################################

        return res


    ############################################################################
    ########################  Effectiveness Validation  ########################
    ############################################################################

    def check_tfpg_diagnosability (self, observables, target_fm_set,
        system_mode_var_name, time_tick_var_name, target_system_mode=None,
        use_bmc=False, bmc_k=10):
        """ Checks whether a bound exists within which any behavior caused by
        failures from target_fm_set can be distinguished from any other behavior
        not including any of these failures. Optionally, this analysis can be
        limited to a given stable system mode (ie mode doesn't change over
        time). Assumption: system mode and time ticks are always observable."""

        # result can be: {error, diagnosable, not_diagnosable, diagnosability_unknown}
        result = {'result': None, 'critical_pair': None}

        # check sanity of parameters

        assert isinstance(bmc_k, int)
        assert bmc_k >= 0

        # create temporary observables file

        obs_fname = self._fname_prefix + "diag_obs.tmp"
        log.system.debug('Creating helper file for observables (%s).' % obs_fname)
        write_obs_file (observables, obs_fname, time_tick_var_name, system_mode_var_name)

        # cmd: set on_failure_script_quits

        log.system.debug("Sending 'set on_failure_script_quits' command to NuSMV.")
        self.send_cmd('set on_failure_script_quits')

        # cmd: build twin plant

        log.system.debug("Sending 'build_twin_plant' command to NuSMV.")
        self.send_cmd('build_twin_plant -o "%s" -a' % (obs_fname))

        # create temporary file for diagnosability conditions

        diagcond_fname = self._fname_prefix + "diag_cond.tmp"
        log.system.debug('Creating helper file for diagnosability condition (%s).' % diagcond_fname)

        # used for the diagnosability conditions file
        target_fm_expression = "("
        # the property to be checked
        smv_prop = "("

        for i in range(0,len(target_fm_set)):        

            target_fm_expression += ("failuremode_%s.is_active" % target_fm_set[i])
            smv_prop += ("twin_L.failuremode_%s.is_active" % target_fm_set[i])

            if i < (len(target_fm_set) - 1):
                target_fm_expression += " | "
                smv_prop += " | "
            else:
                target_fm_expression += ")"
                smv_prop += " ) "

        smv_prop = "((G F !stutter_L) & (G F !stutter_R)) -> G ! (%s & G TRUE)" % smv_prop

        system_mode_expression = None if target_system_mode is None else ("(%s = %s)" % (system_mode_var_name, target_system_mode))
                
        with open(diagcond_fname, "w") as cond_out:

            cond_out.write("CONDITION_L %s\n" % target_fm_expression)
            cond_out.write("CONDITION_R TRUE\n")

            if system_mode_expression is not None:
                cond_out.write("CONTEXT_L %s\n" % system_mode_expression)
                cond_out.write("CONTEXT_R %s & !%s\n" % (system_mode_expression, target_fm_expression))
            else:
                cond_out.write("CONTEXT_R !%s\n" % target_fm_expression)

        # cmd: load_diag_conditions

        log.system.debug("Sending 'load_diag_conditions' command to NuSMV.")
        self.send_cmd('load_diag_conditions -c %s -d -1 -p' % diagcond_fname)

        # cmd: go

        if use_bmc:
            log.system.debug("Sending 'go_bmc' command to NuSMV.")
            self.send_cmd('go_bmc')
        else:
            log.system.debug("Sending 'go' command to NuSMV.")
            self.send_cmd('go')

        # cmd: check_property

        if use_bmc:
            log.system.debug("Sending 'check_ltlspec_bmc' command to NuSMV.")
            self.send_cmd('check_ltlspec_bmc -k %d -p "%s"' % (bmc_k, smv_prop))
        else:
            log.system.debug("Sending 'check_ltlspec' command to NuSMV.")
            self.send_cmd('check_ltlspec -p "%s"' % smv_prop)

        # retrieve results

        grep = TextLineGrepper(".*\s+(is false|is true|no counterexample found with bound %d)" % bmc_k)
        (m, txt) = grep.match(self)

        if m is None:
            result['result'] = 'error'
            result['critical_pair'] = None
        
        elif m.groups()[0] == 'is true':
            result['result'] = 'diagnosable'
            result['critical_pair'] = None

        elif m.groups()[0] == 'is false':
            result['result'] = 'not_diagnosable'
            orig_trace_fname = self._create_trace()
            
            if orig_trace_fname is None: 
                result['result'] = 'error'
                return result
            
            result['critical_pair'] = self.split_effectiveness_tracefile(orig_trace_fname)

        else:
            assert ('no counterexample found with bound %d' % bmc_k) in m.groups()[0]
            result['result'] = 'diagnosability_unknown'
            result['critical_pair'] = None

        return result


    ############################################################################
    ################################  Synthesis  ###############################
    ############################################################################

    def synthesize_tfpg(self, sys_smv_fname, out_tfpg_fname, system_modes, failure_modes, discrepancies, tfpg_name, bmc_k=10):

        log.system.debug("Sending 'set on_failure_script_quits' command to NuSMV.")
        self.send_cmd('set on_failure_script_quits')

        log.system.debug("Sending 'go_msat' command to NuSMV.")
        self.send_cmd('go_msat')

        minimal_cutsets = {}

        # add empty cutsets for failure modes
        for fm in failure_modes:
            minimal_cutsets[fm] = []

        # for each discrepancy
        for d in discrepancies:

            # clear logfile
            open(out_tfpg_fname, 'w').close()

            log.system.debug("Computing minimal cutsets for discrepancy '%s'." % d[0])
            print("Computing minimal cutsets for discrepancy '%s'." % d[0])

            # compute all mcs (read results from stdout)
            log.system.debug("Sending 'check_ltlspec' command to NuSMV.")
            self.send_cmd('compute_fault_tree_msat_bmc -o "%s" -k %d "tfpgsynth_%s"' % (out_tfpg_fname, bmc_k, d[0]))
            self.send_cmd('echo DONE')

            # wait for completion
            grep = TextLineGrepper("DONE")
            (m, txt) = grep.match(self)

            # parse mcs using cutset class
            print "parsing cutset file ...",
            cur_cutsets = CutSet.parse_cutset_list(out_tfpg_fname)
            minimal_cutsets[d[0]] = []

            if cur_cutsets is None:
                print "'%s' is not reachable within given bound" % d[0]
                minimal_cutsets[d[0]] = None
            elif len(cur_cutsets) is 1 and len(cur_cutsets[0].get_events()) is 0:
                print "'%s' has an empty cutset" % d[0]
                minimal_cutsets[d[0]] = []
            else:
                assert len(cur_cutsets) > 0
                print "found %d cutsets for '%s'" % (len(cur_cutsets), d[0])
                for cs in cur_cutsets:
                    tle = cs.get_tle().replace('tfpgsynth_','')
                    assert tle == d[0]
                    events = [e.replace('tfpgsynth_','') for e in cs.get_events()]
                    print " cutset events:",events
                    minimal_cutsets[tle].append(events)

            print "done"

        # call tfpg_synthesizer

        tfpg_synthesizer = TfpgSynthesizer(system_modes, failure_modes, discrepancies, minimal_cutsets)

        print "validating node names and cutsets ..."
        res_nodenames = tfpg_synthesizer.validate_node_names()
        res_cutsets = tfpg_synthesizer.validate_cutsets()
        print "done"

        if res_nodenames == "ok" and res_cutsets == "ok":

            print "running cutset-based graph simplification ...",
            tfpg_synthesizer.simplify_cutsetbased()
            print "done"

            print "calling synthesis ...",
            discrepancy_names = [d[0] for d in discrepancies]
            tfpg_synthesizer.synthesize_causality_graph(discrepancy_names)
            print "done"

            print "calling simplification ..."
            tfpg_synthesizer.simplify_causality_graph(tfpg_synthesizer.causality_graph)
            print "done"

            print "writing to file ...",
            f_out = open(out_tfpg_fname, 'w')
            #tfpg_synthesizer.print_human_readable_causality_graph(f_out, tfpg_name)
            tfpg_synthesizer.print_xml_causality_graph(f_out, tfpg_name)
            f_out.close()
            print "done"

            return "ok"

        elif res_nodenames != "ok":
            return res_nodenames

        else:
            return res_cutsets


    ############################################################################
    #############################  Internal Utils  #############################
    ############################################################################

    def _get_completeness_proof_obligations(self, discrepancies, edges):

        from tfpg2smv.tfpg2smv import compute_smv_edge_clock_name, compute_smv_node_name

        proof_obligations = []

        # for each discrepancy node
        for d in discrepancies:

            node_name = compute_smv_node_name(d)

            node_type = d.get_type().lower()
            assert node_type in ["and", "or"]

            incoming_edges = [e for e in edges if e.get_destNode() is d]
            assert len(incoming_edges) > 0

            ls_below_tmin = []
            ls_exceeding_tmax = []

            # for each incoming edge e
            for e in incoming_edges:

                etime = compute_smv_edge_clock_name(e)
                tmin = e.get_tMin()
                tmax = e.get_tMax()

                ls_below_tmin.append("%s < %s" % (etime, tmin))

                # skip edges with tmax=infty
                if float(tmax) >= 0:
                    ls_exceeding_tmax.append("%s > %s" % (etime, tmax))

            if node_type == "or":

                # missed activation
                if len(ls_exceeding_tmax) > 0:
                    prop = "!(%s)" % ' | '.join(ls_exceeding_tmax)
                    proof_obligations.append({'tfpg_node':d.get_name(), 'type':'missed activation', 'prop':prop})

                # unexpected activation
                prop = "!(%s & (%s))" % (node_name, ' & '.join(ls_below_tmin))
                proof_obligations.append({'tfpg_node':d.get_name(), 'type':'unexpected activation', 'prop':prop})

            if node_type == "and":

                # missed activation
                if len(ls_exceeding_tmax) > 0:
                    prop = "!(%s)" % ' & '.join(ls_exceeding_tmax)
                    proof_obligations.append({'tfpg_node':d.get_name(), 'type':'missed activation', 'prop':prop})

                # unexpected activation
                prop = "!(%s & (%s))" % (node_name, ' | '.join(ls_below_tmin))
                proof_obligations.append({'tfpg_node':d.get_name(), 'type':'unexpected activation', 'prop':prop})

        return proof_obligations

    def split_behval_completeness_tracefile(self, trace_filename):
        """This method takes a trace as generated by NuSMV model checking for
        TFPG completeness analysis, and splits them into two traces: One
        being the system trace, and one being the corresponding trace over the
        TFPG alphabet violating the TFPG constraints.

        The filenames of the two new trace files are returned.
        @param trace_filename: the filename of the original trace
        @type trace_filename: string """

        # new trace files

        sys_trace_fn = trace_filename.replace('.xml', '-SYS.xml')
        tfpg_trace_fn = trace_filename.replace('.xml', '-TFPG.xml')
        sys_trace = open(sys_trace_fn, "w")
        tfpg_trace = open(tfpg_trace_fn, "w")

        # parse trace file line by line

        orig_trace = open(trace_filename, "r")

        while 1:
            line = orig_trace.readline()
            if not line:
                break

            tfpg_line = False
            common_line = False

            tfpg_tokens = ["failuremode_", "or_node_", "and_node_", "ETime_", "EState_", "Mode_"]
            for t in tfpg_tokens:
                match = re.search(r'^\s*<value variable="%s' % t, line)
                if match:
                    tfpg_line = True
                    break

            common_line = not tfpg_line and not "value" in line

            if common_line:
                sys_trace.write(line)
                tfpg_trace.write(line)

            elif tfpg_line:

                # prepend root.sc_ so GUI trace viewer won't ignore variables
                if "failuremode_" in line:
                    line = line.replace('failuremode_','root.sc_')
                elif "or_node_" in line:
                    line = line.replace('or_node_','root.sc_')
                elif "and_node_" in line:
                    line = line.replace('and_node_','root.sc_')
                elif "ETime_" in line:
                    line = line.replace('ETime_','root.sc_ETime_')
                elif "EState_" in line:
                    line = line.replace('EState_','root.sc_EState_')
                elif "Mode_" in line:
                    line = line.replace('Mode_','root.sc_Mode_')
                else:
                    # there should be no other nodes in the tfpg trace variables
                    assert False
                    
                tfpg_trace.write(line)

            else:
                sys_trace.write(line)

        # done

        sys_trace.close()
        tfpg_trace.close()
        orig_trace.close()

        return {'sys_trace':sys_trace_fn, 'tfpg_trace':tfpg_trace_fn}

    def split_effectiveness_tracefile(self, trace_filename):
        """This method takes a trace as generated by diagnosability analysis of
        NuSMV, and splits them into two traces, hiding internal implementation
        details of NuSMV. The filenames of the two new trace files are
        returned.
        @param trace_filename: the filename of the original trace
        @type trace_filename: string """

        import xml.dom
        import xml.dom.minidom

        doc = xml.dom.minidom.parse(trace_filename) 
        system_mode = doc.getElementsByTagName('value')[0]
    
        # new trace files

        trace_fn_1 = trace_filename.replace('.xml', '-A.xml')
        trace_fn_2 = trace_filename.replace('.xml', '-B.xml')
        trace_1 = open(trace_fn_1, "w")
        trace_2 = open(trace_fn_2, "w")

        # parse trace file line by line

        orig_trace = open(trace_filename, "r")

        while 1:
            line = orig_trace.readline()
            if not line:
                break

            # skip lines with path monitors of restrictions
            if 'have_C' in line:
                continue

            # skip lines with edge variables
            if "edge_" in line:
                continue

            # can/must fire flags
            if "must_fire" in line or "can_fire" in line:
                continue

            # replace tfpg2smv translation artifacts from variable names
            if "failuremode_" in line:
                line = line.replace('failuremode_','')
            if "or_node_" in line:
                line = line.replace('or_node_','')
            if "and_node_" in line:
                line = line.replace('and_node_','')
            if ".is_active" in line:
                line = line.replace('.is_active','')   
                
            #LB# The ENUM values of system_mode and instant_trans_type are converted in SMV Format as CONST_    
            if "system_mode" in line:
                line = line.replace(''.join( [node.data for node in system_mode.childNodes]), "const_" + ''.join( [node.data for node in system_mode.childNodes]))          
            if "instant_trans_type" in line and "NODE_ACTIVATION" in line:
                line = line.replace('NODE_ACTIVATION','const_NODE_ACTIVATION')
            if "instant_trans_type" in line and "NONE" in line:
                line = line.replace('NONE','const_NONE')    
            if "instant_trans_type" in line and "MODE_CHANGE" in line:
                line = line.replace('MODE_CHANGE','const_MODE_CHANGE') 
                  
            # redirect line to correct file (and drop twin tags)
            if line.count('twin_L.') > 0:
                trace_1.write(line.replace('twin_L.', 'root.sc_'))
            elif line.count('twin_R.') > 0:
                trace_2.write(line.replace('twin_R.', 'root.sc_'))
            else:
                trace_1.write(line)
                trace_2.write(line)

        # done

        trace_1.close()
        trace_2.close()
        orig_trace.close()

        return (trace_fn_1, trace_fn_2)


    pass # end of class

