# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


import copy

from lxml import etree

from property_management.src import formulae
from property_management.src import properties
import property_management.src.constants as properties_constants
import slim_compiler.src.model_extension.errors as model_extension_errors
from slim_compiler.src.model_extension.util import general
import slim_compiler.src.smv.api as smv_api
import slim_compiler.src.dataStructure.nominalbehaviour as nominal


# Define constants for some tags and attribute names of the XML elements that
# are used in the XML representation of fault trees.
FAULT_TREE_TAG = "fault-tree"
BASIC_EVENT_TAG = "event"
GATE_TAG = "gate"
EVENT_NAME_ATTRIBUTE = "name"
EVENT_DESCRIPTION_ATTRIBUTE = "description"


def parse_xml(data_source):
    """Parses the fault tree XML document from the specified data source and
    returns a tree representation of the document.
    
    @param   data_source: The source of the data to be parsed. This may be a
                          filename or a file-like object.
    @type    data_source: A string or a file-like object.
    @return: A tree representation of the document.
    @rtype:  etree._ElementTree
    """
    # TODO: Document possible parsing exceptions.
    return etree.parse(data_source)


def get_as_xml_string(fault_tree):
    """Returns a string representation of the specified fault tree in XML
    format.
    
    @param   fault_tree: The fault tree to be considered.
    @type    fault_tree: etree._ElementTree
    @return: A string representation of the fault tree in XML format.
    @rtype:  string
    """
    return etree.tostring(fault_tree)


def get_event_names(fault_tree):
    """Returns the set of all event names (basic event names and gate names)
    in the specified fault tree.
    
    @param   fault_tree: The fault tree to be considered.
    @type    fault_tree: etree._ElementTree
    @return: The set of all event names in the specified fault tree.
    @rtype:  A set of strings.
    """
    return set([event_element.get(EVENT_NAME_ATTRIBUTE)
                for event_element in fault_tree.iter()
                if event_element.tag in (BASIC_EVENT_TAG, GATE_TAG)])


def get_top_level_event_name(fault_tree):
    """Returns the name of the top-level event of the specified fault tree.
    
    @param   fault_tree: The fault tree to be considered.
    @type    fault_tree: etree._ElementTree
    @return: The name of the top-level event of the specified fault tree.
    @rtype:  string
    """
    return fault_tree.getroot()[0].get(EVENT_NAME_ATTRIBUTE)


def replace_event_descriptions(fault_tree, replacement_function):
    """Applies the specified replacement function to the event descriptions of
    the specified fault tree.
    
    @param   fault_tree: The fault tree to be modified.
    @type    fault_tree: etree._ElementTree
    @param   replacement_function: A function that returns for each event
                                   description in the fault tree another
                                   description, which shall replace the
                                   original one.
    @type    replacement_function: A function: string -> string
    """
    for xml_element in fault_tree.iter():
        if xml_element.tag in (GATE_TAG, BASIC_EVENT_TAG):
            event_description = xml_element.get(EVENT_DESCRIPTION_ATTRIBUTE)
            xml_element.set(EVENT_DESCRIPTION_ATTRIBUTE,
                            replacement_function(event_description))


def create_basic_event_name_to_rate_dict(fault_tree, root_implementation):
    """Creates a dictionary that maps the basic events of the specified fault
    tree to the corresponding rates.
    
    A basic event in a fault tree represents the firing of an error event port
    in a certain component of the SLIM model. The rate of the basic event will
    be determined by reading the rate parameter of the corresponding error
    event port.

    @param   fault_tree:          The fault tree to be considered.
    @type    fault_tree:          etree._ElementTree
    @param   root_implementation: The implementation of the root component of
                                  the SLIM model. This must be the same root
                                  implementation that was used to create the
                                  specified fault tree.
    @type    root_implementation: nominalbehaviour.ComponentImplementation
    @raise   model_extension_errors.InvalidComponentPathError:
                If some basic event name does not refer to an existing
                component of the SLIM model.
    @raise   InvalidEventPortIdentifier:
                If a basic event does not refer to an existing error event
                port in the specified SLIM model.
    @raise   ErrorEventPortHasNoParameterError:
                If the error event port that is referenced by some basic event
                name has no probability distribution parameter attached.
    """
    event_name_to_rate_dict = {}
    # The fault tree is represented by an XML tree. Consider all XML elements
    # in the tree that represent basic events.
    for event_element in fault_tree.iter(BASIC_EVENT_TAG):
        # The corresponding error event port for the basic event is specified
        # in the description attribute of the XML event element in the form of
        # an SMV identifier.
        smv_port_identifier = event_element.get(EVENT_DESCRIPTION_ATTRIBUTE)
        # In order to find the error event port in the SLIM model, the SMV
        # identifier must be translated to the corresponding SLIM identifier.
        slim_port_identifier = smv_api.convert_smv_id_to_slim(
                                    smv_port_identifier,
                                    can_be_enum_literal=False)
        # The SLIM identifier for the error event port must be a path that
        # consists of a sequence of subcomponent identifiers (the component
        # path), which determines the component in the SLIM model, and the
        # name of an error event port of that component. All parts of the path
        # are separated by dots.
        # Example: "subcomp1.subcomp2.some_error_event".
        # Extract the component path and the event port name.
        if not slim_port_identifier:
            # Note: The event port identifier that is passed as a parameter of
            # the exception is in SMV format, not in SLIM format, because it
            # could not be converted to SLIM format.
            raise InvalidEventPortIdentifier(smv_port_identifier)
        event_port_path = tuple(slim_port_identifier.split("."))
        component_path = event_port_path[:-1]
        event_port_name = event_port_path[-1]
        # Now create an object representing the component that is determined
        # by the component path and try to find the error event port in that
        # component.
        # Note: Creating the general.ComponentInstance object might raise a
        # model_extension_errors.InvalidComponentPathError.
        component = general.ComponentInstance(root_implementation,
                                              component_path)
        try:
            event_port = component.type.out_event_ports[event_port_name]
        except KeyError:
            raise InvalidEventPortIdentifier(slim_port_identifier)
        if not isinstance(event_port, nominal.ErrorEventPort):
            raise InvalidEventPortIdentifier(slim_port_identifier)
        # We have found the error event port. Store its rate in the
        # dictionary, using the name of the basic event as the key.
        rate = event_port.get_parameter()
        if rate is None:
            raise ErrorEventPortHasNoParameterError(event_port, component)
        event_name = event_element.get(EVENT_NAME_ATTRIBUTE)
        event_name_to_rate_dict[event_name] = rate
    
    return event_name_to_rate_dict


def sub_fault_tree_generator(fault_tree):
    """Returns the name of the top-level event of the specified fault tree.
    
    @param   fault_tree: The fault tree to be considered.
    @type    fault_tree: etree._ElementTree
    @return: The name of the top-level event of the specified fault tree.
    @rtype:  string
    """
    # TODO: Add comments to this function.
    visited_event_names = set()
    for xml_element in fault_tree.iter():
        if xml_element.tag in (GATE_TAG, BASIC_EVENT_TAG):
            event_name = xml_element.get(EVENT_NAME_ATTRIBUTE)
            if event_name not in visited_event_names:
                visited_event_names.add(event_name)
                root_element = etree.Element(FAULT_TREE_TAG)
                root_element.append(copy.deepcopy(xml_element))
                xml_tree = etree.ElementTree(element=root_element)
                yield xml_tree


def get_in_fault_tree_plus_format(fault_tree,
                                  event_name_to_probability_dict=None,
                                  event_name_to_criticality_dict=None):
    """Returns a representation of the specified fault tree in FaultTree+
    format.
    
    @param   fault_tree:
                The fault tree to be considered.
    @type    fault_tree:
                etree._ElementTree
    @param   event_name_to_probability_dict:
                (Optional) A dictionary, mapping the names of the events of
                the fault tree to their probabilities. The dictionary may map
                all event names or only a subset of them.
    @type    event_name_to_probability_dict:
                dict: string -> float
    @param   event_name_to_criticality_dict:
                (Optional) A dictionary, mapping the names of the events of
                the fault tree to their criticalities. The dictionary may map
                all event names or only a subset of them.
    @type    event_name_to_criticality_dict:
                dict: string -> float
    @return: The specification of the basic events and the specification of
             the gates in FaultTree+ format.
    @rtype:  A pair of strings.
    """
    # TODO: Add comments to this function.
    if event_name_to_probability_dict is None:
        event_name_to_probability_dict = {}
    if event_name_to_criticality_dict is None:
        event_name_to_criticality_dict = {}
    
    BASIC_EVENT_INPUT_TEMPLATE = "E:%s"
    GATE_INPUT_TEMPLATE = "G:%s"
    BASIC_EVENT_PROBABILITY_TEMPLATE = "%E"
    GATE_PROBABILITY_TEMPLATE = "NuSMV-SA probability: %e"
    GATE_CRITICALITY_TEMPLATE = "NuSMV-SA criticality: %e"
    SUPPORTED_GATE_TYPES = ("NULL", "AND", "OR", "PAND")

    # The maximum number of inputs of the gates in the fault tree must be
    # computed here, because it is used in the closure assemble_gate_line.
    # TODO: The closure might be confusing and might better be avoided.
    maximum_number_of_inputs = max([len(gate_element) for gate_element in
                                    fault_tree.iter(GATE_TAG)])

    def create_line_for_basic_event(basic_event):
        name = basic_event.get(EVENT_NAME_ATTRIBUTE)
        description = basic_event.get(EVENT_DESCRIPTION_ATTRIBUTE)
        try:
            probability_string = (BASIC_EVENT_PROBABILITY_TEMPLATE %
                                  event_name_to_probability_dict[name])
        except KeyError:
            probability_string = ""
        return assemble_basic_event_line(name, "Basic", description, "Fixed",
                                         probability_string)
    
    def create_line_for_gate(gate):
        name = gate.get(EVENT_NAME_ATTRIBUTE)
        description = gate.get(EVENT_DESCRIPTION_ATTRIBUTE)
        try:
            probability_string = (GATE_PROBABILITY_TEMPLATE %
                                  event_name_to_probability_dict[name])
        except KeyError:
            probability_string = ""
        try:
            criticality_string = (GATE_CRITICALITY_TEMPLATE %
                                  event_name_to_criticality_dict[name])
        except KeyError:
            criticality_string = ""
        notes = "%s %s" % (probability_string, criticality_string)
        type = gate.get("type")
        assert type in SUPPORTED_GATE_TYPES
        input_names = []
        for child_element in gate:
            child_name = child_element.get("name")
            if child_element.tag == BASIC_EVENT_TAG:
                input_names.append(BASIC_EVENT_INPUT_TEMPLATE % child_name)
            elif child_element.tag == GATE_TAG:
                input_names.append(GATE_INPUT_TEMPLATE % child_name)
            else:
                raise Exception("Invalid XML element in fault tree.")
        return assemble_gate_line(name, type, description, input_names,
                                  "Off", "On", notes)
    
    def assemble_basic_event_line(name, symbol_type, description,
                                  failure_model_type,
                                  failure_model_parameter):
        return "\t".join([name, symbol_type, description,
                          failure_model_type, failure_model_parameter])
        
    def assemble_gate_line(name, type, description,
                           input_names, page_flag, retain_results, notes):
        complete_input_names = input_names[:]
        for i in range(maximum_number_of_inputs - len(input_names)):
            complete_input_names.append("")
        return "\t".join([name, type, description] + complete_input_names +
                         [page_flag, retain_results, notes])

    basic_events_header = assemble_basic_event_line(
            "Name", "Symbol Type", "Description",
            "Failure Model Type", "Failure Model Parameter 0")
    basic_event_lines = [create_line_for_basic_event(basic_event)
                         for basic_event in fault_tree.iter(BASIC_EVENT_TAG)]
    
    maximum_number_of_inputs = max([len(gate_element) for gate_element in
                                    fault_tree.iter(GATE_TAG)])
    input_name_headers = ["Input  %d Type and Name" % i
                          for i in range(0, maximum_number_of_inputs)]
    gates_header = assemble_gate_line(
            "Name", "Type", "Description", input_name_headers,
            "Page Flag", "Retain Results", "Notes1")
    gate_lines = [create_line_for_gate(gate)
                  for gate in fault_tree.iter(GATE_TAG)]
    
    basic_events_complete_string = "\r\n".join([basic_events_header] +
                                               basic_event_lines) + "\r\n"
    gates_complete_string = "\r\n".join([gates_header] + gate_lines) + "\r\n"
    
    return (basic_events_complete_string, gates_complete_string)


def create_reachability_property(event_name, mission_time, property_name):
    """Returns a probabilitic reachability property (logic: CSL) representing
    the occurrence of the specified event within the specified mission time.
    
    @param   event_name:    The name of the event to be considered. This will
                            be used as an atomic proposition in the property.
    @type    event_name:    string
    @param   mission_time:  The time-bound for the occurrence of the event.
    @type    mission_time:  int
    @param   property_name: The name of the property. This can be an arbitrary
                            string.
    @type    property_name: string
    @return: A probabilistic reachability property in CSL.
    @rtype:  properties.Property
    """
    # TODO: Add comments to this function.
    logic_csl = properties_constants.Logic.CSL
    
    pattern_instance = properties.PatternInstance.create(
                            logic_csl, "probabilisticExistence")
    phi_proposition = formulae.AtomicProposition.create(logic_csl,
                                                         event_name)
    phi_instance = properties.StateFormulaPlaceholderInstance.create(
                                    logic_csl, "Phi", phi_proposition)
    
    lower_bound_instance = properties.TimeBoundPlaceholderInstance.create(
                                        logic_csl, "Time1", 0)
    upper_bound_instance = properties.TimeBoundPlaceholderInstance.create(
                                        logic_csl, "Time2", mission_time)
    
    pattern_instance.add_placeholder_instance(phi_instance)
    pattern_instance.add_placeholder_instance(lower_bound_instance)
    pattern_instance.add_placeholder_instance(upper_bound_instance)
    
    return properties.Property.create(pattern_instance, property_name)

def write_fault_tree_plus_output(fault_tree,
                                 event_name_to_probability_dict,
                                 event_names_to_criticality_dict,
                                 events_filename,
                                 gates_filename):
        (events_file_content, gates_file_content) = \
            get_in_fault_tree_plus_format(fault_tree, event_name_to_probability_dict, event_names_to_criticality_dict)

        with open(events_filename, "w") as events_file:
            events_file.write(events_file_content)
        with open(gates_filename, "w") as gates_file:
            gates_file.write(gates_file_content)


class FaultTreeError(Exception):
    """The base class for errors that are related to fault trees."""
    pass


class InvalidEventPortIdentifier(FaultTreeError):
    """An object indicating that a basic event of a certain fault tree does
    not refer to an existing error event port in the respective SLIM model.
    """

    def __init__(self, event_port_identifier):
        """Initializes the InvalidEventPortIdentifier object.
        
        @param   event_port_identifier: The invalid event port identifier.
        @type    event_port_identifier: string
        """
        self.event_port_identifier = event_port_identifier
    
    def __str__(self):
        """Returns a human-readable string representation of this error.
        
        @return: A human-readable error message.
        @rtype:  string
        """
        return (('Basic event does not refer to an existing error event ' +
                 'port: "%s"' % self.event_port_identifier))
        

class ErrorEventPortHasNoParameterError(FaultTreeError):
    """An object indicating that the error event port referenced by a basic
    event of a fault tree has not probability distribution parameter attached.
    """

    def __init__(self, error_event_port, component):
        """Initializes the ErrorEventPortHasNoParameterError object.
        
        @param   error_event_port:  The error event port.
        @type    error_event_port:  nominal.ErrorEventPort
        @param   component:         The SLIM component containing the error
                                    event port.
        @type    component:         general.ComponentInstance
        """
        self.error_event_port = error_event_port
        self.component = component
    
    def __str__(self):
        """Returns a human-readable string representation of this error.
        
        @return: A human-readable error message.
        @rtype:  string
        """
        if self.component.path:
            path_as_string = ".".join(self.component.path)
        else:
            path_as_string = "<root>"
        return (('The error event port "%s" of the component "%s" has no ' +
                 'probability distribution parameter attached.') %
                (self.error_event_port.name, path_as_string))
