# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


# In Python 2.5, the "with" statement needs to be imported.
# (In Python 2.6, it is enabled by default.)
from __future__ import with_statement

import os.path
import options

from invokers import slim
import property_management.src.api as prop_api

import utils.nusmv_process
import utils.process
from utils.misc import get_random_name
from utils import log
import utils.fault_tree as fault_tree_module

############## CLASS FsapProcess #############################

class FsapProcess (utils.nusmv_process.NuSMVProcess):
    """Process handler for running FSAP interactively"""

    # This is a prefix of generated file names. It must
    # be unusual enough to avoid conflicts with normal files.
    _fname_prefix = "__fsap_prefix__"


    def __init__(self, smv_fname, fail_obs_alm_fname, ord_fname=None, 
                 need_msat=False, args=[],
                 stdout=None, stderr=None, dumpers=[]):
        """Given a smv filename and a file with failures/observables/
        alarms specifications and an optional ordering filename,
        creates an interactive process instance of class
        FsapProcess. If not specified, ord_fname is searched taking
        the smv basename."""

        # check the failure-observables-alarms file existence. smv
        # and ord files are checked in NuSMVProcess
        if not os.path.isfile(fail_obs_alm_fname): 
            msg = "File not found: %s" % fail_obs_alm_fname
            log.system.error(msg)
            raise IOError(msg)
        pass

        args = list(args) + ["-sa_compass", "-sa_ft_layering", 
                             "-sa_compass_task", fail_obs_alm_fname]

        import subprocess
        super(FsapProcess, self).__init__(smv_fname, ord_fname, need_msat, args,
                                          stdout=stdout, stderr=stderr, dumpers=dumpers)

        # it cannot be in degraded mode, as FSAP requires full operations
        assert not self.is_degraded_mode()

        self._fail_obs_alm_fname = fail_obs_alm_fname

        # delete commands which are not expected to be in FSAP
        # Potentially, there is no problems in keeping them
        # [AT] Deleting does not work right now!!!!
        #delattr(self,"model_check_bdd_ctl")
        #delattr(self,"model_check_bdd_ltl")
        #delattr(self,"model_check_bmc_ltl")
        #delattr(self,"model_simulate")
        return


    def set_tmp_filename_prefix(self, prefix):
        """Sets the prefix of temporary generated files to the provided one."""
        self._fname_prefix = prefix
        pass

    from utils.log import log_icd
    @log_icd
    def compute_dyn_fault_tree(self, property, bmc_length=options.NUSMV_BMC_LENGTH):
        """Compute a Fault Tree for a given propositional property,
        outputs the result into a file (one for events and one for gates)
        and returns the names of the generated files (events, gates, xml).
        Raises Exception if a problem in execution is detected.

        @param property: a propositional SMV expression.
        @type property: string
        @param bmc_length: number of transitions used in the hybrid case and BMC
        @type bmc_length: int        
        """
        return self.compute_fault_tree(property, dynamic=True, bmc_length=bmc_length)

    from utils.log import log_icd
    @log_icd
    def compute_fault_tree(self, property, dynamic=False, bmc_length=options.NUSMV_BMC_LENGTH,
                           use_bdd=True, sat_sbmc=False, sbmc_completeness=False):
        """Compute a Fault Tree for a given propositional property,
        outputs the result into a file (one for events and one for gates)
        and returns the names of the generated files (events, gates, xml).
        Raises Exception if a problem in execution is detected.

        @param property: a propositional SMV expression.
        @type property: string
        @param dynamic: a flag to generate a DFT
        @type dynamic: boolean
        @param bmc_length: number of transitions used in the hybrid case and BMC
        @type bmc_length: int        
        @param use_bdd: a flag to generate FT using BDDs (default)
        @type use_bdd: boolean
        @param sat_sbmc: a flag to select SBMC SAT algorithms
        @type sat_sbmc: boolean
        @param sbmc_completeness: a flag to enable completeness check
        @type sat_sbmc: boolean
        """
        # note for developers: here are a list of possible commands
        # to FSAP:
        #compute_fault_tree -t -x prefix_ "root.sc_Ctrl.sc_error_bit_hardware = 1"
        #compute_fault_tree_bmc -t -x prefix_ -k 12 "root.sc_Ctrl.sc_error_bit_hardware = 1"
        #compute_fault_tree_bmc_inc -t -x prefix_ -k 12 "root.sc_Ctrl.sc_error_bit_hardware = 1"
        #compute_fault_tree_sbmc_inc -t -x prefix_ -k 12 "root.sc_Ctrl.sc_error_bit_hardware = 1"

        rnd_prefix = self._fname_prefix + get_random_name(6)

        if self.needs_msat():
            if dynamic: log.system.warning("FsapProcess: dynamic hybrid FT are not supported, falling back to static")
            self.__compute_anything('compute_fault_tree_msat_bmc -k %d -t -x "%s" "%s"' % \
                                        (bmc_length, rnd_prefix, property))
            
        elif use_bdd: #  BDD FT computation (default)
            self.__compute_anything('compute_fault_tree -t %s -x "%s" "%s"' % \
                                        ("-d" if dynamic else "", rnd_prefix, property))
        else:
            bmc = ("bmc" if not sat_sbmc else "sbmc")
            sbmc_closure = ("-c" if (sat_sbmc and  sbmc_completeness) else "")
            cmd = 'compute_fault_tree_%s_inc -k %d -t %s -x "%s" "%s"' % \
                (bmc, bmc_length, sbmc_closure, rnd_prefix, property)

            self.__compute_anything(cmd, use_sat = True)

        return (rnd_prefix + "events.txt",
                rnd_prefix + "gates.txt", 
                rnd_prefix + "ft.xml")

    def compute_dyn_fmea_table(self, property_list, fault_cardinality, 
                               bmc_length=options.NUSMV_BMC_LENGTH, 
                               _compact=options.FSAP_COMPACT_FMEA_TABLE):
        """Compute a FMEA (Failure Mode and Effects Analysis) Table
        for a given set of propositional property,
        outputs the result into a file
        and returns the names of the generated files.
        Raises Exception if a problem in execution is detected.

        @param porperty_list: a list of properties
        @type porperty_list: list of strings

        @param fault_cardinality: a natural number specifying cardinality
            of a set of fault configurations. A positive number.
        @type fault_cardinality: int

        @param bmc_length: number of transitions used in the hybrid case and BMC
        @type bmc_length: int
        """
        return self.compute_fmea_table(property_list, fault_cardinality, 
                                       dynamic=True, bmc_length=bmc_length,
                                       compact=_compact)
    
    
    from utils.log import log_icd
    @log_icd
    def compute_fmea_table(self, property_list, fault_cardinality,
                           dynamic=False, bmc_length=options.NUSMV_BMC_LENGTH,
                           compact=options.FSAP_COMPACT_FMEA_TABLE):
        """Compute a FMEA (Failure Mode and Effects Analysis) Table
        for a given set of propositional property,
        outputs the result into a file
        and returns the names of the generated files.
        Raises Exception if a problem in execution is detected.

        @param property_list: a list of properties
        @type porperty_list: list of strings

        @param fault_cardinality: a natural number specifying cardinality
            of a set of fault configurations. A positive number.
        @type fault_cardinality: int

        @param dynamic: a flag to generate a FMEA
        @type dynamic: boolean

        @param bmc_length: number of transitions used in the hybrid case and BMC
        @type bmc_length: int

        @param compact: whether compact fmea table (without
        redundances) has to be generated. This works only if dynamic
        is False. If dynaic is True, compact is ignored.
        @type compact: boolean
        """
        # note for developers: a typical FSAP command to compute FMEA table
        #compute_fmea_table -t -x prefix_ -N 2 "root.sc_Ctrl.sc_error_bit_hardware = 1" \
        #         "root.sc_Ctrl.sc_error_bit_hardware != 1" "root.sc_Ctrl.mode = mode_completely_failed"

        if compact and not dynamic: compact_opt = "-c"
        else: 
            if dynamic: log.system.warning("FsapProcess: compact & dynamic FMEA are not supported, falling back to dynamic only")
            compact_opt = ""
            pass
        
        if self.needs_msat():
            if dynamic: log.system.warning("FsapProcess: dynamic hybrid FMEA are not supported, falling back to static")
            self.__compute_anything('compute_fmea_table_msat_bmc -k %d -t '
                                    '-x "%s" -N %d %s "%s"' % \
                                        (bmc_length, self._fname_prefix, fault_cardinality, 
                                         compact_opt,
                                         '" "'.join(property_list)))
        else: 
            if dynamic: dyn_cmd = "-d"
            else: dyn_cmd = ""
            self.__compute_anything('compute_fmea_table -t %s -x "%s" -N %d %s "%s"' % \
                                        (dyn_cmd, self._fname_prefix, 
                                         fault_cardinality, 
                                         compact_opt,
                                         '" "'.join(property_list)))
            pass

        return self._fname_prefix + "fmea_table.txt"

    from utils.log import log_icd
    @log_icd    
    def compute_fault_detection(self, property, bmc_length=options.NUSMV_BMC_LENGTH, delay_counter_var=None, detection_delay=None):
        """Perform Fault Detection Analysis and 
        returns a set of alarms for a given propositional property (a fault).
        Outputs the set of alarm into file
        and returns the name of the generated file.
        If empty set is returned then the fault is not detectable.
        Raises Exception if a problem in execution is detected.
        @param property: a fault, i.e. a propositional properties
        @type property: string
        @param bmc_length: number of transitions used in the hybrid case and BMC
        @type bmc_length: int
        @param delay_counter_var: a variable that counts the time/steps after the
               occurrence of some property of interest (incompatible with
               parameter 'property'; requires parameter 'detection_delay')
        @type delay_counter_var: string
        @param detection_delay: a bound within which the alarm has to be
               raised for being considered as a valid fault detection means
               (requires parameter 'delay_counter_var')
        @type detection_delay: double
        """
        # note for developers: a typical FSAP command to run Fault Detection
        #compute_fault_detection -t -x prefix_ "root.sc_Ctrl.sc_error_bit_hardware = 1"

        assert (detection_delay is None and delay_counter_var is None) or (detection_delay is not None and delay_counter_var is not None)
        assert (detection_delay is None or detection_delay is -1 or detection_delay >= 0)

        if self.needs_msat():
            if delay_counter_var is not None:
                self.__compute_anything('compute_fault_detection_msat_bmc -k %d -t -x "%s" -d %f -c %s' % \
                                            (bmc_length, self._fname_prefix, detection_delay, delay_counter_var))
            else:
                self.__compute_anything('compute_fault_detection_msat_bmc -k %d -t -x "%s" "%s"' % \
                                            (bmc_length, self._fname_prefix, property))
        else:
            assert delay_counter_var is None
            self.__compute_anything('compute_fault_detection -t -x "%s" "%s"' % \
                                        (self._fname_prefix, property))
            pass
        
        return self._fname_prefix + "detection_means.txt"


    from utils.log import log_icd
    @log_icd    
    def compute_fault_isolation(self, number_of_alarms, bmc_length=options.NUSMV_BMC_LENGTH):
        """Perform Fault Isolation Analysis and 
        generates a set of Fault Trees, one for each alarm.
        Returns the set of generated file names (set of pairs: events file
        and gates files).
        Raises Exception if a problem in execution is detected.
        @param number_of_alarms: the number of alarms (int)
        @param bmc_length: number of transitions used in the hybrid case and BMC
        @type bmc_length: int
        """
        # note for developers: a typical FSAP command to run Fault Isolation
        #compute_fault_isolation -t -x prefix_
        if self.needs_msat():
            self.__compute_anything('compute_fault_isolation_msat_bmc -k %d -t -x "%s"' % \
                                        (bmc_length, self._fname_prefix))
        else:
            self.__compute_anything('compute_fault_isolation -t -x "%s"' % \
                                        self._fname_prefix)
            pass
        
        # create a list of generated files.
        res = []
        for num in range(1,number_of_alarms+1) :
            res.append(("%s_%d_events.txt" % (self._fname_prefix, num), \
                        "%s_%d_gates.txt" % (self._fname_prefix, num)))
            pass
        return res


    from utils.log import log_icd
    @log_icd
    def check_diagnosability(self, property, context, observables, path_restrictions, delta_t_var_name=None):
        """Performs Diagnosability Check and returns ("True",null,null) if the two given
        propositional properties are guaranteed to be distinguishable given a set of
        observable variables and a set of (zero or more) path restrictions (specified in
        chronological order), where the latter restrict the state space to be considered.
        In case the two properties are not distinguishable, two traces are generated,
        where under the same observable behavior one ends up in property_A and the other
        in property_B. The return value is ("False", trace_A_fn, trace_B_fn), with trace_A/B
        being the names of the xml trace files.
        In the case of hybrid models with an infinite domain, not BDD but BMC is used; if
        no counterexample is found, the return value is ("Unknown", None, None), and
        in case it is, the answer's again ("False", trace_A_fn, trace_B_fn).
        @param property: a propositional smv formula
        @type property: string
        @param context: a propositional smv formula
        @type context: string
        @param observables: all flattened smv names of the observable variables
        @type observables: list of strings
        @param path_restrictions: propositional properties which have to apply to states
                                  of the FSM in the given order before diagnosability is
                                  considered (order in list is important; length of list
                                  can be zero, as this is an optional parameter)
        @type path_restrictions: list of strings
        """

        # create temporary observables file
        obs_filename = self._fname_prefix + "diag_obs.tmp"
        log.system.debug('Creating helper file for observables (%s).' % obs_filename)
        file = open(obs_filename, "w")        
        for obs in observables:
            file.write(obs + "\n")
        file.close()

        # build the twin plant of the model
        log.system.debug('Sending build_twin_plant command to NuSMV.')

        if delta_t_var_name is not None:
            self.send_cmd('build_twin_plant -o "%s" -a -p -t "%s"' % (obs_filename, delta_t_var_name))
        else:
            self.send_cmd('build_twin_plant -o "%s" -a -p' % (obs_filename))

        # create temporary file for diagnosability conditions (NuSMV format)
        diagcond_filename = self._fname_prefix + "diag_cond.tmp"
        log.system.debug('Creating helper file for diagnosability condition (%s).' % diagcond_filename)
        file = open(diagcond_filename, "w")
        file.write("CONDITION_L (" + context + " & (" + property + "))\n")
        file.write("CONDITION_R (" + context + " & !(" + property + "))\n")
        file.write("CONTEXT_R !(" + property + ")\n")
        for res in path_restrictions:
            file.write("RESTRICTION (" + res + ")\n")
        file.close()

        # load the diagnosability condition into the property manager
        # nb: all properties within the property manager are removed first (with '-p')
        log.system.debug('Sending load_diag_conditions command to NuSMV.')

        if options.DIAG_DELAY == -1.0: # NEEDS LTL CHECKING
            self.send_cmd('load_diag_conditions -p -c %s -d -1' % (diagcond_filename))
        elif options.DIAG_DELAY >= 0.0: # NEEDS INVAR CHECKING
            self.send_cmd('load_diag_conditions -p -c "%s" -d %f -r' % (diagcond_filename, options.DIAG_DELAY))
        else:
            ## we should never reach this code!!
            raise Exception("Delay must be >= 0 or 'FINITE'")

        # with hybrid models, bmc has to be used
        if self.needs_msat():
            log.system.debug('Sending go_msat command to NuSMV.')
            self.send_cmd('go_msat')

            if options.DIAG_DELAY == -1.0:
                log.system.debug('Sending msat_check_ltlspec command to NuSMV.')
                self.send_cmd('msat_check_ltlspec_bmc -k %d' % options.NUSMV_BMC_LENGTH)
            elif options.DIAG_DELAY >= 0.0:
                log.system.debug('Sending msat_check_invar_bmc command to NuSMV.')
                self.send_cmd('msat_check_invar_bmc -k %d -i -a falsification' % options.NUSMV_BMC_LENGTH)

            self.send_cmd("echo DONE BMC") # This is for detecting no result found
            grep = utils.process.TextLineGrepper(".*\s+is (false|true)", "DONE BMC")
            (m, txt) = grep.match(self)

            if m is None: result = "Error"
            elif grep.is_match_unwanted(m) and "DONE BMC" in txt: result = "Unknown"
            elif m.groups()[0] == "true": result = "True"
            else:
                assert m.groups()[0] == "false"
                result = "False"

        # discrete models are dealt with using bdd
        else:
            log.system.debug('Sending go command to NuSMV.')
            self.send_cmd('go')
            log.system.debug('Sending check_ltlspec command to NuSMV.')
            self.send_cmd('check_ltlspec')
            grep = utils.process.TextLineGrepper(".*\s+is (false|true)")
            (m, txt) = grep.match(self)

            if m is None: result = "Error"
            elif m.groups()[0] == "true": result = "True"
            else: result = "False"
       
        # return results
        if result == "Error":
            return (result, None, None)

        elif result == "True" or result == "Unknown":
            return (result, None, None)

        orig_trace_fn = self._create_trace()
        (trace_A_fn, trace_B_fn) = self.split_tracefile(orig_trace_fn)
        return (result, trace_A_fn, trace_B_fn)


    def split_tracefile(self, trace_filename):
        """This method takes a trace as generated by NuSMV in check_diagnosability(),
        and splits them into two traces, hiding internal implementation details of NuSMV.
        The filenames of the two new trace files are returned.
        @param trace_filename: the filename of the original trace
        @type trace_filename: string
        """

        # new trace files

        trace_fn_1 = trace_filename.replace('.xml', '-A.xml')
        trace_fn_2 = trace_filename.replace('.xml', '-B.xml')
        trace_1 = open(trace_fn_1, "w")
        trace_2 = open(trace_fn_2, "w")

        # parse trace file line by line

        orig_trace = open(trace_filename, "r")

        while 1:
            line = orig_trace.readline()
            if not line:
                break

            # skip twin plant stutter variables
            if re.search(r'"stutter_.">', line):
                continue
            # skip path restriction monitors
            if re.search(r'"have_C.*">', line):
                continue
            # skip history variable
            if re.search(r'"h_obs_eq">', line):
                continue
            # skip diag delay counter and *_rnd
            if re.search(r'"diag_delay_counter', line):
                continue

            # redirect line to correct file (and drop twin tags)
            if line.count('twin_L.') > 0:
                trace_1.write(line.replace('twin_L.', ''))
            elif line.count('twin_R.') > 0:
                trace_2.write(line.replace('twin_R.', ''))
            else:
                trace_1.write(line)
                trace_2.write(line)

        # done

        trace_1.close()
        trace_2.close()
        orig_trace.close()
        # os.remove(orig_trace)
        return (trace_fn_1, trace_fn_2)

    
    def __compute_anything(self, command, use_sat = False):
        """An internal function which runs an FSAP command with all the
        required checks, etc. May raise Exception.
        @param command: a string with complete one line FSAP command
        @param use_sat: optionally use SAT bmc instead of BDDs for finite systems
        """
        if self.needs_msat(): self.build_msat_fsm()
        elif use_sat: self.build_bmc_fsm()
        else: self.build_bdd_fsm()

        self.send_cmd(command)

        if not self.wait_for_command_finish():
            raise Exception("Unknown error in FSAP execution (advice: run with '-l debug' to see FSAP messages)")
            
        pass

    pass # end of class


########### A set of function to covert result of running FSAP  ################
########### I.e. restoring SLIM identifiers from SMV ones  ###########

    
import csv
import re
import string
import datetime
import os
import functools


#below strings define which characters can be the first and following
#characters in identifiers and constants.
_first_character = string.letters + string.digits + "_"
_second_character = string.letters + string.digits + "_.#@"
_id_pattern = "[" + _first_character + "][" + _second_character + "]*"

from utils.log import log_icd
@log_icd
def convert_fault_tree(ev_fname, gt_fname, smv2slim):
    """Converts the fault tree contained into the given files
    (events and gates).  

    @param ev_fname is an events file name of a fault tree
    generated by FSAP.

    @param gt_fname is an gates file name of a fault tree generated
    by FSAP.

    @param smv2slim is a conversion functions : SMV
    identifier/constant -> SLIM identifier/constant (i.e. string
    ->string, and if there is no corresponding SLIM id then None
    has to be returned).

    Names inside the files will be converted to SLIM language, if
    the option FSAP_KEEP_SMV_IDS is not set to True.

    The two files will be overwritten. An exception will be raised
    if there is an error condition."""

    if not options.FSAP_KEEP_SMV_IDS:
        fsap_process.convert_fault_tree_events(ev_fname, ev_fname, smv2slim)
        fsap_process.convert_fault_tree_gates(gt_fname, gt_fname, smv2slim)
        pass

    return 
    

def convert_fault_tree_events(in_file_name, out_file_name, smv2slim):
    """Takes an events files, converts SMV identifiers to SLIM ones
    and outputs the result.
    "in_file_name" is an events file name of a fault tree generated by FSAP.
    tool. "out_file_name" is a file name for output.
    "smv2slim" is a conversion functions : SMV identifier/constant ->
    SLIM identifier/constant (i.e. string ->string, and if there is
    no corresponding SLIM id then None has to be returned).
    Precondition: input file has to contains only convertibles SMV ids.
    Note: input and output files can be the same.
    """
    #Note for developers about the format:
    #The values are TAB separated, the records are "\r\n" separated,
    #there are 3 values the last of which contains one "\n".
    #The first record is the name of fields.
    #The third field is the one to be modified.
    #A special case: there may be only one record with values "false Basic false"

    # Since input and output files can be the same we have
    # to read everything and only then write back.

    content = [] # a list of records, i.e. a list of lists of values

    log.system.debug("Converting smv2slim(FSAP events): '%s' -> '%s'"%(in_file_name,out_file_name))

    in_file = open(in_file_name, "rb")
    reader = csv.reader(in_file, delimiter="\t")
    iter = reader.__iter__()

    # collect the file content
    try:
        # read the first record with titles
        values = iter.next()
        assert len(values) == 3, "Events file '%s' has to have exactly 3 fields" % in_file_name
        content.append(values)

        # iterate over the file content
        last = None
        while True:            
            try: 
                values = iter.next() if last is None else last                
                last = iter.next()
                
                assert len(values) == 3
                if len(last) == 1:
                    values[2] = values[2] + "\n" + last[0]
                    last = None
                    pass

                # converts the description : the string with expressions, i.e.
                # the third value in every record
                values[2] = _convert_smv_exp_to_slim(values[2], smv2slim)
                content.append(values)
                
            except StopIteration:
                if values: # special case: there is only one line with empty set  
                    if (len(content) == 1 and 
                        (values == ["false", "Basic", "false"] or 
                         values == ["true", "Basic", "true"])): content.append(values)

                    else: # last line is a normal triple
                        values[2] = _convert_smv_exp_to_slim(values[2], smv2slim)
                        content.append(values)
                        pass
                    pass                
                        
                break
            else: values = [] # for debugging in assertion above the format of file
            pass

    except csv.Error, e:\
        raise Exception('file %s, line %d: %s' % (in_file_name, reader.line_num, e))
    in_file.close()
    
    # Now write the file the modified content.
    # Note for developers:
    # standard csv.writer(...).write approach did not work well with
    # special characters such as escapechar, lineterminator and
    # quotechar. As result here we simply use the standard "write".
    out_file = open(out_file_name, "wb")
    for values in content:
        out_file.write("\t".join(values) + "\r\n")
    out_file.close()
    log.system.debug("Conversion smv2slim has been done")
    pass


def convert_fault_tree_gates(in_file_name, out_file_name, smv2slim):
    """Takes a gates files, converts SMV identifiers to SLIM ones
    and outputs the result.
    "in_file_name" is an gates file name of a fault tree generated by FSAP.
    tool. "out_file_name" is a file name for output.
    "smv2slim" is a conversion functions : SMV identifier/constant ->
    SLIM identifier/constant (i.e. string ->string, and if there is
    no corresponding SLIM id then None has to be returned).
    Precondition: input file has to contains only convertibles SMV ids.
    Note: input and output files can be the same.
    """
    #Note for developers about the format:
    #The values are TAB separated, the records are "\r\n" separated,
    #there are 22 values none of which may have newline "\n" character.
    #The first record is the name of fields.
    #The third field is the one to be modified.

    # Since input and output files can be the same we have
    # to read everything and only then write back.

    content = [] # a list of records, i.e. a list of lists of values
    
    log.system.debug("Converting smv2slim(FSAP gates): '%s' -> '%s'"%(in_file_name,out_file_name))

    in_file = open(in_file_name, "rb")

# This code fixes ticket:347 http://compass.informatik.rwth-aachen.de/trac/ticket/347
    # reads the header
    header_str = in_file.readline().strip()
    header = header_str.split("\t")
    cols = len(header)
    assert(cols) == 22
 
    content.append(header)
   
    # removes the newlines, substituting them with tabs
    rest_str = in_file.read().strip().replace("\r\n", "\t")
    rest = rest_str.split("\t")

    # content has to be a multiple of cols
    assert (len(rest) % cols) == 0, "rest="+str(rest)+" len="+str(len(rest)) 

    # divides the rest in chunks of size cols
    for i in range(0, len(rest), cols):
        content.append(rest[i:i+cols])
        pass

# Previous code substitutes this one:
# #     reader = csv.reader(in_file, delimiter="\t")

# #     # collect the file content
# #     try:        
# #         for values in reader:
# #             while len(values) < 22: 
# #                 subline = reader.next()
# #                 values += subline
# #                 print "values:", values, len(values)
# #                 assert len(values) <= 22, "value="+str(values)
# #                 pass
            
# #             content.append(values)
# #             pass
# #     except csv.Error, e:\
# #            raise Exception('file %s, line %d: %s' % (in_file_name, reader.line_num, e))
    in_file.close()

    # convert the description : the string with expressions, i.e.
    # the third value in every record
    for values in content[1:]:
        values[2] = _convert_smv_exp_to_slim(values[2],smv2slim)
    
    # Now write to the file the modified content.
    out_file = open(out_file_name, "wb")
    for values in content:
        out_file.write("\t".join(values) + "\r\n")
    out_file.close()
    log.system.debug("Conversion smv2slim has been done")
    pass

def convert_fault_tree_xml(in_file_name, out_file_name, smv2slim):
    """Takes a fault tree XML file, converts event descriptions from SMV
    format to SLIM format and outputs the result.
    Note: input and output files can be the same.
    
    @precondition:
        All event descriptions in the input file must be convertible to SMV
        identifiers.
    
    @param   in_file_name:  The name of a fault tree XML file generated by
                            FSAP.
    @type    in_file_name:  string
    @param   out_file_name: The output filename.
    @type    out_file_name: string
    @param   smv2slim:      A conversion function: SMV identifier/constant ->
                            SLIM identifier/constant. If there is no
                            corresponding SLIM identifier, then None has to be
                            returned.
    @param   smv2slim:      A function: string -> string/NoneType
    """
    # Define a function that returns for a given event description in SMV
    # format the corresponding SLIM expression.
    def convert_smv_to_slim(smv_description):
        # Special case: A basic event is just "true" or "false".
        # This can happen if the fault tree has only this single basic event.
        if smv_description in ("true", "false"):
            return smv_description
        else:
            return _convert_smv_exp_to_slim(smv_description, smv2slim)
    
    fault_tree = fault_tree_module.parse_xml(in_file_name)
    fault_tree_module.replace_event_descriptions(fault_tree,
                                                 convert_smv_to_slim)
    with open(out_file_name, "w") as out_file:
        out_file.write(fault_tree_module.get_as_xml_string(fault_tree))
        
    pass

from utils.log import log_icd
@log_icd
def convert_fmea_table(in_file_name, out_file_name, smv2slim):
    """Takes a "FMEA table" file, converts SMV identifiers
    to SLIM ones and outputs the result.
    "smv2slim" has to be a string-to-string conversion function which
    converts SMV identifier/constant to SLIM identifier/constant (if there is
    no corresponding SLIM id then None has to be returned).
    Precondition: input file has to contains only convertibles SMV ids.
    Note: input and output files can be the same.
    """
    #Note for developers about the format:
    #Every line is a record of 3 values.
    #The first line is with titles and remains unchanged.
    #The second and third values in consecutive lines are expressions
    #to be modified.
    
    # Since input and output files can be the same we have
    # to read everything and only then write back.

    content = [] # a list of lines

    log.system.debug("Converting smv2slim (FSAP FMEA): '%s' -> '%s'"%(in_file_name,out_file_name))

    in_file = open(in_file_name, "rb")
    reader = csv.reader(in_file, delimiter="\t")

    # collect the file content
    try:
        for values in reader:
            assert len(values) == 4, "\"FMEA table\" file %s, line %d: there must be 3 values in a row (got %d)" % (in_file_name, reader.line_num, len(values))
            content.append(values)
            pass
    except csv.Error, e:\
           raise Exception('file %s, line %d: %s' % (in_file_name, reader.line_num, e))
    in_file.close()

    # convert the description : the string with expressions, i.e.
    # the second and third values in every record
    for values in content[1:]:
        values[1] = _convert_smv_exp_to_slim(values[1],smv2slim)
        values[2] = _convert_smv_exp_to_slim(values[2],smv2slim)
    
    # Now write to the file the modified content.
    out_file = open(out_file_name, "wb")
    for values in content:
        out_file.write("\t".join(values) + "\r\n")
    out_file.close()
    log.system.debug("Conversion smv2slim has been done")
    pass


def convert_fault_detection(in_file_name, out_file_name, smv2slim):
    """Takes a "fault detection means" files, converts SMV identifiers
    to SLIM ones and outputs the result.
    "smv2slim" has to be a string-to-string conversion function which
    converts SMV identifier/constant to SLIM identifier/constant (if there is
    no corresponding SLIM id then None has to be returned).
    Precondition: input file has to contains only convertibles SMV ids.
    Note: input and output files can be the same.
    """
    #Note for developers about the format:
    #The first line is a comment and every consecutive file
    #contains an expression (probably, just one name) to be modified.
    
    # Since input and output files can be the same we have
    # to read everything and only then write back.

    content = [] # a list of lines
 
    log.system.debug("Converting smv2slim (FSAP fault detection): '%s' -> '%s'"%(in_file_name,out_file_name))

    in_file = open(in_file_name, "rb")
    
    content.append(in_file.readline()) # read the first (special) line
    for line in in_file: # read and convert the rest text
        line = _convert_smv_exp_to_slim(line, smv2slim)
        content.append(line)
        pass
    
    # Now write to the file the modified content.
    out_file = open(out_file_name, "wb")
    for values in content: out_file.write(values)
    out_file.close()
    log.system.debug("Conversion smv2slim has been done")
    pass


def _str2str_wrapper(str2str):
    """This is wrapper which takes a str->str conversion function
    and returns a function which performs the same
    conversion but with:
    1. a match object from re module as argument
    2. an assertion that the result is not None"""
    def wrapper(a_match):
        a_str = a_match.group(0) # get the string
        s = str2str(a_str)
        return s if s != None else a_str
    functools.update_wrapper(wrapper, str2str) # updates fun attributes
    return wrapper


def _convert_smv_exp_to_slim(in_str, id_smv2slim):
    """Takes a string with simple SMV expression and returns a
    corresponding expression with SLIM identifiers.  Precondition:
    currently the input expression is expected to have only those SMV
    identifiers which can have SLIM counterparts"""
    return re.sub(_id_pattern, _str2str_wrapper(id_smv2slim), in_str)


# ----------------------------------------------------------------------
class FTDProcess (utils.process.InteractiveProcess):
    """Process handler for running the Fault Tree Viewer"""

    def __init__(self, event_fname, gates_fname, 
                 args=[], stdout=None, stderr=None, dumpers=[]):

        # check that files exist
        for fname in (event_fname, gates_fname):
            if fname and not os.path.isfile(fname): 
                msg = "File not found: %s" % fname
                log.system.error(msg)
                raise IOError(msg)
            pass
        
        args = list(args)
        args += [event_fname, gates_fname]
        
        path = utils.process.get_exec_name(options.FTD)
        super(FTDProcess, self).__init__(path, args, stdout=stdout, stderr=stderr, 
                                         dumpers=dumpers)

        return
    pass # end of class
