# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


import logging
import options
import os

import utils.misc
import functools

__all__ = ["system", "user", "testing", "metrics", "init", "get_level_from_string"]

# Initially, a simple logging is offered for system
system = logging.getLogger("System") # global system logger
testing = logging.getLogger("Testing") # global logger used for testing
metrics = logging.getLogger("Metrics") # global logger used for metrics
user = None # global user logger

# Allow multiple loggers using the same file
# See: http://mail.python.org/pipermail/tutor/2007-December/059158.html
class ManagedFileHandler(logging.FileHandler):
    """Subclass of logging.FileHandler to cope with multiple handlers
    targeting the same file"""

    """Map of open filenames to streams"""
    open_files = {}
    
    def __init__(self, filename, mode='a', encoding=None, delay=0):
        logging.FileHandler.__init__(self, filename, mode, encoding)

    def _open(self):
        # First check map of open streams
        if self.baseFilename in ManagedFileHandler.open_files:
            return ManagedFileHandler.open_files[self.baseFilename]
        # if not present, call super and store result in map
        stream = logging.FileHandler._open(self)
        ManagedFileHandler.open_files[self.baseFilename] = stream
        return stream

def init():
    """Initialises the logging system according to data in
    options. Note: options must be already set before calling it"""
    global system
    global testing
    global metrics
    global user
    
    # ----------------------------------------
    #    System logger 
    # ----------------------------------------
    system = _SystemLogger()

    # ----------------------------------------
    #    Testing logger 
    # ----------------------------------------
    testing = _TestingLogger()
    
    # ----------------------------------------
    #    Metrics logger 
    # ----------------------------------------
    metrics = _MetricsLogger()

    # ----------------------------------------
    #    User logger 
    # ----------------------------------------
    user = _UserLogger()    
    return


def get_level_from_string(level_name):
    """Given a string, returns the corresponding logging
    level. Raise ValueError if bad string is passed"""
    if not level_name: return logging.INFO # default

    LEVELS = {'debug': logging.DEBUG,
              'info': logging.INFO,
              'warning': logging.WARNING,
              'error': logging.ERROR,
              'critical': logging.CRITICAL}

    if level_name not in LEVELS: 
        raise ValueError("Logging level must be in '%s'" % ", ".join(LEVELS.keys()))

    return LEVELS[level_name]


def log_icd(func):
    """This is a decorator which can be used to mark functions
    as logged for the ICD compliance testing. It will result 
	in a message to the testing logger when the function is used."""
    
    @functools.wraps(func)
    def new_func(*args, **kwargs):
        testing.info(
            "Call to ICD function %(funcname)s." % {
            'funcname': func.__module__ + "." + func.__name__, }
        )
        return func(*args, **kwargs)
    return new_func	



# -----------------------------------------------------------
class _Logger (object):
    def __init__(self, logger_name):
        self.logger = logging.getLogger(logger_name)
        self.fch = None

        if options.LOGGING_FNAME:
            logfname = options.LOGGING_FNAME
        elif options.BASE_FILE_NAME:
            logfname = utils.misc.get_out_fname(options.BASE_FILE_NAME + ".log")
        else:
            logfname = utils.misc.get_out_fname("compass.log")

        # file handler (if log name)
        if logfname:
            fch = ManagedFileHandler(logfname, "w")
            fch.setLevel(get_level_from_string(options.LOGGING_SYS_LEVEL))
            formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
            fch.setFormatter(formatter)
            self.logger.addHandler(fch)
            self.fch = fch
        
        return
    pass # end of class
# -----------------------------------------------------------

class _SystemLogger (_Logger):
    def __init__(self, suffix_name=""):
        _Logger.__init__(self, "System" + suffix_name)

        self.logger.setLevel(logging.DEBUG)

        # console handler
        ch = logging.StreamHandler()
        ch.setLevel(logging.ERROR)
        formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
        ch.setFormatter(formatter)
        self.logger.addHandler(ch)

        return

    def debug(self, msg, *args, **kwargs): self.logger.debug(msg, *args, **kwargs)
    def info(self, msg, *args, **kwargs): self.logger.info(msg, *args, **kwargs)
    def warning(self, msg, *args, **kwargs): self.logger.warning(msg, *args, **kwargs)
    def error(self, msg, *args, **kwargs): self.logger.error(msg, *args, **kwargs)
    def critical(self, msg, *args, **kwargs): self.logger.critical(msg, *args, **kwargs)
    pass # end of class
# -----------------------------------------------------------


class _TestingLogger (_Logger):
    def __init__(self, suffix_name=""):
        _Logger.__init__(self, "Testing" + suffix_name)

        self.logger.setLevel(logging.DEBUG)
        # console handler
        ch = logging.StreamHandler()
        ch.setLevel(logging.ERROR)
        formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
        ch.setFormatter(formatter)
        self.logger.addHandler(ch)
        
        if self.fch:
            self.fch.setLevel(get_level_from_string(options.LOGGING_TEST_LEVEL))

        return

    def debug(self, msg, *args, **kwargs): self.logger.debug(msg, *args, **kwargs)
    def info(self, msg, *args, **kwargs): self.logger.info(msg, *args, **kwargs)
    def warning(self, msg, *args, **kwargs): self.logger.warning(msg, *args, **kwargs)
    def error(self, msg, *args, **kwargs): self.logger.error(msg, *args, **kwargs)
    def critical(self, msg, *args, **kwargs): self.logger.critical(msg, *args, **kwargs)
    pass # end of class

class _MetricsLogger (_Logger):
    def __init__(self, suffix_name=""):
        _Logger.__init__(self, "Metrics" + suffix_name)

        self.logger.setLevel(logging.DEBUG)
        # console handler
        ch = logging.StreamHandler()
        ch.setLevel(logging.ERROR)
        formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
        ch.setFormatter(formatter)
        self.logger.addHandler(ch)

        return

    def debug(self, msg, *args, **kwargs): self.logger.debug(msg, *args, **kwargs)
    def info(self, msg, *args, **kwargs): self.logger.info(msg, *args, **kwargs)
    def warning(self, msg, *args, **kwargs): self.logger.warning(msg, *args, **kwargs)
    def error(self, msg, *args, **kwargs): self.logger.error(msg, *args, **kwargs)
    def critical(self, msg, *args, **kwargs): self.logger.critical(msg, *args, **kwargs)
    pass # end of class

# -----------------------------------------------------------


class _UserLogger (_Logger):
    def __init__(self):
        _Logger.__init__(self, "User")

        # sets up the logger
        self.logger.setLevel(logging.INFO)

        # console handler
        ch = logging.StreamHandler()
        ch.setLevel(get_level_from_string(options.LOGGING_USER_LEVEL))
        # Only print simple messages
        formatter = logging.Formatter("%(message)s")
        ch.setFormatter(formatter)
        self.logger.addHandler(ch)
        
        # Note: file logger uses SYS level

        return

    def info(self, msg, *args, **kwargs): self.logger.info(msg, *args, **kwargs)
    def warning(self, msg, *args, **kwargs): self.logger.warning(msg, *args, **kwargs)
    def error(self, msg, *args, **kwargs): self.logger.error(msg, *args, **kwargs)
    pass # end of class
# -----------------------------------------------------------


import gobject
class GuiHandler (logging.Handler):
    """This class is used to send logging info to a gtk.textview
    (in a GUI)"""

    def __init__(self, text_buffer, level=logging.NOTSET):
        """text)buffer is a gtk.TextBuffer instance to be used to
        send log records. It must support tag named 'timestamp' and
        'level'"""
        logging.Handler.__init__(self, level)
        formatter = logging.Formatter("[%(asctime)s] %(levelname)s: \t%(message)s")
        self.setFormatter(formatter)

        self.textbuf = text_buffer
        return

    def emit(self, record):
        def callback(buf, msg): 
            # separates the timestamp and the level
            if msg[0] == '[':
                idx = msg.find(']')+1
                if idx > 0: 
                    buf.insert_with_tags_by_name(buf.get_end_iter(), msg[0:idx], "timestamp")
                    idx2 = msg.find(':', idx)+1
                    if idx2 > 0:
                        buf.insert_with_tags_by_name(buf.get_end_iter(), msg[idx:idx2], "level")
                        buf.insert(buf.get_end_iter(), msg[idx2:])
                    else: # no match for the level
                        buf.insert(buf.get_end_iter(), msg[idx:])
                        pass
                    
                # no match for timestamp
                else: buf.insert(buf.get_end_iter(), msg)
            else: # no match for timestamp
                buf.insert(buf.get_end_iter(), msg)
                pass
            return
        
        # this is done to protect from multithreading calls
        msg = self.format(record)
        gobject.idle_add(callback, self.textbuf, msg + os.linesep)
        return

    pass # end of class
# -----------------------------------------------------------
