# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


# General utilities

import os
import os.path
import tempfile
import string
import random

import options
import utils.log


# This defines relpath function
"""relpath substitutes os.path.relpath offered by python 2.6 and later"""
relpath = None
  
_cnt = 0
def get_progressive_name(length):
    """Returns a progressive counter string of given length. The returned
    string can be used for filename suffixes."""
    global _cnt

    _cnt = (1 + _cnt) % (10 ** length)
    return str(_cnt).zfill(length)

def get_random_name(length):
    """Returns a random string whose length is given. The returned
    string is good for filename prefixes/suffixes"""
    if length <= 0: return ""
    digits = string.ascii_letters + string.digits + "_"
    return (random.choice(string.ascii_letters) + 
            ''.join((random.choice(digits) for x in range(length-1))))

def get_out_fname(fname, override_out_dir=None):
    """This function can be used to retrieve the complete path to a
    file name which it is needed to be generated. It provides a way
    of creating all files in a consistent manner. It considers what
    the user specified in file options.py, in particular
    OUT_DIR_NAME is checked. If empty, then the final part of
    BASE_FILE_NAME is used.

    The fname is separated from the directory part, and only the
    suffix is appended to the directory part supported by
    options.py"""

    # In case absolute file path is provided, do not attempt to merge with directory
    if os.path.isabs(fname):
        return fname

    if options.OUT_DIR_NAME:        
        dirname = options.OUT_DIR_NAME if os.path.isabs(options.OUT_DIR_NAME) \
            else os.path.abspath(options.OUT_DIR_NAME) # absolute path normalization
    elif override_out_dir:
        dirname = override_out_dir
    else:
        def __get_random_dirname(basename):
            """Returns a directory name which has given
            basename. If existing, a suffix is added"""                        
            return "." if not basename else basename

        # the user did not specify an output directory, one is
        # created out of the base name if available.
        dirname = __get_random_dirname(options.BASE_FILE_NAME)
        pass
    
    if not os.path.isdir(dirname): 
        os.makedirs(dirname)
        utils.log.system.info("Created directory '%s' for collecting output files" % dirname)
    else: utils.log.system.info("Keeping already existing directory '%s' for collecting output files" % dirname)

    # setting the tempfile generator to generate tempfiles by default 
    # in dirname
    if not override_out_dir:
        tempfile.tempdir = dirname
    
    return os.path.join(dirname, os.path.basename(fname))


# ----------------------------------------------------------------------
# ----------------------------------------------------------------------

# This is taken from python 2.6 (os.path.relpath is not supported in 2.5)
def __posix_relpath(path, start=os.curdir):
    """Return a relative version of a path"""

    if not path: raise ValueError("no path specified")

    start_list = os.path.abspath(start).split(os.sep)
    path_list = os.path.abspath(path).split(os.sep)

    # Work out how much of the filepath is shared by start and path.
    i = len(os.path.commonprefix([start_list, path_list]))

    rel_list = [os.pardir] * (len(start_list)-i) + path_list[i:]
    if not rel_list: return os.curdir
    return os.path.join(*rel_list)

# This is taken from python 2.6 (os.path.relpath is not supported in 2.5)
# This is for windows
def __nt_relpath(path, start=os.curdir):
    """Return a relative version of a path"""

    if not path: raise ValueError("no path specified")

    start_list = os.path.abspath(start).split(os.sep)
    path_list = os.path.abspath(path).split(os.sep)
    if start_list[0].lower() != path_list[0].lower():
        unc_path, rest = os.path.splitunc(path)
        unc_start, rest = os.path.splitunc(start)
        if bool(unc_path) ^ bool(unc_start):
            raise ValueError("Cannot mix UNC and non-UNC paths (%s and %s)"                                                                % (path, start))
        else: raise ValueError("path is on drive %s, start on drive %s"
                                                % (path_list[0], start_list[0]))
    # Work out how much of the filepath is shared by start and path.
    for i in range(min(len(start_list), len(path_list))):
        if start_list[i].lower() != path_list[i].lower():
            break
    else: i += 1

    rel_list = [os.pardir] * (len(start_list)-i) + path_list[i:]
    if not rel_list: return os.curdir
    return os.path.join(*rel_list)


try: 
    import os.path.relpath
    relpath = os.path.relpath
except ImportError: relpath = __nt_relpath if os.name == 'nt' else __posix_relpath


# ----------------------------------------------------------------------
#               GCD Utils Functions
# ----------------------------------------------------------------------
        
def compute_greatest_common_divisor(list_numbers):
    """Computes the GCD of a list of numbers.
        @param   list_numbers:  The first number.
        @type    list_numbers:  float
      
        @param   result:  The GCD of a list of numbers
        @type    result:  float
    """
    import fractions
    return reduce(fractions.gcd, list_numbers)
