# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================


import os.path
import re

import options
import utils.process
import utils.log as log
from utils.exceptions import NusmvControlledException, TraceException


class NuSMVShellProcess (utils.process.InteractiveProcess):
    """This class wraps a nusmv-based process, but not necessarily
    called through the NuSMV executable."""

    def __init__(self, path, need_msat=False,
                 args=[], stdout=None, stderr=None, dumpers=[]):
        super(NuSMVShellProcess, self).__init__(path, args,
                                                stdout=stdout, stderr=stderr,
                                                dumpers=dumpers)
        self.__build_cmd = set()
        self._need_msat = need_msat

    def needs_msat(self):
        """Returns True if current nusmv process needs mathsat
        (deals with infinite domains)"""
        return self._need_msat

    def __build(self, cmd):
        if cmd not in self.__build_cmd:
            self.send_cmd("%s; echo DONE_CMD" % cmd)
            grep = utils.process.TextLineGrepper("DONE_CMD")
            # wait for echoed message or the process death
            m, txt = grep.match(self)
            self._check_handler_errors(txt)
            self.__build_cmd.add(cmd)

    def build_bdd_fsm(self):
        assert not self._need_msat, "Only MATHSAT can be used"
        self.__build("go")

    def build_bmc_fsm(self):
        assert not self._need_msat, "Only MATHSAT can be used"
        self.__build("go_bmc")

    def build_msat_fsm(self):
        self.__build("go_msat")

    def build_hierarchy(self):
        """Builds the hierarchy if needed. Returns True iff
        everthing went ok"""

        if (not set(("build_hrc", "go", "go_bmc", "go_msat")) &
                self.__build_cmd):
            self.send_cmd("read_model")
            self.send_cmd("flatten_hierarchy")
            self.send_cmd("encode_variables")
            self.__build_cmd.add("build_hrc")

            return self.wait_for_command_finish()

        return True

    def reset(self):
        self.send_cmd("reset")
        self.__build = set()

    def kill(self):
        if not self.is_alive():
            return
        self.send_cmd("quit -x")
        super(NuSMVShellProcess, self).kill()

    def wait_for_command_finish(self):
        """This function waits for the termination of
        previously issued commands and then flushes (empties)
        the stdout and stderr stream of the process.
        Returns True if the process is still running and
        responds the commands, or False otherwise.
        """
        # this messages is used to detect the end of FSAP command
        # execution.  it must be unique (not output by normal NuSMV
        # run), without spaces at the ends and simple enough for python's re.
        _end_message = "END OF NuSMV COMMAND EXECUTION ~"
        self.send_cmd('echo %s' % _end_message)

        grep = utils.process.TextLineGrepper(_end_message, "Division by zero")
        # wait for echoed message or the process death
        m, txt = grep.match(self)

        assert (m is not None) == self.is_alive(), \
            "NuSMV can/must respond iff it is alive (this is guaranteed by" \
            "'set on_failure_script_quits' in __init__)" \
            + str(m) + str(self.is_alive())

        # empty the streams of the process. it is actually not useful here
        self.flush_stderr_stdout()
        if m and grep.is_match_unwanted(m):
            raise NusmvControlledException(m.string)

        return m is not None

    def _check_handler_errors(self, txt):
        """There is a set of failures which can cause nusmv to
        fail. These failures are told to be 'handled' and an
        assertion is prepared and raised in any of these conditions
        is met in the given txt"""

        if self.is_alive():
            return
        handled_errors = ("Division by zero",)

        lines = txt.splitlines()
        for idx, line in zip(range(len(lines)), lines):
            for err in handled_errors:
                # matched: returns from this line to the end
                if err in line:
                    raise NusmvControlledException("\n".join(lines[idx:]))

    pass  # end of class
# ----------------------------------------------------------------------


class NuSMVProcess (NuSMVShellProcess):
    """Process handler for running NuSMV interactively"""

    # these are literals used in is_state_zeno result
    (ZENO, NON_ZENO, UNREACHABLE, UNKNOWN_NOT_FOUND,
     UNKNOWN_NOT_PROVED, ERROR) = range(6)

    # these are literals used in check_clock_divergence
    (BOUNDED, UNBOUNDED, UNKNOWN) = range(3)

    def __init__(self, smv_fname=None, ord_fname=None, need_msat=False,
                 args=[], stdout=None, stderr=None, dumpers=[]):

        if ord_fname is None and smv_fname is not None:
            # search if there is one possible ordering available
            basename = os.path.splitext(smv_fname)[0]
            ord_fname = basename + ".ord"
            if not os.path.isfile(ord_fname):
                ord_fname = None

        # check that files exist
        for fname in (smv_fname, ord_fname):
            if fname and not os.path.isfile(fname):
                msg = "File not found: %s" % fname
                log.system.error(msg)
                raise IOError(msg)

        args = list(args)
        args += ["-int"]

        # first tries the fully operative version of nusmv, then if
        # failing falls back to the standard nusmv
        self.__degraded = False
        try:
            path = utils.process.get_exec_name(options.NUSMV)
        except IOError:
            path = utils.process.get_exec_name(options.DEGRADED_NUSMV)
            self.__degraded = True
            pass

        super(NuSMVProcess, self).__init__(path, need_msat, args,
                                           stdout=stdout, stderr=stderr,
                                           dumpers=dumpers)

        self._smv_fname = smv_fname
        self._ord_fname = ord_fname
        self.__trace_counter = 0

        # a few default commands sent to NuSMV:
        # 1. exit immediately on any error. This is important and
        # used in how commands success is checked.
        # 2. enable dynamic BDD var order. By default it is advantageous
        self.send_cmd("set on_failure_script_quits")
        self.send_cmd("dynamic_var_ordering -e sift")
        if self._smv_fname:
            self.send_cmd("set input_file '%s'" % self._smv_fname)
        if self._ord_fname:
            self.send_cmd("set input_order_file '%s'" % self._ord_fname)

    @classmethod
    def zeno_result_to_string(cls, res):
        _map = dict(zip(
            [cls.ZENO, cls.NON_ZENO, cls.UNREACHABLE, cls.UNKNOWN_NOT_FOUND,
             cls.UNKNOWN_NOT_PROVED, cls.ERROR],
            ["ZENO", "NON-ZENO", "UNREACHABLE", "NOT_FOUND",
             "NOT_PROVED", "ERROR"]))
        return _map[res]

    @classmethod
    def clock_diverge_result_to_string(cls, res):
        _map = dict(zip(
            [cls.BOUNDED, cls.UNBOUNDED, cls.UNKNOWN, cls.ERROR],
            ["BOUNDED", "UNBOUNDED", "UNKNOWN", "ERROR"]))
        return _map[res]

    def is_degraded_mode(self):
        """Returns True if the process supports full operations (namely:
        sa, diag, infinite domains)"""
        return self.__degraded

    def model_check_bdd_ctl(self, formula):
        """Model checks a formula (string), and returns: (True,
        None) if the formula is true, (False, fname) if
        false. fname is the filename of a xml trace containing the
        countexample. Returns (None, None) if no result could be
        found"""
        self.build_bdd_fsm()

        cmd = 'check_ctlspec -p "%s"' % formula
        self.send_cmd(cmd)

        # this grepper checks for TRUE/FALSE, the other for problems in
        # trace generation
        grep = utils.process.TextLineGrepper(".*\s+is (false|true)")
        m, txt = grep.match(self)  # wait for result!
        if m is None:
            log.system.error("Process failed on command:\n> " + cmd)
            self._log_returned_text(txt)
            return (None, None)

        self._check_handler_errors(txt)

        # processes results
        if m.groups()[0] == "true":
            return (True, None)

        return (False, self._create_trace())

    def model_check_bdd_ltl(self, formula):
        """Model checks a formula (string), and returns: (True,
        None) if the formula is true, (False, fname) if
        false. fname is the filename of a xml trace containing the
        countexample. Returns (None, None) if no result could be
        found"""
        self.build_bdd_fsm()
        cmd = 'check_ltlspec -p "%s"' % formula
        self.send_cmd(cmd)

        grep = utils.process.TextLineGrepper(".*\s+is (false|true)")
        m, txt = grep.match(self)  # wait for result!
        if m is None:
            log.system.error("Process failed on command:\n> " + cmd)
            self._log_returned_text(txt)
            return (None, None)

        self._check_handler_errors(txt)

        # processes results
        if m.groups()[0] == "true":
            return (True, None)

        return (False, self._create_trace())

    def model_check_bmc_ltl(self, formula, k):
        """Model checks a formula (string), and returns: (True,
        None) if the formula is true, (False, fname) if
        false. fname is the filename of a xml trace containing the
        countexample. Returns (None, None) if no result could be
        found"""

        # Use MATHSAT if needed
        if self._need_msat:
            return self.model_check_msat_bmc_ltl(formula, k)

        # else use BMC
        self.build_bmc_fsm()

        if options.NUSMV_BMC_USE_SBMC:
            compl_opt = "-c" if options.NUSMV_SBMC_IS_COMPLETE else ""
            cmd = ('check_ltlspec_sbmc_inc %s -p "%s" -k %d' %
                   (compl_opt, formula, k))
        else:
            cmd = ('check_ltlspec_bmc_inc -p "%s" -k %d' %
                   (formula, k))

        self.send_cmd(cmd)
        # This is for detecting no result found
        self.send_cmd("echo DONE_BMC")

        grep = utils.process.TextLineGrepper([".*\s+is (false|true)",
                                              "DONE_BMC"])
        m, txt = grep.match(self)  # wait for result!
        if m is None or grep.is_match_unwanted(m):
            log.system.error("Process failed on command:\n> " + cmd)
            self._log_returned_text(txt)
            return (None, None)

        self._check_handler_errors(txt)

        if not m.groups():
            # no result
            return (None, None)

        if m.groups()[0] == "true":
            return (True, None)

        return (False, self._create_trace())

    def model_check_msat_bmc_ltl(self, formula, k):
        """Model checks a formula (string), and returns: (True,
        None) if the formula is true, (False, fname) if
        false. fname is the filename of a xml trace containing the
        countexample. Returns (None, None) if no result could be
        found"""
        self.build_msat_fsm()

        if options.NUSMV_BMC_USE_SBMC:
            log.system.warning(
                "SBMC with mathsat not yet available, using BMC")

        cmd = 'msat_check_ltlspec_sbmc_inc -p "%s" -k %d' % (formula, k)
        self.send_cmd(cmd)
        self.send_cmd("echo DONE BMC")  # This is for detecting no result found

        grep = utils.process.TextLineGrepper([".*\s+is (false|true)",
                                              "DONE BMC"])
        m, txt = grep.match(self)  # wait for result!

        # processes results
        if m is None or grep.is_match_unwanted(m):
            log.system.error("Process failed on command:\n> " + cmd)
            self._log_returned_text(txt)
            return (None, None)

        self._check_handler_errors(txt)

        if not m.groups():
            # no result found
            return (None, None)
        elif m.groups()[0] == "true":
            return (True, None)

        return (False, self._create_trace())

    def goto_state(self, trace_index, step_index):
        """Picks state from existing trace using BDD/BMC or MSAT, and
        creates a new trace. Returns the filename of the
        newly created trace."""
        # Use MATHSAT if needed

        # TODO: temporary hack: handle the case in which trace_index is invalid
        if trace_index is None or trace_index <= 0:
            raise NusmvControlledException("""Command goto-state received """
                                           """an invalid trace.
This may be due to a known issue in nusmv. However, please report this bug.""")

        if self._need_msat:
            idx = self._goto_state_msat(trace_index, step_index)
        else:
            idx = self._goto_state_bdd(trace_index, step_index)
            if idx is None:
                return None
            idx = self._goto_state_bmc(idx, -1)

        assert self.get_traces_num() == idx, \
            "The generated trace is not the last one as expected"
        res = self._create_trace()
        if res is None:
            raise NusmvControlledException("""Command goto-state seems """
                                           """to be not working correctly.
This may be due to a known issue in nusmv. However, please report this bug.""")
        return res

    def pick_state_random(self, constraints):
        """Picks state randomly using BDD or MSAT.
        @param constraints is a list of smv constraints (list of
        strings).
        @return the file name of the created trace, or
        None if no trace was created."""
        # Use MATHSAT if needed
        idx = self._pick_state_random_msat(constraints) if self._need_msat \
            else self._pick_state_random_bdd(constraints)

        if idx is None:
            return None
        # picks also in BMC if possible
        if not self._need_msat:
            idx = self._goto_state_bmc(idx, 1)

        return self._create_trace()

    def simulate_random(self, constraints, length):
        """Simulate using BDD or MSAT. The state is randomly
        selected taking into account the list of SMV constraints
        (list of strings). Returns the name of the created xml trace, or
        None if a problem occurs"""

        # Use MATHSAT if needed
        if self._need_msat:
            idx = self._simulate_random_msat(constraints, length)
        else:
            idx = self._simulate_random_bdd(constraints, length)
            if idx is None:
                return None
            idx = self._goto_state_bmc(idx, -1)

        return self._create_trace()

    def simulate_check_feasability(self, constraints):
        """Constraints is a list of next formulae
        (strings). Returns a list of boolean values telling
        positionally which constraints are true or false."""
        # Use MATHSAT if needed
        return (self._simulate_check_feasability_msat(constraints)
                if self._need_msat
                else self._simulate_check_feasability_bmc(constraints))

    def get_traces_num(self):
        """Returns the number (integer) of traces currently loaded
        in the trace database. Returns None if an error occurs"""
        grep = utils.process.TextLineGrepper(
            ["There (is|are) (\d+|no) traces? currently available."])
        cmd = 'show_traces -t'
        self.send_cmd(cmd)
        m, txt = grep.match(self)
        if m is None:
            log.system.error("Process failed on command:\n> " + cmd)
            self._log_returned_text(txt)
            return None

        return (0
                if m.group(2) == "no"
                else int(m.group(2)))

    def read_trace(self, fname):
        """Reads given trace from file, and returns the numeric
        index (as integer) which was used to store the trace
        internally. The index can later be used to pick states from
        traces. If an error occurs, returns None"""
        if not self.build_hierarchy():
            return None

        cmd = "build_flat_model"
        self.send_cmd(cmd)

        cmd = 'read_trace "%s"' % fname
        self.send_cmd(cmd)
        grep = utils.process.TextLineGrepper("Trace is stored at (\d+) index",
                                             "Division by zero")
        m, txt = grep.match(self)
        if m is None:
            log.system.error("Process failed on command:\n> " + cmd)
            self._log_returned_text(txt)
            return None

        if grep.is_match_unwanted(m):
            raise TraceException(m.string)

        return int(m.group(1))

    def check_fsm_deadlocks(self):
        """Checks the BDD fsm for deadlocks. Returns True if the
        FSM is deadlock-free, False if it contains deadlocks, or
        None if an error occurs."""
        self.build_bdd_fsm()

        cmd = 'check_fsm'
        self.send_cmd(cmd)
        grep = utils.process.TextLineGrepper(
            ["The transition relation is total: No deadlock state exists",
             "However, all the states without successors are"],
            ["The transition relation is not deadlock-free."],)

        m, txt = grep.match(self)
        if m is None:
            log.system.error("Process failed on command:\n> " + cmd)
            self._log_returned_text(txt)
            return None

        return (False
                if grep.is_match_unwanted(m)
                else True)

    def check_pattern_absence(self, formula, time_bound, time_var,
                              max_search_depth, max_klive_bound):
        """Invokes hycomp_check_pattern_absence.
        Returns:
           (True, None) if property is true
           (False, trace_fname) with trace_fname xml file of generated trace
           (None, None) if checking result is unknown
        """
        return self._check_pattern_aux("absence",
                                       [formula],
                                       time_bound, time_var,
                                       max_search_depth, max_klive_bound)

    def check_pattern_existence(self, formula, time_bound, time_var,
                                max_search_depth, max_klive_bound):
        """Invokes hycomp_check_pattern_existence.
        Returns:
           (True, None) if property is true
           (False, trace_fname) with trace_fname xml file of generated trace
           (None, None) if checking result is unknown
        """
        return self._check_pattern_aux("existence",
                                       [formula],
                                       time_bound, time_var,
                                       max_search_depth, max_klive_bound)

    def check_pattern_universality(self, formula, time_bound, time_var,
                                   max_search_depth, max_klive_bound):
        """Invokes hycomp_check_pattern_universality.
        Returns:
           (True, None) if property is true
           (False, trace_fname) with trace_fname xml file of generated trace
           (None, None) if checking result is unknown
        """
        return self._check_pattern_aux("universality",
                                       [formula],
                                       time_bound, time_var,
                                       max_search_depth, max_klive_bound)

    def check_pattern_precedence(self, formula_p, formula_q,
                                 time_bound, time_var,
                                 max_search_depth, max_klive_bound):
        """Invokes hycomp_check_pattern_precedence.
        Returns:
           (True, None) if property is true
           (False, trace_fname) with trace_fname xml file of generated trace
           (None, None) if checking result is unknown
        """
        return self._check_pattern_aux("precedence",
                                       [formula_p, formula_q],
                                       time_bound, time_var,
                                       max_search_depth, max_klive_bound)

    def check_pattern_response_global(self, formula_p, formula_q,
                                      time_bound, time_var,
                                      max_search_depth, max_klive_bound):
        """Invokes hycomp_check_pattern_response_global.
        Returns:
           (True, None) if property is true
           (False, trace_fname) with trace_fname xml file of generated trace
           (None, None) if checking result is unknown
        """
        return self._check_pattern_aux("responseGlobal",
                                       [formula_p, formula_q],
                                       time_bound, time_var,
                                       max_search_depth, max_klive_bound)

    def is_state_zeno(self, formula, time_var,
                      max_search_depth=options.NUSMV_BMC_LENGTH,
                      max_klive_bound=options.NUSMV_KLIVE_BOUND):
        """Performs state-zeno check.
        Takes:
           formula: for state to check (string)
           time_var: name of time variable
           max_seatch_depth: depth of search bound (integer)
           max_klive_bound: bound for k-liveness check (integer)

        Returns one of these pairs:
           (NuSMVProcess.UNREACHABLE, None)
           (NuSMVProcess.NON_ZENO, trace_fname): trace_name is a
                                                 filename (string)
           (NuSMVProcess.ZENO, trace_fname): trace_name is a filename (string)
           (NuSMVProcess.UNKNOWN_NOT_FOUND, steps): steps is an integer value
           (NuSMVProcess.UNKNOWN_NOT_PROVED, time): time is a rational
                 value (string) in default time unit used by SmvConverter
        """

        self.build_msat_fsm()

        cmd = ('{cmd_name} -s "{formula}" -t "{time_var}" ' +
               '-k {max_search_depth} -l {max_klive_bound}').format(
            cmd_name="hasdel_is_state_zeno",
            formula=formula,
            time_var=time_var,
            max_search_depth=max_search_depth,
            max_klive_bound=max_klive_bound)

        self.send_cmd(cmd)
        ack_msg = "DONE HYCOMP zeno"
        # This is for detecting no result found
        self.send_cmd("echo %s" % ack_msg)

        grep = utils.process.TextLineGrepper(
            ["The state is (unreachable|ZENO|NON-ZENO)",
             "Cannot (find) a zeno-path in (\d+) steps",
             "Cannot (prove) the absence of zeno paths up-to ([f'\d./]+)",
             ack_msg])
        m, txt = grep.match(self)  # wait for result!

        # processes results
        if m is None or grep.is_match_unwanted(m):
            log.system.error("Process failed on command:\n> " + cmd)
            self._log_returned_text(txt)
            return (self.ERROR, None)

        self._check_handler_errors(txt)

        # processes results
        if m.groups()[0] == "unreachable":
            return (self.UNREACHABLE, None)
        elif m.groups()[0] == "NON-ZENO":
            return (self.NON_ZENO, self._create_trace())
        elif m.groups()[0] == "ZENO":
            return (self.ZENO, self._create_trace())
        elif m.groups()[0] == "find":
            return (self.UNKNOWN_NOT_FOUND, int(m.group(2)))
        elif m.groups()[0] == "prove":
            return (self.UNKNOWN_NOT_PROVED, m.group(2))
        assert False, "Bad state" + txt

    def check_clock_divergence(self, clock_name, clock_bound=None,
                               max_search_depth=options.NUSMV_BMC_LENGTH):

        """Takes clock_name (string), clock_bound (string for rational
        number, or None for no-bound-specification),
        max_search_depth (integer, -1 for no bound).

        Searches for clock diverge and returns one pair of:
          * (NuSMVProcess.BOUNDED, bound)
            Where bound is a string with rational expression (smv syntax)

          * (NuSMVProcess.UNBOUNDED, trace)
            Where trace is a file name containing produced trace (iff
            clock_bound was specified), or None if no clock_bound was
            specified.

          * (NuSMVProcess.UNKNOWN, None)
        """

        self.build_msat_fsm()

        clock_bound = ("-s"
                       if clock_bound is None
                       else '-b "%s"' % clock_bound)

        cmd = ('{cmd_name} -c "{clock_name}" {clock_bound} ' +
               '-k {max_search_depth}').format(
            cmd_name="hasdel_check_clock_divergence",
            clock_name=clock_name,
            clock_bound=clock_bound,
            max_search_depth=max_search_depth)

        self.send_cmd(cmd)
        ack_msg = "DONE HYCOMP divergence"
        # This is for detecting no result found
        self.send_cmd("echo %s" % ack_msg)

        grep = utils.process.TextLineGrepper(
            ["The clock %s is (BOUNDED)" % clock_name,
             "The clock %s is (UNBOUNDED)" % clock_name,
             "The check on the clock divergence for %s is (UNKNOWN)"
             % clock_name,
             ack_msg])

        m, txt = grep.match(self)  # wait for result!

        # processes results
        if m is None or grep.is_match_unwanted(m):
            log.system.error("Process failed on command:\n> " + cmd)
            self._log_returned_text(txt)
            return (self.ERROR, None)

        self._check_handler_errors(txt)
        # processes results
        if m.groups()[0] == "BOUNDED":
            # searches the bound:
            m2 = re.search("The bound is ([\d.f' /*]+)", txt)
            assert m2 is not None, "Error matching bound, text is:\n" + txt
            return (self.BOUNDED, m2.group(1))

        elif m.groups()[0] == "UNBOUNDED":
            # if bound is specified, trace is also created
            trace = clock_bound and self._create_trace()
            return (self.UNBOUNDED,  trace)
        elif m.groups()[0] == "UNKNOWN":
            return (self.UNKNOWN, None)
        assert False, "Bad state" + txt


    # ======================================================================
    # Internals
    # ======================================================================

    def _goto_state_bdd(self, trace_index, step_index):
        """Selects the given trace and step index to be the next
        trace used by the simulation (BDD-based). Creates a new
        trace, and returns the index (integer) of the newly created
        trace."""
        self.build_bdd_fsm()
        self.send_cmd('goto_state "%d.%d"' % (trace_index, step_index))
        self.wait_for_command_finish()
        return self.get_traces_num()

    def _goto_state_bmc(self, trace_index, step_index):
        """Selects the given trace and step index to be the next
        trace used by the simulation (BMC-based). Creates a new
        trace, and returns the index (integer) of the newly created
        trace."""
        self.build_bmc_fsm()
        self.send_cmd('bmc_pick_state -s "%d.%d"' % (trace_index, step_index))
        self.wait_for_command_finish()
        return self.get_traces_num()

    def _goto_state_msat(self, trace_index, step_index):
        """Selects the given trace and step index to be the next
        trace used by the simulation (MSAT-based). Creates a new
        trace, and returns the index (integer) of the newly created
        trace."""
        self.build_msat_fsm()
        self.send_cmd('msat_pick_state -s "%d.%d"' % (trace_index, step_index))
        self.wait_for_command_finish()
        return self.get_traces_num()

    def _pick_state_random_bdd(self, constraints):
        """Creates a 1-step trace for starting a simulation (using
        BDDs). The state is randomly selected taking into account
        the list of constraints (list of strings). Returns the index of
        the created trace, or None if no trace was created."""

        self.build_bdd_fsm()

        grep = utils.process.TextLineGrepper(
            ["PICK_DONE"],
            ["The set of initial states is EMPTY"])

        constr_cmd = ('-c "%s"' % " & ".join(("(%s)" % c for c in constraints))
                      if constraints
                      else "")

        cmd = 'pick_state -r %s; echo "PICK_DONE"' % constr_cmd
        self.send_cmd(cmd)
        m, txt = grep.match(self)
        if m is None:
            log.system.error("Process failed on command:\n> " + cmd)
            self._log_returned_text(txt)
            return None

        if grep.is_match_unwanted(m):
            return None

        return self.get_traces_num()

    def _simulate_random_bdd(self, constraints, length):
        """Constraints is a list of propositional formulae
        (strings). Returns the index of the created trace, or
        None if a problem occurs"""
        self.build_bdd_fsm()

        grep = utils.process.TextLineGrepper(
            ["SIM_DONE"],
            ["No future state exists: trace not built",
             'No current state set. Use the "pick_state" command.'])

        constr_cmd = ('-t "%s"' % " & ".join(("(%s)" % c for c in constraints))
                      if constraints
                      else "")

        cmd = 'simulate -r %s -k %d; echo SIM_DONE' % (constr_cmd, length)
        self.send_cmd(cmd)
        m, txt = grep.match(self)
        if m is None:
            log.system.error("Process failed on command:\n> " + cmd)
            self._log_returned_text(txt)
            return None

        if grep.is_match_unwanted(m):
            return None

        return self.get_traces_num()

    def _pick_state_random_msat(self, constraints):
        """Creates a 1-step trace for starting a simulation (using
        mathsat). The state is randomly selected taking into
        account the list of constraints (strings). Returns the
        index of the the created trace, or None if no trace was
        created."""
        self.build_msat_fsm()

        grep = utils.process.TextLineGrepper(
            ["PICK_DONE"],
            ["No trace: initial state is inconsistent",
             "No trace: constraint and initial state are inconsistent"])

        constr = ('"%s"' % " & ".join(("(%s)" % c for c in constraints))
                  if constraints
                  else "")

        cmd = ('msat_pick_state %s; echo "PICK_DONE"' %
               (("-c " + constr)
                if constr
                else ""))

        self.send_cmd(cmd)
        m, txt = grep.match(self)
        if m is None:
            log.system.error("Process failed on command:\n> " + cmd)
            self._log_returned_text(txt)
            return None

        if grep.is_match_unwanted(m):
            return None

        return self.get_traces_num()

    def _simulate_random_msat(self, constraints, length):
        """Constraints is a list of propositional formulae
        (strings). Returns the index of the created trace, or None
        if a problem occurs"""
        self.build_msat_fsm()

        grep = utils.process.TextLineGrepper(
            ["SIM_DONE"],
            ["No trace: constraint and initial state are inconsistent",
             "Simulation is UNSAT"])

        constr = ('"%s"' % " & ".join(("(%s)" % c for c in constraints))
                  if constraints
                  else "")

        cmd = ('msat_simulate %s -k %d; echo SIM_DONE' %
               (("-t " + constr) if constr else "", length))
        self.send_cmd(cmd)
        m, txt = grep.match(self)
        if m is None:
            log.system.error("Process failed on command:\n> " + cmd)
            self._log_returned_text(txt)
            return None

        if grep.is_match_unwanted(m):
            return None

        return self.get_traces_num()

    def _simulate_check_feasability_bmc(self, constraints):
        """Constraints is a list of next formulae
        (strings). Returns a list of boolean values telling
        positionally which constraints are true or false."""

        res = []
        if constraints:
            self.build_bmc_fsm()

            msg = "SIM_FEAS:"
            constr_cmd = " ".join(('-c "%s"' % c for c in constraints))

            grep = utils.process.TextLineGrepper("%s\s*([01]+)" % msg)
            cmd = (
                'echo -n %s; bmc_simulate_check_feasible_constraints -q %s' %
                (msg, constr_cmd))
            self.send_cmd(cmd)
            m, txt = grep.match(self)
            if m is None:
                log.system.error("Process failed on command:\n> " + cmd)
                self._log_returned_text(txt)
                return None

            res = [True if c == "1" else False for c in m.group(1)]

        return res

    def _simulate_check_feasability_msat(self, constraints):
        """Constraints is a list of next formulae
        (strings). Returns a list of boolean values telling
        positionally which constraints are true or false."""

        res = []
        if constraints:
            self.build_msat_fsm()

            msg = "SIM_FEAS:"
            constr_cmd = " ".join(('-c "%s"' % c for c in constraints))

            grep = utils.process.TextLineGrepper("([^0-9]*)([01]+)")
            cmd = (
                'echo -n %s; msat_simulate_check_feasible_constraints -q %s' %
                (msg, constr_cmd))
            self.send_cmd(cmd)
            m, txt = grep.match(self)
            if m is None:
                log.system.error("Process failed on command:\n> " + cmd)
                self._log_returned_text(txt)
                return None

            res = [True if c == "1" else False for c in m.group(0)]

        return res

    def _create_trace(self):
        """Invokes Nusmv to create a xml trace file out of latest
        created trace. Returns the filename containing the trace.
        Raises a utils.exceptions.TraceException if an
        error occurs. Returns None if the trace was not generated
        due to an unknown reason."""
        # searches for a good name for trace
        basename = os.path.splitext(self._smv_fname)[0]
        while True:
            self.__trace_counter += 1
            fname = "%s_trace%d.xml" % (basename, self.__trace_counter)
            if not os.path.isfile(fname):
                break

        # creates the trace fname
        cmd = 'show_traces -p 4 -o "%s"' % fname
        self.send_cmd(cmd)
        self.send_cmd('echo "TRACE_DONE"')

        # wait for the file to be generated:
        grep = utils.process.TextLineGrepper("TRACE_DONE", "Division by zero")
        m, txt = grep.match(self)
        if m is None:
            log.system.error("Process failed on command:\n> " + cmd)
            self._log_returned_text(txt)
            return None

        if grep.is_match_unwanted(m):
            raise TraceException(m.string)

        res = utils.process.wait_for_condition(os.path.isfile, [fname], 5)
        assert res, "A trace file was expected to be generated"

        return fname

    def _check_pattern_aux(self, pattern_name, specs,
                           time_bound, time_var,
                           max_search_depth, max_klive_bound):
        """Service for all time property checking functions, as all of them
        share a common base code. specs is a list of one or two
        formula strings
        """

        self.build_msat_fsm()

        cmd_map = {"absence": "hasdel_check_pattern_absence",
                   "existence": "hasdel_check_pattern_existence",
                   "universality": "hasdel_check_pattern_universality",
                   "precedence": "hasdel_check_pattern_precedence",
                   "responseGlobal": "hasdel_check_pattern_response_global"}

        assert pattern_name in cmd_map, "Unsupported pattern: " + pattern_name

        formulas = ['%s "%s"' % p
                    for p in zip(("-p", "-q"), specs)]

        cmd = ('{cmd_name} {formulas} -T "{time_bound}" -t "{time_var}" ' +
               '-k {max_search_depth} -l {max_klive_bound}').format(
            cmd_name=cmd_map[pattern_name],
            formulas=" ".join(formulas),
            time_bound=time_bound,
            time_var=time_var,
            max_search_depth=options.NUSMV_BMC_LENGTH,
            max_klive_bound=options.NUSMV_KLIVE_BOUND)

        self.send_cmd(cmd)
        ack_msg = "DONE HYCOMP " + pattern_name
        # This is for detecting no result found
        self.send_cmd("echo %s" % ack_msg)

        grep = utils.process.TextLineGrepper(
            [".*\s+The property is (false|true|unknown)",
             ack_msg])
        m, txt = grep.match(self)  # wait for result!

        # processes results
        if m is None or grep.is_match_unwanted(m):
            log.system.error("Process failed on command:\n> " + cmd)
            self._log_returned_text(txt)
            return (None, None)

        self._check_handler_errors(txt)

        # processes results
        if m.groups()[0] == "true":
            return (True, None)
        elif m.groups()[0] == "unknown":
            return (None, None)
        assert m.groups()[0] == "false"
        return (False, self._create_trace())

    def _log_returned_text(self, txt):
        if txt.strip() and txt.strip() != "esmc >":
            log.system.error("Returned text was:\n" +
                             "-" * 70 + "\n" +
                             txt.strip() +
                             "\n" + "-" * 70 + "\n")
        else:
            log.system.error("No output is available: process may "
                             "have badly crashed\n")

    pass  # end of class
