# =============================================================================
#  This file is part of the ``HASDEL'' Toolset.
#  Copyright (C) 2014 by Fondazione Bruno Kessler
#  Copyright (C) 2014 by RWTH Aachen University
#
#
#  The HASDEL Toolset can be copied, modified and/or redistributed
#  under the terms of the HASDEL General Public License, either
#  version 1 of the License, or (at your option) any later version.
#
#  You should have received a copy of the HASDEL General Public
#  License along with this software. If not, please contact the
#  copyright owners.
#
#  The HASDEL Toolset is distributed in the hope that it will be
#  useful, but WITHOUT ANY WARRANTY; without even the implied warranty
#  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  HASDEL General Public License for more details.
#
#  For more information on the HASDEL Project see the Project Web
#  site at <http://compassinternal.informatik.rwth-aachen.de/hasdel-site/> 
#  or email to <hasdel-support@lists.rwth-aachen.de>.  
#
#  Please report bugs to <hasdel-support@lists.rwth-aachen.de>.
# =============================================================================

# this is required in 2.5, not in >= 2.6
from __future__ import with_statement

import subprocess
import threading
import time
import signal
import re
import os
import os.path
import sys
import select
import Queue
import platform

import options
from utils import log


# ----------------------------------------------------------------------
class KillablePipe (subprocess.Popen):

    def kill(self):
        """This is not provided in 2.5, so providing it here"""
        if not self.is_alive():
            return  # already terminated

        # tries terminating the process kindly at first
        self.stdin.close()
        for i in range(0, 3):
            time.sleep(0.1)
            if not self.is_alive():
                return

        if subprocess.mswindows:
            _terminate_process_win32(self)
        else:
            os.kill(self.pid, signal.SIGKILL)
        return

    def is_alive(self):
        """Returns True iff the process is still running"""
        return self.poll() is None

    pass  # end of class


# ----------------------------------------------------------------------
class InteractiveProcess (KillablePipe):
    """Use this class to launch and control interactive programs
    like nusmv. For files, see subprocess.Popen docs.
    The output/error text use functions get_stdout_text and get_stderr_text."""

    def __init__(self, path, args=[], stdout=None, stderr=None,
                 stdin=subprocess.PIPE, dumpers=[]):
        """@param dumpers is a list of file objects (or objects
        supporting method write, used to send process stdout and
        stderr to"""
        self.path = path
        self.name = os.path.basename(path)
        log.system.debug("Running: "+" ".join([self.path] + args))

        super(InteractiveProcess, self).__init__([self.path] + args,
                                                 stdin=stdin,
                                                 stdout=stdout, stderr=stderr)
        # makes non-blocking streams
        #import fcntl
        #if self.stdout: fcntl.fcntl(self.stdout, fcntl.F_SETFL, os.O_NONBLOCK)
        #if self.stderr: fcntl.fcntl(self.stderr, fcntl.F_SETFL, os.O_NONBLOCK)
        #time.sleep(0.5)

        #create a wrapper for log.system.debug stream
        class LoggingWrapper:
            def __init__(self, prefix):
                self._prefix = prefix

            def write(self, text):
                if text and text[-1] == '\n':
                    text = text[:-1]  # remove last "\n"
                log.system.debug(self._prefix + text)

        out_log_wrapper = LoggingWrapper(self.name + " out : ")
        err_log_wrapper = LoggingWrapper(self.name + " err : ")

        self.out_tee = (ProcessTee(self.stdout, dumpers + [out_log_wrapper])
                        if self.stdout is not None
                        else None)

        self.err_tee = (ProcessTee(self.stderr, dumpers + [err_log_wrapper])
                        if self.stderr is not None
                        else None)
        return

    def get_path(self):
        return self.path

    def send_cmd(self, cmd):
        if not self.is_alive():
            return

        log.system.debug('Sending command "%s" to process "%s"' %
                         (cmd, self.name))

        # here we hide a problem in sending a command.
        # it is responsibility of the reader of the command results to detect
        # the problem.
        try:
            self.stdin.write(cmd + "\n")
        except IOError as e:
            if self.is_alive():
                log.system.error("IOError with %s: %s" % (self.name, str(e)))

        except ValueError as e:
            if self.is_alive():
                log.system.error("IOError with %s: %s" % (self.name, str(e)))

    def kill(self):
        """This is not provided in 2.5, so providing it here"""
        #self.flush_stderr_stdout() # flush the streams just in case
        if not self.is_alive():
            return  # already terminated

        # quits the tee if needed
        if self.out_tee:
            self.out_tee.quit_now()
        if self.err_tee:
            self.err_tee.quit_now()

        # tries terminating the process kindly at first
        if self.stdin:
            self.stdin.close()

        for i in range(0, 3):
            time.sleep(0.1)
            if not self.is_alive():
                return

        if subprocess.mswindows:
            _terminate_process_win32(self)
        else:
            os.kill(self.pid, signal.SIGKILL)
        return

    def is_alive(self):
        """Returns True iff the process is still running"""
        return self.poll() is None

    def get_stdout_text(self):
        """Returns the text collected in output stream so far"""
        if self.stdout is None:
            return ""
        return self._get_file_output(self.out_tee)

    def get_stderr_text(self):
        """Returns the text collected in error stream so far"""
        if self.stderr is None:
            return ""
        return self._get_file_output(self.err_tee)

    def flush_stderr_stdout(self):
        """Flush all the text collected in the output and error stream
        so far"""
        self.get_stdout_text()
        self.get_stderr_text()

    def _get_file_output(self, _tee):
        res = _tee.read()
        return res

    pass  # end of class


# ----------------------------------------------------------------------
class ProcessTee (threading.Thread):
    """This class is a pipe tee, that can read from an input file, and
    sends the output to a set of files (or objects supporting method
    write). The class supports 'readline' and 'read' methods such that the
    class can be used as a pipe. After reading the input and before
    writing to the output, data can be transformed optionally by a
    function. This class is used in processes (like external
    executables) to log the output, to send the output to a GUI, etc.

    Important: before exiting the program, it is important to
    kill the process originating the input for this class,
    otherwise the execution will stuck hang."""

    def __init__(self, input_file, out_files,
                 trans_func=None, trans_func_arg=None):
        """
        @param input_file: a file object exposing 'readline' method.
                           Data will be read from this object.
        @param out_files: a list of file objects or objects
             exposing method 'write'. Data will be written to these objects.
        @param trans_func: an optional function taking (data, arg),
            transforming the data and returning it. arg is specified
            with parameter trans_func_arg """

        super(ProcessTee, self).__init__()
        self._input = input_file
        self._out_files = out_files

        # Queue is thread-safe. 1 item in the queue is 1 line from input
        self._queue = Queue.Queue()

        self._trans_func = trans_func
        self._trans_func_arg = trans_func_arg

        self._quit_now = False
        # starts a thread
        self.start()
        return

    def readline(self):
        """Returns one line of input text, and empty line if there is no input.
        This function does not block the execution."""
        #wait a bit in case of some delay
        try:
            line = self._queue.get(block=True, timeout=0.1)
        except Queue.Empty:
            return ""

        return line

    def read(self):
        """Returns all text collected from input so far.
        This function does not block the execution"""
        res = ""
        try:
            while True:
                #wait a bit in case of some delay
                line = self._queue.get(block=True, timeout=0.1)
                res += line

        except Queue.Empty:
            return res
        assert False, "impossible code"

    def _run(self):
        """Returns True if ok, False if a problem occurs"""
        try:
            _line = self._input.readline()
        except IOError as e:
            # in case of a input problem return nothing.
            # the higher level code will detect problem
            log.system.debug(
                "Problem with input stream in ProcessTee" + str(e))
            return False  # stop iteration

        if self._quit_now:
            return False

        # apply the conversion function if any
        if self._trans_func:
            _line = self._trans_func(_line, self._trans_func_arg)

        # write the result to the queue and files (wait if queue if full)
        self._queue.put(_line, block=True)
        for _file in self._out_files:
            _file.write(_line)
        return True

    def quit_now(self):
        if not self.isAlive():
            return
        self._quit_now = True

    def have_to_quit(self):
        return not self.isAlive() or self._quit_now

    def run(self):
        if platform.system() == "Windows":
            keep_on = True
            while (not self.have_to_quit() and keep_on and
                   not self._input.closed and
                   self._input.fileno() >= 0):
                keep_on = self._run()
            return

        poll = select.poll()
        poll.register(self._input)
        keep_on = True
        while keep_on and not self.have_to_quit():
            polls = poll.poll()
            if [pr for pr in polls if pr[1] & select.POLLIN]:
                # read from the input given to the Tee
                keep_on = self._run()
                # exit if there is no read but process terminated or
                # has problems

            elif ([pr
                   for pr in polls
                   if pr[1] & (select.POLLHUP |
                               select.POLLERR |
                               select.POLLNVAL)] or
                  self._input.closed):
                break

            else:
                assert False, "impossible code"

        if not self._input.closed and self._input.fileno() >= 0:
            poll.unregister(self._input)
        return

    pass  # end of class


class InteractiveStreamReader (threading.Thread):
    """Use this class to read interactively from a file which is
    written asynchronously. This can be used within a GUI"""
    def __init__(self, _file, buffer=""):
        super(InteractiveStreamReader, self).__init__()
        self._buf = buffer
        self.file = _file
        self.lock = threading.Lock()
        self.start()
        return

    def _read(self):
        try:
            line = self.file.readline()
        except:
            return ""
        return line

    def run(self):
        line = self._read()
        while line:
            with self.lock:
                self._write(line)
            line = self._read()

    def pop_text(self):
        """Returns what was read in buffer, and reset the buffer to be empty"""
        with self.lock:
            res = self._get_text()
            self._reset()
        return res

    def _write(self, line):
        """This function is used to append into the buffer"""
        self._buf += line
        return

    def _reset(self):
        """This function is used when resetting the buffer"""
        self._buf = ""
        return

    def _get_text(self):
        """This function is used to retrieve the text out of the buffer"""
        return self._buf

    pass


def wait_for_condition(predicate, args=[], timeout=-1):
    """Waits for a condition to occur. Returns True if condition is
    satisfied, False if timeout expired"""
    delta = 0.25
    times = int(timeout / delta) if timeout >= 0 else 0
    while times >= 0:
        if predicate(*args):
            return True
        time.sleep(delta)
        if timeout >= 0:
            times -= 1

    return False


# ----------------------------------------------------------------------
class TextLineGrepper (object):
    """This class can be used for searching regex matching within
    the standard output or error of a running process that is
    running. It is possible to specify patterns into to different
    sets. One set is for wanted (desidered) matches, and the other
    is for unwanted matches. For example when searching for errors,
    unwanted list can be used. The two sets are basically handled
    in the same manner, but after a match is found it is possible
    to ask at which set the matching belongs to. It is up to the
    user to provide the right semantics to the sets."""

    def __init__(self, wanted_line_re, unwanted_line_re=[]):
        """Each of wanted_line_re and unwanted_line_re can be a
        list of strings, or one single string. Each string is a
        regular expression for matching"""
        if type(wanted_line_re) in (tuple, list):
            self.wanted_line_res = map(re.compile, wanted_line_re)
        else:
            self.wanted_line_res = [re.compile(wanted_line_re)]

        if type(unwanted_line_re) in (tuple, list):
            self.unwanted_line_res = map(re.compile, unwanted_line_re)
        else:
            self.unwanted_line_res = [re.compile(unwanted_line_re)]

    def add_wanted_regex(self, line_re):
        """Add a new string (line) to the list of regular
        expressions for matching (wanted set)"""
        self.wanted_line_res.append(re.compile(line_re))
        return

    def add_unwanted_regex(self, line_re):
        """Add a new string (line) to the list of regular
        expressions for matching (unwanted set)"""
        self.unwanted_line_res.append(re.compile(line_re))
        return

    def is_match_wanted(self, match):
        "Returns true if the match is in the wanted patterns"""
        return match.re in self.wanted_line_res

    def is_match_unwanted(self, match):
        "Returns true if the match is in the wanted patterns"""
        return match.re in self.unwanted_line_res

    def match(self, proc, timeout=-1, in_stdout=True):
        """Try to match each line of the stream with the given matches.
        Returns as soon as a first match is found or the process
        terminates/is killed.
        @param proc: a process the output (if in_stdout==True)
                     or error (if in_stdout==False) stream is checked.
        @type proc: InteractiveProcess

        Returns a tuple (match, text) where 'match' is the match found
        (and None otherwise) and 'text' is the total text processed by this
        function call.
        """
        delta = 0.25
        getter = proc.get_stdout_text if in_stdout else proc.get_stderr_text
        times = int(timeout / delta) if timeout >= 0 else 0
        res_text = ""
        while times >= 0:
            # check for aliveness before reading text!
            is_alive = proc.is_alive()
            new_text = getter()
            res_text += new_text

            for _line in new_text.splitlines():
                stripped = _line.rstrip()
                for line_re in self.wanted_line_res + self.unwanted_line_res:
                    m = line_re.search(stripped)
                    if m:
                        return (m, res_text)

            if not is_alive:
                break
            time.sleep(delta)
            if timeout >= 0:
                times -= 1

        return (None, res_text)

    pass  # end of class
# ----------------------------------------------------------------------


def get_exec_name(name):
    """Given the name of an executable like for example "NuSMV",
    returns the complete name that corresponds to the executable to
    be invoked in the detected platform, for example
    "bin/NuSMV-linux32". Raises IOError if no valid executable is
    found. If 64 bit executable is not found when requested, and a
    corresponding version for 32 is found instead, the 32 bit
    version is returned. This function is also robust in those case
    where the full path to an existing executable is passed, or
    relative to BIN_DIR variable."""

    # check if already provided a complete name
    if os.path.isfile(name):
        return name
    if os.path.isfile(os.path.join(options.BIN_DIR, name)):
        return name

    bits = int(platform.architecture()[0][0:-3])
    try:
        system, exeext = {'Linux': ('linux', '',),
                          'Windows': ('win', '.exe'),
                          'Darwin': ('mac', '')}[platform.system()]
    except KeyError:
        log.system.error("System was not recognized: %s", platform.system())
        quit(1)

    # in case the executable is the ftd, we have some issues depending on the
    # machine we are working on (static version does not work on some
    # RedHat local machines

    if name == options.FTD:
        distrib_id, distrib_release, distrib_codename = \
            platform.linux_distribution()
        if distrib_id == "Red Hat Enterprise Linux Workstation":
            # we are on our FBK local machines
            assert system == "linux"
            path = os.path.join(options.BIN_DIR, "%s-%s%d" %
                                (name, system, 32))
            return path
        else:
            # try to use the static executables
            assert system == "linux"
            path = os.path.join(options.BIN_DIR, "%s-%s%d" %
                                (name, system, 32)) # we must always call the 32-bits executable as it is more stable
            return path

    # now searches if it is available
    path = os.path.join(options.BIN_DIR, "%s-%s%d%s" %
                        (name, system, bits, exeext))
    if os.path.isfile(path):
        return path

    # if 64 bit not found, try with 32:
    if bits == 64:
        path = os.path.join(options.BIN_DIR, "%s-%s%d" % (name, system, 32))
        if os.path.isfile(path):
            log.system.warning("Using 32 bit version for executable %s, "
                               "even if bits are %d", name, bits)
            return path

    # no lucky: raise an exception
    raise IOError("Not found a valid executable for %s, "
                  "platform is %s, bits are %d" %
                  (name, system, bits))


def start_process_and_get_results(args, input, allow_output, kill_event=None):
    """Runs an executable with the provided arguments (specified in "args")
    and gives it in input stream text from "input".
    Flag allow_output makes all the output (normal and error)
    to be passed to standard output/error. Otherwise, it is returned
    as a pair (output-text, error-text).
    """
    log.system.debug("Invoking: %s\n with the following input: %s" %
                     (" ".join(args), input))

    # this private class and the following function are used to
    # handle killing the process from the outside set event.
    def __kill_event_handler(process_queue, event):
        assert process_queue.qsize() <= 1
        event.wait()

        # here an event has been received
        try:
            process = process_queue.get_nowait()
            process.kill()
            process_queue.task_done()
            pass

        except Queue.Empty:
            pass  # end of job
        return  # end of event handling

    process = KillablePipe(args,
                           stdin=subprocess.PIPE,
                           stdout=None if allow_output else subprocess.PIPE,
                           stderr=None if allow_output else subprocess.PIPE)

    queue = Queue.Queue(1)
    queue.put(process)

    ev_handler = None
    if kill_event:
        ev_handler = threading.Thread(target=__kill_event_handler,
                                      args=(queue, kill_event))
        ev_handler.start()

    try:
        stdout_res, stderr_res = process.communicate(input)
    except OSError as e:
        log.system.error("Error in running external application:\n %s\n%s" %
                         (" ".join(args), e))
        raise e

    finally:
        # if the monitor is still running, frees it
        if ev_handler and ev_handler.isAlive() and not kill_event.isSet():
            try:
                process = queue.get_nowait()
                kill_event.set()
                kill_event.clear()
            except Queue.Empty:
                pass

    if process.returncode not in (0, -9):
        _process_error(process, args, stdout_res, stderr_res)

    return (stdout_res, stderr_res)


# ----------------------------------------------------------------------
#    Internal functions
# ----------------------------------------------------------------------

def _process_error(process, args, stdout, stderr):
    if stdout or stderr:
        log.system.debug("Error:")
        if stdout:
            log.system.debug("stdout:------------------")
            log.system.debug(str(stdout))
        if stderr:
            log.system.debug("stderr:------------------")
            log.system.debug(str(stderr))
    raise subprocess.CalledProcessError(process.returncode, args)


def _terminate_process_win32(pipe):
    """This internal service can be used under win32 only."""
    assert subprocess.mswindows
    try:
        from subprocess import TerminateProcess
    except:
        try:
            from win32api import TerminateProcess
        except:
            try:
                from ctypes import TerminateProcess
            except:
                # last choice via a windows command
                if os.popen('TASKKILL /PID %d /F' % pipe.pid):
                    import warnings
                    msg = (" or install Python version >= 2.5.0"
                           if sys.version_info < (2, 5, 0)
                           else "")
                    warnings.warn("It was not possible to kill the process. "
                                  "Install module 'win32api'%s." % msg,
                                  RuntimeWarning)
                return

    TerminateProcess(pipe._handle, -1)
    return
